/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.Conductor;

import java.io.*;
import java.util.*;
import java.net.*;

import FMG.Conductor.dataflow.IOManager;
import FMG.Conductor.util.*;
import FMG.Conductor.security.SecurityCoordinator;
import FMG.Conductor.monitor.EnvMonitor;
import FMG.Conductor.dataflow.SessionTable;
import FMG.Conductor.comm.*;
import FMG.Conductor.planning.*;
import FMG.Conductor.dataflow.comm.*;
import FMG.Conductor.iface.Interface;

/**
 * The Conductor class is the main class for the Conductor system.
 * It contains code for
 * <ul>
 *    <li>the main() function
 *    <li>argument handling
 *    <li>property handling
 *    <li>initialization and startup
 * </ul>
 * Much of the initialization and startup occurs in the main thread.  Once
 * this is accomplished, the interface is started up.  The interface module
 * may or may not continue to execute in that thread.
 *
 * @author Mark Yarvis (yarvis@fmg.cs.ucla.edu)
 * @version $Revision: 1.49 $
 */

public class Conductor {
   /** The default name for the properties file. */
   private static final String DEFAULT_PROPERTIES_FILE = 
                                          "Conductor.properties";

   /** The list of properties read from the properties file */
   private static Properties props;

   /**
    * Is reliability enabled?  This is determined by a property, but since
    * it's a widely used property, the main code reads the property
    * and sets the variable.  It is assessible through an accessor method.
    */
   private static boolean reliabilityEnabled;

   /** The name of this host. */
   private static String hostname = null;

   /**
    * This is the first function executed when Conductor is started.
    * <p>
    * TODO: is there any real reason to create an instance of Conductor?
    */
   public static void main(String[] args) {
      (new Conductor()).go(args);
   }

   /**
    * Returns the boolean value of a property from the Conductor property
    * file.
    * @param name The name of the property to test.
    * @param defaultValue The default value to return if the property is 
    *                     not defined.
    * @return If the property is defined, returns <code>true</code> iff the 
    *         property value is "true", "1", or "yes" (case insensitive).
    *         If the property is not defined, returns the value passed
    *         as defaultValue.
    */
   public static boolean getBooleanProperty(String name, 
                                            boolean defaultValue) {
      if (props == null) {
         return defaultValue;
      }

      String value = props.getProperty(name);
      if (value == null) {
         return defaultValue;
      } else {
         return (value.equalsIgnoreCase("true") || 
                 value.equals("1") || 
                 value.equalsIgnoreCase("yes"));
      }
   }

   /**
    * Returns the integer value of a property from the Conductor property
    * file.
    * @param name The name of the property to fetch.
    * @param defaultValue The default value to return if the property is 
    *                     not defined.
    * @return If the property is defined, it's integer value is returned.
    *         If the property is not defined, the value passed
    *         as defaultValue is returned.
    */
   public static int getIntProperty(String name, int defaultValue) {
      if (props == null) {
         return defaultValue;
      }

      String value = props.getProperty(name);
      if (value == null) {
         return defaultValue;
      } else {
         return Integer.parseInt(value);
      }
   }

   /**
    * Obtain the requested property which represents a filename.  If the
    * result contains ${HOSTNAME}, this substring is replaced with the
    * hostname of the local host prior to returning the result.
    * @param name The name of the desired property
    * @param defaultValue The default value to use if the property was not 
    *                     specified.
    * @return The value of the requested property, or the default value.
    */
   public static String getFilenameProperty(String name, String defaultValue) {
      if (hostname == null) {
         try {
            hostname = InetAddress.getLocalHost().getHostName();
         } catch (UnknownHostException e) {
            throw new RuntimeException("Unexpected error: " + e);
         }
      }

      return StringOps.replace(getProperty(name, defaultValue), 
                               "${HOSTNAME}", hostname);
   }

   /**
    * Returns the string value of a property from the Conductor property
    * file.
    * @param name The name of the property to fetch.
    * @return If the property is defined, it's value is returned.  If the
    *         property is not defined, null is returned.
    */
   public static String getProperty(String name) {
      if (props == null) {
         return null;
      }
      return props.getProperty(name);
   }

   /**
    * Returns the string value of a property from the Conductor property
    * file.
    * @param name The name of the property to fetch.
    * @param defaultValue The default value to return if the property is 
    *                     not defined.
    * @return If the property is defined, it's value is returned.
    *         If the property is not defined, the value passed
    *         as defaultValue is returned.
    */
   public static String getProperty(String name, String defaultValue) {
      if (props == null) {
         return null;
      }
      return props.getProperty(name, defaultValue);
   }

   /**
    * Indicates whether the boolean property "ReliabilityEnabled" was set.
    * Since so many different parts of the code depend on this property,
    * it is checked once in this class and is then accessible using this 
    * method.
    * @return True iff the property "ReliabilityEnabled" was true according
    *              to the <code>getBooleanProperty()</code> method.
    */
   public static boolean reliabilityEnabled() {
      return reliabilityEnabled;
   }

   /**
    * Reads the properties file into the <code>props</code> variable.
    * If the file is not found or is inaccessible, no properties are loaded
    * and no warning occurs.
    * @param filename The name of the file to read the properties from.
    * @return No return value.
    */
   private static void readPropertiesFile(String filename) {
      /* if the user specified a null filename, don't read any properties */
      if (filename.length() == 0) {
         props = new Properties();
         return;
      }

      /* read the properties file */
      props = PropsLoader.readProps(filename, true);
      if (props == null) {
         props = new Properties();
         // Verbosity is not set yet, so just go ahead and warn the user
         System.err.println("Warning: error reading properties file");
      }
   }

   /**
    * Update the value of a property using the given string.  The string
    * should either be the name of a property or in the form "propname=value".
    * If only the name of a property is given (ie, no equals sign), then
    * the value is set to "true".
    * @param propString A string containing a property name or both a 
    *                   property name and value.
    */
   private static void updateProperty(String propString) {
      int index = propString.indexOf('=');

      if (index == -1) {
         props.put(propString, "true");
      } else {
         String property = propString.substring(0, index);
         String value = propString.substring(index + 1, propString.length());
         props.put(property, value);
      }
   }

   /**
    * Print out the command-line usage for Conductor.  Additional options
    * may be available through the script that launches Conductor.
    * @return No return value.
    */
   private void usage() {
      System.out.println("Usage:  Conductor [-v] [-m] [-E] [-?] " + 
                         "[-V<module>=<level>]" +
                         "[-L<local-address>] " +
                         "[-P<propsfile>] [-D<propanme>=<value>] [<port> ...]");
      System.out.println("");
      System.out.println("Notes:");
      System.out.println("     <port> is a port number to intercept.");
      System.out.println("           - only root can perform interception");
      System.out.println("           - by number or service name");
      System.out.println("     -v: verbose mode");
      System.out.println("           - more v's give more verbosity");
      System.out.println("     -V: set the verbosity of a module to the " + 
                                                              "given level");
      System.out.println("           - overrides the global verbosity level " +
                                                        "for that module");
      System.out.println("     -m: measurement mode");
      System.out.println("     -?: get this message");
      System.out.println("     -P: set the name of the properties file");
      System.out.println("           - for no properties file, use -P alone");
      System.out.println("     -D: set a given property to the given value");
      System.out.println("           - overrides properties file settings");
      System.out.println("     -L: set Conductor's notion of the local " + 
                                  "address");
      System.out.println("     -E: consider this node an edge node");
   }

   /**
    * Determine the name of the file containing the application properties.
    * If a filename is specified as an option on the command line, it will be
    * returned, otherwise, the default file name will be returned.
    * @param args The arguments passed to the application at startup.
    * @return The name of the file containing the application properties
    */
   private String getPropsFilename(String[] args) {
      String propsFilename = DEFAULT_PROPERTIES_FILE;

      for (int i=0; i < args.length; i++) {
         if (args[i].startsWith("-P")) {
            propsFilename = args[i].substring(2,args[i].length());
         }
      }

      return propsFilename;
   }

   /**
    * Given the command line arguments, initialize the global data 
    * structures, and start the required daemon threads.  In particular,
    * the properites file is read, the verbosity and measurement levels
    * are set, and threads are started to handle interception and planning
    * services.
    * @param args The command line arguments given to Conductor.
    * @return No return value.
    */
   private void initialize(String[] args) {
      int debugLevel = 0;
      boolean measure = false;
      boolean isEdgeNode = false;

      readPropertiesFile(getPropsFilename(args));

      /* look for options */
      for (int i = 0; i < args.length ; i++) {
         if (! args[i].startsWith("-")) {   // its an not an option, skip it
            continue;
         }

         if (args[i].startsWith("-D")) {
            updateProperty(args[i].substring(2, args[i].length()));

         } else if (args[i].startsWith("-P")) {
            // ignore this option, it was handled above

         } else if (args[i].startsWith("-L")) {
            String addr = args[i].substring(2, args[i].length());
            try {
               NodeID.setLocalID(InetAddress.getByName(addr));
            } catch (UnknownHostException e) {
               throw new RuntimeException("Unknown local address: " + addr);
            }

         } else if (args[i].startsWith("-V")) {
            int equalsPos = args[i].indexOf('=');
            if (equalsPos == -1) {
               System.err.println("Missing '=' in option: " + args[i]);
               System.err.println("");
               usage();
               System.exit(1);
            }

            String module = args[i].substring(2, equalsPos);
            String strLevel = args[i].substring(equalsPos+1, args[i].length());

            if (Debug.on) {
               Debug.setLevel(module, Integer.parseInt(strLevel));
            } else {
               System.out.println("Debugging is not enabled, " + 
                                  "verbosity not set!");
            }
         } else {
            char[] buf = new char[args[i].length()-1];
            args[i].getChars(1, args[i].length(), buf, 0);

            for (int j=0; j<buf.length; j++) {
               switch (buf[j]) {
                  case 'v': 
                     debugLevel++;
                     break;
                  case 'm':
                     measure = true;
                     break;
                  case 'E':
                     isEdgeNode = true;
                     break;
                  case '?':
                     usage();
                     System.exit(0);
                  default:
                     System.err.println("Invalid option: " + buf[j]);
                     System.err.println("");
                     usage();
                     System.exit(1);
               }
            }
         }
      }

      reliabilityEnabled = getBooleanProperty("ReliabilityEnabled", true);

      if (debugLevel > 0) {
         if (Debug.on) {
            Debug.setLevel(debugLevel);
            Debug.println("Verbosity level: " + debugLevel);
         } else {
            System.out.println("Debugging not enabled, verbosity not set!");
         }
      }

      if (measure) {
         if (Debug.on) {
            Debug.println("*******************************************");
            Debug.println("WARNING!!! WARNING!!! WARNING!!! WARNING!!!");
            Debug.println("Do not take measurements with Debug mode on.");
            Debug.println("Turn debugging off in FMG.Conductor.util.Debug");
            Debug.println("and recompile!!!");
            Debug.println("*******************************************");
         } else {
            Debug.println("Measurement mode is ON");
         }
         Measure.turnOn();
      }

      /* Create the session table */
      SessionTable sessionTable = new SessionTable();

      /* Initialize main components */
      Path.initialize(isEdgeNode);
      IOManager.initialize(sessionTable);
      SecurityCoordinator.initialize();

      /* Get ready to handle planning paths */
      PlanCommMgr pcm = new PlanCommMgr(sessionTable);
      Path.setPlanCommMgr(pcm);

      /* Get ready to monitor conditions */
      EnvMonitor envMonitor = EnvMonitor.createMonitor();
      ConductorIO.setEnvMonitor(envMonitor);
      Path.setEnvMonitor(envMonitor);

      /* Start accepting streams from other Conductor nodes */
      ConductorServer.createAndStart();

      /* If the user specified one or more ports, start intercepting */
      for (int i = 0; i < args.length ; i++) {
         if (args[i].startsWith("-")) {   // it's an option, skip it
            continue;
         }

         int port = ExternalInfo.getServicePort(args[i]);
         if (port != -1) {
            Interceptor.createAndStart(port, sessionTable);
         } else {
            Debug.println(Debug.WARNING, "WARNING: unknown port " + args[i] +
                                         ", not intercepting");
         }
      }
   }

   /**
    * Start conductor running.  Calls <code>initialize()</code> to get
    * everything ready and then calls the Interface code to handle any
    * command-line or graphics interface that might have been requested.
    * @param args The command line arguments given to Conductor.
    * @return No return value.
    * @see FMG.Conductor.iface.Interface#createAndStart
    */
   void go(String[] args) {
      try {
         initialize(args);
      } catch (Throwable e) {
         e.printStackTrace();
         System.exit(1);
      }

      if (Debug.on) {
         Debug.println("Ready for incoming connections");
      }

      Interface.createAndStart();
   }
}
