/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.Conductor.comm;

import java.io.*;

import FMG.Conductor.net.SocketOps;

/**
 * The BasicMessageInputStream is an implementation of MessageInputStream
 * that can handle a dynamic set of message types.  This class should be 
 * used in conjunction with BasicMessageOutputStream.
 * <p>
 * Before reading or writing a message, the message must be registered
 * using either form of <code>registerMessagePrototype()</code>.  These 
 * calls register a prototype with the BasicMessageInputStream and establish 
 * (either statically or dynamically) an id number for that message type.  
 * The prototype must correctly implement <code>readNewMessage()</code>, as 
 * described in the Message class, to properly support instantiation of 
 * incoming messages.  Each instance of that message type must also return 
 * the assigned message id whenever <code>getMessageID()</code> is called, to
 * support proper identification during output.
 * <p>
 * TODO: should this still be abstract?
 *
 * @author Mark Yarvis (yarvis@fmg.cs.ucla.edu)
 * @version $Revision: 1.14 $
 * @see FMG.Conductor.comm.MessageInputStream
 * @see FMG.Conductor.comm.BasicMessageOutputStream
 */

public abstract class BasicMessageInputStream extends MessageInputStream {

   /** The total number of message types supported */
   private static final int MAX_MESSAGE_TYPES = 200;

   /** 
    * The total number of message types that can be dynamically assigned ids.
    * Ids over this number are available for static assignment.
    */
   private static final int MAX_DYNAMIC_TYPES = 100;

   /** The list of prototypes, indexed by the id of that message type. */
   private static Message[] prototypes = new Message[MAX_MESSAGE_TYPES];

   /** The number of dynamic types assigned so far. */
   private static int numDynamicTypes = 0;

   /** The input stream to read messages from */
   protected DataInputStream is;

   /**
    * The original input stream passed to us.  This should really be a 
    * FileInputStream or <code>shutdown()</code> will fail.
    */
   protected InputStream origIS;

   /**
    * Use an end marker to sanity check message lengths?  The value of
    * this variable is obtained from the variable of the same name in
    * <code>BasicMessageOutputStream</code>.
    */
   static boolean useMessageLengthGuard = 
                         BasicMessageOutputStream.useMessageLengthGuard;

   /**
    * Create a input stream for reading messages.  Before reading any messages,
    * the class must be initialized with the appropriate message types.
    * @param is The underlying input stream to read from.  Note, this should
    *           really be a FileInputStream, or <code>shutdown()</code> won't 
    *           work, but it's convenient to pass an InputStream.
    */
   protected BasicMessageInputStream(InputStream is) {
      origIS = is;
      this.is = new DataInputStream(new BufferedInputStream(is, 10240));
   }

   /**
    * Register a prototype for a message type, and dynamically assign an ID 
    * number to this type.  This method should be called before messages of
    * this type can be read and written.  The return value is the assigned
    * message id which should then be returned when <code>getMessageID()</code>
    * is called on a message of that type.
    * <p>
    * This method should only be used when the registration order for each 
    * message type is guaranteed to be the same at every node.  Otherwise,
    * the sender and receiver will have different id lists.
    * @param m The mssage prototype to register.
    * @return An integer id number for this message type.
    */
   public static int registerMessagePrototype(Message m) {
      int newID = numDynamicTypes;
      numDynamicTypes++;
      prototypes[newID] = m;
      return newID;
   }

   /**
    * Register a prototype and a statically assigned id number for a message 
    * type.  This method should be called before messages of this type can 
    * be read and written.  The message id that is registered should also 
    * be returned when <code>getMessageID()</code> is called on a message 
    * of that type.
    * <p>
    * It is up to the caller to ensure that the message id assigned to each
    * message type is unique.  Also, the message id must be greater than
    * <code>MAX_DYNAMIC_TYPES</code>.
    * <p>
    * This method should be used when the registration order for each 
    * message type is NOT guaranteed to be the same at every node.  This
    * ensures that the sender and receiver will have the same id lists.
    * @param m The mssage prototype to register.
    * @param id The id to assign to messages of this type.
    * @return No return value.
    */
   public static void registerMessagePrototype(Message m, int id) {
      if (id < MAX_DYNAMIC_TYPES) {
         throw new RuntimeException("Cannot register message, invalid id");
      }

      if (prototypes[id] != null) {
         throw new RuntimeException("Cannot register message, id clash!!!");
      }

      prototypes[id] = m;
   }

   /**
    * Read a message from the input stream.  This method determines the
    * message id of the next message in the stream (by reading the id off
    * the stream) and asks the prototype for this message type
    * to instantiate a new message from the data in the input stream.
    * Presumably the message was written using BasicMessageOutputStream.
    * @return A newly created message instance containing the data from the
    *         input stream.
    * @exception java.io.IOException if there is an error reading from the 
    *                                underlying input stream.
    */
   public synchronized Message readMessage() throws IOException {
      int type = is.read();

      if (type<0) {
         throw new IOException("Bad read");
      }

      if ((type >= MAX_MESSAGE_TYPES) || 
          (prototypes[type] == null)) {
         throw new RuntimeException("Got message of unknown type:" + type);
      }

      Message m = prototypes[type].readNewMessage(is);

      if (useMessageLengthGuard) {
         int eomMarker = is.read();

         if (eomMarker != BasicMessageOutputStream.END_OF_MESSAGE_MARKER) {
            throw new IOException("readMessage() end of message signature " + 
                                  "not found.  Wrong number of bytes read " +
                                  "for message " + m);
         }
      }

      return m;
   }

   /**
    * Shutdown the underlying input stream.  Note that exceptions are 
    * discarded.
    * @return No return value.
    */
   public synchronized void shutdown() {
      try {
         SocketOps.shutdown(origIS);
      } catch (IOException e) {
         System.err.println("Caught exception shutting down message " + 
                            "input stream: " + e);
      }
   }

   public synchronized void close() {
      try {
         is.close();
      } catch (IOException ignore) {}
   }
}
