/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.Conductor.comm;

import java.io.*;

import FMG.Conductor.net.OutputCounter;
import FMG.Conductor.net.SocketOps;
import FMG.Conductor.Conductor;

/**
 * The BasicMessageOutputStream is an implementation of MessageOutputStream
 * that can handle a dynamic set of message types.  This class should be used
 * in conjunction with BasicMessageInputStream.
 * <p>
 * See BasicMessageInputStream for information about registering message
 * ids.
 * <p>
 * This class works in to modes: automatic and manual flush.  With automatic
 * flush, the stream is flushed after every message.  With manual flush,
 * the stream shoudl be flushed using the <code>flush()</code> call.  The mode
 * is set by calling <code>isAutoFlush()</code>.  The default is automatic 
 * flush.
 * <p>
 * TODO: should this still be abstract?
 * 
 * @author Mark Yarvis (yarvis@fmg.cs.ucla.edu)
 * @version $Revision: 1.15 $
 * @see FMG.Conductor.comm.BasicMessageInputStream
 * @see FMG.Conductor.comm.MessageOutputStream
 */

public abstract class BasicMessageOutputStream extends MessageOutputStream {

   /** The output stream to write messages to */
   protected DataOutputStream os;

   /** 
    * The original output stream passed to us.  This should really be a 
    * FileOutputStream or <code>shutdown()</code> will fail.
    */
   protected OutputStream origOS;

   /**
    * The counter stream that keeps track of the bytes actually written to 
    * the socket.
    */
   protected OutputCounter counter;

   /** Are we in the automatic flush mode? */
   protected boolean autoFlush = true;

   /** 
    * Use an end marker to sanity check message lengths?  This behavior is 
    * configured using the Debug.UseMessageLengthGuard property. 
    */
   static boolean useMessageLengthGuard = 
      Conductor.getBooleanProperty("Debug.UseMessageLengthGuard", false);

   /**
    * The value to use for the end of message marker.  This marker is only
    * written if <code>useMessageLengthGuard</code> is true.
    */
   static byte END_OF_MESSAGE_MARKER = 0x55;

   /**
    * Create an output stream for writing messages.  Before writing any
    * messages, the appropriate message types must be registered.
    * @param os The underlying output stream to write to.
    */
   protected BasicMessageOutputStream(OutputStream os) {
      origOS = os;
      counter = new OutputCounter(os);
      this.os = new DataOutputStream(new BufferedOutputStream(counter, 10240));
   }

   /**
    * Set automatic flushing after each message on or off.  This call can 
    * be made at any time.
    * @param shouldAutoFlush Indicates whether or not to automatically flush 
    *                        the stream after each message.
    * @return No return value.
    */
   public void setAutoFlush(boolean shouldAutoFlush) {
      autoFlush = shouldAutoFlush;
   }

   /**
    * Write a message to the output stream.  This message notifies the
    * receiver of the message type to expect and then instructs the message
    * to write itself to the stream.  Presumably the message will be read
    * using BasicMessageInputStream.
    * @param m The message to write to the stream.
    * @return No return value.
    * @exception java.io.IOException if there is an error writing to the
    *                                underlying output stream.
    */
   public synchronized void writeMessage(Message m) throws IOException {
      os.write(m.getMessageID());
      m.sendSelf(os);
      if (useMessageLengthGuard) {
         os.write(END_OF_MESSAGE_MARKER);
      }
      if (autoFlush) {
         os.flush();
      }
   }

   /**
    * Write a message to the output stream without any extra framing.  Unlike
    * <code>writeMessage()</code>, this method does not write out a message 
    * type.  It is presumed that the receiver knows the message type to expect
    * and will call <code>receiveSelf()</code> on a message of that type 
    * directly.
    * @param m The message to write to the stream.
    * @return No return value.
    * @exception java.io.IOException if there is an error writing to the
    *                                underlying output stream.
    */
   public synchronized void writeUnframedMessage(Message m) 
                                                   throws IOException {
      m.sendSelf(os);
      if (autoFlush) {
         os.flush();
      }
   }

   /**
    * Flush any messages that might be buffered in the output stream.
    * Note that exceptions are discarded.
    * @return No return value.
    */
   public synchronized void flush() {
      try {
         os.flush();
      } catch (IOException e) {}
   }

   /**
    * Shutdown the underlying output stream.  Not that exceptions are
    * discarded.
    * @return No return value.
    */
   public synchronized void shutdown() {
      try {
         os.flush();
         SocketOps.shutdown(origOS);
      } catch (IOException e) {
         System.err.println("Caught exception shutting down " + 
                            "MessageOutputStream: " + e);
      }
   }

   public synchronized void close() {
      try {
         os.close();            
      } catch (IOException ignore) {}
   }

   /**
    * Return the number of bytes that have been written to the underlying 
    * output stream.  Note that this may be different from the number of
    * bytes written so far by message objects, due to buffering.  Also note 
    * that the counter is <b>not</b> reset after each call.
    * @return The number of bytes written.
    */
   public long bytesSent() {
      return counter.bytesWritten();
   }
}
