/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.Conductor.comm;

import java.io.*;
import java.net.*;

/**
 * The ConnID class is a unique identifier for a connection in Conductor.
 * Users of this class should not depend on the member datastructures of 
 * this object (that which makes the identifier unique), as they may change 
 * in the future.
 *
 * @author Mark Yarvis (yarvis@fmg.cs.ucla.edu)
 * @version $Revision: 1.7 $
 */

public class ConnID implements Serializable {

   /** A cached copy of the id of the local host. */
   private static NodeID localID = NodeID.getNewLocalID();


   /** The node id of the source of this connection */
   private NodeID node = new NodeID();

   /** 
    * The last allocated timestamp.  The next one must be at least 1 
    * greater. 
    */
   private static long lastAllocatedTime = 0;

   /** 
    * A lock to ensure that only only unique connection id is generated at 
    * a time 
    */
   private static Object allocationLock = new Object();

   /**
     * The time (in milliseconds) of connection startup, as measured on 
     * the source node 
     */
   private long time;

   /**
    * Make this NodeID a unique identifier for a connection requested 
    * by this node.
    * @return No return value.
    */
   public void makeNewUnique() {
      node.copy(localID);
      time = System.currentTimeMillis();

      synchronized (allocationLock) {
         if (time <= lastAllocatedTime) {
            lastAllocatedTime++;
            time = lastAllocatedTime;
         } else {
            lastAllocatedTime = time;
         }
      }
   }

   /**
    * Copy to this object the id from another ConnID object.
    * @param other The object to copy from.
    * @return No return value.
    */
   public void copy(ConnID other) {
      node.copy(other.node);
      time = other.time;
   }

   /**
    * Duplicate this connection id into a new object.
    * @return A new object with the same id as this one.
    */
   public ConnID duplicate() {
      ConnID cid = new ConnID();
      cid.copy(this);
      return cid;
   }

   /**
    * Reinitialize this connection id object with data read from a 
    * DataInputStream.  Presumably the data was written using 
    * <code>writeSelf()</code>.
    * @param is The input stream to read from.
    * @exception java.io.IOException is thrown if there is an error
    *                                reading from the stream.
    */
   public void readSelf(DataInputStream is) throws IOException {
      node.readSelf(is);
      time = is.readLong();
   }

   /**
    * Write this connection id to a DataOutputStream.  Presumably the data 
    * will later be read using <code>readSelf()</code>.
    * @param os The output stream to write to.
    * @exception java.io.IOException is thrown if there is an error
    *                                writing to the stream.
    */
   public void writeSelf(DataOutputStream os) throws IOException{
      node.writeSelf(os);
      os.writeLong(time);
   }

   /**
    * Generate a hashcode for this address suitable for use with the Hashtable
    * and other Java container classes.
    * A number suitable for hashing.
    * @see java.util.Hashtable
    */
   public int hashCode() {
      return (node.hashCode() ^ (int) time);
   }

   /**
    * Determine if this ConnID object contains the same id another ConnID 
    * object.
    * @param o An object to compare with.
    * @return True iff the object is a ConnID object that contains the same
    *         id as this one.
    */
   public boolean equals(Object o) {
      if ((o!=null) && (o instanceof ConnID)) {
         ConnID other = (ConnID) o;
         return (node.equals(other.node)) && (time == other.time);
      } else {
         return false;
      }
   }

   /**
    * Generate a string representation of this id.  This function is for
    * debugging use only.
    */
   public String toString() {
      return ("[" + node + "," + time + "]");
   }
}
