/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.Conductor.comm;

import java.io.*;

import FMG.Conductor.util.*;

/**
 * The Message interface provides a specification for a unit of data
 * that can be read and written via a BasicMessage{In/Out}putStream and
 * stored in various message queues.
 * <p>
 * <b>Important:</b> Note that it is the implementer's responsibility to 
 * ensure that <code>sendSelf()</code> and <code>receiveSelf()</code> are 
 * completely symmetric.  In particular, a call to <code>receiveSelf()</code>
 * must read the same number of bytes that were written in the call to 
 * <code>sendSelf()</code>.  If this does not occur, it is impossible for
 * the caller to know, and subsequent messages will be garbled!
 *
 * @author Mark Yarvis (yarvis@fmg.cs.ucla.edu)
 * @version $Revision: 1.12 $
 */

public interface Message {

   /**
    * Create a new message and read in its data from the given input stream.
    * Note that we must already know the type of the message we're about
    * to read so we can pick an object of the right instance to use.
    * This method is meant to allow a prototype Message object to be asked
    * to read in an object of the same type.
    * <p>
    * TODO: this method is really only needed for 
    *       BasicMessage{In/Out}putStream.  Perhaps we need a sub-interface.
    * @param is The input stream to read the object from.
    * @return A new object initialized from the input stream.
    */
   public abstract Message readNewMessage(DataInputStream is) 
                                                     throws IOException;

   /**
    * Get a unique id for this message.  This identifier is intended to
    * be written to the stream so that the reader can determin, in advance,
    * what class to read in.
    * @return A unique id for this message class.
    */
   public abstract int getMessageID();

   /**
    * Initialize this object's data from the input stream.  Presumably the
    * data has been written using a call to <code>sendSelf()</code> on
    * an object of the same type.
    * <p>
    * <b>Important:</b> see the note in the class description.
    * @param is The input stream to read from.
    * @return No return value.
    * @exception java.io.IOException if an error occurs while readiing from
    *                                the stream.
    */
   public abstract void receiveSelf(DataInputStream is) throws IOException;

   /**
    * Write this object's data to the output stream.  Presumably the
    * data will later by read using a call to <code>receiveSelf()</code> on
    * an object of the same type.
    * <p>
    * <b>Important:</b> see the note in the class description.
    * @param os The ouptput stream to write to.
    * @return No return value.
    * @exception java.io.IOException if an error occurs while writing to
    *                                the stream.
    */
   public abstract void sendSelf(DataOutputStream os) throws IOException;

   /**
    * Writes this object's data to a raw stream.  Presumably, the reader
    * has no knowledge of our data structures.
    * <p>
    * TODO: is this necessary anymore?  Is this used by anyone other
    * than DataFragment?
    * @param os The output stream to write to.
    * @return No return value.
    * @exception java.io.IOException if an error occurs while writing to
    *                                the stream.
    */
   public abstract void writeToRaw(OutputStream os) throws IOException;

}
