/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.Conductor.comm;

import java.util.Vector;

/**
 * The MessageSync is a container class and synchronization primitive.
 * It provides a FIFO message queue.  In addition, if the queue is empty,
 * a call to <code>getMessage()</code> will block until another thread calls
 * <code>putMessage()</code>.
 * <p>
 * This class also has a notion of being <i>open</i> or <i>closed</i>.  
 * Initially, the object is <i>open</i>, and messages can be added or removed.
 * Once <code>close()</code> is called, no further messages can be added.
 * Also, a call to <code>getMessage()</code> on a closed and empty MessageSync
 * will cause an exception, notifying the caller of the closure.
 *
 * @author Mark Yarvis (yarvis@fmg.cs.ucla.edu)
 * @version $Revision: 1.6 $
 */

public class MessageSync {

   /** The underlying data storage. */
   MessageQueue messages = new MessageQueue();

   /** Are we in the closed state? */
   boolean closed = false;

   /**
    * Add a message to the end of the queue.  If a thread is blocked waiting 
    * for a message, it will be notified.
    * @param m The message to add.
    * @return No return value.
    * @exception FMG.Conductor.comm.MessageFlowClosed if close() has been 
    *                                                 called.
    */
   public synchronized void putMessage(Message m) throws MessageFlowClosed {
      if (closed) {
         throw new MessageFlowClosed("Cannot enqueue on closed queue");
      }

      messages.enqueue(m);
      notifyAll();
   }

   /**
    * Get a message from the front of the queue.  The thread will block if
    * the queue is empty.
    * @return The oldest message in the queue.
    * @exception FMG.Conductor.comm.MessageFlowClosed if close() has been 
    *                                                 called and the queue is 
    *                                                 empty.
    * @exception java.lang.InterruptedException if the thread had blocked,
    *                                           but was later interrupted.
    */
   public synchronized Message getMessage() 
                          throws InterruptedException, MessageFlowClosed {
      while (messages.isEmpty()) {
         if (closed) {
            throw new MessageFlowClosed("Cannot dequeue on closed queue");
         }

         wait();
      }

      Message m = messages.dequeue();
      return m;
   }

   /**
    * Is the queue empty?
    * @return True iff the queue is empty.
    */
   public synchronized boolean isEmpty() {
      return messages.isEmpty();
   }

   /**
    * Put the queue in the closed state.  Further calls to 
    * <code>putMessage()</code> will throw an exception.  Calls to 
    * <code>getMessage()</code> will throw an exception if the queue is empty.
    * @return No return value.
    */
   public synchronized void close() {
      closed = true;
      notifyAll();
   }
}
