/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.Conductor.comm;

import java.net.*;
import java.io.*;
import java.util.Hashtable;

/**
 * The NodeID class provides a more palatable replacement for the InetAddress
 * class.  It stores the raw address and never performs DNS lookups.  It can
 * be created from an InetAddress or a long value and an InetAddress can be
 * created when needed.  If it is created from an InetAddress, or if it ever 
 * needs to create an InetAddress, it caches a handle to that object for 
 * future requests.  It can also be written to and read from IO streams.
 * <p>
 * <b>Warning:</b> Unline the InetAddress object, this object is mutable
 * by a call to readSelf() or copy()!
 *
 * @author Mark Yarvis (yarvis@fmg.cs.ucla.edu)
 * @version $Revision: 1.17 $
 *
 * @see java.net.InetAddress
 */

public class NodeID {

   /**
    * A cache of InetAddress objects that correspond to the address of
    * each NodeID.
    */
   protected static Hashtable addressCache = new Hashtable();

   /** A cached object representing the local node's address */
   private static NodeID localhost;

   /** The actual address stored as a byte-array */
   private byte[] address = new byte[4];

   /**
    * Create a new NodeID from an InetAddress object.
    * @param addr A representation of the desired host address.
    */
   public NodeID(InetAddress addr) {
      this(addr.getAddress());
   }

   /**
    * Create a new NodeID from a byte array.  The byte array is assumed
    * to be 4 bytes for IPv4.
    * @param address The byte array for address construction.
    */
   public NodeID(byte[] address) {
      for (int i=0; i<4; i++) {
         this.address[i] = address[i];
      }
   }

   /**
    * Create a new NodeID by reading the address from an input stream.
    * Presumably the NodeID was written using writeSelf().
    * @param addr A representation of the desired host address.
    * @exception java.io.IOException if the read operation throws an 
    *                                exception, it is passed along.
    * @see FMG.Conductor.comm.NodeID#writeSelf
    */
   public NodeID(InputStream is) throws IOException {
      readSelf(is);
   }

   /**
    * Create a NodeID that points to an unkown address (0.0.0.0).
    */
   public NodeID() {
      for (int i=0; i<4; i++) {
         address[i]=0;
      }
   }

   /**
    * Create a NodeID from a numeric representation.
    * @param addr An int value representing the address in host byte order.
    */
   public NodeID(int addr) {
      for (int i=3; i>=0; i++) {
         address[i]= (byte) (addr & 0xFF);
         addr = addr / 256;
      }
   }

   /**
    * Create a NodeID from an IP address in dotted-decimal noation.
    */
   public NodeID(String s) {
      byte[] buf = s.getBytes();
      int pos = 0;

      for (int i = 0; i<buf.length; i++) {
         if (buf[i] == '.') {
            pos++;
            continue;
         }
         address[pos] *= 10;
         address[pos] += buf[i] - '0';
      }
   }

   /**
    * Obtain an address object with the same address as the given NodeID.
    * If such an object is cached, it is retreived from the cache.  Otherwise,
    * one is created.
    * @param node The id of the node in question.
    * @return An InetAddress object with the desired address.
    */
   protected static InetAddress getAddress(NodeID node) {
      InetAddress addr = (InetAddress) addressCache.get(node);
      if (addr == null) {
         try {
            addr = InetAddress.getByName(node.toString());
            addressCache.put(node.duplicate(), addr);
         } catch (UnknownHostException e) {
            throw new RuntimeException("Couldn't make InetAddress object " +
                                       "from IP address.");
         }
      }
      return addr;
   }

   /**
    * Obtain a node ID representing our notion of the local host.  If no
    * such notion exists yet, create one based on Java's notion of the local
    * host address.
    * @return A node ID representing the local node.
    */
   protected static NodeID localhost() {
      if (localhost == null) {
         try {
            return new NodeID(InetAddress.getLocalHost());
         } catch (UnknownHostException e) {
            throw new RuntimeException("Couldn't get local address.");
         }
      }

      return localhost;
   }

   /**
    * Does this NodeID represent a local address for this host?
    * @return True iff this node id represents the local node
    */
   public boolean isLocalNode() {
      return equals(localhost());
   }

   /**
    * Get a new NodeID object representing the local host's address.
    * @return A new object representing the local host.
    */
   public static NodeID getNewLocalID() {
      return localhost().duplicate();
   }

   /**
    * Set Conductor's notion of the local host.
    * @param addr The address representing the local host.
    * @return No return value.
    */
   public static void setLocalID(InetAddress addr) {
      localhost = new NodeID(addr);
   }

   /**
    * Determine if this address is a loopback interface.
    * @return True iff the most significant octet has the value 127.
    */
   public boolean isLoopback() {
      return (address[0] == 127);
   }

   /**
    * Copy the contents of another NodeID object.
    * @param other The source to copy from.
    * @return No return value.
    */
   public void copy(NodeID other) {
      address[0] = other.address[0];
      address[1] = other.address[1];
      address[2] = other.address[2];
      address[3] = other.address[3];
   }

   /**
    * Create a new NodeID object pointing to the same address as this one.
    * @return A new object containing the same address contained in this 
    *         object.
    */
   public NodeID duplicate() {
      return new NodeID(address);
   }

   /**
    * Obtain an InetAddress pointing to the address contained in this object.
    * @return An InetAddress object containing the address of the host 
    *         represented by this object.
    */
   public InetAddress getAddress() {
      return getAddress(this);
   }

   /**
    * Update this NodeID with the address from the given input stream.  
    * Presumably this address was written by writeSelf().
    * @param is The input stream to read from.
    * @return No return value.
    * @exception java.io.IOException if the call to read the address from 
    *                                the input stream returns an exception, 
    *                                it is passed along.
    */
   public void readSelf(InputStream is) throws IOException {
      int pos = 0;

      while (pos < 4) {
         pos += is.read(address, pos, 4-pos);
      }
   }

   /**
    * Write the address contained in this object to an output stream.
    * Presumably this address will be read by readSelf() or by the appropriate 
    * constructor.
    * @param is The input stream to read from.
    * @return No return value.
    * @exception java.io.IOException if the call to read the address from 
    *                                the input stream returns an exception, 
    *                                it is passed along.
    */
   public void writeSelf(OutputStream os) throws IOException {
      os.write(address, 0, 4);
   }

   /**
    * Generate a string representation of this address in decimal dot 
    * notation.
    * @return a new string containing a user-understandable representation 
    *         of the address.
    */
   public String toString() {
      /*
       * WARNING: do not change this implemantation without adjusting the 
       * getAddress(NodeID) method.
       */
      return "" + ((int) address[0] & 0xFF) + "." +
                  ((int) address[1] & 0xFF) + "." +
                  ((int) address[2] & 0xFF) + "." +
                  ((int) address[3] & 0xFF);
   }

   /**
    * Get a numerical representation of this address in host byte order.
    * @return a long representing the host address stored in this object.
    */
   public int intValue() {
      int result = 0;

      for (int i=0; i<4; i++) {
         result = address[i] + (result * 256);
      }

      return result;
   }

   /**
    * Get a byte-array representation of this address in host byte order.
    * @return a new byte-array representing the host address stored in this 
    *         object.
    */
   public byte[] getBytes() {
      byte[] ret = new byte[4];

      for (int i=0; i<4; i++) {
         ret[i] = address[i];
      }

      return ret;
   }

   /**
    * Set the value of this node id from an address in byte notation.
    * @param addr An array of bytes representing the address.  It is expected 
    *             to be exactly 4 bytes in length.
    * @return No return value.
    */
   public void setBytes(byte[] addr) {
      for (int i=0; i<4; i++) {
         address[i] = addr[i];
      }
   }

   /**
    * Generate a hashcode for this address suitable for use with the Hashtable 
    * and other Java container classes.
    * @return A number suitable for hashing.
    * @see java.util.Hashtable
    */
   public int hashCode() {
      return (address[0] << 24) | (address[1] << 16) +
             (address[2] << 8) + address[3];
   }

   /**
    * Determine if this NodeID object contains the same host address as
    * another NodeID object.
    * @param o An object to compare with.
    * @return True iff the object is a NodeID object that contains the same
    *         address as this one.
    */
   public boolean equals(Object o) {
      if (o==null) {
         return false;
      }

      if (o instanceof NodeID) {
         NodeID other = (NodeID) o;
         return ((address[0] == other.address[0]) && 
                 (address[1] == other.address[1]) &&
                 (address[2] == other.address[2]) &&
                 (address[3] == other.address[3]));
      } else if (o instanceof InetAddress) {
         return getAddress().equals(o);
      } else {
         return false;
      }
   }

   /**
    * Determine if this NodeID contains an unknown address (all zeros).
    * @return True iff the address is all zeros.
    */
   public boolean isNull() {
      return ((address[0] == 0) && (address[1] == 0) &&
              (address[2] == 0) && (address[3] == 0));
   }
}
