/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.Conductor.dataflow;

import FMG.Conductor.Conductor;
import FMG.Conductor.util.Direction;
import FMG.Conductor.util.IntP;
import FMG.Conductor.dataflow.messages.DataFragment;

/**
 * An AccessorBoundary class deliniates one edge of a StreamAccessor.
 * A pair of AccessorBoundary objects surrounds a portion
 * of the stream that is protected from access by a lock.  Access to
 * the protected segment is provided only through the boundary, which
 * obtains the lock before allowing access.  Notice that this lock also
 * obtains access to the "inner" half of each boundary.
 * <p>
 * AccessorBoundary objects interact with StreamAccessor objects.  Each 
 * AccessorBoundary is associated with a StreamAccessor that controls the 
 * adjacent territory.  When a StreamElement is blocked by a Boundary, the 
 * associated Stream accessor is notified.
 * <p>
 * Finally, a pair of boundaries implements a queuing limit.  A maximum
 * number of data objects may be moved within a pair of boundaries.  Beyond
 * this amount, the operation will block until data is retreived by the
 * downstream boundary.  The queuing limit is controled by the 
 * <i>DataQueuingLimit</i> property in the Conductor properties file.  The
 * default value is <code>Integer.MAX_VALUE</code>
 * <p>
 * To prevent accidental operations on the shared portion of the stream,
 * operations such as <code>insert()</code>, <code>swap()</code>, and
 * <code>next()</code> are partially or completely disabled.  Internally,
 * <code>safeInsert()</code>, <code>safeSwap()</code>, and 
 * <code>safeNext()</code> are used, instead.  Externally, only the
 * higher-level operations should be used.
 *
 * @author Mark Yarvis (yarvis@fmg.cs.ucla.edu)
 * @version $Revision: 1.3 $
 */

public class AccessorBoundary extends Boundary {

   /** 
    * The initial value for queuingLimit.  This value may be set using the 
    * <i>DataQueuingLimit</i> property.  The default value is 
    * <code>Integer.MAX_VALUE</code>
    */
   protected static int initialQueuingLimit = 
             Conductor.getIntProperty("DataQueuingLimit", Integer.MAX_VALUE);

   /*
    * The maximum number of data objects that may be queued between a 
    * pair of boundaries.
    */
   protected int queuingLimit = initialQueuingLimit;

   /** The boundary object paired with this one.  */
   protected AccessorBoundary pair;

   /** The StreamAccessor object associated with this boundary. */
   protected StreamAccessor accessor;

   /** The lock file used to protect the data between the boundaries. */
   protected Object lock;

   /** The side of the boundary that requires locking. */
   protected Direction lockDir;

   /** 
    * The amount of queuing space available.  Initially set to 
    * <code>queuingLimit</code>.
    */
   protected IntP spaceRemaining;

   /**
    * Create a new AccessorBoundary object.
    */
   protected AccessorBoundary() {
      lock = new Object();
      spaceRemaining = new IntP(queuingLimit);
   }

   /**
    * Create a new boundary object and pair it with an existing one.
    * Both will share a lock and a common handle to 
    * <code>spaceRemaining</code>.
    * @param pair An existing boundary to pair with this one.
    */
   protected AccessorBoundary(AccessorBoundary pair) {
      this.pair = pair;
      lock = pair.lock;
      lockDir = pair.lockDir.opposite();
      spaceRemaining = pair.spaceRemaining;
   }

   /**
    * Set the StreamAccessor object for this boundary.
    * @param sa the new StreamAccessor.
    */
   void setAccessor(StreamAccessor sa) {
      accessor = sa;
   }

   /**
    * Adjust the queuing limit.  This method will also adjust the 
    * spaceRemaining by the difference between the old and new limits.
    * @param newLimit The new limit, replacing the old one.
    * @return No return value.
    */
   void setNewQueuingLimit(int newLimit) {
      synchronized (lock) {
         spaceRemaining.val += (newLimit - queuingLimit);
         queuingLimit = newLimit;
      }
   }

   /**
    * Wait for some queue space to become available.  <b>Note</b>, the lock
    * must already be held before this call!
    * @return No return value.
    */
   protected final void waitForQueueSpace() {
      while (spaceRemaining.val <= 0) {
         try {
            lock.wait();
         } catch (InterruptedException e) {}
       }
   }

   /**
    * Adjusts the remaining queue space by the given value.  Notify waiting
    * threads if space has been freed up.  <b>Note</b>, the lock
    * must already be held before this call!
    * @return No return value.
    */
   protected final void adjustQueueSize(int adjustment) {
      if (adjustment == 0) {
         return;
      }

      int prevSpaceRemaining = spaceRemaining.val;
      spaceRemaining.val -= adjustment;

      if ((prevSpaceRemaining <= 0) && (spaceRemaining.val > 0)) {
         lock.notifyAll();
      }
   }

   /**
    * Insert a stream element next to this object.  This method will not
    * allow insertion on the shared portion of the stream.  Instead, a
    * RuntimeException will be thrown.
    * @param p The stream element to insert.
    * @paramm dir The direction, relative to this object, for insertion.
    * @return No return value.
    */
   public void insert(StreamElem p, Direction dir) {
      if (dir.equals(lockDir)) {
         /* Nobody should be doing this.  Danger of violating locking. */
         throw new RuntimeException("Operation not permitted");
      } else {
         super.insert(p, dir);
      }
   }

   /**
    * This operation is disallowed for this object as it will violate
    * locking.  A RuntimeException will be thrown. 
    */
   public void swap(Direction dir) {
      throw new RuntimeException("Operation not permitted");
   }

   /**
    * Determine the adjacent stream object in a particular direction.  This 
    * method will not allow access to the shared portion of the stream.  
    * Instead, a RuntimeException will be thrown.
    * @param p The stream element to insert.
    * @paramm dir The direction, relative to this object, for insertion.
    * @return No return value.
    */
   public StreamElem next(Direction dir) {
      if (dir.equals(lockDir)) {
         /* Nobody should be doing this.  Danger of violating locking. */
         throw new RuntimeException("Operation not permitted");
      } else {
         return super.next(dir);
      }
   }

   /**
    * Obtain a handle for the adjacent stream element in the given direction.
    * This method may only be used when the lock is held.
    * @param dir The direction of the object we're interested in.
    * @return A handle to an adjacent object.
    */
   protected StreamElem safeNext(Direction dir) {
      return super.next(dir);
   }

   /**
    * Swap this stream element with the adjacent element.
    * This method may only be used when the lock is held.
    * @param dir The direction to swap.
    * @return No return value.
    */
   protected void safeSwap(Direction dir) {
      StreamElem target = safeNext(dir);

      if (target.isData()) {
         /* adjust space remaining here for common case efficiency */
         if (dir.equals(lockDir)) {
            if (spaceRemaining.val == 0) {
               lock.notifyAll();
            }
            spaceRemaining.val++;
         } else {
            spaceRemaining.val--;
         }
      }

      target.unlink();
      super.insert(target, dir.opposite());
   }

   /**
    * Insert an element adjacent to this one.
    * This method may only be used when the lock is held.
    * @param se The stream element to insert.
    * @param dir The side on which to insert the new stream element.
    * @return No return value.
    */
   protected void safeInsert(StreamElem se, Direction dir) {
      if (se.isData()) {
         /* adjust space remaining here for common case efficiency */
         if (dir.equals(lockDir)) {
            if (spaceRemaining.val == 0) {
               lock.notifyAll();
            }
            spaceRemaining.val--;
         }
      }

      super.insert(se, dir);
   }

   /**
    * Swap one event out of the window.  This operation can be down
    * for either a downstream boundary or an upstream boundary.
    * @return no return value.
    */
   public void swapEventOut() {
      Object origLock = lock;
      synchronized (lock) {
         if (lock != origLock) {
            swapEventOut();
            return;
         }
         super.swap(lockDir.opposite());
         adjustQueueSize(super.next(lockDir).proceed());
      }
   }

   /**
    * Put a new stream element into the shared portion of the stream.
    * @param se The element to insert.
    * @return No return value.
    */
   public void putOneOut(StreamElem se) {
      Object origLock = lock;
      synchronized (lock) {
         if (! se.isEvent()) {
            waitForQueueSpace();
         }
         if (lock != origLock) {
            putOneOut(se);
            return;
         }

         if (se.isData()) {
            recordSegmentation((DataFragment) se);
         }

         super.insert(se, lockDir);
         adjustQueueSize(se.proceed() + (se.isData()? 1 : 0));
      }
   }

   /**
    * If an element of the specified type is adjacent (in the locking
    * direction), move past it and return a pointer to it.
    * @param msgClass The type of object we're interested in.
    * @return A pointer to the element that was found, or null.
    */
   public StreamElem getEvent(Class targetClass) {
      Object origLock = lock;
      synchronized (lock) {  
         if (origLock != lock) {
            return getEvent(targetClass);
         }

         StreamElem target = safeNext(Direction.up);
         if ((target != null) && targetClass.isInstance(target)) {
            safeSwap(lockDir);
            return target;
         } else {
            accessor.serviceComplete(this);
            return null;
         }
      }
   }

   /**
    * The distance to any byte is always -1.  This prevents searches for
    * bytes from straying beyond boundaries.
    * @param targetByte The byte to look for.
    * @return Always -1.
    */
   public int distanceTo(byte targetByte) {
      return -1;
   }

   /**
    * This method is called whenver an object in motion stops next to this
    * boundary.  If the object is on the shared part of the stream, record
    * the fact that there may be work waiting and notify any waiting threads.
    * @param dir The direction from which the element arrived.
    * @param se The element that is now adjacent to this one.
    * @return An adjustment to the number of data objects as a result
    *         of this operation.  Always 0.
    */
   public int handleNewAdjacent(Direction dir, StreamElem se) {
      /*
       * Note: if the accessor is null, then setUp/DownstreamBoundary() 
       * has not been called yet.  Once it is, needsService() will also be 
       * called.  So, in this case, we can do nothing here.
       */
      if (dir.equals(lockDir) && (accessor != null)) {
         accessor.needsService(this);
      }
      return 0;
   }

   /**
    * Merge this boundary with another one.  The other boundary must
    * belong to an adjacent pair.  This operation is used when an object
    * with exclusive control over a portion of the data stream (ie, a 
    * StreamAccessor) gives up that control.  This leads to two directly
    * adjacent portions of the stream that are controlled by boundaries.
    * these boundary pairs must be merged, leaving one lock and one 
    * handle to spaceRemaining.  Note that the changes in locks is
    * handled entirely internally to this class.  Callers need not
    * take any action.
    * @param other The adjacent boundary to merge.
    * @return No return value.
    */
   public void merge(AccessorBoundary other) {
      Object origLock = lock;
      Object otherOrigLock = other.lock;

      if (! (isBoundary() && other.isBoundary())) {
         throw new RuntimeException("merge performed on non-boundary");
      }

//System.out.println("########## performing a merge!");
      synchronized (lock) {
         synchronized (other.lock) {
            if ((origLock != lock) || (otherOrigLock != other.lock)) {
               merge(other);
               return;
            }

            /* count neighboring space used against us */
            spaceRemaining.val -= (queuingLimit - other.spaceRemaining.val);

            /* replace shared objects */
            other.pair.lock = lock;
            other.pair.spaceRemaining = spaceRemaining;
            other.lock = lock;
            other.spaceRemaining = spaceRemaining;

            /* revise the pairing */
            pair.pair = other.pair;
            other.pair.pair = pair;

            /* figure out who is up and who is down */
            AccessorBoundary upBound;
            AccessorBoundary downBound;
            if (safeNext(Direction.up) == other) {
               downBound = this;
               upBound = other;
            } else if (safeNext(Direction.down) == other) {
               upBound = this;
               downBound = other;
            } else {
               throw new RuntimeException("Attept to merge non-adjacent " + 
                                          "boundaries");
            }

            boolean didSomething = true;
            while (didSomething) {
               didSomething = false;

               /* give upstream elements a chance to flow down */
               for (StreamElem se = upBound.safeNext(Direction.up);
                    (! se.flowDir().isUp());
                    se = upBound.safeNext(Direction.up)) {
                  se.unlink();
                  downBound.safeInsert(se, Direction.down);
                  /**
                   * Adjust for any changes resulting from proceed, AND undo
                   * any change made my safeInsert().
                   */
                  pair.adjustQueueSize(se.proceed() - (se.isData() ? 1 : 0));
                  didSomething = true;
               }

               /* give donwstream elements a chance to flow up */
               for (StreamElem se = downBound.safeNext(Direction.down);
                    (! se.isBoundary()) && (se.flowDir().isUp());
                    se = downBound.safeNext(Direction.down)) {
                  se.unlink();
                  upBound.safeInsert(se, Direction.up);
                  /**
                   * Adjust for any changes resulting from proceed, AND undo
                   * any change made my safeInsert().
                   */
                  pair.adjustQueueSize(se.proceed() - (se.isData() ? 1 : 0));
                  didSomething = true;
               }
            }

            /* get out of the way */
            unlink();
            other.unlink();

         }
      }
   }

   /* don't do this without making it safe */
   public int advance() {
      throw new RuntimeException("This operation is not implemented");
   }

}
