/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.Conductor.dataflow;

import java.lang.reflect.*;

import FMG.Conductor.dataflow.*;
import FMG.Conductor.util.*;

/**
 * The Adaptor abstract class is the base class for all adaptor objects.
 * This class provides a method for instantiating an adaptor of any class.  
 * In addition, this class gives a subclass adaptor the facilities it
 * needs to operate, primarily access to the AdaptorWindow and IAC objects.
 * <p>
 * Any adaptor should subclass this class and override the abstract 
 * <code>handleFlow()</code> method.  The <code>run()</code> method
 * must <b>not</b> be overridden, or Conductor will not function correctly.
 * Each adaptor should also provide a constructor which takes a Window
 * as an argument and at least calls the super constructor provided.
 * <p>
 * Each adaptor has it's own thread of control.  For the adaptor writer's
 * purposes, this thread of control starts with a call to 
 * <code>handleFlow()</code> and ends when that method returns.
 *
 * @author Mark Yarvis (yarvis@fmg.cs.ucla.edu)
 * @version $Revision: 1.18 $
 */

public abstract class Adaptor extends Job {

   /** The last created instance of an adaptor. For debugging purposes only! */
   private static int lastInstance = 0;

   /** The instance number of this adaptor. For debugging purposes only! */
   protected int instanceNum = lastInstance++;

   /** A string identifier for this object. For debugging purposes only! */
   private String objectID;

   /** This adpator's window onto the data stream. */
   protected AdaptorWindow window;

   /** This adaptor's name (currently the class name). */
   private String adaptorName;

   /** The session with which an adaptor is associated. */
   private Session session;

   public static class CompositionType {
      public static final CompositionType NONE = 
                               new CompositionType() {
                                  public boolean isNone() {return true;}
                               };

      public static final CompositionType PUSH = 
                               new CompositionType() {
                                  public boolean isPush() {return true;}
                               };

      public static final CompositionType POP = 
                               new CompositionType() {
                                  public boolean isPop() {return true;}
                               };

      public boolean isNone() { return false; }

      public boolean isPush() { return false; }

      public boolean isPop() { return false; }
   }

   {
      // initialize the object id with the class basename and the istanceID
      String name = getClass().getName();
      int nameStart = name.lastIndexOf('.') + 1;
      name = name.substring(nameStart, name.length());

      objectID = name + '#' + instanceNum;
   }

   /**
    * Create a new adaptor.
    * @param window The adaptor's window onto the data flow.
    * @param parameter The parameter passed to the adaptor (if any).
    */
   public Adaptor(AdaptorWindow window, String parameter) {
      this.window = window;
      session = window.getSession();
   }

   /**
    * This method is the main body that Conductor uses for creating and
    * shutting down adaptors.  Subclasses should not override this method.
    * override <code>handleFlow()</code> instead.
    * @return No return value.
    */
   public final void run() {
      boolean failureOccurred = false;
      if (Debug.on) {
         currentThread.setName(adaptorName);
      }
      try {
         handleFlow();
      } catch (RuntimeException e) {
         Debug.println("**** An adaptor (" + this + ") failed: "); 
         e.printStackTrace();
         failureOccurred = true;
      }
      currentThread.yield();
      window.removeSelf(failureOccurred);
   }

   /**
    * This method provides the main code for user-level adaptors.
    * Override this method and operate on the window to perform
    * adaptations.
    */
   protected abstract void handleFlow();

   /**
    * An accessor method to the inter-adaptor communication object for this
    * data stream.  Adaptors cooperating on a given stream can use this object
    * to communicate.
    * @return A handle for the IAC object for this stream.
    */
   protected IAC getStreamIAC() {
      return session.iac;
   }
   
   /**
    * An accessor method to the global inter-adaptor communication object
    * for all streams on this node.  Adaptors can achieve  global 
    * communication across streams using this object.
    * @return A handle for the IAC object for this node.
    */
   protected IAC getNodeIAC() {
      return IAC.getNodeIAC();
   }

   public static String getNameFromSpec(String adaptorSpec) {
      int endOfName = adaptorSpec.indexOf(' ');
      if (endOfName == -1) {
         return adaptorSpec;
      } else {
         return adaptorSpec.substring(0, endOfName);
      }
   }

   public static String getParamsFromSpec(String adaptorSpec) {
      int endOfName = adaptorSpec.indexOf(' ');
      if (endOfName == -1) {
         return null;
      } else {
         return adaptorSpec.substring(endOfName + 1, adaptorSpec.length());
      }
   }
   
   /**
    * Instantiate and start a given adaptor with the specified window.  The 
    * adaptor must implement a constructor that takes a Window object as an 
    * argument.  Once created, the adaptor's thread is started.
    * @param name The class name of the desired adaptor.
    * @param window The window object to pass to the constructor.
    * @return No return value.
    */
   public static void createNewAdaptor(String adaptorSpec, 
                                       AdaptorWindow window) {
      try {
         String name = getNameFromSpec(adaptorSpec);
         String parameter = getParamsFromSpec(adaptorSpec);

         /* get the appropriate constructor */
         Class adaptorClass = Class.forName(name);
         Class[] constructorArgTypes = { AdaptorWindow.class, String.class };
         Constructor cons = adaptorClass.getConstructor(constructorArgTypes);

         /* call constructor */
         Object[] constructorArgs = new Object[2];
         constructorArgs[0] = window;
         constructorArgs[1] = parameter;
         Adaptor adaptor = (Adaptor) cons.newInstance(constructorArgs);

         adaptor.adaptorName = name;

         /* start a thread for this adaptor */
         ThreadRecycler.startNewJob(adaptor);
      } catch (ClassNotFoundException e) {
         throw new RuntimeException(e.toString());
      } catch (NoSuchMethodException e) {
         throw new RuntimeException(e.toString());
      } catch (InstantiationException e) {
         throw new RuntimeException(e.toString());
      } catch (InvocationTargetException e) {
         throw new RuntimeException(e.getTargetException().toString());
      } catch (IllegalAccessException e) {
         throw new RuntimeException(e.toString());
      }
   }

   public String toString() {
      return objectID;
   }

   /**
    * This static method indicates whether or not this adaptor is paired
    * with another adaptor in the stream.  An unpaired adaptor should
    * return <code>CompositionType.NONE</code>.  The upstream adaptor in an
    * adaptor pair should return <code>CompositionType.PUSH</code>, while
    * the downstream adaptor should return <code>CompositionType.POP</code>.  
    * If an adaptor class does not implement this method, 
    * <code>CompositionType.NONE</code> will be assumed.
    * @param parameters The parameters that were specified in the plan for 
    *                   deployment of this adaptor.
    * @return Either <code>CompositionType.NONE</code>, 
    *         <code>CompositionType.PUSH</code>, or 
    *         <code>CompositionType.POP</code>.
    */
   public static CompositionType compositionType(String parameters) {
      return CompositionType.NONE;
   }
}
