/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.Conductor.dataflow;

import FMG.Conductor.util.*;
import FMG.Conductor.comm.ConnID;
import FMG.Conductor.dataflow.messages.*;
import FMG.Conductor.Conductor;

// Be specific to avoid confusion with java.lang.Shutdown
import FMG.Conductor.dataflow.messages.Shutdown;

/**
 * The AdaptorWindow provides an adaptor with a window into a unidirection
 * portion of the data stream.  A set of operators are provided for 
 * moving the window across the data stream, allowing an adaptor to perform
 * operations on that data (using a different set of objects).
 * <p>
 * Each AdaptorWindow has an associated Adaptor, which is created when the
 * AdaptorWindow is instantiated.  The AdaptorWindow is passed as an argument
 * to the Adaptor's constructor.
 * <p>
 * The window consists of an upstream boundary, a downstream boundary, and
 * a chunk of the stream in between.  The adaptor has exclusive control
 * over the portion of the stream between the boundaries and partial control
 * over the boundaries.  The adaptor can modify the data in this portion 
 * of the stream without obtaining any locks.  The window can also be 
 * expanded and contracted, but locks for the adjacent area of the stream 
 * must first be obtained.  Some window operations may block until data 
 * becomes available.  The window includes a notification system allowing
 * the stream to notify the blocked operation of available data.
 *
 * TODO: We need an interface for acknowledgement and retransmission of
 * segments.  Can we maintain segment transparency?  Can it simply be in
 * terms of input or output bytes?  Perhaps, if we keep track of the mapping
 * between segment start number and byte number.  Could also allow checkpoint
 * objects that the adaptor can create (perhaps out of a pointer object) that
 * remember the segment positioning.  The later idea would have to ensure
 * that it falls on segment boundaries.  How transparent is this?
 *
 * @author Mark Yarvis (yarvis@fmg.cs.ucla.edu)
 * @version $Revision: 1.45 $
 */

public class AdaptorWindow extends StreamAccessor {

   /** 
    * Specifies whether or not recovery should be initiated in response to 
    * failure.  The value is controlled by the property InitiateRecovery and
    * currently defaults to false (until testing is completed).
    */
   public static final boolean INITIATE_RECOVERY = 
      Conductor.getBooleanProperty("InitiateRecovery", false);

   /** Specifies the maxumum level of buffering. */
   public static final int MAX_BUFFERING = BufferingControl.MAX_BUFFERING;

   /** Specifies the minimum level of buffering. */
   public static final int MIN_BUFFERING = BufferingControl.MIN_BUFFERING;

   /** Specifies the default level of buffering. */
   public static final int DEFAULT_BUFFERING = 
                                           BufferingControl.DEFAULT_BUFFERING;

   /** The session object for this connection */
   protected Session session;

   protected Adaptor.CompositionType compositionType;

   protected StreamTypeStack.StreamType myType;

   protected StreamTypeStack upstreamTypeStack = new StreamTypeStack();
   protected StreamTypeStack downstreamTypeStack = new StreamTypeStack();

   protected boolean inhibitTypeRecovery = false;
   protected boolean inhibitDataRecovery = false;

   /** 
    * The name of this adaptor as called for in the plan from which it was
    *  deployed. 
    */
   String adaptorName;

   /**
    * Create a new AdaptorWindow and instantiate the associated adaptor.
    * @param adaptor The name of the adaptor to create.  Currently this is
    *                the full class name of the adaptor.
    * @param session The session object for this connection.
    * @param upstreamBoundary The boundary in which data will flow.
    * @param downstreamBoundary The boundary out of which data will flow.
    */
   public AdaptorWindow(String adaptor, Session session, 
                        UpstreamBoundary upstreamBoundary, 
                        DownstreamBoundary downstreamBoundary) {

      super(upstreamBoundary, downstreamBoundary);
      this.session = session;

      /* create and start adaptor */
      adaptorName = adaptor;
      Adaptor.createNewAdaptor(adaptor, this);
   }

   /**
    * Obtain the session object for the connection with which this window
    * is associated
    * @return The sesison object for this connection.
    */
   Session getSession() {
      return session;
   }

   /**
    * Look to see if there is an event downstream that is trying to move
    * upstream, into this window.  If there is, move it in and handle it.
    * @exception FMG.Conductor.dataflow.AsyncEventException is thrown if an
    *                             event is present, but is not fully handled
    *                             by this routine.
    * @return No return value.
    */
   protected void checkForUpwardEvent() throws AsyncEventException {

      for (StreamElem se = downstreamBoundary.getAnEvent();
           (se != null);
           se = downstreamBoundary.getAnEvent()) {
         se.handleEntry(this);
      }
   }

   /**
    * Handle the case in which a SwitchPlan message has just entered the
    * window.  In particular, any adaptors required prior to this adaptor
    * should be installed.  Also, if this adaptor is slated to be removed,
    * that should be handled here as well.
    * @param sp The SwitchPlan message that just entered the window.
    * @return Never returns ... an exception is always thrown.
    * @exception FMG.Conductor.dataflow.AsyncEventException 
    *                       Will always be thrown to alert the user to a 
    *                       change of plans.
    */
   public boolean handleEventEntry(SwitchPlan sp) throws AsyncEventException {
      if (sp.installAdaptors(adaptorName, session)) {
         setUpstreamBoundary((UpstreamBoundary) sp.next(Direction.up));
         sp.copyTypeStack(upstreamTypeStack);
         throw new PlanChangedException();
      } else {
         // handle deletions
         sp.unlink();
         downstreamBoundary.putOneOut(sp);
         if (Debug.on) {
            Debug.println(Debug.DATAFLOW, Debug.INFO,
                          adaptorName + ": removing self " + 
                          "(not called for in new plan)");
         }

         // don't recover types, the SwitchPlan message will take care of this
         inhibitTypeRecovery = true;

         throw new RuntimeException("This adaptor is no longer needed!");
      }
   }

   public boolean handleEventEntry(RebalanceTypes rt) 
                                    throws AsyncEventException {
      super.handleEventEntry(rt);
      rt.unlink();

      if (Debug.on) {
         Debug.println(Debug.RELIABILITY, Debug.MOREINFO,
                       this + " received " + rt);
      }

      // in most cases, don't do any recovery ... it's already ongoing
      inhibitTypeRecovery = true;
      inhibitDataRecovery = true;

      if (rt.flowDir().isDown()) {          // if it's going downstream
         if (rt.equalsTargetTypeStack(downstreamTypeStack)) {
            inhibitDataRecovery = false;    // do data recovery!!!
            rt.recycle();
         } else {                           // keep going
            downstreamBoundary.putOneOut(rt);
         }            
      } else {                              // must be going upstream
         if (rt.equalsTargetTypeStack(upstreamTypeStack)) {
            rt.recycle();                   // we've achieved the desired type
         } else {
            upstreamBoundary.putOneOut(rt); // keep going
         }
      }

      // fail local adaptor
      throw new RuntimeException("Adaptor failing (to rebalance types)");
   }

   /**
    * Handle the case in which we receive notification of a failure upstream.
    * The message needs to proceed until it reaches the start of a complete 
    * segment.  Either that occurs within our window or beyond it.  In the
    * former case, it will become a <code>RetransRequest</code> and we then
    * have to wait for a <code>RetransStart</code> message.  In the later
    * case, we have to wait for the <code>RetransRequest</code> to be sent
    * from downstream.
    * @param sf The <code>SignalFailure</code> message that just entered 
    *           our window.
    * @return Doesn't ever return.
    * @exception FMG.Conductor.dataflow.AsyncEventException 
    *                       Will always be thrown to alert the user to a 
    *                       change in the data stream.
    */
   public boolean handleEventEntry(SignalFailure sf) 
                                    throws AsyncEventException {
      if (Debug.on) {
         Debug.println(Debug.ADAPTOR, Debug.INFO, 
                       adaptorName + ": Received " + sf);
      }

      sf.proceed();

      if (downstreamBoundary.next(Direction.up) == sf) {
         // it made it to the other side, so push it out
         sf.unlink();
         downstreamBoundary.putOneOut(sf);

         // wait for RetransRequest
         StreamElem rr = waitForMessage(Direction.down, RetransRequest.class);
         rr.handleEntry(this);
      } else {
         // it must have come back as a RetransRequest
         RetransRequest rr = 
            (RetransRequest) upstreamBoundary.next(Direction.down);
         rr.handleExit(this);
         rr.unlink();
         upstreamBoundary.putOneOut(rr);

         // wait for RetransStart
         StreamElem rs = waitForMessage(Direction.up, RetransStart.class);
         rs.handleEntry(this);
      }

      // the above calls to handleEntry() should have resulted in an exception
      throw new RuntimeException("This shouldn't happen");
   }

   /**
    * Handle the case in which we receive a retransmission request from
    * downstream.  We currently assume that the adaptor is incapable of
    * doing any retransmission itself.  Instead, we forward it along
    * and wait for a <code>RetransStart</code>.
    * @param rr The <code>RetransRequest</code> message that just entered 
    *           our window.
    * @return Doesn't ever return.
    * @exception FMG.Conductor.dataflow.AsyncEventException 
    *                       Will always be thrown to alert the user to a 
    *                       change in the data stream.
    */
   public boolean handleEventEntry(RetransRequest rr) 
                                  throws AsyncEventException {
      super.handleEventEntry(rr);

      if (Debug.on) {
         Debug.println(Debug.ADAPTOR, Debug.INFO, 
                       adaptorName + ": Received " + rr);
      }

      rr.proceed();
      rr.handleExit(this);
      rr.unlink();
      upstreamBoundary.putOneOut(rr);

      // wait for RetransStart
      StreamElem rs = waitForMessage(Direction.up, RetransStart.class);
      rs.handleEntry(this);

      // the above call to handleEntry should result in an exception
      throw new RuntimeException("This shouldn't happen");
   }

   /**
    * Handle the case in which we receive the start of retransmission from
    * upstream.  Let it keep going until it finds the point of the 
    * retransmission request.
    * @param rs The <code>RetransStart</code> message that just entered 
    *           our window.
    * @return Doesn't ever return.
    * @exception FMG.Conductor.dataflow.AsyncEventException 
    *                       Will always be thrown to alert the user to a 
    *                       change in the data stream.
    */
   public boolean handleEventEntry(RetransStart rs) 
                                                  throws AsyncEventException {
      if (Debug.on) {
         Debug.println(Debug.ADAPTOR, Debug.INFO, 
                       adaptorName + ": Received " + rs);
      }

      inhibitTypeRecovery = false;
      inhibitDataRecovery = false;

      rs.proceed();

      // if it didn't self destruct, it has further to go, so push it out
      if (downstreamBoundary.next(Direction.up) == rs) {
         rs.unlink();
         downstreamBoundary.putOneOut(rs);
      }

      // service anything we've been neglecting
      needsService(upstreamBoundary);
      needsService(downstreamBoundary);

      throw new RollbackOccurredException();
   }

   /**
    * Handle the case in which a SegmentAck message has just entered the
    * window.  The fact that this ack has been received must be recorded
    * so the adaptor can free any associated state.  The ack is pushed out
    * the other side of the window.
    * @param ack The SegmentAck message that just entered the window.
    * @return Always false, indicating that no further message handling is 
    *         required.
    */
   public boolean handleEventEntry(SegmentAck ack) {
      ack.unlink();
/*
 * TODO: should we add a callback mechanism here?  Ie, call an adaptor 
 *       function and let it do the local handling and push it out the other
 *       side.  Or should wo push it out and then call an adaptor handler 
 *       function?
 */
      upstreamBoundary.putOneOut(ack);
      return false;
   }

   /**
    * Handle the case in which an EOF message has just entered the window.
    * Notice that this can occur from either direction.  The direction from
    * which the EOF message came is marked as closed.  An exception will be
    * thrown to the Adaptor.
    * @param eof The EOF message that entered the window.
    * @return Always true indicating that more handling is required.  
    *         In particular, an exception will be thrown to the adaptor.
    * @exception FMG.Conductor.dataflow.AsyncEventException 
    *                       Will always be thrown to alert the user to a 
    *                       stream closure.
    */
   public boolean handleEventEntry(EOF eof) throws AsyncEventException {
      ClosureException e;
      if (eof.flowDir().isUp()) {
         close(Direction.down);
         e = new OutputClosedException();
      } else {
         close(Direction.up);
         e = new InputClosedException();
      }

      eof.unlink();
      eof.recycle();
      throw e;
   }

   /**
    * Handle the case in which a Shutdown message enters the window.  This
    * should only happen after both sides of data flow have closed.  It
    * indicates that both endpoints are aware of the closure, and no more
    * messages should flow on the stream.
    * @param sd The Shutdown message that was received.
    * @return Always true indicating that further handling is required.
    */
   public boolean handleEventEntry(Shutdown sd) {
      sd.unlink();
      if (sd.flowDir().isUp()) {
         upstreamBoundary.putOneOut(sd);
      } else {
         downstreamBoundary.putOneOut(sd);
      }
      return true;
   }

   /**
    * Handle the case in which a BufferingControl message enters the window.  
    * <p>
    * TODO: Someday we should have a callback to the adaptor.  For now,
    *       just pass it along
    * @param sd The Shutdown message that was received.
    * @return Always false indicating that no further handling is required.
    */
   public boolean handleEventEntry(BufferingControl bc) {
      bc.unlink();
      bc.handleExit(this);
      downstreamBoundary.putOneOut(bc);
      return false;
   }

   public boolean handleEventExit(SwitchPlan sp) {
      super.handleEventExit(sp);
      sp.copyTypeStack(downstreamTypeStack);
      return true;
   }

   /**
    * Check to see if there is an event upstream of the window that is 
    * trying to flow into the window.  If so, move it in and handle it.
    * @exception FMG.Conductor.dataflow.AsyncEventException is thrown
    *                   if an event enters the window and is not
    *                   completely handled within this call.
    * @return No return value.
    */
   protected void checkForDownwardEvent() throws AsyncEventException {
      StreamElem se = upstreamBoundary.getAnEvent();

      if (se != null) {
         se.handleEntry(this);
      }
   }

   /**
    * Set the amount of buffering allowed downstream of this adaptor.
    * Normally this should only be called by the last adaptor on a node.
    * @param level The requested level of buffering.  This should be 
    *              MAX_BUFFERING, MIN_BUFFERING, or DEFAULT_BUFFERING.
    * @return No return value.
    */
   public void requestDownstreamBuffering(int level) {
      BufferingControl bc = BufferingControl.getNewMessage(level);
      bc.handleExit(this);
      downstreamBoundary.putOneOut(bc);
   }

   /**
    * Is there data available upstream?  Warning: this method can return 
    * true if an empty DataFragment is upstream.  Thus, a call to expand() 
    * may still block.  Consider using expand(0) instead.
    */
   public boolean dataAvailable() {
      return upstreamBoundary.dataAvailable();
   }

   /**
    * Wait for a byte to become available upstream and then move all
    * available bytes into the window.
    * @exception FMG.Conductor.dataflow.AsyncEventException is thrown
    *                   if an event enters the window and is not
    *                   completely handled within this call.
    * @return The number of bytes moved into the window.
    */
   public int expand() throws AsyncEventException {
      return expand(1);
   }

   /**
    * Wait for some number of bytes to become available upstream and then 
    * move all available bytes into the window.  To avoid blocking, pass
    * 0 as the number of bytes desired.
    * @exception FMG.Conductor.dataflow.AsyncEventException is thrown
    *                   if an event enters the window and is not
    *                   completely handled within this call.
    * @return The number of bytes moved into the window.
    */
   public int expand(int bytesDesired) throws AsyncEventException {
      int bytesAdded = 0;

      // TODO: This happens if we've previously called expand and discovered 
      //       that the input is closed.  Is this the right thing to do?
      if (upstreamClosed) {
         throw new InputClosedException();
      }

      while (true) {
         if (downNeedsService) {
            checkForUpwardEvent();
         }

         if (upNeedsService) {
            /* bytes available? */
            bytesAdded += upstreamBoundary.movePastBytes();
         }

         /* 
          * We're done if we have at least the requested number of bytes.
          *
          * But, if the user requested 0 (or less) bytes and no bytes
          * were available, check for downward events before returning.
          */
         if ((bytesAdded >= bytesDesired) &&
             ((bytesDesired > 0) || (bytesAdded > 0))) {
            return bytesAdded;
         }

         if (upNeedsService) {
            checkForDownwardEvent();
         }

         if (bytesDesired <= 0) {
            return bytesAdded;  // will this ever be non-zero?
         }

         waitForServiceRequest();
      }
   }

   /**
    * Wait for a particular character to appear upstream and then
    * move all available bytes into the window.  This function just
    * casts the character to a byte and calls <code>expand(byte)</code>, 
    * but it is useful to prevent code like <code>expand('a')</code> from 
    * calling <code>expand(int)</code> inadvertently.
    * @param targetChar The character to wait for before returning.
    * @exception FMG.Conductor.dataflow.AsyncEventException is thrown
    *                   if an event enters the window and is not
    *                   completely handled within this call.
    * @return The number of bytes moved into the window.
    */
   public final int expand(char targetChar) throws AsyncEventException {
      return expand((byte) targetChar);
   }

   /**
    * Wait for a particular byte to appear upstream and then 
    * move all available bytes into the window.
    * @param targetByte The byte to wait for before returning.
    * @exception FMG.Conductor.dataflow.AsyncEventException is thrown
    *                   if an event enters the window and is not
    *                   completely handled within this call.
    * @return The number of bytes moved into the window.
    */
   public int expand(byte targetByte) throws AsyncEventException {
      int bytesAdded = 0;
      StreamElem seen;

      // TODO: This happens if we've previously called expand and discovered 
      //       that the input is closed.  Is this the right thing to do?
      if (upstreamClosed) {
         throw new InputClosedException();
      }

      seen = upstreamBoundary.next(Direction.down);

      while (true) {
         if (downNeedsService) {
            checkForUpwardEvent();
         }

         if (upNeedsService) {
            /* bytes available? */
            bytesAdded += upstreamBoundary.movePastBytes();
         }

         /* Look at newly encorporated bytes.  Are we done? */
         while (seen.next(Direction.up)!=upstreamBoundary) {
            seen = seen.next(Direction.up);
            if (seen.containsByte(targetByte)) {
               return bytesAdded;
            }
         }

         if (upNeedsService) {
            checkForDownwardEvent();
         }

         waitForServiceRequest();
      }
   }

   /**
    * Move the SwitchPlan message out of the window.  It must be the
    * case that the SwitchPlan message is immediately adjacent to the 
    * downstream boundary.  In other words, all data in the window prior
    * to receiving the exception must have already been contracted out
    * of the window.
    * @return No return value.
    */
   public void advanceSwitchPlan() {
      if (downstreamBoundary.next(Direction.up) instanceof SwitchPlan) {
         SwitchPlan sp = (SwitchPlan) downstreamBoundary.next(Direction.up);
         
         /*
          * Adjust the type stack according to this (previously deployed) 
          * adaptor's type.
          */
         if (compositionType.isPush()) {
            sp.typeStackPush(myType);
         } else if (compositionType.isPop()) {
            sp.typeStackPop();
         }

         sp.handleExit(this);
         downstreamBoundary.swapEventOut();
      } else {
         throw new RuntimeException("attempted advanceSwitchPlan on " + 
                                    downstreamBoundary.next(Direction.up));
      }
   }

   /**
    * Contract out the entire contents of the window.
    * @return The number of bytes contracted out.
    */
   public int contract() {
      return contract(upstreamBoundary);
   }

   /**
    * Contracte out all data downstream of the given pointer.  After this
    * operation, the pointer will be inside the window, adjacent to the 
    * donwstream boundary.
    * @param p The pointer that indicates where to contract to.
    * @return The number of bytes contracted out.
    */
   public int contract(Pointer p) {
// TODO: if the downstream is closed, should we just throw away the data?
      return downstreamBoundary.moveToPointer(p);
   }

   /**
    * Contract the given number of bytes out of the window.
    * @param n The number of bytes to push out of the window.
    * @return The number of bytes actually contracted (which could be les than 
    *         <code>n</code>.
    */
   public int contract(int n) {
// TODO: if the downstream is closed, should we just throw away the data?
      return downstreamBoundary.advance(n);
   }

   /**
    * Insert a pointer into the window.  The pointer will be put immediately
    * upstream of the downstream boundary.
    * @param p The pointer to insert.
    * @return No return value.
    */
   public void insertPointer(Pointer p) {
      downstreamBoundary.insertPointer(p);
   }

   /**
    * Determine where the boundary is located in the window.
    * @return The number of data bytes in the window downstream of the 
    *         pointer.
    */
   public int getPosition(Pointer p) {
      return p.distanceTo(downstreamBoundary, Direction.down);
   }

   /**
    * Determine the distance from the pointer to the begining of the window.
    * @return The number of data bytes in the window upstream of the pointer.
    */
   public int countRemaining(Pointer p) {
      return p.distanceTo(upstreamBoundary, Direction.up);
   }

   /**
    * Determine the size (in bytes) of the window.
    * @return The number of bytes in the window.
    */
   public int size() {
      return upstreamBoundary.distanceTo(downstreamBoundary, Direction.down);
   }

   /**
    * Determine if the upstream (input) end of the window is closed.
    * @return True iff the input endpoint is closed.
    */
   public boolean inputClosed() {
      return upstreamClosed;
   }

   /**
    * Determine if the downstream (output) end of the window is closed.
    * @return True iff the output endpoint is closed.
    */
   public boolean outputClosed() {
      return downstreamClosed;
   }

   /**
    * Determine if a SwitchPlan message is in the window.
    * @return True iff a SwitchPlan message is present in the window.
    */
   public boolean switchPlanInWindow() {
      for (StreamElem se = downstreamBoundary; 
           se != upstreamBoundary; 
           se = se.next(Direction.up)) {
         if (se instanceof SwitchPlan) {
            return true;
         }
      }

      return false;
   }

   /**
    * Close the input endpoint.  After this call, the window can no longer
    * be expanded.
    * @return No return value.
    */
   public void closeInput() {
      /* create EOF and push it up */
      close(Direction.up);
      EOF eof = EOF.getNewMessage(Direction.up);

      upstreamBoundary.insert(eof, Direction.down);
      upstreamBoundary.swapEventOut();
   }

   /**
    * Close the output endpoint.  After this call, the window can no longer
    * be contracted.
    * @return No return value.
    */
   public void closeOutput() {
      /* create an EOF and contract it out */
      EOF eof = EOF.getNewMessage(Direction.neutral);

      downstreamBoundary.insert(eof, Direction.up);
      eof.handleExit(this);
      downstreamBoundary.swapEventOut();
   }

   /**
    * This method is called after an adaptor's main code returns.  This
    * window is removed from the stream.
    * @return No return value.
    */
   void removeSelf(boolean failureOccurred) {
// TODO: are these the right conditions?
      if (downstreamClosed && ! upstreamClosed) {
         closeInput();
      }

      if ((upstreamBoundary.next(Direction.down) != downstreamBoundary) ||
          (! upstreamBoundary.isBetweenSegments()) ||
          (! downstreamBoundary.isBetweenSegments()) ||
          failureOccurred) {
         /* window is not empty ... clean up! */
         if (Debug.on) {
            Debug.println(Debug.ADAPTOR, Debug.WARNING, 
                          "WARNING: Adaptor " + adaptorName + 
                          " left non-empty window.");
         }

         if (! downstreamClosed) {
            if (INITIATE_RECOVERY && Conductor.reliabilityEnabled()) {
               doInitiateRecovery();
            } else {
               /* push out data */
               contract();
            }
         }

      }

      /* destroy anything that's left */
      for (StreamElem se = upstreamBoundary.next(Direction.down);
           !se.isBoundary();
           se = upstreamBoundary.next(Direction.down)) {
         se.unlink();
         if (se instanceof SwitchPlan) {
            downstreamBoundary.putOneOut(se);
         } else {
            se.recycle();
         }
      }

      if (upstreamBoundary.next(Direction.down) != downstreamBoundary) {
         throw new RuntimeException("this shouldn't happen");
      }

      /*
       * Before doing a merge, make sure that our segmentation is erased.
       * otherwise messages that should be passed from one side to each
       * other may behave badly (in particular, RetransStart).
       */
      upstreamBoundary.eraseSegmentation();
      downstreamBoundary.eraseSegmentation();

      /* now merge away the boundaries */
      upstreamBoundary.merge(downstreamBoundary);
   }

   /**
    * Start the recovery algorithm in the case that this associated 
    * adaptor has failed and data loss was possible.
    */
   public void doInitiateRecovery() {
      Debug.println("**** " + this + ": Initiating recovery!!!!!");

      if (! inhibitDataRecovery) {
         // prevent downstream data flow until retransmission starts
         if (Debug.on) {
            Debug.println(Debug.RELIABILITY, Debug.INFO, 
                          "Installing data sink");
         }
         downstreamBoundary.putOneOut(
                  new DataSink(downstreamBoundary.isBetweenSegments(), 
                               downstreamBoundary.lastSegmentID()));
      }

      if ((! inhibitTypeRecovery) && compositionType.isPush()) {
         // send retyping message downstream
         RebalanceTypes rt = 
             RebalanceTypes.getNewMessage(Direction.down, upstreamTypeStack);
         if (Debug.on) {
            Debug.println(Debug.RELIABILITY, Debug.INFO, 
                          this + " sending " + rt);
         }
         rt.handleExit(this);
         downstreamBoundary.putOneOut(rt);
      } else {
         if ((! inhibitTypeRecovery) && compositionType.isPop()) {
            // send retyping message upstream
            RebalanceTypes rt = 
               RebalanceTypes.getNewMessage(Direction.up, downstreamTypeStack);
            if (Debug.on) {
               Debug.println(Debug.RELIABILITY, Debug.INFO, 
                             this + " sending " + rt);
            }
            rt.handleExit(this);
            upstreamBoundary.putOneOut(rt);
         }

         if (! inhibitDataRecovery) {

            if (downstreamBoundary.isBetweenSegments()) {
               RetransRequest rr = RetransRequest.getNewMessage(
                                           downstreamBoundary.lastSegmentID());
               if (Debug.on) { 
                  Debug.println(Debug.RELIABILITY, Debug.INFO, 
                                this + " sending " + rr);
               }
               rr.handleExit(this);
               upstreamBoundary.putOneOut(rr);
            } else {
               SignalFailure sf = SignalFailure.getNewMessage();
               if (Debug.on) { 
                  Debug.println(Debug.RELIABILITY, Debug.INFO, 
                                this + " sending " + sf);
               }
               downstreamBoundary.putOneOut(sf);
            }
         }
      }
   }

   public void setTypeStack(StreamTypeStack ts, Adaptor.CompositionType ct) {
      compositionType = ct;
      downstreamTypeStack.copy(ts);
      upstreamTypeStack.copy(ts);

      if (ct.isPush()) {
         myType = upstreamTypeStack.pop();
      } else if (ct.isPop()) {
         myType = downstreamTypeStack.pop();
      }
   }

   /**
    * Obtain a string containing a human-readable representation of this 
    * window and the portion of the stream it contains.  This method should
    * be used for deubgging only.
    * @return A new string containing a human-readable representation of 
    *         this window.
    */
   public String toString() {
      StringBuffer sb = new StringBuffer();

      sb.append("AdaptorWindow[");
      sb.append(adaptorName);
      sb.append("]");
      if (Debug.on && (Debug.getLevel(Debug.DATAFLOW) > Debug.DETAIL)) {
         sb.append("{");
         for (StreamElem se = downstreamBoundary; 
              (se != upstreamBoundary); 
              se = se.next(Direction.up)) {
            sb.append(se.toString());
            sb.append(", ");
         }
         sb.append(upstreamBoundary.toString());
         sb.append("}");
      }
      return sb.toString();
   }
}
