/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.Conductor.dataflow;

import FMG.Conductor.dataflow.messages.DataFragment;
import FMG.Conductor.util.Direction;

/**
 * The BlockModPointer is a special type of DataAccessPointer that allows
 * modifications to be made that semantically extend across multiple 
 * operations.  Among other things, this allows modifications to extend
 * beyond what might otherwise be placed in a window at one time.  Also,
 * this allows an adaptor to describe semantically contiguous sets of
 * operations, which preserves proper semantic segmentation without requiring
 * explicit adaptor knowledge of segmentation.
 * <p>
 * The BlockModPointer has two modes.  When in normal mode, this object
 * behaves like a DataAccessPointer.  When in block modification mode,
 * all modifications and movement lead to data that is in one large
 * segment.
 * <p>
 * An adaptor should create and insert the BlockModPointer in the same way
 * as a DataAccessPointer.  Before performing multiple operations that
 * are semantically related, the adaptor should call 
 * <code>blockModStart()</code> all data that is changed, and subsequently
 * moved past the pointer are then combined into the same segment.
 * Once the related operations are complete and the pointer has moved past
 * the modified data, the adaptoer should call <code>blockModEnd()</code>.
 * <p>
 * TODO: deal with segment numbering issues
 *
 * @author Mark Yarvis (yarvis@fmg.cs.ucla.edu)
 * @version $Revision: 1.10 $
 */

public class BlockModPointer extends DataAccessPointer {

   /** Are we currently performing a block modification. */
   boolean blockModMode = false;

   /** Did we just pass the end of the segment? */
   boolean atEOS = true;

   /** The current segment ID, when in block mod mode. */
   int segmentId;

   /**
    * Start a block modification operation.  This method must not be called
    * when already in block mod mode.
    * @return No return value.
    */
   public void blockModStart() {
      if (blockModMode) {
         throw new RuntimeException("Operation not permitted in block mode");
      }
      blockModMode = true;
      atEOS = true;
   }

   /**
    * End a block modification operation.  This method can only be called 
    * while in block mod mode.  The current segment is finished off and
    * numbered appropriately.
    * @return No return value.
    */
   public void blockModEnd() {
      if (!blockModMode) {
         throw new RuntimeException("Operation only permitted in block mode");
      }

      if (atEOS) {
         DataFragment newFragment = DataFragment.getNewMessage(null, 0, 0);
         newFragment.makeBlockEnd(segmentId);
         insert(newFragment, Direction.down);
// TODO: is this always OK?
         newFragment.proceed();
      }

      blockModMode = false;
   }

   /**
    * Insert a set of bytes downstream of the pointer.
    * @param bytes The array containing the bytes to insert.
    * @param sourceStart The frist byte to copy from the source array.
    * @param sourceLen The number of bytes to insert.
    * @return No return value.
    */
   public void insertBytes(byte[] bytes, int sourceStart, int sourceLen) {
      if (!blockModMode) {
         throw new RuntimeException("Operation only permitted in block mode");
      }

      while (sourceLen > 0) {
         DataFragment newFragment = DataFragment.getNewMessage(bytes,
                                                       sourceStart, sourceLen);
         newFragment.makeBlock();
         sourceLen -= newFragment.length();
         sourceStart += newFragment.length();
         insert(newFragment, Direction.down);
      }
   }

   /**
    * Delete a set of bytes prior to (upstream of) the pointer.
    * @param len The numbe of bytes to delete
    * @return No return value.
    */
   public void deleteBytes(int len) {
      if (len == 0) {
         return;
      }

      // do the operation and find out what happened
      super.replaceBytes(null, 0, 0, 0, len);

      if (blockModMode) {
         DataFragment deleted = (DataFragment) next(Direction.up);

         // record where we ended up
         atEOS = deleted.isEndOfSegment();
         segmentId = deleted.segmentID();

         // no longer need this guy
         deleted.unlink();
         deleted.recycle();
      }
   }

   /**
    * Replace the specified portion of the stream with a set of
    * bytes from a byte array.  The pointer will move past the
    * replaced bytes.  This method should not be used for deletion (ie
    * sourceLen == 0).
    * @param bytes The source array.
    * @param sourceStart The index of the first byte to copy from the
    *                    source array.
    * @param sourceLen The number of bytes to copy from the source array.
    * @param targetStart The position in the stream, relative to the pointer,
    *                    to put the first byte
    * @param len The number of bytes to replace in the stream.
    * @return The data stream size change in bytes.
    */
   public int replaceBytes(byte[] bytes, int sourceStart, int sourceLen, 
                            int targetStart, int targetLen) {
      if ((sourceLen == 0) && blockModMode) {
         throw new RuntimeException("should call deleteBytes instead");
      } else {
         return super.replaceBytes(bytes, sourceStart, sourceLen, 
                                   targetStart, targetLen);
      }
   }

   /**
    * Replace the specified portion of the stream with a set of
    * bytes from a byte array.  The pointer will move past the replaced
    * bytes.  This method should not be used for deletion 
    * (ie sourceLen == 0).
    * @param bytes The source array.
    * @param sourceStart The index of the first byte to copy from the
    *                    source array.
    * @param sourceLen The number of bytes to copy from the source array.
    * @param postP A pointer that specifies the end of the portion to
    *              replace.
    * @return The data stream size change in bytes.
    */
   public int replaceBytes(byte[] bytes, int sourceStart, int sourceLen,
                           DataAccessPointer postP) {
      if ((sourceLen == 0) && blockModMode) {
         throw new RuntimeException("should call deleteBytes instead");
      } else {
         return super.replaceBytes(bytes, sourceStart, sourceLen, postP);
      }
   }

   /**
    * Advance this pointer until it passes a given number of bytes.  Notice
    * that the bytes past will be combined into the current segment if block
    * modification mode is on.
    * @param distance The number of bytes to advance past.
    * @return No return value.
    */
   public int advance(int distance) {
      if (blockModMode) {
         // remember the boundary of what was already downstream
         StreamElem oldNext = next(Direction.down);

         // advance the requested distance
         int distMoved = super.advance(distance);

// TODO: what if segment combination occurs?  oldNext will still be there

         // get the last segment Id
         if (oldNext != next(Direction.down)) {
            DataFragment df = (DataFragment) next(Direction.down);
            atEOS = df.isEndOfSegment();
            segmentId = df.segmentID();
         }

         // mark the segments that passed by as a block
         for (StreamElem se = next(Direction.down); 
              se != oldNext; se = se.next(Direction.down)) {
            if (se instanceof DataFragment) {
               ((DataFragment) se).makeBlock();
            }
         }
         return distMoved;
      } else {
         return super.advance(distance);
      }
   }

   public int handleNewAdjacent(Direction dir, StreamElem se) {
      /**
       * The super call is going to allow this message to pass by, 
       * so, we have to keep track of the segmentation stuff.
       */
      if (blockModMode && dir.isUp() && se.isData() && (se.length() == 0)) {
         DataFragment df = (DataFragment) se;
         atEOS = df.isEndOfSegment();
         segmentId = df.segmentID();
      }
      return super.handleNewAdjacent(dir, se);
   }

// really should protect users from doing this in block mode
   public void removeSelf() {
      if (blockModMode) {
         throw new RuntimeException("Operation not permitted in block mode");
      }

      super.removeSelf();
   }

}
