/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.Conductor.dataflow;

import FMG.Conductor.util.Direction;
import FMG.Conductor.dataflow.messages.DataFragment;

/**
 * The Boundary class provides a StreamElement that blocks other
 * StreamElments from moving past.
 *
 * @author Mark Yarvis (yarvis@fmg.cs.ucla.edu)
 * @version $Revision: 1.16 $
 */

public abstract class Boundary extends Pointer {

   /** Is this boundary in between two distinct segments? */
   protected boolean amBetweenSegments = true;

   /** The ID of the last complete segment to pass by. */
   protected int lastSegmentID = -1;

   /**
    * Create a new boundary object.
    */
   protected Boundary() {
   }

   /**
    * Record the segmentation of the given data fragment.  This method should
    * be called when the given data fragment is passing by.  This boundary
    * wants to remember the ID of the last full segment that passed and whether
    * the last fragment that passwd was the end of a segment.
    * @param df The data fragment in question.
    * @return No return value.
    */
   protected final void recordSegmentation(DataFragment df) {
      amBetweenSegments = df.isEndOfSegment();
      if (amBetweenSegments) {
         lastSegmentID = df.segmentID();
      }
   }

   protected final void eraseSegmentation() {
      amBetweenSegments = true;
      lastSegmentID = -2;
   }

   /**
    * Is this boundary in between two segments?  In other words, a fragment
    * ending the segment was the last thing to pass by.
    * @return True if this boundary is between distinct segments, false 
    *         otherwise.
    */
   public boolean isBetweenSegments() {
      return amBetweenSegments;
   }

   /**
    * What was the id of the last complete segment to pass?
    * @return The id of the last complete segment that passed this boundary.
    */
   public int lastSegmentID() {
      return lastSegmentID;
   }

   /**
    * Determine if this is a boundary.
    * @return Always return true.
    */
   public boolean isBoundary() {
      return true;
   }

   /**
    * Determine the flow direction of this object.
    * @return Alwasy returns <code>Direction.up</code>.
    */
   public Direction flowDir() {
      return Direction.up;
   }

}
