/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.Conductor.dataflow;

import java.util.Vector;

import FMG.Conductor.util.Debug;
import FMG.Conductor.util.Direction;
import FMG.Conductor.dataflow.messages.*;

public class Cache extends Boundary {
   Vector cache = new Vector();

   public boolean isBoundary() {
      return true;
   }

   public Direction flowDir() {
      return Direction.up;
   }

   protected void cacheData(DataFragment df) {
      df = df.duplicate();
      cache.addElement(df);
   }

   protected void handleAck(int ackedID) {
      while (true) {
         if (cache.isEmpty()) {
            return;
         }

         DataFragment df = (DataFragment) cache.firstElement();

         if (df.segmentID() <= ackedID) {
            df.recycle();
            cache.removeElementAt(0);
         } else {
            return;
         }
      }
   }

   public int handleNewAdjacent(Direction dir, StreamElem se) {
      if (dir.isUp()) {
         if (se instanceof DataFragment) {
            DataFragment df = (DataFragment) se;
            cacheData(df);
            recordSegmentation(df);
         }

      } else { /* dir is down */

         // handle ack
         if (se instanceof SegmentAck) {
            handleAck(((SegmentAck) se).segmentID);

         // handle retransmission request
         } else if (se instanceof RetransRequest) {
            // pull out retrans request and recycle it (save segment id)
            int lastReceivedSegmentID = 
                        ((RetransRequest) se).lastReceivedSegmentID();
            se.unlink();
            se.recycle();

            // forward 
            return retransmitData(lastReceivedSegmentID);
         }
      }

      if (se.isBoundary()) {
         return 0;
      }

      swap(dir);
      return se.proceed() + next(dir).proceed();
   }

   protected int retransmitData(int lastReceivedSegmentID) {
      int adjustment = 0;

      if (Debug.on) {
         Debug.println(Debug.DATAFLOW, Debug.INFO, 
                       this + ": Retransmission starting after segment " + 
                       lastReceivedSegmentID);
      }

      RetransStart rs = RetransStart.getNewMessage(lastReceivedSegmentID);
      if (Debug.on) {
         Debug.println(Debug.RELIABILITY, Debug.INFO, 
                       this + " sending " + rs);
      }
      insert(rs, Direction.down);
      adjustment += rs.proceed();

      int index = 0;
      DataFragment df = null;

      while (index < cache.size()) {
         df = (DataFragment) cache.elementAt(index);
         if (df.segmentID() > lastReceivedSegmentID) {
            break;
         }
         index++;
      }

      // no data needed to be retransmitted
      if (index == cache.size()) {
         return adjustment;
      }

      df = df.duplicate();   // duplicate for retransmission
      if (!df.isBlock() && 
          (df.segmentID() - df.length() < lastReceivedSegmentID)) {
         // split of the unnecessary part of the segment and throw it away
         DataFragment tail = df.split(df.length() - df.segmentID() + 
                                      lastReceivedSegmentID);
         df.unlink();
         df.recycle();
         df = tail;
      }

      if (Debug.on) {
         Debug.println(Debug.DATAFLOW, Debug.DETAIL,
                       this + ": Retransmitting " + df);
      }
      insert(df, Direction.down);
      adjustment++;

      while (++index < cache.size()) {
         df = ((DataFragment) cache.elementAt(index)).duplicate();
         if (Debug.on) {
            Debug.println(Debug.DATAFLOW, Debug.DETAIL,
                          this + ": Retransmitting " + df);
         }
         insert(df, Direction.down);
         adjustment++;
      }

      if (Debug.on) {
         Debug.println(Debug.DATAFLOW, Debug.INFO, 
                       this + ": Retransmitted " + adjustment + 
                              " DataFragment objects");
      }

      return adjustment;
   }
}
