/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.Conductor.dataflow;

import FMG.Conductor.dataflow.messages.DataFragment;
import FMG.Conductor.dataflow.messages.InterfaceControl;
import FMG.Conductor.util.*;

/**
 * The DataAccessPointer is a user-level pointer that helps an adaptor
 * operate on the data stream.  The DataAccessPointer points to a
 * position within the window.  A set of operations for moving the
 * pointer are provided.  The user can then read and modify
 * the data stream by providing an index relative to the pointer.
 * <p>
 * To use this object, an adaptor should insert it into the window
 * using AdaptorWindow.insertPointer()
 * <p>
 * Many of the operations provided affect the segmentation of the data stream.
 * The rules of segmentation are automatically preserved by these operators.
 * The adaptor need not participate in segmentation at this level.
 *
 * @author Mark Yarvis (yarvis@fmg.cs.ucla.edu)
 * @version $Revision: 1.21 $
 *
 * @see FMG.Conductor.dataflow.AdaptorWindow#insertPointer
 */

public class DataAccessPointer extends Pointer {

   /**
    * Obtain a byte from the stream at a position relative to the pointer.
    * Attempting to access a byte outside the window will result in a
    * runtime exception.
    * @param targetPos The relative position of the desired byte.
    * @return True iff the byte was found and the pointer moved.
    */
   public byte getByte(int targetPos) {
      StreamElem se = next(Direction.up);
      int pos = 0;

      while (!se.isBoundary()) {
         if (se instanceof DataFragment) {
            int len = se.length();
            if (pos + len > targetPos) {
               return ((DataFragment) se).getByte(targetPos - pos);
            }
            pos += len;
         }
         se = se.next(Direction.up);
      }
      throw new RuntimeException("getByte(): accessed data outside of window");
   }

   /**
    * Move the pointer to the first occurrance of the given byte.
    * If the byte is not in the stream, the pointer will not move
    * and false will be returned.
    * @param b The byte to move to.
    * @return The distance the pointer was moved in bytes, or -1 if the 
    *         target byte is not present in the window.
    */
   public int moveToByte(byte b) {
      int dist = distanceTo(b);
      if (dist == -1) {
         return -1;
      } else {
         advance(dist);
         return dist;
      }
   }

   /**
    * Move the pointer to the character after the first occurrance of 
    * the given byte.
    * If the byte is not in the stream, the pointer will not move
    * and false will be returned.
    * @param b The byte to move past.
    * @return The distance the pointer was moved in bytes, or -1 if the 
    *         target byte is not present in the window.
    */
   public int movePastByte(byte b) {
      int len = moveToByte(b);
      if (len == -1) {
         return -1;
      } else {
         advance(1);  // this will work or the byte would not have been found
         return len + 1;
      }
   }

   /**
    * Replace the byte at the given position relative to the position of the
    * pointer.
    * @param targetPos The position, relative to the pointer, to effect change.
    * @param b The new value to record at the specified location.
    */
   public void replaceByte(int targetPos, byte b) {
      StreamElem se = next(Direction.up);
      int pos = 0;

      while (!se.isBoundary()) {
         int len = se.length();
         if (pos + len > targetPos) {
            ((DataFragment) se).setByte(targetPos - pos, b);
            return;
         }
         pos += len;
         se = se.next(Direction.up);
      }
      throw new RuntimeException("getByte(): accessed data outside of window");
   }

   /**
    * Copy the specified bytes into a byte array.
    * @param bytes The target array.
    * @param targetPos The position in the array where the first byte 
    *                  should be copied.
    * @param start The position in the stream, relative to the pointer, 
    *              of the first byte to copy
    * @param len The number of bytes to copy.
    * @return No return value.
    */
   public void getBytes(byte[] bytes, int targetPos, int start, int len) {
      StreamElem se = next(Direction.up);
      int bytesCopied = 0;

      if (len <= 0) {
         if (len == 0) {
            return;
         }
         throw new RuntimeException("Bad argumenet.");
      }

      while (!se.isBoundary()) {
         if (se instanceof DataFragment) {
            DataFragment df = (DataFragment) se;
            int segLen = df.length();

            if (segLen > start) {  /* found start? */
               /* get available bytes */
               int count = Math.min(len - bytesCopied, segLen - start);
               df.getBytes(bytes, targetPos, start, count);
               bytesCopied += count;
               targetPos += count;
               start = 0;   /* subsequent copies start from first byte */

               if (bytesCopied >= len) {
                  if (bytesCopied > len) {
                     throw new RuntimeException("this shouldn't happen");
                  }
                  return;
               }

            } else {   /* still looking for start */
               start -= segLen;
            }
         }
         se = se.next(Direction.up);
      }
      throw new RuntimeException("getByte(): accessed data outside of window");
   }

   /**
    * Replace the specified portion of the stream with a set of bytes
    * from a byte array.
    * @param bytes The source array.
    * @param sourceStart The index of the first byte to copy from the 
    *                    source array.
    * @param sourceLen The number of bytes to copy from the source array.
    * @param targetStart The position in the stream, relative to the pointer, 
    *                    to put the first byte
    * @param len The number of bytes to replace in the stream.
    * @return The net change in stream size in bytes.
    */
   public int replaceBytes(byte[] bytes, int sourceStart, int sourceLen, 
                           int targetStart, int targetLen) {
      StreamElem se;
      DataFragment first = null;  /* segment starting with first byte of data 
                                    to replace */
      StreamElem last = null;   /* segment starting after last byte of data 
                                    to replace */

      if ((sourceLen == 0) && (targetLen == 0)) {
         return 0;
      }

      if ((sourceLen < 0) || (targetLen < 0)) {
         throw new RuntimeException("Bad argument");
      }

      if (targetLen == 0) {
         throw new RuntimeException("Insertion is now allowed");
      }

      /* find first segment */
      for (se = next(Direction.up); (!se.isBoundary()); 
                                              se = se.next(Direction.up)) {
         if (se instanceof DataFragment) {
            DataFragment df = (DataFragment) se;
            int segLen = df.length();

            if (segLen <= targetStart) {
               /* not there yet */
               targetStart -= segLen;
            } else {
               if (targetStart != 0) {
                  /* split the segment if we didn't fall at the start of one */
                  first = df.split(targetStart);
               } else {
                  first = df;
               }
               break;
            }
         }
      }

      if (first == null) {
         throw new RuntimeException("This shouldn't happen");
      }

      /* find last segment (actually segment after last) */
      for (se = first; !se.isBoundary(); se = se.next(Direction.up)) {
         if (se instanceof DataFragment) {
            DataFragment df = (DataFragment) se;
            int segLen = df.length();

            if (segLen < targetLen) {
               /* not there yet */
               targetLen -= segLen;
            } else {
               if (targetLen != segLen) {
                  /* split the segment if we didn't fall at the start of one */
                  last = df.split(targetLen);
               } else {
                  last = df.next(Direction.up);
               }
               break;
            }
         }
      }

      if (last == null) {
         last = se;
      }

      return doReplaceBytes(bytes, sourceStart, sourceLen, first, last);
   }

   /**
    * Replace the specified portion of the stream with a set of
    * bytes from a byte array.
    * @param bytes The source array.
    * @param sourceStart The index of the first byte to copy from the
    *                    source array.
    * @param sourceLen The number of bytes to copy from the source array.
    * @param postP A pointer that specifies the end of the portion to
    *              replace.
    * @return The data stream size change in bytes.
    */
   public int replaceBytes(byte[] bytes, int sourceStart, int sourceLen,
                           DataAccessPointer postP) {
      // find the first DataFragment
      StreamElem first = next(Direction.up);
      while (! (first instanceof DataFragment)) {
         first = next(Direction.up);
      }

      if (first.distanceTo(postP, Direction.up) == 0) {
         throw new RuntimeException("Insertion is not allowed");
      }

      // make sure we're next to some data
      StreamElem last = postP;
      StreamElem nextToLast = postP.next(Direction.down);
      while (! (nextToLast instanceof DataFragment)) {
         last = nextToLast;
         nextToLast = postP.next(Direction.down);
      }

      return doReplaceBytes(bytes, sourceStart, sourceLen, 
                            (DataFragment) first, last);
   }

   /**
    * Do the actual replacement work for the various flavors of 
    * ReplaceBytes.
    * @param bytes The source array.
    * @param sourceStart The index of the first byte to copy from the
    *                    source array.
    * @param sourceLen The number of bytes to copy from the source array.
    * @param first A pointer that points to the first data block to
    *              be replaced.
    * @param last A pointer that points to a part of the stream <em>after</em>
    *             the last byte to be replaced.
    * @return The data stream size change in bytes.
    */
   protected int doReplaceBytes(byte[] bytes, int sourceStart, int sourceLen, 
                                DataFragment first, StreamElem last) {
      boolean segmentIDIsSet = false;   // ensure we set the segmentID
      int segmentID = 0;
      boolean atEOS = true;
      int replacedLen = 0;

      /* remove everything between first and last */
      StreamElem se = first;
      while (se != last) {
         StreamElem nextElem = se.next(Direction.up);
         replacedLen += se.length();
         if (se instanceof DataFragment) {
            DataFragment df = (DataFragment) se;
            segmentID = df.segmentID();
            segmentIDIsSet = true;      // segmentID has been set!
            atEOS = df.isEndOfSegment();
         } else {
// TODO: do we really want to unlink if it's not a DataFragment?
            Debug.println(Debug.DATAFLOW, Debug.WARNING, 
                          "WARNING: poorly supported case in " + 
                          "DataAccessPointer.replaceBytes()");
         }
         se.unlink();
         se.recycle();
         se = nextElem;
      }
      /* Note: variable first is no longer valid here */

      /* segmentID should now be set to the last segment id */
      if (! segmentIDIsSet) {
         throw new RuntimeException("This shouldn't happen");
      }

      /* handle deletion case */
      if (sourceLen == 0) {
         DataFragment newFragment = DataFragment.getNewMessage(null, 0, 0);
         if (atEOS) {
            newFragment.makeBlockEnd(segmentID);
         } else {
            newFragment.makeBlock();
         }
         last.insert(newFragment, Direction.down);
         return (- replacedLen);
      }

      /* insert the new data */
      while (sourceLen > 0) {
         DataFragment newFragment = DataFragment.getNewMessage(bytes, 
                                                       sourceStart, sourceLen);
         sourceLen -= newFragment.length();
         if ((sourceLen == 0) && atEOS) {
            newFragment.makeBlockEnd(segmentID);
         } else {
            newFragment.makeBlock();
         }
         sourceStart += newFragment.length();
         last.insert(newFragment, Direction.down);
      }

      return (sourceLen - replacedLen);
   }

   /**
    * Insert a marker into the stream that tells the network interface to 
    * turn on.
    * @return No return value.
    */
   public void interfaceOn() {
      insert(InterfaceControl.getNewMessage(true), Direction.down);
   }

   /**
    * Insert a marker into the stream that tells the network interface to 
    * turn off.
    * @return No return value.
    */
   public void interfaceOff() {
      insert(InterfaceControl.getNewMessage(false), Direction.down);
   }

   /**
    * Determine if this is a user-level object.  Because it is, this object
    * can never be moved (even accidentally) out of the window.
    * @return Always returns true.
    */
   public boolean isUserLevel() {
      return true;
   }

   /**
    * Determine the direction of flow of this object.
    * @return Always returns Direction.neutral
    */
   public Direction flowDir() {
      return Direction.neutral;
   }

   public int advance(int dist) {
      StreamElem origDownstream = next(Direction.down);

      int ret = super.advance(dist);

      if (origDownstream != null) {
         origDownstream.handleNewAdjacent(Direction.up, 
                                          origDownstream.next(Direction.up));
      }

      return ret;
   }

   public int handleNewAdjacent(Direction dir, StreamElem se) {
      // let 0 length data blocks pass us by
      if (dir.isUp() && se.isData() && (se.length() == 0)) {
         swap(Direction.up);
         return se.proceed();
      } else {
         return super.handleNewAdjacent(dir, se);
      }
   }

   public int advance() {
      StreamElem origDownstream = next(Direction.down);

      int ret = super.advance();

      if (origDownstream != null) {
         origDownstream.handleNewAdjacent(Direction.up, 
                                          origDownstream.next(Direction.up));
      }

      return ret;
   }

   public void removeSelf() {
      // save the identities of my neighbor
      StreamElem origDownstream = next(Direction.down);

      // remove myself
      unlink();

      /*
       * Tell these guys they are now neighbors.  We only do this so that
       * DataFragments can rejoin (right?).
       */
      if (origDownstream != null) {
         origDownstream.handleNewAdjacent(Direction.up, 
                                          origDownstream.next(Direction.up));
      }
   }

}
