/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.Conductor.dataflow;

import FMG.Conductor.util.*;
import FMG.Conductor.dataflow.messages.*;

/**
 * The DataSink object is deployed after a failure and sucks up all 
 * data flowing until the start of retransmission arrives.  This object
 * is required because it is otherwise possible for the input side of
 * the ConductorIO (that is upstream from the point of failure) to forward 
 * data after a RetransRequest has passed by but before the output side can 
 * set a flag that prevents such forwarding.
 *
 * @author Mark Yarvis (yarvis@fmg.cs.ucla.edu)
 * @version $Revision: 1.2 $
 */

public class DataSink extends Boundary {

   /**
    * Create a new data sink.  Notice that the parameters indicate a point in
    * the data stream immediately downstream, which never changes since we're 
    * only throwing away data.
    * @param amBetweenSegments Was the point downstream between segments?
    * @param lastSegmentID The last segment id downstream.
    */
   public DataSink(boolean amBetweenSegments, int lastSegmentID) {
      super();

      // these don't change ... they're just set to match the creator
      this.amBetweenSegments = amBetweenSegments;
      this.lastSegmentID = lastSegmentID;
   }

   /**
    * This method is called when a new object bumps into us (which happens
    * a lot, since we're a boundary).
    * @param dir The direction of the newly adjacent object.
    * @param se The handle of the newly adjacent object.
    * @return The change in the number of data objects as a result of this
    *         operation.
    */
   public int handleNewAdjacent(Direction dir, StreamElem se) {
      if (se == null) {
         return 0;
      } else if (se.isData()) {
         se.unlink();
         se.recycle();
         return handleNewAdjacent(dir, next(dir)) - 1;
      } else if (se.isBoundary()) {
         return 0;
      } else if (se instanceof RetransStart) {
         if (Debug.on) {
            Debug.println(Debug.RELIABILITY, Debug.INFO, 
                          "Data sink going away at start of retransmission");
         }
         unlink();
         return se.proceed();
      } else {
         if ((dir.isUp() && (!se.flowDir().isUp())) || 
             (dir.isDown() && (se.flowDir().isUp()))) {
            swap(dir);
            return se.proceed() + handleNewAdjacent(dir, next(dir));
         } else {
            return 0;
         }
      }
   }
}
