/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.Conductor.dataflow;

import FMG.Conductor.util.Direction;
import FMG.Conductor.util.IntP;
import FMG.Conductor.dataflow.messages.DataFragment;
import FMG.Conductor.dataflow.messages.InterfaceControl;

/**
 * The DownstreamBoundary is intended to be one of a pair of boundaries that
 * constists of an UpstreamBoundary and a DownstreamBoundary.  Each of these
 * specialized boundaries includes particular operations that are applicable
 * to each.
 *
 * @author Mark Yarvis (yarvis@fmg.cs.ucla.edu)
 * @version $Revision: 1.20 $
 *
 * @see FMG.Conductor.dataflow.UpstreamBoundary
 */

public class DownstreamBoundary extends AccessorBoundary {

   /**
    * Create a new DownstreamBoundary with an associated UpstreamBoundary.
    * The two are linked together in a linked list.
    */
   public DownstreamBoundary() {
      super();
      lockDir = Direction.down;
      pair = new UpstreamBoundary(this);
      safeInsert(pair, Direction.down);
   }

   /**
    * Create a DownstreamBoundary that is linked to and paried with the
    * given UpstreamBoundary.  This method is normally called by the 
    * UpstreamBoundary constructor.
    * @param other The UpstreamBoundary to link this object to
    * @see FMG.Conductor.dataflow.UpstreamBoundary#UpstreamBoundary()
    */
   public DownstreamBoundary(UpstreamBoundary other) {
      super(other);
   }

   /**
    * Obtain the pair to this boundary.
    * @return The UpstreamBoundary that is paired with this object.
    */
   UpstreamBoundary pair() {
      return (UpstreamBoundary) pair;
   }

   /**
    * Used during contract operations, this method moves the boundary
    * past the given number of bytes of data.  If the window is smaller
    * than the given number of bytes, the pointer will be advanced
    * as far as possible.  Any non-data stream elements encountered will 
    * be left in the window in the order they were encountered.
    * @param distance The number of bytes to pass.
    * @return The number of bytes passe (which can be less than 
    *         <code>distance</code>
    * @see FMG.Conductor.dataflow.AdaptorWindow#contract(int)
    */
   public int advance(int distance) {
      Object origLock = lock;
      synchronized(lock) {
         waitForQueueSpace();

         /* make sure the lock wasn't swapped on us while we waited */
         if (origLock != lock) {
            return advance(distance);
         }

         int downCount = 0;
         StreamElem prevPos = this;

         int distToGo = distance;

         while (distToGo > 0) {
            StreamElem se = prevPos.next(Direction.up);
            int seLen = se.length();

            if (se.isBoundary()) {
               break;
            }

            if (! se.isData()) {
               prevPos = se;     // skip this one to keep it in the window
            } else if (distToGo > seLen) {

               distToGo -= seLen;
               se.unlink();
               safeInsert(se, Direction.down);
               downCount += se.proceed();
            } else {
               // This must be a data fragment!!!
               DataFragment df = (DataFragment) se;
               if (! df.migrateBytesTo(safeNext(Direction.down), distToGo)) {

                  if (distToGo < seLen) {
                     df.split(distToGo);
                  }
                  df.unlink();
                  safeInsert(df, Direction.down);

                  recordSegmentation(df);

                  downCount += df.proceed();
               }
               distToGo = 0;
            }
         }

         adjustQueueSize(downCount);
         return (distance - distToGo);
      }
   } 

   /**
    * Used during contract operations, this method moves the boundary
    * up to (but not past) the given pointer.  It is assumed that the pointer
    * is present in the stream prior to the next boundary.
    * Any non-data stream elements encountered will be left in the window
    * in the order they were encountered.
    * @param p The pointer to move to.
    * @return The number of bytes passed
    * @see FMG.Conductor.dataflow.AdaptorWindow#contract(FMG.Conductor.dataflow.Pointer)
    */
   public int moveToPointer(Pointer p) {
      Object origLock = lock;
      synchronized (lock) {
         int downCount = 0;

         waitForQueueSpace();

         if (origLock != lock) {
            return moveToPointer(p);
         }

         int distance = 0;

         StreamElem prevPos = this;
         for (StreamElem target=prevPos.next(Direction.up); target != p;
              target=prevPos.next(Direction.up)) {
            if (target.isData()) {
               recordSegmentation((DataFragment) target);

               distance += target.length();
               target.unlink();
               safeInsert(target, Direction.down);
               adjustQueueSize(target.proceed());
            } else if (target instanceof InterfaceControl) {
               target.unlink();
               safeInsert(target, Direction.down);
               target.proceed();
            } else {
               prevPos = target;
            }
         }

         return distance;
      }
   }  

   /**
    * Used to obtain upward flowing events during expand() operations and
    * for transmission by an IO object, this method moves this boundary 
    * past the next upward moving event object.
    * @return The event that the boundary was moved past, or null if none 
    *         was present.
    * @see FMG.Conductor.dataflow.comm.IO#getAMessage()
    * @see FMG.Conductor.dataflow.AdaptorWindow#checkForUpwardEvent()
    */
   public StreamElem getAnEvent() {
      Object origLock = lock;
      synchronized (lock) {
         if (origLock != lock) {
            return getAnEvent();
         }

         StreamElem target = safeNext(Direction.down);
         if ((! target.isBoundary()) && (target.flowDir().isUp())) {
            safeSwap(Direction.down);

            // See if there will be more work to do in the future
            StreamElem nextTarget = safeNext(Direction.down);
            if ((nextTarget.isBoundary()) || (! nextTarget.flowDir().isUp())) {
               accessor.serviceComplete(this);
            }

            return target;
         } else {
            accessor.serviceComplete(this);
            return null;
         }
      }
   }

   public int handleNewAdjacent(Direction dir, StreamElem se) {
      // let a zero sized data frag move out on its own
      if (dir.isUp() && se.isData() && (se.length() == 0)) {
         se.unlink();
         putOneOut(se);
         return 0;   // return value shouldn't matter in this case
      } else {
         return super.handleNewAdjacent(dir, se);
      }
   }

}
