/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.Conductor.dataflow;

import FMG.Conductor.comm.ConnID;
import java.util.Hashtable;

/**
 * The IAC class provides a primitive form a Intra-Adaptor Communication.
 * Communication takes the form of objects.  An object handle can be stored
 * into the an IAC object with an associated key.  The key is simply a
 * string which has meaning only to the user.  Object handles can then
 * be retreived by the same or a different entity.  Both blocking and
 * non-blocking retreival methods are provided.
 * <p>
 * The IAC class also maintains one IAC object as a node-golbal communication
 * mechanism.  It is created at initialization time.
 *
 * @author Mark Yarvis (yarvis@fmg.cs.ucla.edu)
 * @version $Revision: 1.10 $
 */
public class IAC {

   /** The stream-global IAC object */
   private static IAC nodeIAC = new IAC();

   /** The data stored in an IAC instance, indexed by user-slected strings */
   private Hashtable entries = new Hashtable();

   /**
    * Get the global IAC object.
    * @return The IAC object global to all streams.
    */
   static IAC getNodeIAC() {
      return nodeIAC;
   }

   /**
    * Get an object that was previously recorded under this key string.
    * If no such value was recorded, null is returned.
    * @param key The key used to store the desired object.
    * @return The object stored under the desired key, or null if there is
    *         no such beast.
    */
   public Object lookup(String key) {
      return entries.get(key);
   }

   /**
    * Get an object that was previously recorded under this key string.
    * If no such value has yet been recorded, wait for it to be recorded.
    * @param key The key used to store the desired object.
    * @return The object stored under the desired key.
    * @exception java.lang.InterruptedException if the thread is interrupted
    *                                           while blocked waiting on a 
    *                                           particular key.
    */
   public synchronized Object waitFor(String key) throws InterruptedException {
      while (true) {
         Object entry = entries.get(key);
         if (entry != null) {
            return entry;
         }
         wait();
      }
   }

   /**
    * Store an object under a particular key.  Notify all threads that might
    * be waiting for an object to be stored under this key.
    * @param key The key for storage and retreival
    * @param o The object to store
    * @return No return value.
    */
   public synchronized void store(String key, Object o) {
      entries.put(key, o);
      notifyAll();
   }

   /**
    * Store an object under a particular key.  Only overwrite existing
    * objects if the overwrite parameter is true.  Notify all threads that
    * might be waiting for an object to be stored under this key.
    * @param key The key for storage and retreival
    * @param o The object to store
    * @param overwrite Indicates whether to overwrite an existing value
    * @return No return value.
    */
   public synchronized void store(String key, Object o, boolean overwrite) {
      if (entries.containsKey(key) && ! overwrite) {
         return;
      }

      store(key, o);
   }

   /**
    * Remove an object stored under a particular key.
    * @return No return value.
    */
   public void remove(String key) {
      entries.remove(key);
   }
}
