/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.Conductor.dataflow;

import FMG.Conductor.Conductor;
import FMG.Conductor.planning.Path;
import FMG.Conductor.dataflow.comm.*;
import FMG.Conductor.dataflow.messages.*;
import FMG.Conductor.util.*;
import FMG.Conductor.comm.*;

import java.util.*;
import java.net.*;
import java.io.*;

/**
 * The IOManager object acts as a single point of focus for communication
 * to the next node along a data path.  For each connection, there will
 * be communication torward the client and communication toward the server.
 * An IOManager manages IO objects communicating in one direction or the
 * other.  Typically, both directions of data flow to and from the 
 * adjacent node are handled.
 * <p>
 * The IOManager also includes routines allowing other objects like the
 * Interceptor, the ConductorServer, and the Path object to interact with 
 * IO objects and for IO objects to report events that occur.
 * <p>
 * IOManager objects are always created in pairs.
 * <p>
 * TODO: Perhaps a better way to handle IO objects is this:
 *       The IOManager creates two threads, one for input and one for output.
 *       When a connection is needed, an appropriate object is created and
 *       a call is made to the handler.
 *       When a failure occurs, or if a SwitchPlan is received, the IO object
 *       returns a message back to the IOManager, which can then create a new
 *       object.
 *                 THIS CAN'T WORK
 *                     even with one direction of flow, we need two threads
 *       Also, make them in pairs, but make them unspecialized.  Only 
 *       specialize once a call is made to establish function (ie, handle this
 *       raw socket, or handle this Conductor socket, or handle this 
 *       SwitchPlan message).
 *
 *        9/19/00: Now that we have a session object, do we need to create
 *                 IO objects in pairs?
 * 
 * @author Mark Yarvis (yarvis@fmg.cs.ucla.edu)
 * @version $Revision: 1.34 $
 *
 * @see FMG.Conductor.dataflow.comm.Interceptor
 * @see FMG.Conductor.dataflow.comm.ConductorServer
 * @see FMG.Conductor.planning.Path
 */

public class IOManager {

   /*
    * Static data
    */

   /** The list of all sessions in the system. */
   protected static SessionTable sessionTable = null;


   /*
    * Instance variables.
    */

   /** The boundary for outbound data flow */
   protected UpstreamBoundary upstreamBoundary;

   /** The boundary for inbound data flow */
   protected DownstreamBoundary downstreamBoundary;

   /** The current input object.  Might be the same as outputIO. */
   protected IO inputIO;

   /** The current output object.  Might be the same as inputIO. */
   protected IO outputIO;

   /**
    * Which direction does this object point (up is toward the server, down
    * is toward the client).
    */
   protected Direction dir;

   /** The object holding handles to the various parts of this session */
   protected Session session;

   /**
    * Initialize datapath related data structuresa and messages.  In 
    * particular, register all message types we intend to send.
    * @return No return value.
    */
   public static void initialize(SessionTable st) {
      ConductorIO.registerMessages();
      if (sessionTable == null) {
         sessionTable = st;
      }
   }

   /**
    * Handle a new connection from an adjacent Conductor node.  First,
    * a ConnectionInit message is obtained from the socket to determine
    * which IOManager to give it to.  If no IOManager exists, one is 
    * created.  The new socket is then established as the new input
    * socket for that manager.
    * @param s A socket for a newly established connection to an adjacent
    *          Conductor node.
    * @return No return value.
    * @see FMG.Conductor.dataflow.messages.ConnectionInit
    */
   public static void handleIncoming(Socket s) {
      ConnectionInit initMsg;
      try {
         /* do passive handshake */
         initMsg = ConnectionInit.getNewMessage(s);
      } catch (IOException e) {
         throw new RuntimeException("case not handled");
      }

// MDY: to duplicate or not to duplicate?
      Session newSession = sessionTable.getSession(initMsg.connID);

      if (newSession.getIOManager(initMsg.dir.opposite()) == null) {
         new IOManager(newSession, initMsg.dir.opposite());
      }

      newSession.getIOManager(initMsg.dir.opposite()).replaceInputSocket(s);

      initMsg.recycle();
   }

   /**
    * This method is called on the server endpoint to establish a new 
    * connection to an application server.  An IOManager is created in
    * each direction.
    * @param sourceAddr The address of the client node.
    * @param sourcePort The client port number.
    * @param targetAddr The address of the server node.
    * @param targetPort The server port number.
    * @param session The object holding the parts of this connection.
    * @return no return value.
    */
   public static void createRawServer(InetAddress sourceAddr, int sourcePort, 
                                      InetAddress targetAddr, int targetPort,
                                      Session session) {
      new IOManager(sourceAddr, sourcePort, targetAddr, targetPort, session);
   }

   /**
    * This method is called on the client endpoint when a connection has
    * been intercepted from a client application.  An IOManager is created
    * in each direction.
    * @param s The intecepted socket.
    * @param session The object holding the parts of this connection.
    * @return No return value.
    */
   public static void createRawClient(Socket s, Session session) {
      new IOManager(s, session);
   }

   /**
    * Construct an IOManager that manages the connection with the client
    * application.  Note that the paired IOManager is also created in order
    * to match up the data paths between the two (one is the output to the 
    * other, and visa-versa).  It is assumed that the client node is not 
    * also the server node.
    * @param s The intecepted socket.
    * @param id The id of this connection.
    */
   private IOManager(Socket s, Session session) {
      upstreamBoundary = new UpstreamBoundary();
      downstreamBoundary = new DownstreamBoundary();

      if (Conductor.reliabilityEnabled()) {
         downstreamBoundary.safeInsert(new Cache(), Direction.down);
      }

      /* creates IOManager in other direction */
      new IOManager(session, downstreamBoundary.pair(), upstreamBoundary.pair(),
                                                             Direction.up);
      this.dir = Direction.down;
      this.session = session;
      session.downIOManager = this;

      outputIO = new RawIO(s, session, upstreamBoundary);
   }

   /**
    * Construct an IOManager to manage the connection to the server endpoint.
    * Note that the paired IOmanager is also created to match up the data
    * paths between the two (one is the input to the other and visa-versa).
    * @param sourceAddr The address of the client node.
    * @param sourcePort The client port number.
    * @param targetAddr The address of the server node.
    * @param targetPort The server port number.
    * @param id The id of this connection.
    */
   private IOManager(InetAddress sourceAddr, int sourcePort, 
                     InetAddress destAddr, int destPort, Session session) {
      upstreamBoundary = new UpstreamBoundary();
      downstreamBoundary = new DownstreamBoundary();

      if (Conductor.reliabilityEnabled()) {
         downstreamBoundary.safeInsert(new Cache(), Direction.down);
      }

      /* creates IOManager in other direction */
      new IOManager(session, downstreamBoundary.pair(), 
                    upstreamBoundary.pair(), Direction.down);

      this.dir = Direction.up;
      this.session = session;
      this.session.upIOManager = this;

      outputIO = new RawIO(sourceAddr, sourcePort, destAddr, destPort, 
                           session, upstreamBoundary);
   }

   /**
    * Create an IOmanager that manages a connection with an adjacent 
    * Conductor node.  The paired IOManager is also created in order
    * to match up the data structures (one is an input to the other, and
    * visa-versa).
    * @param id The id of this connection.
    * @param dir The direction this IOManager is communicating (up is toward
    *            the server, down is toward the client).
    */
   private IOManager(Session session, Direction dir) {
      upstreamBoundary = new UpstreamBoundary();
      downstreamBoundary = new DownstreamBoundary();

      /* creates IOManager in other direction */
      new IOManager(session, downstreamBoundary.pair(), 
                    upstreamBoundary.pair(), dir.opposite());

      this.dir = dir;
      this.session = session;
      this.session.setIOManager(dir, this);

      outputIO = new ConductorIO(session, upstreamBoundary, dir);
   }

   /**
    * Create a pair to another IOManager.  This constructor is always called
    * from one of the others.
    * @param id The id of this connection.
    * @param upstreamBoundary The boundary to the outgoing data flow.
    * @param downstreamBoundary The boundary to the incoming data flow.
    * @param dir The direction this IOManager is communicating (up is toward
    *            the server, down is toward the client).
    */
   private IOManager(Session session, UpstreamBoundary upstreamBoundary,
                     DownstreamBoundary downstreamBoundary, Direction dir) {
      this.upstreamBoundary = upstreamBoundary;
      this.downstreamBoundary = downstreamBoundary;

      this.dir = dir;
      this.session = session;
      if (dir.isUp()) {
         this.session.upIOManager = this;
      } else {
         this.session.downIOManager = this;
      }

      outputIO = new ConductorIO(session, upstreamBoundary, dir);
   }

   /**
    * Establish the given socket as the new socket for the inputIO.
    * @param s A newly connected socket form another Conductor node.
    * @return No return value.
    */
   private void replaceInputSocket(Socket s) {

      if ((inputIO instanceof RawIO) || (outputIO instanceof RawIO)) {
          throw new RuntimeException("This shouldn't happen");
      }

      if ((inputIO == null) && !((ConductorIO)outputIO).isConnected()) {
         makeInputIO(outputIO);
      } else if (((ConductorIO) inputIO).isConnected()) {
         makeInputIO(new ConductorIO(session, upstreamBoundary, dir));
      }
      ((ConductorIO)inputIO).setSocket(s);
   }

   /**
    * Make the given IO object the manager of the input data flow.
    * The downstream boundary is set so that data read from the socket can
    * flow down the data stream.
    * @param io The object that will now be the input IO.
    * @return No return value.
    */
   public void makeInputIO(IO io) {
      if (inputIO != io) {
         inputIO = io;
         inputIO.setDownstreamBoundary(downstreamBoundary);
      }
   }

   /**
    * Forward a SwitchPlan object down the data path.  This method is called
    * on the client and server nodes to start plan installation in each
    * direction of data flow.
    * @param m The SwitchPlan message.
    * @return No return value.
    */
   public void forwardSwitchPlan(SwitchPlan m) {
      if (Debug.on) {
         Debug.println(Debug.DATAFLOW, Debug.MOREINFO, 
                       "In IOManager:forwardSwitchPlan");
      }

      m.handleExit(outputIO);
      downstreamBoundary.putOneOut(m);
   }

   /**
    * This method is called by an IO object when a connection has failed.
    * The side of the break closest to the client will always trigger 
    * replanning.
    * @return No return value.
    */
   public void connectionFailed() {
      if (dir.isUp()) {
         session.path.nextNodeFailed();
         session.path.triggerPlanning();
      }
   }

   /**
    * This method is called as the path is being cleanly shut down to
    * cleanup local data structures, including removal of the IOManager 
    * itself.  On the client node, a planning path shutdown should also
    * be triggered.
    * @param shutdownPlanningPath True if the planning path should also be 
    *                             shut down.
    * @return No return value.
    */
   public void connectionEnded(boolean shutdownPlanningPath) {
      if (shutdownPlanningPath) {
         session.path.shutdownPath();
      }
      if (Debug.on) {
         Debug.println(Debug.DATAFLOW, Debug.MOREINFO, 
                       "IOManager(" + dir + "): going away");
      }
      session.removeIOManager(dir);
   }

   public void setUpstreamBoundary(UpstreamBoundary bound) {
      upstreamBoundary = bound;
   }

   /**
    * Dump a verbose listing of the objects in each datapath to the standard
    * output device.  This method should be used for debugging only!
    * @return No return value.
    */
   public static void dumpAllDatapaths(PrintStream out) {
      boolean none = true;
      Enumeration e = sessionTable.getSessions();
      while (e.hasMoreElements()) {
         IOManager manager = ((Session) e.nextElement()).downIOManager;
         // make sure it didn't get deleted
         if (manager != null) {
            out.println(">>>> " + manager.session.connID);
            out.print("downstream flow: ");
            manager.upstreamBoundary.printList(out);
            out.print("upstream flow:   ");
            manager.downstreamBoundary.printReverseList(out);
            out.println("");

            none = false;
         }
      }

      if (none) {
         out.println("None.");
      }
           
   }

   public boolean isRootNode() {
      return session.path.isRootNode();
   }
}
