/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.Conductor.dataflow;

import FMG.Conductor.dataflow.messages.DataFragment;
import FMG.Conductor.util.Direction;

/**
 * A Pointer is a StreamElement that can be advanced efficiently along
 * that data flow in an upstream manner.
 *
 * @author Mark Yarvis (yarvis@fmg.cs.ucla.edu)
 * @version $Revision: 1.14 $
 */

abstract class Pointer extends StreamElem {

   /**
    * Insert a pointer immediately upstream of this one.
    * @param p The pointer to insert.
    * @return No return value.
    */
   public void insertPointer(Pointer p) {
      insert(p, Direction.up);
   }

   /**
    * Advance the pointer upstream by a certain number of bytes.  If the
    * given number of bytes are not present, this call with advance the
    * pointer to the edge of the window.
    * @param distance The number of bytes to move past.
    * @return The distance actually traveled, which may be less than 
    *         <code>distance</code>.
    */
   public int advance(int distance) {
      int distToGo = distance;

      while (distToGo > 0) {
         StreamElem se = next(Direction.up);
         int seLen = se.length();

         if (se.isBoundary()) {
            return (distance - distToGo);
         }

         if (distToGo > seLen) {

            distToGo -= seLen;
            super.swap(Direction.up);
         } else {
            // This must be a data fragment!!!
            DataFragment df = (DataFragment) se;
            if (! df.migrateBytesTo(next(Direction.down), distToGo)) {
               if (distToGo < seLen) {
                  df.split(distToGo);
               }
               super.swap(Direction.up);
            }
            distToGo = 0;
         }
      }

      return distance;
   }

   /**
    * Advance the pointer upstream to the edge of the window.
    * @return The number of bytes passed.
    */
   public int advance() {
      int dist = 0;

      StreamElem se = next(Direction.up);
      while (!se.isBoundary()) {
         dist += se.length();
         super.swap(Direction.up);
         se = next(Direction.up);
      }

      return dist;
   }
}
