/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.Conductor.dataflow;

import FMG.Conductor.comm.ConnID;
import FMG.Conductor.planning.Path;
import FMG.Conductor.util.Direction;

/**
 * The Session object maintains handles for several objects that are 
 * associated with a particular stream: the Path object, an IAC object,
 * and two IOManager objects.  The Session object thus allows various
 * pieces of code to get access to the important objects related to a
 * particular stream.  A Session object is frequently held by stream-specific 
 * objects, allowing them access to other stream-specific objects.  Other
 * non-stream-specific objects may obtain a particular Stream object from
 * a StreamTable object.
 * <p>
 * A Session object creates and maintains an IAC object for this stream.
 * Since the IAC object can contain stream-private information (most notably
 * session keys), it is important that objects from other streams not have
 * access to this session object.
 * <p>
 * A Session object should only be created by the SessionTable object.  The
 * SessionTable object holds a handle for Session objects so that other
 * objects can gain access.  When the session object is no longer needed 
 * (the session has shut down), it removes itself from the table.  This 
 * happens when there is no Path object or IOManager objects associated 
 * with the Session object.
 *
 * @author Mark Yarvis (yarvis@fmg.cs.ucla.edu)
 * @version $Revision: 1.3 $ $Date: 2001/10/17 16:45:29 $
 */

final public class Session {

   /** The conneciton id for this session. */
   protected ConnID connID = new ConnID();

   /** The inter-adaptor communication object for this session. */
   protected IAC iac = new IAC();

   /** The Path object for this session. */
   protected Path path;

   /** The IOManager for this session that is in the direction of the server */
   protected IOManager upIOManager;

   /** The IOManager for this session that is in the direction of the client */
   protected IOManager downIOManager;

   /** The session table which stores a handle to this Session object */
   private SessionTable table;

   /**
    * Create a new Session object with a (new) unique session id.  This
    * constructor should only be called by a SessionTable object and 
    * should only be used for newly intercepted connections.
    * @param table The session table that is creating this Session object.
    */
   Session(SessionTable table) {
      connID.makeNewUnique();
      this.table = table;
   }

   /**
    * Create a Session object with the given connection id.  This constructor
    * should only be called by a SessionTable object.
    * @param table The session table that is creating this Session object.
    */
   Session(ConnID connID, SessionTable table) {
      this.connID.copy(connID);
      this.table = table;
   }

   /**
    * Obtain the connection id for this session.
    * @return The connection id for this session;
    */
   public ConnID getConnID() {
      return connID;
   }

   /**
    * Set the identity of one of the two IOManager objects associated with
    * this session.
    * @param dir The direction indicating which IOManager to set.
    * @param manager The handle to the new IOManager.
    * @return No return value.
    */
   public void setIOManager(Direction dir, IOManager manager) {
      if (dir.isUp()) {
         upIOManager = manager;
      } else {
         downIOManager = manager;
      }
   }

   /**
    * Obtain the handle for one of the two IOManager objects for this session.
    * @param dir The direction indicating which IOManager to get.
    * @return The requested IOManager for this session.
    */
   public IOManager getIOManager(Direction dir) {
      return (dir.isUp() ? upIOManager : downIOManager);
   }

   /**
    * Remove the handle for one of the two IOManagers.  Presumably, the
    * connection is shutting down.  Once both IOManagers and the Path object
    * have removed themselves, this object will remove itself from its
    * associated table, allowing it to be garbage collected.
    * @param dir The direction indicating which IOManager to remove.
    * @return No return value.
    */
   public void removeIOManager(Direction dir) {
      setIOManager(dir, null);

      if ((upIOManager == null) && (downIOManager == null) && (path == null)) {
         table.removeSession(connID);
      }
   }

   /**
    * Set the identity of the path object for this session.
    * @param path The handle to this session's Path object.
    * @return No return value.
    */
   public void setPath(Path path) {
      this.path = path;
   }

   /**
    * Obtain a handle to this session's Path object.
    * @return A handle to the path object for this session.
    */
   public Path getPath() {
      return path;
   }
   
   /**
    * Remove the handle for the Path object for this session.  Presumably
    * the connection is shutting down.  Once the Path object and both
    * IOManager objects are removed, the connection is closed, so the
    * Session object removes itself from the table object, allowing garbage
    * collection of this object.
    * @return No return value;
    */
   public void removePath() {
      path = null;

      if ((upIOManager == null) && (downIOManager == null)) {
         table.removeSession(connID);
      }
   }

   /**
    * Obtain the inter-adaptor communication object for this session.
    * @return A handle to the inter-adaptor communication object for this
    *         session.
    */
   public IAC getIAC() {
      return iac;
   }

   public String toString() {
      return "Session[" + connID + "]";
   }

}
