/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.Conductor.dataflow;

import java.util.Hashtable;
import java.util.Enumeration;

import FMG.Conductor.comm.ConnID;

/**
 * The SessionTable object is a collection of Session objects.  It allows
 * any object that has a handle to a particular SessionTable object to share
 * handles to those sessions.  This object allows sessions to be created,
 * accessed by connection ID, or destroyed.
 * <p>
 * Access to a particular SessionTable object should be limited.  Any object
 * that can access a SessionTable object can obtain all of the Session objects
 * it contains.  See the note in the Session object for further security 
 * concerns.
 * <p>
 * Note that in Conductor, there is normally only one instantiated SessionTable
 * at a time.
 *
 * @author Mark Yarvis (yarvis@fmg.cs.ucla.edu)
 * @version $Revision: 1.3 $ $Date: 2001/10/17 16:45:29 $
 */

public class SessionTable {

   /** The list of sessions handled by a given SessionTable */
   protected Hashtable sessions = new Hashtable();

   /**
    * Create a session object with a unique connection ID for a newly 
    * intercepted connection.
    * @return The handle to a new session object.
    */
   public Session createUniqueSession() {
      Session ret = new Session(this);
      sessions.put(ret.connID, ret);

      return ret;
   }

   /**
    * Get the session object with a particular connection ID.  If none exists,
    * create one.
    * @param connID The connection id of interest.
    * @return The handle to the session object for this connection ID.
    */
   public synchronized Session getSession(ConnID connID) {
      Session ret = (Session) sessions.get(connID);

      if (ret == null) {
         connID = connID.duplicate();
         ret = new Session(connID, this);
         sessions.put(ret.connID, ret);
      }

      return ret;
   }

   /**
    * Remove a session with a particular connection ID.  This method is 
    * normally only called from the session object itself.
    * @param connID The id of the session to remove.
    * @return No return value.
    */
   public void removeSession(ConnID connID) {
      sessions.remove(connID);
   }

   /**
    * Obtain a list of all session objects known to this SessionTable.
    * This method is intended for debugging purposes only.
    * @return An enumeration of all Session objects known to this object.
    */
   public Enumeration getSessions() {
      return sessions.elements();
   }

}
