/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.Conductor.dataflow;

import FMG.Conductor.util.Direction;
import FMG.Conductor.dataflow.messages.*;

// Be specific to avoid confusion with java.lang.Shutdown
import FMG.Conductor.dataflow.messages.Shutdown;

/**
 * The StreamAccessor class provides a skeleton for an object that
 * has access to the data stream.  A StreamAccessor has two associated
 * boundary objects which divide the stream, giving the StreamAccessor
 * exclusive access to a portion.
 * <p>
 * The StreamAccessor works in conjunction with the boundaries.  Whenever
 * a stream element moves adjacent to one of the boundaries, and is stopped,
 * the StreamAccessor is notified.
 *
 * @author Mark Yarvis (yarvis@fmg.cs.ucla.edu)
 * @version $Revision: 1.10 $
 *
 * @see FMG.Conductor.dataflow.Boundary
 */

public class StreamAccessor {

   /** The boundary from which data will flow down.  */
   protected UpstreamBoundary upstreamBoundary;

   /** The boundary from which data will flow up. */
   protected DownstreamBoundary downstreamBoundary;

   /** The upstream boundary has a new adjacent element of interest. */
   protected boolean upNeedsService = false;

   /** The downstream boundary has a new adjacent element of interest. */
   protected boolean downNeedsService = false;

   /** The upstream boundary is closed. */
   protected boolean upstreamClosed = false;

   /** The downstream boundary is closed. */
   protected boolean downstreamClosed = false;

   /**
    * The synchronization object used to wait for data to become available
    * and to notify waiting threads when it does.  <b>WARNING</b>, it is
    * important that we not ask a boundary to do anything while holding
    * this lock, or we will get lock inversion!
    */
   private Object sync = new Object();

   /**
    * Create a new StreamAccessor with the given boundaries.
    * @param upstreamBoundary The boundary from which downward flowing 
    *                         elements will come.
    * @param downstreamBoundary The boundary from which upward flowing 
    *                           elements will come.
    */
   protected StreamAccessor(UpstreamBoundary upstreamBoundary, 
                  DownstreamBoundary downstreamBoundary) {
      if (upstreamBoundary != null) {
         setUpstreamBoundary(upstreamBoundary);
      }

      if (downstreamBoundary != null) {
         setDownstreamBoundary(downstreamBoundary);
      }
   }

   /**
    * Revise the upstream boundary.  Assume that there may be
    * interesting stream elements next to this boundary.
    * @param newBound The new boundary to use as the upstream boundary.
    * @return no return value.
    */
   public void setUpstreamBoundary(UpstreamBoundary newBound) {
      upstreamBoundary = newBound;
      if (newBound != null) {
         newBound.setAccessor(this);
         needsService(newBound);
      }
   }

   /**
    * Revise the downstream boundary.  Assume that there may be
    * interesting stream elements next to this boundary.
    * @param newBound The new boundary to use as the downstream boundary.
    * @return no return value.
    */
   public void setDownstreamBoundary(DownstreamBoundary newBound) {
      downstreamBoundary = newBound;
      if (newBound != null) {
         newBound.setAccessor(this);
         needsService(newBound);
      }
   }

   /**
    * Record the fact that there are interesting stream elements adjacent
    * to the given boundary.  Notify any waiting threads.
    * @param b The boundary that is now next to interesting stream elements.
    */
   protected void needsService(Boundary b) {
      synchronized (sync) {
         if (upstreamBoundary==b) {
            upNeedsService = true;
            sync.notifyAll();
            return;
         }
         if (downstreamBoundary==b) {
            downNeedsService = true;
            sync.notifyAll();
            return;
         }
         throw new RuntimeException("Unexpected event");
      }
   }

   /**
    * Record the fact that there are no interesting stream elements adjacent
    * to this boundary.
    * <p>
    * Notice that this should be called by the Boundary while holding
    * the boundary lock.  Otherwise, a race condition is possible.
    * @param b The boundary in question.
    */
   protected void serviceComplete(Boundary b) {
      synchronized (sync) {
         if (upstreamBoundary==b) {
            upNeedsService = false;
            sync.notifyAll();
            return;
         }
         if (downstreamBoundary==b) {
            downNeedsService = false;
            sync.notifyAll();
            return;
         }
         throw new RuntimeException("Unexpected event");
      }
   }

   /**
    * Wait for interesting information to move adjacent to one of the 
    * boundaries.  This call will return immediately if this object has
    * already been notified of an interesting stream element
    * adjacent to one of the boundaries.  This call will block until
    * shuch notification comes or until the stream is closed in both
    * directions.
    * @return No return value.
    */
   protected void waitForServiceRequest() {
      synchronized (sync) {
         while (true) {
            if (upNeedsService || downNeedsService) {
               return;
            }

            if (upstreamClosed && downstreamClosed) {
               return;
            }

            try {
               sync.wait();
            } catch (InterruptedException e) {}
         }
      }
   }

   /**
    * Wait for a service request, but only in the given direction.
    * @param dir The direction of interest.
    */
   protected void waitForServiceRequest(Direction dir) {
      synchronized (sync) {
         while (true) {
            if ((dir.isUp() && upNeedsService) || 
                (dir.isDown() && downNeedsService)) {
               return;
            }

            if (upstreamClosed && downstreamClosed) {
               return;
            }

            try {
               sync.wait();
            } catch (InterruptedException e) {}
         }
      }
   }

   /**
    * Wait until an element of the specified type is adjacent (in the 
    * specified direction), swap it in, and return it.
    * <p>
    * Warning: After using this method, you will probably want to call 
    * <code>needsService()</code> on one or both boundaries.
    * @param dir The direction of interest.
    * @param msgClass The type of object we're interested in.
    * @return A pointer to the element that was found.
    */
   protected StreamElem waitForMessage(Direction dir, Class msgClass) {
      StreamElem se = null;
    
      while (true) {
         waitForServiceRequest(dir); 
    
         if (upstreamClosed && downstreamClosed) {
            throw new RuntimeException("This shouldn't happen");
         }
   
         if (dir.isUp()) {
            se = upstreamBoundary.getEvent(msgClass);
         } else {
            se = downstreamBoundary.getEvent(msgClass);
         }

         // when the retrans request comes, handle it
         if (se != null) {
            return se;
         }
      }
   }  

   /**
    * Close the data flow in the given direction.  This call notifies
    * blocked threads, in case both directions become closed.
    * <p>
    * TODO: should notifyAll() only be called if both directions are closed?
    * @param dir The direction to close, either up or down.
    * @return No return value.
    */
   protected void close(Direction dir) {
      synchronized (sync) {
         if (dir.isUp()) {
            upstreamClosed = true;
         } else {
            downstreamClosed = true;
         }

         sync.notifyAll();
      }
   }

   /**
    * Handle the case in which a DataFragment message has just been
    * swapped into this StreamAccessor's portion of the stream.  Note
    * that this must be defined here in order for the correct method
    * to be called in a subclass.
    * <p>
    * Note that unlike the other message types, this method is not
    * always called for a DataFragment.
    * @param se The DataFragment message in question.
    * @return True iff the event is still present in the stream upon return.
    *         In this base class, true is always returned.
    */
   public boolean handleEventEntry(DataFragment se) {
      return true;
   }

   /**
    * Handle the case in which a EOF message has just been
    * swapped into this StreamAccessor's portion of the stream.  Note
    * that this must be defined here in order for the correct method
    * to be called in a subclass.
    * @param eof The EOF message in question.
    * @return True iff the event is still present in the stream upon return.
    *         In this base class, true is always returned.
    * @exception FMG.Conductor.dataflow.AsyncEventException 
    *                 Can be thrown by a subclass to alert the consumer 
    *                 of the EOF.  Never thrown in this implementation.
    */
   public boolean handleEventEntry(EOF eof) throws AsyncEventException {
      return true;
   }

   /**
    * Handle the case in which a Shutdown message has just been
    * swapped into this StreamAccessor's portion of the stream.  Note
    * that this must be defined here in order for the correct method
    * to be called in a subclass.
    * @param sd The Shutdown message in question.
    * @return True iff the event is still present in the stream upon return.
    *         In this base class, true is always returned.
    */
   public boolean handleEventEntry(Shutdown sd) {
      return true;
   }

   /**
    * Handle the case in which a RebalanceTypes message has just been
    * swapped into this StreamAccessor's portion of the stream.  Note
    * that this must be defined here in order for the correct method
    * to be called in a subclass.
    * @param sd The Shutdown message in question.
    * @return True iff the event is still present in the stream upon return.
    *         In this base class, true is always returned.
    */
   public boolean handleEventEntry(RebalanceTypes sf)
                                    throws AsyncEventException {
      return true;
   }

   /**
    * Handle the case in which a SignalFailure message has just been
    * swapped into this StreamAccessor's portion of the stream.  Note
    * that this must be defined here in order for the correct method
    * to be called in a subclass.
    * @param sd The Shutdown message in question.
    * @return True iff the event is still present in the stream upon return.
    *         In this base class, true is always returned.
    */
   public boolean handleEventEntry(SignalFailure sd) 
                                        throws AsyncEventException {
      return true;
   }

   /**
    * Handle the case in which a RetransRequest message has just been
    * swapped into this StreamAccessor's portion of the stream.  Note
    * that this must be defined here in order for the correct method
    * to be called in a subclass.
    * @param sd The Shutdown message in question.
    * @return True iff the event is still present in the stream upon return.
    *         In this base class, true is always returned.
    */
   public boolean handleEventEntry(RetransRequest rr)
                                        throws AsyncEventException {
      downstreamBoundary.eraseSegmentation();
      return true;
   }

   /**
    * Handle the case in which a RetransStart message has just been
    * swapped into this StreamAccessor's portion of the stream.  Note
    * that this must be defined here in order for the correct method
    * to be called in a subclass.
    * @param sd The Shutdown message in question.
    * @return True iff the event is still present in the stream upon return.
    *         In this base class, true is always returned.
    */
   public boolean handleEventEntry(RetransStart sd)
                                        throws AsyncEventException {
      return true;
   }

   /**
    * Handle the case in which a InterfaceControl message has just been
    * swapped into this StreamAccessor's portion of the stream.  Note
    * that this must be defined here in order for the correct method
    * to be called in a subclass.
    * @param se The InterfaceControl message in question.
    * @return True iff the event is still present in the stream upon return.
    *         In this base class, true is always returned.
    */
   public boolean handleEventEntry(InterfaceControl se) {
      return true;
   }

   /**
    * Handle the case in which a SegmentAck message has just been
    * swapped into this StreamAccessor's portion of the stream.  Note
    * that this must be defined here in order for the correct method
    * to be called in a subclass.
    * @param se The SegmentAck message in question.
    * @return True iff the event is still present in the stream upon return.
    *         In this base class, true is always returned.
    */
   public boolean handleEventEntry(SegmentAck se) {
      return true;
   }

   /**
    * Handle the case in which a SwitchPlan message has just been
    * swapped into this StreamAccessor's portion of the stream.  Note
    * that this must be defined here in order for the correct method
    * to be called in a subclass.
    * @param se The SwitchPlan message in question.
    * @return True iff the event is still present in the stream upon return.
    *         In this base class, true is always returned.
    * @exception FMG.Conductor.dataflow.AsyncEventException 
    *                 Can be thrown by a subclass to alert the consumer 
    *                 of the plan change.  Never thrown in this 
    *                 implementation.
    */
   public boolean handleEventEntry(SwitchPlan se) throws AsyncEventException {
      setUpstreamBoundary((UpstreamBoundary) se.next(Direction.up));
      return true;
   }

   /**
    * Handle the case in which a BufferingControl message has just been
    * swapped into this StreamAccessor's portion of the stream.  Note
    * that this must be defined here in order for the correct method
    * to be called in a subclass.
    * @param se The BufferingControl message in question.
    * @return True iff the event is still present in the stream upon return.
    *         In this base class, true is always returned.
    * @exception FMG.Conductor.dataflow.AsyncEventException 
    *                 Can be thrown by a subclass to alert the consumer 
    *                 of the plan change.  Never thrown in this 
    *                 implementation.
    */
   public boolean handleEventEntry(BufferingControl se) 
                                               throws AsyncEventException {
      return true;
   }

   /**
    * Handle the case in which a DataFragment message is about to be
    * swapped out of the window.  Note that this method must be defined
    * here in order for the correct method to be called in a subclass.
    * <p>
    * Note that unlike the other message types, this method is not
    * always called for a DataFragment.
    * @param se The DataFragment message in question.
    * @return True iff the caller should check that the direction of
    *         flow is not closed before swappin the message past the boundary.
    */
   public boolean handleEventExit(DataFragment se) {
      return true;
   }

   /**
    * Handle the case in which a EOF message is about to be
    * swapped out of the window.  Note that this method must be defined
    * here in order for the correct method to be called in a subclass.
    * @param eof The EOF message in question.
    * @return True iff the caller should check that the direction of
    *         flow is not closed before swappin the message past the boundary.
    */
   public boolean handleEventExit(EOF eof) {
      return false;
   }

   /**
    * Handle the case in which a Shutdown message is about to be
    * swapped out of the window.  Note that this method must be defined
    * here in order for the correct method to be called in a subclass.
    * @param sd The Shutdown message in question.
    * @return True iff the caller should check that the direction of
    *         flow is not closed before swappin the message past the boundary.
    */
   public boolean handleEventExit(Shutdown sd) {
      return false;
   }

   /**
    * Handle the case in which a RetransRequest message is about to be
    * swapped out of the window.  Note that this method must be defined
    * here in order for the correct method to be called in a subclass.
    * @param sd The RetransRequest message in question.
    * @return True iff the caller should check that the direction of
    *         flow is not closed before swappin the message past the boundary.
    */
   public boolean handleEventExit(RetransRequest rr) {
      upstreamBoundary.eraseSegmentation();
      return false;
   }

   /**
    * Handle the case in which a RetransStart message is about to be
    * swapped out of the window.  Note that this method must be defined
    * here in order for the correct method to be called in a subclass.
    * @param sd The RetransStart message in question.
    * @return True iff the caller should check that the direction of
    *         flow is not closed before swappin the message past the boundary.
    */
   public boolean handleEventExit(RetransStart sd) {
      return false;
   }

   /**
    * Handle the case in which a RebalanceTypes message is about to be
    * swapped out of the window.  Note that this method must be defined
    * here in order for the correct method to be called in a subclass.
    * @param sd The FailurOccurred message in question.
    * @return True iff the caller should check that the direction of
    *         flow is not closed before swappin the message past the boundary.
    */
   public boolean handleEventExit(RebalanceTypes sd) {
      return false;
   }

   /**
    * Handle the case in which a SignalFailure message is about to be
    * swapped out of the window.  Note that this method must be defined
    * here in order for the correct method to be called in a subclass.
    * @param sd The FailurOccurred message in question.
    * @return True iff the caller should check that the direction of
    *         flow is not closed before swappin the message past the boundary.
    */
   public boolean handleEventExit(SignalFailure sd) {
      return false;
   }

   /**
    * Handle the case in which a InterfaceControl message is about to be
    * swapped out of the window.  Note that this method must be defined
    * here in order for the correct method to be called in a subclass.
    * @param se The InterfaceControl message in question.
    * @return True iff the caller should check that the direction of
    *         flow is not closed before swappin the message past the boundary.
    */
   public boolean handleEventExit(InterfaceControl se) {
      return true;
   }

   /**
    * Handle the case in which a SegmentAck message is about to be
    * swapped out of the window.  Note that this method must be defined
    * here in order for the correct method to be called in a subclass.
    * @param se The SegmentAck message in question.
    * @return True iff the caller should check that the direction of
    *         flow is not closed before swappin the message past the boundary.
    */
   public boolean handleEventExit(SegmentAck se) {
      return true;
   }

   /**
    * Handle the case in which a SwitchPlan message is about to be
    * swapped out of the window.  Note that this method must be defined
    * here in order for the correct method to be called in a subclass.
    * @param se The SwitchPlan message in question.
    * @return True iff the caller should check that the direction of
    *         flow is not closed before swappin the message past the boundary.
    */
   public boolean handleEventExit(SwitchPlan se) {
      return true;
   }

   /**
    * Handle the case in which a BufferingControl message is about to be
    * swapped out of the window.  Note that this method must be defined
    * here in order for the correct method to be called in a subclass.
    * @param se The BufferingControl message in question.
    * @return True iff the caller should check that the direction of
    *         flow is not closed before swappin the message past the boundary.
    */
   public boolean handleEventExit(BufferingControl se) {
      downstreamBoundary.setNewQueuingLimit(se.getQueuingLimit());
      return true;
   }

}
