/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.Conductor.dataflow;

import FMG.Conductor.util.*;

/**
 * The StreamElem class provides an abstract super-class for objects in 
 * the data stream.  The stream is implemented as a linked-list, so this 
 * object is a subclass of LLElem.  Each subclass of StreamElem implements
 * a piece of data or an event flowing in the stream or a control structure.
 * <p>
 * Each stream element has three binary atributes.  It can be a boundary,
 * meaning that other stream elements may not move past it and must stop.
 * it can be an event, as distinct from data.  And it can be user-level,
 * meaning that it cannot be allowed to stray outside of a window.
 * <p>
 * Each stream element also flows in one of three directions.  It can
 * flow in a neutral direction, in which case, it moves behind some piece
 * of data and in front of some other piece of data (ie, with the data).  It
 * can also flow upstream, in which case, it moves against the data flow
 * and can pass by data items in the reverse direction.  Finally, it
 * can flow downstream, in which case it moves with the flow of
 * data but can flow more quickly than the data, passing data items.
 * <p>
 * A stream element also has a size.  A nonzero size only makes sense for
 * a stream element that contains user data.
 *
 * @author Mark Yarvis (yarvis@fmg.cs.ucla.edu)
 * @version $Revision: 1.16 $
 */

public abstract class StreamElem extends LLElem {

   /**
    * Determine if this object is a section of the data stream.  At this
    * time, only a DataFragment should return true.
    * @return True iff this object contains a section of the data stream.
    */
   public boolean isData() {
      return false;
   }

   /** 
    * Deterine if this object is a boundary.
    * @return True iff this object is a boundary.
    */
   public boolean isBoundary() {
      return false;
   }

   /** 
    * Deterine if this object is an event.
    * @return True iff this object is an event.
    */
   public boolean isEvent() {
      return false;
   }

   /** 
    * Deterine if this object is restricted to the user level.  If so, it
    * should not be allowed to escape a window.
    * @return True iff this object is a user-level structure.
    */
   public boolean isUserLevel() {
      return false;
   }

   /**
    * Determine the direction this item flows.
    * @return The direction of flow.
    */
   public abstract Direction flowDir();

   /**
    * Get the next item in the data stream in a particular direction.
    * @param dir The direction in which to look for the next element, either
    *            up or down.
    * @return the adjacent element in the requested direction.
    */
   public StreamElem next(Direction dir) {
      return (StreamElem) (dir.isUp() ? next : prev);
   }

   /**
    * Insert the given element into the data stream in the direction given.
    * The item can also be part of a linked list, in which case the entire
    * list is inserted.
    * @param p The item(s) to insert.
    * @param dir The direction in which to insert the item(s).
    * @return No return value.
    */
   public void insert(StreamElem p, Direction dir) {
      if (dir.isUp()) {
         insertAfter(p);
      } else { 
         insertBefore(p);
      }
   }

   /**
    * Find the distance from this stream element to another one that is
    * in the given direction.  The "distance" is the sum of the sizes of
    * this element and all elements between the two.  A RuntimeException
    * will occur if that element is not found as the list is traversed
    * in the given direction.
    * @param other The handle of the target element.
    * @param dir The direction in which to find the other element in the 
    *            stream.
    * @return The sum of the sizes of this element and the elements between
    *         this element and the given element.
    */
   public int distanceTo(StreamElem other, Direction dir) {

      if (other == this) {
         return 0;
      }

      return length() + next(dir).distanceTo(other, dir);
   }

   /**
    * The length of this object.  Only data objects should return a non-zero
    * result.
    * @return This base class returns 0.
    */
   public int length() {
      return 0;
   }

   /**
    * Determine if this stream element contains the given byte.  Data-carrying
    * sub-classes should override this method.
    * @return This base class always returns false.
    */
   public boolean containsByte(byte targetByte) {
      return false;
   }

   /**
    * Determine the distance upstream to locate the given byte.
    * Data-carrying sub-classes should override this method.  This base class
    * simply queries the next element upstream.  Boundary objects should
    * return -1.
    * @return The distance, in bytes, to the given byte in the upstream
    *         direction, or -1 if it is not found before the next boundary.
    */
   public int distanceTo(byte targetByte) {
      return next(Direction.up).distanceTo(targetByte);
   }

   /**
    * Find the distance from this stream element to another one in the stream.
    * The "distance" is the sum of the sizes of
    * this element and all elements between the two.  A RuntimeException
    * will occur if that element is not in the stream.  Note that it is
    * more efficient to give the direction to the target object if it is
    * known.
    * @param other The handle of the target element.
    * @return The sum of the sizes of this element and the elements between
    *         this element and the given element.
    */
   public int distanceTo(StreamElem other) {
      if (other == this) {
         return 0;
      }

      /* assume down */
      StreamElem se = next(Direction.down);
      while (true) {
         // found it, we were right
         if (se == other) {
            return distanceTo(other, Direction.down);
         }

         // not found, we were wrong, it must be down
         if ((se == null) || se.isBoundary()) {
            return distanceTo(other, Direction.up);
         }

         // keep looking
         se = se.next(Direction.down);
      }
   }

   /**
    * Swap the location of this stream element with the element in the
    * given direction.
    * @param dir The direction for performing the swap.
    * @return No return value.
    */
   public void swap(Direction dir) {
      StreamElem target=next(dir);
      unlink();
      target.insert(this, dir);
   }

   /**
    * Perform a commonly needed movement operation.  All neutral-moving
    * stream elements should call this method to implement the 
    * <code>proceed()</code> method.
    * @return The change in the number of data elements in the stream upon 
    *         performing this operation.
    */
   protected int neutralProceed() {
      StreamElem prevNeighbor = null; 
      int ret = 0;
   
      for (StreamElem neighbor = next(Direction.down);
           (neighbor!=null) && (neighbor != prevNeighbor);
           neighbor = next(Direction.down)) {
         ret += neighbor.handleNewAdjacent(Direction.up, this);
         prevNeighbor = neighbor;
      }
      return ret;
   }

   /**
    * Move this object in it's natural direction until it can move no
    * further.  An element will normally, but not necessarily stop when 
    * it reaches an object moving in the same direction.  An element must
    * stop moving if it reaches a boundary.  When it stops, an element
    * should call <code>handleNewAdjacent()</code> on the element that is
    * now its neighbor in the direction of motion.
    * @return The change in the number of data elements in the stream as
    *         a result of this operation.
    */
   public int proceed() {
      return 0;
   }

   /**
    * This method is called immediately after this object is moved into an 
    * area of the stream controlled by a StreamAccessor.
    * @return True iff this element is still in the stream and needs to
    *         be dealt with.
    * @exception FMG.Conductor.dataflow.AsyncEventException
    *                Can be thrown to tell the StreamAccessor (who usually
    *                makes this call) that something unusual happened.
    */
   public boolean handleEntry(StreamAccessor sa) throws AsyncEventException {
      return true;
   }

   /**
    * This method is called immediately before this object is moved out
    * of an area of the stream controlled by a StreamAccessor.
    * @return True iff it should first be verified that the stream is
    *         still open before proceeding.
    */
   public boolean handleExit(StreamAccessor sa) {
      return true;
   }

   /**
    * This method is called whenever an element moves adjacent to this
    * element.
    * @param dir The direction of the newly adjacent element.
    * @param se The handle of the newly adjacent element.
    * @return The change in the number of data elements in the stream as
    *         a result of this operation.
    */
   public int handleNewAdjacent(Direction dir, StreamElem se) {
      return 0;
   }

}
