/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.Conductor.dataflow;

import FMG.Conductor.util.Direction;
import FMG.Conductor.util.IntP;
import FMG.Conductor.dataflow.messages.DataFragment;

/**
 * The UpstreamBoundary is intended to be one of a pair of boundaries that
 * constists of an UpstreamBoundary and a DownstreamBoundary.  Each of these
 * specialized boundaries includes particular operations that are applicable
 * to each.
 *
 * @author Mark Yarvis (yarvis@fmg.cs.ucla.edu)
 * @version $Revision: 1.13 $
 *
 * @see FMG.Conductor.dataflow.DownstreamBoundary
 */

public class UpstreamBoundary extends AccessorBoundary {

   /**
    * Create a new UpstreamBoundary with an associated DownstreamBoundary.
    * The two are linked together in a linked list.
    */
   public UpstreamBoundary() {
      super();
      lockDir = Direction.up;
      pair = new DownstreamBoundary(this);
      safeInsert(pair, Direction.up);
   }

   /**
    * Create an UpstreamBoundary that is linked to and paired with the given
    * DownstreamBoundary.  This method is normally called by the 
    * DownstreamBoundary constructor.
    * @param other The DownstreamBoundary to link this object to.
    * @see FMG.Conductor.dataflow.DownstreamBoundary#DownstreamBoundary()
    */
   public UpstreamBoundary(DownstreamBoundary other) {
      super(other);
   }

   /**
    * Obtain the pair to this boundary.
    * @return The DownstreamBoundary that is paired with this object.
    */
   public DownstreamBoundary pair() {
      return (DownstreamBoundary) pair;
   }

   /* don't use this without making it safe */
   public int advance(int distance) {
      throw new RuntimeException("Operation not permitted");
   }

   /**
    * Used during expand operations, this method moves the boundary past 
    * all upstream data objects.  Motion stops when any non-data stream 
    * element is encountered.
    * @return The number of bytes passed.
    * @see FMG.Conductor.dataflow.AdaptorWindow#expand()
    * @see FMG.Conductor.dataflow.AdaptorWindow#expand(int)
    * @see FMG.Conductor.dataflow.AdaptorWindow#expand(byte)
    */
   public int movePastBytes() {
      Object origLock = lock;
      int distance = 0;
      StreamElem target;
      StreamElem lastData = null;

      synchronized (lock) {

         if (origLock != lock) {
            return movePastBytes();
         }
   
         while (true) {
            target = safeNext(Direction.up);
            if ((target == null) || target.isEvent() || target.isBoundary()) {
               if (target.flowDir().isUp()) {
                  accessor.serviceComplete(this);
               }
               if (lastData != null) {
                  recordSegmentation((DataFragment) lastData);
               }
               return distance;
            }
            safeSwap(Direction.up);
            distance += target.length();

            if (target.isData()) {
               lastData = target;
            }

            /*
             * special case for 0-length blocks ...
             *    give up lock, tell them to proceed, and then continue
             */
            if (target.length()==0) {
               break;
            }
         }
   
      }

      recordSegmentation((DataFragment) target);

      // we only get here in a special case ... see above!
      target.proceed();
      return distance + movePastBytes();
   }
   
   /**
    * Obtain the next downward flowing or neutral event from the stream and
    * move past it.  If none is available, return null.
    * @return The next appropriate event or null.
    * @see FMG.Conductor.dataflow.AdaptorWindow#checkForDownwardEvent()
    */
   public StreamElem getAnEvent() {
      Object origLock = lock;
      synchronized (lock) {
         if (origLock != lock) {
            return getAnEvent();
         }

         StreamElem target = safeNext(Direction.up);
         if ((target == null) || (target.flowDir().isUp())) {
            accessor.serviceComplete(this);
            return null;
         } else if (target.isEvent()) {
            safeSwap(Direction.up);

            // is there more work to do in the future?
            StreamElem nextTarget = safeNext(Direction.up);
            if ((nextTarget == null) || (nextTarget.flowDir().isUp())) {
               accessor.serviceComplete(this);
            }

            return target;
         } else {
            return null;  /* this won't normally happen */
         }
         
      }
   }

   /** 
    * Called by the IO object, this method obtains the next downward or 
    * neutral traveling message from the stream, if one is available.  
    * The element is swapped past the boundary and a handle is returned.
    * @return A handle to the next element or null if none is available.
    * @see FMG.Conductor.dataflow.comm.IO#getAMessage()
    */
   public StreamElem getOne() {
      Object origLock = lock;
      synchronized (lock) {
         if (origLock != lock) {
            return getOne();
         }

         StreamElem target = safeNext(Direction.up);
         if ((target != null) && (! target.flowDir().isUp())) {
            safeSwap(Direction.up);

            if (target.isData()) {
               recordSegmentation((DataFragment) target);
            }

            // is there more work to do in the future?
            StreamElem nextTarget = safeNext(Direction.up);
            if ((nextTarget == null) || (nextTarget.flowDir().isUp())) {
               accessor.serviceComplete(this);
            }

            return target;
         } else {
            accessor.serviceComplete(this);
            return null;
         }
      }
   }

   /**
    * Returns true if there are any data bytes upstream of this boundary.  
    * WARNING: there is no guarantee that the data bytes contained are not
    * of zero-length.
    */
   public boolean dataAvailable() {
      return (spaceRemaining.val != queuingLimit);
   }

}
