/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.Conductor.dataflow.comm;

import FMG.Conductor.util.*;
import FMG.Conductor.dataflow.*;
import FMG.Conductor.dataflow.messages.*;
import FMG.Conductor.comm.*;
import FMG.Conductor.monitor.*;
import FMG.Conductor.Conductor;
import FMG.Conductor.net.SocketOps;

// Be specific to avoid confusion with java.lang.Shutdown
import FMG.Conductor.dataflow.messages.Shutdown;

import java.net.*;
import java.io.*;

/**
 * The ConductorIO class provides a mechanism for moving data from the data
 * path on one node, across the net, to the data path on another node.  Hence
 * it manages a connection to a neighboring node that is either in the 
 * direction of the server or the client node.  A ConductorIO 
 * manages a single socket which allows two way communication with a 
 * neighboring node.  Each ConductorIO can communicate with one adjacent
 * neighbor, and can handle either one datapath or two datapaths (one in 
 * each direction of data flow).  If a different neighbor exists for each 
 * data path, two ConductorIO objects are needed.  This can occur for
 * asymmetric channels or during plan installation.
 * <p>
 * A ConductorIO object is initially unconnected.  It will become connected
 * if either 
 * <ul>
 *   <li> the output thread receives a SwitchPlan message, which indicates
 *        who the neighbor is, or
 *   <li> the IOManager calls <code>setSocket()</code> after ConductorServer
 *        receives a new connection.
 *
 * @author Mark Yarvis (yarvis@fmg.cs.ucla.edu)
 * @version $Revision: 1.53 $
 *
 * @see FMG.Conductor.dataflow.IOManager
 * @see FMG.Conductor.dataflow.comm.ConductorServer
 */
public class ConductorIO extends IO {

   /**      
    * Specifies whether or not recovery should be initiated in response to
    * failure.  The value is controlled by the property InitiateRecovery and
    * currently defaults to false (until testing is completed).
    */      
   public static final boolean INITIATE_RECOVERY =
      Conductor.getBooleanProperty("InitiateRecovery", false);

   /** The system environment monitor to which we report stream activity. */
   protected static EnvMonitor envMonitor = null;

   /** 
    * The size to set the sockets send buffer.  The default, -1, means 
    * don't set it. 
    */
   protected static int initialOutputBufferLimit = 
                    Conductor.getIntProperty("SocketSendBufferLimit", -1);

   /** The input stream used to read messages from the socket */
   protected BasicMessageInputStream is;

   /** The output stream used to write messages to the socket */
   protected BasicMessageOutputStream os;

   /** The connection id for this connection. */
   protected ConnID connID;

   /** The id of the node we're connected to. */
   protected NodeID nextNode = new NodeID();

   /**
    * Does the output stream need to be flushed?  In other words, did we
    * write anything since the last flush?
    */
   protected boolean flushNeeded = false;

   protected Shutdown prematureShutdownMessage = null;

   protected StreamTypeStack upstreamTypeStack = new StreamTypeStack();

   protected StreamTypeStack downstreamTypeStack = new StreamTypeStack();

   protected boolean recoveryMode = false;

   /**
    * Create a new ConductorIO object.  Initially, it will not be connected
    * to anything.
    * @param session The session object for this connection.
    * @param upstreamBoundary the boundary that we are downstream of.
    * @param dir The direction we are facing (up is toward the server, down
    *            is toward the client).
    */
   public ConductorIO(Session session, UpstreamBoundary upstreamBoundary,
                      Direction dir) {
      super(session, upstreamBoundary, dir);
      connID = session.getConnID();
      startThreads();
   }

   /** 
    * Set the identity of our EnvMonitor object which we use to report
    * stream activity.  This method can be called only once, for
    * security reasons.  Subsequent calls will result in a runtime exception.
    * @param envMonitor The new EnvMonitor object.
    * @return No return value.
    */
   public static void setEnvMonitor(EnvMonitor newEnvMonitor) {
      if (envMonitor == null) { 
         envMonitor = newEnvMonitor;
      } else {
         throw new RuntimeException("This shouldn't happen");
      }
   }  

   /**
    * Tell each message type handled by this class to register itself
    * with the BasicMessageInputStream.  This method must be called before
    * this class can be used and should normally be called in a deterministic
    * order with other registration operations.
    * @return No return value.
    */
   public static void registerMessages() {
      SwitchPlan.registerSelf();
      DataFragment.registerSelf();
      EOF.registerSelf();
      ConnectionInit.registerSelf();
      SegmentAck.registerSelf();
      Shutdown.registerSelf();
      RebalanceTypes.registerSelf();
      SignalFailure.registerSelf();
      RetransRequest.registerSelf();
      RetransStart.registerSelf();
   }

   protected void setOutputBufferingLimit(int newLimit) {
      try {
         SocketOps.setOutputBuffer(s, newLimit);
      } catch (IOException e) {
         if (Debug.on) {
            Debug.println(Debug.DATAFLOW, Debug.WARNING,
                          "ConductorIO: failed to set ouput buffering limit");
         }
      }
   }           

   /**
    * Create the input and output streams required to communicate through
    * the socket.  This method is called by the IO superclass shortly after
    * setting a new socket.
    * @return No return value.
    */
   protected void createStreams() {
      if (s == null) {
         return;
      }
      try {
         if (initialOutputBufferLimit != -1) {
            setOutputBufferingLimit(initialOutputBufferLimit);
         }

         nextNode = new NodeID(s.getInetAddress());
         is = new DataFlowInputStream(s.getInputStream());

         if (os != null) {
            envMonitor.unregisterOutputStream(os);
         }
         os = new DataFlowOutputStream(s.getOutputStream());
         os.setAutoFlush(false);
         envMonitor.registerOutputStream(nextNode, os);
      } catch (IOException e) {
         throw new RuntimeException(e.toString());
      }
   }

   /**
    * Perform the active startup protocol on a newly created socket.
    * Essentially, this just means sending a ConnectionInit message to
    * the other end, which helps it identify which ConductorIO to connect
    * this new socket with.
    * @param connID The connection id for this data path.
    * @param dir The direction we're currently facing.  We expect to connect
    *            to a ConductorIO that faces the opposite direction.
    * @return No return value.
    */
   protected void doActiveStartup(ConnID connID, Direction dir) 
                                                        throws IOException {
      ConnectionInit m = ConnectionInit.getNewMessage(connID,dir);
      os.writeUnframedMessage(m);
      flushNeeded = true;
      m.recycle();
   }
   
   /**
    * This method forms the main code for the output thread.  While it
    * is the cast that either the upstream or downstream direction of
    * data flow is not closed, pull messages that want to flow toward
    * the next node and send them.  Note that the acceptance of a SwitchPlan
    * message will cause this object to be connected or re-connected.  If a 
    * failure occurs, the manager will be notified and this method will
    * be called recursively for further handling.
    * @return No return value.
    */
   public void handleOutputStream() {
      if (upstreamClosed) {
         return;
      }

      // For debugging only
      int messagesSent = 0;

      while (true) {    // Look to handle socket failures
         if (Debug.on) {
            Debug.println(Debug.DATAFLOW, Debug.MOREINFO, 
                          "** Conductor output starting up");
         }

         try {
            while ((! upstreamClosed) || (! downstreamClosed)) {
               Message m = getAMessage();
               if (Debug.on && (Debug.getLevel(Debug.DATAFLOW) > 
                                     Debug.EXTRAINFO)) {
                  messagesSent++;
                  if ((messagesSent % 10) == 0) {
                     Debug.println("ConductorIO: sent " + 
                                   messagesSent + " messages");
                  }
               }
               if (Debug.on) {
                  Debug.println(Debug.DATAFLOW, Debug.DETAIL,
                                "ConductorIO: sending " + m);
               }

               if (m != null) {
                  if (s!=null) {
                     os.writeMessage(m);
                     flushNeeded = true;
                  } else {
                     if (m instanceof Shutdown) {
                        ((Recyclable) m).retain();
                        triggerPrematureShutdown((Shutdown) m);
                     }
                  }
                  ((Recyclable) m).recycle();
               }
            }
            if (os != null) {
               os.flush();
               closeSocket();
            }

            if (Debug.on) {
               Debug.println(Debug.DATAFLOW, Debug.MOREINFO,
                             "** ConductorIO: output handler closed socket");
            }
            manager.connectionEnded(false);
            return;
         } catch (IOException e) {
            if (upstreamClosed && downstreamClosed) {
               // connection is done, forget the failure
               envMonitor.unregisterOutputStream(os);
               envMonitor.unregisterDatapath(connID, nextNode);

               System.err.println("** ConductorIO(output): ending on " + 
                                  "close or error");
               closeSocket();
               manager.connectionEnded(false);
               return;
            }

            System.err.println("ConductorIO(output): " + e);
            envMonitor.unregisterOutputStream(os);
            envMonitor.unregisterDatapath(connID, nextNode);
            os.flush();
            closeSocket();
         }
      }
   }

   protected void triggerPrematureShutdown(Shutdown m) {
      prematureShutdownMessage = m;
      // kludge
      setSocket(null);
   }

   /**
    * This method contains the main body of code for the input thread.
    * Once a socket has been set, read messages from that socket and
    * deliver them down the appropriate data path.  If a failure occurs
    * while the data path is alive, loop around and wait for a new socket
    * to be set.
    * @return No return value.
    */
   public void handleInputStream() {
      // For debugging only
      int messagesReceived = 0;

      Thread currentThread = Thread.currentThread();

      while (true) {   // loop to handle socket failures
         waitForSocket();

         if (prematureShutdownMessage != null) {
            manager.makeInputIO(this);
            putAMessage(prematureShutdownMessage);
         }

         if (Debug.on) {
            Debug.println(Debug.DATAFLOW, Debug.MOREINFO,
                          "** Conductor input starting up");
         }
         try {
            while ((! upstreamClosed) || (! downstreamClosed)) {
               Message m = is.readMessage();

               if (upstreamClosed && downstreamClosed) {
                  throw new RuntimeException("Got a message when we should " +
                                             "be shut down");
               }

               /* establish a segment ID on non-block segments */
               if (Debug.on) {
                  Debug.println(Debug.DATAFLOW, Debug.DETAIL,
                                "ConductorIO: receiving " + m);

                  if (Debug.getLevel(Debug.DATAFLOW) > Debug.EXTRAINFO) {
                     messagesReceived++;
                     if ((messagesReceived % 10) == 0) {
                        Debug.println("ConductorIO: received " + 
                                      messagesReceived + " messages");
                     }
                  }
               }

               putAMessage(m);
            }
            if (Debug.on) {
               Debug.println(Debug.DATAFLOW, Debug.MOREINFO, 
                             "** ConductorIO: input handler ended");
            }
            return;
         } catch (IOException e) {
            if ((! upstreamClosed) || (! downstreamClosed)) {
               System.err.println("ConductorIO(input): " + e);
               manager.connectionFailed();
               closeSocket();
            } else {
               if (Debug.on) {
                  Debug.println(Debug.DATAFLOW, Debug.MOREINFO, 
                                "** ConductorIO(input): ending on " + 
                                  "close or error");
               }

               closeSocket();
               return;
            }
         }

         if (INITIATE_RECOVERY && Conductor.reliabilityEnabled()) {
            if (Measure.on && ((Measure.mask & Measure.RECOVERY) != 0)) {
               Measure.println("RECOVERY: Detected failure");
            }

            recoveryMode = true;
            if (Debug.on) {
               Debug.println("**** " + this + ": Initiating recovery!!!!!");
               Debug.println(Debug.RELIABILITY, Debug.INFO, 
                             this + "Inserting DataSink");
            }
            downstreamBoundary.putOneOut(
                     new DataSink(downstreamBoundary.isBetweenSegments(), 
                                  downstreamBoundary.lastSegmentID()));
         }
      }
   }

   /**
    * Handle the case in which a SwitchPlan message is pulled from
    * the data path for transmission.  First, any local adaptors
    * that need installing are installed by calling the super method.
    * Next, the socket is either connected or re-connected (in case
    * of re-planning, which can change the node order).
    * <p>
    *
    * @param sp The SwitchPlan message being handled.
    * @return Always true, indicating that this message needs to be sent 
    *         along to the next node.
    */
   public boolean handleEventEntry(SwitchPlan sp) {
      super.handleEventEntry(sp);

      // copy old type into to SwitchPlan
      sp.setOldUpstreamTypeStack(upstreamTypeStack);

      // set local upstream type from SwitchPlan's current type
      sp.copyTypeStack(upstreamTypeStack);

      NodeID newNextNode = sp.getNextInterface(session);
      try {
         if (s == null) {
            /* create socket and streams */
            setSocket(new Socket(newNextNode.getAddress(), 
                                 ConductorServer.CONDUCTOR_PORT));

   /* TODO: should register actual link param desires (from plan) */
            envMonitor.registerDatapath(sp.connID, nextNode, 
                                        LinkParams.MAX_PARAMS);
   
            /* do active startup */
            doActiveStartup(sp.connID, sp.direction());
         } else {
// BUG: nextNode should be the nodeID, not the interface!
//      also need a newNextNode and a newNextInterface
            /* are we already connected to the right place? */
            if (! nextNode.equals(newNextNode)) {
               /* nope! */
               Debug.println("Re-connecting data socket");
               envMonitor.unregisterDatapath(sp.connID, nextNode);
               os.flush();
               closeSocket();

               /* create socket and streams */
               setSocket(new Socket(newNextNode.getAddress(), 
                                    ConductorServer.CONDUCTOR_PORT));

/* TODO: should register actual link param desires (from plan) */
               envMonitor.registerDatapath(sp.connID, nextNode, 
                                           LinkParams.MAX_PARAMS);

               /* do active startup */
               doActiveStartup(sp.connID, sp.direction());
            } else {
/* TODO: should register actual link param desires (from plan) */
               envMonitor.registerDatapath(sp.connID, nextNode, 
                                           LinkParams.MAX_PARAMS);
            }
         }
      } catch (IOException e) {
         s = null;
         socketSet = false;
      }
      return true;
   }

   /**
    * Handle the case in which an Shutdown message has been received
    * locally.  This message indicates that no more messages should be
    * expected on a particular endpoint.  So, that endpoint is closed.
    * @param sd The Shutdown message to be handled.
    * @return Always returns true, indicating that this message should
    *         be forwarded across the socket.
    */
   public boolean handleEventEntry(Shutdown sd) {
      if (sd.flowDir().isUp()) {
         close(Direction.down);
      } else {
         close(Direction.up);
      }
      return true;
   }

   /**
    * Handle the case in which an BufferingControl message has been 
    * received locally.  Set the output buffering for the socket.
    * @param bc The BufferingControl message to be handled.
    * @return Always returns false, indicating that this message has
    *         been handled and should not be forwarded across the socket.
    */
   public boolean handleEventEntry(BufferingControl bc) {
      setOutputBufferingLimit(bc.getSocketBufferingLimit());
      bc.unlink();
      bc.recycle();
      return false;
   }

   public boolean handleEventEntry(RebalanceTypes rt) 
                                      throws AsyncEventException {
      super.handleEventEntry(rt);
      if (Debug.on) {
         Debug.println(Debug.RELIABILITY, Debug.INFO, 
                       this + " received " + rt);
      }

      if (rt.flowDir().isDown()) {
         // copy local upstream type from requested type
         rt.copyTargetTypeStack(upstreamTypeStack);
      } else {
         // copy local downstream type from requested type
         rt.copyTargetTypeStack(downstreamTypeStack);
      }

      return true;
   }

   public boolean handleEventExit(RebalanceTypes rt) {
      super.handleEventExit(rt);

      if (rt.flowDir().isUp()) {
         // copy local upstream type from requested type
         rt.copyTargetTypeStack(upstreamTypeStack);
      } else {
         // copy local downstream type from requested type
         rt.copyTargetTypeStack(downstreamTypeStack);
      }

      return true;
   }
   
   /**
    * Handle the case in which an Shutdown object is leaving this object and
    * entering the local data stream.  Close the appropriate boundary
    * (depending on the flow direction of the message).
    * @param sd The Shutdown message to be handled.
    * @return Always returns false, indicating that the boundary should not be
    *         checked for closure first.
    */
   public boolean handleEventExit(Shutdown sd) {
      if (sd.flowDir().isUp()) {
         close(Direction.up);
      } else {
         close(Direction.down);
      }
      return false;
   }

   public boolean handleEventExit(SwitchPlan sp) {
      super.handleEventExit(sp);

      if (recoveryMode) {
         recoveryMode = false;

         // get the old upstream type stack
         StreamTypeStack oldUpstreamTypeStack = new StreamTypeStack();
         sp.copyOldUpstreamTypeStack(oldUpstreamTypeStack);

         // compute the "old" GCT
         int gctSize = StreamTypeStack.numCommonElements(downstreamTypeStack,
                                                         oldUpstreamTypeStack);
         StreamTypeStack oldGCT = new StreamTypeStack();
         oldGCT.copy(downstreamTypeStack);
         oldGCT.setSize(gctSize);

         // set up newGCT ... may be altered below
         StreamTypeStack newGCT = new StreamTypeStack();
         sp.copyTypeStack(newGCT);

         if (! oldUpstreamTypeStack.equals(oldGCT)) {
            /*
             * alter "new" GCT
             */
            // remove each "extra" element in the old upstream type from the 
            // current type
            int badElements = 
                    oldUpstreamTypeStack.size() - 
                    StreamTypeStack.
                              numCommonElements(oldUpstreamTypeStack, oldGCT);
            for (int i=0; i<badElements; i++) {
               StreamTypeStack.StreamType badType = oldUpstreamTypeStack.pop();
               newGCT.removeType(badType);
            }
         }

         if (! sp.equalsTypeStack(newGCT)) {
            /*
             * Attain new GCT
             */
            // Set the new GCT as a plan target
            if (Debug.on) {
               Debug.println(Debug.RELIABILITY, Debug.INFO, 
                             this + " setting new GCT target in PlanDistro " + 
                             newGCT);
            }
            sp.setTargetTypeStack(newGCT);
            // Send a retyping message upstream to obtain new GCT
            try {
               RebalanceTypes rt = 
                    RebalanceTypes.getNewMessage(Direction.up, newGCT);
               if (Debug.on) {
                  Debug.println(Debug.RELIABILITY, Debug.INFO, 
                                this + " sending " + rt);
               }
               os.writeMessage(rt);
            } catch (IOException e) {
               // just another failure ... will be caught
            }
         }

         if (! downstreamTypeStack.equals(oldGCT)) {
            // send retyping message downstream
            RebalanceTypes rt = 
                  RebalanceTypes.getNewMessage(Direction.down, oldGCT);
            if (Debug.on) {
               Debug.println(Debug.RELIABILITY, Debug.INFO, 
                             this + " sending " + rt);
            }
            putAMessage(rt);

         } else {
            if (Debug.on) {
               Debug.println(Debug.RELIABILITY, Debug.INFO, 
                             this + " sending SignalFailure");
            }
            // send SignalFailure downstream
            putAMessage(SignalFailure.getNewMessage());
         }

         downstreamTypeStack.copy(newGCT);

      } else {        // if not doing recovery ...
         sp.copyTypeStack(downstreamTypeStack);
      }

      return true;
   }

   /**
    * Close one of the boundaries, and hence a direction of data flow.
    * Besides calling the super method to do the expected work, we also
    * have to notify any environment monitor that this data path is shutting
    * down.
    * @param dir The direction of flow that is to be closed.
    * @return No return value.
    */
   public void close(Direction dir) {
      super.close(dir);

      /* unregister os with monitor */
      envMonitor.unregisterOutputStream(os);
      envMonitor.unregisterDatapath(connID, nextNode);
   }

   /**
    * This method is called whenever getting the next message to write
    * will cause the output thread to block.  When this occurs, if a
    * flush is needed, it it performed now.
    * @return No return value.
    */
   public void prepareToBlock() {
      if (flushNeeded) {
         // sync any outstanding messages
         os.flush();
         flushNeeded = false;
      }
   }

   /**
    * Close the socket currently managed by this object.
    * @return No return value.
    */
   protected synchronized void closeSocket() {
      // if it's already closed, do nothing
      if (s == null) {
         return;
      }

      /** 
       * IMPORTANT NOTE: The IBM JDK v1.1.8 seems to have a problem where
       * if one thread is blocked on a read call and the other thread closes
       * the socket, the read can start reading from some other socket.
       * The following ordering of closes (input stream, then output stream,
       * then socket) seems to get rid of the problem.  Take care when
       * changing this function!
       */
      is.close();
      os.close();

      try {
         s.close();
      } catch (IOException ignore) {}

      s = null;
      socketSet = false;
   }

   public String toString() {
      return "ConductorIO@" + hashCode() + "[" + connID + "," + dir + "]";
   }

}
