/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.Conductor.dataflow.comm;

import java.io.*;
import java.net.*;

import FMG.Conductor.Conductor;
import FMG.Conductor.util.Debug;
import FMG.Conductor.dataflow.*;

/**
 * The ConductorServer is a thread that handles incoming data-path connections
 * from other Conductor nodes.  These connections are generated by ConductorIO
 * objects when the data path is forming.
 * <p>
 * Normally, only one of these objects will be present in the system at a
 * time, since only one server socket can be bound to the port.
 * <p>
 * The port number for these connections is configurable using the 
 * <code>DataPort</code> property in the Conductor properties file.
 *
 * @author Mark Yarvis (yarvis@fmg.cs.ucla.edu)
 * @version $Revision: 1.17 $
 *
 * @see FMG.Conductor.dataflow.comm.ConductorIO
 */

public class ConductorServer extends Thread {

   /** The port number to use for this type of communication */
   public static final int CONDUCTOR_PORT = 
                             Conductor.getIntProperty("DataPort", 4446);

   /**
    * Create a new ConductorServer and start the associated thread.
    * @return No return value.
    */
   public static void createAndStart() {
      (new ConductorServer()).start();
   }

   /**
    * Create a new ConductorServer object.  This consutructor should normally
    * be called by <code>createAndStart()</code> so the associated thread
    * is started.
    */
   protected ConductorServer() {
      super();
      if (Debug.on) {
         setName("ConductorServer");
      }
      setPriority(MIN_PRIORITY + 1);
   }

   /**
    * Do the main loop for this thread.  For each new socket connection
    * accepted, pass the new socket to the IOManager, which will
    * create an appropriate handler.
    * @return No return value.  In fact, it should never return.
    */
   public void run() {
      ServerSocket ss;
      try {
         ss = new ServerSocket(CONDUCTOR_PORT);
      } catch (IOException e) {
         throw new RuntimeException("Can't start data connection server");
      }

      while (true) {

         try {
            Socket s = ss.accept();

            if (Debug.on) {
               Debug.println(Debug.DATAFLOW, Debug.MOREINFO,
                             "Got data connection from " + s.getInetAddress());
            }

            IOManager.handleIncoming(s);
         } catch (IOException e) {
             System.err.println("Exception accepting data connection: " + e);
         }
      }
   }
}
