/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.Conductor.dataflow.comm;

import java.io.*;
import java.net.*;

import FMG.Conductor.net.SocketOps;
import FMG.IPcept.SocketControl;
import FMG.Conductor.util.*;

public class FastPathIO extends Job {

   protected InputStream is;
   protected OutputStream os;
   protected String dir;

   protected static final int BUFFER_SIZE = 10240;

   public static void createDatapath(Socket clientSocket,
                                     InetAddress sourceAddr, int sourcePort,
                                     InetAddress destAddr, int destPort) {
      if (Debug.on) {
         Debug.println(Debug.DATAFLOW, Debug.MOREINFO, 
                       "FastpathIO: Bypassing conductor.");
         Debug.println(Debug.DATAFLOW, Debug.MOREINFO,
                       "Attempting to masquerade as " + sourceAddr);

      }

      Socket serverSocket;
      try {
         serverSocket = 
               SocketControl.createUnproxiedSocket(destAddr, destPort, 
                                                   sourceAddr, sourcePort);
      } catch (Exception e) {
         // Fall back on normal method
         if (Debug.on) {
            Debug.println(Debug.WARNING,
                          "Warning: Failed to masquerade as source");
            Debug.println(Debug.DATAFLOW, Debug.MOREINFO, "Error was: " + e);
         }
         try {
            serverSocket = new Socket(destAddr, destPort);
         } catch (IOException e2) {
            if (Debug.on) {
               Debug.println(Debug.WARNING,
                             "FastPathIO: Connection failed.");
            }
            try {
               clientSocket.close();
            } catch (IOException ignore) {}
            return;
         }
      }

      createAndStart(clientSocket, serverSocket, "upstream");
      createAndStart(serverSocket, clientSocket, "downstream");
     
   }

   protected static void createAndStart(Socket inSock, Socket outSock, 
                                                           String dir) {
      FastPathIO fp = new FastPathIO(inSock, outSock, dir);
      ThreadRecycler.startNewJob(fp, Thread.MAX_PRIORITY);
   }

   protected FastPathIO(Socket inSock, Socket outSock, String dir) {
      try {
         is = inSock.getInputStream();
         os = outSock.getOutputStream();
         this.dir = dir;
      } catch (IOException e) {
         throw new RuntimeException("This shouldn't happen");
      }
   }


   public void run() {
      byte[] buffer = new byte[BUFFER_SIZE];
      int len;

      if (Debug.on) {
         currentThread.setName("FastPathIO(" + dir + ")");
      }

      while (true) {
         try {
            len = is.read(buffer);
         } catch (IOException e) {
            // couldn't read ... do a half close on output stream
            shutdown(os);
            return;
         }

         if (len == -1) {
            shutdown(os);
            return;
         }

         try {
             os.write(buffer, 0, len);
             os.flush();
         } catch (IOException e) {
            // couldn't write ... do a half close on input stream
            shutdown(os);
            return;
         }
      }
   }

   protected void shutdown(InputStream is) {
      try {
         SocketOps.shutdown(is);
      } catch (IOException e) {
         if (Debug.on) {
            Debug.println(Debug.WARNING, "WARNING: shudown failed: " + e);
         }
      }
   }

   protected void shutdown(OutputStream os) {
      try {
         SocketOps.shutdown(os);
      } catch (IOException e) {
         if (Debug.on) {
            Debug.println(Debug.WARNING, "WARNING: shudown failed: " + e);
         }
      }
   }

}
