/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.Conductor.dataflow.comm;

import java.net.*;

import FMG.Conductor.dataflow.*;
import FMG.Conductor.comm.*;
import FMG.Conductor.util.*;
import FMG.Conductor.dataflow.messages.*;

/**
 * The IO class is the base class for objects that pass data in and out of 
 * the Conductor node.  Two threads are associated with each IO object,
 * an InputJob and an OutputJob.  Each job executes the abstract
 * methods intended for handling input and output, 
 * <code>handleInputStream()</code> and <code>handleOutputStream()</code>.
 * <p>
 * A given connection will have at least two IO objects.  If the upstream
 * and downstream data paths are the same, it will have exactly two IO 
 * objects.  If, however, the next node toward the server (for example) is
 * different for data flowing upstream and data flowing downstream, then
 * there will be a different IO object for each direction of flow at that
 * endpoint.  Thus, there can be up to 4 IO objects for each connection.
 * also, note that if the datapath is changing as a result of replanning
 * there may temporarily be a different next node for each direction of flow.
 * There is one IOManager object that controls the (one or two) IO objects
 * on the side toward the server, and another for the side toward
 * the client.
 * <p>
 * This class extends StreamAccessor, so it has an upstream and a 
 * downstream boundary.  However, it is somewhat different in two
 * respects.  First, rather than sitting along a unidirectional portion
 * of the data flow, it sits at the end of two data flows.  Second,
 * the downstream endpoint is initially not set.
 * <p>
 * TODO: the fact that the downstream endpoint is not initially set is very 
 * ugly!
 *
 * @author Mark Yarvis (yarvis@fmg.cs.ucla.edu)
 * @version $Revision: 1.22 $
 *
 * @see FMG.Conductor.dataflow.comm.InputJob
 * @see FMG.Conductor.dataflow.comm.OutputJob
 * @see FMG.Conductor.dataflow.IOManager
 */

public abstract class IO extends StreamAccessor {

   /** The thread that is handling the input task. */
   protected InputJob reader = new InputJob(this);

   /** The thread that is handling the output task. */
   protected OutputJob writer = new OutputJob(this);

   /** This lock synchronizes around the setting of the socket. */
   private Object socketLock = new Object();

   /** Is the socket set yet? */
   protected boolean socketSet = false;

   /** The socket on which we will operate. */
   protected Socket s;

   /** The number of the last segment processed */
   private int lastSegmentID = 0;

   /** 
    * Which end is this IO object on, upstream (toward the server) or 
    * downstream (toward the client)? 
    */
   Direction dir;

   /** Our manager, to which we report when things happen. */
   protected IOManager manager;

   /** The session object for this session. */
   protected Session session;

   /**
    * Create a new IO object to handle outflowing data.
    * @param session The session object for this stream
    * @param upstreamBoundary The boundary where data to be output will appear.
    * @param dir The side on which this IO is located, upstream or down.
    */
   public IO(Session session, UpstreamBoundary upstreamBoundary, 
             Direction dir) {
      super(upstreamBoundary, (DownstreamBoundary) null);
      this.dir = dir;
      manager = session.getIOManager(dir);
      this.session = session;
   }

   /**
    * Startup the input and output threads.
    * @return No return value.
    */
   protected void startThreads() {
      ThreadRecycler.startNewJob(reader);
      ThreadRecycler.startNewJob(writer);
   }

   /**
    * Set the socket that we will be using for communication.  Notify any
    * thread that might be waiting for the socket to be set.
    * @param s The new socket to use.
    * @return No return value.
    */
   public void setSocket(Socket s) {
      synchronized (socketLock) {
         this.s = s;
         createStreams();
         socketSet = true;
         socketLock.notifyAll();
      }
   }  

   /**
    * Create any streams that are required by a subclass of IO.  This
    * method is called after the socket is set, but before the socket
    * lock has been given up.  This method must be implemented by all
    * subclasses.
    * @return No return value.
    */
   protected abstract void createStreams();

   /**
    * Wait for another thread to set the socket.  Interrupts are ignored.
    * @return No return value.
    */
   protected void waitForSocket() {
      synchronized (socketLock) {
         while (socketSet == false) {
            try {
               socketLock.wait();
            } catch (InterruptedException e) {}
         }
      }
   }

   /**
    * Determine if the socket is set for this object.
    * @return True iff the socket is already set.
    */
   public boolean isConnected() {
      return socketSet;
   }

   /**
    * Get the next message from the upstream or downstream data flows.
    * This method will ask the upstream boundary for the next message (one
    * traveling in the down or neutral direction).  If none is available,
    * it will ask the downstream boundary for an upstream traveling event.
    * If no messages are pending, it will wait for a service request.
    * <p>
    * Note that the boundary objects are responsible for doing the actual
    * locking, and for notifying the StreamAccessor (us) that new data
    * is available.
    * @return A message ready to be transmitted, or null if both the 
    *         upstream and downstream flows have been closed.
    */
   public Message getAMessage() {
      try {
         while (true) {
            StreamElem ret = null;

            if (upNeedsService) {
               ret = upstreamBoundary.getOne();
            }

            if ((ret == null) && (downNeedsService)) {
               ret = downstreamBoundary.getAnEvent();
            }

            if ((ret != null) && (ret.handleEntry(this))) {
               ret.unlink();
               return (Message) ret;
            }

            if (upstreamClosed && downstreamClosed) {
               return null;
            }

            if ((!upNeedsService) && (!downNeedsService)) {
               prepareToBlock();
               waitForServiceRequest();
            }
         }
      } catch (AsyncEventException e) {
         throw new RuntimeException("This should never happen");
      }
   }

   /**
    * Put a message that was received from the outside onto either the upstream
    * or downstream flow, depending on the direction the message intends to 
    * flow.  Upstream flowing messages are handed to the upstream endpoint.
    * Downstream and neutral flowing messages are handed to the downstream 
    * endpoint.
    * <p>
    * Note that the boundary objects are responsible for doing the actual
    * locking.
    * @param m The incoming message that needs to be added to the local 
    *          data flow.
    * @return No return value.
    */
   public void putAMessage(Message m) {
      StreamElem newElem = (StreamElem) m;

      if (newElem.flowDir().isUp()) {

         if (newElem.handleExit(this) && 
             ((upstreamBoundary == null) || (upstreamClosed))) {
            throw new RuntimeException("Case not handled");
         }

         upstreamBoundary.putOneOut(newElem);

      } else {  /* not an upward moving stream element */
         if (newElem.handleExit(this) && 
             ((downstreamClosed) || (downstreamBoundary == null))) {
            throw new RuntimeException("Case not handled");
         }

         downstreamBoundary.putOneOut(newElem);
      }
   }

   /**
    * Handle the case in which a SwitchPlan message has entered this
    * object (ie, it was fetched by <code>getAMessage()</code>).
    * In this case, install any additional adaptors required on this
    * node and reset our upstream boundary.
    * @param sp The PwitchPlan object that was fetched.
    * @return True, indicating that the message still needs to be handled.
    */
   public boolean handleEventEntry(SwitchPlan sp) {
      sp.installAdaptors(session);

      /*
       * If any adaptors were installed, we gave away our upstream boundary.
       * Make sure it's set correctly again.
       */
      setUpstreamBoundary((UpstreamBoundary) sp.next(Direction.up));
      manager.setUpstreamBoundary(upstreamBoundary);

      return true;
   }

   /**
    * Handle the case in which a DataFragment message is about to exit
    * this object (ie, in <code>putAMessage()</code>).
    * In this case indicate the previous segment id to this object so it
    * can set its segment id.  This allows us to not transmit segment ids
    * when a DataFragment consists of a bunch of one-byte segments and
    * generates the initial segment ids for raw data.
    * @param df The DataFragment that was received and is entering the 
    *           data stream.
    * @return True, indicating that the caller should ensure that the
    *         exit is not closed.
    */
   public boolean handleEventExit(DataFragment df) {
      df.informOfPreviousSegmentID(lastSegmentID);
      lastSegmentID = df.segmentID();
      return true;
   }

   /**
    * Handle the case in which the start of retransmission is leaving
    * this IO object.  In this case, we need to record the segment id so
    * we can attach the right segment id to the next message that comes 
    * through.
    * @param rs The retransmission message.
    */
   public boolean handleEventExit(RetransStart rs) {
      lastSegmentID = rs.lastReceivedSegmentID();
      return true;
   }

   /**
    * Handle the case in which a SwitchPlan message is about to exit
    * this object (ie, in <code>putAMessage()</code>).
    * In this case, tell the IOManager to make this IO object the input
    * IO for this end of this stream.
    * @return True, indicating that the caller should ensure that the
    *         exit is not closed.
    */
   public boolean handleEventExit(SwitchPlan sp) {
      manager.makeInputIO(this);
      return true;
   }

   /**
    * This method implements the main code executed by the InputJob for this
    * object.  It is abstract and must be implemented by a sublcass.
    * @return no return value.
    * @see FMG.Conductor.dataflow.comm.InputJob
    */
   public abstract void handleInputStream();

   /**
    * This method implements the main code executed by the OutputJob for this
    * object.  It is abstract and must be implemented by a sublcass.
    * @return no return value.
    * @see FMG.Conductor.dataflow.comm.OutputJob
    */
   public abstract void handleOutputStream();

   /**
    * This method is called whenever <code>getAMessage()</code> is about
    * to block because no data is waiting to be output in either direction of
    * data flow.  This is useful for performing a flush on the output stream
    * only when no more data is available for output.
    */
   public abstract void prepareToBlock();

}
