/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.Conductor.dataflow.comm;

import java.net.*;
import java.io.*;

import FMG.IPcept.*;
import FMG.Conductor.dataflow.*;
import FMG.Conductor.comm.*;
import FMG.Conductor.util.*;
import FMG.Conductor.planning.Path;

/**
 * The Interceptor class implements a thread that intercepts connections 
 * from the local machine that are bound for a particular remote port.
 * The IPcept mechanism is used to actually intercept connections.  This
 * thread simply accepts connections from a properly configured socket
 * and hands them off to the IOManager and Path objects.
 * <p>
 * There may be more than one Interceptor object in the system at a time.
 * In particular, there will be one per port number being intercepted.
 *
 * @author Mark Yarvis (yarvis@fmg.cs.ucla.edu)
 * @version $Revision: 1.25 $
 *
 * @see FMG.Conductor.dataflow.IOManager
 * @see FMG.Conductor.planning.Path
 */

public class Interceptor extends Thread {

   /** The server socket upon which to accept connections */
   protected ServerSocket server;

   /** The address of the local node */
   private NodeID localhost = NodeID.getNewLocalID();

   /** The global object holding the list of sessions. */
   protected SessionTable sessionTable;

   /**
    * Create a new interceptor to grab connections bound for a specific
    * remote port number.  A server socket is created and the IPcept
    * classes are used to set up proxying of the desired port.  This 
    * constructor should normally be called by <code>createAndStart()</code> 
    * so that the thread is actually started up.
    * @param port The number of the remote port of interest.
    * @param sessionTable The global object holding the list of sessions.
    */
   protected Interceptor(int port, SessionTable sessionTable) {
      if (Debug.on) {
         setName("Interceptor(" + port + ")");
      }

      this.sessionTable = sessionTable;

      try {
         server = new ServerSocket(0);
         ServerSocketControl.addProxyPort(server, port);
      } catch (IOException e) {
         System.err.println("Couldn't create interceptor socket: " + e);
         if (System.getProperty("user.name").equals("root")) {
            System.err.println("   Probable solution: start the IPcept " + 
                                  "module!");
         } else {
            System.err.println("   Probable solution: run as root!");
         }
         System.exit(1);
      }
   }

   /**
    * Create a new Interceptor thread and start it up.
    * @param port The number of the remote port of interest.
    * @param sessionTable The global object holding the list of sessions.
    * @return No return value.
    */
   public static void createAndStart(int port, SessionTable sessionTable) {
      Interceptor cept = new Interceptor(port, sessionTable);
      cept.setPriority(MIN_PRIORITY);
      cept.start();
   }

   /**
    * This method is the main loop for the interceptor thread.  For each 
    * loop iteration, wait for a connection, give it a new unique id, 
    * figure out where it came from and where it wants to go to, notify the 
    * IOManager, create a Path object, and trigger planning.
    * @return No return value.  In fact, it should never return.
    */
   public void run() {

      while (true) {
         try {
            Socket s = server.accept();

            if (Debug.on) {
               Debug.println(Debug.DATAFLOW, Debug.MOREINFO,
                               "Intercepted a connection for " +
                               SocketControl.getLocalAddress(s) + " " +
                               SocketControl.getLocalPort(s));
            }

            if (Measure.on && ((Measure.mask & Measure.DATAFLOW) != 0)) {
               Measure.println("Interceptor trapped a client connection");
            }

            Session session = sessionTable.createUniqueSession();

            NodeID sourceAddr = new NodeID(SocketControl.getPeerAddress(s));
            if (sourceAddr.isLoopback()) {
               sourceAddr = localhost;
            }
            int sourcePort = SocketControl.getPeerPort(s);
            NodeID destAddr = new NodeID(SocketControl.getLocalAddress(s));
            int destPort = SocketControl.getLocalPort(s);

            /* setup path and start planning */
            Path p = new Path(session, sourceAddr, sourcePort, 
                                       destAddr, destPort, s);
            p.triggerPlanning();
         } catch (IOException e) {
            System.err.println("interceptor accept failed: " + e);
         }
      }
   }
}
