/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.Conductor.dataflow.comm;

import java.io.*;

import FMG.Conductor.comm.*;
import FMG.Conductor.dataflow.messages.*;
import FMG.Conductor.net.SocketOps;
import FMG.Conductor.util.*;

/**
 * The RawDataInputStream class is designed to allow messages to be read
 * from a raw stream (ie, directly from an unaware user application).  It 
 * bundles the raw byte stream into DataFragment objects and produces an 
 * EOF object when the end of the stream is reached.
 *
 * @author Mark Yarvis (yarvis@fmg.cs.ucla.edu)
 * @version $Revision: 1.16 $
 *
 * @see FMG.Conductor.dataflow.messages.DataFragment
 * @see FMG.Conductor.dataflow.messages.EOF
 */

public class RawDataInputStream extends MessageInputStream {

   /** The underlying input stream to read bytes from */
   protected InputStream is;

   /** Has the underlying input stream already closed? */
   protected boolean closed = false;

   /**
    * Create a new RawDataInputStream that will generate data flow messages
    * by reading from a play byte stream generated by an unaware application.
    * @param is The input stream to read bytes from.
    */
   public RawDataInputStream(InputStream is) {
      this.is = is;
   }

   /**
    * Get a message by reading from the stream.  The message with either be
    * a DataFragment or an EOF message.
    * @return A message representing data or an eof condition generated by
    *         reading the stream.
    * @exception java.io.IOException thrown if <code>readMessage()</code> 
    *                                is called after an EOF message has 
    *                                already been returned.
    * @see FMG.Conductor.dataflow.messages.DataFragment
    * @see FMG.Conductor.dataflow.messages.EOF
    */
   public Message readMessage() throws IOException {
      if (closed) {
         throw new IOException("readMessage() called closed input stream");
      }

      DataFragment m = DataFragment.getNewMessage();
      try {
         m.readRawBytes(is);
         return m;
      } catch (IOException e) {
         m.recycle();
         closed=true;
         if (Debug.on) {
            Debug.println(Debug.DATAFLOW, Debug.MOREINFO, 
                          "Raw input closing: " + e);
         }
         return EOF.getNewMessage(Direction.neutral);
      }
   }

   /**
    * Shut down the input portion of the underlying socket.  Exceptions are
    * not expected to occur and will generate a RuntimeException.
    * @return No return value.
    */
   public void shutdown() {
      try {
         SocketOps.shutdown(is);
      } catch (IOException e) {
         throw new RuntimeException("Unexpected IOException on shutdown: " + 
                                    e);
      }
   }
}
