/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.Conductor.dataflow.comm;

import java.io.*;

import FMG.Conductor.comm.*;
import FMG.Conductor.dataflow.messages.*;
import FMG.Conductor.util.Debug;
import FMG.Conductor.net.SocketOps;

/**
 * The RawDataOutputStream class is designed to allow messages to be written
 * to a raw stream (ie, directly to an unaware user application).  It is 
 * assumed that only messages that can handle a call to 
 * <code>writeToRaw()</code> should be passed to this object.
 * <p>
 * This object will, by default, automatically flush the stream after each
 * message is written.  This option can be turned off, in which case, manual
 * flushes are recommended.
 *
 * @author Mark Yarvis (yarvis@fmg.cs.ucla.edu)
 * @version $Revision: 1.14 $
 *
 * @see FMG.Conductor.comm.Message
 */

public class RawDataOutputStream extends MessageOutputStream {

   /** The underlying output stream to write bytes to. */
   OutputStream os;

   /** Is autoflush on?  Default is true */
   protected boolean autoFlush = true;

   /** 
    * Create a new RawDataOutputStream that will write data to the given
    * underlying output stream.
    * @param os The underlying output stream to write data to.
    */
   public RawDataOutputStream(OutputStream os) {
      this.os = os;
   }

   /**
    * Set the autoflush mode on or off.
    * @param autoFlush The new stat for the autoflushing mode.
    * @return No return value.
    */
   protected void setAutoFlush(boolean autoFlush) {
      this.autoFlush = autoFlush;
   }

   /**
    * Write the corresponding raw bytes for a message to the output
    * stream and, if autoflushing is turned on, flush the output stream.
    * The message must be able to accept a <code>writeToRaw()</code> call.
    * @param m The message to write.
    * @return No return value.
    * @exception java.io.IOException thrown if an error occurs while writing
    *                                of flushing the output stream.
    */
   public final void writeMessage(Message m) throws IOException {
      m.writeToRaw(os);
      if (autoFlush) {
         os.flush();
      }
   }

   /**
    * Flush the underlying output stream.  Exceptions are ignored.
    * @return No return value.
    */
   public void flush() {
      try {
         os.flush();
      } catch (IOException e) {}
   }

   /**
    * Shut down the output portion of the underlying socket.  Exceptions
    * are ignored (but may produce debugging output.
    * @return No return value.
    */
   public void shutdown() {
      try {
         flush();
         SocketOps.shutdown(os);
      } catch (IOException e) {
         if (Debug.on) {
            Debug.println(Debug.WARNING, 
                               "WARNING: shutdown failed: " + e);
         }
      }
   }
}
