/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.Conductor.dataflow.comm;

import FMG.Conductor.util.*;
import FMG.Conductor.Conductor;
import FMG.IPcept.*;
import FMG.Conductor.comm.*;
import FMG.Conductor.dataflow.*;
import FMG.Conductor.dataflow.messages.*;

// Be specific to avoid confusion with java.lang.Shutdown
import FMG.Conductor.dataflow.messages.Shutdown;

import java.net.*;
import java.io.*;

/**
 * The RawIO object is used to handle both input and output from a socket
 * connected to a user application (either the client or server endpoint).
 * Since this is an endpoint, this object will always handle both input and 
 * output.  Like any IO object, two threads are used to perform actual 
 * intput and output functions.
 * <p>
 * Remember, a failure here means real failure.  Start shutting things down.
 *
 * @author Mark Yarvis (yarvis@fmg.cs.ucla.edu)
 * @version $Revision: 1.47 $
 */
public class RawIO extends IO {

   /*
    * Client/server identification (only used on the server endpoint).
    */

   /** The address of the client node.  Only valid on the server endpoint. */
   private InetAddress sourceAddr;

   /** The client port number.  Only valid on the server endpoint. */
   private int sourcePort;

   /** The address of the server node.  Only valid on the server endpoint. */
   private InetAddress destAddr;

   /** The server port number.  Only valid on the server endpoint. */
   private int destPort;


   /*
    * Startup synchronization.
    */

   /** Tells the input thread when to start going */
   private boolean startInput = false;

   private Object startSync = new Object();

   private boolean checkForPrematureClose = 
                Conductor.getBooleanProperty("CheckForPrematureClose", true);


   /*
    * Streams
    */

   /** Output stream to write messages to the raw socket */
   protected RawDataOutputStream os;

   /** Input stream to read messages from the raw socket */
   protected RawDataInputStream is;

   /*
    * other stuff
    */

   /** An ack waiting to be sent on its way */
   private int ackID = -1;

   private Thread myThread;

   /** A queue of messages yet to be delivered to the user. */
   private MessageQueue messageQueue = new MessageQueue();

   /** is reliability enabled in this instance of Conductor? */
   private static boolean reliabilityEnabled = Conductor.reliabilityEnabled();

   /** Messages read from the socket before the data stream has been set up. */
   MessageQueue preStartQueue = new MessageQueue();


   /*
    * For measurement use only!!!
    */

   /** Number of bytes so far in the current segment.  Measurement use only! */
   int bytesInCurrentSegment = 0;

   /** Number of block segments received.  Measurement use only! */
   int numBlockSegments = 0;

   /** Total size of received block segments.  Measurement use only! */
   int totalBlockSegmentSize = 0;

   /** Total size of the original stream.  Measurement use only! */
   int originalStreamLen = 0;

   /** Number of single-byte segments received.  Measurement use only! */
   int numSingleSegments = 0;

   /** 
    * Number of chunks of single-byte segments received.  Measurement 
    * use only! 
    */
   int numSingleSegmentChunks = 0;

   /**
    * The number of acks received.  Measurement use only.
    */
   int acksReceived = 0;

   /**
    * The number of bytes that received delayed delivery as part of a 
    * partial segment.  Measurement use only!
    */
   long bytesDelayed = 0;

   /**
    * The total delay time for all bytes delivered as part of a 
    * partial segment.  Measurement use only!
    */
   long totalDelayTime = 0;


   /**
    * Create a RawIO object to handle the client endpoint.  Since the client
    * endpoint is created when a connection is received, the main input here
    * is a socket that is already open.  The input and output threads are
    * created in this constructor.
    * @param s The socket connected to the client application.
    * @param session The session object for this connection.
    * @param upstreamBoundary The boundary from which data will be received
    *                         for output.
    */
   public RawIO(Socket s, Session session, UpstreamBoundary upstreamBoundary) {
      super(session, upstreamBoundary, Direction.down);
      setSocket(s);
      startThreads();
   }

   /**
    * Create a RawIO object to handle the server endpoint.  This endpoint
    * is expected to create a connection to the server, so appropriate 
    * addresses for the connection are given.  The socket is actually
    * created by the output thread, allowing the object to be created
    * quickly.  The input and output threads are created by this constructor.
    * @param sourceAddr The address of the client.
    * @param sourceAddr The client port number.
    * @param destAddr The address of the intended server.
    * @param destPort The port number of the intended server.
    * @param session The session object for this connection.
    * @param upstreamBoundary The boundary we are downstream of.
    */
   public RawIO(InetAddress sourceAddr, int sourcePort, 
                InetAddress destAddr, int destPort, 
                Session session, UpstreamBoundary upstreamBoundary) {
      super(session, upstreamBoundary, Direction.up);
      this.sourceAddr = sourceAddr;
      this.sourcePort = sourcePort;
      this.destAddr = destAddr;
      this.destPort = destPort;
      startThreads();
   }

   /**
    * Create the input and output streams needed for reading and writing
    * messages to and from this raw socket.  This method is called after 
    * the socket is set but before other threads are notified of this fact.
    * Any resulting exception will cause a RuntimeExecption to be thrown.
    * @return no return value.
    */
   protected void createStreams() {
      try {
         os = new RawDataOutputStream(s.getOutputStream());
         is = new RawDataInputStream(s.getInputStream());
      } catch (IOException e) {
         throw new RuntimeException(e.toString());
      }
   }

   /**
    * Notify the input thread that it is ok to start reading from the
    * socket.
    * @return No return value.
    */
   public void startInput() {
      if (Measure.on && ((Measure.mask & Measure.PLANNING) != 0)) {
         Measure.TimeOfDay tod = Measure.getTimeOfDay();
         Measure.println("Data flow started at " + 
                         tod.secs + "." + tod.usecsString());
      }

      // push out any queued messages
      while (! preStartQueue.isEmpty()) {
         putAMessage(preStartQueue.dequeue());
      }

      synchronized (startSync) {
         startInput = true;
         startSync.notify();
      }
   }

   /**
    * This method is the main function for the input thread.
    * Once the socket has been set, and we are notified that we can start
    * reading, messages are read and forwarded to the appropriate endpoint
    * until the input is shut down.
    * @return No return value.
    * @see FMG.Conductor.dataflow.comm.InputJob
    */
   public void handleInputStream() {
      Thread currentThread = Thread.currentThread();

      waitForSocket();

      if (Measure.on && ((Measure.mask & Measure.DATAFLOW) != 0)) {
         Measure.println("RawIO data flow begins ");
      }

      if (Debug.on) {
         Debug.println(Debug.DATAFLOW, Debug.MOREINFO, 
                       "** RawIO: input starting up");
      }
      try {
// TODO: what if a user pushes out his request and then does a half close
//       ... maybe I should look for his input to close?
//           -- perhaps do this AFTER I notice the ouptut closed?

         if (! checkForPrematureClose) {
            synchronized (startSync) {
               while (! startInput) {
                  try {
                     startSync.wait();
                  } catch (InterruptedException ignore) {}
               }
            }
         }

         // path isn't set up yet ... queue messages
         while ((is!=null) && !startInput) {
            Message m = is.readMessage();
            if (!startInput && (m instanceof EOF) && manager.isRootNode()) {
               /*
                * The client closed before it got started.
                * Get rid of the data path.  Once it's gone, the path
                * object will also be killed.
                */

               if (Debug.on) {
                  Debug.println(Debug.DATAFLOW, Debug.WARNING,
                                "RawIO: Detected pre-mature close!");
               }

               manager.makeInputIO(this);
               putAMessage(Shutdown.getNewMessage(Direction.neutral));

               return;
            }
            preStartQueue.enqueue(m);
         }

         // push out any queued messages
         while (! preStartQueue.isEmpty()) {
            putAMessage(preStartQueue.dequeue());
         }

         while (is!=null) {
            Message m = is.readMessage();
            if (Debug.on) {
               Debug.println(Debug.DATAFLOW, Debug.DETAIL, 
                             "RawIO: got " + m);
            }
            putAMessage(m);
         }
         if (Debug.on) {
            Debug.println(Debug.DATAFLOW, Debug.MOREINFO,
                          "** RawIO: input handler ended");
         }
      } catch (IOException e) {
         throw new RuntimeException("Case not handled");
      }
   }

   /**
    * This method is the main code for the output thread.  If we're the
    * server endpoint, a socket connection to the server is created
    * (masqueraded if possible).  After startup, messages are pulled
    * one-by-one from the boundaries, handled locally, and pushed out the
    * socket until the upstream boundary is closed.
    * @return No return value.
    * @see FMG.Conductor.dataflow.comm.OutputJob
    */
   public void handleOutputStream() {
      myThread = Thread.currentThread();
      boolean messagesAreQueued = true;
      try {
         if (s == null) {   /* am I the server end? */
            try {
               // Try to masquerade as the source
               if (Debug.on) {
                  Debug.println(Debug.DATAFLOW, Debug.MOREINFO,
                                "Attempting to masquerade as " + sourceAddr);
               }
               setSocket(SocketControl.createUnproxiedSocket(
                                 destAddr, destPort, sourceAddr, sourcePort));
            } catch (Exception e) {
               // Fall back on normal method
               if (Debug.on) {
                  Debug.println(Debug.DATAFLOW, Debug.WARNING,
                                "Warning: Failed to masquerade as source");
                  Debug.println(Debug.DATAFLOW, Debug.MOREINFO, 
                                "Error was: " + e);
               }
               setSocket(new Socket(destAddr, destPort));
            }
         }

         if (Debug.on) {
            Debug.println(Debug.DATAFLOW, Debug.MOREINFO,
                          "** RawIO: output starting up");
         }
         while (!upstreamClosed) {
            Message m = getAMessage();

            if (m != null) {
               if (upstreamBoundary.isBetweenSegments() || 
                   (! reliabilityEnabled)) {

                  // if there might be messages queued, send them first
                  if (messagesAreQueued) {
                     writeQueuedMessages();
                     messagesAreQueued = false;
                  }

                  writeMessage(m);
               } else {
                  enqueueMessage(m);
                  messagesAreQueued = true;
               }
            }
         }
      } catch (IOException e) {
         putAMessage(EOF.getNewMessage(Direction.up));
      } finally {
         if (Debug.on) {
            Debug.println(Debug.DATAFLOW, Debug.MOREINFO,
                          "** RawIO: output handler ended");
         }

         if (Measure.on && ((Measure.mask & Measure.SEGMENTATION) != 0)) {
            Measure.println("Original stream length: " + 
                                                      originalStreamLen);
            Measure.println("Number of one-byte segments: " + numSingleSegments
                            + " (in " + numSingleSegmentChunks + " chunks)");
            Measure.println("Number of block segments: " + 
                                                      numBlockSegments);
            Measure.println("Total bytes received in block segments: " + 
                                                      totalBlockSegmentSize);
            Measure.println("Number of acks received: " + acksReceived);
            Measure.println("Number of partial-segment bytes that received " +
                            "delayed delivery: " + bytesDelayed);
            Measure.println("Total delivery delay for all bytes that " +
                            "received delayed delivery: " + totalDelayTime);
         }
      }
      manager.connectionEnded(dir.isDown());
   }

   /**
    * Deliver a message to the user.
    * @return No return value.
    * @exception java.io.IOException if an error occurs while writing the 
    *                                data stream.
    */
   protected final void writeMessage(Message m) throws IOException {
      if (Debug.on) {
         Debug.println(Debug.DATAFLOW, Debug.DETAIL, "RawIO: delivering " + m);
      }

      os.writeMessage(m);

      ((Recyclable) m).recycle();
   }

   /**
    * Deliver any messages previously queued.  Normally queuing occurs until
    * the end of a segment is reached.
    * @return No return value.
    * @exception java.io.IOException if an error occurs while writing the 
    *                                data stream.
    */
   protected final void writeQueuedMessages() throws IOException {
      if (Debug.on && ! messageQueue.isEmpty()) {
         Debug.println(Debug.DATAFLOW, Debug.DETAIL, 
                       "RawIO: End of segment reached, " + 
                       "sending queued messages.");
      }

      long now = 0;
      if (Measure.on && ((Measure.mask & Measure.SEGMENTATION) != 0)) {
         now = System.currentTimeMillis();
      }

      while (! messageQueue.isEmpty()) {
         Message m = messageQueue.dequeue();

         if (Measure.on && ((Measure.mask & Measure.SEGMENTATION) != 0)) {
            if (((StreamElem) m).isData()) {
               DataFragment df = (DataFragment) m;
               bytesDelayed += df.length();
               totalDelayTime += df.length() * (now - df.timestamp);
            }
         }

         writeMessage(m);
      }
   }

   /**
    * Discard (and recycle) any message that were queued for delivery.
    * @return no return value.
    */
   protected void clearQueuedMessages() {
      while (! messageQueue.isEmpty()) {
         Message m = messageQueue.dequeue();
         ((Recyclable) m).recycle();
      }
   }

   /**
    * Enqueue a message for later delivery to the user.  This method is used
    * to deal with partial data segments and control messages received in
    * the middle of segments (are there any of these?).
    * @param m The message to queue.
    * @return No return value.
    */
   protected final void enqueueMessage(Message m) {
      if (Debug.on) {
         Debug.println(Debug.DATAFLOW, Debug.DETAIL, 
                       "RawIO: enqueuing message until end of segment " + m);
      }

      if (Debug.on && (Debug.getLevel(Debug.DATAFLOW) > Debug.WARNING)) {
         if (! ((StreamElem) m).isData()) {
            Debug.println("WARNING: RawIO enqueuing non data object " + m + 
                          " ... is this OK?");
         }
      }

      if (Measure.on && ((Measure.mask & Measure.SEGMENTATION) != 0)) {
         if (((StreamElem) m).isData()) {
            ((DataFragment) m).timestamp = System.currentTimeMillis();
         }
      }

      messageQueue.enqueue(m);
   }

   /**
    * Handle the case in which a SwitchPlan message has been received
    * locally.  First, the super function is called to install local
    * adaptors.  Then, the message is either turned around or discarded.
    * Notice, it is never delivered to the application!
    * @param sp The SwitchPlan message we are handling.
    * @return Always false, indicating that this message has been handled
    *         and should not be delivered.
    */
   public boolean handleEventEntry(SwitchPlan sp) {
      super.handleEventEntry(sp);

      if (Debug.on) {
         Debug.println(Debug.DATAFLOW, Debug.EXTRAINFO,
                       "RawIO: turning around SwitchPlan");
      } 

      sp.unlink();
      if (sp.nextLeg()) {
         manager.forwardSwitchPlan(sp);
      } else {
         sp.recycle();
      }
      return false;
   }

   /**
    * Handle the case in which a SegmentAck message has been received
    * locally.  All SegmentAcks can be discarded.
    * @param incomingAck The SegmentAck message we are handling.
    * @return Always false, indicating that this message has been handled
    *         and should not be delivered.
    */
   public boolean handleEventEntry(SegmentAck incomingAck) {
      if (Debug.on) {
         Debug.println(Debug.DATAFLOW, Debug.DETAIL,
                       "RawIO: got ack: " + incomingAck);
      }
      incomingAck.unlink();
      incomingAck.recycle();

      if (Measure.on && ((Measure.mask & Measure.SEGMENTATION) != 0)) {
         acksReceived++;
      }

      return false;
   }

   /**
    * Handle the case in which a DataFragment message has been received
    * locally.  An ack is created for later delivery.  Note that the ack
    * is not actually sent until the data is deliverd by the ouptut thread.
    * @param df The DataFragment message we are handling.
    * @return Always true, indicating that this message should be delivered
    *         to the user.
    */
   public boolean handleEventEntry(DataFragment df) {

      if (Measure.on && ((Measure.mask & Measure.SEGMENTATION) != 0)) {

         bytesInCurrentSegment += df.length();

         if (df.isEndOfSegment()) {
            if (df.isBlock()) {
               numBlockSegments++;
               totalBlockSegmentSize += bytesInCurrentSegment;
            } else {
               numSingleSegmentChunks++;
               numSingleSegments += bytesInCurrentSegment;
            }

            originalStreamLen = df.segmentID();
            bytesInCurrentSegment = 0;
         }
      }

      if (df.isEndOfSegment()) {
         if (reliabilityEnabled) {
            ackID = df.segmentID();
         }
      }
      return true;
   }

   /**
    * Handle the case in which an EOF message has been received
    * locally.  If this is an upward flowing EOF, the input part of
    * the socket is closed, otherwise, the output part is closed.
    * Note that this message is allowed to pass to the raw output
    * stream, though it will just be destroyed there.
    * @param eof The EOF message we are handling.
    * @return Always true, indicating that this message should be delivered
    *         to the user.  Kind of stupid, since the output stream will
    *         just throw it away.
    * @exception FMG.Conductor.dataflow.AsyncEventException
    *                        Will never actually be thrown.
    */
   public boolean handleEventEntry(EOF eof) throws AsyncEventException {
      super.handleEventEntry(eof);

      if (eof.flowDir().isUp()) {
         if (Debug.on) {
            Debug.println(Debug.DATAFLOW, Debug.MOREINFO,
                          "** RawIO: shutdown(input) for reverse close");
         }
         is.shutdown();
         putAMessage(Shutdown.getNewMessage(Direction.neutral));
         return true;
      } else {
         os.shutdown();
         putAMessage(Shutdown.getNewMessage(Direction.up));
         return true;
      }
   }

   /**
    * Handle the case in which an Shutdown message has been received
    * locally.  This message indicates that no more messages should be
    * expected on a particular endpoint.  So, that endpoint is closed.
    * @param sd The Shutdown message to be handled.
    * @return Always returns false, indicating that this message has
    *         been handled.
    */
   public boolean handleEventEntry(Shutdown sd) {
      if (sd.flowDir().isUp()) {
         close(Direction.down);
      } else {
         close(Direction.up);
      }

      sd.unlink();
      sd.recycle();
      return false;
   }

   /**
    * Handle the case in which a failure has been signaled from upstream.
    * When this occurs, it must be the case that we have received part of
    * a segment.  Throw away the partial segment and request retransmission
    * starting after the last full segment received.  Then wait for the
    * start of retransmission.
    * @param sf The message signaling the failure.
    * @return Always returns false indicating that the message has been 
    *         handled.
    */
   public boolean handleEventEntry(SignalFailure sf) {
      sf.unlink();
      sf.recycle();

      clearQueuedMessages();

      RetransRequest rr = 
         RetransRequest.getNewMessage(upstreamBoundary.lastSegmentID());

      upstreamBoundary.putOneOut(rr);

      // wait for RetransStart
      StreamElem rs = waitForMessage(Direction.up, RetransStart.class);
      rs.unlink();

      // service anything we've been neglecting
      needsService(upstreamBoundary);
      needsService(downstreamBoundary);

      return false;
   }

   public boolean handleEventEntry(RetransStart rs) {
      rs.unlink();
      return false;
   }

   /**
    * A RawIO should never receive a RetransRequest message.
    */
   public boolean handleEventEntry(RetransRequest rr) {
      throw new RuntimeException("This should never happen!");
   }

   /**
    * Handle the case in which an EOF object is leaving this object and 
    * entering the local data stream.  If this is a neutral EOF, record
    * that fact the input stream has closed.
    * @param eof The EOF message we are handling.
    * @return Always returns true, indicating that the boundary should be
    *         checked for closure first.
    */
   public boolean handleEventExit(EOF eof) {
      if (eof.flowDir().isNeutral()) {
         is = null;
      }
      return true;
   }

   /**
    * Handle the case in which an Shutdown object is leaving this object and 
    * entering the local data stream.  Close the appropriate boundary 
    * (depending on the flow direction of the message).
    * @param sd The Shutdown message to be handled.
    * @return Always returns false, indicating that the boundary should not be
    *         checked for closure first.
    */
   public boolean handleEventExit(Shutdown sd) {
      if (sd.flowDir().isUp()) {
         close(Direction.up);
      } else {
         close(Direction.down);
      }

      return false;
   }

   /**
    * If we're about to block, do nothing.  The output stream uses auto
    * flushing, so we don't have to flush it.
    * @return No return value.
    */
   public void prepareToBlock() {
      if (ackID != -1) {
         myThread.yield();
         putAMessage(SegmentAck.getNewMessage(ackID));
         ackID = -1;
      }
   }

}
