/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.Conductor.dataflow.messages;

import java.io.*;
import java.net.*;

import FMG.Conductor.comm.*;
import FMG.Conductor.util.*;

/**
 * The ConnectionInit object is a message that travels along the data path.
 * It is used when a data connection between neighboring Conductor nodes is 
 * created to indicate which user-level connection (ie, which datapath) the
 * "link-level" connection serves.
 * <p>
 * This object is fully recyclable to avoid frequent re-allocation.
 *
 * @author Mark Yarvis (yarvis@fmg.cs.ucla.edu)
 * @version $Revision: 1.6 $
 */

public class ConnectionInit extends Recyclable implements Message {

   /** The id of the connection this datapath serves */
   public ConnID connID = new ConnID();

   /** 
    * The direction this message is traveling.  Upstream is toward the server
    * and downstream is toward the client 
    */
   public Direction dir;

   /** The recycler for this class.  */
   private static Recycler recycler = new Recycler(ConnectionInit.class);

   /** The dynamically assigned id for this message type. */
   private static int messageID = -1;

   /**
    * Register a prototype of this message class with BasicMessageInputStream
    * so that it can be read and written.
    * @return No return value.
    * @see FMG.Conductor.comm.BasicMessageInputStream
    */
   public static void registerSelf() {
      messageID = BasicMessageInputStream.
                        registerMessagePrototype((Message) recycler.getOne());
   }

   /**
    * Return the id that was assigned to messages of this type by the 
    * BasicMessageInputStream.
    * @return The dynamically assigned id of this message type.
    * @see FMG.Conductor.comm.BasicMessageInputStream
    */
   public int getMessageID() {
      return messageID;
   }

   /**
    * Read a message of this type from the input stream.  Presumably the 
    * message was written using <code>sendSelf()</code>.  This method is
    * normally called on the prototype that was registered with 
    * BasicMessageInputStream.
    * @param is The input stream to read from.
    * @return A new message of type ConnectionInit that contains the data
    *         read from the stream.
    * @exception java.io.IOException if an error occurs while reading from the
    *                                stream.
    * @see FMG.Conductor.comm.BasicMessageInputStream
    */
   public Message readNewMessage(DataInputStream is) throws IOException {
      return ConnectionInit.getNewMessage(is);
   }
   
   /**
    * Create a new (or recycled) ConnectionInit message.
    * @param connID The id of the connection this message pertains to.
    * @param dir The direction this message will be traveling.
    * @return A new (or recycled) ConnectionInit message, initialized
    *         from the parameters.
    */
   public static ConnectionInit getNewMessage(ConnID connID, Direction dir) {
      ConnectionInit m = (ConnectionInit) recycler.getOne();
      m.dir = dir;
      m.connID.copy(connID);
      return m;
   }

   /**
    * Read a new ConnectionInit message from a socket.
    * <p>
    * <b>Note:</b> This is a big HACK allowing us to read a message
    * from the socket before a permanent DataFlowInputStream is created.
    * This is useful because we need this message to decide which connection
    * the new socket belongs to.  So, no ConductorIO object has yet been
    * created or identified.
    * @param s The socket to obtain the message from.
    * @return A new (or recycled) ConnectionInit object initialized by
    *         reading from the socket.
    * @exception java.io.IOException if an error occurs while reading from
    *                                the socket.
    */
   public static ConnectionInit getNewMessage(Socket s) throws IOException {
      DataInputStream is = new DataInputStream(s.getInputStream());
      return getNewMessage(is);
   }

   /**
    * Read a message of this type from the input stream.  Presumably the 
    * message was written using <code>sendSelf()</code>.
    * @param is The input stream to read from.
    * @return A new message of type ConnectionInit that contains the data
    *         read from the stream.
    * @exception java.io.IOException if an error occurs while reading from the
    *                                stream.
    */
   public static ConnectionInit getNewMessage(DataInputStream is) 
                                                      throws IOException {
      ConnectionInit m = (ConnectionInit) recycler.getOne();
      m.receiveSelf(is);
      return m;
   }

   /**
    * Initialize the data members of this object from the input steram.  
    * Presumably the data was written using <code>sendSelf()</code>.
    * @param is The input stream to read from.
    * @return No return value.
    * @exception java.io.IOException if an error occurs while reading from the
    *                                stream.
    */
   public void receiveSelf(DataInputStream is) throws IOException {
      connID.readSelf(is);
      dir = Direction.readDir(is);
   }

   /**
    * Write the data members of this object to the output stream.
    * Presumably the data will be read using <code>receiveSelf()</code>.
    * @param os The output stream to write to.
    * @return No return value.
    * @exception java.io.IOException if an error occurs while writing to the
    *                                stream.
    */
   public void sendSelf(DataOutputStream os) throws IOException {
      connID.writeSelf(os);
      dir.writeSelf(os);
   }

   /**
    * This is not a valid call for objects of this type.  A RuntimeException
    * will result.
    * @param os The output stream to write to.
    * @return No return value.
    * @exception java.io.IOException if an error occurs while writing to the
    *                                stream.
    */
   public void writeToRaw(OutputStream os) throws IOException {
      throw new RuntimeException("This souldn't happen!!!");
   }
}
