/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.Conductor.dataflow.messages;

import java.io.*;

import FMG.Conductor.util.*;

/**
 * The DataBlock object stores a block of data as it moves through the
 * system.  Typically, DataFragment objects will point to subsections of
 * a given DataBlock.
 * <p>
 * Care must be taken to not change the contents of a DataBlock.  You never
 * know who (maybe another DataFragment, maybe a cache) might be pointing
 * at it.
 * <p>
 * The DataBlock object is recyclable so we don't have to keep re-allocating
 * the large byte array.
 *
 * @author Mark Yarvis
 * @version $Revision: 1.11 $
 *
 * @see FMG.Conductor.dataflow.messages.DataBlock
 */

public class DataBlock extends Recyclable {

   /** The recycler for this class. */
   private static Recycler recycler = new Recycler(DataBlock.class);

   /** The maximum number of bytes that can be stored in a DataBlock */
   protected static final int MAX_LEN = 10240;

   /**
    * The actual data.  Warning, for efficiency, this is sometimes directly
    * accessed by DataFragment object.
    */
   byte[] data = new byte[MAX_LEN];

   /**
    * Get a new or recycled DataBlock object.
    */
   public static DataBlock getNewMessage() {  
      return (DataBlock) recycler.getOne();
   }

   /**
    * Create a new data block with the given portion of the data block copied.
    * @param first The position of the first byte to copy.
    * @param last The position of the last byte to copy.
    * @return A new DataBlock object with the specified portion of the data.
    */
   public DataBlock duplicate(int first, int last) {
      DataBlock db = getNewMessage();
      System.arraycopy(data, first, db.data, first, last - first + 1);
      return db;
   }

   /**
    * Read some number of bytes from an input stream and store them in this
    * object.  Note that exactly <code>len</code> bytes are read.
    * @param is The input stream to read from.
    * @param len The number of bytes to read.
    * @return The number of bytes read, which is always <code>len</code>
    * @exception java.io.IOException if an error occurs while reading from
    *                                the stream.
    */
   int readBytes(InputStream is, int len) throws IOException {
      if (len > MAX_LEN) {
         throw new RuntimeException("Message.readBytes(): " +
                                    "Received more bytes than could be read");
      }

      int pos = 0;
      int readCount;
      while (pos < len) {
         readCount = is.read(data, pos, len-pos);
         if (readCount < 0) {
            throw new EOFException();
         }
         pos += readCount;
      }

      return len;
   }
   
   /**
    * Read some bytes and store them in this object.  Whatever bytes are
    * available will be read.
    * @param is The input stream to read from.
    * @return The number of bytes read.
    * @exception java.io.IOException if an error occurs while reading from
    *                                the stream.
    */
   int readBytes(InputStream is) throws IOException {
      int len = is.read(data, 0, MAX_LEN);
      if (len < 0) {
         throw new EOFException();
      }
      return len;
   }

   /**
    * Copy bytes from part of a byte array into this object.  Since this 
    * object has a finite storage capacity, it may not be able to store all 
    * of the bytes given to it.
    * @param bytes The source array to copy from.
    * @param start The position of the first byte to copy in the source array.
    * @param len The number of bytes to copy.
    * @return The number of bytes copied.  This will not be more than MAX_LEN.
    */
   public int storeBytes(byte[] bytes, int start, int len) {
      if (len > MAX_LEN) {
         len = MAX_LEN;
      }
      if (len > 0) {
         System.arraycopy(bytes, start, data, 0, len);
      }

      return len;
   }

   /**
    * Copy a portion of the contents of this object into a byte array.  It 
    * is assumed that the target array is big enough to receive the requested
    * data.
    * @param bytes The target byte array.
    * @param targetPos The position in the array to receive the first byte.
    * @param start The position of the first byte of source data to copy.
    * @param len The number of bytes to copy.
    * @return No return value.
    */
   public void getBytes(byte[] bytes, int targetPos, int start, int len) {
      System.arraycopy(data, start, bytes, targetPos, len);
   }
   
   /**
    * Write a portion of the contents of this object to the output stream.
    * @param os The output stream to write to.
    * @param first The index of the first byte to write.
    * @param last The index of the last byte to write.
    * @return No return value.
    * @exception java.io.IOException if an error occurs while writing to 
    *                                the stream.
    */
   void writeBytes(OutputStream os, int first, int last) 
                                throws IOException {
      os.write(data, first, last - first + 1);
   }

   /**
    * Create a string from a portion of the data in this object.
    * @param first The index of the first byte to copy to the string.
    * @param last The index of the last byte to copy to the string.
    * @return A new String object containing the requested  portion of the 
    *         data from this object.
    */
   public String toString(int first, int last) {
      return new String(data, first, last-first+1);
   }

}
