/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.Conductor.dataflow.messages;

import java.io.*;

import FMG.Conductor.comm.*;
import FMG.Conductor.util.*;
import FMG.Conductor.dataflow.*;
import FMG.Conductor.dataflow.comm.IO;

/**
 * The DataFragment class is a pointer to a portion of a DataBlock.  It 
 * doesn't actually store data, but it represents a particular chunk of 
 * data traveling along the data path.  By using indirection, a block of
 * data may be broken up and reassembled without doing any data copying.
 * <p>
 * This object also plays a crucial role in <i>semantic segmentation</i>.
 * Care must be taken when combining adjacent DataFragments that proper
 * segmentation is preserved.  Each chunk of data is also either a part of 
 * a block-<i>segment</i> (when segment combination has already occurred) or 
 * a bunch of one-byte <i>segments</i>.  So, this object must keep track of 
 * which type of segment it is part of.  If the bytes are a part of a 
 * block-<i>segment</i>, this object must also keep track of the id of
 * the <i>segment</i> the bytes belong to and whether this is the last chunk 
 * of that segment.  In actuality, we only keep track of the segment id 
 * in the last chunk of the segment.  If the bytes are a bunch of one-byte 
 * <i>segments</i>, we must keep track of the last byte of the segment.  
 * <p>
 * Note that for efficiency when transmitting a DataFragment, the segment 
 * id is only transmitted in the last chunk of a block-<i>segment</i>.  It is
 * assumed that chunks of one-byte <i>segments</i> begin with the next
 * segment id after the previous chunk.  So, the receiver can keep
 * track for those chunks.  Further, chunks that do not end a 
 * block-<i>segment</i> do not require an id, since we only care about the
 * id upon complete reception of a segment.
 * <p>
 * This object is fully recyclable to prevent frequent re-allocation.
 *
 * @author Mark Yarvis (yarvis@fmg.cs.ucla.edu)
 * @version $Revision: 1.35 $
 * @see FMG.Conductor.dataflow.messages.DataBlock
 */

public class DataFragment extends StreamElem implements Message {

   /** The recycler for objects of this class */
   private static Recycler recycler = new Recycler(DataFragment.class);

   /** The DataBlock object holding the data we're pointing to */
   DataBlock data;

   /** 
    * Is the underlying data block write protected?  If so, we must duplicate
    * it before we can modify the bytes contained.
    */
   boolean writeProtected;

   /**
    * The index of the first byte in the DataBlock we're pointing to.<p>
    * TODO: should this be a short?
    */
   int first;

   /**
    * The index of the last byte in the DataBlock we're pointing to.<p>
    * TODO: should this be a short?
    */
   int last;

   /** 
    * This mask is used for encoding and deocding the blockSegment boolean 
    * for transmission in a byte.
    */
   private static final byte IS_BLOCKSEG = 0x1;

   /** 
    * This mask is used for encoding and deocding the endOfSegment boolean 
    * for transmission in a byte.
    */
   private static final byte IS_EOS      = 0x2;

   /**
    * True if this is a block-segment.  False if this is a bunch of 
    * single-byte segments.
    */
   protected boolean blockSegment;

   /* 
    * True if this is the last fragment in a block-segment as well as for 
    * all non-block-segments.
    */
   protected boolean endOfSegment;

   /**
    * The segment id for this bunch of bytes.  For block-segments, this is 
    * the id of the last combined segment, but it is only valid when
    * endOfSegment is true.  If these bytes are single-byte segments, this
    * is the id of the last byte.
    */
   protected int segmentID;

   /**
    * A timestamp that can be used for measurement purposes only.  The value
    * is never trasnsmitted.
    */
   public long timestamp;

   /** The message id dynamicall assigned by BasicMessageInputStream */
   private static int messageID = -1;

   /** 
    * Register a prototype of this message class with BasicMessageInputStream
    * so that messages of this type can be read and written.  Record the
    * messageID to be returned by <code>getMessageID()</code>.
    * @return No return value.
    * @see FMG.Conductor.comm.BasicMessageInputStream
    */
   public static void registerSelf() {
      messageID = BasicMessageInputStream.
                        registerMessagePrototype((Message) recycler.getOne());
   }

   /**
    * Get the id assiged to messages of this type by BasicMessageInputStream.
    * @return The id of this message type.
    * @see FMG.Conductor.comm.BasicMessageInputStream
    */
   public int getMessageID() {
      return messageID;
   }

   /**
    * Read a message of this type from the input steram.  Presumably the
    * message was written using <code>sendSelf()</code>.  This method is
    * normally called on the prototype that was registered with
    * BasicMessageInputStream.
    * @param is The input stream to read from.
    * @return A new message of type DataFragment that contains the data
    *         read from the stream.
    * @exception java.io.IOException if an error occurs while reading from the
    *                                stream.
    * @see FMG.Conductor.comm.BasicMessageInputStream
    */
   public Message readNewMessage(DataInputStream is) throws IOException {
      return DataFragment.getNewMessage(is);
   }

   /**
    * Create an empty DataFragment object, pointing to no bytes in an
    * uninitialized DataBlock.
    * @return A new, empty, DataFragment.
    */
   public static DataFragment getNewMessage() {
      DataFragment df = (DataFragment) recycler.getOne();
      df.data = DataBlock.getNewMessage();
      df.writeProtected = false;
      df.first = 0;
      df.last = 0;

      df.segmentID = 0;

      return df;
   }

   /**
    * Read a DataFragment from the input stream.  Presumably the message
    * was written using <code>sendSelf()</code>.
    * @param is The input stream to read from.
    * @return A new message of type DataFragment that contains the data
    *         read from the stream.
    * @exception java.io.IOException if an error occurs while reading from the
    *                                stream.
    */
   public static DataFragment getNewMessage(DataInputStream is)
                                                throws IOException {
      DataFragment df = (DataFragment) recycler.getOne();
      df.receiveSelf(is);
      df.writeProtected = false;

      return df;
   }

   /**
    * Get a new DataFragment object initialized with a portion of the
    * contents of a byte array.  Note that, if the length of the source
    * array is too large, the new DataFragment may only contain a part
    * of the requested data.  You should call <code>length()</code> on
    * the returned object to see how many bytes were actually copied. 
    * <p>
    * The new segment will be a block segment, but it will not be the end of 
    * the segment.  To make it the end of a segment, call 
    * <code>makeBlockEnd()</code>.
    * @param bytes The byte array to copy data from.
    * @param start The first byte in the source arry to copy.
    * @param len The number of bytes to copy.
    * @return A new DataFragment object containing as many of the requested
    *         data bytes as would fit.
    */
   public static DataFragment getNewMessage(byte[] bytes, int start, int len) {
      if ((len < 0) || (start < 0)) {
         throw new RuntimeException("This shouldn't happen");
      }

      DataFragment df = (DataFragment) recycler.getOne();
      df.data = DataBlock.getNewMessage();
      df.writeProtected = false;
      df.first = 0;
      len = df.data.storeBytes(bytes, start, len);
      df.last = len - 1;
      df.blockSegment=true;
      df.endOfSegment=false;   // by default
      df.segmentID = -1;

      return df;
   }

   public DataFragment duplicate() {
      DataFragment df = (DataFragment) recycler.getOne();

      // Share the same data block, but use copy-on-write
      df.data = data;
      data.retain();             // keep a handle on this until we're done
      df.writeProtected = true;  // don't allow the data to be modified
      writeProtected = true;     //   ... here too

      df.first = first;
      df.last = last;

      df.blockSegment = blockSegment;
      df.endOfSegment = endOfSegment;
      df.segmentID = segmentID;

      return df;
   }

   /**
    * Split a fragment into two.  This fragment will now consist of
    * the first <i>length</i> bytes.  A new fragment is returned containing
    * the tail bytes.
    * @param length The length to make the current object.
    * @return A new DataFragment object containing the bytes removed from
    *         the tail of this object.
    */
   public DataFragment split(int length) {
      if ((length <= 0) || (length >= last - first + 1)) {
         throw new RuntimeException("This shouldn't happen");
      }

      // Create tail fragment
      DataFragment df = (DataFragment) recycler.getOne();
      df.data = data;
      df.data.retain();     /* retain data until fragment is recycled */
      df.first = first + length;
      df.last = last;

      df.writeProtected = writeProtected;

      // set reliability attributes of tail
      df.blockSegment = blockSegment;
      df.segmentID = segmentID;
      df.endOfSegment = endOfSegment;

      // Truncate original fragment
      last = first + length - 1;

      // fix reliability attributes of original fragment
      if (blockSegment) {
         endOfSegment = false;
         segmentID = -1;
      } else {
         segmentID = segmentID - (df.last - last);
         if (Debug.on && (segmentID < 0)) {
            Debug.println("WARNING: In DataFragment.split(), segmentID = " + 
                           segmentID);
         }
      }

      insert(df, Direction.up);

      return df;
   }

   /**
    * Get the byte at a given position in this fragment.
    * @param pos The index of the desired byte.
    * @return The byte at the given index.
    */
   public byte getByte(int pos) {
      return data.data[first + pos];
   }

   /**
    * Copy a bunch of bytes from this fragment into a byte array.
    * @param bytes The target byte array.
    * @param targetPos The position of the first byte to copy in the target.
    * @param start The index of the first byte to copy from the DataFragment.
    * @param len The number of bytes to copy.
    * @return No return value.
    */
   public void getBytes(byte[] bytes, int targetPos, int start, int len) {
      if (len > last - first + 1) {
         throw new RuntimeException("this shouldn't happen");
      }
      data.getBytes(bytes, targetPos, start + first, len);
   }

   /**
    * Set the byte at the given position in the DataFragment to a new value.
    * @param pos The index of the byte to modify.
    * @param b The new value for the byte at the given position.
    * @return No return value.
    */
   public void setByte(int pos, byte b) {
      // if the underlying data is write protected, copy it first
      if (writeProtected) {
         DataBlock oldData = data;
         data = data.duplicate(first, last);
         oldData.recycle();      // stop holding a pointer to the old block
         writeProtected = false;
      }
      data.data[first + pos] = b;
   }

   /**
    * Initialize the data members of this object from the input steram.
    * Presumably the data was written using <code>sendSelf()</code>.
    * @param is The input stream to read from.
    * @return No return value.
    * @exception java.io.IOException if an error occurs while reading from the
    *                                stream.
    */ 
   public void receiveSelf(DataInputStream is) throws IOException {
      byte flags = is.readByte();

      /* interpret flags, and decode related data */
      blockSegment = ((flags & IS_BLOCKSEG) > 0);
      endOfSegment = ((flags & IS_EOS) > 0);
      if (blockSegment && endOfSegment) {
         segmentID = is.readInt();
      } else {
         /**
          * If this is a block segment but not end-of-segment, then this
          * is undefined.  If this is a non-block segment, then this variable
          * will be set later by informOfPrevioiusSegmentID().
          */
         segmentID = -1;
      }

      /* get the data */
      int len = (int) is.readShort();

      if (len < 0) {
         throw new RuntimeException("This shouldn't happen");
      }
      data = DataBlock.getNewMessage();
      data.readBytes(is, len);
      first = 0;
      last = len - 1;
   }

   /**
    * Write the data members of this object to the output stream.
    * Presumably the data will be read using <code>receiveSelf()</code>.
    * @param os The output stream to write to.
    * @return No return value.
    * @exception java.io.IOException if an error occurs while writing to the
    *                                stream.
    */
   public void sendSelf(DataOutputStream os) throws IOException {
      byte flags = (byte) ((endOfSegment ? IS_EOS : 0) | 
                           (blockSegment ? IS_BLOCKSEG : 0));

      os.writeByte(flags);

      if (blockSegment && endOfSegment) {
         os.writeInt(segmentID);
      }

      /* send data */
      os.writeShort((short) (last - first + 1));
      data.writeBytes(os, first, last);
   }

   /**
    * Initialize this object by simply reading raw bytes from an input
    * stream.  The DataFragment will be considered to be a collection of
    * one-byte <i>segments</i>
    * @param is The input stream to read bytes from.
    * @return No return value.
    * @exception java.io.IOException if an error occurs while reading from
    *                                the stream.
    */
   public void readRawBytes(InputStream is) throws IOException {
      first = 0;
      last = data.readBytes(is) - 1;

      blockSegment = false;
      endOfSegment = true;

      segmentID = -1; // will be set later in informOfPreviousSegmentID()
   }

   /**
    * Generate a string representing the contents of this DataFragment.
    * This method should only be used for debugging.
    * @return A new string representing the DataFragment.
    */
   public String toString() {
      if (data == null) {
         return "DataFragment[]";
      }

      /* create a string buffer of approximately the right size */
      StringBuffer sb = new StringBuffer(last - first + 20);

      // add byte range
      sb.append("DataFragment[");
      sb.append(first);
      sb.append(',');
      sb.append(last);

      // add segmentation info
      sb.append(':');
      sb.append((blockSegment ? 'b' : 'i'));
      if (endOfSegment) {
         sb.append(':');
         sb.append(segmentID);
      }

      sb.append(']');
      int length = length();
      if (length > 0) {
         sb.append("(");
         if (Debug.on && ((Debug.getLevel(Debug.DATA) > Debug.DETAIL) || 
                          (length < 5))) {
            sb.append(Debug.printableString(data.data, first, last));
         } else {
            sb.append(Debug.printableChar(data.data[first]));
            sb.append("..");
            sb.append(Debug.printableChar(data.data[last]));
         }
         sb.append(")");
      }

      return sb.toString();
   }

   /**
    * Find out how many bytes are in this DataFragment.
    * @return The length of this fragment.
    */
   public int length() {
      return last - first + 1;
   }

   /**
    * Determine if this fragment contains a particular byte value at any
    * position.
    * @param targetByte The byte value we're looking for.
    * @return True iff one of the bytes within this fragment has the given 
    *         value.
    */
   public boolean containsByte(byte targetByte) {
      for (int i = first; i <= last; i++) {
         if (data.data[i] == targetByte) {
            return true;
         }
      }
      return false;
   }

   /**
    * Determine the distance, from the start of this fragment, to the given
    * byte.  If the byte is in this fragment, it's index will be returned.
    * Otherwise, ask the next element in the stream to tell us how far away
    * the byte is, and add our length to the return value.
    * @param targetByte The byte value that we're looking for.
    * @return The distance along the data path to that byte, or -1 if the byte
    *         is not present before the next boundary.
    */
   public int distanceTo(byte targetByte) {
      for (int i = first; i <= last; i++) {
         if (data.data[i] == targetByte) {
            /* found here, return position */
            return (i - first);
         }
      }

      int res = next(Direction.up).distanceTo(targetByte);
      if (res == -1) {
         /* wasn't found further on */
         return -1;
      } else {
         /* found further on, add my length to the distance */
         return last - first + 1 + res;
      }
   }

   /**
    * This method is called when the recycler is getting ready to
    * recycle this object.  Perform cleanup by recycling the underlying
    * DataBlock as well.
    * @return No return value.
    * @see FMG.Conductor.util.Recycler
    */
   public void recycling() {
      super.recycling();
      data.recycle();
      data = null;
   }

   /**
    * Determine the direction this object flows along the stream.  Since
    * the flow direction is measured relative to the data, by definition
    * this object always has a neutral direction of flow.
    * @return Always returns <code>Direction.neutral</code>.
    * @see FMG.Conductor.util.Direction
    */
   public Direction flowDir() {
      return Direction.neutral;
   }

   /**
    * Determine if this object is a section of the data stream.  At this
    * time, only a DataFragment should return true.
    * @return True iff this object contains a section of the data stream.
    */
   public boolean isData() {
      return true;
   }

   /**
    * Proceed along the datapath under my own control.  Essentially, do
    * the standard proceed functionality for neutral objects as defined
    * by <code>neutralProceed()</code>.
    * @return The net change in the number of DataFragment objects 
    *         during this operation.
    * @see FMG.Conductor.dataflow.StreamElem#neutralProceed
    */
   public int proceed() {
      return neutralProceed();
   }

   /**
    * Determine the given DataFrag (representing the next one in sequence)
    * can be combined with this one, without changing the way the data
    * is segmented. <p>
    * Essentially, return true as long as 
    * <ul>
    *   <li>both DataFrags are block segments we're not the last part of a 
    *       block segment, or
    *   <li>both DataFrags are chunks of one-byte segments and there is
    *       not a gap in segment ids between the two
    * @param nextFrag The next fragment in the stream, which we want to 
    *                 combine with this one.
    * @return True iff the given segment can be combined with this one.
    */
   protected final boolean isNextSegmentFrag(DataFragment nextFrag) {
      return ((nextFrag.blockSegment && blockSegment && ! endOfSegment) || 
              (! nextFrag.blockSegment && ! blockSegment && 
                (nextFrag.segmentID == (segmentID + nextFrag.last - 
                                                    nextFrag.first + 1))));
   }

   /**
    * Migrate some number of bytes from the begining of this DataFragment
    * to the end of the given one.  Such an operation frequently allows us
    * to move a pointer that's between two DataFragment objects without
    * creating or deleting any objects.  It is only permissable to do this
    * if the two objects point to adjacent portions of the same underlying 
    * DataBlock and the segmenting rules can be preserved.
    * @param se The element to try to move some bytes to.
    * @param length The number of bytes to move.
    * @return True iff the operation succeeded.  If false is returned,
    *         nothing happened.
    */
   public boolean migrateBytesTo(StreamElem se, int length) {
      if (! (se instanceof DataFragment)) {
         return false;
      }
      DataFragment targetFrag = (DataFragment) se;
      /* do the frags point to contiguous data blocks? */
      if ((data != targetFrag.data) || (first != targetFrag.last + 1)) {
         return false;
      }

      // ensure compatible segmenting
      if (! targetFrag.isNextSegmentFrag(this)) {
         return false;
      }

      first += length;
      targetFrag.last += length;
      if (! blockSegment) {
         targetFrag.segmentID += length;
      }

      if (length() == 0) {
         unlink();
         recycle();
         return true;
      }

      if (Debug.on && (last < first)) {
         Debug.println("WARNING: In DataFragment.migrateBytesTo(), " + 
                       "last = " + last + ", first = " + first + 
                       "(should have last>=first)");
      }
      if (Debug.on && (targetFrag.first<0)) {
         Debug.println("WARNING: In DataFragment.migrateBytesTo(), " + 
                       "first = " + first + "(should be >=0)");
      }

      return true;
   }  

   /**
    * Deal with the fact that a new object has just been inserted next to us.
    * If it's a DataFragment has been added upstream from us, we want to
    * try to combine it with this one.  Otherwise, nothing is done.
    * @param dir The position of the new object relative to us.
    * @param se A pointer to the new object.
    * @return The net change in the number of DataFragment objects.
    */
   public int handleNewAdjacent(Direction dir, StreamElem se) {
      if (dir.isDown()) {
         return 0;
      }

      /* try to merge with a following StreamElem */
      if (se instanceof DataFragment) {
         DataFragment nextFrag = (DataFragment) se;

         // ensure compatible segmenting
         if (isNextSegmentFrag(nextFrag)) {
            // is next fragment empty?
            if (nextFrag.last-nextFrag.first+1 == 0) {
               if (nextFrag.blockSegment && nextFrag.endOfSegment) {
                  makeBlockEnd(nextFrag.segmentID);
               }
               nextFrag.unlink();
               nextFrag.recycle();
               return -1;

            // is next frag the next part of this one?
            } else if ((nextFrag.data == data) && 
                       (nextFrag.first == last + 1)) {
               last = nextFrag.last;
               segmentID = nextFrag.segmentID;
               nextFrag.unlink();
               nextFrag.recycle();
               return -1;
            } 
         }
      }
      return 0;
   }

   /**
    * This method is called immediately after this object is swapped into
    * an area controlled by an IO object.  In this case, just call the
    * StreamAccessor's method for handling this object.
    * <p>
    * Note: for other objects, this method is called for entry into the area
    * controlled by any StreamAccessor.  This is not true for the DataFragment.
    * This method is only called when this message enters an IO object. 
    * @param sa The StreamAccessor that controls the part of the stream we 
    *           just entered.
    * @return True iff this object still needs to be dealt with upon return.
    * @exception FMG.Conductor.dataflow.AsyncEventException
    *                     Can be thrown to alert the StreamAccessor (who is the
    *                     usual caller) that somethign unusual happened.
    */
   public boolean handleEntry(StreamAccessor sa) throws AsyncEventException {
      return sa.handleEventEntry(this);
   }

   /**
    * This method is called immediately before this object is swapped out of
    * an area controlled by an IO object.
    * <p>
    * Note: for other objects, this method is called before exiting any area
    * controlled by any StreamAccessor.  This is not true for the DataFragment.
    * This method is only called when this message exits an IO object. 
    * @param sa The StreamAccessor that controls the part of the stream we 
    *           are about to enter.
    * @return True iff we should care whether the stream is already closed
    *         before exiting.
    */
   public boolean handleExit(StreamAccessor sa) {
      return sa.handleEventExit(this);
   }

   /**
    * Force this chunk of data to form part of a block segment.  It will 
    * not, however be the end of the segment.
    * @return No return value.
    */
   public void makeBlock() {
      blockSegment = true;
      endOfSegment = false;
      segmentID = -1;
   }

   /**
    * Force this chunk of data to form the end of a block segment.  The
    * given segment id will be assigned.
    * @param segmentID The segment ID to assign to the segment we are ending.
    * @return No return value.
    */
   public void makeBlockEnd(int segmentID) {
      this.segmentID = segmentID;
      blockSegment = true;
      endOfSegment = true;
   }

   /**
    * Obtain the segment id of this segment.  Warning, the segment id is not
    * defined on fragment block segments that are not the end of the segment.
    * @return The assigned segment id.
    */
   public int segmentID() {
      return segmentID;
   }

   /**
    * Is this the end of a segment?
    * @return False iff this is a fragment that ends in the middle of a block 
    *         segment.
    */
   public boolean isEndOfSegment() {
      return endOfSegment;
   }

   /**
    * Is this a block segment?
    * @return True iff this is part of a block segment.
    */
   public boolean isBlock() {
      return blockSegment;
   }

   /**
    * Inform this fragment of the segment id of the last fragment.
    * If this isn't a block segment, we can assign our segment id based
    * on this segment id and the length of this fragment.
    * @param The segment id of the segment before this fragment.
    * @return no return value.
    */
   public void informOfPreviousSegmentID(int prevSegmentID) {
      if (! blockSegment) {
         segmentID = prevSegmentID + last - first + 1;
      }
   }

   /**
    * Write the raw data of this fragment to a socket.  Presumably, the user
    * will actually receive the resulting data.
    * @param os The stream to write the bytes to.
    * @return No return value.
    * @exception java.io.IOException if an error occurs while writing the data.
    */
   public final void writeToRaw(OutputStream os) throws IOException {
      os.write(data.data, first, last - first + 1);
   }
}
