/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.Conductor.dataflow.messages;

import java.io.*;

import FMG.Conductor.comm.*;
import FMG.Conductor.util.*;
import FMG.Conductor.dataflow.*;
import FMG.Conductor.dataflow.comm.IO;

/**
 * The InterfaceControl class provides a mechanism for forwarding, to
 * an IO object, an adaptor's request to turn on or off the interface.
 * This object will always travel with the data (neutral direction).
 * <p>
 * This object is fully recyclable to preven rapid re-creation of object.s
 * <p>
 * TODO: This should probably be generalized to other controls over IO, for
 *       instance, establishing or changing reservations.
 * <p>
 * TODO: Currently, this object must be generated by the last adaptor 
 *       before the IO object.  Eventually, it should be possible for
 *       this object to enter a window.
 *
 * @author Mark Yarvis (yarvis@fmg.cs.ucla.edu)
 * @version $Revision: 1.16 $
 */

public class InterfaceControl extends StreamElem implements Message {

   /** The recycler for this class */
   private static Recycler recycler = new Recycler(InterfaceControl.class);

   /** Is this a request to turn on the interface? (otherwise it's turn off) */
   protected boolean turnOn;

   /** The id assigned to messages of this type. */
   private static int messageID = -1;

   /**
    * Register a prototype of this message class with BasicMessageInputStream
    * so that it can be read and written.
    * @return No return value.
    * @see FMG.Conductor.comm.BasicMessageInputStream
    */
   public static void registerSelf() {
      messageID = BasicMessageInputStream.
                        registerMessagePrototype((Message) recycler.getOne());
   }
   
   /**
    * Get the id that is assigned to messages of this type. 
    * @return The id for this message type.
    */
   public int getMessageID() {
      return messageID;
   }

   /**
    * Read a message of this type from the input stream.  Presumably the
    * message was written using <code>sendSelf()</code>.  This method is
    * normally called on the prototype that was registered with
    * BasicMessageInputStream.
    * @param is The input stream to read from.
    * @return A new message of type EOF that contains the data
    *         read from the stream.
    * @exception java.io.IOException if an error occurs while reading from the
    *                                stream.
    * @see FMG.Conductor.comm.BasicMessageInputStream
    */
   public Message readNewMessage(DataInputStream is) throws IOException {
      return InterfaceControl.getNewMessage(is);
   }

   /**
    * Create a new (or recycled) InterfaceControl object that tells the
    * IO to either turn on or turn off.
    * @param turnOn Should I tell the IO to turn on?  Otherwise, it's turn off.
    * @return A new or recycled InterfaceControl object, initialized as 
    *         specified.
    */
   public static InterfaceControl getNewMessage(boolean turnOn) {
      InterfaceControl m = (InterfaceControl) recycler.getOne();
      m.turnOn = turnOn;
      return m;
   }

   /**
    * Read a message of this type from the input stream.  Presumably the
    * message was written using <code>sendSelf()</code>.
    * @param is The input stream to read from.
    * @return A new message of type InterfaceControl that contains the data
    *         read from the stream.
    * @exception java.io.IOException if an error occurs while reading from the
    *                                stream.
    */
   public static InterfaceControl getNewMessage(DataInputStream is) 
                                                      throws IOException {
      throw new RuntimeException("This shouldn't happen");
   }

   /**
    * Initialize the data members of this object from the input steram.
    * Presumably the data was written using <code>sendSelf()</code>.
    * @param is The input stream to read from.
    * @return No return value.
    * @exception java.io.IOException if an error occurs while reading from the
    *                                stream.
    */
   public void receiveSelf(DataInputStream is) throws IOException {
      throw new RuntimeException("This shouldn't happen");
   }

   /**
    * Write the data members of this object to the output stream.
    * Presumably the data will be read using <code>receiveSelf()</code>.
    * @param os The output stream to write to.
    * @return No return value.
    * @exception java.io.IOException if an error occurs while writing to the
    *                                stream.
    */
   public void sendSelf(DataOutputStream os) throws IOException {
      throw new RuntimeException("This shouldn't happen");
   }

   /**
    * This is not a valid call for objects of this type.  A RuntimeException
    * will result.
    * @param os The output stream to write to.
    * @return No return value.
    * @exception java.io.IOException if an error occurs while writing to the
    *                                stream.
    */
   public void writeToRaw(OutputStream os) throws IOException {
      throw new RuntimeException("This shouldn't happen");
   }

   /**
    * Determine the direction of flow for this object.
    * @return Always <code>Direction.neutral</code>
    */
   public Direction flowDir() {
      return Direction.neutral;
   }

   /**
    * Is this object an event?
    * @return Always returns true.
    */
   public boolean isEvent() {
      return true;
   }

   /**
    * Proceed along the datapath under my own control.  Essentially, I need
    * to do the standard proceed funcationality for neutral objects as
    * defined by <code>neutralProceed()</code>.
    * @return The net change in the number of DataFragment objects.
    * @see FMG.Conductor.dataflow.StreamElem#neutralProceed
    */
   public int proceed() {
      return neutralProceed();
   }

   /**
    * Handle the case in which this object enters an area controlled by a 
    * StreamAccessor.  When entering an area controlled by an IO object,
    * handle turning it on or off.<p>
    * TODO: Someday this should operate for real.  Currently it's just 
    *       println's.
    * @param sa The stream accessor who's area of the segment this object 
    *           just entered.
    * @return True iff this object still needs to be dealt with upon return.
    *         Currently, always return false.  This should probably change.
    * @exception FMG.Conductor.dataflow.AsyncEventException
    *                     Will actually never be thrown.
    */
   public boolean handleEntry(StreamAccessor sa) throws AsyncEventException {
      if (sa instanceof IO) {
         if (Measure.on) {
            String state = (turnOn ? "ON" : "OFF");

            Measure.TimeOfDay tod = Measure.getTimeOfDay();

            Measure.println("InterfaceControl: Turning interface " + state +
                         " at " + tod.secs + "." + tod.usecsString());
         }
         unlink();
         recycle();
         return false;
      }
      return true;
   }
}
