/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.Conductor.dataflow.messages;

import java.io.*;

import FMG.Conductor.comm.*;
import FMG.Conductor.util.*;
import FMG.Conductor.dataflow.*;
import FMG.Conductor.net.*;

/**
 * The RebalanceTypes object is a notification of a change in stream typing
 * due to a failure.  This object will travel either upstream or downstream, 
 * checking the types at each AdaptorWindow.  If the upstream or downstream 
 * type (depending on the direction of flow) does not match the target type,
 * the adaptor is destroyed.
 *
 * <p>
 * This object is fully recyclable to prevent rapid re-creation of objects.
 *
 * @author Mark Yarvis
 * @version $Revision: 1.2 $
 */

public class RebalanceTypes extends StreamElem implements Message {

   /** The recycler for this class of objects */
   private static Recycler recycler = new Recycler(RebalanceTypes.class);

   /** The id assigned to messages of this type. */
   private static int messageID = -1;

   protected StreamTypeStack typeStack = new StreamTypeStack();

   protected Direction dir;

   /**
    * Register a prototype of this message class with BasicMessageInputStream
    * so that it can be read and written.
    * @return No return value.
    * @see FMG.Conductor.comm.BasicMessageInputStream
    */
   public static void registerSelf() {
      messageID = BasicMessageInputStream.
                        registerMessagePrototype((Message) recycler.getOne());
   }

   /**
    * Get the id that is assigned to messages of this type.
    * @return The id for this message type.
    */
   public int getMessageID() {
      return messageID;
   }

   /**
    * Read a message of this type from the input stream.  Presumably the
    * message was written using <code>sendSelf()</code>.  This method is
    * normally called on the prototype that was registered with
    * BasicMessageInputStream.
    * @param is The input stream to read from.
    * @return A new message of type RebalanceTypes that contains the data
    *         read from the stream.
    * @exception java.io.IOException if an error occurs while reading from the
    *                                stream.
    * @see FMG.Conductor.comm.BasicMessageInputStream
    */
   public Message readNewMessage(DataInputStream is) throws IOException {
      return RebalanceTypes.getNewMessage(is);
   }

   /**
    * Create a new (or recycled) RebalanceTypes object.
    * @param dir The direction of travel for this object.
    * @return A new (or recycled) RebalanceTypes object, initialized as 
    *         specified.
    */
   public static RebalanceTypes getNewMessage(Direction dir, 
                                              StreamTypeStack sts) {
      RebalanceTypes m = (RebalanceTypes) recycler.getOne();
      m.typeStack.copy(sts);
      m.dir = dir;
      return m;
   }

   /**
    * Read a message of this type from the input stream.  Presumably the
    * message was written using <code>sendSelf()</code>.
    * @param is The input stream to read from.
    * @return A new message of type RebalanceTypes that contains the data
    *         read from the stream.
    * @exception java.io.IOException if an error occurs while reading from the
    *                                stream.
    */
   public static RebalanceTypes getNewMessage(DataInputStream is) 
                                                         throws IOException {
      RebalanceTypes m = (RebalanceTypes) recycler.getOne();
      m.receiveSelf(is);
      return m;
   }

   /**
    * Initialize the data members of this object from the input steram.
    * Presumably the data was written using <code>sendSelf()</code>.
    * @param is The input stream to read from.
    * @return No return value.
    * @exception java.io.IOException if an error occurs while reading from the
    *                                stream.
    */
   public void receiveSelf(DataInputStream is) throws IOException {
      dir = Direction.readDir(is);
      typeStack.readSelf(is);
   }

   /**
    * Write the data members of this object to the output stream.
    * Presumably the data will be read using <code>receiveSelf()</code>.
    * @param os The output stream to write to.
    * @return No return value.
    * @exception java.io.IOException if an error occurs while writing to the
    *                                stream.
    */
   public void sendSelf(DataOutputStream os) throws IOException {
      dir.writeSelf(os);
      typeStack.writeSelf(os);
   }

   /**
    * This is not a valid call for objects of this type.  A RuntimeException
    * will result.
    * @param os The output stream to write to.
    * @return No return value.
    * @exception java.io.IOException if an error occurs while writing to the
    *                                stream.
    */
   public void writeToRaw(OutputStream os) throws IOException {
      throw new RuntimeException("This shouldn't happen!!!");
   }

   /**
    * Determine the direction of flow for this object.
    * @return The direction of flow, normally <code>Direction.neutral</code>
    *         or <code>Direction.up</code>
    */
   public Direction flowDir() {
      return dir;
   }

   /**
    * Is this object an event?
    * @return Always returns true.
    */
   public boolean isEvent() {
      return true;
   }

   /**
    * Does the given type stack equal the target we're supposed to reach?
    * @param sts The type stack to test.
    * @return True, iff the given type stack is the same as the target.
    */
   public boolean equalsTargetTypeStack(StreamTypeStack sts) {
      return typeStack.equals(sts);
   }

   /**
    * Copy our target type stack into the object given as an argument.
    * @param sts The target for the copy.
    * @return No return value.
    */
   public void copyTargetTypeStack(StreamTypeStack sts) {
      sts.copy(typeStack);
   }

   /**
    * Cause the message to proceed along the datapath under its own control.  
    * Proceed downstream, removing data objects and user-level pointers, and 
    * passing all others.  If an "end of segment" is reached, either at a 
    * DataFragment or at a boundary, replace self with RetransRequest and 
    * tell it to proceed.  Stop if a boundary that isn't at the EoS is reached.
    * @return The net change in the number of DataFragment objects in the
    *         stream.
    */
   public int proceed() {
      int adjustment = 0;

      // keep moving until we hit a boundary
      while (true) {
         StreamElem se = next(dir);

         if (se.isBoundary()) {             // we hit a boundary
            return se.handleNewAdjacent(dir.opposite(), this) + adjustment;

         } else if (se.isData()) {         // we hit data
            se.unlink();
            se.recycle();
            adjustment--;

         } else if (se instanceof SwitchPlan) {
            if (dir.isUp()) {
               SwitchPlan sp = (SwitchPlan) se;
               if (sp.equalsTypeStack(typeStack)) {
                  unlink();
                  recycle();
                  return sp.proceed() + adjustment;
               } else {
                  throw new RuntimeException("Case not handled");
               }
            } else {
               return 0;
            }

         } else if ((se instanceof RebalanceTypes) && 
                    (se.flowDir().equals(dir.opposite()))) {
            RebalanceTypes rt = (RebalanceTypes) se;

            if (rt.typeStack.equals(typeStack)) {
               rt.unlink();
               rt.recycle();

               se = next(dir.opposite());
               unlink();
               recycle();
 
               return se.proceed() + adjustment;
            } else {
               int newSize = StreamTypeStack.numCommonElements(typeStack, 
                                                               rt.typeStack);
               typeStack.setSize(newSize);
               rt.typeStack.setSize(newSize);
               swap(dir);
               adjustment += rt.proceed();
            }

         } else if (! se.flowDir().equals(dir)) {
            // skip non-data (is this always right?)
            swap(dir);
            adjustment += se.proceed();
         }
      }
   }

   /**
    * Handle a new object that is now adjacent to me.  If the new object is
    * in my direction of motion, just call <code>proceed()</code>.  Otherwise,
    * do nothing.
    * @param dir The position, relative to me, of the new object.
    * @param se A pointer to the new object.
    * @return The net change in the number of DataFragment objects in the
    *         stream.
    */
   public int handleNewAdjacent(Direction dir, StreamElem se) {
      if (dir.equals(this.dir)) {
         return proceed();
      } else {
         return 0;
      }
   }

   /**
    * Handle the case in which this object enters an area controlled by a 
    * StreamAccessor.  Just call the StreamAccessor's method for handling 
    * this object.
    * @param sa The StreamAccessor that this object just entered.
    * @return True iff this object still needs to be dealt with upon return.
    * @exception FMG.Conductor.dataflow.AsyncEventException
    *                     Can be thrown to alert the StreamAccessor (who is the
    *                     usual caller) that somethign unusual happened.
    */
   public boolean handleEntry(StreamAccessor sa) throws AsyncEventException {
      return sa.handleEventEntry(this);
   }

   /**
    * Handle the case in which this object is about to exit a StreamAccessor.
    * Just call the StreamAccessor's method for handling exits of this object.
    * @param sa The StreamAccessor that this object is about to exit.
    * @return True iff the caller should care that the exit path is closed.
    */
   public boolean handleExit(StreamAccessor sa) {
      return sa.handleEventExit(this);
   }

   /**
    * Generate a human-meaningful representation of this object.  Use for
    * debugging purposes only.
    * @return A new string representing this object.
    */
   public String toString() {
      return "RebalanceTypes(" + dir + "," + typeStack + ")";
   }
}
