/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.Conductor.dataflow.messages;

import java.io.*;

import FMG.Conductor.comm.*;
import FMG.Conductor.util.*;
import FMG.Conductor.dataflow.*;
import FMG.Conductor.net.*;

/**
 * The Shutdown object is the last message that will flow in a particular
 * direction of data flow.  This message is sent in response to an EOF
 * message, and hence travels in the reverse direction form that object.  
 * Notice that, like EOF, this message affects only half of the data stream,
 * one direction of data flow.
 * <p>
 * This object is fully recyclable to prevent rapid re-creation of objects.
 *
 * @author Mark Yarvis
 * @version $Revision: 1.8 $
 * @see FMG.Conductor.dataflow.messages.EOF
 */

public class Shutdown extends StreamElem implements Message {

   /** The recycler for this class of objects */
   private static Recycler recycler = new Recycler(Shutdown.class);

   /** The direction that this messsage is traveling on the stream */
   protected Direction dir;

   /** The id assigned to messages of this type. */
   private static int messageID = -1;

   /**
    * Register a prototype of this message class with BasicMessageInputStream
    * so that it can be read and written.
    * @return No return value.
    * @see FMG.Conductor.comm.BasicMessageInputStream
    */
   public static void registerSelf() {
      messageID = BasicMessageInputStream.
                        registerMessagePrototype((Message) recycler.getOne());
   }

   /**
    * Get the id that is assigned to messages of this type.
    * @return The id for this message type.
    */
   public int getMessageID() {
      return messageID;
   }

   /**
    * Read a message of this type from the input stream.  Presumably the
    * message was written using <code>sendSelf()</code>.  This method is
    * normally called on the prototype that was registered with
    * BasicMessageInputStream.
    * @param is The input stream to read from.
    * @return A new message of type Shutdown that contains the data
    *         read from the stream.
    * @exception java.io.IOException if an error occurs while reading from the
    *                                stream.
    * @see FMG.Conductor.comm.BasicMessageInputStream
    */
   public Message readNewMessage(DataInputStream is) throws IOException {
      return Shutdown.getNewMessage(is);
   }

   /**
    * Create a new (or recycled) Shutdown object.
    * @param dir The direction of travel for this object.
    * @return A new (or recycled) Shutdown object, initialized as specified.
    */
   public static Shutdown getNewMessage(Direction dir) {
      Shutdown m = (Shutdown) recycler.getOne();
      m.dir = dir;
      return m;
   }

   /**
    * Read a message of this type from the input stream.  Presumably the
    * message was written using <code>sendSelf()</code>.
    * @param is The input stream to read from.
    * @return A new message of type Shutdown that contains the data
    *         read from the stream.
    * @exception java.io.IOException if an error occurs while reading from the
    *                                stream.
    */
   public static Shutdown getNewMessage(DataInputStream is) 
                                                      throws IOException {
      Shutdown m = (Shutdown) recycler.getOne();
      m.receiveSelf(is);
      return m;
   }

   /**
    * Initialize the data members of this object from the input steram.
    * Presumably the data was written using <code>sendSelf()</code>.
    * @param is The input stream to read from.
    * @return No return value.
    * @exception java.io.IOException if an error occurs while reading from the
    *                                stream.
    */
   public void receiveSelf(DataInputStream is) throws IOException {
      dir = Direction.readDir(is);
   }

   /**
    * Write the data members of this object to the output stream.
    * Presumably the data will be read using <code>receiveSelf()</code>.
    * @param os The output stream to write to.
    * @return No return value.
    * @exception java.io.IOException if an error occurs while writing to the
    *                                stream.
    */
   public void sendSelf(DataOutputStream os) throws IOException {
      dir.writeSelf(os);
   }

   /**
    * This is not a valid call for objects of this type.  A RuntimeException
    * will result.
    * @param os The output stream to write to.
    * @return No return value.
    * @exception java.io.IOException if an error occurs while writing to the
    *                                stream.
    */
   public void writeToRaw(OutputStream os) throws IOException {
      throw new RuntimeException("Should not happen");
   }

   /**
    * Determine the direction of flow for this object.
    * @return The direction of flow, normally <code>Direction.neutral</code>
    *         or <code>Direction.up</code>
    */
   public Direction flowDir() {
      return dir;
   }

   /**
    * Is this object an event?
    * @return Always returns true.
    */
   public boolean isEvent() {
      return true;
   }

   /**
    * Cause this message to proceed along the datapath.  In this case,
    * there shouldn't be much left on the path, certainly, nothing
    * coming in the opposite direction.  So, just notify the newly adjacent
    * object of this object's presence.
    * @return The net change in the number of DataFragment objects in the
    *         stream.
    */
   public int proceed() {
      return next(dir).handleNewAdjacent(dir.opposite(), this);
   }

   /**
    * This method shouldn't ever be called.  Nothing can flow behind
    * this object (in the same direction), and nothing should be flowing
    * in the other direction, either.
    * @param dir The position, relative to me, of the new object.
    * @param se A pointer to the new object.
    * @return Since it's never called, the return value is of no consequence.
    */
   public int handleNewAdjacent(Direction dir, StreamElem se) {
      throw new RuntimeException("This shouldn't happen");
   }

   /**
    * Handle the case in which this object enters an area controlled by a
    * StreamAccessor.  Just call the StreamAccessor's method for handling
    * this object.
    * @param sa The StreamAccessor that this object just entered.
    * @return True iff this object still needs to be dealt with upon return.
    * @exception FMG.Conductor.dataflow.AsyncEventException
    *                     Can be thrown to alert the StreamAccessor (who is the
    *                     usual caller) that somethign unusual happened.
    */
   public boolean handleEntry(StreamAccessor sa) throws AsyncEventException {
      return sa.handleEventEntry(this);
   }

   /**
    * Handle the case in which this object is about to exit a StreamAccessor.
    * Just call the StreamAccessor's method for handling exits of this object.
    * @param sa The StreamAccessor that this object is about to exit.
    * @return True iff the caller should care that the exit path is closed.
    */
   public boolean handleExit(StreamAccessor sa) {
      return sa.handleEventExit(this);
   }

   /**
    * Generate a human-meaningful representation of this object.  Use for
    * debugging purposes only.
    * @return A new string representing this object.
    */
   public String toString() {
      StringBuffer sb = new StringBuffer();
      sb.append("Shutdown(");
      if (dir != null) {
         sb.append(dir.toString());
      } else {
         sb.append("???");
      }
      sb.append(")");
      return sb.toString();
   }
}
