/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.Conductor.dataflow.messages;

import java.io.*;

import FMG.Conductor.comm.*;
import FMG.Conductor.util.*;
import FMG.Conductor.dataflow.*;
import FMG.Conductor.net.*;

/**
 * The SignalFailure object is a notification of an upstream failure.
 * This object will travel downstream, seeking the end of a segment.
 * Once found, a retransmission request for the following segments will
 * be generated.
 * <p>
 * This object is fully recyclable to prevent rapid re-creation of objects.
 *
 * @author Mark Yarvis
 * @version $Revision: 1.3 $
 */

public class SignalFailure extends StreamElem implements Message {

   /** The recycler for this class of objects */
   private static Recycler recycler = new Recycler(SignalFailure.class);

   /** The id assigned to messages of this type. */
   private static int messageID = -1;

   /**
    * Register a prototype of this message class with BasicMessageInputStream
    * so that it can be read and written.
    * @return No return value.
    * @see FMG.Conductor.comm.BasicMessageInputStream
    */
   public static void registerSelf() {
      messageID = BasicMessageInputStream.
                        registerMessagePrototype((Message) recycler.getOne());
   }

   /**
    * Get the id that is assigned to messages of this type.
    * @return The id for this message type.
    */
   public int getMessageID() {
      return messageID;
   }

   /**
    * Read a message of this type from the input stream.  Presumably the
    * message was written using <code>sendSelf()</code>.  This method is
    * normally called on the prototype that was registered with
    * BasicMessageInputStream.
    * @param is The input stream to read from.
    * @return A new message of type SignalFailure that contains the data
    *         read from the stream.
    * @exception java.io.IOException if an error occurs while reading from the
    *                                stream.
    * @see FMG.Conductor.comm.BasicMessageInputStream
    */
   public Message readNewMessage(DataInputStream is) throws IOException {
      return SignalFailure.getNewMessage(is);
   }

   /**
    * Create a new (or recycled) SignalFailure object.
    * @param dir The direction of travel for this object.
    * @return A new (or recycled) SignalFailure object, initialized as 
    *         specified.
    */
   public static SignalFailure getNewMessage() {
      SignalFailure m = (SignalFailure) recycler.getOne();
      return m;
   }

   /**
    * Read a message of this type from the input stream.  Presumably the
    * message was written using <code>sendSelf()</code>.
    * @param is The input stream to read from.
    * @return A new message of type SignalFailure that contains the data
    *         read from the stream.
    * @exception java.io.IOException if an error occurs while reading from the
    *                                stream.
    */
   public static SignalFailure getNewMessage(DataInputStream is) 
                                                         throws IOException {
      SignalFailure m = (SignalFailure) recycler.getOne();
      m.receiveSelf(is);
      return m;
   }

   /**
    * Initialize the data members of this object from the input steram.
    * Presumably the data was written using <code>sendSelf()</code>.
    * @param is The input stream to read from.
    * @return No return value.
    * @exception java.io.IOException if an error occurs while reading from the
    *                                stream.
    */
   public void receiveSelf(DataInputStream is) throws IOException {
   }

   /**
    * Write the data members of this object to the output stream.
    * Presumably the data will be read using <code>receiveSelf()</code>.
    * @param os The output stream to write to.
    * @return No return value.
    * @exception java.io.IOException if an error occurs while writing to the
    *                                stream.
    */
   public void sendSelf(DataOutputStream os) throws IOException {
   }

   /**
    * This is not a valid call for objects of this type.  A RuntimeException
    * will result.
    * @param os The output stream to write to.
    * @return No return value.
    * @exception java.io.IOException if an error occurs while writing to the
    *                                stream.
    */
   public void writeToRaw(OutputStream os) throws IOException {
      throw new RuntimeException("This shouldn't happen!!!");
   }

   /**
    * Determine the direction of flow for this object.
    * @return The direction of flow, normally <code>Direction.neutral</code>
    *         or <code>Direction.up</code>
    */
   public Direction flowDir() {
      return Direction.down;
   }

   /**
    * Is this object an event?
    * @return Always returns true.
    */
   public boolean isEvent() {
      return true;
   }

   /**
    * Cause the message to proceed along the datapath under its own control.  
    * Proceed downstream, removing data objects and user-level pointers, and 
    * passing all others.  If an "end of segment" is reached, either at a 
    * DataFragment or at a boundary, replace self with RetransRequest and 
    * tell it to proceed.  Stop if a boundary that isn't at the EoS is reached.
    * @return The net change in the number of DataFragment objects in the
    *         stream.
    */
   public int proceed() {
      int adjustment = 0;

      int lastFullSegment;

      // keep moving until we hit an EOS or a boundary
      while (true) {
         StreamElem se = next(Direction.down);

         if (se.isBoundary()) {             // we hit a boundary
            Boundary bound = (Boundary) se;
            if (bound.isBetweenSegments()) {
               lastFullSegment = bound.lastSegmentID();
               break;                          // done, break out of loop!
            } else {
               // must pass the boundary ... return
               return se.handleNewAdjacent(Direction.up, this) + adjustment;
            }

         } else if (se.isData()) {         // we hit data
            DataFragment df = (DataFragment) se;
            if (df.isEndOfSegment()) {
               // full segment reached
               lastFullSegment = df.segmentID();
               break;                          // done, break out of loop!
            } else {
               // delete partial segment
               se.unlink();
               se.recycle();
               adjustment--;
            }

         } else if (se.isUserLevel()) {
            // unlink user-level pointers
            se.unlink();

         } else {
            // skip non-data (is this always right?)
            swap(Direction.down);
         }
      }

      // We hit an "end of segment".  Replace self with RetransRequest.
      RetransRequest rr = RetransRequest.getNewMessage(lastFullSegment);
      if (Debug.on) {
         Debug.println(Debug.RELIABILITY, Debug.INFO, 
                       this + " reached segment edge, switching to " + rr);
      }
      insert(rr, Direction.up);
      unlink();
      return adjustment + rr.proceed();
   }

   /**
    * Handle a new object that is now adjacent to me.  If the new object is
    * in my direction of motion, just call <code>proceed()</code>.  Otherwise,
    * do nothing.
    * @param dir The position, relative to me, of the new object.
    * @param se A pointer to the new object.
    * @return The net change in the number of DataFragment objects in the
    *         stream.
    */
   public int handleNewAdjacent(Direction dir, StreamElem se) {
      if (dir.isDown()) {
         return proceed();
      } else {
         return 0;
      }
   }

   /**
    * Handle the case in which this object enters an area controlled by a 
    * StreamAccessor.  Just call the StreamAccessor's method for handling 
    * this object.
    * @param sa The StreamAccessor that this object just entered.
    * @return True iff this object still needs to be dealt with upon return.
    * @exception FMG.Conductor.dataflow.AsyncEventException
    *                     Can be thrown to alert the StreamAccessor (who is the
    *                     usual caller) that somethign unusual happened.
    */
   public boolean handleEntry(StreamAccessor sa) throws AsyncEventException {
// in an adaptor, 
//      adaptor window calls proceed
//      if we're then at the end of window, push it out and wait for a response
//      if a retrans request is at the start, wait for data
      return sa.handleEventEntry(this);
   }

   /**
    * Handle the case in which this object is about to exit a StreamAccessor.
    * Just call the StreamAccessor's method for handling exits of this object.
    * @param sa The StreamAccessor that this object is about to exit.
    * @return True iff the caller should care that the exit path is closed.
    */
   public boolean handleExit(StreamAccessor sa) {
      return sa.handleEventExit(this);
   }

   /**
    * Generate a human-meaningful representation of this object.  Use for
    * debugging purposes only.
    * @return A new string representing this object.
    */
   public String toString() {
      return "SignalFailure()";
   }
}
