/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.Conductor.dataflow.messages;

import FMG.Conductor.comm.NodeID;

import java.io.*;

public class StreamTypeStack {

   public static final int MAX_COMPOSITION_LEVELS = 15;

   public static class StreamType {
      public NodeID planner = new NodeID();
      public int planVersion;
      public int adaptorNum;

      protected StreamType() {
      }

      protected void setType(NodeID planner, int planVersion, int adaptorNum) {
         this.planner.copy(planner);
         this.planVersion = planVersion;
         this.adaptorNum = adaptorNum;
      }

      protected boolean equals(StreamType other) {
         return (planner.equals(other.planner) && 
                 (planVersion == other.planVersion) &&
                 (adaptorNum == other.adaptorNum));
      }

      protected void writeSelf(DataOutputStream os) throws IOException {
         planner.writeSelf(os);
         os.writeInt(planVersion);
         os.writeInt(adaptorNum);
      }

      protected void readSelf(DataInputStream is) throws IOException {
         planner.readSelf(is);
         planVersion = is.readInt();
         adaptorNum = is.readInt();
      }

      protected void copy(StreamType other) {
         planner.copy(other.planner);
         planVersion = other.planVersion;
         adaptorNum = other.adaptorNum;
      }

      public String toString() {
         return planner.toString() + "," + planVersion + "," + adaptorNum;
      }
   }

   int size = 0;
   StreamType[] stack = new StreamType[MAX_COMPOSITION_LEVELS];

   public StreamTypeStack() {
      for (int i=0; i<stack.length; i++) {
         stack[i] = new StreamType();
      }
   }

   public StreamTypeStack(StreamTypeStack other) {
      this();
      size = other.size;
      for (int i=0; i<size; i++) {
         stack[i].copy(other.stack[i]);
      }
   }

   public void push(StreamType st) {
      stack[size++].copy(st);
   }

   void push(NodeID planner, int version, int adaptorNum) {
      stack[size++].setType(planner, version, adaptorNum);
   }

   public StreamType pop() {
      return stack[--size];
   }

   public StreamType peek() {
      return stack[size - 1];
   }

   public void writeSelf(DataOutputStream os) throws IOException {
      os.writeInt(size);
      for (int i=0; i<size; i++) {
         stack[i].writeSelf(os);
      }
   }

   public void readSelf(DataInputStream is) throws IOException {
      size = is.readInt();
      for (int i=0; i<size; i++) {
         stack[i].readSelf(is);
      }
   }

   public void clear() {
      size = 0;
   }

   public int size() {
      return size;
   }

   public void setSize(int newSize) {
      size = newSize;
   }

   public boolean equals(StreamTypeStack other) {
      if (size != other.size) {
         return false;
      }

      for (int i=0; i<size; i++) {
         if (! stack[i].equals(other.stack[i])) {
            return false;
         }
      }

      return true;
   }

   public static int numCommonElements(StreamTypeStack one, 
                                       StreamTypeStack two) {
      int len = Math.min(one.size, two.size);
      for (int i=0; i<len; i++) {
         if (! one.stack[i].equals(two.stack[i])) {
            return i;
         }
      }
      return len;
   }

   public void copy(StreamTypeStack other) {
      size = other.size;
      for (int i=0; i<size; i++) {
         stack[i].copy(other.stack[i]);
      }
   }

   /**
    * Remove all elements down to the specified one.
    */
   public void removeType(StreamType st) {
      for (int i=(size-1); i>=0; i--) {
         if (stack[i].equals(st)) {
            size = i;
            return;
         }
      }
   }

   public String toString() {
      StringBuffer sb = new StringBuffer(100);
      for (int i=(size-1); i>=0; i--) {
         sb.append(stack[i].toString());
         if (i > 0) {
            sb.append('/');
         }
      }
      return sb.toString();
   }
}
