/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.Conductor.dataflow.messages;

import java.io.*;
import java.util.*;
import java.lang.reflect.*;

import FMG.Conductor.dataflow.*;
import FMG.Conductor.dataflow.comm.RawIO;
import FMG.Conductor.comm.*;
import FMG.Conductor.util.*;
import FMG.Conductor.planning.messages.PlanDistro;
import FMG.Conductor.planning.Path;

/**
 * The SwitchPlan message causes the installation of a new plan along
 * the data path.  Data following this message will be adapted according
 * to a new plan.  Plan changes are executed when the object enters a new
 * StreamAccessor.  This message flows with the flow of data from the 
 * planner node, to the root node, and back to the planner node.
 * <p>
 * This object indicates the version of the plan to install.  The actual
 * plan is transmitted via the planning path.  This object includes a method
 * that aids in getting a handle on the proper plan.  Once obtained, a copy 
 * of this handle is cached for future use.
 * <p>
 * This object is fully recyclable to prevent rapid re-creation of objects.
 *
 * @author Mark Yarvis
 * @version $Revision: 1.25 $
 * @see FMG.Conductor.dataflow.StreamAccessor
 * @see FMG.Conductor.planning.messages.PlanDistro
 */

public class SwitchPlan extends StreamElem implements Message {

   /** The recycler for messages of this class */
   private static Recycler recycler = new Recycler(SwitchPlan.class);


   /** The id of this connection.  Usef to fetch the plan. */
   public ConnID connID = new ConnID();

   /** The version number of the plan to switch to. */
   public int version;

   /** Is this the second leg of the round trip journey? */
   public boolean secondLeg;

   /** The type stack developed so far during deployment */
   protected StreamTypeStack typeStack = new StreamTypeStack();

   /**
    * The type stack at the upstream IO object, before the new plan was
    * delivered.  This field is only used by IO objects.
    */
   protected StreamTypeStack oldUpstreamTypeStack = new StreamTypeStack();

   /**
    * The type stack we're hoping to attain.  This field is only used after 
    * a node failure and is only valid when useTargetTypeStack is true.
    */
   protected StreamTypeStack targetTypeStack = new StreamTypeStack();

   /**
    * Indicates whether a prior failure is requiring us to alter the plan
    * so as to attain a target type stack.
    */
   protected boolean useTargetTypeStack = false;


   /** The plan we're switching to */
   protected PlanDistro plan;

   /** 
    * The list of adaptors, specified in the plan, to be installed on this 
    * node. 
    */
   protected String[] localAdaptors = null;

   /**
    * The number of adaptors already installed on this node.
    */
   protected int installed;

   /**
    * The number of adaptors intalled before this node.
    */
   protected int previouslyInstalled;

   /** Used in a kludge to remember the last RawIO object we exited. */
   private RawIO lastRawIO = null;

   /** The id assigned to messages of this type */
   private static int messageID = -1;

   /** 
    * A cache of composition types (push, pop, or none) that have been 
    * previously determined for adaptor spacs. 
    */
   protected static Hashtable compositionTypeCache = new Hashtable();

   /**
    * Register a prototype of this message class with BasicMessageInputStream
    * so that it can be read and written.
    * @return No return value.
    * @see FMG.Conductor.comm.BasicMessageInputStream
    */
   public static void registerSelf() {
      messageID = BasicMessageInputStream.
                        registerMessagePrototype((Message) recycler.getOne());
   }
   
   /**
    * Get the id that is assigned to messages of this type.
    * @return The id for this message type.
    */
   public int getMessageID() {
      return messageID;
   }

   /**
    * Read a message of this type from the input stream.  Presumably the
    * message was written using <code>sendSelf()</code>.  This method is
    * normally called on the prototype that was registered with
    * BasicMessageInputStream.
    * @param is The input stream to read from.
    * @return A new message of type SwitchPlan that contains the data
    *         read from the stream.
    * @exception java.io.IOException if an error occurs while reading from the
    *                                stream.
    * @see FMG.Conductor.comm.BasicMessageInputStream
    */
   public Message readNewMessage(DataInputStream is) throws IOException {
      return SwitchPlan.getNewMessage(is);
   }

   /**
    * Create a new (or recycled) SwitchPlan object.
    * @param connID The id of this connection.
    * @param version The version number of the plan to install.
    * @return A new (or recycled) SwitchPlan object, initialized as specified.
    */
   public static SwitchPlan getNewMessage(ConnID connID, int version) {  
      SwitchPlan m = (SwitchPlan) recycler.getOne();
      m.connID.copy(connID);
      m.version = version;
      m.secondLeg = false;
      m.plan = null;
      m.localAdaptors = null;
      m.useTargetTypeStack = false;
      m.targetTypeStack.clear();
      m.typeStack.clear();
      m.oldUpstreamTypeStack.clear();
      return m;
   }
   
   /**
    * Read a message of this type from the input stream.  Presumably the
    * message was written using <code>sendSelf()</code>.
    * @param is The input stream to read from.
    * @return A new message of type SwitchPlan that contains the data
    *         read from the stream.
    * @exception java.io.IOException if an error occurs while reading from the
    *                                stream.
    */
   public static SwitchPlan getNewMessage(DataInputStream is) 
                                                       throws IOException {
      SwitchPlan m = (SwitchPlan) recycler.getOne();
      m.receiveSelf(is);
      return m;
   }
 
   /**
    * Initialize the data members of this object from the input steram.
    * Presumably the data was written using <code>sendSelf()</code>.
    * @param is The input stream to read from.
    * @return No return value.
    * @exception java.io.IOException if an error occurs while reading from the
    *                                stream.
    */
   public void receiveSelf(DataInputStream is) throws IOException {
      connID.readSelf(is);
      version = is.readInt();
      secondLeg = is.readBoolean();
      typeStack.readSelf(is);
      oldUpstreamTypeStack.readSelf(is);
      useTargetTypeStack = is.readBoolean();
      if (useTargetTypeStack) {
         targetTypeStack.readSelf(is);
      } else {
         targetTypeStack.clear();
      }

      plan = null;
      localAdaptors = null;
   }

   /**
    * Write the data members of this object to the output stream.
    * Presumably the data will be read using <code>receiveSelf()</code>.
    * @param os The output stream to write to.
    * @return No return value.
    * @exception java.io.IOException if an error occurs while writing to the
    *                                stream.
    */
   public void sendSelf(DataOutputStream os) throws IOException {
      connID.writeSelf(os);
      os.writeInt(version);
      os.writeBoolean(secondLeg);
      typeStack.writeSelf(os);
      oldUpstreamTypeStack.writeSelf(os);
      os.writeBoolean(useTargetTypeStack);
      if (useTargetTypeStack) {
         targetTypeStack.writeSelf(os);
      }
   }

   /**
    * This is not a valid call for objects of this type.  A RuntimeException
    * will result.
    * @param os The output stream to write to.
    * @return No return value.
    * @exception java.io.IOException if an error occurs while writing to the
    *                                stream.
    */
   public void writeToRaw(OutputStream os) throws IOException {
      throw new RuntimeException("This souldn't happen!!!");
   }

   /**
    * Get the plan that is supposed to be installed form the appropriate 
    * Path object.  Note that this call can block until such time as the 
    * plan is actually received on the planning path.
    * @param session The session object for this stream.
    * @return A hanlde to the PlanDistro object
    */
   private PlanDistro getPlan(Session session) {
      if (plan == null) {
         plan = session.getPath().getPlan(version);
      }
      return plan;
   }

   private String[] getLocalAdaptors(Session session) {
      if (localAdaptors == null) {
         localAdaptors = getPlan(session).getLocalAdaptors(direction());
         installed = 0;
         previouslyInstalled = getPlan(session).numPrevious(direction());
      }
      return localAdaptors;
   }

   /**
    * Get the name of the next adaptor to install on this node.
    * <p>
    * TODO: this is really meant as a temporary hack.
    * @param session The session object for this stream.
    * @return A string containing the class name of the next adaptor to 
    *         install.
    */
   private String getNextAdaptor(Session session) {
      String[] adaptors = getLocalAdaptors(session);

      /*
       * HACK: currently this works because the array is guaranteed to be big
       * enough and has a null element at the end
       */
      return adaptors[installed];
   }

   /**
    * Push a new type onto the current type stack.
    * @param st The new type to push.
    * @return No return value.
    */
   public void typeStackPush(StreamTypeStack.StreamType st) {
      typeStack.push(st);
   }

   /**
    * Pop a type off of the current type stack.
    * @return No return value.
    */
   public void typeStackPop() {
      typeStack.pop();
   }

   /**
    * Copy the current type stack into the object passed as a parameter.
    * @param sts The target of the copy.
    * @return No return value.
    */
   public void copyTypeStack(StreamTypeStack sts) {
      sts.copy(typeStack);
   }

   /**
    * Determine if the given type stack is equal to the current type stack.
    * @param sts The stack to test against.
    * @return True, iff the stacks are the same.
    */
   public boolean equalsTypeStack(StreamTypeStack sts) {
      return sts.equals(typeStack);
   }

   /**
    * Set the old upstream type stack to the given stack.
    * @param sts The source to copy the old type stack from.
    * @return No return value.
    */
   public void setOldUpstreamTypeStack(StreamTypeStack sts) {
      oldUpstreamTypeStack.copy(sts);
   }

   /**
    * Copy the old upstream type stack into the stack given as the parameter.
    * @param sts The target of the copy.
    * @return No return value.
    */
   public void copyOldUpstreamTypeStack(StreamTypeStack sts) {
      sts.copy(oldUpstreamTypeStack);
   }

   /**
    * Set the target type stack to the one given in the argument.  As adaptor
    * deployment proceeds from this point, no adaptors will be deployed until
    * this type is achieved.  We are guaranteed (I think), however, that no
    * existing adaptors will have to be removed (they will already have been
    * removed for us by a RebalanceTypes message).
    * @param sts The type to achieve.
    * @return No return value.
    */
   public void setTargetTypeStack(StreamTypeStack sts) {
      useTargetTypeStack = true;
      targetTypeStack.copy(sts);
   }

   /**
    * Is the named adaptor going to be called for in the plan (ie, is it 
    * coming up in the plan)?
    * @param session The session object for this stream.
    * @param targetAdaptor The class name of the adaptor in question.
    * @return True iff the given adaptor is named in the plan as being 
    *         installed on this node sometime after the previously installed 
    *         adaptors.
    */
   private boolean isAdaptorComingUp(String targetAdaptor, Session session) {
      String[] adaptors = getLocalAdaptors(session);

      for (int i=installed; adaptors[i]!=null; i++) {
         if (adaptors[i].equals(targetAdaptor)) {
            return true;
         }
      }
      return false;
   }

   /**
    * Which half of the user flow are we currently installing on.  During the
    * first leg, this will be <code>Direction.down</code>.  During the second
    * leg, this will be <code>Direction.up</code>.
    * <p>
    * <b>Warning:</b> Do not confuse this with the direction, relative to 
    *                 the data flow.
    * @return Either Direction.up, or Direction.down depending on which leg
    *         of the journey we're on.
    */
   public Direction direction() {
      return (secondLeg ? Direction.up : Direction.down);
   }

   /**
    * Check off the current adaptor as installed.
    * <p>
    * TODO: this is really meant as a temporary hack.
    * @return A string containing the class name of the next adaptor to 
    *         install.
    */
   private void checkOffNextAdaptor() {
      installed++;
   }

   /**
    * Called when the recycler is about the recycle this object.  In 
    * particular, if a plan was taken from the Path object, it has been
    * slated to be retained.  So, we have to recycle it, to give up our
    * claim on this object.
    */
   public void recycling() {
      super.recycling();
      if (plan != null) {
         plan.recycle();
      }
   }

   /**
    * Install all remaining adaptors at the point where this object is 
    * currently located.  This method is normally called from an IO object
    * immediately before leaving a node.  Note that this method changes the 
    * boundaries of the calling AdaptorWindow or IO object.
    * @param session The session object for this connection.
    * @return No return value.
    */
   public void installAdaptors(Session session) {
      installAdaptors(null, session);
   }

   /**
    * Determine if two adaptors are the same.  This is currently implemented
    * as a string comparison.
    * @param a1 The first adaptor name to compare.
    * @param a2 The first adaptor name to compare.
    * @return True, iff both strings are non-null and contain the same value.
    *         False, otherwise.
    */
   private boolean sameAdaptors(String a1, String a2) {
      if (a1 == null) {
         return (a2 == null);
      } else {
         return a1.equals(a2);
      }
   }

   /**
    * Get the composition type that a deployed adaptor would have.  The
    * argument, an adpator specification, contains the adaptor class name
    * and the arguments to be used in its deployment.  The adaptor 
    * specification is used to make a call to compositionType(String) for
    * the appropriate adaptor class.  If no such method exists, 
    * <code>CompositionType.NONE</code> is returned.
    * @param The specification of the adaptor we're interested in.
    * @return The composition type of the specified adaptor.
    */
   protected Adaptor.CompositionType getCompositionType(String adaptorSpec) {
      Adaptor.CompositionType ret = 
              (Adaptor.CompositionType) compositionTypeCache.get(adaptorSpec);

      if (ret == null) {
         try {
            String className = Adaptor.getNameFromSpec(adaptorSpec);
            String params = Adaptor.getParamsFromSpec(adaptorSpec);

            Class[] paramTypes = { String.class };
            Method action = Class.forName(className).
                                    getMethod("compositionType", paramTypes);
            Object[] args = { params };
            ret = (Adaptor.CompositionType) action.invoke(null, args);
         } catch (Exception e) { }

         if (ret ==  null) {
            ret = Adaptor.CompositionType.NONE;
         }

         compositionTypeCache.put(adaptorSpec, ret);
      }

      return ret;
   }

   /**
    * Install any adaptors, up to the named adaptor, where this object is 
    * currently located.  This method is normally called from an 
    * AdaptorWindow.  Note that this method changes the boundaries of the
    * calling AdaptorWindow or IO object.
    * @param currentAdaptor The name of the adaptor after the last one 
    *             that should be installed.  If null, install all adaptors.
    * @param session The session object for this connection.
    * @return Returns true iff the current adaptor was called for in this 
    *         plan.  If false is returned, the current adaptor had better 
    *         forward this message and then die.
    */
   public boolean installAdaptors(String currentAdaptor, Session session) {

      // There is a current adaptor, but it's not coming up in the plan!
      // It must be a deletion.
      if ((currentAdaptor != null) && 
          (!isAdaptorComingUp(currentAdaptor, session))) {
         return false;
      }

      for (String adaptor = getNextAdaptor(session); 
           ! sameAdaptors(adaptor, currentAdaptor);
           adaptor = getNextAdaptor(session)) {

         if (Debug.on) {
            Debug.println(Debug.DATAFLOW, Debug.MOREINFO,
                          "In installAdaptors, next adaptor is " + adaptor);
         }

         AdaptorWindow newWindow = null;

         if (! useTargetTypeStack) {
            UpstreamBoundary oldUB = (UpstreamBoundary) next(Direction.up);

            /* create a new pair, and insert it */
            UpstreamBoundary newUB = new UpstreamBoundary();
            insert(newUB, Direction.up);

            /* create the new adaptor with appropriate bounds */
            newWindow = new AdaptorWindow(adaptor, session, 
                                          oldUB, newUB.pair());
         }

         Adaptor.CompositionType ct = getCompositionType(adaptor);

         if (ct.isPush()) {
            PlanDistro plan = getPlan(session);
            typeStack.push(plan.planner, plan.version, 
                           previouslyInstalled + installed);
            if (newWindow != null) {
               newWindow.setTypeStack(typeStack, ct);
            }
         } else if (ct.isPop()) {
            if (newWindow != null) {
               newWindow.setTypeStack(typeStack, ct);
            }
            typeStack.pop();
         } else {
            if (newWindow != null) {
               newWindow.setTypeStack(typeStack, ct);
            }
         }

         checkOffNextAdaptor();

         if (useTargetTypeStack && targetTypeStack.equals(typeStack)) {
            // we've reached our goal!
            useTargetTypeStack = false;
         }
      }

      if (currentAdaptor != null) {
         checkOffNextAdaptor();
      }

      return true;
   }

   /**
    * Get the address of the interface for the next node after this one, 
    * as specified by the plan.
    * @param session The session object for this stream.
    * @return The node id of the next node.
    */
   public NodeID getNextInterface(Session session) {
      return getPlan(session).getNextInterface(direction());
   }

   /**
    * Determine the direction of flow for this object.
    * @return Always <clode>Direction.neutral</code>
    */
   public Direction flowDir() {
      return Direction.neutral;
   }

   /**
    * Is this object an event?
    * @return Always returns true.
    */
   public boolean isEvent() {
      return true;
   }

   /**
    * Switch to the next leg of the journey.
    * @return True if we're now on the second leg, False if we completed the
    *         second leg (ie, should we continue?).
    */
   public boolean nextLeg() {
      if (typeStack.size() != 0) {
         if (Debug.on) {
            Debug.println(Debug.PLANNING, Debug.WARNING, 
                          "WARNING: Detected unbalanced adaptor " + 
                          "composition!!!!");
         }
         typeStack.clear();
      }

      if (secondLeg) {
         return false;
      } else {
         secondLeg = true;
         localAdaptors = null;
         return true;
      }
   }

   /**
    * Proceed along the data strea in a neutral direction.  This includes
    * causing any RawIO we've just left to start allowing a data flow in
    * this direction and a call to <code>neutralProceed()</code>.
    * @return The net change in the number of DataFragment objects in the
    *         stream.
    */
   public int proceed() {
      // TODO: fix this kludge!!!
      if (lastRawIO != null) {
         lastRawIO.startInput();
         lastRawIO = null;
      }

      return neutralProceed();
   }

   /**
    * Handle the case in which this object enters an area controlled by a
    * StreamAccessor.  Just call the StreamAccessor's method for handling
    * this object.
    * @param sa The StreamAccessor that this object just entered.
    * @return True iff this object still needs to be dealt with upon return.
    * @exception FMG.Conductor.dataflow.AsyncEventException
    *                     Can be thrown to alert the StreamAccessor (who is the
    *                     usual caller) that somethign unusual happened.
    */
   public boolean handleEntry(StreamAccessor sa) throws AsyncEventException {
      if (Debug.on) {
         Debug.println(Debug.DATAFLOW, Debug.EXTRAINFO, 
                       sa + ": got SwitchPlan");
      }

      return sa.handleEventEntry(this);
   }

   /**
    * Handle the case in which this object is about to exit a StreamAccessor.
    * Just call the StreamAccessor's method for handling exits of this object.
    * @param sa The StreamAccessor that this object is about to exit.
    * @return True iff the caller should care that the exit path is closed.
    */
   public boolean handleExit(StreamAccessor sa) {
      if (Debug.on) {
         Debug.println(Debug.DATAFLOW, Debug.EXTRAINFO, 
                       sa + ": SwitchPlan exiting");
      }

      if (sa instanceof RawIO) {
         lastRawIO = (RawIO) sa;
      }
      return sa.handleEventExit(this);
   }

   public String toString() {
      return "SwitchPlan[" + connID + "," + version + "," + secondLeg + "]";
   }
}
