/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.Conductor.iface;

import FMG.Conductor.Conductor;

/**
 * The Interface class allows an extension module to provide an interface for 
 * Conductor.  The default interface simply prints a message indicating
 * that conductor is initialized and ready for connections before returning.
 * <p>
 * More sophistocated interfaces can be constructed by subclassing the 
 * Interface class and over-riding the <code>runInterface</code> method.
 *
 * @author Mark Yarvis (yarvis@fmg.cs.ucla.edu)
 * @version $Revision: 1.3 $
 */

public class Interface {

   /**
    * Creates a new interface of the class specified in the properties
    * file and starts it running by calling runInterface.  If no interface
    * is specified, an instance of this class is created by default.  This 
    * method may or may not return, so it must be called as the last function
    * in the Conductor startup sequence.
    * @return No return value.  May not return.
    */
   public static void createAndStart() {
      String interfaceName = Conductor.getProperty("InterfaceClass");
      Interface iface = null;

      if (interfaceName == null) {
         iface = new Interface();
      } else {
         try {
            Class interfaceClass = Class.forName(interfaceName);
            iface = (Interface) interfaceClass.newInstance();
         } catch (Exception e) {
            System.out.println("Failure loading interface: " + interfaceName);
            e.printStackTrace();
            System.exit(1);
         }
      }
      iface.runInterface();
   }

   /**
    * Run the interface.  This call may or may not return.  This method
    * should only be called by <code>createAndStart()</code>.  You should
    * override this method in a sub-class to implement a new interface.
    * @return No return value.  May not return.
    */
   protected void runInterface() {
   }

}
