/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.Conductor.iface;

import java.io.*;
import java.net.*;
import java.lang.reflect.*;

import FMG.Conductor.Conductor;
import FMG.Conductor.util.Debug;

public class RCInterface extends TimestampInterface {

   protected static String PID_FILE =
       Conductor.getFilenameProperty("RCInterface.PidFile", 
                                     "/var/run/conductor.pid");

   protected static String SOCKET_INTERFACE_CLASS =
       Conductor.getProperty("RCInterface.SocketInterface");

   protected static int INTERFACE_PORT =
       Conductor.getIntProperty("RCInterface.InterfacePort", -1);

   private static class SocketInterfaceServer extends Thread {
      Constructor interfaceConstructor;
      ServerSocket ss;

      private static class SocketInterfaceRunner extends Thread {
         Interface iface;
         SocketInterfaceRunner(Interface iface) {
            this.iface = iface;
         }

         public void run() {
            setPriority(Thread.MIN_PRIORITY);
            if (Debug.on) {
               setName("RCInterface.SocketInterfaceRunner");
            }

            iface.runInterface();
         }
      }

      SocketInterfaceServer(Class interfaceClass, int port) 
                                                   throws Exception {
         ss = new ServerSocket(port);

         Class[] paramTypes = new Class[2];
         paramTypes[0] = InputStream.class;
         paramTypes[1] = PrintStream.class;

         interfaceConstructor = interfaceClass.getConstructor(paramTypes);
      }

      public void run() {
         setPriority(Thread.MIN_PRIORITY);
         if (Debug.on) {
            setName("RCInterface.SocketInterfaceServer");
         }

         try {
            while (true) {
               Socket s = ss.accept();

               if (Debug.on) {
                  Debug.println("RCInterface: received connection from " + 
                                s.getInetAddress());
               }

               if (!s.getInetAddress().equals(s.getLocalAddress())) {
                  if (Debug.on) {
                     Debug.println("RCInterface: disallowed non-local " + 
                                   "socket connection");
                  }
                  s.close();
                  continue;
               }

               Object[] params = new Object[2];
               params[0] = s.getInputStream();
               params[1] = new PrintStream(s.getOutputStream());

               Interface iface = (Interface) interfaceConstructor.
                                                      newInstance(params);

               (new SocketInterfaceRunner(iface)).start();
            }
         } catch (Exception e) {
            Debug.println(Debug.WARNING, 
                          "RCInterface: can't accept " + 
                          "connections for socket interface");
         }
      }
   }

   protected void touchPidFile() {
      try {
         FileWriter fw = new FileWriter(PID_FILE);
         fw.close();
      } catch (IOException e) {
         Debug.println("Error writing PID file '" + PID_FILE + "': " + e);
         System.exit(1);
      }
   }

   protected void runInterface() {
      Thread.currentThread().setPriority(Thread.MIN_PRIORITY);
      if (Debug.on) {
         Thread.currentThread().setName("RCInterface");
      }

      touchPidFile();

      if ((SOCKET_INTERFACE_CLASS != null) && (INTERFACE_PORT != -1)) {
         startSocketInterface();
      }

      if (Debug.on) {
         runTimer();
      }
   }

   protected void startSocketInterface() {
      try {
         Class socketInterfaceClass = Class.forName(SOCKET_INTERFACE_CLASS);

         (new SocketInterfaceServer(socketInterfaceClass, 
                                    INTERFACE_PORT)).start();
      } catch (Exception e) {
         Debug.println(Debug.WARNING, 
                           "Warning: could not create socket interface " + 
                           SOCKET_INTERFACE_CLASS);
         e.printStackTrace();
      }
   }
}
