/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.Conductor.monitor;

import FMG.Conductor.comm.*;
import FMG.Conductor.planning.messages.PlanningInfo;

/**
 * This class implements the default monitor, which does very little.
 * No actual node monitoring is done, and only the information passed
 * to <code>getLocalInfo()</code> is returned for planning.
 * This class is primarily intended for use with the StaticPlanner which 
 * needs no node state information to accomplish planning.
 *
 * @author Mark Yarvis (yarvis@fmg.cs.ucla.edu)
 * @version $Revision: 1.7 $
 *
 * @see FMG.Conductor.planning.StaticPlanner
 */

public class DefaultMonitor extends EnvMonitor {

   /**
    * This method returns a PlanningInfo message containing only the
    * data passed to it.  It is called by the Path object to gather
    * the status of the node and links for forwarding to the planner.
    * @param connID The id of the connection for which planning is occurring.
    * @param nodeID The local nodeID.
    * @param prevNode The id of the previous node in the planning path.
    * @param nextNode The id of the next node in the planning path. 
    * @param prevInterface The interface to use to talk with the previous node.
    * @param nextInterface The interface to use to talk with the next node.
    * @param srcNode The id of the ultimate connection source.
    * @param srcPort The port number of the ultimate connection source.
    * @param dstNode The id of the ultimate connection destination.
    * @param dstPort The port number of the ultimate connection destination.
    * @param isPlanner A boolean indicating whether this is the last node
    *                  in the planning path.  If so, nextNode is probably 
    *                  null or invalid.
    * @param isRoot A boolean indicating whether this is the first node
    *               in the planning path.  If so, nextNode is probably null
    *               or invalid.
    * @return A message object containing the planning data.  In this case, 
    *         very little.
    */
   public PlanningInfo getLocalInfo(ConnID connID, NodeID nodeID, 
                                    NodeID prevNode, NodeID prevInterface,
                                    NodeID nextNode, NodeID nextInterface,
                                    NodeID srcNode, int srcPort,
                                    NodeID dstNode, int dstPort,  
                                    boolean isPlanner, boolean isRoot) {
      return PlanningInfo.getNewMessage(connID, nodeID, 
                                        prevNode, prevInterface, 
                                        nextNode, nextInterface, 
                                        isPlanner, isRoot);
   }

   /**
    * This method is called to register a new output stream.  In this case,
    * it does nothing.
    * @param nextNode The id of the node to which this output stream flows.
    * @param outputStream The output stream that was created.
    * @return No return value.
    */
   public synchronized void registerOutputStream(NodeID nextNode, 
                                    BasicMessageOutputStream outputStream) {
   }

   /**
    * This method is called to unregister an output stream.  In this case,
    * it does nothing.
    * @param outputStream The output stream that has closed.
    * @return No return value.
    */
   public synchronized void unregisterOutputStream(
                                    BasicMessageOutputStream outputStream) {
   }

   /**
    * This method is called to register a new datapath.  In this case, it
    * does nothing.
    * @param connID The id of the connection this datapath is associated with.
    * @param nextNode The next node along the path.
    * @param request The expected data parameters for this connection.
    * @return No return value.
    */
   public synchronized void registerDatapath(ConnID connID, NodeID nextNode, 
                                             LinkParams request) {
   }

   /**
    * This method is called to unregister a datapath.  In this case, it
    * does nothing.
    * @param connID The id of the connection this datapath is associated with.
    * @param nextNode The next node along the path.
    * @return No return value.
    */
   public synchronized void unregisterDatapath(ConnID connID,  
                                                   NodeID nextNode) {
   }

}
