/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.Conductor.monitor;

import FMG.Conductor.Conductor;
import FMG.Conductor.util.Debug;
import FMG.Conductor.comm.*;
import FMG.Conductor.planning.messages.PlanningInfo;

/**
 * The EnvMonitor is an abstract class which can be sub-classed to create
 * an environment monitor for Conductor.  There is one environment monitor
 * in the system at a time.  The user can select the class of the desired 
 * environment monitor class using the MonitorClass property.  If no 
 * environment monitor class is specified, DefaultMonitor is used.
 * <p>
 * An environment monitor is expected to provide information to the planner
 * about the local node and adjacent links.  It can get this information 
 * either from external sources or by monitoring local activities, such as
 * the number of bytes transmitted along a particular link over some time
 * period.
 *
 * @author Mark Yarvis (yarvis@fmg.cs.ucla.edu)
 * @version $Revision: 1.11 $
 *
 * @see FMG.Conductor.monitor.DefaultMonitor
 */

public abstract class EnvMonitor {

   /**
    * Determine the correct type of environment monitor and create one.
    * The user can controll the type using the MonitorClass property.
    * If none is psecified, the default monitor class is DefaultMonitor.
    * @return A new environment monitor of the type requested by the user.
    */
   public static EnvMonitor createMonitor() {
      /* 
       * Normally this is only called once, so we're not too concerned
       * with efficiency.
       */
      String monitorName = Conductor.getProperty("MonitorClass");

      if (monitorName == null) {
         if (Debug.on) {
            Debug.println(Debug.WARNING, "Using default environment monitor");
         }
         return new DefaultMonitor();
      } else {
         try {
            Class monitorClass = Class.forName(monitorName);
            return (EnvMonitor) monitorClass.newInstance();
         } catch (Exception e) {
            System.out.println("Failure loading environment monitor");
            e.printStackTrace();
            System.exit(1);
            return null;  // to allow compilation
         }
      }
   }

   /**
    * This method is called by the planner to obtain the current conditions
    * for the local node and the links to the previous and next nodes.  This
    * class must be overriden to provide appropriate information and may
    * return a subclass of PlanningInfo to allow planner-specific information
    * to be delivered.
    * @param connID The id of the connection for which planning is occurring.
    * @param nodeID The local nodeID.
    * @param prevNode The id of the previous node in the planning path.
    * @param nextNode The id of the next node in the planning path.
    * @param nextInterface The interface to use to talk with the next node.
    * @param srcNode The id of the ultimate connection source.
    * @param srcPort The port number of the ultimate connection source.
    * @param dstNode The id of the ultimate connection destination.
    * @param dstPort The port number of the ultimate connection destination.
    * @param isPlanner A boolean indicating whether this is the last node 
    *                  in the planning path.  If so, nextNode is probably 
    *                  null or invalid.
    * @param isRoot A boolean indicating whether this is the first node 
    *               in the planning path.  If so, nextNode is probably null 
    *               or invalid.
    * @return A message object containing the planning data.
    */
   public abstract PlanningInfo getLocalInfo(ConnID connID, NodeID nodeID, 
                                         NodeID prevNode, NodeID prevInterface,
                                         NodeID nextNode, NodeID nextInterface,
                                             NodeID srcNode, int srcPort,
                                             NodeID dstNode, int dstPort,
                                             boolean isPlanner, 
                                             boolean isRoot);

   /**
    * This method is called by the datapath code whenever a new output stream
    * is created.  This allows the monitor data being transferred.
    * @param nextNode The id of the node to which this output stream flows.
    * @param outputStream The output stream that was created.
    * @return No return value.
    */
   public abstract void registerOutputStream(NodeID nextNode, 
                                    BasicMessageOutputStream outputStream);

   /**
    * This method is called by the datapath code whenever an output stream
    * is closing.  Any monitoring of this output stream should be
    * discontinued.
    * @param outputStream The output stream that has closed.
    * @return No return value.
    */
   public abstract void unregisterOutputStream(
                                    BasicMessageOutputStream outputStream);

   /**
    * This method is called by the datapath code whenever a new datapath
    * is starting up.  The call informs the monitor of the expected 
    * characteristics of the data stream.
    * @param connID The id of the connection this datapath is associated with.
    * @param nextNode The next node along the path.
    * @param request The expected data parameters for this connection.
    * @return No return value.
    */
   public abstract void registerDatapath(ConnID connID, NodeID nextNode, 
                                                      LinkParams request);

   /**
    * This method is called by the datapath code whenever an existing
    * datapath is shutting down.  The monitor can clean up any state 
    * associated with this connection.
    * @param connID The id of the connection this datapath is associated with.
    * @param nextNode The next node along the path.
    * @return No return value.
    */
   public abstract void unregisterDatapath(ConnID connID,  NodeID nextNode);

}
