/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.Conductor.net;

/**
 * This class holds one entry including address and mask.  An entry
 * may be "inclusive" (matches are included in the set) or "exclusive"
 * (matches are not included in the set).
 * @author Mark Yarvis (yarvis@fmg.cs.ucla.edu)
 * @version $Revision: 1.3 $
 */
public class AddressEntry {

   /** The address of this entry in byte format. */
   protected byte[] address = new byte[4];

   /**
    * The mask of this entry in byte format.  Ones indicate significant 
    * bits.
    */
   protected byte[] mask = new byte[4];

   /** Is this an inclusive entry (or an exclusive one)? */
   protected boolean inclusiveEntry = true;

   /**
    * Create a new entry by parsing a string.  The string should be
    * in the format described above.
    * @param entry The string to parse to create this entry.
    * @exception java.lang.NumberFormatException If the string does not 
    *                                            coform to the expected 
    *                                            format.
    */
   public AddressEntry(String entry) throws NumberFormatException {

      if (entry.charAt(0) == '!') {
         entry = entry.substring(1, entry.length());
         inclusiveEntry = false;
      }

      int slashPos = entry.indexOf('/');

      String addrStr = (slashPos == -1 ? entry : 
                                         entry.substring(0, slashPos));
      String maskStr = (slashPos == -1 ? "" : 
                                         entry.substring(slashPos+1, 
                                                         entry.length()));

      for (int i=0; i<4; i++) {
         if (i < 3) {
            int nextDot = addrStr.indexOf(".");
            if (nextDot == -1) {
               throw new NumberFormatException("Wrong number of octets");
            }
            address[i] = (byte) Integer.parseInt(addrStr.
                                                    substring(0,nextDot));
            addrStr = addrStr.substring(nextDot+1, addrStr.length());
         } else {
            address[i] = (byte) Integer.parseInt(addrStr);
         }
      }

      int maskVal = (maskStr.length()==0 ? 32 : Integer.parseInt(maskStr));
      for (int i=0; i<4; i++) {
         if (maskVal >= 8) {
            mask[i] = (byte) 255;
            maskVal -= 8;
         } else if (maskVal ==0) {
            mask[i] = 0;
         } else {
            int zeroBits = 8-maskVal;
            mask[i] = (byte) (255 << zeroBits);
            maskVal = 0;
         }
      }
   }

   /**
    * Does the given address (in byte notation) match this entry?
    * @param target The address to test (in byte notation).
    * @return Returns true iff the significant bits (as determined by the 
    *         mask of this entry) of the entry matches those of
    *         the argument.
    */
   public boolean isMatch(byte[] target) {
      int len = target.length;

      if (len != address.length) {
         return ! inclusiveEntry;
      }

      for (int i=0; i<len; i++) {
         if ((target[i]&mask[i]) != (address[i]&mask[i])) {
            return ! inclusiveEntry;
         }
      }

      return inclusiveEntry;
   }

   /**
    * Create a human-readable string representing this object.
    * @return A new String representing the contents of this object.
    */
   public String toString() {
      StringBuffer sb = new StringBuffer();

      sb.append(inclusiveEntry ? '+' : '-');

      for (int i=0; i<4; i++) {
         sb.append((int) address[i] & 0xFF);
         if (i < 3) {
            sb.append(".");
         }
      }

      sb.append('(');

      for (int i=0; i<4; i++) {
         sb.append((int) mask[i] & 0xFF);
         if (i < 3) {
            sb.append(".");
         }
      }

      sb.append(')');

      return sb.toString();
   }
}
