/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.Conductor.net;

import java.util.Vector;
import java.io.*;

/**
 * This class allows a set of addresses to be defined.  Each AddressList 
 * instance contains a list of entries.  Each entry consists of an address
 * and a mask.  The maks indicates which address bits are significant.  
 * Each entry is either positive entry (describing what's in the set) or a 
 * negative entry (describing what's outside of the set).  
 * <p>
 * The AddressList class alows an address to be tested to see if it's in or
 * out of the described set.  The address is checked against each entry
 * sequentially.  The first entry that matches determines whether the
 * address is in the set (if it's a positive entry) or out of the set (if
 * it's a negative entry).  If not match is made, the address is out of
 * the set.  This default can be altered by simply using a positive entry
 * with a 0 mask as the last entry.
 * <p>
 * An AddressList can be created from a File or by successive calls to
 * addEntry().  When a file is used, each entry is expected to be on
 * a new line.  Comments (beginging with #) and blank lines are ignored.
 * <p>
 * Each entry is a string in the fillowing format:
 * <p>
 *      <code>[!]xxx.xxx.xxx.xxx[/nn]</code>
 * <p>
 * Each xxx is an octet of the address in decimal notation.  An entry 
 * <b>without</b> the "!" indicates a positive entry (the set includes 
 * this address).  If the "!" is specified, matching addresses are not 
 * included in the set.  If trailing mask is included, nn bits are 
 * considered significant.  If the mask is omitted, the entire address is
 * significant (equivalent to specifying "/24").
 * <p>
 * Note: It doesn't matter if you use network byte order or host byte
 * order, but be consistant.
 *
 * @author Mark Yarvis (yarvis@fmg.cs.ucla.edu)
 * @version $Revision: 1.9 $
 */
public class AddressList {

   /** 
    * The entries to be checked when isIncludedAddress() is called.  This 
    * array is populated from <code>entryVector</code> when 
    * <code>populateArray()</code> is called. 
    */
   protected AddressEntry[] entries;

   /** The location where new entries are placed. */
   private Vector entryVector = new Vector();

   /** 
    * Create an empty AddressList object.
    */
   public AddressList() {
   }

   /**
    * Remove any comments and leading or trailing whitespace from a string.
    * Comments begin with a # and terminate at the end of the line.
    * @param s The input string.
    * @return The input string with commentds and leading and trailing 
    *         whitespace removed.
    */
   protected String removeCommentsAndWhitespace(String s) {
      // remove comments
      int commentStart = s.indexOf('#');
      if (commentStart != -1) {
         s = s.substring(0, commentStart);
      }

      // no whitespace is here
      if ((s.indexOf(' ') == -1) && (s.indexOf('\t') == -1)) {
         return s;
      }

      // remove leading and trailing whitespace
      int startPos;
      int endPos;
      for (startPos = 0; ; startPos++) {
         char c = s.charAt(startPos);
         if ((c!=' ') && (c!='\t')) {
            break;
         }
      }
      for (endPos = s.length(); ; endPos--) {
         char c = s.charAt(endPos-1);
         if ((c!=' ') && (c!='\t')) {
            break;
         }
      }
      return s.substring(startPos, endPos);
   }

   /**
    * Create an address list from a file.  Open and read the given
    * file.  The file format is described above.
    * @param f The file to read.
    */
   public AddressList(File f) throws IOException {
      BufferedReader in = new BufferedReader(new FileReader(f));
      while (true) {
         String s = in.readLine();

         if (s == null) {
            break;
         }

         s = removeCommentsAndWhitespace(s);

         if (s.length() != 0) {
            addEntry(s);
         }
      }
   }

   /**
    * Add an entry to an addresslist from the given string.  The
    * string format is given above.
    * @param s A string representing the new entry.
    * @return no return value.
    */
   public void addEntry(String s) {
      entryVector.addElement(new AddressEntry(s));
      entries = null;
   }

   /**
    * If a new entry has been added since the last call, create a 
    * new entry array from the entry vector.  This method should
    * be called before accessing the entry array.
    * @return no return value.
    */
   private final void populateArray() {
      if (entries == null) {
         entries = new AddressEntry[entryVector.size()];
         entryVector.copyInto(entries);
      }
   }

   /**
    * Determine if the given address is included in the set
    * described by this object.  The algorithm for determining
    * what addresses match is described above.
    * @param targetAddr The address to test.
    * @return true iff the address is included in the addresses described 
    *         by this object.
    */
   public boolean isIncludedAddress(byte[] targetAddr) {
      populateArray();

      for (int i=0; i<entries.length; i++) {
         if (entries[i].isMatch(targetAddr)) {
            return true;
         }
      }

      // non entries are excluded (change by having 0.0.0.0/0 as last entry)
      return false;
   }

   /**
    * Generate a human-readable representation of this object.
    * @return A new string representing this object.
    */
   public String toString() {
      StringBuffer sb = new StringBuffer("AddressList[");

      populateArray();

      for (int i=0; i<entries.length; i++) {
         sb.append(entries[i].toString());
         sb.append(",");
      }

      sb.append(']');
      return sb.toString();
   }

}
