/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.Conductor.net;

import java.io.*;

/**
 * This class implements an output stream that counts the number of
 * bytes output.
 *
 * @author Mark Yarvis (yarvis@fmg.cs.ucla.edu)
 * @version $Revision: 1.4 $
 */

public class OutputCounter extends FilterOutputStream {

   /** The number of bytes written so far. */
   private long count = 0;

   /**
    * Create a new counter stream wraped around another output stream.
    * @param out The original output stream through which data will actuall 
    *            flow.
    */
   public OutputCounter(OutputStream out) {
      super(out);
   }

   /**
    * Write a byte array to the underlying output stream and increment the 
    * counter by the size of the array.
    * @param b The byte array to write.
    * @exception java.io.IOException If writing to the underlying output
    *                                array throws an exception, it is 
    *                                forwarded.
    */
   public void write(byte b[]) throws IOException {
      count+=b.length;
      out.write(b);
   }

   /**
    * Write part of a byte array to the underlying output stream and 
    * increment the counter by the amount of data written.
    * @param b The byte array to write from.
    * @param off The starting position of the data to write.
    * @param len The number of bytes to write.
    * @exception java.io.IOException If writing to the underlying output
    *                                array throws an exception, it is 
    *                                forwarded.
    */
   public void write(byte b[], int off, int len) throws IOException {
      count+=len;
      out.write(b, off, len);
   }

   /**
    * Write a byte to the underlying output stream and increment the counter.
    * @param b The byte to write.
    * @exception java.io.IOException If writing to the underlying output
    *                                array throws an exception, it is 
    *                                forwarded.
    */
   public void write(int b) throws IOException {
      count++;
      out.write(b);
   }

   /**
    * Determine the number of bytes written since this stream was created.
    * @return The number of bytes written so far.
    */
   public long bytesWritten() {
      return count;
   }

}
