/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.Conductor.net;

import java.io.*;
import java.net.*;

import FMG.Conductor.util.Debug;

public class SocketOps {

   public static final int TOS_LOWDELAY = 0x10;
   public static final int TOS_THROUGHPUT = 0x08;
   public static final int TOS_RELIABILITY = 0x04;
   public static final int TOS_LOWCOST = 0x02;

   /**
    * Load the native code.
    */
   static {
      System.loadLibrary("condnet");
   }

   /**
    * Performs a half-close on the input side of the file or socket
    * associated with this input stream.  This only works on FileInputStream
    * objects, but it's more convenient to pass an InputStream here.
    * @param is The input stream of the file or socket to shut down.
    * @return No return value
    * @exception java.io.IOException if the input stream is not a
    *                                FileInputStream or if the shutdown()
    *                                call returns an error
    **/
   public static void shutdown(InputStream is) throws IOException {
      if (is instanceof FileInputStream) {
         int fd = getFD((FileInputStream) is);

         if (Debug.on) {
            Debug.println(Debug.SOCKET, Debug.EXTRAINFO, 
                          "SocketOps: calling shutdown(" + fd + ", 0)");
         }

         shutdownFD(fd, 0);
      } else {
         throw new IOException("Invalid operation");
      }
   }

   /**
    * Performs a half-close on the output side of the file or socket
    * associated with this input stream.  This only works on FileOutputStream
    * objects, but it's more convenient to pass an OutputStream here.
    * @param is The output stream of the file or socket to shut down.
    * @return No return value
    * @exception java.io.IOException if the input stream is not a
    *                                FileOutputStream or if the shutdown()
    *                                call returns an error
    **/
   public static void shutdown(OutputStream os) throws IOException {
      if (os instanceof FileOutputStream) {
         int fd = getFD((FileOutputStream) os);

         if (Debug.on) {
            Debug.println(Debug.SOCKET, Debug.EXTRAINFO, 
                          "SocketOps: calling shutdown(" + fd + ", 1)");
         }

         shutdownFD(fd, 1);
      } else {
         throw new IOException("Invalid operation");
      }
   }

   /**
    * Performs a half-close on the input side of a socket.
    * @param is The socket to shut down.
    * @return No return value
    * @exception java.io.IOException if the shutdown() call returns an error
    **/
   public static void shutdownInput(Socket s) throws IOException {
      int fd = getFD(s);

      if (Debug.on) {
         Debug.println(Debug.SOCKET, Debug.EXTRAINFO, 
                       "SocketOps: calling shutdown(" + fd + ", 0)");
      }

      shutdownFD(fd, 0);
   }

   /**
    * Performs a half-close on the output side of a socket.
    * @param is The socket to shut down.
    * @return No return value
    * @exception java.io.IOException if the shutdown() call returns an error
    **/
   public static void shutdownOutput(Socket s) throws IOException {
      int fd = getFD(s);

      if (Debug.on) {
         Debug.println(Debug.SOCKET, Debug.EXTRAINFO, 
                       "SocketOps: calling shutdown(" + fd + ", 1)");
      }

      shutdownFD(fd, 1);
   }

   public static void setInputBuffer(Socket s, int size) throws IOException {
      int fd = getFD(s);

      if (Debug.on) {
         Debug.println(Debug.SOCKET, Debug.EXTRAINFO, 
                       "SocketOps: setting input buffer of file descriptor " + 
                       fd + " to " + size);
      }

      setInputBuffer(fd, size);
   }

   public static void setOutputBuffer(Socket s, int size) throws IOException {
      int fd = getFD(s);

      if (Debug.on) {
         Debug.println(Debug.SOCKET, Debug.EXTRAINFO, 
                       "SocketOps: setting output buffer of file descriptor " + 
                       fd + " to " + size);
      }

      setOutputBuffer(fd, size);
   }

   // valid priorities are 0 to 6
   public static void setPriority(Socket s, int pri) throws IOException {
      int fd = getFD(s);

      if (Debug.on) {
         Debug.println(Debug.SOCKET, Debug.EXTRAINFO,
                       "SocketOps: setting priority on file desciptor " + 
                       fd + " to " + pri);
      }

      setPriority(fd, pri);
   }

   public static void setTOS(Socket s, int tos) throws IOException {
      int fd = getFD(s);

      if (Debug.on) {
         Debug.println(Debug.SOCKET, Debug.EXTRAINFO,
                       "SocketOps: setting TOS on file desciptor " + 
                       fd + " to " + tos);
      }

      setTOS(fd, tos);
   }

   /**
    * Call <code>shutdown()</code> on a file descriptor.
    * @param fd The file descriptor to operate on.
    * @param how The second argument to the <code>shutdown()</code> system call
    * @return No return value.
    */
   protected static native void shutdownFD(int fd, int how) throws IOException;

   /**
    * Call <code>setsockopt()</code> to set the value of the SO_RECVBUF 
    * option to the specified socket size.
    * @param fd The file descriptor to operate on.
    * @param how The size of the desired input buffer.
    * @return No return value.
    */
   protected static native void setInputBuffer(int fd, int size) 
                                                         throws IOException;

   /**
    * Call <code>setsockopt()</code> to set the value of the SO_SNDBUF 
    * option to the specified socket size.
    * @param fd The file descriptor to operate on.
    * @param how The size of the desired output buffer.
    * @return No return value.
    */
   protected static native void setOutputBuffer(int fd, int size) 
                                                         throws IOException;

   protected static native void setPriority(int fd, int pri)
                                                         throws IOException;

   protected static native void setTOS(int fd, int tos) throws IOException;

   /**
    * Obtain the file descriptor (in the C/Unix sense) from a given
    * FileDescriptor object.
    * @param fdes The file descriptor oject from which we want the 
    *             underlying file descriptor.
    * @return An integer representing the desired file descriptor.
    */
   protected static native int getFD(FileDescriptor fdes);

   /**
    * Obtain the file descriptor (in the C/Unix sense) from a given
    * Socket object.
    * @param fdes The file descriptor oject from which we want the 
    *             underlying file descriptor.
    * @return An integer representing the desired file descriptor.
    */
   protected static native int getFD(Socket s);

   /**
    * Obtain the file descriptor (in the C/Unix sense) from a given
    * DatagramSocket object.
    * @param fdes The file descriptor oject from which we want the 
    *             underlying file descriptor.
    * @return An integer representing the desired file descriptor.
    */
   protected static native int getFD(DatagramSocket s);

   /**
    * Obtain the file descriptor (in the C/Unix sense) from a given
    * FileInputStream object.
    * @param fdes The file descriptor oject from which we want the 
    *             underlying file descriptor.
    * @return An integer representing the desired file descriptor.
    */
   protected static native int getFD(FileInputStream is);

   /**
    * Obtain the file descriptor (in the C/Unix sense) from a given
    * FileOutputStream object.
    * @param fdes The file descriptor oject from which we want the 
    *             underlying file descriptor.
    * @return An integer representing the desired file descriptor.
    */
   protected static native int getFD(FileOutputStream os);
}
