/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.Conductor.planning;

import java.net.*;
import java.io.*;
import java.util.Timer;
import java.util.TimerTask;

import FMG.Conductor.Conductor;
import FMG.Conductor.net.AddressList;
import FMG.Conductor.util.*;
import FMG.Conductor.comm.*;
import FMG.Conductor.planning.comm.*;
import FMG.Conductor.planning.messages.*;
import FMG.Conductor.dataflow.*;
import FMG.Conductor.dataflow.messages.SwitchPlan;
import FMG.Conductor.monitor.EnvMonitor;
import FMG.Conductor.security.SecurityCoordinator;
import FMG.Conductor.dataflow.comm.FastPathIO;


/**
 * The Path object implements the heart of the architectural support for
 * planning.  It executes the local part of the distributed algorithm that 
 * creates an end-to-end connection across Conductor-enabled nodes, gathers
 * planning information at a central location, and distributes the generated
 * plan.
 * <p>
 * Planning occurs over a planning path.  The planning path is entirely 
 * different from the data path.  The planning path is made up of all 
 * Conductor-enabled nodes along the path between the client and server 
 * nodes.  Planning messages flow between adjacent nodes through a TCP
 * connection that is shared by all coincident planning paths.  For example,
 * if there are two paths A-X-Y-B and C-X-Y-D, a single TCP connection between
 * X and Y is created to handle planning traffic for both planning paths.
 * <p>
 * The actual planning activity gathers all information about local conditions
 * from individual nodes to a single node, known as the planner node.  The
 * planner node is normally the server endpoint of the connection.  The
 * plan is then forwarded and distributed back along the path toward the
 * other endpoint (the client), known as the root.  In the case of a
 * partitioned network, a different pair of root and planner nodes may
 * be used as needed.
 * <p>
 * The algorithms implemented by this object are as follows:
 * <ul>
 *    <li>Trigger Planning
 *       <ul>
 *          <li>A node can trigger planning by sending a TriggerPlanning
 *              message toward the root node.
 *          <li>When TriggerPlanning is received at a non-root node, it is
 *              forwarded toward the root node.
 *          <li>When TriggerPlanning is received at the root node, information
 *              gathering begins.
 *          <li>Initial information gathering is triggered by a 
 *              TriggerPlanning message that is delivered to the Path object 
 *              at startup.
 *       </ul>
 *    <li>Path Creation
 *       <ul>
 *          <li>A PathRequest is sent (via UDP) addressed to the planner, 
 *               whenever ...
 *             <ul>
 *                <li>A PathRequest is intercepted by a node that isn't the 
 *                    planner.
 *                <li>A message needs to be sent (in the direction of the 
 *                    planner), but there is no path.
 *                <li>A timeout has expired waiting to receive PlanPathInit
 *                    in response to PathRequest.
 *             </ul>
 *          <li>When PathRequest is received, ...
 *             <ul>
 *                <li>If no connection exists to the previous node, 
 *                    one is created
 *                <li>A PlanPathInit message is sent to the previous node in
 *                    response
 *             </ul>
 *          <li>When PathPathInit is received, panding messages
 *              (those messages that could not be handled because there was
 *              no output path) are processed.
 *       </ul>
 *    <li>Information Gathering and Plan Distribution
 *       <ul>
 *          <li>When the root node receives TriggerPlanning, or when any node
 *              receives PlanningDone, ...
 *             <ul>
 *                <li>A PlanningInfo message is generated by the local 
 *                    monitor and sent to the next node.
 *                <li>A PlanningDone message is sent to the next node.
 *             </ul>
 *          <li>Any PlanningInfo message received is forwarded toward the 
 *              planner.
 *          <li>When the planner receives a PlanningDone message, it formulate 
 *              a plan and sends it in a PlanDistro along the revese path
 *          <li>Each node forwards PlanDistro toward the root.
 *          <li>When the root node receives PlanDistro, planning is complete.
 *       </ul>
 * </ul>
 * <p>
 * The path object itself runs in its own thread, allowing it to block while
 * writing messages to adjacent nodes without blocking other paths from
 * receiving messages.  Messages may be delivered to a Path by calling
 * <code>handleMessage()</code>.  These messages are dropped in a queue
 * and picked up later by the Path thread.  The path processes each message
 * in turn.  If the resulting action requires sending a message, and an
 * error occurs, the originally received message is placed on a pending queue 
 * for future processing.  Once a failed connection is re-established, the 
 * pending messages are moved back into the main queue and are processed
 * again.
 *
 * @author Mark Yarvis (yarvis@fmg.cs.ucla.edu)
 * @version $Revision: 1.75 $
 */

public class Path extends Job {

   protected class TimeoutHandler extends TimerTask {
      public void run() {
         timeExpired();
      }
   }

   /*
    * static info for Path class
    */

   /** Maximum timeout to wait for a PlanPathInit (in millisec) */
   private static final long MAX_UDP_TIMEOUT = 
                   Conductor.getIntProperty("MaxPlanPathInitTimeout", 10000);

   /** Minimum timeout to wait for a PlanPathInit (in millisec) */
   private static final long MIN_UDP_TIMEOUT = 
                   Conductor.getIntProperty("MinPlanPathInitTimeout", 500);

   private static final boolean NAT_IN_USE = 
                   Conductor.getBooleanProperty("NATinUse", true);

   /** The timer object used to implement a timeout */
   private static Timer timer = new Timer();

   /** The commMgr used to communicate with the comm objects */
   protected static PlanCommMgr commMgr = null;

   /** A cached copy of the local host id */
   private static NodeID localhost = NodeID.getNewLocalID();

   /** The local enivronment monitor */
   protected static EnvMonitor envMonitor = null;

   /** Is this node at the edge of the Conductor network? */
   protected static boolean isEdgeNode = false;

   /** Name of the file from which to read conductor network membership */
   private static final String CONDUCTOR_NET_FILE = 
          Conductor.getFilenameProperty("ConductorNetFile",
                                        "./config/ConductorNet");

   /** 
    * The list of nodes/nets that are part of the Conductor net.  This 
    *  variable is only valid for edge nodes.
    */
   protected static AddressList conductorNet;

   /*
    * Instance info for a specific connection.
    */

   /** The connection id of this path */
   protected ConnID connID;

   /** The object that holds the parts of this session */
   protected Session session;

   /** The address of the client endpoint of the user-level connection */
   protected NodeID sourceAddr = new NodeID();

   /** The port number of the client endpoint of the user-level connection */
   protected int sourcePort;

   /** The address of the server endpoint of the user-level connection */
   protected NodeID targetAddr = new NodeID();

   /** The port number of the client endpoint of the user-level connection */
   protected int targetPort;

   /*
    * Communication mechanisms
    */

   /** The object to use to send along the path toward the client. */
   protected PlanComm outC;

   /** The object to use to send along the path toward the server. */
   protected PlanComm outS;

   /** New messages, yet to be processed. */
   private MessageSync incoming = new MessageSync();

   /** Messages which can't be processed until outC!=null */
   private MessageQueue pendingOutC = new MessageQueue(); 

   /** Messages which can't be processed until outS!=null */
   private MessageQueue pendingOutS = new MessageQueue();

   /** The security input and output filter for this path. */
   private SecurityCoordinator securityCoordinator;

   /*
    * Structures for path operations
    */

   /** The planner object (on the planner node only) */
   private Planner planner;

   /** The last distributed plan. */
   private PlanDistro currentPlan;

   /** An object used to synchronize setting and getting the current plan.  */
   private Object planSync = new Object();

   /** 
    * An interface to the datapath assocaited with this planning path.  This
    * is used to tell the datapath to implement a new plan.
    */
   private IOManager serverManager;

   /** The previous node along the path.  */
   protected NodeID prevNode = new NodeID();

   /** The next node along the path.  */
   protected NodeID nextNode = new NodeID();

   /** The interface to use to communicate with the next node alont the path */
   protected NodeID interfaceToNext = new NodeID();

   /** 
    * The interface to use to communicate with the previous node alont 
    * the path.
    */
   protected NodeID interfaceToPrev = new NodeID();

   /** Is this the root node for the current planning operation? */
   protected boolean isRootNode = false;

   /** Is this the planner node for the current planning operation? */
   protected boolean isPlannerNode = false;

   /** Are we waiting for PlanPathInit? */
   private boolean pathRequestOutstanding = false;

   /** 
    * The time at which planning began.  Only valid on the root node.  
    * Only used for measurement purposes. 
    */
   private long planningStartTime = 0;

   /** The amount of time the UDP timer was set for the last request. */
   private long lastUdpTimeoutLength = MIN_UDP_TIMEOUT;

   /**
    * Create a new Path object for a connection specified by the connection id
    * and the source and destination endpoints.  The thread associated with
    * this path object is started automatically.  If this is the planner
    * node, the server endpoint of the datapath is started up.
    * @param session The object holding the parts of this session.
    * @param sourceAddr The address of the client node.
    * @param sourcePort The port number used at the client endpoint.
    * @param targetAddr The address of the server node.
    * @param targetPort The port number used at the server endpoint.
    * @param rawSocket The raw client socket from which this connection came 
    *                  (null if this isn't the client node).
    */
   public Path(Session session, NodeID sourceAddr, int sourcePort, 
                                NodeID targetAddr, int targetPort,
                                Socket rawSocket) {
      this.session = session;
      this.sourceAddr.copy(sourceAddr);
      this.sourcePort = sourcePort;
      this.targetAddr.copy(targetAddr);
      this.targetPort = targetPort;

      this.session.setPath(this);

      connID = session.getConnID();

      if (Debug.on) {
         Debug.println(Debug.PLANNING, Debug.INFO, 
                       "Creating path from " + sourceAddr + "/" + sourcePort + 
                            " to " + targetAddr + "/" + targetPort +
                            "(ID=" + connID + ")");
      }

      isRootNode = (rawSocket != null);
      isPlannerNode = (targetAddr.isLocalNode() || 
                       (isEdgeNode && ! conductorNet.isIncludedAddress(
                                                      targetAddr.getBytes())));

      if (isPlannerNode && isRootNode) {
         // Conductor should get out of the way
         FastPathIO.createDatapath(rawSocket, 
                                   sourceAddr.getAddress(), sourcePort, 
                                   targetAddr.getAddress(), targetPort);
         // Don't do any more startup
         return;

      } else if (isPlannerNode) {
         planner = Planner.createPlanner(connID, sourceAddr, sourcePort, 
                                                 targetAddr, targetPort);

         // connect to server
         IOManager.createRawServer(sourceAddr.getAddress(), sourcePort, 
                                   targetAddr.getAddress(), targetPort, 
                                   session);
         serverManager = session.getIOManager(Direction.up);

      } else if (isRootNode) {
         IOManager.createRawClient(rawSocket, session);
      }

      securityCoordinator = new SecurityCoordinator(session,
                                                    sourceAddr, sourcePort, 
                                                    targetAddr, targetPort);

      // Start thread to process messages
      ThreadRecycler.startNewJob(this, Thread.MAX_PRIORITY - 1);
   }

   /**
    * Get the planning system ready to go.  In particular, the 
    * BasicMessageInputStream is informed of the types of messages we intend
    * to send.
    * @return No return value.
    * @see FMG.Conductor.comm.BasicMessageInputStream
    */
   public static void initialize(boolean makeEdgeNode) {
      PlanComm.registerMessages();
      if (makeEdgeNode) {
         isEdgeNode = true;

         try {
            File netFile = new File(CONDUCTOR_NET_FILE);
            conductorNet = new AddressList(netFile);
            if (Debug.on) {
               Debug.println(Debug.PLANNING, Debug.INFO, 
                             "Conductor network for this edge node is " + 
                             conductorNet);
            }
         } catch (IOException e) {
            throw new RuntimeException("Failure reading conductor " + 
                                       "network file '" + CONDUCTOR_NET_FILE +
                                       "': " + e);
         } catch (NumberFormatException e) {
            throw new RuntimeException("Failure reading conductor " + 
                                       "network file '" + CONDUCTOR_NET_FILE +
                                       "': " + e);
         }
      }
   }

   /**
    * Set the identity of our PlanCommMgr object which we use to interface
    * with the comm objects and talk to Path objects on other nodes.  This
    * method can be called only once, for security reasons.  Subsequent
    * calls will result in a runtime exception.
    * @param pcm The new PlanCommMgr object.
    * @return No return value.
    */
   public static void setPlanCommMgr(PlanCommMgr pcm) {
      if (commMgr == null) {
         commMgr = pcm;
      } else {
         throw new RuntimeException("This shouldn't happen");
      }
   }

   /**
    * Set the identity of our EnvMonitor object which we use to get
    * planning information.  This method can be called only once, for 
    * security reasons.  Subsequent calls will result in a runtime exception.
    * @param envMonitor The new EnvMonitor object.
    * @return No return value.
    */
   public static void setEnvMonitor(EnvMonitor newEnvMonitor) {
      if (envMonitor == null) {
         envMonitor = newEnvMonitor;
      } else {
         throw new RuntimeException("This shouldn't happen");
      }
   }

   /**
    * Send a message, through the "security coordinator", and then across the
    * network to the next Conductor node in the direction of the client.
    * The "security coordinator" will eventually call <code>sendCToNet()</code>
    * to accomplish the actual sending.
    * @param m The message to send.
    * @return True iff the message was successfully written (but that doesn't
    *         guarantee it will arrive), false if there is no connection in
    *         that direction or if an error occurs.
    */
   protected boolean sendC(PlanningMessage m) {
      return securityCoordinator.handleOutgoingToClient(m);
   }

   /**
    * Try to send a message to the next Conductor node in the direction of
    * the client node.  If a failure occurs, record the fact that we can no
    * longer communicate in that direction and wait for my neighbor to 
    * conatact me.
    * <p>
    * Warning: Upon success, the message will be recycled.  If you want to
    *          to maintain a handle to the message, you must call 
    *          <code>retain()</code>
    * <p>
    * Warning: The path object should normally use sendC, which will pass
    *          all messages to the security coordinator.  The security 
    *          coordinator will call this method to do the actual sending.
    *          
    * @param m The message to send.
    * @return True iff the message was successfully written (but that doesn't
    *         guarantee it will arrive), false if there is no connection in
    *         that direction or if an error occurs.
    */
   public boolean sendCToNet(PlanningMessage m) {
      if (outC == null) {
// TODO: when this occurs, should recycle be called?
         return false;
      }

      try {
         outC.sendMessage(m);
         return true;
      } catch (IOException e) {
         outC = null;
// TODO: when this occurs, should recycle be called?
         return false;
      }
   }

   /**
    * Send a message, through the "security coordinator", and then across the
    * network to the next Conductor node in the direction of the server.
    * The "security coordinator" will eventually call <code>sendSToNet()</code>
    * to accomplish the actual sending.
    * @param m The message to send.
    * @return True iff the message was successfully written (but that doesn't
    *         guarantee it will arrive), false if there is no connection in
    *         that direction or if an error occurs.
    */
   protected boolean sendS(PlanningMessage m) {
      return securityCoordinator.handleOutgoingToServer(m);
   }

   /**
    * Try to send a message to the next Conductor node in the direction of
    * the client node.  If a failure occurs, record that fact that we no longer
    * have a connection in that direction and send a PathRequest to try to
    * establish a new connection.
    * <p>
    * Warning: Upon success, the message will be recycled.  If you want to
    *          to maintain a handle to the message, you must call 
    *          <code>retain()</code>
    * <p>
    * Warning: The path object should normally use sendS, which will pass
    *          all messages to the security coordinator.  The security 
    *          coordinator will call this method to do the actual sending.
    *          
    * @param m The message to send.
    * @return True iff the message was successfully written (but that doesn't
    *         guarantee it will arrive), false if there is no connection in
    *         that direction or if an error occurs.
    */
   public boolean sendSToNet(PlanningMessage m) {
      try {
         if (outS!=null) {
            outS.sendMessage(m);
            return true;
         }
      } catch (IOException e) {
         outS = null;
      }

      if (Debug.on) {
         Debug.println(Debug.PLANNING, Debug.MOREINFO, 
                       "Requesting planning path to next node.");
      }

      requestPath(false);

// TODO: when this occurs, should recycle be called?
      return false;
   }

   /**
    * Determin the timeout for the current UDP request.  If this is the
    * first request, use the minimum timeout.  If a request was already
    * outstanding, double the last timeout, up to the maximum timeout.
    * @return The value (in milliseconds) to set the UDP timeout timer
    *         for the next request.
    */
   private long getUdpTimeout() {
      if (pathRequestOutstanding) {
         lastUdpTimeoutLength *= 2;
      } else {
         lastUdpTimeoutLength = MIN_UDP_TIMEOUT;
      }

      if (lastUdpTimeoutLength > MAX_UDP_TIMEOUT) {
         lastUdpTimeoutLength = MAX_UDP_TIMEOUT;
      }

      return lastUdpTimeoutLength;
   }

   protected void scheduleTimeout(long delay) {
      try {
         timer.schedule(new TimeoutHandler(), delay);
      } catch (IllegalStateException e) {
         // handle the case in which the timer thread terminated
         if (Debug.on) {
            Debug.println(Debug.PLANNING, Debug.WARNING, 
                          "WARNING: Timer thread went away ... "
                          + "creating new timer");
         }
         timer = new Timer();
         timer.schedule(new TimeoutHandler(), delay);
         // it shouldn't fail again, so if it does, let it thrown an exception
      }
   }

   /**
    * Send a PathRequest message to the next node along the path.
    * A timeout is set in case it's lost.  When the timeout occurs,
    * <code>timeExpired()</code> will be called.  Normally, however,
    * a call to this method will not send a message if a previously
    * sent PathRequest message is still outstanding (the timeout has
    * yet to be reached and we haven't yet received PlanPathInit).
    * @param force If true, send the request even though there is an
    *              outstanding PathRequest.
    * @return No return value.
    */
   protected synchronized void requestPath(boolean force) {
      if (force || !pathRequestOutstanding) {
         commMgr.seekPath(connID, sourceAddr, sourcePort, 
                                  targetAddr, targetPort);
         scheduleTimeout(getUdpTimeout());
         pathRequestOutstanding = true;
      } else {
         if (Debug.on) {
            Debug.println(Debug.PLANNING, Debug.INFO,
                          this + ": delaying new path request ... one is " + 
                                 "currently outstanding.");
         }
      }
   }

   /**
    * This method implements the timeout functionality for the PathRequest
    * message.  It checks to see if a response ever came for the last 
    * PathRequest method.  If not, a new request is sent.  This method 
    * is called by the TimeoutHandler inner class.
    * @return No return value.
    */
   public synchronized void timeExpired() {
      if (!pathRequestOutstanding) {
         return;
      }

      if (Debug.on) {
         Debug.println(Debug.PLANNING, Debug.INFO, 
                       "Timer expired, retransmitting PathRequest");
      }

      requestPath(true);
   }

   /*
    * Indicates a possible failure in the next node.  We must find a new
    * channel for communicating toward the server.
    */
   public void nextNodeFailed() {
      outS = null;
   }

   /**
    * Trigger planning for this connection.  A TriggerPlanning message is
    * queued for processing by this Path object.
    * @return No return value.
    */
   public void triggerPlanning() {
      // if I'm both root an planner, don't plan
      if (isRootNode && isPlannerNode) {
         return;
      }

      handleMessage(TriggerPlanning.getNewMessage(connID));
   }

   /**
    * Shut down this path.  A PathClose message is queued for processing.
    * It will be forwarded if necessary and then the Path object will remove
    * itself from the list of active paths.  Presumably the data path has
    * completely gone away at this point.
    * @return No return value.
    */
   public void shutdownPath() {
      handleMessage(PathClose.getNewMessage(connID));
   }

   public void handleIncomingFromNet(PlanningMessage m) {
      securityCoordinator.handleIncoming(m);
   }

   /**
    * Queue a message for local processing.  This is required when a message
    * is received from the net for this path object or when a local object
    * needs to trigger somethign to happen.
    * @see #shutdownPath
    * @see #triggerPlanning
    */
   public void handleMessage(PlanningMessage m) {
      try {
         incoming.putMessage(m);
      } catch (MessageFlowClosed e) {
         throw new RuntimeException("This shouldn't happen");
      }
   }

   /**
    * This is the main body for the thread that implements the Path
    * algorithms.  It pulls one message at a time off of the incoming
    * queue and calls <code>processMessage()</code> to do the work.
    * @return No return value.  Only retruns once the path is closed.
    */
   public void run() {
      if (Debug.on) {
         currentThread.setName("Path");
      }

      try {
         while (true) {
            if (processMessage((PlanningMessage) incoming.getMessage())) {
               return;
            }
         }
      } catch (InterruptedException e) {
         throw new RuntimeException("This shouldn't happen");
      } catch (MessageFlowClosed e) {
         throw new RuntimeException("This shouldn't happen");
      }
   }

   /**
    * This method implements the logic of handling one incoming message.
    * Each method is either handled, usually by forwarding some message
    * to somebody else, or queued as a pending message for later 
    * processiing.  The later only occurs if handling a message means sending
    * a message, but the communication channel is down.  In this case,
    * once the channel is up again, the pending messages will move to the
    * incoming queue for processing.  The algorithms implemented in this
    * method are described in the header for this class.
    * @param m A message to process.
    * @return True iff this path is closed and should be shut down.
    */
   public boolean processMessage(PlanningMessage m) {

      if (Debug.on) {
         Debug.println(Debug.PLANNING, Debug.DETAIL, "Path: handling " + m);
      }
      
      if (m instanceof TriggerPlanning) {

         if (isRootNode) {
            if (Measure.on && ((Measure.mask & Measure.PLANNING) != 0)) {
               planningStartTime = Measure.getTime();
               Measure.println("Path: Planning started", planningStartTime);
            }

            m.recycle();

            if (isPlannerNode) {
               throw new RuntimeException("Not implemented");
            }

            // Start planning
            processMessage(PlanningDone.getNewMessage(connID));
         } else {  /* not the root node */
            // Forward to the next node toward the client
            if (! sendC(m)) {
               m.recycle();
               // just drop the message, somone else will trigger planning
            }
         }
      } else if (m instanceof PathRequest) {
         PathRequest pr = (PathRequest) m;
         /* record identity of previous node */
         prevNode.copy(pr.prevNode);

         /* record previous node's interface address */
         NodeID prevInterface = pr.sourceInterface.duplicate();

         if (Debug.on) {
            Debug.println(Debug.PLANNING, Debug.INFO,
                          "Path: Got PathRequest from " + prevNode + 
                          "(" + prevInterface + ")");
         }

         m.recycle();

         // if we don't have an outgoing path, establish one now
         if ((outS == null) && !isPlannerNode) {
            requestPath(false);
         }

         /* get the queue for existing or new connection to prev node */
         outC = commMgr.getCommObj(prevNode, prevInterface);

         /* send init to that prev node */
         PlanPathInit ppi = PlanPathInit.getNewMessage(connID, localhost, 
                                                       prevInterface);

         if (sendC(ppi)) {

            interfaceToPrev.copy(outC.getLocalInterface());

            // put pending messages back into the queue for processing
            try {
               while (!pendingOutC.isEmpty()) {
                  incoming.putMessage(pendingOutC.dequeue());
               }
            } catch (MessageFlowClosed e) {
               throw new RuntimeException("This shouldn't happen");
            }
         } else {
            // we failed, so wait for another init to try again
            ppi.recycle();
         }

      } else if (m instanceof PlanPathInit) {
         PlanPathInit ppi = (PlanPathInit) m;
         /* record the identity of the next node */
         nextNode.copy(ppi.nextNode);

         /* record the approriate comm object */
         outS = commMgr.lookupCommObj(nextNode);
         pathRequestOutstanding = false;

         /* record the interface we use to talk to the next node */
         if (NAT_IN_USE) {
            // with NAT, only the downstream node can know our address
            interfaceToNext.copy(ppi.yourInterface);
         } else {
            interfaceToNext.copy(outS.getLocalInterface());
         }

         m.recycle();

         // put pending messages back into the queue for processing
         try {
            while (!pendingOutS.isEmpty()) {
               incoming.putMessage(pendingOutS.dequeue());
            }
         } catch (MessageFlowClosed e) {
            throw new RuntimeException("This shouldn't happen");
         }
      } else if (m instanceof PlanningInfo) {

         if (isPlannerNode) {

            // TODO: expand planning info beyond node id
            planner.addInfo((PlanningInfo) m);
            m.recycle();
         } else {
            // forward info to planner
            if (! sendS(m)) {
               pendingOutS.enqueue(m);
            }
         }
      } else if (m instanceof PlanningDone) {

         PlanningInfo localInfo = 
                         envMonitor.getLocalInfo(connID, localhost, 
                                            prevNode, interfaceToPrev,
                                            nextNode, interfaceToNext,
                                                 sourceAddr, sourcePort,
                                                 targetAddr, targetPort,
                                                 isPlannerNode, isRootNode);
         if (isPlannerNode) {
            m.recycle();

            // complete and distribute plan
            planner.addInfo(localInfo);
            localInfo.recycle();

//FMG.ConductorExt.iface.TestInterface.doTest();

            Measure.TimeOfDay startTime = null;
            if (Measure.on && ((Measure.mask & Measure.PLANNING) != 0)) {
               startTime = Measure.getTimeOfDay();
            }

            PlanDistro pd = planner.formulatePlan();

            if (Measure.on && ((Measure.mask & Measure.PLANNING) != 0)) {
               Measure.TimeOfDay endTime = Measure.getTimeOfDay();
               Measure.println("Plan formulation started at " + 
                               startTime.secs + "." + startTime.usecsString());
               Measure.println("Plan formulation ended at " + 
                               endTime.secs + "." + endTime.usecsString());
            }

            if (currentPlan == null) {
               pd.version = 0;
            } else {
               pd.version = currentPlan.version+1;
            }

            processMessage(pd);
         } else {
            if (!sendS(localInfo)) {
               localInfo.recycle();
               pendingOutS.enqueue(m);
               return false;
            }
            if (!sendS(m)) {
               pendingOutS.enqueue(m);
               return false;
            }
         }
      } else if (m instanceof PlanDistro) {

         if (Debug.on) {
            Debug.println(Debug.PLANNING, Debug.INFO, 
                          this + " got plan: " + m);
         }

         // as long as we're not re-handling, install this plan
         if ((currentPlan == null) || 
             (currentPlan.version != ((PlanDistro) m).version)) {  
            // install plan locally
            setCurrentPlan((PlanDistro) m);

            if (isPlannerNode) {
               // tell data path to switch plans
               serverManager.forwardSwitchPlan(
                       SwitchPlan.getNewMessage(connID, 
                                                ((PlanDistro) m).version));
            }
         }

         // forward to the next node
         if (! isRootNode) {
            if (!sendC(m)) {
               pendingOutC.enqueue(m);
            }
         } else {
            // Doesn't really recycle, since setCurrentPlan called retain(),
            // but if we don't send it, we have to call recycle.
            m.recycle();
         }

         if (Measure.on && ((Measure.mask & Measure.PLANNING) != 0) && 
             isRootNode) {
            long planningDoneTime = Measure.getTime();

            Measure.println("Path: Planning completed after " + 
                            (planningDoneTime - planningStartTime) + 
                            " ms", planningDoneTime);
         }
      } else if (m instanceof PathClose) {
         if (Debug.on) {
            Debug.println(Debug.PLANNING, Debug.INFO, 
                          "Path: closing the path (ID=" + connID + ")");
         }

         if (! isPlannerNode) {
            sendS(m);          // don't worry about success
         } else {
            m.recycle();
         }

         emptyMessageQueues();
         pathRequestOutstanding = false;  // don't keep trying

         setCurrentPlan(null);   // release the plan
         session.removePath();
         return true;          // path thread should exit
      }

      return false;
   }

   protected void emptyMessageQueues() {
      while (! pendingOutS.isEmpty()) {
         ((Recyclable) pendingOutS.dequeue()).recycle();
      }

      while (! pendingOutC.isEmpty()) {
         ((Recyclable) pendingOutC.dequeue()).recycle();
      }

      while (! incoming.isEmpty()) {
         try {
            ((Recyclable) incoming.getMessage()).recycle();
         } catch (InterruptedException ignore) {
         } catch (MessageFlowClosed ignore) {}
      }
   }

   /**
    * Establish a new current plan.  If a thread was waiting for a new plan
    * to arrive, it is notified.
    * @param plan The new current plan.
    * @return No return value.
    */
   // install a new plan on the path
   protected void setCurrentPlan(PlanDistro plan) {
      synchronized (planSync) {
         // let go of the old plan
         if (currentPlan!=null) {
            currentPlan.recycle();
         }

         // install this plan as the new one
         currentPlan = plan;
         if (currentPlan != null) {
            currentPlan.retain();

            // notify any waiting threads
            planSync.notifyAll();
         }
      }
   }

   /**
    * Wait for a plan with a particular version number to be established
    * as the current plan.  If the current plan has this version number,
    * that plan is returned immediately.  If not the current thread blocks
    * until a plan with that version number is delivered.
    * @param version The version number of the desired plan.
    * @return A PlanDistro object containing a plan with the requested version 
    *         number.
    */
   public PlanDistro getPlan(int version) {
      synchronized (planSync) {
         while ((currentPlan == null) || (currentPlan.version != version)) {
            try {
               planSync.wait();
            } catch (InterruptedException e) {}
         }

         // Prevent this message from being re-used until the caller is done
         currentPlan.retain();

         return currentPlan;
      }
   }

   public boolean isPlannerNode() {
      return isPlannerNode;
   }

   public boolean isRootNode() {
      return isRootNode;
   }

   // for debugging use only
   public String toString() {
      return "Path" + connID;
   }
}
