/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.Conductor.planning;

import java.net.InetAddress;
import java.net.Socket;
import java.util.Hashtable;
import java.io.IOException;

import FMG.Conductor.comm.*;
import FMG.Conductor.planning.comm.*;
import FMG.Conductor.dataflow.SessionTable;
import FMG.Conductor.util.Debug;

/**
 * The PlanCommMgr object is the glue for the planning comm objects and 
 * provides the interface between these objects and Path objects.  More
 * specifically, this object 
 * <ul>
 *    <li>Handles new connections accepted by the PathAcceptor object.
 *    <li>Provides an interface for finding new paths.
 *    <li>Maintains a list of currently open planning connections.
 *    <li>Provides accessors allowing Path objects and PlanComm objects
 *        to obtain handles for each other.
 * </ul>
 * <p>
 * There will normally be only one PlanCommMgr in the system at a time,
 * however, there is no reason why there could not be different PlanCommMgr
 * objects for distinct sets of comm objects.
 * <p>
 * TODO: Is it safe to store handles to comm objs given that there may
 *       be more than one connection between nodes?  Perhaps we should
 *       just fix the real problem here (see next TODO).
 * <p>
 * TODO: See the race condition outlined in the description of getCommObj().
 *
 * @author Mark Yarvis (yarvis@fmg.cs.ucla.edu)
 * @version $Revision: 1.27 $
 */

public class PlanCommMgr {

   /** The global table of active sessions */
   protected static SessionTable sessionTable;

   /** A list of active PlanComm objects, indexed by remote node id */
   protected Hashtable commObjs = new Hashtable();

   /** A PathFinder object to do the actual work of finding new paths. */
   protected PathFinder pathFinder = new PathFinder();

   /**
    * Create a new PlanCommMgr.  This also starts up a PathAcceptor and 
    * PlanCommServer to handle new path finding and path creation actions.
    * @see FMG.Conductor.planning.comm.PlanCommServer
    * @see FMG.Conductor.planning.comm.PathAcceptor
    */
   public PlanCommMgr(SessionTable sessionTable) {
      this.sessionTable = sessionTable;
      PlanCommServer.createAndStart(this);
      PathAcceptor.createAndStart(sessionTable);
   }

   /**
    * Handle a newly accepted planning connection.  A new PlanComm object
    * is created to manage this connection and this object's handle is
    * recorded for future use.
    * @param s The socket of a newly accepted planning connection.
    * @return No return value.
    */
   public void acceptIncoming(Socket s) {
      try {
         PlanComm co = new PlanComm(s, this, sessionTable);

         NodeID node = co.getRemoteNode();

         if (Debug.on) {
            Debug.println(Debug.PLANNING, Debug.INFO, 
                          "Accepted new path to " + node);
         }

         synchronized(commObjs) {
            // if it's already in the list, replace it
            PlanComm existing = (PlanComm) commObjs.get(node);
            if (existing != null) {
               existing.closeConnection();
               commObjs.remove(node);
            }

            commObjs.put(node, co);
         }
      } catch (IOException e) {
         // failed to create new PlanComm();
         if (Debug.on) {
            Debug.println(Debug.PLANNING, Debug.WARNING,
                          "WARNING: failure accepting planning connection");
         }
      }
   }

   /**
    * This method removes a CommObj from the list of active connections.
    * Normally this method is called by the CommObj itself when it is
    * shutting down.
    * @param node The id of the node the we no longer are connected with.
    * @return No return value.
    */
   public void removeCommObj(NodeID node, PlanComm co) {
      synchronized (commObjs) {
         if (commObjs.get(node) == co) {
            commObjs.remove(node);
         }
      }
   }

   /**
    * Obtain a handle to a comm object connected to the given node ID.
    * If no connection currently exists to that node, a new comm object
    * is created and connected to that node.
    * <p>
    * <b>Note:</b> There is a race condition here.  It may be that the 
    * remote node connects with us after we determine that there is
    * not currently a connection to that node, but before he accepts
    * our new connection.  We will then end up with two connections between
    * the same pair of nodes.
    * <p>
    * TODO: we need to fix the above problem.  One of the two connections
    * should shut down.  We must make sure we return a good handle to
    * the user, though.
    * @param node The id of the node with which we want to communicate.
    * @return A PlanComm object connected to the requested node.
    */
   PlanComm getCommObj(NodeID node, NodeID address) {
      synchronized (commObjs) {
         PlanComm co = (PlanComm) commObjs.get(node);

         if (co == null) {
            try {
               if (Debug.on) {
                  Debug.println(Debug.PLANNING, Debug.INFO, 
                                "Creating new path to " + node);
               }

               co = new PlanComm(node, address, this, sessionTable);
               commObjs.put(node, co);
            } catch (IOException e) {
               if (Debug.on) {
                  Debug.println(Debug.PLANNING, Debug.INFO, "Failed (" + e +
                                ") to create planning connection to " + node);
               }
               return null;
            }
         } else if (! co.verifyInterfaceAddress(address)) {
            // the interface address changed ... start over
            if (Debug.on) {
               Debug.println(Debug.PLANNING, Debug.INFO, 
                             "Interface changed: address for node " + node + 
                             " is now " + address);
            }

            co.closeConnection();
            commObjs.remove(node);
            return getCommObj(node, address);
         }

         return co;
      }
   }

   /**
    * Obtain a handle for a PlanComm object that is connected to a particular
    * node.  If no such object exists, return null.
    * @param node The id of the node with which we want to communicate.
    * @return A PlanComm object connected to the requested node, or null if no
    *         such object exists.
    */
   synchronized PlanComm lookupCommObj(NodeID node) {
      return (PlanComm) commObjs.get(node);
   }

   /**
    * Probe for the creation of a path connecting the source node to the
    * destination node.  The PathFinder object is used to send a request
    * toward the destination (target) node.  This request will presumably
    * be picked up by a Conductor-enabled node and cause a connection
    * to be created with this node.
    * <p>
    * If an error occurs sending the first packet, try sending another one.
    * It is likely that the error was not with this send, but with a previous
    * send (once the ICMP message reaches us).
    * @param connID The id of the connection we are trying to form a path for.
    * @param sourceAddr The source address of the connection.
    * @param sourcePort The source port number of the connection.
    * @param targetAddr The target (destination) address of the connection.
    * @param targetPort The target port number of the connection.
    * @return No return value.
    * @see FMG.Conductor.planning.comm.PathFinder
    */
   public void seekPath(ConnID connID, NodeID sourceAddr, int sourcePort,
                        NodeID targetAddr, int targetPort) {
      // Try twice.  The first error may be from a previous request.
      for (int i = 0; i < 2; i++) {
         try {
            pathFinder.sendPathRequest(connID, sourceAddr, sourcePort, 
                                               targetAddr, targetPort);
            // success, we're done
            break;
         } catch (IOException e) {
            if (Debug.on) {
               Debug.println(Debug.PLANNING, Debug.WARNING, 
                             "Exception sending path request: " + e);
            }
         }
      }
   }
}
