/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.Conductor.planning;

import java.lang.reflect.*;
import java.util.*;

import FMG.Conductor.util.Debug;
import FMG.Conductor.util.Direction;
import FMG.Conductor.Conductor;
import FMG.Conductor.planning.messages.*;
import FMG.Conductor.comm.*;

/**
 * The abstract Planner class is a template and support structure for
 * implementing plug-in planner objects in Conductor.  A planner class
 * is expected to be a subclass of Planner and to implement at least the 
 * following methods:
 * <ul>
 * <li>The <code>addInfo()</code> method should accept PlanningInfo objects 
 *     (or subclasses) describing the conditions at individual nodes along 
 *     the data path.
 * <li>The <code>formulatePlan()</code> method should generate a plan after
 *     all planning information has been received.
 * <li>A constructor should be provided that takes five arguments: a 
 *     connection id, a source address, a source port, a destination address, 
 *     and a destination port.
 * </ul>
 * Note that a given planner may be asked to participate in several rounds
 * of re-planning for a given connection.
 * <p>
 * The actual planning class used by the system is determined by the property
 * <i>PlannerClass</i> in the Conductor properties file.  This property 
 * should contain a fully qualified class name.  The default planner
 * class is StaticPlanner.  A new planner, of the
 * type specified, can be generated for a particular connection by
 * calling <code>createPlanner()</code>.
 *
 * @author Mark Yarvis (yarvis@fmg.cs.ucla.edu)
 * @version $Revision: 1.19 $
 *
 * @see FMG.Conductor.planning.StaticPlanner
 */

public abstract class Planner {

   /** The id of the connection that this planner is planning for */
   protected ConnID connID = new ConnID();

   /** The constructor to use when creating new planners */
   private static Constructor plannerConstructor;

   /**
    * The initilization of this class determines which class has been
    * selected for use and records the appropriate constructor.
    */
   static {
      String plannerName = Conductor.getProperty("PlannerClass");
      if (plannerName == null) {
         if (Debug.on) {
            Debug.println(Debug.PLANNING, Debug.WARNING, 
                          "Using default planner");
         }
         setPlannerClass(StaticPlanner.class);
      } else {
         try {
            setPlannerClass(Class.forName(plannerName)); 
         } catch (Exception e) {
            System.out.println("Failure loading planner");
            e.printStackTrace();
            System.exit(1);
         }
      }
   }

   /**
    * Create a new planner for a given connection.
    * @param connID The ID of the connection this planner is associated with.
    */
   protected Planner(ConnID connID) {
      this.connID.copy(connID);
   }

   /**
    * Create a new planner of the class that was determined at startup.
    * The class-specific constructor will be called here.
    * @param connID The id of the connection that needs a new planner.
    * @return A new planner object for this connection.
    */
   public static Planner createPlanner(ConnID connID, 
                                       NodeID sourceAddr, int sourcePort, 
                                       NodeID destAddr, int destPort) {
      if (plannerConstructor == null) {
         return new StaticPlanner(connID, sourceAddr, sourcePort, 
                                          destAddr, destPort);
      } else {
         try {
            Object[] args = new Object[5];
            args[0] = connID;
            args[1] = sourceAddr;
            args[2] = new Integer(sourcePort);
            args[3] = destAddr;
            args[4] = new Integer(destPort);

            return (Planner) plannerConstructor.newInstance(args);
         } catch (Exception e) {
            throw new RuntimeException("Could not instantiate planner: " + e);
         }
      }
   }

   /**
    * Set the class to use when creating planners.  This method is only
    * called by the static initializer once it has determined the class
    * that the user has selected.  From that, we can determine the
    * constructor that should be used to create such a planner and
    * record it for future use.
    * @param plannerClass An object representing the class we want 
    *                     to use for planning.
    * @return No return value.
    */
   private static void setPlannerClass(Class plannerClass) {
      try {
         Class[] args = new Class[5];
         args[0] = ConnID.class;
         args[1] = NodeID.class;
         args[2] = int.class;
         args[3] = NodeID.class;
         args[4] = int.class;

         plannerConstructor = plannerClass.getConstructor(args);
      } catch (Exception e) {
         throw new RuntimeException("Could not get constructor for planner: " 
                                           + e);
      }
   }

   /**
    * This method is called by the Path object when it receives information
    * about a node during the information gathering process.  This method
    * will be called once for each node along the path.  If a custom 
    * environmental monitor is being used, the passed object may be a
    * subclass of Planning Info.  This allows the construction of planners
    * that work together with particular environment monitors.
    * @param pi The planning information for one node along the path.
    * @see FMG.Conductor.monitor.EnvMonitor
    */
   public abstract void addInfo(PlanningInfo pi);

   /**
    * This method is called by the Path object once all planning
    * information has been received.  The planner should generate and
    * return a plan based on the preceeding calls to <code>addInfo()</code>.
    * @return A PlanDistro object containing our new plan.
    */
   public abstract PlanDistro formulatePlan();

   /**
    * Take an unsorted list of planning information and sort it so that the
    * client (root) node is first and the server (planner) node is last.
    * This method assumes that all PlanningInfo messages are present in
    * the list (otherwise, the results will be unpredictable).  The 
    * <code>isRoot</code> member is used to determine the first node.  
    * The <code>nextNode</code> member is used to order the remaining nodes.
    * <p>
    * Warning: this method assumes that the path in each direction is the
    * same.
    * @param infoList A vector of PlanningInfo objects.
    * @return Returns true on success, false if sorting failed because some
    *         exected nodes were not present.
    */
   protected boolean sortPlanningInfo(Vector infoList) {
      PlanningInfo targetElem = null;

      // find the root node
      for (int i=0; i<infoList.size(); i++) {
         PlanningInfo elem = (PlanningInfo) infoList.elementAt(i);
         if (elem.isRoot) {
            if (targetElem != null) { // more than one root found!
               return false;
            }
            targetElem = elem;
         }
      }

      if (targetElem == null) {
         return false;
      }

      // put the root node at the end
      infoList.removeElement(targetElem);
      infoList.addElement(targetElem);

      int nodesHandled = 1;

      // when there is no next node, we're done
      while (! targetElem.nextNode.isNull()) {
         NodeID nextNode = targetElem.nextNode;
         targetElem = null;

         // look for the next node
         for (int i=0; i < (infoList.size() - nodesHandled); i++) {
            PlanningInfo elem = (PlanningInfo) infoList.elementAt(i);
            if (elem.node.equals(nextNode)) {
               targetElem = elem;
               break;
            }
         }

         if (targetElem == null) {       // wasn't found
            return false;
         }

         // append to the end of the list
         infoList.removeElement(targetElem);
         infoList.addElement(targetElem);

         nodesHandled++;
      }

      // if the root node isn't first, then there were some extra nodes
      if (((PlanningInfo) infoList.firstElement()).isRoot == false) {
         return false;
      }

      return true;
   }

   /**
    * Take a complete plan and account for the <code>interfaceToNext</code>
    * field in <code>PlanningInfo</code> messages.  This method helps planners
    * take into account the fact that some nodes have interfacs that do not
    * correspond to their node id.
    * @param plan The completed plan.
    * @param infoList A vector of <code>PlanningInfo</code> message used 
    *                 to generate the plan.
    * @return No return value.
    */
   protected void setUpstreamAddresses(PlanDistro plan, Vector infoList) {
      Enumeration e = infoList.elements();
      while (e.hasMoreElements()) {
         PlanningInfo pi = (PlanningInfo) e.nextElement();
         // setting is somewhat expensive, so make sure we want to do it
         if (! pi.interfaceToNext.isNull() && 
             ! pi.interfaceToNext.equals(pi.node)) {
            plan.setUpstreamAddress(pi.node, pi.interfaceToNext, 
                                                Direction.down);
         }
         if (! pi.interfaceToPrev.isNull() && 
             ! pi.interfaceToPrev.equals(pi.node)) {
            plan.setUpstreamAddress(pi.node, pi.interfaceToPrev, 
                                                Direction.up);
         }
      }
   }

}
