/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.Conductor.planning;

import java.util.*;
import java.io.*;

import FMG.Conductor.planning.messages.*;
import FMG.Conductor.comm.*;
import FMG.Conductor.util.*;
import FMG.Conductor.Conductor;

/**
 * The StaticPlanner is a simple example of a planner that is very useful
 * for debugging and demonstration purposes.  The plan, in this case, is a
 * generated from a pre-selected set of adaptors called a plan configuration.
 * The plan configuration specifies which adaptors to put on each node, but
 * not what specific nodes to put them on.  A plan is generated by matching
 * the first node in the connection with the first node in the configuration,
 * and so on.
 * <p>
 * The plan configuration can be specified in several ways:
 * 1. The attribute <code>UserPrefs.StaticPlanner.configFile</code> in the 
 *    <code>PlanningInfo</code> message from the client can specify the name 
 *    of a file (on the planner node) that contains the configuration.
 * 2. The <code>PlanningInfo</code> message from the client can contain the 
 *    configuration.
 * 3. The <code>StaticPlanner.ConfigFile.N</code> Conductor property can 
 *    specify the name of a file to read the configurations from for a
 *    particular protocol (in this case, the protocol used when the 
 *    destination port number is N).
 * 4. A default configuration is contained in a file on the server.  The name 
 *    of this file can optionally be specified by the 
 *    <code>StaticPlanner.ConfigFile</code> Conductor property.  The default
 *    file name is "./adaptors.cfg".
 * For each connection, the StaticPlanner selects the first configuration
 * available from the four sources by checking each, in the order indicated
 * above.
 * <p>
 * Options 1, 3, and 4 above allow a configuration to be read from a file.
 * The file consists of two sections.  Lines in the first section begin with 
 * a plus ('+') and indicate the adaptors to deply in the upstream direction.  
 * Each line contains a space-delimeted list of fully-qualified adaptor class 
 * names.  Lines in the second section begin with a minus ('-') and indicate 
 * the adpators to deploy in the downstream direction.  The format is the 
 * same as the first section.  There <b>MUST</b> be the same number of lines 
 * in the first and second sections.
 * <p>
 * Option 2 above allows a configuration to be encoded in a 
 * <code>PlanningInfo</code> message.  The <code>PlanningInfo</code> message
 * can carry any number of atribute-value pairs.  The adaptors to be deployed
 * for a particular direction of flow on a particular node are stored as
 * a value for a single attributed.  The attribute name is 
 * <code>UserPrefs.StaticPlanner.upConfig.N</code> or 
 * <code>UserPrefs.StaticPlanner.downConfig.N</code>, where N is the node
 * number (starting at 0).  An empty value should be used for any node
 * that does not require adpators.
 * <p>
 * Each time planning occurs, the StaticPlanner instance keeps track of the
 * nodes that are involved in the connection as indicated by calls to 
 * <code>addInfo()</code>.  Adaptors listed in the configuration 
 * are assigned positionally to the nodes identified in the connection.
 * If there are more entries in the configuration than nodes,
 * some adpators will not be deployed.  If there are more nodes than entries
 * in the configuration, an error will result.  The resulting plan is returned
 * in the call to <code>formulatePlan()</code>.
 * <p>
 * <b>Warning:</b> Since the StaticPlanner is intended for debuggin and
 * demo purposes, it is not very robust or forgiving, nor is it very
 * user-friendly.  Use it at your own risk.
 *
 * @author Mark Yarvis (yarvis@fmg.cs.ucla.edu)
 * @version $Revision: 1.20 $
 */

public class StaticPlanner extends Planner {

   /**
    * This class is a container for a planning configuration that has been
    * read out of a configuration file.  Two vectors are provided, one for
    * each direction of flow.  Each vector element is an array of strings
    * specifying the adaptors to execute on a given node.
    */
   private class PlanningConfig {
      Vector upConfig = new Vector();
      Vector downConfig = new Vector();
   }

   /** The name of the configuration file for this class. */
   private static final String DEFAULT_CONFIG_FILE = 
                    Conductor.getFilenameProperty("StaticPlanner.ConfigFile", 
                                                  "./config/adaptors.cfg");

   /** 
    * The prefix for the property name for obtaining protocol-specific 
    * plan file names.
    */
   private static final String PROT_SPECIFIC_CONFIG_FILE_PREFIX = 
                                           "StaticPlanner.ConfigFile.";

   /** 
    * Indicates whether or not all nodes should be included in the plan.  
    * If false, any node that doesn't have adaptors in either direction 
    * will not be included in the plan.  The default value is false.
    */
   private static final boolean INCLUDE_ALL_NODES =
          Conductor.getBooleanProperty("StaticPlanner.IncludeAllNodes", false);

   /**
    * The tag that can appear in PlanningInfo messages that indexes the name
    * of the configuration file to use for planning.
    */
   protected static final String CONFIG_FILE_INFO_TAG = 
                                  "UserPrefs.StaticPlanner.configFile";

   /**
    * This prefix is used to index individual pieces of a plan configuration 
    * that can be included directly in a PlanningInfo message.  Normally,
    * an integrer indicating the node number for this part of the 
    * configuration is appended to this string.  This prefix is used for
    * the upstream configuration.
    */
   protected static final String INLINE_CONFIG_UP_PREFIX =
                                  "UserPrefs.StaticPlanner.upConfig.";

   /**
    * This prefix is used to index individual pieces of a plan configuration 
    * that can be included directly in a PlanningInfo message.  Normally,
    * an integrer indicating the node number for this part of the 
    * configuration is appended to this string.  This prefix is used for
    * the downstream configuration.
    */
   protected static final String INLINE_CONFIG_DOWN_PREFIX =
                                  "UserPrefs.StaticPlanner.downConfig.";

   /**
    * This index is used to determine whether a plan configuration is
    * contained directly in this PlanningInfo message.
    */
   protected static final String INLINE_CONFIG_PRESENCE_INFO_TAG =
                                          INLINE_CONFIG_UP_PREFIX + "0";

   /** The information received so far for the current planning process. */
   protected Vector infoList = new Vector();

   /** The config to use for this instance of the planner. */
   protected PlanningConfig selectedConfig;

   /** The destination port number for this connection. */
   protected int destPort;

   /** The defalt config to use if no specific config was specified */
   static PlanningConfig defaultConfig;

   /** 
    * An table, indexed by port number, of previously determined 
    * configuration file names for particular ports.
    */
   static SparseTable configByDestPort = new SparseTable();

   /** A table, indexed by file name, of previously read configurations. */
   static Hashtable configByFileName = new Hashtable();

   /**
    * Create a StaticPlanner.  Obtain the configuration for this object.
    */
   public StaticPlanner(ConnID connID, NodeID sourceAddr, int sourcePort, 
                                       NodeID destAddr, int destPort) {
      super(connID);

      this.destPort = destPort;
   }

   /**
    * Read a list of adaptors from a line in the configuration file and
    * digest it into a set of strings, one for each adaptor class name.  
    * Note that the leading plus or minus has already been read.  This 
    * method creates an array of the tokenized adaptor class names and
    * then adds that array to the vector (which is really either upConfig
    * or downConfig).
    * @param config A reader already opened on the config file and positioned
    *               just after the leading plus or minus.
    * @param list A vector to fill in with the tokenized strings from the line.
    * @return No return value.
    * @exception java.io.IOException if an error occurs while reading the
    *                                config file.
    */
   private static void readAdaptorsForNode(BufferedReader config, Vector list) 
                                                        throws IOException {
      list.addElement(StringOps.tokenizeString(config.readLine()));
   }

   /**
    * Read the config file and populate upConfig and downConfig with
    * arrays of classes to install on individual nodes.
    * @return The configuration read from the file.
    */
   private PlanningConfig readConfigFile(String fileName) throws IOException {
      PlanningConfig config = new PlanningConfig();

      BufferedReader in = new BufferedReader(new FileReader(fileName));
      int numNodes = 0;

      while (in.read() == '+') {
         numNodes++;
         readAdaptorsForNode(in, config.upConfig);
      }

      for (int i = 0; i<numNodes; i++) {
         if (i > 0) {
            in.read();
         }
         readAdaptorsForNode(in, config.downConfig);
      }

      in.close();

      return config;
   }

   /**
    * Obtain the default configuration.  If that configuration has yet to
    * be read from a file, read it.
    * @return a PlanningConfig object representing the default configuration.
    */
   private PlanningConfig getDefaultConfig() {
      if (defaultConfig == null) {
         try {
            defaultConfig = readConfigFile(DEFAULT_CONFIG_FILE);
         } catch (IOException e) {
            if (Debug.on) {
               Debug.println(Debug.WARNING,
                             "StaticPlanner was unable to read it's default " +
                               "configuration file (" + DEFAULT_CONFIG_FILE + 
                               "): " + e);
            }
            // use the null plan instead
            defaultConfig = new PlanningConfig();
         }
      }

      return defaultConfig;
   }

   /**
    * Obtain a configuration embedded in a PlanningInfo message.
    * @param pi The message to obtain the configuration from.
    * @return A configuration, if one was present in the message.  null
    *         otherwise.
    */
   protected PlanningConfig getConfigFromPlanningInfo(PlanningInfo pi) {
      if (pi.info.get(INLINE_CONFIG_PRESENCE_INFO_TAG) != null) {
         PlanningConfig config = new PlanningConfig();
         String adaptorList;

         for (int i = 0; true; i++) {
            adaptorList = (String) pi.info.get(INLINE_CONFIG_UP_PREFIX + i);
            if (adaptorList == null) {
               break;
            }
            config.upConfig.addElement(StringOps.tokenizeString(adaptorList));
         }

         for (int i = 0; true; i++) {
            adaptorList = (String) pi.info.get(INLINE_CONFIG_DOWN_PREFIX + i);
            if (adaptorList == null) {
               break;
            }
            config.downConfig.addElement(StringOps.tokenizeString(adaptorList));
         }

         return config;
      } else {
         return null;
      }
   }

   /**
    * Read a configuration from a configuration file.  Configurations are
    * cached by file name.  If the given configuration file has been read
    * before, the same configuration object is returned.  Otherwise, the file
    * is read and a configuration object is created, cached, and returned.
    * @param filename The name of the file to read.
    * @return A configuration object obtained from the file.
    * @exception java.io.IOException if an error occurs while reading the
    *                                config file.
    */
   protected PlanningConfig getConfigFromFile(String filename) 
                                                      throws IOException {
      // look up in hashtable first
      PlanningConfig config = (PlanningConfig) configByFileName.get(filename);

      // not cached, read it from the file
      if (config == null) {
         config = readConfigFile(filename);

         // record back in hash table for future requests
         configByFileName.put(filename, config);
      }

      return config;
   }

   /**
    * Get a configuration for a particular connection.  The right configuration
    * is based on the PlanningInfo message from the client node and the
    * destination port number.  The configuration is obtained from
    * (1) a file specified by the PlanningInfo message,
    * (2) the PlanningInfo message itself,
    * (3) a file specified for this protocol (destination port), or
    * (4) the default configuration (in that order of preference).
    * @param pi The PlanningInfo message obtained from the client node.
    * @param destPort The destination port number for this connection.
    * @return A PlanningConfig object appropriate for this connection.
    */
   protected PlanningConfig getConfig(PlanningInfo pi, int destPort) {
      String fileName = null;

      try {
         // Try to get config file name from planning info
         fileName = (String) pi.info.get(CONFIG_FILE_INFO_TAG);

         if (fileName != null) {
            // substitute port number of ${PORT} appers in filename
            fileName = StringOps.replace(fileName, "${PORT}", "" + destPort);
            return getConfigFromFile(fileName);
         }

         // Look for previously determined port-specific file name
         PlanningConfig config = 
                    (PlanningConfig) configByDestPort.get(destPort);

         if (config != null) {
            return config;
         }

         // look for port number based property
         fileName = Conductor.getFilenameProperty(
                            PROT_SPECIFIC_CONFIG_FILE_PREFIX + destPort, null);
         if (fileName != null) {
            config = getConfigFromFile(fileName);
         } else {
            config = getDefaultConfig();
         }

         configByDestPort.put(destPort, config);

         return config;
      } catch (IOException e) {
         System.err.println("Warning: StaticPlanner failed reading '" + 
                            fileName + "'. Using default config.");
         return getDefaultConfig();
      }
   }

   /**
    * This method is called upon receipt of each PlanningInfo message.
    * The messages are stored for later sorting and use in the planning
    * process.  When the PlanningInfo message from the client node is
    * obtained, the planning configuration for this connection is determined.
    * @param pi The planning information for one of the nodes along the path.
    * @return No return value.
    */
   public void addInfo(PlanningInfo pi) {
      // hold on to this message for later processing
      infoList.addElement(pi);
      pi.retain();   // don't let this message be recycled

      /*
       * If this message is from the root node, figure out what configuration
       * to use.  By doing this now, we may read a config file, if necessary,
       * in parallel with other messages being transmitted.
       */
      if (pi.isRoot) {
         // look for the config spec in the user prefs
         selectedConfig = getConfigFromPlanningInfo(pi);

         if (selectedConfig == null) {
            // config is not in the planning info, get it from a local file
            selectedConfig = getConfig(pi, destPort);
         }
      }
   }

   /**
    * Forumlate a plan based on the planning information received.  The
    * PlanningInfo messages cached during previous calls to 
    * <code>addInfo()</code> are sorted and then applied to the previously 
    * determined planning configuration.
    * @return The plan we formulated.
    */
   public PlanDistro formulatePlan() {
      /*
       * TODO: once we can handle different nodes in each direction 
       * (reliably), we really should consider each direction 
       * individuallly
       */

      // Create a new plan distro message
      PlanDistro plan = PlanDistro.getNewMessage(connID);

      // Sort the list of messages received
      if ((selectedConfig == null) || (sortPlanningInfo(infoList) == false)) {
         throw new RuntimeException("This shouldn't happen");
      }

      for (int i=0; i<infoList.size(); i++) {
         PlanningInfo pi = (PlanningInfo) infoList.elementAt(i);

         String[] upAdaptors = null;
         if (selectedConfig.upConfig.size() > i) {
            upAdaptors = (String[]) selectedConfig.upConfig.elementAt(i);
         }
         String[] downAdaptors = null;
         if (selectedConfig.downConfig.size() > i) {
            downAdaptors = (String[]) selectedConfig.downConfig.elementAt(i);
         }

         //
         // Other than first and last node, don't include nodes that don't have
         // adaptors in either direction
         if (INCLUDE_ALL_NODES ||
             ((upAdaptors != null) && (upAdaptors.length != 0)) || 
             ((downAdaptors != null) && (downAdaptors.length != 0)) || 
             (pi.nextNode.isNull()) || (pi.prevNode.isNull())) {
            // add info into plan
            plan.addNode(pi.node, pi.interfaceToPrev, upAdaptors, 
                                                     Direction.up);
            plan.insertNode(pi.node, pi.interfaceToNext, downAdaptors, 
                                                     Direction.down);
         }

         pi.recycle();
      }

      infoList.removeAllElements();

      return plan;
   }
}
