/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.Conductor.planning.comm;

import java.net.*;
import java.io.*;

import FMG.Conductor.planning.messages.PathRequest;
import FMG.Conductor.planning.Path;
import FMG.Conductor.Conductor;
import FMG.Conductor.util.Debug;
import FMG.Conductor.dataflow.Session;
import FMG.Conductor.dataflow.SessionTable;
import FMG.Conductor.comm.NodeID;

/**
 * The PathAcceptor class implements a server thread that intercepts
 * or receives PathRequest messages arriving in UDP packets.  Such a
 * packet may be recieved at the destination address or intercepted
 * by a conductor-enabled node between the sender and the final destination.
 * In either case, the message is delivered to a local path object
 * (which is created if it does not already exist).  The local path
 * object should create a planning connection back to the sender of
 * this message and may forward the PathRequest toward the destination (if
 * this isn't it).
 * <p>
 * The UDP packets we're receiving should have been sent by a PathRequest
 * object.
 * <p>
 * The port number used by the PathAcceptor is configurable using the
 * <i>AcceptorPort</i> property in the Conductor properties file.
 *
 * @author Mark Yarvis (yarvis@fmg.cs.ucla.edu)
 * @version $Revision: 1.21 $
 *
 * @see FMG.Conductor.planning.Path
 * @see FMG.Conductor.planning.messages.PathRequest
 * @see FMG.Conductor.planning.comm.PathFinder
 */

public class PathAcceptor extends Thread {

   /** The port number on which to expect incoming messages */
   public static final int ACCEPTOR_PORT = 
                             Conductor.getIntProperty("AcceptorPort", 4444);


   /** We assume the packet data will be no bigger than this. */
   private static final int BUFSIZE = 1000;

   /** The global table of session objects */
   protected SessionTable sessionTable;

   /** The socket with which we receive incoming messages */
   protected DatagramSocket in;

   /** A reusable buffer to hold the contents of each packet */
   private byte[] buf = new byte[BUFSIZE];

   /**
    * Create a new PathAcceptor.  Normally there will be only one of these
    * since there can only be one socket bound to the port.  This method
    * should normally only be called from <code>createAndStart()</code> so
    * we can ensure that the thread is started up.
    * @param sessionTable The handle of the global session table.
    */
   protected PathAcceptor(SessionTable sessionTable) {
      this.sessionTable = sessionTable;
      if (Debug.on) {
         setName("PathAcceptor");
      }

      try {
         in = new DatagramSocket(ACCEPTOR_PORT);
      } catch (SocketException e) {
         throw new RuntimeException("Couldn't create path finder socket " + e);
      }
   }

   /**
    * Create a PathAcceptor and start the server thread to handle incoming
    * packets.  Normally, this should only be called once in the program
    * because there is only one socket to be managed.
    * @param sessionTable The handle of the global session table.
    * @return No return value.
    */
   public static void createAndStart(SessionTable sessionTable) {
      PathAcceptor pa = new PathAcceptor(sessionTable);
      pa.setPriority(MAX_PRIORITY - 1);
      pa.start();
   }

   /**
    * Get the next message sent to the socket.  A PathRequest will be
    * decoded from the next UDP packet received and returned.
    * @return A PathRequest object that was created from data received
    *         in a UDP packet.
    * @exception java.io.IOException if an error occurs while receiving
    *                                or decoding the packet.
    */
   protected PathRequest getNextRequest() throws IOException {
      //TODO: see if we can avoid creating a new packet every time
      DatagramPacket pack = new DatagramPacket(buf,buf.length);
      in.receive(pack);

      DataInputStream is = new DataInputStream(
                                   new ByteArrayInputStream(pack.getData(), 0, 
                                                            pack.getLength()));

      PathRequest pr = (PathRequest) PathRequest.getNewMessage(is);

      /*
       * Remember where we got this from ... it's our neighbor's interface
       * for talking with us.
       */
      pr.setSourceInterface(pack.getAddress());

      return pr;
   }

   /**
    * This is the main code for the server thread.  For each PathRequest
    * recived, an associated Path object is either found or created.
    * The message is then delivered to that object.  Rinse, later, repeat.
    * @return No return value.  Unless there is an error, this method will
    *         not return.
    */
   public void run() {
      PathRequest pr;
      Session session;

      while (true) {
         try {
            pr = getNextRequest();

            session = sessionTable.getSession(pr.connID);
            if (session.getPath() == null) {
               new Path(session, pr.sourceAddr, pr.sourcePort, 
                                 pr.targetAddr, pr.targetPort, null);
            }

            // session.path will be filled in during path construction
            session.getPath().handleMessage(pr);

         } catch (IOException e) {
            System.err.println("Error reading from finder socket:" + e);
         }
      }
   }
}
