/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.Conductor.planning.comm;

import java.io.*;
import java.net.*;

import FMG.Conductor.planning.messages.PathRequest;
import FMG.Conductor.comm.*;

/**
 * The PathFinder class is the agent that looks for Conductor-enabled
 * nodes along the way to a designated endpoint.  For each call to
 * sendPathRequest(), a PathRequest message is sent in a UDP packet 
 * toward a target node.  It is assumed that any Conductor-enabled nodes 
 * between us and the target will intercept the packet, forward it along 
 * to the server, and create a planning connection back to this node.
 * The interception capability is implemented by the PathAcceptor 
 * object/thread.
 * <p>
 * Note: An alternative design would be to simply create a new planning
 * connection to the endpoint and expect any Conductor-enabled
 * nodes along the path to capture the connection through a transparent
 * proxy capability.  However, this would create a new connection for
 * each planning path.  By sending a UDP packet, we can discover our
 * neighbor without creating a TCP connection.  The neighbor can then
 * determine whether to use an existing TCP connection or to create a new
 * one.
 *
 * @author Mark Yarvis (yarvis@fmg.cs.ucla.edu)
 * @version $Revision: 1.10 $
 *
 * @see FMG.Conductor.planning.comm.PathAcceptor
 * @see FMG.Conductor.planning.messages.PathRequest
 */

public class PathFinder {

   /** The socket used for sending the UDP packets */
   DatagramSocket out;

   /** A cached copy of our local address */
   NodeID localAddress = NodeID.getNewLocalID();

   /**
    * Create a new PathFinder object.  Normally, only one of these will
    * be needed in the system.
    */
   public PathFinder() {

      /* set up outgoing socket */
      try {
         out = new DatagramSocket();
      } catch (SocketException e) {
          throw new RuntimeException("Could not set up socket: " + e);
      }
   }

   /**
    * Package a PathRequest into a DatagramPacket.  The message is 
    * "serialized" by asking it to write itself to a ByteArrayOutputStream,
    * the result of which is placed in the packet.  A new DatagramPacket is 
    * created for each message to avoid a Java bug.
    * @param pr The PathRequest message that we are trying to package into
    *           a datagram packet.
    * @return A new DatagramPacket object containing a "serialized" form of
    *         the given PathRequest message.
    */
   protected DatagramPacket makePacket(PathRequest pr) {
      ByteArrayOutputStream baos = new ByteArrayOutputStream();
      ObjectOutput oo;
      byte[] buf;

      try {
         DataOutputStream os = new DataOutputStream(baos);
         pr.sendSelf(os);
         os.flush();
      } catch (IOException e) {
         System.err.println("Error writing packetizing PathRequest: " + e);
      }
      buf = baos.toByteArray();

      /* put buf in a packet and return */
      return new DatagramPacket(buf, buf.length, 
                                pr.targetAddr.getAddress(), 
                                PathAcceptor.ACCEPTOR_PORT);
   }

   /**
    * Send a "path request" toward the target node.  The request will
    * contain the key details of the connection: the id, the source and
    * source port, and the destination (target) and destination port.
    * A UDP packet will be generated, containing the request, and it
    * will be sent toward the target node.  It is expected that if any
    * Conductor-enabled nodes are between the local node and the destination,
    * this packet will be intercepted and eventually forwarded.
    * @param connID The id of the connection we are trying to form a path for.
    * @param sourceAddr The source address of the connection.
    * @param sourcePort The source port number of the connection.
    * @param targetAddr The target (destination) address of the connection.
    * @param targetPort The target port number of the connection.
    * @return No return value.
    * @exception java.io.IOException if the send() call fails.
    * @see java.net.DatagramSocket#send
    */
   public synchronized void sendPathRequest(ConnID connID, 
                                     NodeID sourceAddr, int sourcePort,
                                     NodeID targetAddr, int targetPort) 
                                               throws IOException {
      PathRequest pr = (PathRequest) PathRequest.getNewMessage(connID, 
                                       sourceAddr, sourcePort, 
                                       targetAddr, targetPort, localAddress);
      out.send(makePacket(pr));
      pr.recycle();
   }
}
