/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.Conductor.planning.comm;

import java.io.*;
import java.net.*;

import FMG.Conductor.Conductor;
import FMG.Conductor.util.*;
import FMG.Conductor.comm.*;
import FMG.Conductor.net.SocketOps;
import FMG.Conductor.planning.messages.*;
import FMG.Conductor.planning.*;
import FMG.Conductor.dataflow.SessionTable;
import FMG.Conductor.dataflow.Session;

/**
 * Each PlanComm object handles the reading and writing of messages along
 * a given planning connection.  Recall that planning connections
 * carry planning messages between adjacent nodes along planning paths. 
 * A given planning connection handles planning messages for all planning
 * paths that include the link between these two nodes.
 * <p>
 * A PlanComm object may be created to build and manage a new planning
 * connections or to handle a newly accepted planning connection.  Upon 
 * creation of a PlanComm object, a dynamically allocated thread is created
 * to handle the input stream.
 * <p>
 * Messages that are read are presumed to be of type PlanningMessage.  They
 * can, therefore, be routed to the appropriate Path object using the
 * connection id member of each PlanningMessage object.
 * <p>
 * Methods are also provided for sending planning messages to the adjacent 
 * node.
 *
 * @author Mark Yarvis (yarvis@fmg.cs.ucla.edu)
 * @version $Revision: 1.30 $
 *
 * @see FMG.Conductor.planning.Path
 * @see FMG.Conductor.planning.messages.PlanningMessage
 */

public class PlanComm extends Job {

   /** 
    * The amount of time (in seconds) after which the connection will be 
    * dropped if it hasn't been active.  This value is controlled by the 
    * Conductor property <code>PlanCommInactivityTimeout</code>.  The default 
    * is 0 (never).
    */
   protected static final int INACTIVITY_TIMEOUT = 
               Conductor.getIntProperty("PlanCommInactivityTimeout", 0);

   /**
    * The amount of time after which the connection will be dropped if
    * it hasn't been active (in msec).
    */
   private static final int MSEC_INACTIVITY_TIMEOUT = 
                                         INACTIVITY_TIMEOUT * 1000;

   private static final int PLAN_COMM_PRIORITY = 
               Conductor.getIntProperty("PlanCommPriority", -1);

   private static final boolean PLAN_COMM_LOWDELAY = 
               Conductor.getBooleanProperty("PlanCommLowDelay", false);

   /** The ID of the node on the remote end of this connection. */
   protected NodeID node = new NodeID();

   /** The address of the node we're connected to (may be != node) */
   protected NodeID address = new NodeID();

   /** Our local interface address for this connection. */
   protected NodeID localInterface = new NodeID();

   /** The socket used to communicate with the remote endpoint. */
   protected Socket s;

   /** The output stream for sending messages */
   protected PlanningOutputStream output;

   /** The input stream for receiving messages */
   protected PlanningInputStream input;

   /** The object which keeps track of currently active connections */
   protected PlanCommMgr commMgr;

   /** The object which keeps track of session objects */
   protected SessionTable sessionTable;

   private static NodeID localhost = NodeID.getNewLocalID();

   /**
    * Create a new PlanComm object and a new connection to an adjacent
    * Conductor node.  The thread is also obtained from the ThreadRecycler
    * to handle input from the new socket.
    * @param node The id of the node to connect to.
    * @param address The address of the node to connect to.
    * @param commMgr The object keeping track of open connections
    * @param sessionTable The list of all session objects.
    * @exception java.io.IOException if we fail to connect to the remote node.
    * @see FMG.Conductor.util.ThreadRecycler
    */
   public PlanComm(NodeID node, NodeID address, PlanCommMgr commMgr, 
                   SessionTable sessionTable) throws IOException {
      this.node.copy(node);
      this.commMgr = commMgr;
      this.sessionTable = sessionTable;
      this.address.copy(address);

      if (Debug.on) {
         Debug.println(Debug.PLANNING, Debug.INFO, 
                       "PlanComm: Creating new planning comm channel to node "
                       + node + "(" + address + ")");
      }

      s = new Socket(address.getAddress(), 
                     PlanCommServer.PLAN_COMM_SERVER_PORT);

      localInterface.setBytes(s.getLocalAddress().getAddress());

      if (MSEC_INACTIVITY_TIMEOUT > 0) {
         s.setSoTimeout(MSEC_INACTIVITY_TIMEOUT);
      }

      s.setKeepAlive(true);

      if (PLAN_COMM_PRIORITY >= 0) {
         SocketOps.setPriority(s, PLAN_COMM_PRIORITY);
      }

      if (PLAN_COMM_LOWDELAY) {
         SocketOps.setTOS(s, SocketOps.TOS_LOWDELAY);
      }

      /* on active startup side, write local host id as first data */
      OutputStream rawOut = s.getOutputStream();
      localhost.writeSelf(rawOut);

      output = new PlanningOutputStream(rawOut);

      input = new PlanningInputStream(s.getInputStream());

      /* start input job */
      ThreadRecycler.startNewJob(this, Thread.MAX_PRIORITY - 1);
   }

   /**
    * Create a new PlanComm object to handle an incoming connection from
    * an adjacent Conductor node.  The thread is also obtained from the 
    * ThreadRecycler to handle input from the socket.
    * @param s The socket that is already connected to the adjacent node.
    * @param node The id of the node we're connected to.
    * @param commMgr The object keeping track of open connections.
    * @param sessionTable The list of all session objects.
    * @exception java.io.IOException if an error occurs obtaining the address 
    *            ofthe socket or an output stream for the socket.
    * @see FMG.Conductor.util.ThreadRecycler
    */
   public PlanComm(Socket s, PlanCommMgr commMgr, SessionTable sessionTable) 
                                                        throws IOException {
      this.s = s;
      this.commMgr = commMgr;
      this.sessionTable = sessionTable;

      // get the interface address from the socket
      address.setBytes(s.getInetAddress().getAddress());

      localInterface.setBytes(s.getLocalAddress().getAddress());

      if (MSEC_INACTIVITY_TIMEOUT > 0) {
         s.setSoTimeout(MSEC_INACTIVITY_TIMEOUT);
      }

      s.setKeepAlive(true);

      if (PLAN_COMM_PRIORITY >= 0) {
         SocketOps.setPriority(s, PLAN_COMM_PRIORITY);
      }

      if (PLAN_COMM_LOWDELAY) {
         SocketOps.setTOS(s, SocketOps.TOS_LOWDELAY);
      }

      /* on passive startup side, read local host id as first data */
      InputStream rawInput = s.getInputStream();
      node.readSelf(rawInput);

      if (Debug.on) {
         Debug.println(Debug.PLANNING, Debug.INFO, 
                       "PlanComm: Accepted a new planning comm channel " + 
                       "from node " + node + "(" + address + ")");
      }

      input = new PlanningInputStream(rawInput);

      output = new PlanningOutputStream(s.getOutputStream());

      /* start input job */
      ThreadRecycler.startNewJob(this, Thread.MAX_PRIORITY - 1);
   }

   /**
    * Register the message types that we currently forsee sending or receiving
    * on any given connection.  Note that other message types can later be
    * added.
    * @return No return value.
    */
   public static void registerMessages() {
      PlanPathInit.registerSelf();
      PlanDistro.registerSelf();
      PlanningInfo.registerSelf();
      PlanningDone.registerSelf();
      TriggerPlanning.registerSelf();
      PathClose.registerSelf();
   }

   /**
    * This method handles the main functionality of the input thread.
    * It reads messages, one at a time, from the socket, and passes them
    * to the appropriate path object using the connection id contained
    * in each message.  It is assumed that all received messages will
    * be of type PlanningMessage.  This method, and hence the thread,
    * will continue to run until an error occurs.
    * @return No return value.  Note, this method will not return unless
    *         an error occurs.
    */
   protected void handleInput() {
      try {
         while (true) {
            PlanningMessage m = (PlanningMessage) input.readMessage();
            if (Debug.on) {
               Debug.println(Debug.PLANNING, Debug.DETAIL, "## Got: " + m);
            }

            Session session = sessionTable.getSession(m.connID);
            Path path = null;
            if (session != null) {
               path = session.getPath();
            }

            if (path != null) {
               path.handleIncomingFromNet(m);
            } else {
               if (Debug.on) {
                  Debug.println(Debug.PLANNING, Debug.WARNING, 
                                "Warning: No path to handle message " + m);
               }
            }
         }
      } catch (IOException e) {
         if (Debug.on) {
            if (e instanceof InterruptedIOException) {
               Debug.println(Debug.PLANNING, Debug.INFO, 
                             "PlanComm: connection to " + node + 
                             " shut down due to inactivity.");
            } else {
               Debug.println(Debug.PLANNING, Debug.INFO, 
                             "PlanComm: input stream from " + node + 
                             " failed on input.");
            }
         }
         // remove self from eligible list
         commMgr.removeCommObj(node, this);

         closeConnection();
      }
   }

   /**
    * Send a message to the node that this object is connected to.
    * <p>
    * Warning: Unless an exception occurs, this call will recycle the message.
    *          If you want to maintain a handle to this message, 
    *          <code>retain()</code> must be called.
    * @param m The message to send.
    * @return No return value.
    * @exception java.io.IOException if an error occurs while writing the
    *                                message to the socket.
    */
   public void sendMessage(PlanningMessage m) throws IOException {
      try {
         if (Debug.on) {
            Debug.println(Debug.PLANNING, Debug.DETAIL, "## Sending: " + m);
         }
         output.writeMessage(m);
         m.recycle();
      } catch (IOException e) {
         if (Debug.on) {
            Debug.println(Debug.PLANNING, Debug.INFO, 
                          "PlanComm: input stream from " + node + 
                          " failed on output");
         }
         // remove self from eligible list
         commMgr.removeCommObj(node, this);

         closeConnection();

         throw e;
      }
   }

   /**
    * This method is the wrapper for the input thread.  It sets up the
    * thread and then calls <code>handleInput()</code>.
    * @return No return value.  Note, this method will not return unless
    *         an error occurs.
    */
   public void run() {
      if (Debug.on) {
         currentThread.setName("PlanComm(" + node + ")");
      }
      handleInput();
   }

   public boolean verifyInterfaceAddress(NodeID otherAddress) {
      return address.equals(otherAddress);
   }

   public void closeConnection() {
      if (s != null) {
         try {
            s.close();
         } catch (IOException ignore) {}
      }
   }

   public NodeID getRemoteNode() {
      return node.duplicate();
   }

   public NodeID getLocalInterface() {
      return localInterface.duplicate();
   }
}
