/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.Conductor.planning.comm;

import java.net.Socket;
import java.net.ServerSocket;
import java.io.IOException;

import FMG.Conductor.Conductor;
import FMG.Conductor.planning.*;
import FMG.Conductor.util.Debug;

/**
 * The PlanCommServer class creates and monitors a server socket that
 * accepts incoming planning connections.  When a new connection is 
 * received, the new socket is passed to the systems's PlanCommMgr, which
 * will create a PlanComm object to manage it.
 * <p>
 * The PlanCommServer runs in its own dedicated thread for the life of
 * Conductor.
 * <p>
 * The port that PlanCommServer listens to can be configured using the
 * <i>PlanningPort</i> property in Conductor's properties file.
 *
 * @author Mark Yarvis (yarvis@fmg.cs.ucla.edu)
 * @version $Revision: 1.14 $
 * 
 * @see FMG.Conductor.planning.comm.PlanComm
 */

public class PlanCommServer extends Thread {

   /** The port number to accept planning connections from. */
   final static int PLAN_COMM_SERVER_PORT = 
                             Conductor.getIntProperty("PlanningPort", 4445);

   /** The bound socket on which we will wait for connections */
   protected ServerSocket s;

   /** The object to which we report new connections */
   protected PlanCommMgr commMgr;

   /**
    * Create a new PlanCommServer object.  This includes binding a socket
    * to the appropriate port for incoming connections.  This method is 
    * normally only called by <code>createAndStart()</code> to ensure that 
    * our thread gets started.
    * @param pcm The PlanCommMgr object to report new connections to.
    * @see #createAndStart
    */
   protected PlanCommServer(PlanCommMgr pcm) {
      if (Debug.on) {
         setName("PlanCommServer");
      }

      commMgr = pcm;
      try {
         s = new ServerSocket(PLAN_COMM_SERVER_PORT);
      } catch (IOException e) {
         throw new RuntimeException("Could not bind to plan comm port: " + e);
      }
   }

   /**
    * Create a new PlanCommServer and start it up.  This method should 
    * normally only be called once during program execution.  Further calls
    * will not be able to bind to the server socket.
    * @param pcm The object to report new connections to.
    * @return No return value.
    */
   public static void createAndStart(PlanCommMgr pcm) {
      PlanCommServer pcs = new PlanCommServer(pcm);
      pcs.setPriority(MIN_PRIORITY+1);
      pcs.start();
   }

   /**
    * This method implements the main body of the server thread.  It waits
    * for new connections on the previously bound port and reports them
    * to the designated PlanCommMgr.  On error, the thread prints an error
    * message and dies.
    * return No return value.  In fact, unless an error occurs, this method
    *        will never return.
    */
   public void run () {
      while (true) {
         try {
            commMgr.acceptIncoming(s.accept());
         } catch (IOException e) {
            System.err.println("Error accepting connection to plann comm port "
                               + e);
         }
      }
   }
}
