/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.Conductor.planning.messages;

import java.io.*;
import java.net.InetAddress;

import FMG.Conductor.comm.*;
import FMG.Conductor.util.*;

/**
 * The PathRequest message is sent during path initialization, in search of
 * the next node in the path.  It is sent during initial startup, as well
 * as during failure recovery.
 * <p>
 * This message is one of the few (if not the only) messages to be packaged
 * into a UDP packet (by the PathFinder object), rather than sent along a 
 * socket stream.  Because of this, the methods required to interface
 * with BasicMessage{In,Out}putStream are not fully functional.
 * <p>
 * The PathRequest object is fully recyclable. It adds (to the data in the
 * PlanningMessage class) identity information for source and destination 
 * nodes as well as the identity of the previous hop.  
 *
 * @author Mark Yarvis (yarvis@fmg.cs.ucla.edu)
 * @version $Revision: 1.15 $
 *
 * @see FMG.Conductor.planning.comm.PathFinder
 * @see FMG.Conductor.comm.BasicMessageInputStream
 * @see FMG.Conductor.comm.BasicMessageOutputStream
 */

public class PathRequest extends PlanningMessage {

   /** The recycler used for messages of this type. */
   private static Recycler recycler = new Recycler(PathRequest.class);

   /** The identity of the source endpoint for this connection. */
   public NodeID sourceAddr = new NodeID();

   /** The port number used at the source endpoint for this connection. */
   public int sourcePort;

   /** The identity of the source endpoint for this connection. */
   public NodeID targetAddr = new NodeID();

   /** The port number used at the source endpoint for this connection. */
   public int targetPort;

   /** The previous hop along the path (ie, the sender of this message). */
   public NodeID prevNode = new NodeID();

   /** 
    * The interface that the source node used to send me this message.  Do
    * not transmit this value.  Instead set it upon reciept.
    */
   public NodeID sourceInterface = new NodeID();

   /**
    * Generate a new (or recycled) PathRequest message with the appropriate
    * data members initilized.
    * @param connID The id of the connection for which we are planning.
    * @param sourceAddr The address of the source endpoint of the connection
    * @param sourcePort The port number used at the source
    * @param targetAddr The address of the target endpoint of the connection
    * @param targetPort The port number used at the target
    * @param prevNode The identity of the sender of this message
    * @return A new (or recycled) and initialized PathRequest object.
    */
   public static PathRequest getNewMessage(ConnID connID, 
                                           NodeID sourceAddr, int sourcePort, 
                                           NodeID targetAddr, int targetPort,
                                           NodeID prevNode) {
      PathRequest m = (PathRequest) recycler.getOne();
      m.initMessage(connID, sourceAddr, sourcePort, 
                            targetAddr, targetPort, prevNode);
//      System.out.println("Set prevNode in PR to :" + prevNode.toString());
      return m;
   }

   /**
    * This method returns a new (or recycled) PathRequest object initialized
    * from the input stream.  This message is not normally called because
    * a PathRequest is sent only in packets, so the BasicMessageInputStream
    * is not used.
    * @param is The input stream to read the message's data from
    * @return a new data object initialized from the data stream.
    * @exception java.io.IOException if an error occurs while reading from 
    *                                the stream.
    * @see FMG.Conductor.comm.BasicMessageInputStream
    */
   public Message readNewMessage(DataInputStream is) throws IOException {
      return PathRequest.getNewMessage(is);
   }

   /**
    * This method should not be used.  Since this message does not interface
    * with the BasicMessageOutputStream, no id needs to be returned.
    * @return -1 is always returned.
    */
   public int getMessageID() {
      return -1;
   }

   /**
    * Crate a new (or recycled) PathRequest object and initialize it from 
    * an input stream.
    * @param is The input stream to read the message's data from.
    * @return A new object initialized from the input stream.
    * @exception java.io.IOException if an error occurs while reading from 
    *                                the stream.
    */
   public static PathRequest getNewMessage(DataInputStream is) 
                                                      throws IOException {
      PathRequest m = (PathRequest) recycler.getOne();
      m.receiveSelf(is);
      return m;
   }

   /**
    * Initilize the data members of this object.
    * @param connID The id of the connection for which we are planning.
    * @param sourceAddr The address of the source endpoint of the connection
    * @param sourcePort The port number used at the source
    * @param targetAddr The address of the target endpoint of the connection
    * @param targetPort The port number used at the target
    * @param prevNode The identity of the sender of this message
    * @return No return value.
    */
   protected void initMessage(ConnID connID, NodeID sourceAddr, int sourcePort, 
                                             NodeID targetAddr, int targetPort,
                                             NodeID prevNode) {
      initMessage(connID);
      this.targetAddr.copy(targetAddr);
      this.targetPort = targetPort;
      this.sourceAddr.copy(sourceAddr);
      this.sourcePort = sourcePort;
      this.prevNode.copy(prevNode);
   }

   /**
    * Initialize the data members of this object from the given input stream.
    * Presumably the data originated from a call to sendSelf().
    * @param is The input stream to read the message's data from.
    * @return No return value.
    * @exception java.io.IOException if an error occurs while reading from 
    *                                the stream.
    */
   public void receiveSelf(DataInputStream is) throws IOException {
      super.receiveSelf(is);
      targetAddr.readSelf(is);
      targetPort = is.readInt();
      sourceAddr.readSelf(is);
      sourcePort = is.readInt();
      prevNode.readSelf(is);
   }

   /**
    * Write the data members of this object to the given ouptut stream.
    * Presumably the data will later be read by a call to receiveSelf().
    * @param os The output stream to write the message's data to.
    * @return No return value.
    * @exception java.io.IOException if an error occurs while writing to 
    *                                the stream.
    */
   public void sendSelf(DataOutputStream os) throws IOException {
      super.sendSelf(os);
      targetAddr.writeSelf(os);
      os.writeInt(targetPort);
      sourceAddr.writeSelf(os);
      os.writeInt(sourcePort);
      prevNode.writeSelf(os);
   }

   /**
    * Set the interface this message was sent from.
    * @param addr The address of the interface this message originated from.
    * @return No return value.
    */
   public void setSourceInterface(InetAddress addr) {
      sourceInterface.setBytes(addr.getAddress());
   }

   /** 
    * Determine the direction of motion for this object.
    * @return Always returns <code>Direction.up</code>    
    */
   public Direction flowDir() { 
      return Direction.up; 
   }  
   
   /**
    * Determine whether this object is understood by the Path object.
    * @return Always returns true.
    */
   public boolean handledByPath() {
      return true;
   }

   /**
    * Create a string representation of this object.  This method is for
    * debuggin use only.
    * @return A new string containing the data from this object in a 
    *         human-readable form
    */
   public String toString() {
      return "PathRequest[" + sourceAddr + "(" + sourcePort + "), " + 
                              targetAddr + "(" + targetPort + "), " + 
                              prevNode + "]";
   }
}
