/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.Conductor.planning.messages;

import java.io.*;
import java.util.Vector;
import java.util.Enumeration;

import FMG.Conductor.comm.*;
import FMG.Conductor.util.*;

/**
 * The PlanDistro message is sent from the planner node to the root node,
 * along the data path, to deliver a new plan to all nodes along the path.
 * A plan consists of a set of nodes that will participate in each direction
 * of data flow and a set of adaptors that will be present on each node.
 * Each adaptor is specified as a fully-qualified adaptor class name.
 * <p>
 * This message requests a dynamically assigned message id from the
 * BasicMessageInputStream at startup time.
 * <p>
 * The PathClose object is fully recyclable.
 * <p>
 * TODO: a single version number is probably not enough.  Consider what
 *       happens when a connection is partitioned and planning continues.
 *
 * @author Mark Yarvis (yarvis@fmg.cs.ucla.edu)
 * @version $Revision: 1.23 $
 *
 * @see FMG.Conductor.comm.BasicMessageInputStream
 */

public class PlanDistro extends PlanningMessage {

   /**
    * The maximum number of adaptors that can be deployed at a single node.
    * <p>
    * TODO: This is a completely artifical limit, set for simplicity of 
    *       implementation.
    */
   public static final int MAX_ADAPTORS = 10;

   /**
    * This class describes the portion of a plan that applies to a single node.
    */
   protected class NodePlan {
      /** The node id that this NodePlan pertains to */
      NodeID node = new NodeID();

      /** The address to contact this node from upstream. */
      NodeID upstreamAddress = new NodeID();

      /** The number of adaptors to deploy on this node */
      int numAdaptors = 0;

      /** The list of adaptors to deploy on this node */
      String[] adaptors = new String[MAX_ADAPTORS];

      /**
       * Create a NodePlan for a particular node with the given list of 
       * adaptors.
       * @param node The node to which this NodePlan pertains.
       * @param adaptors a list of fully-qualified adaptor class names.
       */
      NodePlan(NodeID node, NodeID upstreamAddress, String[] adaptors) {
         this.node.copy(node);
         this.upstreamAddress.copy(upstreamAddress);

         if (adaptors == null) {
            numAdaptors = 0;
            return;
         }

         numAdaptors = adaptors.length;

         if (numAdaptors > MAX_ADAPTORS) {
            throw new RuntimeException("Too many adaptors");
         }

         for (int i=0; i<numAdaptors; i++) {
            this.adaptors[i] = adaptors[i];
         }
      }

      /**
       * Create a NodePlan by reading data from the input stream.  Presumably
       * <code>writeSelf</code> was used to write the data to the stream.
       * @param is The input stream from which to read. 
       * @exception java.io.IOException if an error occurs while reading.
       */
      NodePlan(DataInputStream is) throws IOException {
         node = new NodeID(is);
         upstreamAddress = new NodeID(is);
         numAdaptors = is.readInt();
         for (int i = 0; i < numAdaptors; i++) {
            adaptors[i] = is.readUTF();
         }
      }

      /**
       * Write the contents of a NodePlan to an output stream.  Presumably
       * the data will be read by the appropriate constructor.
       * @param os The output stream to which to write.
       * @return No return value.
       * @exception java.io.IOException if an error occurs while writing.
       */
      public void writeSelf(DataOutputStream os) throws IOException {
         node.writeSelf(os);
         upstreamAddress.writeSelf(os);
         os.writeInt(numAdaptors);
         for (int i = 0; i < numAdaptors; i++) {
            os.writeUTF(adaptors[i]);
         }
      }

      /**
       * Generate a string containing a human-readable version of the portion
       * of the plan represented by this object.  This method is for debugging
       * use only.
       * return A new string containing this part of the plan in a 
       *        human-readable format.
       */
      public String toString() {
         StringBuffer s = new StringBuffer(node.toString());
         if (! upstreamAddress.isNull() && ! upstreamAddress.equals(node)) {
            s.append("[");
            s.append(upstreamAddress.toString());
            s.append("]");
         }
         s.append("{");
         for (int i=0; i<numAdaptors; i++) {
            s.append(adaptors[i]);
            if (i < (numAdaptors-1)) {
               s.append(',');
            }
         }
         s.append('}');
         return s.toString();
      }

   }

   /** 
    * This list contains the plan for the upstream (from client to server) 
    * direction of the data flow.  Each element in the vector is a NodePlan 
    * object, representing the adaptors to deploy on a single ndoe.
    */
   protected Vector upNodes = new Vector();

   /** 
    * This list contains the plan for the downstream (from server to client) 
    * direction of the data flow.  Each element in the vector is a NodePlan 
    * object, representing the adaptors to deploy on a single ndoe.
    */
   protected Vector downNodes = new Vector();

   /** The version number for this plan */
   public int version;

   /** The identity of the planner */
   public NodeID planner = new NodeID();

   /** The recycler for messages of this message type. */
   private static Recycler recycler = new Recycler(PlanDistro.class);

   /** A cached copy of the identity of the local node */
   protected static NodeID localhost = NodeID.getNewLocalID();

   /** The message ID, which is dynamically assigned */
   private static int messageID = -1;
   
   /**
    * Register a prototype of this message class with the 
    * BasicMessageInputStream.
    * @return No return value.
    * @see FMG.Conductor.comm.BasicMessageInputStream
    */
   public static void registerSelf() {
      messageID = BasicMessageInputStream.
                        registerMessagePrototype((Message) recycler.getOne());
   }
   
   /**
    * This message returns the message id that was assigned to messages of
    * this type by BasicMessageInputStream.
    * @return The message id for this type of message.
    * @see FMG.Conductor.comm.BasicMessageInputStream
    */
   public int getMessageID() {
      return messageID;
   }

   /**
    * This method returns a new (or recycled) PlanDistro object initialized
    * from the input stream.  Normally, this method is only called on the
    * prorotype given to BasicMessageInputStream.
    * @param is The input stream to read the message's data from
    * @return a new object initialized from the data stream.
    * @exception java.io.IOException if an error occurs while reading from
    *                                the stream.
    * @see FMG.Conductor.comm.BasicMessageInputStream
    */
   public Message readNewMessage(DataInputStream is) throws IOException {
      return PlanDistro.getNewMessage(is);
   }
   
   /**
    * Generate a new (or recycled) PlanDistro message that is partially
    * initialized.  Initially, there are no adaptors and no nodes
    * specified.  Further calls to addNode() and insertNode() should be
    * made to fill out the plan.
    * @param connID The id of the connection for which we are planning.
    * @return A new (or recycled) and partially initialized PlanDistro message.
    */
   public static PlanDistro getNewMessage(ConnID connID) {
      PlanDistro m = (PlanDistro) recycler.getOne();
      m.initMessage(connID);
      return m;
   }

   /**
    * Initialize a PlanDistro object.
    * @param connID The connection id to set for this object.
    * @return no return value.
    */
   protected void initMessage(ConnID connID) {
      super.initMessage(connID);
      planner.copy(localhost);
   }

   /**
    * This method returns a new (or recycled) PlanDistro object initialized
    * from the input stream.
    * @param is The input stream to read the message's data from
    * @return A new (or recycled) and initialized PlanDistro message.
    * @exception java.io.IOException if an error occurs while reading from
    *                                the stream.
    */
   public static PlanDistro getNewMessage(DataInputStream is) 
                                                     throws IOException {
      PlanDistro m = (PlanDistro) recycler.getOne();
      m.receiveSelf(is);
      return m;
   }

   /**
    * Initialize this message from the input stream.  Presumably the
    * data was written using <code>sendSelf()</code>.
    * @param is The stream to read the message contents from.
    * @return No return value.
    * @exception java.io.IOException if an error occurs while reading from
    *                                the stream.
    */
   public void receiveSelf(DataInputStream is) throws IOException {
      super.receiveSelf(is);

      int num = is.readInt();
      for (int i=0; i<num; i++) {
         upNodes.addElement(new NodePlan(is));
      }

      num = is.readInt();
      for (int i=0; i<num; i++) {
         downNodes.addElement(new NodePlan(is));
      }

      version = is.readInt();

      planner.readSelf(is);
   }

   /**
    * Write the contents of this message to the output stream.  Presumably the
    * data will be read using <code>receiveSelf()</code>.
    * @param os The stream to write the message contents to.
    * @return No return value.
    * @exception java.io.IOException if an error occurs while writing to
    *                                the stream.
    */
   public void sendSelf(DataOutputStream os) throws IOException {
      super.sendSelf(os);

      int num = upNodes.size();
      os.writeInt(num);
      for (int i=0; i<num; i++) {
         ((NodePlan) upNodes.elementAt(i)).writeSelf(os);
      }

      num = downNodes.size();
      os.writeInt(num);
      for (int i=0; i<num; i++) {
         ((NodePlan) downNodes.elementAt(i)).writeSelf(os);
      }

      os.writeInt(version);

      planner.writeSelf(os);
   }

   /** 
    * Determine the direction of motion for this object.
    * @return Always returns <code>Direction.down</code>    
    */
   public Direction flowDir() { 
      return Direction.down; 
   }  
   
   /**
    * Determine whether this object is understood by the Path object.
    * @return Always returns true.
    */
   public boolean handledByPath() {
      return true;
   }

   /**
    * Add a new part to the plan.  This new part consists of a node that
    * will be participating, an address at which that node can be contacted,
    * and the complete list of adaptors that will be required on that node.
    * This node is added to the <b>end</b> of the list of nodes that will be 
    * present in the given direction of data flow.
    * @param node The identity of the node that is to be added to the plan.
    * @param upstreamAddress The address through which this node can be 
    *                        contacted from upstream.
    * @param adaptors An array of fully qualified adaptor class names to be 
    *                 instantiated on the given node.
    * @param direction The direction of flow to which this node specification
    *                  will be added.
    * @return No return value.
    */
   public void addNode(NodeID node, NodeID upstreamAddress, 
                       String[] adaptors, Direction direction) {
      if (direction.isUp()) {
         upNodes.addElement(new NodePlan(node, upstreamAddress, adaptors));
      } else {
         downNodes.addElement(new NodePlan(node, upstreamAddress, adaptors));
      }
   }

   /**
    * Add a new part to the plan.  This new part consists of a node that
    * will be participating and the complete list of adaptors that will 
    * be required on that node.  This node is added to the <b>end</b> of 
    * the list of nodes that will be present in the given direction of data 
    * flow.
    * @param node The identity of the node that is to be added to the plan.
    * @param adaptors An array of fully qualified adaptor class names to be 
    *                 instantiated on the given node.
    * @param direction The direction of flow to which this node specification
    *                  will be added.
    * @return No return value.
    */
   public void addNode(NodeID node, String[] adaptors, 
                          Direction direction) {
      addNode(node, node, adaptors, direction);
   }

   /**
    * Add a new part to the plan.  This new part consists of a node that
    * will be participating, an address for contacting that node, and the 
    * complete list of adaptors that will be required on that node.  This 
    * node is added to the <b>beginning</b> of the list of nodes that will 
    * be present in the given direction of data flow.
    * @param node The identity of the node that is to be added to the plan.
    * @param upstreamAddress The address through which this node can be 
    *                        contacted from upstream.
    * @param adaptors An array of fully qualified adaptor class names to be 
    *                 instantiated on the given node.
    * @param direction The direction of flow to which this node specification
    *                  will be added.
    * @return No return value.
    */
   public void insertNode(NodeID node, NodeID upstreamAddress, 
                          String[] adaptors, Direction direction) {
      if (direction.isUp()) {
         upNodes.insertElementAt(new NodePlan(node, upstreamAddress, 
                                                            adaptors), 0);
      } else {
         downNodes.insertElementAt(new NodePlan(node, upstreamAddress, 
                                                              adaptors), 0);
      }
   }

   /**
    * Add a new part to the plan.  This new part consists of a node that
    * will be participating and the complete list of adaptors that will be 
    * required on that node.  This node is added to the <b>beginning</b> of the 
    * list of nodes that will be present in the given direction of data flow.
    * @param node The identity of the node that is to be added to the plan.
    * @param adaptors An array of fully qualified adaptor class names to be 
    *                 instantiated on the given node.
    * @param direction The direction of flow to which this node specification
    *                  will be added.
    * @return No return value.
    */
   public void insertNode(NodeID node, String[] adaptors, 
                          Direction direction) {
      insertNode(node, node, adaptors, direction);
   }

   /**
    * Modify a plan so that a different address is used to contact the given
    * node when forming a data path.  Since the path is formed in a downstream
    * manner, we need to know the upstream-facing address that corresponds
    * to that node.
    * @param node The node in question.
    * @param upstreamAddress The address to use in contacting this node from 
    *                        downstream.
    * @param dir The direction of data flow to adjust.
    */
   public void setUpstreamAddress(NodeID node, NodeID upstreamAddress, 
                                  Direction dir) {
      Vector nodes = (dir.isUp() ? upNodes : downNodes);
      int numNodes = nodes.size();

      for (int i=0; i<numNodes; i++) {
         NodePlan np = (NodePlan) nodes.elementAt(i);
         if (np.node.equals(node)) {
            np.upstreamAddress.copy(upstreamAddress);
         }
      }
   }

   /**
    * Get the identity of the node that follows this one in the list of
    * nodes for this direction of the plan.
    * @param direction The direction of flow we care about.
    * @return The identity of the next node after us, as specified by the plan,
    *         in the requested direction of flow.
    */
   public NodeID getNextNode(Direction direction) {
      Vector nodes = (direction.isUp() ? upNodes : downNodes);

      int num = nodes.size();
      for (int i=0; i<(num - 1); i++) {
         if (((NodePlan) nodes.elementAt(i)).node.isLocalNode()) {
            return ((NodePlan) nodes.elementAt(i+1)).node;
         }
      }

      return null;
   }

   /**
    * Get the interface address for contacting the node that follows this 
    * one in the list of nodes for this direction of the plan.
    * @param direction The direction of flow we care about.
    * @return The interface address for the next node after us, as specified 
    *         by the plan, in the requested direction of flow.
    */
   public NodeID getNextInterface(Direction direction) {
      Vector nodes = (direction.isUp() ? upNodes : downNodes);

      int num = nodes.size();
      for (int i=0; i<(num - 1); i++) {
         if (((NodePlan) nodes.elementAt(i)).node.isLocalNode()) {
            return ((NodePlan) nodes.elementAt(i+1)).upstreamAddress;
         }
      }

      return null;
   }

   /**
    * Obtain a list of all nodes called for in this plan for the given 
    * direction of data flow.
    * @param dir The direction of data flow of interest.  Should be either
    *            <code>Direction.up</code> or <code>Direction.down</code>
    * @return An Enumeration of node IDs.
    * @see FMG.Conductor.comm.NodeID
    */
   public Enumeration getNodes(Direction dir) {
      Vector nodePlans = (dir.isUp() ? upNodes : downNodes);
      Vector nodes = new Vector();

      int num = nodePlans.size();
      for (int i=0; i<num; i++) {
         nodes.addElement(((NodePlan) nodePlans.elementAt(i)).node);
      }

      return nodes.elements();
   }

   /**
    * Obtain a list of all adaptors called for in this plan for a particular
    * node in a particular direction of data flow.
    * Note that each adaptor is represented by a string that consists
    * of a class name and an optional parameter, separated by a space.
    * @param node The id of the node of interest.
    * @param dir The direction of data flow of interest.  Should be either
    *            <code>Direction.up</code> or <code>Direction.down</code>
    * @return An Enumeration of adaptor class names (and optional parameters).
    */
   public Enumeration getAdaptors(NodeID node, Direction dir) {
      Vector nodePlans = (dir.isUp() ? upNodes : downNodes);

      for (int i=0; i<nodePlans.size(); i++) {
         if (((NodePlan) nodePlans.elementAt(i)).node.equals(node)) {
            NodePlan plan = (NodePlan) nodePlans.elementAt(i);
            Vector adaptors = new Vector();

            for (int j=0; j<plan.numAdaptors; j++) {
               adaptors.addElement(plan.adaptors[j]);
            }

            return adaptors.elements();
         }
      }

      return null;
   }

   /**
    * Get the list of adaptors to deploy on this node for this direction of 
    * flow.  Note that each adaptor is represented by a string that consists
    * of a class name and an optional parameter, separated by a space.
    * @param direction The direction of flow we care about.
    * @return A list of fully qualified adaptor class names (and optional 
    *         parameters) to be deployed on this node as specified by the 
    *         plan.
    */
   public String[] getLocalAdaptors(Direction direction) {
      Vector nodes = (direction.isUp() ? upNodes : downNodes);

      int num = nodes.size();
      for (int i=0; i<num; i++) {
         if (((NodePlan) nodes.elementAt(i)).node.isLocalNode()) {
            return ((NodePlan) nodes.elementAt(i)).adaptors;
         }
      }

      return null;
   }

   /**
    * Determine how many adaptors were deployed prior to the current one by 
    * this plan in this direction.  This value is used to provide a unique
    * "type" for that adaptor, in case it's a paired adaptor.
    * @param direction The direction of data flow that's of interest.
    * @return The number of adaptors deployed previously in this direction
    *         by this plan.
    */
   public int numPrevious(Direction direction) {
      Vector nodes = (direction.isUp() ? upNodes : downNodes);

      int ret = 0;

      int num = nodes.size();
      for (int i=0; i<num; i++) {
         if (((NodePlan) nodes.elementAt(i)).node.isLocalNode()) {
            return ret;
         } else {
            ret += ((NodePlan) nodes.elementAt(i)).numAdaptors;
         }
      }

      throw new RuntimeException("This shouldn't happen");
   }

   /**
    * Get a string representing this plan in a human-readable format.  This
    * method should be used for debugging purposes only.
    * @return a human-readable string representing this plan.
    */
   public String toString() {
      int num = upNodes.size();
      StringBuffer s = new StringBuffer("PlanDistro(up:");

      for (int i=0; i<num; i++) {
         s.append((NodePlan) upNodes.elementAt(i));

         if (i<(num-1)) {
            s.append(", ");
         }
      }

      s.append(";down:");
      num = downNodes.size();
      for (int i=0; i<num; i++) {
         s.append((NodePlan) downNodes.elementAt(i));

         if (i<(num-1)) {
            s.append(", ");
         }
      }

      s.append(")");
      return s.toString();
   }

   /**
    * Perform cleanup before allowing recycling to occur.  This method
    * is called by the Recycler.
    * @return No return value.
    * @see FMG.Conductor.util.Recycler
    */
   public void recycling() {
      super.recycling();
      upNodes.removeAllElements();
      downNodes.removeAllElements();
   }
}
