/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.Conductor.planning.messages;

import java.io.*;

import FMG.Conductor.comm.*;
import FMG.Conductor.util.*;

/**
 * This message is sent in response to a PathRequest message.  Unlike
 * the PathRequest, but like most other messages, this message is sent
 * along a socket stream.  This message identifies, to the receiver, the 
 * identity of the next node along the path and confirms the setup of 
 * communication with that node (and therefore acknowledges the PathRequest
 * message).
 * <p>
 * This message requests a dynamically assigned message id from the
 * BasicMessageInputStream at startup time.
 * <p>
 * The PlanPathInit object is fully recyclable.
 *
 * @author Mark Yarvis (yarvis@fmg.cs.ucla.edu)
 * @version $Revision: 1.10 $
 *
 * @see FMG.Conductor.comm.BasicMessageInputStream
 */

public class PlanPathInit extends PlanningMessage {

   /** The recycler used for messages of this type */
   private static Recycler recycler = new Recycler(PlanPathInit.class);

   /**
    * The identity of the next node along the path (ie, the sender of this 
    * message). 
    */
   public NodeID nextNode = new NodeID();

   /**
    * The interface from which we received a PathRequest message that 
    * triggered this response.
    */
   public NodeID yourInterface = new NodeID();

   /** The message id assigned to messages of this type */
   private static int messageID = -1;
   
   /**
    * Register a prototype of this message class with the
    * BasicMessageInputStream.
    * @return No return value.
    * @see FMG.Conductor.comm.BasicMessageInputStream
    */
   public static void registerSelf() {
      messageID = BasicMessageInputStream.
                        registerMessagePrototype((Message) recycler.getOne());
   }
   
   /**
    * This method returns the message id that was assigned to messages of this
    * type by BasicMessageInputStream.
    * @return The message id for this type of message.
    * @see FMG.Conductor.comm.BasicMessageInputStream
    */
   public int getMessageID() {
      return messageID;
   }
   
   /**
    * This method returns a new (or recycled) PlanPathInit object initialized
    * from the input stream.  Normally, this method is only called on the
    * prorotype given to BasicMessageInputStream.
    * @param is The input stream to read the message's data from
    * @return a new object initialized from the data stream.
    * @exception java.io.IOException if an error occurs while reading from
    *                                the stream.
    * @see FMG.Conductor.comm.BasicMessageInputStream
    */
   public Message readNewMessage(DataInputStream is) throws IOException {
      return PlanPathInit.getNewMessage(is);
   }
   
   /**
    * Generate a new (or recycled) PlanPathInit message with the appropriate
    * data members initialized.
    * @param connID The id of the connection for which we are planning.
    * @param addr The id of the node sending the message
    * @return A new (or recycled) and initialized PlanPathInit message.
    */
   public static PlanPathInit getNewMessage(ConnID connID, NodeID addr, 
                                            NodeID yourInterface) {
      PlanPathInit m = (PlanPathInit) recycler.getOne();
      m.initMessage(connID, addr, yourInterface);
      return m;
   }

   /**
    * This method returns a new (or recycled) PlanPathInit object initialized
    * from the input stream.
    * @param is The input stream to read the message's data from
    * @return A new (or recycled) and initialized PlanPathInit message.
    * @exception java.io.IOException if an error occurs while reading from
    *                                the stream.
    */
   public static PlanPathInit getNewMessage(DataInputStream is) 
                                                        throws IOException {
      PlanPathInit m = (PlanPathInit) recycler.getOne();
      m.receiveSelf(is);
      return m;
   }

   /**
    * This method is used to initialize the data elements of this object.
    * @param connID The id of the connection for which we are planning.
    * @param addr The id of the node sending the message
    * @return No return value.
    */
   protected void initMessage(ConnID connID, NodeID addr, 
                              NodeID yourInterface) {
      initMessage(connID);

      nextNode.copy(addr);
      this.yourInterface.copy(yourInterface);
   }

   /**
    * This method is used to read the data elements of this object from
    * the input stream.  Presumably the information was generated by 
    * <code>sendSelf()</code>.
    * @param is The input stream to read the message's data from.
    * @return No return value.
    * @exception java.io.IOException if an error occurs while reading from the 
    *                                stream.
    */
   public void receiveSelf(DataInputStream is) throws IOException {
      super.receiveSelf(is);
      nextNode.readSelf(is);
      yourInterface.readSelf(is);
   }

   /**
    * This method is used to write the data elements of this object to
    * an output stream.  Presumably the information will be read by
    * <code>receiveSelf()</code>.
    * @param os The output stream to write the message's data to.
    * @return No return value.
    * @exception java.io.IOException if an error occurs while writing to the 
    *                                stream.
    */
   public void sendSelf(DataOutputStream os) throws IOException {
      super.sendSelf(os);
      nextNode.writeSelf(os);
      yourInterface.writeSelf(os);
   }

   /** 
    * Determine the direction of motion for this object.
    * @return Always returns <code>Direction.down</code>    
    */
   public Direction flowDir() { 
      return Direction.down; 
   }  
   
   /**
    * Determine whether this object is understood by the Path object.
    * @return Always returns true.
    */
   public boolean handledByPath() {
      return true;
   }

}
