/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.Conductor.planning.messages;

import java.io.*;
import java.util.*;

import FMG.Conductor.comm.*;
import FMG.Conductor.util.*;

/**
 * This message is sent from each node (and forwarded along the path) toward
 * the planner.  It includes local node information involved in planning
 * like whether this node is the planner or root node, the node's id, and
 * the id's of the neighbor nodes.  
 * <p>
 * This class can also be subclassed to allow subclasses of EnvMonitor to 
 * produce PlanningInfo objects that include more extensive local node
 * information.  If this is done, these subclasses must register themselves
 * with BasicMessageInputStream (probably statically).
 * This message requests a dynamically assigned message id from the
 * BasicMessageInputStream at startup time.
 * <p>
 * The PlanningInfo object is fully recyclable.
 *  
 * @author Mark Yarvis (yarvis@fmg.cs.ucla.edu)
 * @version $Revision: 1.21 $
 * 
 * @see FMG.Conductor.comm.BasicMessageInputStream
 * @see FMG.Conductor.monitor.EnvMonitor
 */

public class PlanningInfo extends PlanningMessage {

   /** The recycler used for messages of this type */
   private static Recycler recycler = new Recycler(PlanningInfo.class);


   /** Is this the planner node? */
   public boolean isPlanner;

   /** Is this the root node? */
   public boolean isRoot;

   /** The node's id */
   public NodeID node = new NodeID();

   /** The id of the node's next neighbor */
   public NodeID nextNode = new NodeID();

   /** The address of the interface used to communicate with the next node. */
   public NodeID interfaceToNext = new NodeID();

   /** The address of the interface used to communicate with the next node. */
   public NodeID interfaceToPrev = new NodeID();

   /** The id of the node's previous neighbor */
   public NodeID prevNode = new NodeID();

   /** 
    * Planning info from this node.  Convention: prefix indexes in this
    * table with "UserPrefs", "NextLinkProps", "PrevLinkProps", or 
    * "NodeProps".
    */
   public Hashtable info = new Hashtable();

   /** The message id assigned to messages of this type */
   private static int messageID = -1;
   
   /**
    * Register a prototype of this message class with the
    * BasicMessageInputStream.
    * @return No return value.
    * @see FMG.Conductor.comm.BasicMessageInputStream
    */
   public static void registerSelf() {
      messageID = BasicMessageInputStream.
                        registerMessagePrototype((Message) recycler.getOne());
   }
   
   /**
    * This method returns the message id that was assigned to messages of this
    * type by BasicMessageInputStream.
    * @return The message id for this type of message.
    * @see FMG.Conductor.comm.BasicMessageInputStream
    */
   public int getMessageID() {
      return messageID;
   }
   
   /**
    * This method returns a new (or recycled) PlanningInfo object initialized
    * from the input stream.  Normally, this method is only called on the
    * prorotype given to BasicMessageInputStream.
    * @param is The input stream to read the message's data from
    * @return a new object initialized from the data stream.
    * @exception java.io.IOException if an error occurs while reading from
    *                                the stream.
    * @see FMG.Conductor.comm.BasicMessageInputStream
    */
   public Message readNewMessage(DataInputStream is) throws IOException {
      return PlanningInfo.getNewMessage(is);
   }
   
   /**
    * Generate a new (or recycled) PlanningInfo message with the appropriate
    * data members initialized.
    * @param connID The id of the connection for which we are planning.
    * @param addr The id of the node being described
    * @param prev The id of the node's previous neighbor
    * @param next The id of the node's next neighbor
    * @param interfaceToprev The address used to communicate with the previous 
    *                        node.
    * @param interfaceToNext The address used to communicate with the next 
    *                        node.
    * @param isPlanner Is this node the planner?
    * @param isRoot Is this node the root?
    * @return A new (or recycled) and initialized PlanningInfo message.
    */
   public static PlanningInfo getNewMessage(ConnID connID, NodeID addr, 
                                          NodeID prev, NodeID interfaceToPrev, 
                                          NodeID next, NodeID interfaceToNext, 
                                            boolean isPlanner, boolean isRoot) {
      PlanningInfo m = (PlanningInfo) recycler.getOne();
      m.initMessage(connID, addr, prev, interfaceToPrev, next, interfaceToNext,
                                           isPlanner, isRoot);
      return m;
   }

   /**
    * This method returns a new (or recycled) PlanningInfo object initialized
    * from the input stream.
    * @param is The input stream to read the message's data from
    * @return A new (or recycled) and initialized PlanningInfo message.
    * @exception java.io.IOException if an error occurs while reading from
    *                                the stream.
    */
   public static PlanningInfo getNewMessage(DataInputStream is) 
                                                     throws IOException {
      PlanningInfo m = (PlanningInfo) recycler.getOne();
      m.receiveSelf(is);
      return m;
   }

   /**
    * This method is used to initialize the data elements of this object.
    * @param connID The id of the connection for which we are planning.
    * @param addr The id of the node being described
    * @param prev The id of the node's previous neighbor
    * @param next The id of the node's next neighbor
    * @param interfaceToPrev The address used to communicate with the previous 
    *                        node.
    * @param interfaceToNext The address used to communicate with the next 
    *                        node.
    * @param isPlanner Is this node the planner?
    * @param isRoot Is this node the root?
    * @return No return value.
    */
   protected void initMessage(ConnID connID, NodeID addr, 
                              NodeID prev, NodeID interfaceToPrev,
                              NodeID next, NodeID interfaceToNext, 
                              boolean isPlanner, boolean isRoot) {
      initMessage(connID);
      node.copy(addr);
      this.interfaceToPrev.copy(interfaceToPrev);
      this.interfaceToNext.copy(interfaceToNext);
      prevNode.copy(prev);
      nextNode.copy(next);
      this.isPlanner = isPlanner;
      this.isRoot = isRoot;
      info.clear();
   }

   /**
    * Copy the planning information in the given <code>Hashtable</code> 
    * into the payload of this message.
    * @param newInfo The hashtable to copy from.
    * @return No return value.
    */
   public void copyInfo(Hashtable newInfo) {
      Enumeration keys = newInfo.keys();
      while (keys.hasMoreElements()) {
         String aKey = (String) keys.nextElement();
         info.put(aKey, newInfo.get(aKey));
      }
   }

   public void addInfo(String key, String value) {
      info.put(key, value);
   }

   /**
    * This method is used to read the data elements of this object from
    * the input stream.  Presumably the information was generated by
    * <code>sendSelf()</code>.
    * @param is The input stream to read the message's data from.
    * @return No return value.
    * @exception java.io.IOException if an error occurs while reading from the
    *                                stream.
    */
   public void receiveSelf(DataInputStream is) throws IOException {
      super.receiveSelf(is);
      node.readSelf(is);
      interfaceToPrev.readSelf(is);
      interfaceToNext.readSelf(is);
      prevNode.readSelf(is);
      nextNode.readSelf(is);
      isRoot = is.readBoolean();
      isPlanner = is.readBoolean();
      info.clear();
      readHashTable(is, info);
   }

   /**
    * This method is used to write the data elements of this object to
    * an output stream.  Presumably the information will be read by
    * <code>receiveSelf()</code>.
    * @param os The output stream to write the message's data to.
    * @return No return value.
    * @exception java.io.IOException if an error occurs while writing to the
    *                                stream.
    */
   public void sendSelf(DataOutputStream os) throws IOException {
      super.sendSelf(os);
      node.writeSelf(os);
      interfaceToPrev.writeSelf(os);
      interfaceToNext.writeSelf(os);
      prevNode.writeSelf(os);
      nextNode.writeSelf(os);
      os.writeBoolean(isRoot);
      os.writeBoolean(isPlanner);
      writeHashTable(os, info);
   }

   /** 
    * Determine the direction of motion for this object.
    * @return Always returns <code>Direction.down</code>    
    */
   public Direction flowDir() { 
      return Direction.up; 
   }  
   
   /**
    * Determine whether this object is understood by the Path object.
    * @return Always returns true.
    */
   public boolean handledByPath() {
      return true;
   }

   /**
    * This method generates a human-readable string representing the contents
    * of this object.  It should be used for debugging purposes only.
    * @return A new string containing a human-readable representation of this 
    * object.
    */
   public String toString() {
      StringBuffer sb = new StringBuffer("PlanningInfo(");
      sb.append(node.toString());
      if (Debug.on && (Debug.getLevel(Debug.PLANNING) > Debug.DETAIL)) {
         sb.append(",");
         sb.append(info.toString());
      }
      sb.append(")");
      return sb.toString();
   }
}
