/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.Conductor.planning.messages;

import java.io.*;
import java.util.*;

import FMG.Conductor.comm.*;
import FMG.Conductor.util.*;

/**
 * The PlanningMessage class is the abstract base class for all messages 
 * involved in planning.  It implements the functions of the Message 
 * interface, and includes the one data object, the connection id, that is
 * included in all planning messages.
 * <p>
 * This class is a subclass of Recyclable, allowing planning messages to
 * potentiall by recycled.  Implementers of subclasses should implement the 
 * appropriate recycling functionality.
 * <p>
 * Subclasses will normally add data items required for individual messages.
 * To handle these data items, a new version of <code>initMessage()</code> 
 * should be written with additional arguments.  In addition, 
 * <code>readSelf()</code> and <code>writeSelf()</code> should be overridden
 * to handle input and output of these additional data items.
 *
 * @author Mark Yarvis (yarvis@fmg.cs.ucla.edu)
 * @version $Revision: 1.13 $
 */

public abstract class PlanningMessage extends Recyclable implements Message {

   /** The id of the connection for which we are planning */
   public ConnID connID = new ConnID();

   /** 
    * Initialize a previously created or recycled object with a given
    * connection ID.
    * @param connID The connection id to record in this message.
    * @return No return value.
    */
   protected void initMessage(ConnID connID) {
      this.connID.copy(connID);
   }

   /**
    * Read the data for this message from an input stream.
    * @param is The input stream to read from.
    * @return No return value.
    * @exception java.io.IOException if an error occurs while reading from 
    *                                the stream.
    */
   public void receiveSelf(DataInputStream is) throws IOException {
      connID.readSelf(is);
   }

   /**
    * Write the data content of this message to an output stream.
    * @param os The output stream to write to.
    * @return No return value.
    * @exception java.io.IOException if an error occurs while writing to 
    *                                the stream.
    */
   public void sendSelf(DataOutputStream os) throws IOException {
      connID.writeSelf(os);
   }

   /**
    * This is not a valid call for objects of this type.  A RuntimeException
    * will result.
    */
   public void writeToRaw(OutputStream os) throws IOException {
      throw new RuntimeException("This shouldn't happen!");
   }

   /**
    * Determine the direction of motion for this object.  Direction.up
    * indicates that this object is flowing toward the server.  Direction.down
    * indicates that this object is flowing toward the client.
    * @return Either <code>Direction.up</code> or <code>Direction.down</code>.
    */
   public abstract Direction flowDir();

   /**
    * Determine whether this object is understood by the Path object.
    * @return True iff the Path object understands messages of this type.
    */
   public abstract boolean handledByPath();

   /**
    * Write data from a hashtable to an output stream.
    * @param is The output stream to write to.
    * @param table The hash table to write data from.
    * @return no return value.
    * @exception java.io.IOException if an error occurs while writing to 
    *                                the stream.
    */
   protected void writeHashTable(DataOutputStream os, Hashtable table)
                                                          throws IOException {
      os.writeInt(table.size());

      Enumeration keys = table.keys();
      while (keys.hasMoreElements()) {
         String aKey = (String) keys.nextElement();
         os.writeUTF(aKey);
         os.writeUTF((String) table.get(aKey));
      }
   }
 
   /**
    * Add data to a hashtable from an input stream.  NOTE: the table is
    * not cleared.  Clear it ahead of time if you want the table to contain
    * only the data read from the stream.
    * @param is The input stream to read from.
    * @param table The hash table to add read data to.
    * @return no return value.
    * @exception java.io.IOException if an error occurs while reading from
    *                                the stream.
    */
   protected void readHashTable(DataInputStream is, Hashtable table)
                                                        throws IOException {
      int size = is.readInt();
 
      for (int i=0; i < size; i++) {
         String aKey = is.readUTF(); 
         String aValue = is.readUTF();
         table.put(aKey, aValue);
      }
   }

}
