/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.Conductor.planning.messages;

import java.io.*;

import FMG.Conductor.comm.*;
import FMG.Conductor.util.*;

/**
 * This message is generated at any node along the path and is forwarded
 * toward the root node to trigger planning activities.  The message includes
 * no additional data items.
 * <p>
 * This message requests a dynamically assigned message id from the
 * BasicMessageInputStream at startup time.
 * <p>
 * The TriggerPlanning object is fully recyclable.
 * 
 * @author Mark Yarvis (yarvis@fmg.cs.ucla.edu)
 * @version $Revision: 1.6 $
 *  
 * @see FMG.Conductor.comm.BasicMessageInputStream
 */

public class TriggerPlanning extends PlanningMessage {

   /** The recycler used for messages of this type */
   private static Recycler recycler = new Recycler(TriggerPlanning.class);

   /** The message id assigned to messages of this type */
   private static int messageID = -1;
   
   /**
    * Register a prototype of this message class with the
    * BasicMessageInputStream.
    * @return No return value.
    * @see FMG.Conductor.comm.BasicMessageInputStream
    */
   public static void registerSelf() {
      messageID = BasicMessageInputStream.
                        registerMessagePrototype((Message) recycler.getOne());
   }
   
   /**
    * This method returns the message id that was assigned to messages of this
    * type by BasicMessageInputStream.
    * @return The message id for this type of message.
    * @see FMG.Conductor.comm.BasicMessageInputStream
    */
   public int getMessageID() {
      return messageID;
   }

   /**
    * This method returns a new (or recycled) TriggerPlanning object 
    * initialized from the input stream.  Normally, this method is only 
    * called on the prorotype given to BasicMessageInputStream.
    * @param is The input stream to read the message's data from
    * @return a new object initialized from the data stream.
    * @exception java.io.IOException if an error occurs while reading from
    *                                the stream.
    * @see FMG.Conductor.comm.BasicMessageInputStream
    */
   public Message readNewMessage(DataInputStream is) throws IOException {
      return TriggerPlanning.getNewMessage(is);
   }
   
   /**
    * Generate a new (or recycled) TriggerPlanning message with the 
    * appropriate data members initialized.
    * @param connID The id of the connection for which we are planning.
    * @param addr The id of the node sending the message
    * @return A new (or recycled) and initialized TriggerPlanning message.
    */
   public static TriggerPlanning getNewMessage(ConnID connID) {
      TriggerPlanning m = (TriggerPlanning) recycler.getOne();
      m.initMessage(connID);
      return m;
   }

   /**
    * This method returns a new (or recycled) TriggerPlanning object 
    * initialized from the input stream.
    * @param is The input stream to read the message's data from
    * @return A new (or recycled) and initialized TriggerPlanning message.
    * @exception java.io.IOException if an error occurs while reading from
    *                                the stream.
    */
   public static TriggerPlanning getNewMessage(DataInputStream is) 
                                                       throws IOException {
      TriggerPlanning m = (TriggerPlanning) recycler.getOne();
      m.receiveSelf(is);
      return m;
   }

   /** 
    * Determine the direction of motion for this object.
    * @return Always returns <code>Direction.down</code>    
    */
   public Direction flowDir() { 
      return Direction.down; 
   }  
   
   /**
    * Determine whether this object is understood by the Path object.
    * @return Always returns true.
    */
   public boolean handledByPath() {
      return true;
   }

}
