/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.Conductor.planning.topology;

import java.util.*;
import java.lang.*;
import java.lang.reflect.*;
import FMG.Conductor.comm.*;
import FMG.Conductor.Conductor;
import FMG.Conductor.util.*;

/**
 * The PathTopology interface specifies the information that can be retrieved 
 * from pluggable topology modules.  A discovery algorithm could be built
 * on this interface, or simpler static topologies can also be implemented.
 *
 * @author Kevin Eustice (kfe@cs.ucla.edu)
 * @version $Revision: 1.2 $
 */

public class PathTopology {
    
    /** Who we are logically connected to */
    protected Vector neighbors; 
    /** The real source */
    protected NodeID src; 
    /** Real destination */
    protected NodeID dest;
    /** Nodes we report topology for */
    protected Vector forwardingNodes; 

    
    public static PathTopology createAndStart(NodeID sourceAddr, NodeID targetAddr) {
	
	PathTopology pathTopology = null;
	String pathTopologyName = Conductor.getProperty("PathTopologyClass");
	
	
	if (pathTopologyName == null) {
	    pathTopologyName = new String("FMG.Conductor.planning.topology.LinearPathTopology");
	} 
	
	try {
	    
	    Class[] parameterTypes = new Class[2];     //parameter types for 
	    //getConstructor
	    Object[] pathTopologyArgs = new Object[2]; //args to the constructor
	    
	    try {
		
		parameterTypes[0] = Class.forName("FMG.Conductor.comm.NodeID");
		parameterTypes[1] = Class.forName("FMG.Conductor.comm.NodeID");
		
		pathTopologyArgs[0] = sourceAddr;
		pathTopologyArgs[1] = targetAddr;
		
	    } catch (ClassNotFoundException cnfe) { 
		System.out.println("Failure loading FMG.Conductor.comm.NodeID");
		cnfe.printStackTrace();
		System.exit(1);
	    }
	    
	    
	    Class pathTopologyClass = Class.forName(pathTopologyName);
	    Constructor pathTopologyConstructor = pathTopologyClass.getConstructor(parameterTypes);
	    pathTopology = (PathTopology) pathTopologyConstructor.newInstance(pathTopologyArgs);
	    
	} catch (Exception e) {
	    System.out.println("Failure loading Path Topology class " + pathTopologyName);
	    
	    
	    
	}
	
	return pathTopology;
	
    }
    

    /** 
     * Default constructor -- set src and dest 
     **/

    public PathTopology(NodeID source, NodeID destination) {
	if(Debug.on) {
	    Debug.println("Started default PathTopology");
	}
	this.src = source;
	this.dest = destination;
	neighbors = new Vector(2);
	forwardingNodes = new Vector(1);
    }

    /** 
     * Set next node -- adds the specified node to the neighbors list 
     */
    public void addNext(NodeID n) {
	// add n to neighbors
	neighbors.addElement(n);
    }

    /** 
     * Set prev node  -- adds specified node to the neighbors list
     * while also defining it as a node who forwards through the local node.
     */
    public void addPrev(NodeID n) {
	// add n to neighbors
	// add n to forwardingNodes
	neighbors.addElement(n);
	forwardingNodes.addElement(n);
    }

    /** 
     *  This method returns the "server-side" nodes.  effectively,
     *  it returns {neighbors - forwardingNodes }
     */
        public NodeID[] getNext() {
	    
	    int vectorcap = neighbors.size();
	    int forwardingcap = forwardingNodes.size();
	    NodeID[] nodeArray  = new NodeID[vectorcap - forwardingcap];
	    int ctrarray = 0;

	    boolean isForward = false;

		for(int i = 0; i < vectorcap; i++) {
		    if(!forwardingNodes.contains(neighbors.elementAt(i)))
			nodeArray[ctrarray++] = (NodeID) neighbors.elementAt(i);
		    //assert(ctrarray < nodeArray.length);
		}
		return nodeArray;
	} 
	
    /** 
     * Return all forwarding nodes
     */
    public NodeID[] getPrev() { //"client" side
	NodeID nodeArray[] = new NodeID[forwardingNodes.size()];
	forwardingNodes.copyInto(nodeArray);
	return nodeArray;
    }
    
    /** 
     * getPRDest returns the nodes to forward Path Requests towards.  In the
     * linear network case, this is just the server.  In more complex nets
     * this will require some discovery/static configuration of 
     * the topology.   
     */
    public NodeID[] getPRDest() { return null; }


}
    


