/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.Conductor.security;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;

import FMG.Conductor.util.Debug;

/**
 * The BoxConstructor class represents one security model (one line) from 
 * the security configuration file and is capable of creating a security
 * box that implements that model.
 *
 * @author Mark Yarvis (yarvis@fmg.cs.ucla.edu)
 * @version $Revision: 1.4 $
 */

class BoxConstructor {

   /** The name of the class that implements the given security model. */
   String className;

   /** The constructor that can be used to generate the security box. */
   Constructor constructor;

   /** The parameters specified in the configuration file for this model. */
   String params;

  /**
   * Create an object that can construct a security box that implements a
   * security model desired by the user.  The description of a security 
   * model consists of a class name and an optional argument, separated by 
   * a space.
   * @param boxDescrption A string specifying the desired security model.
   *                      The string is either a class name or a class name
   *                      and an argument, separated by a space.
   * @exception java.lang.NoSuchMethodException is thrown if the specified 
   *            class doesn't have the expected constructor.
   * @exception java.lang.ClassNotFoundException is thrown if the specified
   *            class doesn't exist.
   */
   BoxConstructor(String boxDescription) 
                     throws NoSuchMethodException, ClassNotFoundException {
      int index = boxDescription.indexOf(' ');

      if (index == -1) {
         className = boxDescription;
         params = "";
      } else {
         className = boxDescription.substring(0, index);
         params = boxDescription.substring(index+1, boxDescription.length());
      }

      Class[] constrParamTypes = new Class[2];
      constrParamTypes[0] = SecurityCoordinator.class;
      constrParamTypes[1] = String.class;

      Class boxClass = Class.forName(className);
      constructor = boxClass.getConstructor(constrParamTypes);
   }

   /**
    * Create a new SecurityBox instance that implements the specified model.
    * The SecurityBox object will be initialized using the parameter specified
    * in the model description and the handle to the SecurityCoordinator that
    * is provided in this call.
    * @param sc A handle to the security coordinator to pass to the new object
    *           during initilization.
    * @return A new SecurityBox or null, if instantiation fails.
    */
   SecurityBox newInstance(SecurityCoordinator sc) {
      Object[] args = new Object[2];
      args[0] = sc;
      args[1] = params;

      try {
         return (SecurityBox) constructor.newInstance(args);
      } catch (Exception e) {
         // The constructor failed!
         if (Debug.on) {
            Debug.println(Debug.SECURITY, Debug.WARNING, 
                          "Warning: could not instantiate security box: " +
                          className + " (" + e + ")");
            if (e instanceof InvocationTargetException) {
               Debug.println(Debug.SECURITY, Debug.WARNING, 
                             "Exception was: " + 
                             ((InvocationTargetException) e).
                                             getTargetException());
            }
         }
         return null;
      }
   }
}
