/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.Conductor.security;

import FMG.Conductor.planning.messages.PlanningMessage;
import FMG.Conductor.util.Debug;

/**
 * The NullSecurityBox is an security box implementation that provides
 * no actual security.  All incoming and outoing messages are passed
 * along unmodified.
 * <p>
 * The parameter that can be used to customize some security boxes has no
 * affect on teh NullSecurityBox.
 * 
 * @author Mark Yarvis (yarvis@fmg.cs.ucla.edu)
 * @version $Revision: 1.6 $
 */

public class NullSecurityBox extends SecurityBox {

   /**
    * Create a new null security box object.
    * @param coordinator The coordinator for this object.
    * @param param The parameter specified by the user in the config file
    *              used to customize this instance of the class.
    */
   public NullSecurityBox(SecurityCoordinator coordinator, String param) {
      super(coordinator, param);
   }

   /**
    * Handle messages that come in from the network.  In particular, all
    * messages are forwarded as-is to the path object.
    * @param m The message that has been received from the network.
    * @return No return value.
    */
   public void handleIncoming(PlanningMessage m) {
      if (Debug.on) {
         Debug.println(Debug.SECURITY, Debug.MOREINFO, 
                       "NullSecurityBox.handleIncoming(" + m + ")");
      }
      coordinator.deliverToPath(m);
   }

   /**
    * Handle messages that the path object is trying to send over the
    * network toward the server.  In particular, all messages will
    * be transmitted as-is.
    * @param m The message that is to be transmitted.
    * @return False if a failure occurs while sending.  True if it appears
    *         that transmission was successful (not guaranteed).
    */
   public boolean handleOutgoingToServer(PlanningMessage m) {
      if (Debug.on) {
         Debug.println(Debug.SECURITY, Debug.MOREINFO, 
                       "NullSecurityBox.handleOutgoingToServer(" + m + ")");
      }
      return coordinator.sendSToNet(m);
   }

   /**
    * Handle messages that the path object is trying to send over the
    * network toward the client.  In particular, all messages will
    * be transmitted as-is.
    * @param m The message that is to be transmitted.
    * @return False if a failure occurs while sending.  True if it appears
    *         that transmission was successful (not guaranteed).
    */
   public boolean handleOutgoingToClient(PlanningMessage m) {
      if (Debug.on) {
         Debug.println(Debug.SECURITY, Debug.MOREINFO, 
                       "NullSecurityBox.handleOutgoingToClient(" + m + ")");
      }
      return coordinator.sendCToNet(m);
   }

}
