/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.Conductor.security;

import java.lang.reflect.*;
import java.util.*;

import FMG.Conductor.dataflow.Adaptor;
import FMG.Conductor.planning.messages.PlanDistro;
import FMG.Conductor.planning.messages.PlanningMessage;
import FMG.Conductor.comm.NodeID;
import FMG.Conductor.dataflow.IAC;
import FMG.Conductor.util.*;

/**
 * This abstract class is the superclass for all security box.  Each
 * security box filters all incoming and outgoing messages, allowing
 * each message to be authenticated and/or encrypted by replacing and/or
 * adding messages to be sent.
 * <p>
 * Each SecurityBox is associated with a SecurityCoordinator.  The 
 * SecurityCoordinator acts as implementation-independent glue between 
 * each SecurityBox and the Path object.  The particular subclass of 
 * SecurityBox that is desired is loaded by the SecurityCoordinator.
 * Messages as subsequently received from and delivered back to the 
 * SecurityCoordinator.
 * <p>
 * Each SecurityBox can be customized by providing a string parameter in
 * the security configuration file.  This string is passed to the
 * constructor and can be used to customize the instance of the class.
 * For instance, the string might specify the signature algorithm to use.
 * <p>
 * A SecurityBox may also implement key distribution to support security
 * adaptors.  For convenience of implementation, the method 
 * <code>getRequiredSessionKeys()</code> is provided to obtain a list of 
 * keys expected by the deployed adaptors.  If this mechanism is used, each 
 * adaptor that requires a key must implement the following methods:<br>
 * <ul>
 *    <li><code>public static String getKeyLabel()</code><br>
 *        <table width="90%" align=center><tr><td>
 *        This method should return a string which identifies the type
 *        of key required for this adaptor.  A pair of adaptors sharing a key
 *        must return the same key label.  Typically, each pair of adaptors
 *        returns a unique key label.  If multiple pairs of adaptors in a given
 *        session  share the same key label only one key will be generated and 
 *        distributed to each of these adaptors during a given session.
 *        </td></tr></table>
 *    <li><code>public static byte[] generateKey()</code><br>
 *        <table width="90%" align=center><tr><td>
 *        This method should return a key appropriate for use by any adaptor
 *        that returns the same key label.  Each adaptor requiring a key must
 *        implement this method (though the decryptor, for instance, might
 *        simply call the implementation in the encryptor).
 *        </td></tr></table>
 * </ul>
 * A set of keys needed by the adaptors in a given plan are returned
 * by <code>getRequiredSessionKeys()</code> and should be securely 
 * distributed to the specified nodes.  At each node, the session
 * keys should be delivered to the appropriate adaptors by storing them
 * in the IAC object for that stream using the key label as the identifier.
 *
 * @author Mark Yarvis (yarvis@fmg.cs.ucla.edu)
 * @version $Revision: 1.13 $
 *
 * @see FMG.Conductor.security.SecurityCoordinator
 * @see FMG.Conductor.security.SessionKey
 * @see FMG.Conductor.dataflow.IAC
 */

public abstract class SecurityBox {

   /** Our security coordinator. */
   protected SecurityCoordinator coordinator;

   /** Key labels previously obtained for given class names. */
   private Hashtable keyLabelCache = new Hashtable();

   /** Adaptor classes that have been found to not need keys. */
   private Hashlist nonSecurityAdaptorClasses = new Hashlist();

   /**
    * Create a new security box.
    * @param coordinator Our security coordinator.
    * @param param A customization paramater taken from the security 
    *              configuration file.
    */
   public SecurityBox(SecurityCoordinator coordinator, String param) {
      this.coordinator = coordinator;
      if (Debug.on) {
         Debug.println(Debug.SECURITY, Debug.MOREINFO, 
                       "Security box " + this.getClass().getName() + 
                       " instantiated with parameter " + param);
      }
   }

   /**
    * Obtain a handle to the IAC object for this stream.
    * @return An IAC object used to communicate with other elements of this 
    *         stream.
    */
   protected IAC getStreamIAC() {
      return coordinator.getStreamIAC();
   }

   /**
    * Obtain a key label for a given adaptor class name.  Look the class
    * name up in the cache.  If it's present (because it has already
    * been obtained), then use it.  Otherwise, obtain a class object for
    * this adaptor class and ask it for its key label.
    * @param adaptorName The name of the adaptor for which we need the key 
    *                    label.
    * @return A string that is the key label for the named adaptor, or null
    *         if the given adaptor does not implement the required function.
    */
   private String getKeyLabelForAdaptor(String adaptorSpec) {

      /* It this class already known to not be a security adaptor class */
      if (nonSecurityAdaptorClasses.isElement(adaptorSpec)) {
         return null;
      }

      /* Is the key label already known for this class? */
      if (keyLabelCache.containsKey(adaptorSpec)) {
         return (String) keyLabelCache.get(adaptorSpec);
      }

      try {
         // get the class
         Class adaptorClass = Class.forName(
                                Adaptor.getNameFromSpec(adaptorSpec));

         // ask for the key label
         Class[] paramTypes = { String.class };
         Method action = adaptorClass.getMethod("getKeyLabel", paramTypes);
         Object[] args = { Adaptor.getParamsFromSpec(adaptorSpec) };
         String keyLabel = (String) action.invoke(null, args);

         // cache and return the key label
         keyLabelCache.put(adaptorSpec, keyLabel);
         return keyLabel;
      } catch (Exception e) {
         // Couldn't call getKeyLabel(), must not be a security adaptor
         nonSecurityAdaptorClasses.addElement(adaptorSpec);
         return null;
      }
   }

   /**
    * Ask the named adaptor class to generate a key.
    * @param adaptorName The name of the adaptor for which we need a key.
    * @return A byte array that is the key for the named adaptor.
    */
   private byte[] getKeyForAdaptor(String adaptorSpec) {
      try {
         // get the class
         Class adaptorClass = Class.forName(
                                     Adaptor.getNameFromSpec(adaptorSpec));

         // ask for the key
         Class[] paramTypes = { String.class };
         Method action = adaptorClass.getMethod("generateKey", paramTypes);
         Object[] args = { Adaptor.getParamsFromSpec(adaptorSpec) };
         return (byte[]) action.invoke(null, args);
      } catch (Exception e) {
         // Couldn't call generateKey(), must not be a security adaptor
         nonSecurityAdaptorClasses.addElement(adaptorSpec);
         return null;
      }
   }

   /**
    * Examine a plan and obtain the session keys required by the specified 
    * adaptors.  Adaptors that require a session key must implement the
    * methods <code>getKeyLabel()</code> and <code>generateKey()</code> as
    * described above.
    * @param m The PlanDistro message containing the plan.
    * @return An Enumeration of SessionKey objects required for this 
    *         connection.
    * @see FMG.Conductor.security.SessionKey
    */
   protected Enumeration getRequiredSessionKeys(PlanDistro m) {
      Hashtable keyHash = new Hashtable();

      // Loop once for dir = up, and once for dir = down
      for (int i = 0; i < 2; i++) {
         Direction dir = (i==0 ? Direction.up : Direction.down);

         // Loop over all of the nodes in this direction
         Enumeration nodes = m.getNodes(dir); 
         while (nodes.hasMoreElements()) {
            NodeID thisNode = (NodeID) nodes.nextElement();

            // Loop over all adaptors for this node
            Enumeration adaptors = m.getAdaptors(thisNode,dir);
            while (adaptors.hasMoreElements()) {
               String adaptorSpec = (String) adaptors.nextElement();

               // what kind of key does this adaptor need?
               String keyLabel = getKeyLabelForAdaptor(adaptorSpec);

               if (keyLabel != null) {

                  if (keyHash.containsKey(keyLabel)) {
                     // add this node to the list that needs this key
                     SessionKey sk = (SessionKey) keyHash.get(keyLabel);
                     sk.addNode(thisNode);
                  } else {
                     // create a new key and store it
                     byte[] key = getKeyForAdaptor(adaptorSpec);
                     if (key != null) {
                        SessionKey sk = new SessionKey(keyLabel, 
                                                       thisNode, key);
                        keyHash.put(keyLabel, sk);
                     }
                  }

               }
            }
         }
      }

      // package up the SessionKey objects
      return keyHash.elements();
   }

   /**
    * Handle messages that have entered the node from the net.  Presumably,
    * some security-related messages will be received in addition to or
    * instead of the normal planning messages.  The security box should
    * turn these messages into normal planning messages and deliver them
    * to the path object by calling 
    * <code>SecurityCoordinator.deliverToPath()</code>.
    * Note that zero, one, or several messages can be delvered to the path
    * object for each call to this method.
    * @param m The message that has been received from the network.
    * @return No return value.
    */
   public abstract void handleIncoming(PlanningMessage m);

   /**
    * Handle messages that a Path object wants to send over the network
    * toward the server.  The message passed in the argument will be a
    * standard plannign message.  The security box will typically supplement
    * or replace this message with security messages containing authentication
    * information or encyrypted versions of data.  The security box can make
    * zero, one or several calls to 
    * <code>SecurityCoordinator.sendSToNet()</code> to deliver messages
    * to the network for each call to this method.
    * @param m The message to send over the network.
    * @return Returns false if message delivery failed, true if message
    *         delivery seemed to work (no guarantee).
    */
   public abstract boolean handleOutgoingToServer(PlanningMessage m);

   /**
    * Handle messages that a Path object wants to send over the network
    * toward the client.  The message passed in the argument will be a
    * standard plannign message.  The security box will typically supplement
    * or replace this message with security messages containing authentication
    * information or encyrypted versions of data.  The security box can make
    * zero, one or several calls to 
    * <code>SecurityCoordinator.sendCToNet()</code> to deliver messages
    * to the network for each call to this method.
    * @param m The message to send over the network.
    * @return Returns false if message delivery failed, true if message
    *         delivery seemed to work (no guarantee).
    */
   public abstract boolean handleOutgoingToClient(PlanningMessage m);

}
