/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.Conductor.security;

import java.io.*;
import java.net.*;
import java.lang.reflect.*;
import java.util.Hashtable;

import FMG.Conductor.planning.messages.PlanningMessage;
import FMG.Conductor.planning.Path;
import FMG.Conductor.Conductor;
import FMG.Conductor.util.Debug;
import FMG.Conductor.dataflow.Session;
import FMG.Conductor.dataflow.IAC;
import FMG.Conductor.comm.*;

/**
 * The SecurityCoordinator acts as glue between a Path object and a 
 * SecurityBox object.  In addition, the SecurityCoordinator selects
 * and instantiates an appropraite SecurityBox implementation based
 * on user preferences and the method selected by the client node.
 * <p>
 * The SecurityCoordinator accepts messages that have been received
 * from the network and forwards them to the SecurityBox and provides
 * the methods which allow the SecurityBox to deliver one or more
 * resulting messages to the Path object.
 * <p>
 * The SecurityCoordinator also accepts messages from the Path object that
 * are intended to be transmitted across the network, gives them to
 * the SecurityBox, and provides the SecurityBox with the methods required
 * to effect transmission of one or more resulting messages.
 * <p>
 * The name of the configuration file for this class is specified in the
 * property "SecurityConfig".  The value for this property can contain
 * the string "${HOSTNAME}" which will be replaced by the name of the actual
 * host.  The configuration file specifies the security models that are 
 * acceptable to the user.  Each line of the file describes a security
 * model that is acceptable to the user.  Each escurity model consists of
 * a class name (which should be a subclass of SecurityBox) followed by an
 * optional argument (separated by a space).  If present the argument is 
 * passed to the class at initialization and is intended to specialize 
 * the actions of that class (for example, the argument might specify a 
 * signature algorithm to use).  The first line of the configuration file 
 * specifies the default security model to be used for connections initiated 
 * by the local user (ie, at the client).  Other nodes involved in the 
 * connection will use the model specified by the client node so long as that
 * model appears somewhere in the configuration file.  An exact match must 
 * occur in both the class name and the argument.
 * <p>
 * TODO: The reliability level for the SecuritySelector is very low.  We
 *       simply expect that if SendSToNet returns true, then it was received.
 *       This should be fixed.
 *
 * @author Mark Yarvis (yarvis@fmg.cs.ucla.edu)
 * @version $Revision: 1.14 $
 */

public final class SecurityCoordinator {

   /** Should a security module be used during planning? The default is no. */
   private static boolean securityEnabled = 
                     Conductor.getBooleanProperty("UseSecurePlanning", false);

   /** 
    * The name of the security configuration file.  The default name is 
    * "./security.cfg"
    */
   private static final String CONFIG_FILE = configFileName();

   /**
    * The class name of the default security box to use for new connections.
    */
   private static String defaultBoxDescription;

   /**
    * A hashed list of BoxConstructor objects of all allowable 
    * types indexed by a box description (from the config file).
    */
   private static Hashtable boxConstructors = new Hashtable();

   /** The path object with which this SecurityCoordinator is assocaited.  */
   private Path path;

   /** The object the holds the parts of this connection */
   private Session session;

   /** The address of the source node. */
   private NodeID sourceAddr;

   /** The source port number. */
   private int sourcePort;

   /** The address of the target node. */
   private NodeID targetAddr;

   /** The target port number. */
   private int targetPort;

   /** The security box for this connection, if one is instantiated. */
   private SecurityBox box;

   /** A selector that has yet to be successfully sent to our neighbor. */
   private SecuritySelector selector;

   /** Is this the root node? */
   private boolean isRoot;

   /** Is this the planner node? */
   private boolean isPlanner;

   /**
    * Read the configuration file and record the name of the default
    * SecurityBox class as well as constructors for all permissible
    * SecurityBox classes.
    */
   static {
      Class[] paramTypes = new Class[1];
      paramTypes[0]=SecurityCoordinator.class;

      if (securityEnabled) {
         try {
            BufferedReader config = new BufferedReader(
                                     new FileReader(CONFIG_FILE));
            for (String line = config.readLine(); line != null; 
                 line = config.readLine()) {
               try {
                  if (! line.startsWith("#")) {  // skip comments
                     /* 
                      * Create a BoxConstructor object and record it for 
                      * future use.
                      */
                     boxConstructors.put(line, new BoxConstructor(line));

                     /* Record the first one as the default */
                     if (defaultBoxDescription == null) {
                        defaultBoxDescription = line;
                     }
                  }
               } catch (Throwable e) {
                  Debug.println("Warning: Error loading SecurityBox for '" + 
                                line + "'");

                  // Help writers of security boxes find errors!
                  if (Debug.on && (Debug.getLevel(Debug.SECURITY) > 
                                                    Debug.MOREINFO)) {
                     Debug.println("The above error was caused by:");

                     if (e instanceof ExceptionInInitializerError) {
                        ExceptionInInitializerError ee = 
                                        (ExceptionInInitializerError) e;
                        ee.getException().printStackTrace();
                     } else {
                        e.printStackTrace();
                     }
                  }

                  /* 
                   * If there's no default yet, warn the user.  He thinks
                   * the first one was going to be the default.
                   */
                  if (defaultBoxDescription ==  null) {
                     Debug.println("         '" + line + 
                                   "' can't be made default!!!");
                  }
               }
            }

         } catch (IOException e) {
            if (Debug.on) {
               Debug.println(Debug.WARNING,
                             "Warning: Error reading security config file (" + 
                             CONFIG_FILE + ")");
            }
            /* security will be disabled below */
         }

         /*
          * If there's no default, then no boxes were found, so disable 
          * security.
          */
         if (defaultBoxDescription == null) {
            if (Debug.on) {
               Debug.println(Debug.WARNING,
                             "Warning: No security boxes found.  " + 
                             "Security disabled!");
            }
            securityEnabled = false;
         }
      }
   }

   /**
    * Create a new SecurityCoordinator associated with a particular path.
    * @param path The path object with which we communicate.
    * @param session The object that holds the parts of this connection
    * @param isRoot Is this the root node?
    */
   public SecurityCoordinator(Session session,
                              NodeID sourceAddr, int sourcePort, 
                              NodeID targetAddr, int targetPort) {
      this.session = session;
      this.sourceAddr = sourceAddr.duplicate();
      this.targetAddr = targetAddr.duplicate();
      this.sourcePort = sourcePort;
      this.targetPort = targetPort;

      path = session.getPath();
      isRoot = path.isRootNode();
      isPlanner = path.isPlannerNode();

      if (isRoot && securityEnabled) {
         box = createSecurityBox(defaultBoxDescription);

         // Try to tell our neighbor what kind of security box to use
         selector = SecuritySelector.getNewMessage(session.getConnID(), 
                                                   defaultBoxDescription);

         // The selector will be sent in sendSToNet before the first message 
         // is sent.
      }
   }

   /**
    * Obtain the identifier for this connection.  This identifier is 
    * frequently needed in planning messages.
    * @return a handle to a ConnID object for this connection.
    */
   public ConnID connectionID() {
      return session.getConnID();
   }

   /**
    * Obtain the iac for this session/stream.
    * @return The IAC object for communicating with adaptors in this stream.
    */
   IAC getStreamIAC() {
      return session.getIAC();
   }

   /**
    * Obtain the node id for the node that is the source of the connection.  
    * Typically, this node is called the "client."
    * @return A handle to a node id object containing the identity of the 
    *         source node.
    */
   public NodeID sourceNode() {
      return sourceAddr;
   }

   /**
    * Obtain the port number used at the source (i.e., client) endpoint of the 
    * connection.
    * @return The desired port number.
    */
   public int sourcePort() {
      return sourcePort;
   }

   /**
    * Obtain the node id for the node that is the target of the connection.  
    * Typically, this node is called the "server."
    * @return A handle to a node id object containing the identity of the 
    *         target node.
    */
   public NodeID targetNode() {
      return targetAddr;
   }

   /**
    * Obtain the port number used at the target (i.e., server) endpoint of the 
    * connection.
    * @return The desired port number.
    */
   public int targetPort() {
      return targetPort;
   }

   /**
    * Determine if this is the root node of the connection.  This node is the
    * point at which information gathering begins.
    * @return True iff this is the root node.
    */
   public boolean isRootNode() {
      return isRoot;
   }

   /**
    * Determine if this is the planner node for the connection.  This node is 
    * the node to which planning information flows and where a plan is
    * generated.
    * @return True iff this is the planner node.
    */
   public boolean isPlannerNode() {
      return isPlanner;
   }

   /**
    * Initialize Conductor security.  In particular, register the message
    * types used by the SecurityCoordinator.
    * @return No return value.
    */
   public static void initialize() {
      SecuritySelector.registerSelf();
   }

   /**
    * Obtain the configuration file name.  The base name is obtained from
    * the SecurityConfig property in the properties file.  If the property is
    * not found, the default value of "./security.cfg" is used.  If the
    * base name contains "${HOSTNAME}", this string is replaced with the
    * hostname of the local host.
    * @return The name of the security configuration file as specified by
    *         the user or the default value.
    */
   private static String configFileName() {
      return Conductor.getFilenameProperty("SecurityConfig", 
                                           "./config/security.cfg");
   }

   /**
    * Create a new security box from the given specifier (a class name and 
    * argument).  If this SecurityBox type is among those allowed by the 
    * config file, one is created using the previously created BoxConstructor 
    * object.  If the specified type of SecurityBox is not allowed, null is 
    * returned.
    * @param boxDescription A specifier for the type of security box desired 
    *                       (the class name).
    * @return A new Security box of the specified type if this type is allowed
    *         by the user, or null otherwise.
    */
   private SecurityBox createSecurityBox(String boxDescription) {
         BoxConstructor constructor = (BoxConstructor) 
                                           boxConstructors.get(boxDescription);
         if (constructor == null) {
            /* Wasn't listed in the config file! */
            return null;
         } else {
            return constructor.newInstance(this);
         }
   }

   /**
    * Is the specified security model one of the ones allowed by the
    * user (ie, specified in the config file)?
    * @param boxDescription A specifier for the type of security box desired 
    *                       (the class name).
    * @return True iff the specified security box type is acceptable to the 
    *         user.
    */
   private boolean allowedBox(String boxDescription) {
      return boxConstructors.containsKey(boxDescription);
   }

   /**
    * This method is called by the Path object for messages received from
    * the network.  If security is disabled these messages are delivered
    * back to the Path object for processing.  If security is enabled,
    * the messages are delivered to the security box.  This method also
    * handles the SecuritySelector object which causes an appropraite
    * security box to be instantiated if allowed by the config file.
    * @param m The mssage received from the network.
    * @return No return value.
    */
   public void handleIncoming(PlanningMessage m) {
      if (securityEnabled) {

         if (m instanceof SecuritySelector) {

            if (box == null) {
               // Try to instantiate a new box

               String requestedBoxDescription = 
                               ((SecuritySelector) m).boxDescription;

               // Is the requested box allowed?
               if (allowedBox(requestedBoxDescription)) {
                  if (Debug.on) {
                     Debug.println(Debug.SECURITY, Debug.MOREINFO,
                                   "Instantiating security box: " + 
                                   requestedBoxDescription);
                  }
                  // instantiate the new box
                  box = createSecurityBox(requestedBoxDescription);
               } else {
                  if (Debug.on) {
                     Debug.println(Debug.WARNING, 
                                   "Warning: The requested security box (" + 
                                   requestedBoxDescription + ") was denied");
                  }
               }

               if (! isPlanner) {
                  // Forward to the next node
                  if (! path.sendSToNet(m)) {
                     // If failure occurs, try again later
                     selector = (SecuritySelector) m;
                  }
              } else {
                 m.recycle();
              }
            } else {
               // we already got a selector previously, throw this one away
               m.recycle();
            }
         } else {
            if (box != null) {
               // normal case, security is in place, pass message to box
               box.handleIncoming(m);

            } else {
               /*
                * There is no security box.  This only happens if no box
                * has been requested (perhaps security is disabled on the 
                * client) or if the requested box has been rejected.
                */
               deliverOrForward(m);
            }
         }

      } else {     // Security is disabled!
         deliverOrForward(m);
      }
   }

   /**
    * Handle the case in which a message is received, but there is no
    * security in place.  Such a message should either be delivered to
    * the path object or, if the message is security-related, forwarded
    * to the next node.
    * <p>
    * TODO: When forwarding messages, there is no reliability mechanism!
    *       If sendSToNet() fails, we will never detect it!  But, this 
    *       should only happen when some nodes are using security and others 
    *       are not.
    * @param m The message that has been received.
    * @return No return value.
    */
   private void deliverOrForward(PlanningMessage m) {
      if (m.handledByPath()) {   // Can we deliver to the Path?
         path.handleMessage(m);
      } else {
         // Can't deliver to path, so forward to the next node
         if (m.flowDir().isUp()) {
            // use local function in case we haven't sent the selector
            sendSToNet(m);
         } else {
            path.sendCToNet(m);
         }
      }
   }

   /**
    * This method is used by implementations of SecurityBox to delver messages
    * to the Path object for processing.
    * @param m The message to deliver to the path object.
    * @return No return value.
    */
   public void deliverToPath(PlanningMessage m) {
      path.handleMessage(m);
   }

   /**
    * This method is called by the Path object to send a message over the
    * network in the direction of the server.  If security is disabled,
    * the message is delivered back to the Path object for transmission.
    * If security is enabled, the message is delivered to the security box.
    * @param m The message that the Path object wants to transmit.
    * @return False if an error occurs during tranmission.  True if no error
    *         occurs (but delivery is not guaranteed).
    */
   public boolean handleOutgoingToServer(PlanningMessage m) {
      if (securityEnabled && (box != null)) {
         return box.handleOutgoingToServer(m);
      } else {
         // use local function in case we haven't yet sent the selector
         return sendSToNet(m);
      }
   }

   /**
    * This method is used by the SecurityBox to cause actual transmission
    * of messages, across the net, toward the server.  Transmission is 
    * accomplished by passing the message to the Path object for transmission.
    * @param m The message to transmit.
    * @return False if an error occurs during tranmission.  True if no error
    *         occurs (but delivery is not guaranteed).
    */
   public boolean sendSToNet(PlanningMessage m) {
      // If the selector has not been successfully sent yet, send it first
      if (selector != null) {
         if (path.sendSToNet(selector)) {
            selector = null;
         } else {
            return false;
         }
      }

      // Ask the path object to send the message
      return path.sendSToNet(m);
   }

   /**
    * This method is called by the Path object to send a message over the
    * network in the direction of the client.  If security is disabled,
    * the message is delivered back to the Path object for transmission.
    * If security is enabled, the message is delivered to the security box.
    * @param m The message that the Path object wants to transmit.
    * @return False if an error occurs during tranmission.  True if no error
    *         occurs (but delivery is not guaranteed).
    */
   public boolean handleOutgoingToClient(PlanningMessage m) {
      if (securityEnabled && (box != null)) {
         return box.handleOutgoingToClient(m);
      } else {
         return path.sendCToNet(m);
      }
   }

   /**
    * This method is used by the SecurityBox to cause actual transmission
    * of messages, across the net, toward the client.  Transmission is 
    * accomplished by passing the message to the Path object for transmission.
    * @param m The message to transmit.
    * @return False if an error occurs during tranmission.  True if no error
    *         occurs (but delivery is not guaranteed).
    */
   public boolean sendCToNet(PlanningMessage m) {
      return path.sendCToNet(m);
   }

}
