/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.Conductor.security;

import java.util.*;

import FMG.Conductor.comm.NodeID;

/**
 * The SessionKey object represents a single encryption key that needs to
 * be distributed to adaptors on multiple nodes.  This object contains
 * the key, a key label, and a list of nodes that require that key.
 * The key should be securely distributed to each of the listed nodes.
 * On each node, the key should be delivered to the adaptors by placing
 * it in the IAC object for that stream, using the label as the identifier.
 *
 * @author Mark Yarvis (yarvis@fmg.cs.ucla.edu)
 * @version $Revision: 1.2 $ $Date: 2001/10/17 16:45:39 $
 *
 * @see FMG.Conductor.security.SecurityBox
 * @see FMG.Conductor.dataflow.IAC
 */

public class SessionKey {

   /** 
    * The key label supplied by the adaptor that generated this key.
    * The key label allows all adaptors interested in this key to identify
    * it in the IAC table.
    */
   protected String label;

   /** An encryption key generated by the adaptor */
   protected byte[] key;

   /** A list of nodes that need this key */
   protected Vector nodes = new Vector();

   /**
    * Create a new session key.  Initially, it only needs to be distributed
    * to one node.  Additional nodes can be added later.
    * @param label The key label.
    * @param node A node that needs this key.
    * @param key The key.
    */
   SessionKey(String label, NodeID node, byte[] key) {
      this.label = label;
      nodes.addElement(node);
      this.key = key;
   }

   /**
    * Add an additional node that needs this key.  If this node already
    * exists in the list of nodes, no action is taken.
    * @param node An additional node that needs this key.
    * @return No return value.
    */
   void addNode(NodeID node) {
      // Add a node only once
      if (! nodes.contains(node)) {
         nodes.addElement(node);
      }
   }

   /**
    * Obtain the key.
    * @return A handle for the key.
    */
   public byte[] getKey() {
      return key;
   }

   /**
    * Obtain the key label.
    * @return A handle for the key label.
    */
   public String getKeyLabel() {
      return label;
   }

   /**
    * Obtain a list of nodes that need the key.
    * @return An enumeration of NodeID objects representing the nodes that 
    * need the key.
    * @see FMG.Conductor.comm.NodeID
    */
   public Enumeration getNodes() {
      return nodes.elements();
   }

   /**
    * Generate a human readable version of this object.  For debugging use
    * only!
    * @return A string representing this object in human readable form.
    */
   public String toString() {
      StringBuffer sb = new StringBuffer();
      sb.append("[SessionKey: ");
      sb.append(label);
      sb.append(", (");
      for (int i=0; i<nodes.size(); i++) {
         sb.append(nodes.elementAt(i).toString());
         if (i < nodes.size()-1) {
            sb.append(", ");
         }
      }
      sb.append(")]");

      return sb.toString();
   }

}
