/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.Conductor.util;

/**
 * This class provides a variety of useful operations for bytes and byte 
 * arrays.
 *
 * @author Mark Yarvis (yarvis@fmg.cs.ucla.edu)
 * @version $Revision: 1.8 $
 */

public abstract class Bytes {

   /**
    * Determine whether the given byte is whitespace.
    * @param b The byte in question.
    * @return True, iff the byte is a space, tab, newline, or return.
    */
   public static boolean isWhitespace(byte b) {
      return ((b == ' ') || (b == '\t') || (b == '\r') || (b == '\n'));
   }     

   /**
    * Perform a forward search of a byte array for whitespace (space or tab)
    * from the given starting position.
    * @param buf The array to search.
    * @param startPos The index of the first byte to test.
    * @return The position of the first whitespace at or after the given 
    *         starting position, or -1 if no whitespace is found.
    */
   public static int findWhitespace(byte[] buf, int startPos) {
      if (startPos == -1) {
         return -1;
      }

      int pos = startPos;
      while (pos < buf.length) {
         if (isWhitespace(buf[pos])) {
            return pos;
         }
         pos++;
      }
      return -1;
   }

   /**
    * Perform a forward search of a byte array for non-whitespace (anything
    * except space or tab) from the given starting position.
    * @param buf The array to search.
    * @param startPos The index of the first byte to test.
    * @return The position of the first whitespace at or after the given 
    *         starting position, or -1 if no non-whitespace is found.
    */
   public static int findNonWhitespace(byte[] buf, int startPos) {
      if (startPos == -1) {
         return -1;
      }

      int pos = startPos;
      while (pos < buf.length) {
         if (! isWhitespace(buf[pos])) {
            return pos;
         }
         pos++;
      }
      return -1;
   }

   /**
    * Performa reverse search of a byte array for whitespace (space or tab)
    * from the given starting position.
    * @param buf The array to search.
    * @param startPos The index of the first byte to test.
    * @return The position of the first whitespace at or before the given 
    *         starting position, or -1 if no whitespace is found.
    */
   public static int revfindWhitespace(byte[] buf, int startPos) {
      if (startPos == -1) {
         return -1;
      }

      int pos = startPos;
      while (pos > 0) {
         if (isWhitespace(buf[pos])) {
            return pos;
         }
         pos--;
      }
      return -1;
   }

   /**
    * Perform a reverse search of a byte array for non-whitespace (anything
    * except space or tab) from the given starting position.
    * @param buf The array to search.
    * @param startPos The index of the first byte to test.
    * @return The position of the first whitespace at or before the given 
    *         starting position, or -1 if no non-whitespace is found.
    */
   public static int revfindNonWhitespace(byte[] buf, int startPos) {
      if (startPos == -1) {
         return -1;
      }

      int pos = startPos;
      while (pos > 0) {
         if (! isWhitespace(buf[pos])) {
            return pos;
         }
         pos--;
      }
      return -1;
   }

   /**
    * Find the position of the first instance of a given byte in an array 
    * of bytes.
    * @param bytes The array to search.
    * @param b The byte to locate in the array.
    * @return The position of the first instance of the given byte, or -1 if 
    *         it does not appear in the byte array.
    */
   public static int findByte(byte[] bytes, int b) {
      return findByte(bytes, b, 0);
   }

   /**
    * Find the position at or after the given starting position of the 
    * first instance of a given byte in an array of bytes.
    * @param bytes The array to search.
    * @param b The byte to locate in the array.
    * @param startPos The index of the first array element to test.
    * @return The position of the first instance of the given byte at or
    *         after the given starting position, or -1 if no such byte
    *         is present.
    */
   public static int findByte(byte[] bytes, int b, int startPos) {
      if (startPos == -1) {
         return -1;
      }

      byte target = (byte) b;
      for (int i=startPos; i<bytes.length; i++) {
         if (bytes[i] == target) {
            return i;
         }
      }
      return -1;
   }

   /**
    * Find the position of the last instance of a given byte in an array 
    * of bytes.
    * @param bytes The array to search.
    * @param b The byte to locate in the array.
    * @return The position of the last instance of the given byte, or -1 if 
    *         it does not appear in the byte array.
    */
   public static int revfindByte(byte[] bytes, int b) {
      return revfindByte(bytes, b, bytes.length-1);
   }

   /**
    * Find the position of the last instance of a given byte that is also
    * at or prior to the given index.
    * @param bytes The array to search.
    * @param b The byte to locate in the array.
    * @param startPos The upper bound where the search should start.
    * @return The index of the desired byte, or -1 if the byte is not
    *                 present.
    */
   public static int revfindByte(byte[] bytes, int b, int startPos) {
      if (startPos == -1) {
         return -1;
      }

      byte target = (byte) b;
      for (int i=startPos; i>=0; i--) {
         if (bytes[i] == target) {
            return i;
         }
      }
      return -1;
   }

   /**
    * Obtain a lower-case version of a byte.
    * @param b The byte to convert to lower-case.
    * @return If the input is an upper-case letter, the lower case equivalent
    *         is returned.  Otherwise, the input byte is returned.
    */
   public static byte toLowerCase(byte b) {
      if ((b >= 'A') && (b <= 'Z')) {
         b += (byte) 'a' - (byte) 'A';
         return b;
      } else {
         return b; 
      }
   }  

   /**
    * Determine if the specified portions of two byte arrays are identical.
    * @param buf1 The first array to compare.
    * @param start1 The starting index in the first array for comparsion.
    * @param buf2 The second array to compare.
    * @param start2 The starting index in the second array for comparsion.
    * @param len The number of bytes to compare.
    * @return True iff the specified portion of each array are identical.
    *         False is also returned if an array insufficienly long to perform
    *         the comparison.
    */
   public static boolean match(byte[] buf1, int start1, 
                               byte[] buf2, int start2,
                               int len) {
      if ((start1 + len > buf1.length) ||
          (start2 + len > buf2.length)) {
         return false;
      }

      for (int i=0; i<len; i++) {
         if (buf1[i+start1] != buf2[i+start2]) {
            return false;
         }
      }

      return true;
   }

   /**
    * Determine if the specified portions of two byte arrays are identical
    * except for differences in case.
    * @param buf1 The first array to compare.
    * @param start1 The starting index in the first array for comparsion.
    * @param buf2 The second array to compare.
    * @param start2 The starting index in the second array for comparsion.
    * @param len The number of bytes to compare.
    * @return True iff the specified portion of each array are identical
    *         except for differences in case.
    *         False is also returned if an array insufficienly long to perform
    *         the comparison.
    */
   public static boolean matchIgnoreCase(byte[] buf1, int start1, 
                                         byte[] buf2, int start2,
                                         int len) {
      if ((start1 + len > buf1.length) ||
          (start2 + len > buf2.length)) {
         return false;
      }

      for (int i=0; i<len; i++) {
         if (toLowerCase(buf1[i+start1]) != toLowerCase(buf2[i+start2])) {
            return false;
         }
      }

      return true;
   }

   /**
    * Create a new string that contains the bytes from a byte array.
    * @param buf The byte array to turn into a string.
    * @return A new String object with the contents of the input byte array.
    */
   public static String toString(byte[] buf) {
      return toString(buf, 0, buf.length);
   }

   /**
    * Create a new string that contains a portion of the bytes from a 
    * byte array.
    * @param buf The byte array to turn into a string.
    * @param start The index of the first desired byte.
    * @param len The number of desired bytes.
    * @return A new String object with a portion of the contents of the 
    *         input byte array.
    */
   public static String toString(byte[] buf, int start, int len) {
      StringBuffer sb = new StringBuffer(len);
      for (int i=start; i < start + len; i++) {
         sb.append((char) buf[i]);
      }
      return sb.toString();
   }

   /**
    * Determine if the specified portions of the two given byte arrays are
    * equal.
    * @param buf1 The first arry to test.
    * @param start1 The desired starting position in the first array.
    * @param buf2 The second arry to test.
    * @param start2 The desired starting position in the second array.
    * @param len The number of bytes to test.
    * @return No return value.
    */
   public static boolean areEqual(byte[] buf1, int start1, 
                                  byte[] buf2, int start2, int len) {
      for (int i = 0; i < len; i++) {
         if (buf1[i+start1] != buf2[i+start2]) {
            return false;
         }
      }

      return true;
   }

}
