/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.Conductor.util;

import java.util.Hashtable;

/**
 * The Debug class is intended (surprise, surprise) to support debugging of
 * the system.  The system will, at any moment, be at a given debugging 
 * "level".  In addition, each "module" can also have its own debuggin 
 * "level."  A module is simply defined as a unique text string passed to
 * the debug methods.  Several constants are provided for large pieces of the
 * Conductor code.  Arbitrary modules can also be specified for any 
 * granularity of debugging.  A number of constant values are specified 
 * which determin what level of verbosity is needed before triggering debug 
 * output on particular types of events.  Methods are then provided to help 
 * create the debugging output.
 *
 * @author Mark Yarvis (yarvis@fmg.cs.ucla.edu)
 * @version $Revision: 1.28 $
 */

public final class Debug {
   /**
    * Is debugging enabled (is it compiled in)?
    */
   public static final boolean on = true;

   /**
    * The current global debug level of the system.
    */
   protected static int globalLevel;

   /** Debug levels for particular modules.  */
   protected static Hashtable levels = new Hashtable();

   /*
    * Module names.
    */

   /** The name of the planning module */
   public static final String PLANNING="Planning";

   /** The name of the dataflow module */
   public static final String DATAFLOW="Dataflow";

   /** The name to use for reliability-related events */
   public static final String RELIABILITY="Reliability";

   /** The name of the dataflow module */
   public static final String DATA="Data";

   /** The name of the recycling module */
   public static final String RECYCLER="Recycler";

   /** The name of the module for adaptors */
   public static final String ADAPTOR="Adaptor";

   /** The name of the security module */
   public static final String SECURITY="Security";

   /** The name to use for socket-related operations */
   public static final String SOCKET="Socket";

   /*
    * Debug levels.
    */

   /** The level above which sanity checks (ie, assert) are preformed */
   public static final int SANITY_CHECK = 0;

   /** The level for printing basic debugging info */
   public static final int INFO = 1;

   /** The level for printing slightly more verbose debugging info */
   public static final int MOREINFO = 2;

   /** The level for printing even more verbose debugging info */
   public static final int EXTRAINFO = 3;

   /** The level for printing detailed debugging info */
   public static final int DETAIL = 4;

   /** The level above which warning messages are produced */
   public static final int WARNING = 0;
   
   /**
    * Sets the default debugging level of the system.
    * @param newLevel The new debugging level.
    * @return No return value
    */
   public static void setLevel(int newLevel) {
      globalLevel = newLevel;
   }

   /**
    * Sets the debugging level of the system for a particular module.
    * @param module The name of the module.
    * @param newLevel The new debugging level.
    * @return No return value
    */
   public static void setLevel(String module, int newLevel) {
      levels.put(module, new Integer(newLevel));
   }

   /**
    * Obtain the global debugging level of Conductor.
    * @return The global debuggin level of Conductor.
    */
   public static final int getLevel() {
      return globalLevel;
   }

   /**
    * Obtain the debugging level of a particular module.
    * @param module The name of the module.
    * @return The debuggin level of the named module.
    */
   public static final int getLevel(String module) {
      if (module == null) {
         return globalLevel;
      }

      Integer ret = (Integer) levels.get(module);
      if (ret == null) {
         return globalLevel;
      } else {
         return ret.intValue();
      }
   }

   /**
    * Print the given debugging message if the treshold of the given module
    * exceeds the given level.
    * @param module The module in question.
    * @param threshold The threshold over which to print the message.
    * @param msg The mssage to print.
    */
   public static final void print(String module, int threshold, String msg) {
      print(module, threshold, msg, true, false);
   }

   /**
    * Print the message if the global level is above the threshold.
    * @param threshold The verbosity threshold above which to print messages.
    * @param msg The string to print.
    * @return No return value
    */
   public static final void print(int threshold, String msg) {
      print(null, threshold, msg, true, false);
   }

   /**
    * Unconditionally prints the given string without adding a newline.
    * @param msg The string to print.
    * @return No return value
    */
   public static final void print(String msg) {
      System.out.print(msg);
      System.out.flush();
   }

   /**
    * Print the given debugging message if the treshold of the given module
    * exceeds the given level.  A newline is printed after the message.
    * @param module The module in question.
    * @param threshold The threshold over which to print the message.
    * @param msg The mssage to print.
    * @return No return value.
    */
   public static final void println(String module, int threshold, String msg) {
      print(module, threshold, msg, true, false);
   }

   /**
    * Print the message and a newline if the current level is above the 
    * threshold.
    * @param threshold The verbosity threshold above which to print messages.
    * @param msg The string to print.
    * @return No return value
    */
   public static final void println(int threshold, String msg) {
      print(null, threshold, msg, true, false);
   }

   /**
    * Prints the given string with a newline at the end.
    * @param msg The string to print.
    * @return No return value
    */
   public static void println(String msg) {
      System.out.println(msg);
   }

   /**
    * Print the given debugging message if the treshold of the given module
    * exceeds the given level.  A newline is optionally printed after the 
    * message.  A timestamp can optionally preceed the message.  If 
    * <code>null</code> is passed for the module, the global threshold is 
    * used.
    * @param module The module in question.
    * @param threshold The threshold over which to print the message.
    * @param msg The mssage to print.
    * @param newline Should a newline be printed?
    * @param timestamp Should a timestamp be printed?
    * @return No return value.
    */
   public static final void print(String module, int threshold, 
                                  String msg,
                                  boolean newline, boolean timestamp) {
      if (getLevel(module) > threshold) {
         print(msg, newline, timestamp);
      }
   }

   /**
    * Unconditionally print the given debugging message.
    * A newline is optionally printed after the 
    * message.  A timestamp can optionally preceed the message.
    * @param msg The mssage to print.
    * @param newline Should a newline be printed?
    * @param timestamp Should a timestamp be printed?
    * @return No return value.
    */
   public static final void print(String msg,
                                  boolean newline, boolean timestamp) {
      if (timestamp) {
         System.out.print("%% ");
         System.out.print(System.currentTimeMillis());
         System.out.print(": ");
      }
      if (newline) {
         System.out.println(msg);
      } else {
         System.out.print(msg);
         System.out.flush();
      }
   }

   /**
    * Creates a string, containint only printable characters, from the
    * given byte array.  Non-printable characters will be mapped to a period.
    * @param bytes The source byte array.
    * @return A new string object representing the entire byte array.
    */
   public static String printableString(byte[] bytes) {
      return printableString(bytes, 0, bytes.length - 1);
   }

   /**
    * Creates a string, containing only printable characters, from the
    * specified part of the given byte array.  Non-printable characters 
    * will be mapped to a period.
    * @param bytes The source byte array.
    * @param first The position of the first byte to copy.
    * @param last The position of the last byte to copy.
    * @return A new string object representing the specified segment of the
    *         input byte array.
    */
   public static String printableString(byte[] bytes, int first, int last) {
      StringBuffer sb = new StringBuffer(last - first + 1);
      for (int i=first; i<=last; i++) {
         sb.append(printableChar(bytes[i]));
      }
      return sb.toString();
   }

   /**
    * Returns a printable character representing given byte.  If the input 
    * byte represents a non-pritable character, a "period" character is 
    * returned.  Otherwise, the byte is cast to a character.
    * @param b The byte to make printable.
    * @return A character representation of the input byte.
    */
   public static char printableChar(byte b) {
      if ((b >= 32) && (b <= 126)) {
         return (char) b;
      } else {
         return '.';
      }
   }

   public static int percentChange(int origSize, int newSize) {
      if (origSize == 0) {
         return 0;
      } else {
         return (((origSize - newSize) * 100) / origSize);
      }
   }

}
