/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.Conductor.util;

import java.io.*;

/**
 * The Direction class implements a simple data object which can hold a
 * direction.  A direction is either up, down, or neutral.  A direction
 * object is immutable.
 * <p>
 * A direction object cannot be created by the user.  Instead, one of
 * the pre-allocated objects can be referenced from a Direction class 
 * variable or be returned by one of the class methods.
 *
 * @author Mark Yarvis (yarvis@fmg.cs.ucla.edu)
 * @version $Revision: 1.9 $
 */

public final class Direction {

   /** Internally speficies the up direction. */
   private static final int UP = 0;

   /** Internally speficies the down direction. */
   private static final int DOWN = 1;

   /** Internally speficies the neutral direction. */
   private static final int NEUTRAL = 2;


   /** A pre-instantiated up object. */
   public static Direction up = new Direction(UP);

   /** A pre-instantiated down object. */
   public static Direction down = new Direction(DOWN);

   /** A pre-instantiated neutral object. */
   public static Direction neutral = new Direction(NEUTRAL);


   /** The direction of this object */
   private int dir;

   /**
    * Direction objects cannot be created by the user.
    * @param dir The direction of the object to create.
    */
   private Direction(int dir) {
      this.dir = dir;
   }

   /**
    * Determines if the direction of this object is up.
    * @return True iff the direction is up.
    */
   public boolean isUp() {
      return (dir == UP);
   }

   /**
    * Determines if the direction of this object is down.
    * @return True iff the direction is down.
    */
   public boolean isDown() {
      return (dir == DOWN);
   }

   /**
    * Determines if the direction of this object is neutral.
    * @return True iff the direction is neutral.
    */
   public boolean isNeutral() {
      return (dir == NEUTRAL);
   }

   /**
    * Returns a direction object representing the opposite direction.
    * @return A down object if the this object is up.  An up object, 
    *         otherwise.
    */
   public Direction opposite() {
// TODO: what happens if dir is neutral?  BUG?
      return (dir == UP ? down : up);
   }

   /**
    * Determines if two objects are the same direction.
    * @param  other The object to compare against.
    * @return True if this object is the same direction as other.  False 
    *         otherwise.
    */
   public boolean equals(Direction other) {
      return (dir == other.dir);
   }

   /**
    * Turn this object into a string.
    * @return A new string object representing this object textually.
    */
   public String toString() {
      switch (dir) {
         case UP: return "up";
         case DOWN: return "down";
         case NEUTRAL: return "neutral";
         default: throw new RuntimeException("This shouldn't happen");
      }
   }

   /**
    * Read a direction from a DataInputStream.  The direction was presumably
    * written using <code>writeSelf()</code>.
    * @param is The stream to read from.
    * @return A Direction object representing the direction read
    *         from the stream.
    * @exception java.io.IOException if an exception occurs while reading 
    *            from the input stream.
    */
   public static Direction readDir(DataInputStream is) throws IOException {
      int dir = is.readInt();
// TODO: this could really be a byte rather than an int
//       that would then mean we could allow just an InputStream

      switch (dir) {
         case UP: return up;
         case DOWN: return down;
         case NEUTRAL: return neutral;
         default: throw new RuntimeException("This shouldn't happen");
      }
   }

   /**
    * Write this objects direction to a DataOutputStream.  The direction will
    * presumably be read using <code>readSelf()</code>.
    * @param os The stream to write to.
    * @return A Direction object representing the direction read
    *         from the stream.
    * @exception java.io.IOException if an exception occurs while writing 
    *                                to the output stream.
    */
   public void writeSelf(DataOutputStream os) throws IOException {
      os.writeInt(dir);
// TODO: see comment in readDir()
   }

}
