/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.Conductor.util;

/**
 * Obtain system information or other information external to Conductor.
 *
 * @author Mark Yarvis (yarvis@fmg.c.sucla.edu)
 * @version $Revision: 1.8 $
 */

public class ExternalInfo {

   /**
    * Load the utilities native library.
    */
   static {
      System.loadLibrary("condutil");
   }

   /**
    * Is this string a number?
    * @param s The string to test.
    * @return True iff the sting only contains decimal digits.
    */
   private static boolean isNumber(String s) {
      if (s.length() == 0) {
         return false;
      }

      byte[] buf = s.getBytes();

      for (int i=0; i<buf.length; i++) {
         if ((buf[i] < (byte) '0') || (buf[i] > (byte) '9')) {
            return false;
         }
      }

      return true;
   }

   /**
    * Obtain the port number associated with the given service name.  If 
    * the given string contains only a number, the number is simply parsed
    * and returned.  If any characters are in the name, it is looked up
    * using doServiceLookup().
    * @param name A string containint the name or number of a service port.
    * @return Either the numeric port number of the given service name, or
    *         -1 if no such service exists.
    */
   public static int getServicePort(String name) {
      if (name.length() == 0) {
         return -1;
      }

      if (isNumber(name)) {
         return Integer.parseInt(name);
      }

      return doServiceLookup(name);
   }

   /**
    * Use the native library function getservbyname() to obtain the port 
    * number for the given name.
    * @param name The name of the desired service.
    * @return Either the numeric port number of the given service name, or
    *         -1 if no such service exists.
    */
   private static native int doServiceLookup(String name);

   /**
    * Use the native library function getpwnam() to obtain a user's home 
    * directory
    * <p>
    * TODO: this should be able to be private soon.
    * @param username The name of the desired user.
    * @return The home directory of that user or NULL if the user is not known.
    */
   public static native String getUserHomedir(String username);

   /**
    * Verify that username is a valid user of the system and that his 
    * password is the given password.
    * @param username The name of the user to verify.
    * @param password A guess as to the user's password.
    * @return true iff the given user is valid and has the given password.
    */
   static native boolean verifyPassword(String username, String password);

   /**
    * Get the uid for the named user.
    * @param username The name of a user of the system.
    * @return The uid for the named user, or -1 if no such user exists.
    */
   static native int getUsersID(String username);

   /**
    * Get the gid of the default group for the named user.
    * @param username The name of a user of the system.
    * @return The gid of the default group for the named user, 
    *         or -1 if no such user exists.
    */
   static native int getUsersDefaultGroup(String username);

}
