/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.Conductor.util;

import java.io.PrintStream;

/**
 * This class provides an element for a linked list.
 *
 * @author Mark Yarvis (yarvis@fmg.cs.ucla.edu)
 * @version $Revision: 1.13 $
 */

public class LLElem extends Recyclable {
   /**
    * The handle for the next element in the list.  Subclasses should
    * only read this variable.  Direct access is provided for efficiency 
    * only.
    */
   protected LLElem next;

   /**
    * The handle for the previous element in the list.  Subclasses should
    * only read this variable.  Direct access is provided for efficiency 
    * only.
    */
   protected LLElem prev;

   /**
    * Insert a LLElem after this object.  If the element being inserted
    * is part of a list, the entire list is inserted after this object.
    * @param newElem The element (or an element in the list) to insert.
    * @return No return value.
    */
   public void insertAfter(LLElem newElem) {
      /* newElem may be a list, find first and last elem */
      LLElem first = newElem;
      while (first.prev != null) {
         first = first.prev;
      }
      LLElem last = newElem;
      while (last.next != null) {
         last = last.next;
      }

      /* now insert the list or element */
      if (next != null) {
         next.prev = last;
         last.next = next;
      }
      first.prev = this;
      next = first;
   }

   /**
    * Insert a LLElem before this object.  If the element being inserted
    * is part of a list, the entire list is inserted before this object.
    * @param newElem The element (or an element in the list) to insert.
    * @return No return value.
    */
   public void insertBefore(LLElem newElem) {
      /* newElem may be a list, find first and last elem */
      LLElem first = newElem;
      while (first.prev != null) {
         first = first.prev;
      }
      LLElem last = newElem;
      while (last.next != null) {
         last = last.next;
      }

      if (prev != null) {
         prev.next = first;
         first.prev = prev;
      }
      last.next = this;
      prev = last;
   }

   /**
    * Removes this element from the list it is currently in.
    * @return No return value.
    */
   public void unlink() {
      if (prev != null) {
         prev.next = next;
      }

      if (next != null) {
         next.prev = prev;
      }

      next = null;
      prev = null;
   }

   /**
    * Swaps positions between this element and the next element in the list.
    * @return No return value.
    */
   public void swapWithNext() {
      LLElem target = next;

      if (target == null) {
         throw new RuntimeException("Can't swap with null element");
      }

      /* fix my next to be what target's used to be */
      next = target.next;
      if (next != null) {
         next.prev = this;
      }

      /* fix target's prev to what mine used to be */
      target.prev = prev;
      if (prev != null) {
         prev.next = target;
      }

      /* fix pointers to each other */
      target.next = this;
      prev = target;
   }

   /**
    * Swaps positions between this element and the previous element in the 
    * list.
    * @return No return value.
    */
   public void swapWithPrev() {
      LLElem target = prev;

      /* fix my prev to what target's used to be */
      prev = target.prev;
      if (prev != null) {
         prev.next = this;
      }

      /* fix target's next to what mine used to be */
      target.next = next;
      if (next != null) {
         next.prev = target;
      }

      /* fix pointers to each other */
      target.prev = this;
      next = target;
   }

   /**
    * Do a sanity check just before we allow recycling.
    */
   public void recycling() {
      if ((Debug.on) && (Debug.getLevel() > Debug.SANITY_CHECK)) {
         if ((next != null) || (prev != null)) {
            throw new RuntimeException("Shouldn't recycle linked objects (" + 
                                       this + ")!!!");
         }
      }
   }

   /**
    * Prints this list via the Debug object.  Elements are printed in 
    * reverse order.
    * @return No return value.
    * @see FMG.Conductor.util.Debug
    */
   public void printReverseList(PrintStream out) {
      LLElem p = this;
   
      // find start
      while (p.next != null) {
         p = p.next;  
      }

      out.print("[");

      // run through list
      while (p != null) {
         out.print("" + p);
         if (p.prev != null) {
            out.print(", ");
         }
         p = p.prev;
      }
   
      out.println("]");
   }

   /**
    * Prints this list via the Debug object.
    * @return No return value.
    * @see FMG.Conductor.util.Debug
    */
   public void printList(PrintStream out) {
      LLElem p = this;
   
      // find start
      while (p.prev != null) {
         p = p.prev;  
      }

      out.print("[");

      // run through list
      while (p != null) {
         out.print("" + p);
         if (p.next != null) {
            out.print(", ");
         }
         p = p.next;
      }
   
      out.println("]");
   }

}
