/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.Conductor.util;

import FMG.Conductor.Conductor;

/**
 * The Measure class provides some useful functions for measuring the 
 * performance of the code.  In particular, measurement can be turned
 * on and off, and a mask can control which portions of code are measured.
 * Functions are also provided for obtaining timstamps and generating 
 * measurement output in a standardized way.
 *
 * By default, measurement is turned on for all parts of the code.  To
 * limit measurement to portions of the code, the <code>MeasurementMask</code> 
 * property may be set.  The value of this variable may be decimal, octal 
 * (if preceeded by a zero), or hex (if preceeded by "0x").
 *
 * A native code library is used to obtain timstamps.  It is not clear 
 * whether or not this is more efficient than System.getTimeMillis().
 *
 * @author Mark Yarvis (yarvis@fmg.cs.ucla.edu)
 * @version $Revision: 1.10 $
 */

public final class Measure {

   /** 
    * The name of the property that controlls the value of the measurment 
    * mask.
    */
   public static final String MEASURE_MASK_PROP_NAME = "MeasurementMask";

   /** 
    * The mask for determining whether measurement is turned on for planning 
    * code. 
    */
   public static final long PLANNING = 0x1;

   /** 
    * The mask for determining whether measurement is turned on for dataflow 
    * code. 
    */
   public static final long DATAFLOW = 0x2;

   /**
    * The mask for turning on measurement of semantic segmentation.
    */
   public static final long SEGMENTATION = 0x4;

   /**
    * The mask for turning on measurement of failure recovery.
    */
   public static final long RECOVERY = 0x8;

   /** 
    * A mask with all measurement turned on.  This is the default mask value.
    */
   private static final long MASK_ALL = -1;

   /** 
    * Indicates whether measurement is currently on or off.  It is 
    * publically accessible so the code can check whether measurement is 
    * turned on before calling costly functions (like println).  Direct
    * access to this variable is provided for efficiency.
    */
   public static boolean on = false;

   /** 
    * Indicates which portions of the code shoudl be measured.  This variable 
    * is set when measurement is turned on using the MeasurementMask property.
    * The default value is MASK_ALL.  This value of this variable is not valid
    * if <code>on</code> is set to false.  Direct access to this variable is 
    * provided for efficiency.
    */
   public static long mask = MASK_ALL;

   /**
    * Load the native function implementations.
    */
   static {
      System.loadLibrary("condutil");
   }

   /**
    * Determine the mask value specified in the given string.  The string
    * may specify the mask in decimal, octal (if preceeded by 0), or hex 
    * (if precceded by 0x) notation.  
    * @return The long value of the given mask.
    */
   private static long interpretMaskString(String s) {
      if (s.startsWith("0x")) {
         // interpret as hex
         return Long.parseLong(s.substring(2), 16);
      } else if (s.startsWith("0")) {
         // interpret as octal
         return Long.parseLong(s.substring(1), 8);
      } else {
         // interpret as decimal
         return Long.parseLong(s);
      }
   }

   /**
    * Set the measurement mask based on a string specification.  This method
    * is public so that it can be called by an interface (usually a subclass 
    * of Interface).
    * @param value A string specifying the value of the mask to set.
    * @return No return value.
    */
   protected static void setMask(String value) {
      if (value == null) {
         mask = MASK_ALL;
      } else {
         try {
            mask = interpretMaskString(value);
         } catch (NumberFormatException e) {
            System.err.println("Warning, attempt to set invalid mask: " + 
                               value);
            mask = MASK_ALL;
         }
      }
   }

   /**
    * Switches the status of measurement to on and sets the measurement mask
    * according to the user's specified property.
    * @return No return value.
    */
   public static void turnOn() {
      turnOn(Conductor.getProperty(MEASURE_MASK_PROP_NAME));
   }

   /**
    * Switches the status of measurement to on and sets the measurement mask
    * according to the given argument.
    * @param mask The measurment mask to set.
    * @return No return value.
    */
   public static void turnOn(String mask) {
      on = true;
      setMask(mask);
   }

   /**
    * Switches the status of measurement to off.
    * @return No return value.
    */
   public static void turnOff() {
      on = false;
   }

   /**
    * Output a measurement checkpoint.  The timestamp will be the current time.
    * @param s The string which identifies this timestamp.
    * @return No return value.
    */
   public static void println(String s) {
      println(s, getTime());
   }

   /**
    * Output a measurement checkpoint with the given timestamp.
    * @param s The string which identifies this timestamp.
    * @param time A previously measured timestamp to use for this checkpoint.
    * @return No return value.
    */
   public static void println(String s, long time) {
      StringBuffer sb = new StringBuffer();

      sb.append("MEASUREMENT (");
      sb.append(time);
      sb.append(" ms): ");
      sb.append(s);

      System.out.println(sb.toString());
   }

   /**
    * Get the current time in milliseconds.  It is not clear whether this
    * is superior to System.getTimeMillis(), but it is convenient to roll it
    * into this class.
    * @return No return value.
    */
   public static native long getTime();

   public static class TimeOfDay {
      public int secs;
      public int usecs;

      private String usecsString;

      TimeOfDay(int secs, int usecs) {
         this.secs = secs;
         this.usecs = usecs;
      }

      public String usecsString() {
         if (usecsString == null) {
            StringBuffer sb = new StringBuffer(6);
            int val = usecs;
            for (int i=0; i<6; i++) {
               sb.append(val % 10);
               val /= 10;
            }
            sb.reverse();
            usecsString = sb.toString();
         }
         return usecsString;
      }
   }

   public static native TimeOfDay getTimeOfDay();

}
