/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.Conductor.util;

import java.util.*;
import java.io.*;

/**
 * This class simplifies the loading of properties files.  It provides
 * some static methods for loading individual files.  An instance of
 * PropsLoader allows caching for files that may be read multiple times.
 *
 * Any loaded properties file can also include a property called 
 * <code>IncludeDefaults</code> which indicates another file to load
 * as the defaults for that properties file.
 *
 * @author Mark Yarvis (yarvis@fmg.cs.ucla.edu)
 * @version $Revision: 1.4 $
 */

public class PropsLoader {

   /** The list of files that are known to be unloadable */
   protected Hashlist badList = new Hashlist();

   /** 
    * Previously loaded properties indexed by the name of the file they 
    * were loaded from.
    */
   protected Hashtable cachedProps = new Hashtable();

   /** Should we fix the quoted values of files read? */
   protected boolean fixQuotedValues = false;

   /** 
    * The name of the property in a loaded properties file whose value is the
    * name of another file to load that contains a set of default properties.
    */
   protected static final String DEFAULTS_PROP_NAME = "IncludeDefaults";

   /**
    * Read properties from a file with the given name.
    * Also, look for property named "IncludeDefaults".  If found, load
    * that file as a set of default values and remove the property.
    * @param name The name of the file to read.
    * @return A new properties object representing the contents of the file.
    */
   public static Properties readProps(String name) {
      return readProps(new File(name), false);
   }

   /**
    * Read properties from a file with the given name.  
    * Optionally, remove quotes from around quoted values.
    * Also, look for property named "IncludeDefaults".  If found, load
    * that file as a set of default values and remove the property.
    * @param name The name of the file to read.
    * @param fixQuotedValues Remove quotes from around quoted values.
    * @return A new properties object representing the contents of the file.
    */
   public static Properties readProps(String name, boolean fixQuotedValues) {
      return readProps(new File(name), fixQuotedValues);
   }

   /**
    * Read properties from a file specified by a file object.
    * Also, look for property named "IncludeDefaults".  If found, load
    * that file as a set of default values and remove the property.
    * @param name The file object representing the file to read.
    * @return A new properties object representing the contents of the file.
    */
   public static Properties readProps(File f) {
      return readProps(f, false);
   }

   /**
    * Given a properties list and a set of defaults, apply any default values
    * that don't already exist in the properties list.
    * @param target The list to add defaults to.
    * @param defaults The set of defaults to add to the target.
    * @return No return value.
    */
   private static void setDefaults(Properties target, Properties defaults) {
      if (defaults == null) {
         return;
      }

      Enumeration e = defaults.keys();
      while (e.hasMoreElements()) {
         String k = (String) e.nextElement();
         if (! target.containsKey(k)) {
            target.put(k, (String) defaults.get(k));
         }
      }
   }

   /**
    * Read properties from a file specified by a file object.
    * Optionally, remove quotes from around quoted values.
    * @param f The file object representing the file to read.
    * @param fixQuotedValues Remove quotes from around quoted values.
    * @return A new properties object representing the contents of the file.
    */
   protected static Properties doReadProps(File f, boolean fixQuotedValues) {
      Properties props = new Properties();
                                                
      try {
         props.load(new FileInputStream(f)); 
      } catch (IOException e) {
         if (Debug.on) {
            Debug.println("Warning: failure loading properties file " + 
                          f + ":" + e);
         }
         return null;
      }

      if (fixQuotedValues) {
         fixQuotedValues(props);
      }

      return props;
   }

   /**
    * Read properties from a file specified by a file object.
    * Optionally, remove quotes from around quoted values.
    * Also, look for property named "IncludeDefaults".  If found, load
    * that file as a set of default values and remove the property.
    * @param f The file object representing the file to read.
    * @param fixQuotedValues Remove quotes from around quoted values.
    * @return A new properties object representing the contents of the file.
    */
   public static Properties readProps(File f, boolean fixQuotedValues)  {
      Properties props = doReadProps(f, fixQuotedValues);

      if (props == null) {
         return null;
      }

      String defaults = (String) props.get(DEFAULTS_PROP_NAME);
      if (defaults != null) {
         setDefaults(props, readProps(new File(defaults), fixQuotedValues));
         props.remove(DEFAULTS_PROP_NAME);
      }

      return props;
   }

   /**
    * Look through the values in the given hashtable and remove any
    * quotation marks that surround them.
    * @param table The hashtable to fix.
    * @return No return value.
    */
   public static void fixQuotedValues(Hashtable table) {
      Enumeration keys = table.keys();
      while (keys.hasMoreElements()) {
         String k = (String) keys.nextElement();
         String val = (String) table.get(k); 

         int len = val.length();
      
         if ((len > 1) && 
             ( ((val.charAt(0) == '"') && (val.charAt(len-1) == '"')) ||
               ((val.charAt(0) == '\'') && (val.charAt(len-1) == '\''))
            )) {
            table.put(k, val.substring(1, len-1));
         }
      }
   }

   /**
    * Create a PropsLoader.
    * @param fixQuotedValues Should quotes be removed from all read values?
    */
   public PropsLoader(boolean fixQuotedValues) {
      this.fixQuotedValues = fixQuotedValues;
   }

   /**
    * Obtain a set of properties from the given file.  If this file has
    * been read before, the cached copy of the file data will be read.
    * If the file cannot be read (perhaps because it doesn't exist),
    * an exception is thrown.
    * @param name The name of the file to read.
    * @return The properties from this file.  <strong>Warning</strong>, 
    *         the same object may be returned on subsequent calls.
    */
   public Properties getProps(String name) {
      if (badList.isElement(name)) {
         return null;
      }

      Properties props = (Properties) cachedProps.get(name);
      if (props == null) {

         props = doReadProps(new File(name), fixQuotedValues);

         if (props == null) {
            badList.addElement(name);
            return null;
         }

         String defaults = (String) props.get(DEFAULTS_PROP_NAME);
         if (defaults != null) {
            setDefaults(props, getProps(defaults));
            props.remove(DEFAULTS_PROP_NAME);
         }

         cachedProps.put(name, props);
      }
      return props;
   }

   /**
    * Obtain a set of properties from the given file.  If this file has
    * been read before (based on file name), the cached copy of the file 
    * data will be read.  If the file cannot be read (perhaps because it 
    * doesn't exist), an exception is thrown.
    * @param f A file object representing the file to read.
    * @return The properties from this file.  <strong>Warning</strong>, 
    * the same object may be returned on subsequent calls.
    */
   public Properties getProps(File f) {
      return getProps(f.getName());
   }

   /**
    * Obtain a refreshed copy of the properties stored in the named file.
    * Other than that no cached copy (positive or negative) will be used,
    * the semantics of this method is the same as <code>getProps()</code>.
    * @param name The name of the file to read.
    * @return The properties from this file.  <strong>Warning</strong>, 
    * the same object may be returned on subsequent calls.
    */
   public Properties refresh(String name) {
      badList.removeElement(name);
      cachedProps.remove(name);
      return getProps(name);
   }

   /**
    * Obtain a refreshed copy of the properties stored in the file specified.
    * Other than that no cached copy (positive or negative) will be used,
    * the semantics of this method is the same as <code>getProps()</code>.
    * @param f A file object representing the file to read.
    * @return The properties from this file.  <strong>Warning</strong>, 
    *         the same object may be returned on subsequent calls.
    */
   public Properties refresh(File f) {
      return refresh(f.getName());
   }

   /**
    * Delete any cached information (both positive and negative).
    * @return No return value.
    */
   public void flushCaches() {
      badList.removeAllElements();

      Enumeration e = cachedProps.elements();
      while (e.hasMoreElements()) { 
         cachedProps.remove(e.nextElement());
      }
   }
}
