/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.Conductor.util;

/**
 * The Recyclable class is the super-class of all objects that can be recycled
 * using the Recycler.  Objects that will be created and destroyed many
 * times or that are expensive to create should be managed by a recycler.  
 * Once the object is obtained from the recycler, it can be recycled by 
 * calling <code>recycle()</code>.  
 * <p>
 * To maintain control of an object and prevent it from being pre-maturely 
 * recycled by another entity in the system, call <code>retain()</code>.  
 * There must be one more call to <code>recycle()</code> than calls to 
 * <code>retain()</code> before the object will be recycled.
 * <p>
 * Before an object is actually recycled, it's <code>recycling()</code> 
 * method will be called by the recycler.
 * <p>
 * <b>Note:</b> It is important that a recyclable object not be initialized
 * in the constructor.  The constructor is only called by the recycler when
 * the object is created.  An object should be initialized after it is 
 * returned by the recycler.
 *
 * @author Mark Yarvis (yarvis@fmg.cs.ucla.edu)
 * @version $Revision: 1.8 $
 *
 * @see FMG.Conductor.util.Recycler
 */

public class Recyclable {
   /** 
    * The number of entities which currently have a hold of this object 
    * either by obtaining it from the recycler or by making a call to 
    * <code>retain()</code>.
    */
   private int handles = 1;

   /** A pointer to this object's recycler. */
   private Recycler recycler;

   /**
    * Record a handle to this object's recycler.  This method should normally 
    * only be called by the recycler itself.
    * @param r The object's recycler.
    * @return No return value.
    */
   void setRecycler(Recycler r) {
      recycler = r;
   }

   /**
    * Prevent this object from being recycled by another entity.  This method
    * helps resolve issue of ownership of objects and protects objects with
    * multiple "owners" from being recycled.  For each call to 
    * <code>retain</code> there must be an extra call to <code>recycle()</code>
    * before recycling will occur.
    * @return No return value.
    */
   public final synchronized void retain() {
      handles++;
   }

   /**
    * A call to this method indicates that the caller no longer has an
    * interest in this method.  If there are no other interests in this
    * object (indicated by calls to <code>retain()</code>), then the object
    * will be recycled.
    * @return No return value.
    */
   public final synchronized void recycle() {
      if (recycler == null) {
         if (Debug.on) {
            Debug.println(Debug.RECYCLER, Debug.WARNING, 
                          "WARNING: Attempt to recycle object (" + this + 
                          ") that was not created by a recycler");
         }
         return;
      }

      if (handles == 1) {
         recycling();
         recycler.reclaim(this);
      } else {
         handles--;
      }
   }

   /**
    * This method is called immediately before an object is actually
    * available for reuse.  It allows for any cleanup that needs to
    * occur.
    * @return No return value.
    */
   public void recycling() {
   }

}
