/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.Conductor.util;

/**
 * The RecyclableThread class is used internally by the ThreadRecycler
 * to dispatch and execute jobs.  Once created and started, this thread
 * is always running.  In-between jobs, it is waiting.  When a job arrives,
 * the job's <code>run()</code> method is executed until it returns.
 *
 * @author Mark Yarvis (yarvis@fmg.cs.ucla.edu)
 * @version $Revision: 1.9 $
 *
 * @see FMG.Conductor.util.ThreadRecycler
 * @see FMG.Conductor.util.Job
 */

class RecyclableThread extends Thread {
   /**
    * The job that is currently executing.
    */
   protected Runnable job;

   /**
    * Should we stop waiting for jobs?
    */
   protected boolean stopThread = false;

   /**
    * Synchronizer for notifying the thread when new work is ready.
    */
   private Object sync = new Object();

   /**
    * Create a new recyclable thread.  This should normally only be called 
    * by the thread recycler.
    */
   RecyclableThread() {
      super("RecyclableThread(unassigned)");
   }

   /**
    * This method causes a waiting thread to start executing the given new job.
    * @param newJob The job to execute.
    * @return No return value.
    */
   void startJob(Runnable newJob) {
      if (job != null) {
         throw new RuntimeException("Tried to start a new job before " + 
                                    "completion of prior job");
      }
      synchronized (sync) {
         job = newJob;
         if (job instanceof Job) {
            ((Job) job).setThread(this);
         }
         sync.notify();
      }
   }

   /**
    * Stop this thread from waiting for jobs.  If a job is currently running,
    * stop after it's done.
    * @return No return value.
    */
   void stopThread() {
      synchronized (sync) {
         stopThread = true;
         sync.notify();
      }
   }

   /**
    * Do the actual work cycle here.  Wait for a job.  Run the job.  Wait for
    * the next job.
    * @return No return value.
    */
   public void run() {
      while (true) {
         synchronized (sync) {
            // wait for a new job
            while (job == null) {
               if (stopThread) {
                  return;
               }

               try {
                  sync.wait();
               } catch (InterruptedException e) {}
            }
         }

         // run new job
         job.run();

         if (stopThread) {
            return;
         }

         // We're done, clear job and recycle self
         if (Debug.on) {
            setName("RecyclableThread(unassigned)");
         }
         if (job instanceof Job) {
            ((Job) job).setThread(null);
         }
         job = null;
         ThreadRecycler.recycle(this);
      }
   }
}
