/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.Conductor.util;

import java.util.Vector;
import java.util.Enumeration;
import java.io.PrintStream;

/**
 * The Recycler class allows the cost of creating objects to be saved by
 * caching them for reuse.  This is useful for objects that are created and
 * discarded frequently or for objects that are expensive to create.
 * <p>
 * One or more recyclers is created for each class of objects that are
 * to be reused.  When an object is requested from the recycler, one is either
 * created or one is taken from the cache.  Objects to be recycled must
 * be a sublcass of Recyclable.
 * <p>
 * TODO: There should be two separate limits: a creation limit (number of 
 *       outstanding objects) and a caching limit (number of objects saved 
 *       for reuse).
 *
 * @author Mark Yarvis (yarvis@fmg.cs.ucla.edu)
 * @version $Revision: 1.17 $
 *
 * @see FMG.Conductor.util.Recyclable
 */

public final class Recycler {

   /** A list of all recyclers in the system. For debugging only. */
   protected static Vector allRecyclers = new Vector();


   /** The cache of previously created objects */
   protected Vector cache = new Vector();

   /** The class of objects created by this recycler */
   protected Class objectClass;


   /** Does this Recycler have a creation limit? */
   protected boolean hasLimit = false;

   /** The number of new objects that may be created by this Recycler */
   protected int instanceLimit;

   /** The number of new objects that have been created by this Recycler */
   protected int instanceCount = 0;

   /** Another object instantiation count, used for debuggin only!!! */
   private int numInstantiated = 0;

   /** 
    * Creates a new recycler for a particular class of objects.  This recycler
    * will have not instance limit.
    * @param objectClass the class of objects to produce from this recycler.
    */
   public Recycler(Class objectClass) {
      if (Debug.on && (Debug.getLevel() > Debug.SANITY_CHECK)) {
         for (Class c = objectClass; 
              (c != null) && !c.equals(Recyclable.class); 
              c = c.getSuperclass());
         if (objectClass == null) {
            throw new RuntimeException("Recycler can only accept " + 
                                       "subclasses of Recyclable.");
         }
      }

      this.objectClass = objectClass;

      // For debugging
      allRecyclers.addElement(this);
   }

   /** 
    * Creates a new recycler for a particular class of objects with a specified
    * instance limit.  Once this many objects is created, further requests
    * for new objects will block.
    * @param objectClass The class of objects to produce from this recycler.
    * @param instanceLimit The number of new objects that may be created.
    */
   public Recycler(Class objectClass, int instanceLimit) {
      this(objectClass);
      hasLimit = true;
      this.instanceLimit = instanceLimit;
   }

   /**
    * Creates a new object of the class specified for this Recycler.
    * @return A newly instantiated object of the correct class.
    */
   private Recyclable createOne() {
      try {

         // for debugging only
         if (Debug.on && (Debug.getLevel(Debug.RECYCLER) >= Debug.INFO)) {
            numInstantiated++;
            if ((numInstantiated % 100) == 0) {
               Debug.println("" + numInstantiated + " instances of " +
                                  objectClass + " created");
            }
         }

         Recyclable r = (Recyclable) objectClass.newInstance();
         r.setRecycler(this);
         return r;
      } catch (InstantiationException e) {
         throw new RuntimeException("Could not instantiate class " + 
                                    objectClass + ": " + e);
      } catch (IllegalAccessException e) {
         throw new RuntimeException("Could not instantiate class " + 
                                    objectClass + ": " + e);
      }
   }

   /**
    * This method allows the caller to get an object of the type handled
    * by this recycler.  It can either be a newly created object or one
    * previously recycled.  This call can block if the instance limit has
    * been met and there are no objects in the cache.
    * @return An object of the type handled by this recycler.
    */
   public Recyclable getOne() {
      if (hasLimit) {
         synchronized (this) {
            while (true) {
               if (cache.isEmpty()) {
                  if (instanceCount < instanceLimit) {
                     instanceCount++;
                     return createOne();
                  } else {
                     try {
                        wait();
                     } catch (InterruptedException e) {}
                  }
               } else {
                  Recyclable o = (Recyclable) cache.firstElement();
                  cache.removeElementAt(0);
                  return o;
               }
            }
         }
      } else {
         synchronized (cache) {
            if (cache.isEmpty()) {
               return createOne();
            } else {
               Recyclable o = (Recyclable) cache.firstElement();
               cache.removeElementAt(0);
               return o;
            }
         }
      }
   }

   /**
    * This method is called by a Recyclable object when that object
    * is ready to be recycled.  The object is added to the cache.
    * @param o The object to be recycled.
    * @return No return value.
    */
   public void reclaim(Recyclable o) {
      if (hasLimit) {
         synchronized (this) {
            cache.addElement(o);
            notify();
         }
      } else {
         cache.addElement(o);
      }
   }

   /**
    * Create a human readable string representing the state of this recycler.
    * For debugging use only.
    * @return A new string representing this object.
    */
   public String toString() {
      StringBuffer sb = new StringBuffer("Recycler[");
      sb.append(objectClass.getName());
      sb.append(", created: ");
      sb.append(numInstantiated);
      sb.append(", cached: ");
      sb.append(cache.size());
      sb.append("]");
      return sb.toString();
   }

   /**
    * Dump human-readable info about the status of all recycler objects.
    * For debugging use only.
    * @return No return value.
    */
   public static void dumpRecyclingInfo(PrintStream os) {
      Enumeration e = allRecyclers.elements();
      while (e.hasMoreElements()) {
         os.println(((Recycler) e.nextElement()).toString());
      }
   }
}
