/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.Conductor.util;

import java.io.*;

import FMG.Conductor.Conductor;

/**
 * This class is intended to provide an adpator with all allowed access to 
 * the filesystem.  All other access to the filesystem should be disallowed.
 * <p>
 * An instance of this class allows both authenticated and unauthenticated 
 * access to the file system.  Unauthenticated access is allowed only within 
 * a Conductor "working directory," which is, by default, located in 
 * <code>/tmp/conductor/anonymous</code>.  Authenitcate access to
 * a user's personal working directory is also allowed.  A user's personal
 * working directory is typically a sub-directory called "conductor" in
 * their home directory.  If the user's home directory is not available,
 * a sub-directory is created for the user in the working area, typically
 * <code>/tmp/conductor/<var>username</var></code>.  In addition, 
 * authenticated users may access any arbitrary files in the filesystem that 
 * are accessible by the authenticated user.
 * <p> 
 * The working directory can (optionally) be divided according to Conductor
 * module.  In particular, each adaptor may wish to use a unique module name
 * to prevent storage conflicts.  Four constructors are available for the
 * StorageAccess object to create unauthenticated or authenticated objects
 * with or without a specific module name.
 * <p>
 * Once created, a StorageAccess object allows files to be open for reading
 * and writing.  An authenticated user can use <code>writeFile()</code>
 * and <code>readFile()</code> to gain access to arbitrary files in the file 
 * system (subject to the rights of that user).  All users can use
 * <code>writeWorkingFile()</code> and <code>readWorkingFile()</code>
 * to obtain access to files in their working area.  These methods return 
 * objects of type <code>StorageAccess.ReadAccess</code> or 
 * <code>StorageAccess.WriteAccess</code>, which simply provide access
 * to an input or output stream and the absolute name of the file.
 * <p>
 * @author Mark Yarvis (yarvis@fmg.cs.ucla.edu)
 * @version $Revision: 1.6 $
 */

public final class StorageAccess {

   /** 
    * The location of the global working directory.  When access is 
    * authenticated, this is a fallback for the case in which the user's 
    * home is not found.  The value of this variable is controlled by
    * the "GlobalWorkingDir" property and has a default value of 
    * "/tmp/conductor".
    */
   private static final String GLOBAL_WORKING_DIR = 
          Conductor.getFilenameProperty("GlobalWorkingDir", "/tmp/conductor");

   /**
    * The  name of the subdirectory in a user's home in which working files
    * are stored.  The value of this variable is controlled by the 
    * "HomeWorkingDir" property and has a default value of "conductor".
    */
   private static final String USER_WORKING_DIR_NAME = 
          Conductor.getFilenameProperty("HomeWorkingDir", "conductor");

   /**
    * The name of the currently authenticated user, or null if no user is
    * authenticated.
    */
   private String authenticatedUser = null;

   /** The uid of the currently authenticated user. */
   int authenticatedUID = -1;

   /** The gid of the currently authenticated user. */
   int authenticatedGID = -1;

   /** The current working directory. */
   private File workingDir;

   /** The directory that this module wants to write in. */
   private File moduleDir;

   /** The name of the module working directory. */
   private String moduleName;

   /**
    * This class provides a handy structure related to a file that the
    * user has been granted access to read.
    */
   public static class ReadAccess {
      ReadAccess(InputStream is, String filename) {
         this.is = is;
         this.filename = filename;
      }

      /** An input stream from which the file contents can be read */
      public InputStream is;

      /** The absolute pathname of the file. */
      public String filename;
   }

   /**
    * This class provides a handy structure related to a file that the
    * user has been granted access to write.
    */
   public static class WriteAccess {
      WriteAccess(OutputStream os, String filename) {
         this.os = os;
         this.filename = filename;
      }

      /** An output stream with which the file can be written. */
      public OutputStream os;

      /** The absolute pathname of the file. */
      public String filename;
   }

   /**
    * Create an unauthenticed StorageAccess object that is not tied to a 
    * particular module.  Files written will be in the global working 
    * directory.
    */
   public StorageAccess() {
      this(null, null, null);
   }

   /**
    * Create an unauthenticed StorageAccess object that is tied to a 
    * module with the given name.  Files written will be in the global
    * working directory under a directory named for the particular module.
    * @param moduleName The name of the module writing the files.  This
    *                   parameter is only used to choose the subdirectory in
    *                   which the files should be placed.
    */
   public StorageAccess(String moduleName) {
      this(null, null, moduleName);
   }

   /**
    * Create an authenticed StorageAccess object that is not tied to a 
    * particular module.  If authentication succeeds, this object will
    * provide access to any files in the filesystem accessible to the user
    * and will provide a working area either within the user's home or in
    * an area set aside for the user within the global working area.
    * @param username The name of the user to authenticate.
    * @param password The user's password.
    */
   public StorageAccess(String username, String password) {
      this(username, password, null);
   }

   /**
    * Create an unauthenticed StorageAccess object that is tied to a 
    * module with the given name.  If authentication succeeds, this object 
    * will provide access to any files in the filesystem accessible to the 
    * user and will provide a working area either within the user's home or 
    * in an area set aside for the user within the global working area.
    * Files will be store in a subdirectory of the working area set aside
    * for the named module.
    * @param username The name of the user to authenticate.
    * @param password The user's password.
    * @param moduleName The name of the module writing the files.  This
    *                   parameter is only used to choose the subdirectory in
    *                   which the files should be placed.
    */
   public StorageAccess(String username, String password, String moduleName) {
      this.moduleName = moduleName;
      authenticate(username, password);
   }

   /**
    * Attempt to authenticate the given user using his password.  
    * Authentication is performed by encrypting the password and comparing
    * it to the one stored in the password database for this user.  This
    * method can be used to authenticate a different user, to reauthenticate
    * after an authentication failure, or to deauthenticate (pass null for
    * the username or password).  Once authenticated, or unauthenticated,
    * an appropriate working directory will be set.
    * @param username The name of the user to authenticate.  Pass null to 
    *                 deauthenticate.
    * @param password The user's password.
    * @return Returns true if the object is now authenticated to the specified 
    *         user.  Returns false if the object is now deauthenticated 
    *         (either due to a failed authentication or because the username 
    *         was null).
    */
   public boolean authenticate(String username, String password) {
      if ((username != null) && (password != null) && 
          ExternalInfo.verifyPassword(username, password)) {

         authenticatedUser = username;
         authenticatedUID = ExternalInfo.getUsersID(authenticatedUser);
         authenticatedGID = 
                       ExternalInfo.getUsersDefaultGroup(authenticatedUser);

         setAuthenticatedWorkingDir();
         moduleDir = getModuleDir(moduleName);
         return true;
      } else {
         authenticatedUser = null;
         authenticatedUID = -1;
         authenticatedGID = -1;
         setUnauthenticatedWorkingDir();
         moduleDir = getModuleDir(moduleName);
         return false;
      }
   }

   /**
    * Obtain a File object pointing to the module directory within the
    * working directory.  If no module was specified, the working directory
    * is returned.  If a module is specified, a subdirectory of the working 
    * directory is created (if necessary) and the approrpiate file object 
    * is returned.
    * @param module The requested module name.
    * @return A new file object for the requested module or 
    *         <var>workingDir</var> if no module was requested.
    */
   private File getModuleDir(String module) {
      if ((module == null) || (module.equals(""))) {
         return workingDir;
      }

      // at this point, we assume workingDir is present and writable
      File moduleDir = new File(workingDir, module);
      if (! moduleDir.exists()) {
         if (! moduleDir.mkdir()) {
            return null;
         }
         if (isAuthenticated()) {
            if (! changeOwnership(moduleDir.getAbsolutePath(), 
                                  authenticatedUID, authenticatedGID)) {
               return null;
            }
         }
      }

      return moduleDir;
   }

   /**
    * Obtain the name of the directory in which working files will be written.
    * @return A new string representing the absolute path of the directory 
    *         where working files will be written.
    */
   public String baseDir() {
      return moduleDir.getAbsolutePath();
   }

   /**
    * Set the working directory to GLOBAL_WORKING_DIR/anonymous for the case
    * in which a user is not authenticated.
    * @return No return value.
    */
   private void setUnauthenticatedWorkingDir() {
      workingDir = new File(GLOBAL_WORKING_DIR + "/anonymous");
      workingDir.mkdirs();
      if (! workingDir.exists() || ! workingDir.isDirectory() || 
          ! workingDir.canWrite()) {
         // nowhere to write!
         workingDir = null;
      }
   }

   /**
    * Set the working directory to the appropriate subdirectory of the 
    * authenitcated user's home directory, or to the user's private subdir
    * of the global working directory (if the user has no home).  Create
    * such a directory if necessary.
    * @return No return value.
    */
   private void setAuthenticatedWorkingDir() {
      if (authenticatedUser == null) { 
         throw new RuntimeException("This shouldn't happen");
      }

      String home = ExternalInfo.getUserHomedir(authenticatedUser);

      // look for the proper directory in the user's home
      File dir = new File(home, USER_WORKING_DIR_NAME);
      if (dir.exists() && dir.isDirectory() && dir.canWrite()) {
         workingDir = dir;
         return;
      }

      // if home is present, try to create a working directory in user's home
      File homeDir = new File(home);
      if (homeDir.exists() && homeDir.isDirectory() && homeDir.canWrite()) {
         workingDir = new File(homeDir, USER_WORKING_DIR_NAME);
         if (workingDir.mkdir()) {
            changeOwnership(workingDir.getAbsolutePath(), authenticatedUID, 
                                                          authenticatedGID);
            return;
         }
      }

      // look for user subdir in global area
      dir = new File(GLOBAL_WORKING_DIR, authenticatedUser);
      if (dir.exists() && dir.isDirectory() && dir.canWrite()) {
         workingDir = dir;
         return;
      }

      // create user subdir in global area
      dir = new File(GLOBAL_WORKING_DIR);
      dir.mkdirs();
      if (! dir.exists() || ! dir.isDirectory() || ! dir.canWrite()) {
         // no place to write files!
         workingDir = null;
         return;
      }
      workingDir = new File(dir, authenticatedUser);
      workingDir.mkdir();
      changeOwnership(workingDir.getAbsolutePath(), authenticatedUID,
                                                    authenticatedGID);
   }

   /**
    * Determine if this object is currently authenticated.
    * @return true iff the object is authenticated.
    */
   public boolean isAuthenticated() {
      return (authenticatedUser != null);
   }

   /**
    * Obtain the name of the authenticated user.
    * @return The name of the authenticated user or 
    *         null if no user is authenticated.
    */
   public String authenticatedUser() {
      return authenticatedUser;
   }

   /**
    * Obtain write access to an arbitrary file.  This will only succeed if
    * the object is authenticated and if the authenticated user would 
    * normally have access to the specified file.  If the file exists,
    * it will be overwritten.
    * @param filename The absolute name of the file for which write access 
    *                 is desired.
    * @reutrn A new WriteAccess object if access is allowed, or null if access
    *         is denied.
    * @exception java.io.FileNotFoundException Thrown if the directory path
    *                                          to the file does not exist.
    */
   public WriteAccess writeFile(String filename) throws FileNotFoundException {
      return writeFile(filename, false);
   }

   /**
    * Obtain write access to an arbitrary file.  This will only succeed if
    * the object is authenticated and if the authenticated user would 
    * normally have access to the specified file.
    * @param filename The absolute name of the file for which write access 
    *                 is desired.
    * @param append Should the file be appended, if it exists?
    * @reutrn A new WriteAccess object if access is allowed, or null if access
    *         is denied.
    * @exception java.io.FileNotFoundException Thrown if the directory path
    *                                          to the file does not exist.
    */
   public WriteAccess writeFile(String filename, boolean append) 
                                              throws FileNotFoundException {
      if (isAuthenticated() && userCanWrite(authenticatedUser, filename)) {
         WriteAccess access;
         try {
            access = new WriteAccess(new FileOutputStream(filename, append), 
                                     filename);
         } catch (IOException e) {
            throw new FileNotFoundException(e.toString());
         }

         changeOwnership(filename, authenticatedUID, authenticatedGID);

         return access;

      } else {
         return null;
      }
   }

   /**
    * Obtain read access to an arbitrary file.  This will only succeed if
    * the object is authenticated and if the authenticated user would 
    * normally have access to the specified file.
    * @param filename The absolute name of the file for which read access 
    *                 is desired.
    * @reutrn A new ReadAccess object if access is allowed, or null if access
    *         is denied.
    * @exception java.io.FileNotFoundException Thrown if the file does
    *                                          not exist.
    */
   public ReadAccess readFile(String filename) throws FileNotFoundException {
      if (isAuthenticated() && userCanRead(authenticatedUser, filename)) {
         return new ReadAccess(new FileInputStream(filename), filename);
      } else {
         return null;
      }
   }

   /**
    * Determine if the named file already exists in the working directory.
    * @param filename A relative filename.
    * @return True iff the file exists in the working directory.
    */
   public boolean workingFileExists(String filename) {
      File file = new File(moduleDir, filename);
      return file.exists();
   }

   /**
    * Obtain write access to a file in the working directory.  If the
    * file exists, it is overwritten.
    * @param filename A relative filename.
    * @return A new WriteAccess object.
    */
   public WriteAccess writeWorkingFile(String filename) {
      return writeWorkingFile(filename, false);
   }


   /**
    * Obtain write access to a file in the working directory.
    * @param filename A relative filename.
    * @param append Should we append to the file, if it exists?
    * @return A new WriteAccess object.
    */
   public WriteAccess writeWorkingFile(String filename, boolean append) {
      if (moduleDir == null) {
         return null;
      }

      File file = new File(moduleDir, filename);
      WriteAccess access;
      try {
         access = new WriteAccess(new FileOutputStream(file.getAbsolutePath(),
                                                       append), 
                                  file.getAbsolutePath());
      } catch (IOException e) {
         throw new RuntimeException("This shouldn't happen.");
      }

      if (isAuthenticated()) {
         changeOwnership(file.getAbsolutePath(), authenticatedUID, 
                                                 authenticatedGID);
      }

      return access;
   }

   /**
    * Obtain read access to a file in the working directory.
    * @param filename A relative filename.
    * @return A new ReadAccess object.
    * @exception java.io.FileNotFoundException Thrown if the file does not 
    *                                          exist.
    */
   public ReadAccess readWorkingFile(String filename) 
                                         throws FileNotFoundException {
      if (moduleDir == null) {
         return null;
      }

      File file = new File(moduleDir, filename);
      ReadAccess access = new ReadAccess(new FileInputStream(file), 
                                         file.getAbsolutePath());

      return access;
   }

   /**
    * Determine whether the named user would normally be able to read the
    * named file.
    * <p>
    * <b>Warning:</b> At the current time, this method is not implemented
    * and always returns true.
    * @param username The name of the user to check access for.
    * @param filename An absolute pathname.
    * @return True iff the user would normally be able to read the file.
    */
   private static native boolean userCanRead(String username, 
                                             String filename);

   /**
    * Determine whether the named user would normally be able to write the
    * named file.
    * <p>
    * <b>Warning:</b> At the current time, this method is not implemented
    * and always returns true.
    * @param username The name of the user to check access for.
    * @param filename An absolute pathname.
    * @return True iff the user would normally be able to write the file.
    */
   private static native boolean userCanWrite(String username,
                                              String filename);

   /**
    * Change the user and/or group ownership of a file.
    * @param filename An absolute path.
    * @param uid The new user id for this file.  If -1 is passed, the user
    *            owner will not be changed.
    * @param gid The new group id for this file.  If -1 is passed, the group
    *            owner will not be changed.
    * @return True iff the operation was successful.
    */
   private static native boolean changeOwnership(String filename, 
                                                 int uid, int gid);
}
