/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.Conductor.util;

import FMG.Conductor.Conductor;

import java.util.Vector;

/**
 * The ThreadRecycler reduces the cost of transient threads by allowing
 * Thread objects to be reused.  Jobs can be requested to be run by calling
 * <code>startNewJob()</code>.  Jobs should normally be a sublcass of the 
 * Job class, but must at least implement Runnable.  If a thread is available, 
 * the job is allocated to the thread.  Execution begins with the job's 
 * <code>run()</code> method.  If no thread is available, one is created
 * and started.  When the job's <code>run()</code> method returns, the
 * thread moves to a "wait" state and is queued for use by a new job.
 * <p>
 * The threads used to execute the supplied jobs are of type RecyclableThread.
 *
 * @author Mark Yarvis (yarvis@fmg.cs.ucla.edu)
 * @version $Revision: 1.14 $
 *
 * @see FMG.Conductor.util.Job
 * @see FMG.Conductor.util.RecyclableThread
 */

public final class ThreadRecycler {
   /**
    * The list of inactive threads, ready to execute jobs.  They are all of
    * class RecyclableThread.
    * @see FMG.Conductor.util.RecyclableThread
    */
   protected static Vector ready = new Vector();

   /**
    * the limit to the number of threads that will be cached.  Once this
    * many threads are cached, any other threads whose jobs end will be
    * destroyed.
    */
   protected static int cacheLimit = 
               Conductor.getIntProperty("ThreadCacheLimit", Integer.MAX_VALUE);

   /**
    * The number of threads created.  For debugging use only.
    */
   private static int numInstantiated = 0;

   /**
    * Lock which protects the ready list.
    */
   private static Object lock = new Object();

   /**
    * Thread recycling is a global affair.  There is no need to create an 
    * instance of ThreadRecycler.
    */
   private ThreadRecycler() {
   }

   /**
    * This method is called to start the execution of a job at normal
    * priority.  If an available thread exists, it will be used.  Otherwise
    * a new thread is created.  The method returns immediately, without
    * waiting for the job to complete.
    * @param job The job to run.
    * @return No return value.
    */
   public static void startNewJob(Runnable job) {
      startNewJob(job, Thread.NORM_PRIORITY);
   }

   /**
    * This method is called to start the execution of a job at a given
    * priority.  If an available thread exists, it will be used.  Otherwise
    * a new thread is created.  The method returns immediately, without
    * waiting for the job to complete.
    * @param job The job to run.
    * @param priority The job's priority (use the priorities in Thread).
    * @return No return value.
    * @see java.lang.Thread
    */
   public static void startNewJob(Runnable job, int priority) {
      RecyclableThread rthread;

      synchronized (lock) {
         if (ready.isEmpty()) {
            if (Debug.on && (Debug.getLevel(Debug.RECYCLER) > Debug.INFO)) {
               numInstantiated++;
               if ((numInstantiated % 10) == 0) {
                  Debug.println("ThreadRecycler created " + numInstantiated + 
                                " threads");
               }
            }
            rthread = new RecyclableThread();
            rthread.start();
         } else {
               rthread = (RecyclableThread) ready.firstElement();
               ready.removeElementAt(0);
         }
      }

      rthread.setPriority(priority);

      rthread.startJob(job);
   }

   /**
    * This method removes threads from the list of available threads and 
    * stops them from running.  This is probably only useful for debugging.
    */
   public static void killUnusedThreads() {
      synchronized (lock) {
         while (! ready.isEmpty()) {
            ((RecyclableThread) ready.firstElement()).stopThread();
            ready.removeElementAt(0);
         }
      }
   }

   /**
    * This method either places the thread on the list of threads 
    * available to run jobs or destroys the thread (if the cache limit
    * is reached).  It should only be called by the RecyclableThread when 
    * the job's run method returns and it is ready to run a new job.
    * @param rthread The thread that is now ready for a new job.
    * @return No return value.
    */
   static void recycle(RecyclableThread rthread) {
      synchronized (lock) {
         if (ready.size() < cacheLimit) {
            ready.addElement(rthread);
         } else {
            rthread.stopThread();
         }
      }
   }
}
