/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.ConductorExt.adaptors;

import FMG.Conductor.dataflow.*;
import FMG.Conductor.util.*;

import java.util.*;

/**
 * This class makes a useful base for adaptors that work on actual protocols.
 * It provides the notion of the current position in the stream and includes
 * some inner classes that are useful for delineating portions of the stream.
 * <p>
 * @author Mark Yarvis (yarvis@fmg.cs.ucla.edu)
 * @version $Revision: 1.12 $      
 */

public abstract class ProtocolAdaptor extends Adaptor {

   protected DataAccessPointer currentPos = new DataAccessPointer();

   // This is a kludge!  Due to a bug, subclasses, can't refer to "window".
   protected class WindowAccessor {
      protected AdaptorWindow myWindow = getWindow();
   }

   protected class Line extends WindowAccessor {
      protected DataAccessPointer preP;
      protected DataAccessPointer postP;
      protected int len;
      protected byte[] bytes;

      protected Line() {
      }

      public Line(DataAccessPointer startPointer) throws AsyncEventException {
         super();
         try {
            // insert boundary markers
            preP = startPointer;
            postP = new DataAccessPointer();
            preP.insertPointer(postP);

            // move postP to end of line
            len = findEndOfLine(postP);
         } catch (AsyncEventException e) {
            postP.removeSelf();
            throw e;
         }
      }

      public DataAccessPointer getPrePointer() {
         return preP;
      }
      
      public DataAccessPointer getPostPointer() {
         return postP;
      }

      // returns distance moved
      protected int findEndOfLine(DataAccessPointer p)
                                              throws AsyncEventException {
         int len = p.movePastByte((byte) '\n');
         if (len == -1) {
            // oops, no more end of lines in the window
            myWindow.expand((byte) '\n');   // expand window to at least eol

            // this time, success is guaranteed
            len = p.movePastByte((byte) '\n');
            if (len == -1) {
               throw new RuntimeException("This shouldn't happen!");
            }
         }

         return len;
      }

      // for use when len is not yet valid
      protected boolean isEmptyLine(int prelimLen) {
         // empty header line is only \r\n or \n
         return (((prelimLen==1) && (preP.getByte(0) == '\n')) ||
                 ((prelimLen==2) && (preP.getByte(0) == '\r') &&
                                    (preP.getByte(1) == '\n')));
      }
      
      public boolean isEmptyLine() {
         // empty header line is only \r\n or \n
         return (((len==1) && (preP.getByte(0) == '\n')) ||
                 ((len==2) && (preP.getByte(0) == '\r') &&
                              (preP.getByte(1) == '\n')));
      }

      
      protected byte[] getBytes() {
         if (bytes == null) {
            bytes = new byte[len];
            preP.getBytes(bytes, 0, 0, len);
         }
         return bytes;
      }

      protected void replaceBytes(byte[] newBytes) {
         preP.replaceBytes(newBytes, 0, newBytes.length, postP);
         invalidateCache();
      }

      // force the entire command to be a semantic segment
      public void makeSegment() {
         replaceBytes(getBytes());
      }

      public void cacheSelf() {
         getBytes();
      }
      
      protected void invalidateCache() {
         bytes = null;
         len = preP.distanceTo(postP, Direction.up);
      }

      public String toString() {
         return Debug.printableString(getBytes());
      }

      public int length() {
         return len;
      }
   }  

   protected class ElementalLine extends Line {
      int numElements = -1;

      public ElementalLine(DataAccessPointer startPointer) 
                                             throws AsyncEventException {
         super(startPointer);
      }

      protected int getNumElements() {
         if (numElements == -1) {
            byte[] buf = getBytes();

            numElements = 0;
            int pos = Bytes.findNonWhitespace(buf, 0);
            while (pos != -1) {
               numElements++;
               pos = Bytes.findNonWhitespace(buf, 
                  Bytes.findWhitespace(buf, pos));
            }
         }

         return numElements;
      }

      protected String getElement(int elemNum) {
         byte[] buf = getBytes();
         int startPos = Bytes.findNonWhitespace(buf, 0);

         for (int i=0; i < elemNum; i++) {
            startPos = Bytes.findNonWhitespace(buf,
                                Bytes.findWhitespace(buf, startPos));
            if (startPos == -1) {
               throw new RuntimeException("Case not handled");
            }
         }

         int endPos = Bytes.findWhitespace(buf, startPos) - 1;
         if (endPos < 0) { // last arg?
            endPos = buf.length - 1;
         }

         return new String(buf, startPos, endPos - startPos + 1);
      }

      protected void replaceElement(int elemNum, byte[] newBytes) {
         byte[] buf = getBytes();
         int startPos = Bytes.findNonWhitespace(buf, 0);

         for (int i=0; i < elemNum; i++) {
            startPos = Bytes.findNonWhitespace(buf,
                                Bytes.findWhitespace(buf, startPos));
            if (startPos == -1) {
               throw new RuntimeException("Case not handled");
            }
         }

         int endPos = Bytes.findWhitespace(buf, startPos) - 1;
         if (endPos < 0) { // last arg?
            endPos = buf.length - 1;
         }

         preP.replaceBytes(newBytes, 0, newBytes.length, 
                           startPos, endPos - startPos + 1);
         invalidateCache();
      }

      protected void invalidateCache() {
         super.invalidateCache();
         numElements = -1;
      }
   }

   protected class HeaderLine extends Line {
      private String name;
      private String value;
   
      private int colonPos;
   
      HeaderLine(DataAccessPointer startPointer) throws AsyncEventException {
         super(startPointer);
   
         setColonPos();
      }

      // insert a new header before this one and return it
      HeaderLine insertHeaderLine(String newName, String newValue) {
         byte[] oldData = getBytes();

         // space for old data, new data plus ": " and "\r\n"
         byte[] newData = new byte[oldData.length + newName.length() + 
                                   newValue.length() + 4];

         /*
          * Add the new data to the old data and replace it.
          */

         int pos = 0;

         System.arraycopy(newName.getBytes(), 0, 
                          newData, pos, newName.length());
         pos += newName.length();

         newData[pos++] = (byte) ':';
         newData[pos++] = (byte) ' ';

         System.arraycopy(newValue.getBytes(), 0, 
                          newData, pos, newValue.length());
         pos += newValue.length();

         newData[pos++] = (byte) '\r';
         newData[pos++] = (byte) '\n';

         System.arraycopy(oldData, 0, 
                          newData, pos, oldData.length);

         replaceBytes(newData);

         /*
          * Create a new header line on the first half of the data and 
          * re-delimit this one on the second half.
          */

         try {
            HeaderLine newHeader = new HeaderLine(preP);
            preP = newHeader.getPostPointer();
            invalidateCache();

            return newHeader;
         } catch (AsyncEventException e) {
            throw new RuntimeException("This shouldn't happen");
         }
      }
   
      private void setColonPos() {
         colonPos = preP.distanceTo((byte) ':');
         if (colonPos >= len) {   // be sure it's within this header
            colonPos = -1;
         }
      }
   
      private boolean isSpaceOrTab(byte b) {
         return ((b == ' ') || (b == '\t'));
      }

      // returns distance moved
      protected int findEndOfLine(DataAccessPointer p)
                                               throws AsyncEventException {
         int dist = 0;
   
         do {
            dist += super.findEndOfLine(p);
   
            // disallow continuation on empty lines
            if (isEmptyLine(dist)) {
               return dist;
            }
   
            // make sure there's at least 1 byte available
            if (myWindow.countRemaining(p) < 1) {
               myWindow.expand(1);
            }
         } while (isSpaceOrTab(p.getByte(0))); // handle continuation
   
         return dist;
      }
   
      String getName() {
         if (name == null) {
            if (colonPos != -1) {
               byte[] nameBytes = new byte[colonPos];
               preP.getBytes(nameBytes, 0, 0, colonPos);
               name = new String(nameBytes);
            } else {
               name = "";
            }
         }
         return name;
      }
   
      String getValue() {
         if (value == null) {
            if (colonPos != -1) {
               byte[] bytes = getBytes();
   
               // ignore leading and trailing whitespace
               int valueStart = Bytes.findNonWhitespace(bytes, colonPos+1);
               int valueEnd = Bytes.revfindNonWhitespace(bytes,
                                                         bytes.length - 1);

               value = new String(bytes, valueStart,
                                  valueEnd - valueStart + 1);
            } else {
               value = "";
            }
         }
         return value;
      }

      public void cacheSelf() {
         getName();
         getValue();
      }
   
      void replaceValue(String newValue) {
         if (colonPos == -1) {
            throw new RuntimeException("This shouldn't happen");
         }
         byte[] newBytes = newValue.getBytes();
   
         // ignore leading and trailing whitespace
         byte[] oldBytes = getBytes();
         int oldStart = Bytes.findNonWhitespace(oldBytes, colonPos+1);
         int oldEnd = Bytes.revfindNonWhitespace(oldBytes, oldBytes.length-1);
   
         preP.replaceBytes(newBytes, 0, newBytes.length,
                           oldStart, oldEnd - oldStart + 1);
   
         invalidateCache();
      }
   
      void deleteProp() {
         byte[] buf = new byte[0];
         preP.replaceBytes(buf, 0, 0, 0, len);
         invalidateCache();
      }
   
      protected void invalidateCache() {
         super.invalidateCache();
         name = null;
         value = null;
         setColonPos();
      }
   }

   protected class Header extends WindowAccessor {
      protected Vector headerLines = new Vector();

      public Header() {
      }

      public Header(DataAccessPointer currentPos) throws AsyncEventException {
         this();
         readHeader(currentPos);
      }

      public void readHeader(DataAccessPointer currentPos) 
                                                 throws AsyncEventException {
         headerLines.removeAllElements();

         DataAccessPointer endOfPrevious = new DataAccessPointer();
         currentPos.insertPointer(endOfPrevious);

         try {
            // get headers
            while (true) {
               HeaderLine header = new HeaderLine(endOfPrevious);
               endOfPrevious = header.getPostPointer();

               headerLines.addElement(header);

               if (header.isEmptyLine()) {
                  break;
               }
            }
         } catch (AsyncEventException e) {
            removeSelf();
            throw e;
         }
      }

      public DataAccessPointer getPrePointer() {
         HeaderLine h = (HeaderLine) headerLines.firstElement();
         return h.getPrePointer();
      }

      public DataAccessPointer getPostPointer() {
         HeaderLine h = (HeaderLine) headerLines.lastElement();
         return h.getPostPointer();
      }

      int length() {
         return getPrePointer().distanceTo(getPostPointer());
      }

      public void cacheSelf() {
         Enumeration e = headerLines.elements();
         while (e.hasMoreElements()) {
            ((HeaderLine) e.nextElement()).cacheSelf();
         }
      }

      public void removeSelf() {
         Enumeration e = headerLines.elements();
         while (e.hasMoreElements()) {
            HeaderLine h = (HeaderLine) e.nextElement();
            // pull all pre-pointers
            h.getPrePointer().removeSelf();

            // pull last post-pointer
            if (! e.hasMoreElements()) {
               h.getPostPointer().removeSelf();
            }
         }
      }

      // ensure that all of the listed lines are in the same semantic segment
      private void createSemanticSegment(Vector lines) {
         // find the first and last pointers
         DataAccessPointer firstPointer = null;
         DataAccessPointer lastPointer = null;
         int firstPos = Integer.MAX_VALUE;
         int lastPos = Integer.MIN_VALUE;
         for (int i=0; i<lines.size(); i++) {
            Line l = ((Line) lines.elementAt(i));
            int pos = window.getPosition(l.getPrePointer());
            if (pos < firstPos) {
               firstPos = pos;
               firstPointer = l.getPrePointer();
            }
            pos = window.getPosition(l.getPostPointer());
            if (pos > lastPos) {
               lastPos = pos;
               lastPointer = l.getPostPointer();
            }
         }
   
         // create a block mod pointer and move it from the first to the last
         BlockModPointer p = new BlockModPointer();
         firstPointer.insertPointer(p);
         p.blockModStart();
         p.advance(p.distanceTo(lastPointer));
         p.blockModEnd();
         p.removeSelf();
      }

      public Enumeration getHeaderNames() {
         // interested in all but the last (empty) one
         int numHeaders = headerLines.size() - 1; 
         Vector names = new Vector(numHeaders);

         for (int i=0; i<numHeaders; i++) {
            names.addElement(((HeaderLine) headerLines.elementAt(i)).getName());
         }
         return names.elements();
      }

      public String getHeaderValue(String name) {
         StringBuffer result = new StringBuffer();
         boolean resultFound = false;

         Enumeration e = headerLines.elements();
         while (e.hasMoreElements()) {
            HeaderLine h = (HeaderLine) e.nextElement();
            if (h.getName().equalsIgnoreCase(name)) {
               if (resultFound) {
                  result.append(',');
               }
               result.append(h.getValue());
               resultFound=true;
            }
         }

         if (resultFound) {
            return result.toString();
         } else {
            return null;
         }
      }

      // this func can only be used in between readHeaders() and writeHeaders()
      // return true iff such a header was found, false if a new one was added
      public boolean replaceHeaderValue(String name, String newValue) {
         Vector headersModified = new Vector();

         Enumeration e = headerLines.elements();
         while (e.hasMoreElements()) {
            HeaderLine h = (HeaderLine) e.nextElement();
            if (h.getName().equalsIgnoreCase(name)) {
               if (headersModified.size() == 0) {
                  h.replaceValue(newValue);
               } else {
                  h.deleteProp();
               }
               headersModified.addElement(h);
            }
         }

         if (headersModified.size() == 0) {
            addHeader(name, newValue);
         } else if (headersModified.size() > 1) {
            /*
             * When modifying a prop that has been spefied using multiple
             * entries for the same name, we really should make a semantic
             * segment around them (or around all the headers if needed
             */
            createSemanticSegment(headersModified);
         }
         return (headersModified.size() != 0);
      }

      public void addHeader(String name, String value) {
         // get blank header line at the end
         HeaderLine lastHeader = 
            (HeaderLine) headerLines.lastElement();

         // create a new header line usiung the last one
         HeaderLine newHeaderLine = lastHeader.insertHeaderLine(name, value);

         // augment our list with the new element
         headerLines.insertElementAt(newHeaderLine, headerLines.size() -1);
      }

      public boolean deleteHeader(String name) {
         Vector headersRemoved = new Vector();

         Enumeration e = headerLines.elements();
         while (e.hasMoreElements()) {
            HeaderLine h = (HeaderLine) e.nextElement();
            if (h.getName().equalsIgnoreCase(name)) {
               h.deleteProp();
               headersRemoved.addElement(h);
            }
         }

         if (headersRemoved.size() > 1) {
            /*
             * When modifying a prop that has been spefied using multiple
             * entries for the same name, we really should make a semantic
             * segment around them (or around all the headers if needed
             */
            createSemanticSegment(headersRemoved);
         }
         return (headersRemoved.size() != 0);
      }

   }

   public ProtocolAdaptor(AdaptorWindow window, String parameter) {
      super(window, parameter);
   }

   // This is a kludge!  Due to a bug, subclasses, can't refer to "window".
   private AdaptorWindow getWindow() {
      return window;
   }

   public void handleFlow() {
      window.insertPointer(currentPos);
      handleProtocol();
      currentPos.removeSelf();
   }

   protected abstract void handleProtocol();

}
