/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.ConductorExt.adaptors;

import java.io.*;

import FMG.Conductor.util.*;
import FMG.Conductor.comm.*;
import FMG.Conductor.dataflow.*;

/**
 * This adaptor saves the data stream into a file named 
 * /tmp/StreamSaver.out.<N>, where <N> is chosen to make the filename
 * unique.
 * <p>
 * @author Mark Yarvis (yarvis@fmg.cs.ucla.edu)
 * @version $Revision: 1.5 $      
 */

public class StreamSaver extends Adaptor {

   private static String FILE_PREFIX = "/tmp/StreamSaver.out.";

   protected FileOutputStream out;

   public StreamSaver(AdaptorWindow window, String parameter) {
      super(window, parameter);
      File outFile = pickNewFile();
      if (Debug.on) {
         Debug.println(Debug.ADAPTOR, Debug.INFO, 
                       this + " saving to file: " + outFile);
      }

      try {
         out = new FileOutputStream(outFile);
      } catch (IOException e) {
         throw new RuntimeException("Case not handled");
      }
   }

   public File pickNewFile() {
      File file;

      for (int i = 0; true; i++) {
         file = new File(FILE_PREFIX + i);
         if (! file.exists()) {
            return file;
         }
      }
   }

   public void handleFlow() {
      try {
         if (Debug.on) {
            Debug.print(Debug.ADAPTOR, Debug.INFO, "(");
         }

         byte[] buf = new byte[500];

         DataAccessPointer p = new DataAccessPointer();
         window.insertPointer(p);

         while (true) {
            try {
               int len = window.expand();
               if (Debug.on) {
                  Debug.println(Debug.ADAPTOR, Debug.DETAIL,
                                this + " saving " + len + " bytes");
               }

               while (len > 0) {
                  int writeLen = (len < buf.length ? len : buf.length);
                  p.getBytes(buf, 0, 0, writeLen);
                  out.write(buf, 0, writeLen);
                  p.advance(writeLen);
                  window.contract(p);
                  len -= writeLen;
               }
            } catch (AsyncEventException e) {
               if (window.inputClosed() || window.outputClosed()) {
                  if (Debug.on) {
                     Debug.print(Debug.ADAPTOR, Debug.INFO, ")");
                  }
                  p.removeSelf();
                  out.close();
                  window.closeOutput();
                  return;
               }
               if (window.switchPlanInWindow()) {
                  throw new RuntimeException("Case not handled");
               }
            }
         }
      } catch (IOException e) {
         throw new RuntimeException("Case not handled");
      }
   }
}
