/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.ConductorExt.adaptors.http;

import FMG.Conductor.dataflow.*;
import FMG.Conductor.util.Bytes;

/**
 * This class provides a useful base class for adaptors that operate
 * on HTML data transmitted over HTTP.  It provides access to the
 * HTTP header and inner classes that help find and modify tags in the
 * data stream.
 * <p>
 * TODO: These cases are not handled properly:
 * <ol>
 *    <li>Spaces around the equals sign in an attribute
 *    <li>Leading spaces before the tag name (both in terms of finding
 *        an attribute value and in terms of identifying attribute names)
 *    <li>A greater than sign in a (quoted) attribute value
 * </ol>
 * The later issue is particularly hard because we're only guaranteed to
 * be have data up to the first greater than sign in the window!
 * <p>
 * TODO: Perhaps it would be easiest to parse all the attributes if any
 * of the attributes is requested.
 * <p>
 * @author Mark Yarvis (yarvis@fmg.cs.ucla.edu)
 * @version $Revision: 1.11 $      
 */

public class HtmlAdaptor extends HttpResponseAdaptor {

   protected class TagParser {
      private DataAccessPointer p;
      private byte[] bytes;
      private String name;

      private class AttrAccess {
         int nameStart;
         int nameEnd;
         int valueStart;
         int valueEnd;
         boolean quotedValue;
         boolean noValue;

         AttrAccess(byte[] bytes, int startPos) {
            nameStart = startPos;
            int spacePos = Bytes.findWhitespace(bytes, startPos);
            int equalsPos = Bytes.findByte(bytes, '=', startPos);

            noValue = ((equalsPos == -1) || ((spacePos != -1) && 
                                      (equalsPos > spacePos)));

            if (noValue) {
               if (spacePos == -1) {   // this is the last attribute
                  nameEnd = bytes.length - 1;
               } else {                // there are more attributes
                  nameEnd = spacePos;
               }
            } else {
               nameEnd = equalsPos-1;
               valueStart = equalsPos+1;
               quotedValue = ((bytes[valueStart] == '"') || 
                              (bytes[valueStart] == '\''));
               if (quotedValue) {
                  valueEnd = Bytes.findByte(bytes, bytes[valueStart], 
                                            valueStart+1);
               } else {
                  valueEnd = spacePos-1;
               }

               if (valueEnd < 0) {
                  valueEnd = bytes.length-1;
               }
            }
         }
      }

      protected TagParser() {
      }

      protected TagParser(DataAccessPointer p) {
         this();
         setTagStart(p);
      }

      // reset any cached info and parse at the given position
      protected void setTagStart(DataAccessPointer p) {
         this.p = p;
         getBytes();
         name = null;
      }

      protected String getName() {
         if (name == null) {
            // name len is same as position of first space
            int nameLen = Bytes.findWhitespace(bytes, 0);

            if (nameLen == -1) {  // no spaces => no attributes
               nameLen = bytes.length;
            }

            name = new String(bytes, 0, nameLen);
         }

         return name;
      }

      private void getBytes() {
         int len = p.distanceTo((byte) '>');
         bytes = new byte[len-1];
         p.getBytes(bytes, 0, 1, len-1);
      }

      private AttrAccess findAttr(String attrName) {
         byte[] name = attrName.getBytes();

         int start = 0;

         while (true) {
            start = Bytes.findNonWhitespace(bytes, 
                       Bytes.findWhitespace(bytes, start));
            if (start == -1) {          // no more attributes
               return null;
            }

            if (Bytes.matchIgnoreCase(name, 0, bytes, start, name.length) &&
                (bytes[start + name.length] == (byte) '=')) {
               return new AttrAccess(bytes, start);
            }
         }
      }

      protected String getAttrValue(String attrName) {
         AttrAccess aa = findAttr(attrName);
         if (aa == null) {
            return null;
         }

         if (aa.noValue) {
            return "";
         }

         if (aa.quotedValue) {
            return new String(bytes, aa.valueStart + 1, 
                                     aa.valueEnd - aa.valueStart - 1);
         } else {
            return new String(bytes, aa.valueStart, 
                           aa.valueEnd - aa.valueStart + 1);
         }
      }

      protected int replaceAttrValue(String attrName, String newVal) {
         return removeAttr(attrName) + addAttr(attrName, newVal);
      }

      protected int addAttr(String attrName, String attrVal) {
         byte[] nameBytes = attrName.getBytes();
         byte[] valBytes = attrVal.getBytes();

         // name_len + val_len + 1 space byte + 1 equals sign + 2 quotes
         int sizeChange = nameBytes.length + valBytes.length + 4;

         byte[] newBytes = new byte[bytes.length + sizeChange];

         // copy the old stuff
         System.arraycopy(bytes, 0, newBytes, 0, bytes.length);

         // add a space
         int pos = bytes.length;
         newBytes[pos++] = (byte) ' ';

         // copy in the name
         System.arraycopy(nameBytes, 0, newBytes, pos, nameBytes.length);
         pos += nameBytes.length;

         // add an equals and a start quote
         newBytes[pos++] = (byte) '=';
         newBytes[pos++] = (byte) '"';

         // copy in the value
         System.arraycopy(valBytes, 0, newBytes, pos, valBytes.length);
         pos += valBytes.length;

         // add a final quote
         newBytes[pos] = (byte) '"';

         p.replaceBytes(newBytes, 0, newBytes.length, 1, bytes.length);
         bytes = newBytes;
         return sizeChange;
      }

      protected int removeAttr(String attrName) {
         AttrAccess aa = findAttr(attrName);

         if (aa==null) {
            return 0;
         }

         int startPos = aa.nameStart;
         int endPos = (aa.noValue ? aa.nameEnd : aa.valueEnd);

         int sizeChange = 0 - (endPos - startPos + 1);

         byte[] newBytes = new byte[bytes.length + sizeChange];

         // copy all before attr
         System.arraycopy(bytes, 0, newBytes, 0, startPos);
         // copy all after attr
         System.arraycopy(bytes, endPos+1, newBytes, startPos, 
                          bytes.length-endPos-1);

         p.replaceBytes(newBytes, 0, newBytes.length, 1, bytes.length);
         bytes = newBytes;
         return sizeChange;
      }

   }

   public HtmlAdaptor(AdaptorWindow window, String parameter) {
      super(window, parameter);
   }

   /*
    * advances to before the next tag
    */
   protected final int advanceToTag(DataAccessPointer p, int maxLen) 
                                                throws AsyncEventException {
      int distanceTraveled = 0;
      int available = window.countRemaining(p);

      while (true) {
         int distToLT = p.distanceTo((byte) '<');

         if (distToLT == -1) {   // it's not in the window
            if (available < maxLen - distanceTraveled) {  // keep looking
               p.advance(available);
               distanceTraveled += available;
               available = window.expand();
            } else {        // didn't find it within the specified distance
               p.advance(maxLen - distanceTraveled);
               return maxLen;            
            }
         } else if (distToLT > maxLen - distanceTraveled) {
            // didn't find it within maxLen
            p.advance(maxLen - distanceTraveled);
            return maxLen;
         } else {
            // found it
            p.advance(distToLT);
            distanceTraveled += distToLT;
            break;
         }
      }

      // get the end marker in the window
      // HACK: assume it's within the length of the document
      if (p.distanceTo((byte) '>') == -1) {
         window.expand((byte) '>');
      }

      return distanceTraveled;
   }

   protected final int advanceToTag(DataAccessPointer p) 
                                               throws AsyncEventException {
      return advanceToTag(p, Integer.MAX_VALUE);
   }

}
