/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.ConductorExt.adaptors.http;

import FMG.ConductorExt.adaptors.tools.PriorityWriter;
import FMG.Conductor.dataflow.*;

/**
 * This adaptor uses the PriorityWriter utility to prioritize HTTP fetches
 * based on data size.  Smaller content is given priority over larger 
 * content.
 * <p>
 * Options to this adaptor take the form <code>[delay [size-margin]]</code>
 * where <code>delay</code> is the amount of time to wait before allowing 
 * lower-priority traffic to flow after higher priority traffic has stopped,
 * and <code>size-margin</code> is the size difference that must be present
 * before the delay will occur.
 * <p>
 * @author Mark Yarvis (yarvis@fmg.cs.ucla.edu)
 * @version $Revision: 1.10 $      
 */

public class HttpPrioritizer extends HttpResponseAdaptor {

   private static final long DEFAULT_DELAY = 5000;
   private static final int DEFAULT_MARGIN = 1000000;

   private long delay = DEFAULT_DELAY;

   private int margin = DEFAULT_MARGIN;

   PriorityWriter priorityWriter;

   public HttpPrioritizer(AdaptorWindow window, String parameter) {
      super(window, parameter);

      if (parameter != null) {
         int spacePos = parameter.indexOf(' ');
         if (spacePos == -1) {  // only one param?
            delay = Long.parseLong(parameter);
         } else {               // more than one param
            delay = Long.parseLong(parameter.substring(0, spacePos));
            margin = Integer.parseInt(parameter.substring(spacePos+1, 
                                                          parameter.length()));
         }
      }
   }

   protected void handleContent(int contentLength, String contentType, 
                                                   String contentEncoding) 
                                               throws AsyncEventException {
      // figure the length including headers
      int totalLen = contentLength + 
                       window.getPosition(header.getPostPointer());

      // remove header pointers
      firstLine.getPrePointer().removeSelf();
      firstLine.getPostPointer().removeSelf();
      header.removeSelf();

      // write the data at the correct priority level
      priorityWriter.setPriority(-totalLen);
      priorityWriter.writeData(totalLen);
   }

   public void handleFlow() {
      priorityWriter = new PriorityWriter(window, 0, 
                                          delay, margin, getNodeIAC());
      window.requestDownstreamBuffering(AdaptorWindow.MIN_BUFFERING);
      super.handleFlow();
   }
}
