/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.ConductorExt.adaptors.http;

import java.util.*;
import FMG.Conductor.dataflow.*;
import FMG.Conductor.util.*;

/**
 * This class provides a useful base class for adaptors that operate
 * on the request (upstream) part of HTTP streams.  It reads each
 * request header and calls the handleRequest() method handle the request.
 * The base implementation just forwards the request; override this method
 * to make a useful adaptor.
 * <p>
 * @author Mark Yarvis (yarvis@fmg.cs.ucla.edu)
 * @version $Revision: 1.12 $      
 */

public class HttpRequestAdaptor extends HttpAdaptor {

   protected RequestLine requestLine;

   // This class assumes that the target can have spaces, and the version
   // can not!
   protected class RequestLine extends Line {
      private String command;
      private String target;
      private String version;

      RequestLine(DataAccessPointer startPointer) throws AsyncEventException {
         super(startPointer);
      }

      public String getCommand() {
         if (command == null) {
            byte[] buf = getBytes();
            int commandLen = Bytes.findWhitespace(buf, 0);

            command = new String(buf, 0, commandLen);
         }
         return command;
      }

      public void replaceCommand(String newCommand) {
         replaceCommand(newCommand.getBytes());
      }

      public void replaceCommand(byte[] newCommandBytes) {
         byte[] buf = getBytes();
         int commandLen = Bytes.findWhitespace(buf, 0);

         preP.replaceBytes(newCommandBytes, 0, newCommandBytes.length, 
                           0, commandLen);

         // a little overkill, but basically a good thing
         invalidateCache();
      }

      public String getTarget() {
         if (target == null) {
            byte[] buf = getBytes();

            // find byte after first space
            int firstByte = Bytes.findNonWhitespace(buf, 
                                           Bytes.findWhitespace(buf, 0));

            // find byte before last space
            int lastByte = Bytes.revfindNonWhitespace(buf, 
                              Bytes.revfindWhitespace(buf, 
                                 Bytes.revfindNonWhitespace(buf, 
                                                            buf.length - 1)));

            target = new String(buf, firstByte, lastByte-firstByte+1);
         }
         return target;
      }

      public void replaceTarget(String newTarget) {
         replaceTarget(newTarget.getBytes());
      }

      public void replaceTarget(byte[] newTargetBytes) {
         byte[] buf = getBytes();

         // find byte after first space
         int firstByte = Bytes.findNonWhitespace(buf, 
                                          Bytes.findWhitespace(buf, 0));

         // find byte before last space
         int lastByte = Bytes.revfindNonWhitespace(buf, 
                           Bytes.revfindWhitespace(buf, 
                              Bytes.revfindNonWhitespace(buf, 
                                                         buf.length - 1)));

         preP.replaceBytes(newTargetBytes, 0, newTargetBytes.length,
                           firstByte, lastByte - firstByte + 1);

         // a little overkill, but basically a good thing
         invalidateCache();
      }

      public String getVersion() {
         if (version == null) {
            byte[] buf = getBytes();

            // find byte after last space
            int firstByte = Bytes.revfindWhitespace(buf, 
                               Bytes.revfindNonWhitespace(buf, 
                                                          buf.length - 1)) + 1;

            version = new String(buf, firstByte, buf.length - firstByte);
         }
         return version;
      }

      public void replaceVersion(String newVersion) {
         replaceVersion(newVersion.getBytes());
      }

      public void replaceVersion(byte[] newVersionBytes) {
         byte[] buf = getBytes();

         // find byte after last space
         int firstByte = Bytes.revfindWhitespace(buf, 
                            Bytes.revfindNonWhitespace(buf, 
                                                       buf.length - 1)) + 1;

         preP.replaceBytes(newVersionBytes, 0, newVersionBytes.length,
                           firstByte, buf.length - firstByte);

         // a little overkill, but basically a good thing
         invalidateCache();
      }

      public void cacheSelf() {
         getCommand();
         getTarget();
         getVersion();
      }

      protected void invalidateCache() {
         super.invalidateCache();
         command = null;
         target = null;
         version = null;
      }
   }

   public HttpRequestAdaptor(AdaptorWindow window, String parameter) {
      super(window, parameter);
   }

   protected DataAccessPointer readFirstLine(DataAccessPointer endOfPrevious)
                                                 throws AsyncEventException {
      // get request line
      requestLine = new RequestLine(endOfPrevious);

      firstLine = requestLine;

      return requestLine.getPostPointer();
   }

   protected void handleProtocol() {
      while (true) {
         try {
            if (Debug.on) {
               Debug.println(Debug.ADAPTOR, Debug.MOREINFO,
                             this + ": reading headers");
            }

            readHeaders();

            if (Debug.on) {
               Enumeration names = header.getHeaderNames();
               while (names.hasMoreElements()) {
                  String name = (String) names.nextElement();
                  Debug.println(Debug.ADAPTOR, Debug.EXTRAINFO,
                                this + ": Header: " +
                                name + "=" + header.getHeaderValue(name));
               }

               Debug.println(Debug.ADAPTOR, Debug.MOREINFO, this +
                             ": handling content");
            }

            handleRequest(requestLine.getTarget());
         } catch (OutputClosedException e) {
            Debug.println(Debug.ADAPTOR, Debug.INFO,
                          this + ": output closed unexpectedly");
            return;
         } catch (InputClosedException e) {
            if (Debug.on) {
               Debug.print(Debug.ADAPTOR, Debug.INFO, ")");
            }
            window.contract();
            window.closeOutput();
            Debug.println(Debug.ADAPTOR, Debug.INFO, this + ": closed");
            return;
         } catch (PlanChangedException e) {
            throw new RuntimeException("Case not handled.");
         } catch (AsyncEventException e) {
             Debug.println(Debug.ADAPTOR, Debug.WARNING,
                           this + ": Got an unexpected exception");
             return;
         }
      }
   }

   // default implementation, do nothing
   public void handleRequest(String url) {
      if (Debug.on) {
         Debug.println(Debug.ADAPTOR, Debug.INFO, 
                       this + ": Got request for " + url);
      }

      writeHeaders();
   }
}
