/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.ConductorExt.adaptors.http;

import java.util.*;
import FMG.Conductor.dataflow.*;
import FMG.Conductor.util.*;

/**
 * This class provides a useful base class for adaptors that operate
 * on the response (downstream) part of HTTP streams.  It reads each
 * response header and calls the handleContent() method to handle any
 * resulting content.  The default implementation does nothing with
 * the content; override this method to write a useful adaptor.
 * <p>
 * @author Mark Yarvis (yarvis@fmg.cs.ucla.edu)
 * @version $Revision: 1.25 $      
 */

public class HttpResponseAdaptor extends HttpAdaptor {

   // valid only after a call to readHeaders() and before a call to 
   // writeHeaders()
   protected StatusLine statusLine;

   protected class StatusLine extends Line {

      private String version;
      private byte[] code;
      private String reason;

      StatusLine(DataAccessPointer startPointer) throws AsyncEventException {
         super(startPointer);
      }

      public String getVersion() {
         if (version == null) {
            byte[] buf = getBytes();

            // find first space
            int versionLen = Bytes.findWhitespace(buf, 0);

            if (Debug.on && (versionLen == -1)) {
               throw new RuntimeException("This shouldn't happen");
            }

            version = new String(buf, 0, versionLen);
         }
         return version;
      }

      public void replaceVersion(String newVersion) {
         replaceVersion(newVersion.getBytes());
      }

      public void replaceVersion(byte[] newVersionBytes) {
         byte[] buf = getBytes();

         // find first space
         int versionLen = Bytes.findWhitespace(buf, 0);

         preP.replaceBytes(newVersionBytes, 0, newVersionBytes.length,
                           0, versionLen);

         invalidateCache();
      }

      public byte[] getCode() {
         if (code == null) {
            byte[] buf = getBytes();

            // find first byte of second field
            int firstByte = Bytes.findNonWhitespace(buf, 
                                     Bytes.findWhitespace(buf, 0));

            code = new byte[3];
            code[0]=buf[firstByte];
            code[1]=buf[firstByte+1];
            code[2]=buf[firstByte+2];
         }
         return code;
      }

      public void replaceCode(byte[] newCodeBytes) {
         byte[] buf = getBytes();

         // find first byte of second field
         int firstByte = Bytes.findNonWhitespace(buf, 
                                  Bytes.findWhitespace(buf, 0));

         preP.replaceBytes(newCodeBytes, 0, 3, firstByte, 3);

         invalidateCache();
      }

      public String getReason() {
         if (reason == null) {
            byte[] buf = getBytes();

            // find first byte of second field
            int firstByte = Bytes.findNonWhitespace(buf, 
                                            Bytes.findWhitespace(buf, 0));
            // find first byte of third field
            firstByte = Bytes.findNonWhitespace(buf, 
                                        Bytes.findWhitespace(buf, firstByte));

            // stip off newline
            int lastByte = Bytes.revfindNonWhitespace(buf, buf.length - 1);

            reason = new String(buf, firstByte, lastByte - firstByte + 1);
         }
         return reason;
      }

      public void replaceReason(String newReason) {
         replaceReason(newReason.getBytes());
      }

      public void replaceReason(byte[] newReasonBytes) {
         byte[] buf = getBytes();

         // find first byte of second field
         int firstByte = Bytes.findNonWhitespace(buf, 
                                         Bytes.findWhitespace(buf, 0));
         // find first byte of third field
         firstByte = Bytes.findNonWhitespace(buf, 
                                     Bytes.findWhitespace(buf, firstByte));

         // stip off newline
         int lastByte = Bytes.revfindNonWhitespace(buf, buf.length - 1);

         preP.replaceBytes(newReasonBytes, 0, newReasonBytes.length,
                           firstByte, lastByte - firstByte + 1);

         invalidateCache();
      }

      public void cacheSelf() {
         getVersion();
         getCode();
         getReason();
      }

      protected void invalidateCache() {
         super.invalidateCache();
         code = null;
         version = null;
         reason = null;
      }
   }

   public HttpResponseAdaptor(AdaptorWindow window, String parameter) {
      super(window, parameter);
   }

   // currentPos now points the the spot after the last \n or \r\n
   protected DataAccessPointer readFirstLine(DataAccessPointer endOfPrevious) 
                                                  throws AsyncEventException {
      // get status line
      statusLine = new StatusLine(endOfPrevious);

      if (Debug.on && (Debug.getLevel(Debug.ADAPTOR) > Debug.EXTRAINFO)) {
         byte[] code = statusLine.getCode();
         Debug.println(this + ": Status is " + 
                       (char) code[0] + (char) code[1] + (char) code[2]);
      }

      firstLine = statusLine;

      return statusLine.getPostPointer();
   }

   // the formatshould be : HTTP/1*D.1*D SP NNN SP
   // we assume it's all in the buffer!
   private boolean isStartOfStatusLine(byte[] buf) {
      if ((buf[0] != 'H') || (buf[1] != 'T') || 
          (buf[2] != 'T') || (buf[3] != 'P') || (buf[4] != '/')) {
         return false;
      }

      // look for a version number
      int i = 5;
      while (buf[i] != ' ') {
         if (i >= buf.length) {
            return false;
         }
         if ((buf[i] != '.') && ((buf[i]<'0') || (buf[i]>'9'))) {
            return false;
         }
         i++;
      }

      i++;  // skip the space

      // look for the 3-digit numeric code
      for (int j=0; j<3; j++) {
         if (i+j >= buf.length) {
            return false;
         }
         if ((buf[i+j] < '0') || (buf[i+j] > '9')) {
            return false;
         }
      }

      i+=3;  // skip the code

      // there should be a space after the code
      if ((i < buf.length) && (buf[i] != ' ')) {  
         return false;
      }

      return true;
   }

   // this is a KLUDGE ... this is hard or impossible to do right
   private void findResponseStart() throws AsyncEventException {
      while (true) {
         int hPos = currentPos.distanceTo((byte) 'H');
         if (hPos == -1) {
            window.contract();
            window.expand();
         } else {
            currentPos.advance(hPos);
            window.contract(currentPos);

            /* 
             * Get "enough" bytes assuming that the version numbers have 
             * more than two digits in each position.
             */
            int windowSize = window.size();
            if (windowSize < 15) {
               window.expand(15);
            }

            byte[] buf = new byte[15];
            currentPos.getBytes(buf, 0, 0, 15);

            if (isStartOfStatusLine(buf)) {
               return;
            }

            currentPos.advance(1);
         }
      }
   }

   protected void handleProtocol() {
      boolean firstTimeThrough = true;

      while (true) {
         try {
            if (firstTimeThrough) {
               /**
                * When the adaptor is started up, we may not be at the start 
                * of the stream (due to re-planning), so we have to find a 
                * good place to start adapting!
                */
               findResponseStart();
               firstTimeThrough = false;
            }

            if (Debug.on) {
               Debug.println(Debug.ADAPTOR, Debug.MOREINFO,
                             this + ": reading headers");
            }

            readHeaders();

            if (Debug.on) {
               Enumeration names = header.getHeaderNames();
               while (names.hasMoreElements()) {
                  String name = (String) names.nextElement();
                  Debug.println(Debug.ADAPTOR, Debug.EXTRAINFO,
                                this + ": Header: " +
                                name + "=" + header.getHeaderValue(name));
               }

               Debug.println(Debug.ADAPTOR, Debug.MOREINFO, 
                             this + ": handling content");
            }

            String contentType = header.getHeaderValue("Content-Type");
            String contentEncoding = header.getHeaderValue("Content-Encoding");
            if (Debug.on) {
               Debug.println(Debug.ADAPTOR, Debug.MOREINFO, 
                             this + ": Content type is " + contentType);
            }

            if (contentType == null) {
               writeHeaders();
               continue;
            }

            String contentLengthStr = header.getHeaderValue("Content-Length");
            int contentLength = (contentLengthStr == null ?
                                      Integer.MAX_VALUE :
                                      Integer.parseInt(contentLengthStr));

            // make string values lower case
            contentType = (contentType == null ? null : 
                                                 contentType.toLowerCase());
            contentEncoding = (contentEncoding == null ? null : 
                                                contentEncoding.toLowerCase());

            handleContent(contentLength, contentType, contentEncoding);

         } catch (OutputClosedException e) {
            Debug.println(Debug.ADAPTOR, Debug.INFO,
                          this + ": output closed unexpectedly");
            return;
         } catch (InputClosedException e) {
            if (Debug.on) {
               Debug.print(Debug.ADAPTOR, Debug.INFO, ")");
            }
            window.contract();
            window.closeOutput();
            Debug.println(Debug.ADAPTOR, Debug.INFO,
                          this +": closed");
            return;
         } catch (PlanChangedException e) {
            throw new RuntimeException("Case not handled.");
         } catch (RollbackOccurredException e) {
            if (Debug.on) {
               Debug.println(this + ": rollback occurred.");
            }
            firstTimeThrough = true;
            currentPos.unlink();
            window.insertPointer(currentPos);
         } catch (AsyncEventException e) {
             Debug.println(Debug.ADAPTOR, Debug.WARNING,
                           this + ": Got an unexpected exception: " + e);
             return;
         }
      }
   }

   protected void handleContent(int contentLength, String contentType, 
                                                   String contentEncoding) 
                                               throws AsyncEventException {
      writeHeaders();
      forwardContent(contentLength);
   }

   // the default content handler
   protected void forwardContent(int contentLen) throws AsyncEventException {
      int availableBytes = window.countRemaining(currentPos);

      while (contentLen > 0) {
         if (availableBytes == 0) {
            availableBytes = window.expand();
         }

         if (contentLen < availableBytes) {
            availableBytes = contentLen;
         }

         contentLen -= availableBytes;

         currentPos.advance(availableBytes);
         window.contract(currentPos);
         availableBytes = 0;
      }
   }  

}
