/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.ConductorExt.adaptors.http;

import java.util.Hashtable;
import FMG.Conductor.dataflow.*;
import FMG.Conductor.util.*;

/**
 * This adaptor caches the sizes of images that pass by and uses this 
 * information to augment future HTML pages.  When an image size is known,
 * width and height attributes are added to IMG tags.
 * <p>
 * This adaptor should be used with a URLIdentifier adaptor on the same node.
 *
 * @author Mark Yarvis (yarvis@fmg.cs.ucla.edu)
 * @version $Revision: 1.5 $
 */

public class ImageSizeCache extends HtmlAdaptor {

   Hashtable sizeCache;
   IAC iac;

   protected static final String CACHE_KEY = "ImageSizeCache";

   private class Dimensions {
      int width;
      int height;

      Dimensions(int width, int height) {
         this.width = width;
         this.height = height;
      }
   }

   public ImageSizeCache(AdaptorWindow window, String parameter) {
      super(window, parameter);
 
      iac = getStreamIAC();

      IAC nodeIAC = getNodeIAC();

      sizeCache = (Hashtable) nodeIAC.lookup(CACHE_KEY);
      if (sizeCache == null) {
         nodeIAC.store(CACHE_KEY, new Hashtable(), false);
         // avoid the race condition and use whatever is actually in the IAC
         sizeCache = (Hashtable) nodeIAC.lookup(CACHE_KEY);
      }
   }

   protected void handleContent(int contentLength, String contentType, 
                                                   String contentEncoding) 
                                               throws AsyncEventException {
      if (contentEncoding == null) {
         if (contentType.startsWith("image/gif")) {
            writeHeaders();
            String currentURL = (String) iac.lookup(URLIdentifier.COMM_TAG);
            // cache image size
            cacheGifImageSize(currentURL);
         } else if (contentType.startsWith("text/html")) {
            // add cached sizes to IMG tags
            String contentLengthStr = header.getHeaderValue("Content-Length");
            if (header.getHeaderValue("Content-Length") == null) {
               writeHeaders();
               addImageSizes(contentLength);
            } else {
               contentLength += addImageSizes(contentLength);
               header.replaceHeaderValue("Content-Length", "" + contentLength);
               writeHeaders();
            }
         } else {
            writeHeaders();
         }
      } else {
         writeHeaders();
      }

      forwardContent(contentLength);
   }

   protected void cacheGifImageSize(String currentURL) 
                                             throws AsyncEventException {
      int availableBytes = window.countRemaining(currentPos);
      if (availableBytes < 10) {
         window.expand(10-availableBytes);
      }

      byte[] buf = new byte[10];

      currentPos.getBytes(buf, 0, 0, 10);

      int width = (buf[6] & 0xFF) | ((buf[7] & 0xFF) << 8);
      int height = (buf[8] & 0xFF) | ((buf[9] & 0xFF) << 8);

      if (Debug.on) {
         Debug.println(Debug.ADAPTOR, Debug.INFO, 
                       this + "Cached: " + currentURL + 
                       " size " + width + "x" + height);
      }

      sizeCache.put(currentURL, new Dimensions(width, height));
   }

   public String getAbsoluteURL(String targetURL) {
      if ((targetURL.length() >= 7) && 
          targetURL.substring(0, 7).equalsIgnoreCase("http://")) {
         // full URL
         return targetURL;
      }

      String currentURL = (String) iac.lookup(URLIdentifier.COMM_TAG);

      if (targetURL.charAt(0) == '/') {
         // absolute without host name
         String hostname = currentURL.substring(0, currentURL.indexOf('/'));
         return "http://" + hostname + targetURL;
      } else {
         // relative
         String directory = 
                   currentURL.substring(0, currentURL.lastIndexOf('/') + 1);
         return "http://" + directory + targetURL;
      }
   }

   // returns change in size
   protected int addImageSizes(int contentLength) throws AsyncEventException {
      int sizeChange = 0;

      DataAccessPointer p = new DataAccessPointer();
      currentPos.insertPointer(p);
      TagParser tag = new TagParser();

      try {
         while (true) {
            try {
               contentLength -= advanceToTag(p, contentLength);
            } catch (ClosureException e) {
               return sizeChange;
            }

            if (contentLength <= 0) {
               return sizeChange;
            }

            tag.setTagStart(p);

            if (tag.getName().equalsIgnoreCase("img")) {
               String src = getAbsoluteURL(tag.getAttrValue("src"));
               if (src.substring(0,7).equalsIgnoreCase("http://")) {
                  Dimensions dim = 
                         (Dimensions) sizeCache.get(src.substring(7));
                  if (dim != null) {
                      if (Debug.on) {
                         Debug.println(Debug.ADAPTOR, Debug.INFO, this + 
                                       ": Found cached size for " + src);
                      }

                      if (tag.getAttrValue("width") == null) {
                         if (Debug.on) {
                            Debug.println(Debug.ADAPTOR, Debug.INFO, this + 
                                          ": Adding width attribute of  " + 
                                          dim.width);
                         }

                         int mod = tag.addAttr("width", "" + dim.width);
                         contentLength += mod;
                         sizeChange += mod;
                      }
                      if (tag.getAttrValue("height") == null) {
                         if (Debug.on) {
                            Debug.println(Debug.ADAPTOR, Debug.INFO, this + 
                                          ": Adding height attribute of  " + 
                                          dim.height);
                         }

                         int mod = tag.addAttr("height", "" + dim.height);
                         contentLength += mod;
                         sizeChange += mod;
                      }
                  }
               }
            }

            contentLength -= p.movePastByte((byte) '>');
         }
      } finally {
         p.removeSelf();
      }
   }
}
