/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.ConductorExt.adaptors.http;

import FMG.Conductor.dataflow.*;
import FMG.Conductor.util.Debug;

/**
 * This adaptor performs a dual adaptation on HTTP response streams toward 
 * a single purpose.  First, it takes any HTML streams an looks for "img"
 * tags.  For each "img" tag, it adds a "lowsrc" attributed with a value
 * that is a mangled version of the "src" attribute.  This mangled URL,
 * when fetched by the browser is identified by the LowsrcDetector adaptor.
 * As it's second function, when a JPEG response is received, and the 
 * LowsrcDetector has detected a mangled URL, the JPEG will be reduced in 
 * quality (and therefore size).
 * <p>
 * This adaptor is typically run on the same node oas a LowsrcDetector.
 * <p>
 * @author Mark Yarvis (yarvis@fmg.cs.ucla.edu)
 * @version $Revision: 1.9 $      
 */

public class LowsrcEnabler extends HtmlAdaptor {

   static final String LOW_SRC_SUFFIX = "__LoWsRc";
   static final String COMM_TAG = LowsrcDetector.COMM_TAG;
   static final String TRUE = LowsrcDetector.TRUE;

   protected IAC iac;

   public LowsrcEnabler(AdaptorWindow window, String parameter) {
      super(window, parameter);

      iac = getStreamIAC();
   }

   protected void handleContent(int contentLength, String contentType,
                                                   String contentEncoding)  
                                              throws AsyncEventException {
      try {
         boolean reductionRequested = TRUE.equals(iac.lookup(COMM_TAG));

         if (contentType.startsWith("text/html") &&
             contentEncoding == null) {
            if (Debug.on) {
               Debug.println(Debug.ADAPTOR, Debug.INFO, 
                             this + "Modifying HTML content");
            }

            String contentLengthStr = header.getHeaderValue("Content-Length");

            if (contentLengthStr == null) {
               writeHeaders();
               addLowsrcAttrs(contentLength);
            } else {
               contentLength += addLowsrcAttrs(contentLength);
               header.replaceHeaderValue("Content-Length", "" + contentLength);
               writeHeaders();
            }

            forwardContent(contentLength);
         } else if (reductionRequested && 
                    contentType.startsWith("image/jpeg")) {
            if (reductionRequested) {
               if (Debug.on) {
                  Debug.println(Debug.ADAPTOR, Debug.INFO, 
                                this + "Reducing Jpeg");
               }

               contentLength = ReduceJpeg.reduceJpegQuality(window, currentPos, 
                                                            contentLength, 
                                                            (float) 0.1);

               header.replaceHeaderValue("Content-Length", "" + contentLength);
            } else {
               if (Debug.on) {
                  Debug.println(Debug.ADAPTOR, Debug.INFO, 
                                this + "Forwarding unreduced Jpeg");
               }
            }

            writeHeaders();
            forwardContent(contentLength);
         } else {
            if (Debug.on) {
               Debug.println(Debug.ADAPTOR, Debug.INFO, 
                             this + "Forwarding unchanged content");
            }

            writeHeaders();
            forwardContent(contentLength);
         }
      } finally {
         if (Debug.on) {
            Debug.println(Debug.ADAPTOR, Debug.INFO, 
                          this + "Done with content.");
         }
      }
   }

   protected int addLowsrcAttrs(int contentLength) throws AsyncEventException {
      int sizeChange = 0;

      DataAccessPointer p = new DataAccessPointer();
      currentPos.insertPointer(p);
      TagParser tag = new TagParser();

      try {
         while (true) {
            try {
               contentLength -= advanceToTag(p, contentLength);
            } catch (ClosureException e) {
               return sizeChange;
            }

            if (contentLength == 0) {
               return sizeChange;
            }

            tag.setTagStart(p);
            if (tag.getName().equalsIgnoreCase("img")) {
               String src = tag.getAttrValue("src");
               String lowsrc = tag.getAttrValue("lowsrc");

               if ((src != null) && (lowsrc == null)) {
                  lowsrc = src + LOW_SRC_SUFFIX;

                  if (Debug.on) {
                     Debug.println(Debug.ADAPTOR, Debug.EXTRAINFO, this + 
                                   " adding lowsrc attribute to image tag: " + 
                                   lowsrc);
                  }

                  int mod = tag.addAttr("lowsrc", lowsrc);
                  contentLength += mod;
                  sizeChange += mod;
               }
            }

            contentLength -= p.movePastByte((byte) '>');
         }
      } finally {
         p.removeSelf();
      }
   }
}
