/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.ConductorExt.adaptors.http;

import java.util.Enumeration;
import java.io.*;

import FMG.Conductor.dataflow.*;
import FMG.Conductor.util.*;

/**
 * This adpator looks for HTML content in HTTP streams and adds a string
 * to the page title.  Presumably this would help the user know that 
 * Conductor is adapting the stream.
 * <p>
 * The default string is "(C)", but a custom string can be passed as a 
 * parameter.
 * @author Mark Yarvis (yarvis@fmg.cs.ucla.edu)
 * @version $Revision: 1.9 $      
 */
public class MarkTitle extends HtmlAdaptor {

   private static final String DEFAULT_MARK = "(C)";

   protected byte[] markBytes;

   public MarkTitle(AdaptorWindow window, String parameter) {
      super(window, parameter);

      String mark = parameter;
      if (mark == null) {
         mark = DEFAULT_MARK;
      }

      markBytes = new byte[mark.length() + 1];
      markBytes[0] = (byte) ' ';
      mark.getBytes(0, mark.length(), markBytes, 1);
   }

   protected void handleContent(int contentLength, String contentType,
                                                   String contentEncoding)  
                                              throws AsyncEventException {
      if (contentType.startsWith("text/html") &&
          contentEncoding == null) {
System.out.println(this + ": Handling HTML");
         String contentLengthStr = header.getHeaderValue("Content-Length");

         if (contentLengthStr == null) {
            writeHeaders();
            markHtmlTitle(Integer.MAX_VALUE);
         } else {
            contentLength += markHtmlTitle(contentLength);
            header.replaceHeaderValue("Content-Length", "" + contentLength);
            writeHeaders();
         }

      } else {
System.out.println(this + " Ignoring content type: " + contentType + " encoding: " + contentEncoding);
         writeHeaders();
      }

      // forward any and all content
      forwardContent(contentLength);

System.out.println(this + ": Done with content");
   }

   int markHtmlTitle(int contentLength) throws AsyncEventException {
      DataAccessPointer startP = new DataAccessPointer();
      DataAccessPointer endP = null;

      currentPos.insertPointer(startP);

      TagParser tag = new TagParser();
      String tagName;

      try {
         while (true) {
            contentLength -= advanceToTag(startP, contentLength);
            if (contentLength == 0) {    // no more tags, stop looking
               return 0;
            }

            tag.setTagStart(startP);
            tagName = tag.getName();
//System.out.println("Got tag: " + tagName);
 
            if (tagName.equalsIgnoreCase("title")) { // found!
               break;
            }

            if (tagName.equalsIgnoreCase("body")) {  // stop looking
               return 0;
            }
            contentLength -= startP.movePastByte((byte) '>');
         }

         // at this point, startP is before <title>

         endP = new DataAccessPointer();
         startP.insertPointer(endP);

         while (true) {
            // find next tag
            contentLength -= endP.movePastByte((byte) '>');
            contentLength -= advanceToTag(endP, contentLength);

            if (contentLength == 0) {  // we didn't find </title>
               return 0;
            }
         
            tag.setTagStart(endP);
            tagName = tag.getName();

            if (tagName.equalsIgnoreCase("/title")) {
               break;                                  // found!
            }

            if (tagName.equalsIgnoreCase("body")) {
               return 0;                               // give up!
            }
         }

         // at this point endP follows the title text.

         int titleLen = startP.distanceTo(endP, Direction.up);
         byte[] buf = new byte[titleLen + markBytes.length];

         startP.getBytes(buf, 0, 0, titleLen);

         // add "(C)" to the end of the title
         System.arraycopy(markBytes, 0, buf, titleLen, markBytes.length);

         startP.replaceBytes(buf, 0, titleLen + markBytes.length, 0, titleLen);

         // return the change in data size;
         return markBytes.length;
      } finally {
         // clean up
         startP.removeSelf();
         if (endP != null) {
            endP.removeSelf();
         }
      }
   }
}
