/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.ConductorExt.adaptors.http;

import java.util.Enumeration;
import java.io.*;

import com.sun.image.codec.jpeg.*;
import java.awt.image.BufferedImage;

import FMG.Conductor.dataflow.*;
import FMG.Conductor.util.*;

import FMG.ConductorExt.adaptors.tools.*;

/**
 * This adaptor reduces the quality of JPEG images in an HTTP stream.
 * Standard tools from Sun are used to decode and then reencode the
 * images (with a lower quality level).
 * <p>
 * The quality level can be dynamically specified by the planner and passed
 * to the adaptor as a parameter.  The parameter should be a single float
 * value greater than 0 and less than or equal to 1.  If no parameter is 
 * specified, a default quality level will be used.
 * <p>
 * @author Mark Yarvis (yarvis@fmg.cs.ucla.edu)
 * @version $Revision: 1.10 $      
 */
public class ReduceJpeg extends HttpResponseAdaptor {

   protected float qualityLevel = (float) 0.1;

   public ReduceJpeg(AdaptorWindow window, String parameter) {
      super(window, parameter);

      if (parameter != null) {
         try {
            qualityLevel = Float.parseFloat(parameter);
         } catch (NumberFormatException e) {
            // just use the default value
            if (Debug.on) {
               Debug.println(Debug.ADAPTOR, Debug.WARNING, 
                             "WARNING: ReduceJpeg failed to parse " + 
                             "float parameter '" + parameter + "'");
            }
         }
      }
   }

   protected void handleContent(int contentLength, String contentType,
                                                   String contentEncoding)  
                                              throws AsyncEventException {
      if (contentType.startsWith("image/jpeg")) {
System.out.println(this + "Handling JPEG");

         handleJpeg(contentLength);

      } else {  // I don't know what it is, just forward it
System.out.println(this + "Handling unknown data");
         writeHeaders();
         forwardContent(contentLength);
      }

System.out.println(this + ": Done with content");
   }

   protected void handleJpeg(int contentLength) throws AsyncEventException {
      contentLength = reduceJpegQuality(window, currentPos, contentLength, 
                                                               qualityLevel);

      header.replaceHeaderValue("Content-Length", "" + contentLength);

      writeHeaders();
      forwardContent(contentLength);
   }

   static int reduceJpegQuality(AdaptorWindow window, 
                                DataAccessPointer currentPos, 
                                int contentLength, float qualityLevel) {
      WindowInputStream is = new WindowInputStream(window, 
                                           currentPos, contentLength);
      WindowOutputStream os = new WindowOutputStream(is);
      os.setSupressFlush(true);
      os.setAutoFlush(false);

      try {
System.out.println("Decoding image");
         JPEGImageDecoder decoder = JPEGCodec.createJPEGDecoder(is);
         BufferedImage image = decoder.decodeAsBufferedImage();

         is.close();

System.out.println("Encoding image (at low quality)");
         JPEGEncodeParam param = JPEGCodec.getDefaultJPEGEncodeParam(image);
         param.setQuality(qualityLevel, false);
         JPEGImageEncoder encoder = JPEGCodec.createJPEGEncoder(os, param);
         encoder.encode(image);

         os.close();

         return os.bytesWritten();
      } catch (IOException e) {
         throw new RuntimeException("Case not handled.");
      } catch (ImageFormatException e) {
         e.printStackTrace();
         throw new RuntimeException("Case not handled.");
      }
   }
}
