/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.ConductorExt.adaptors.imagedb;

import java.util.zip.*;

import FMG.Conductor.util.*;
import FMG.Conductor.comm.*;
import FMG.Conductor.dataflow.*;

/*
 * NOTE: this adaptor probably still violates the rules of segmentation.
 *       we need more consistent use of BlockModPointer
 */

public class ImageCompress extends ImageDBAdaptor {

   static int instances = 0;
   static char echochars[] = { '.', ',', ';', ':', '|'};

   char e;
   Deflater deflater = new Deflater(Deflater.BEST_COMPRESSION, true);
   int buflen = 1000000;
   byte[] inbuf = new byte[buflen];
   byte[] outbuf = new byte[buflen];
   boolean powerSavingProtocol = false;

   public ImageCompress(AdaptorWindow window, String parameter) {
      super(window, parameter);
      if (Debug.on) {
         if (Debug.getLevel(Debug.ADAPTOR) > Debug.INFO) {
            Debug.println("New ImageCompress instantiated");
         }
         if (Debug.getLevel(Debug.ADAPTOR) > Debug.DETAIL) {
            e = echochars[(instances++) % echochars.length];
         }
      }

      if ((parameter != null) && parameter.equals("power")) {
         powerSavingProtocol = true;
      }
   }

   public void compressBlock(BlockModPointer pointer, int len)
                         throws InterruptedException, AsyncEventException {

      if (window.countRemaining(pointer) == 0) {
//System.out.println("Expanding window");
         window.expand();
//System.out.println("Expanded");
      }

      while (len > 0) {
         int inlen = min(len, buflen, window.countRemaining(pointer));
//System.out.println("Handling " + inlen + " bytes");
         pointer.getBytes(inbuf, 0, 0, inlen);

         deflater.setInput(inbuf, 0, inlen);
         int outlen = deflater.deflate(outbuf, 0, buflen);
         if (outlen == 0) {
            pointer.deleteBytes(inlen);
         } else {
            pointer.replaceBytes(outbuf, 0, outlen, 0, inlen);
            pointer.advance(outlen);
            window.contract(pointer);
         }

         while (! deflater.needsInput()) {
            outlen = deflater.deflate(outbuf, 0, buflen);
            pointer.insertBytes(outbuf, 0, outlen);
            window.contract(pointer);
         }
         len -= inlen;

//System.out.println("Still need " + len + " bytes");
         if (len > 0) {
//System.out.println("Expanding window");
            window.expand();
//System.out.println("Expanded");
         }
      }
//System.out.println("Done with input bytes ... finishing up");
      deflater.finish();
      while (! deflater.finished()) {
         int outlen = deflater.deflate(outbuf, 0, buflen);
         pointer.insertBytes(outbuf, 0, outlen);
         window.contract(pointer);
      }

      if (Debug.on) {
         Debug.println(Debug.ADAPTOR, Debug.INFO, 
                       "Compressed " + deflater.getTotalIn() + 
                       " bytes to " + deflater.getTotalOut() + " bytes");
      }

      deflater.reset();
   }

   public void handleFlow() {

      if (Debug.on) {
         Debug.print(Debug.ADAPTOR, Debug.INFO, "(");
      }

      BlockModPointer pointer = new BlockModPointer();
      window.insertPointer(pointer);

      try {
         while (true) {
            try {
               getNextLine(pointer);               /* get query/result id */
//System.out.println("Handling a query or result record");
               pointer.movePastByte((byte) '\n');
               window.contract(pointer);           /* push it out */

               getNextLine(pointer);               /* get # of samples */
               int numSamples = readInt(pointer);
               pointer.movePastByte((byte) '\n');
               window.contract(pointer);           /* push it out */

//System.out.println("There are " + numSamples + " images in this record");

               for (int i=0; i<numSamples; i++) {
                  getNextLine(pointer);             /* get image dimensions */
                  int size = 3;
                  size *= readInt(pointer);
                  pointer.movePastByte((byte) ' ');
                  size *= readInt(pointer);
                  pointer.movePastByte((byte) '\n');
                  window.contract(pointer);         /* push them out */

//System.out.println("Image " + i + " is " + size + " bytes");

                  pointer.blockModStart();
                  compressBlock(pointer, size);
                  pointer.blockModEnd();
                  window.contract(pointer);
               }

               if (powerSavingProtocol) {
                  getNextLine(pointer);               /* get adjustment */
                  pointer.movePastByte((byte) '\n');
                  window.contract(pointer);           /* push it out */
               }
//System.out.println("Finished handling a query or result record");


            } catch (InputClosedException e) {
               pointer.removeSelf();
               window.contract();
               if (Debug.on) {
                  Debug.print(Debug.ADAPTOR, Debug.INFO, ")");
               }
               window.closeOutput();
               return;
            } catch (OutputClosedException e) {
               pointer.removeSelf();
               window.closeInput();
               return;
            } catch (AsyncEventException e) {
               throw new RuntimeException("Not implemented");
            }
         }
      } catch (InterruptedException ignore) {
         throw new RuntimeException("Case not handled");
      }
   }
}
