/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.ConductorExt.adaptors.imagedb;

import java.util.zip.*;

import FMG.Conductor.util.*;
import FMG.Conductor.comm.*;
import FMG.Conductor.dataflow.*;

/*
 * NOTE: this adaptor probably still violates the rules of segmentation.
 *       we need more consistent use a BlockModPointer.
 */

public class ImageDecompress extends ImageDBAdaptor {

   static int instances = 0;
   static char echochars[] = { '.', ',', ';', ':', '|'};

   char e;
   Inflater inflater = new Inflater(true);
   int buflen = 1000000;
   byte[] inbuf = new byte[buflen];
   byte[] outbuf = new byte[buflen];
   boolean powerSavingProtocol = false;

   public ImageDecompress(AdaptorWindow window, String parameter) {
      super(window, parameter);
      if (Debug.on) {
         Debug.println(Debug.ADAPTOR, Debug.INFO,
                       "New ImageDecompress instantiated");

         if (Debug.getLevel(Debug.ADAPTOR) > Debug.DETAIL) {
            e = echochars[(instances++) % echochars.length];
         }
      }

      if ((parameter != null) && parameter.equals("power")) {
         powerSavingProtocol = true;
      }
   }

   public void uncompressBlock(BlockModPointer pointer, int len)
                         throws InterruptedException, AsyncEventException {

      if (window.countRemaining(pointer) == 0) {
//System.out.println("Getting more bytes");
         window.expand();
//System.out.println("Got em");
      }

      try {
         while (len > 0) {
//System.out.println("Still looking for " + len + " bytes");
            int inlen = min(buflen, window.countRemaining(pointer));
            pointer.getBytes(inbuf, 0, 0, inlen);
            inflater.setInput(inbuf, 0, inlen);
   
            int outlen = inflater.inflate(outbuf, 0, buflen);
            pointer.insertBytes(outbuf, 0, outlen);
            window.contract(pointer);
            len -= outlen;

           while (outlen > 0) {
//System.out.println("Processing remaining bytes from buffer");
               outlen = inflater.inflate(outbuf, 0, buflen);
               pointer.insertBytes(outbuf, 0, outlen);
               window.contract(pointer);
               len -= outlen;
            }

//System.out.println("Deleting used up bytes");
            if (len == 0) {
               int bytesUsed = inlen - inflater.getRemaining();
               pointer.deleteBytes(bytesUsed);
            } else if (len > 0) {
               pointer.deleteBytes(inlen);
               window.expand();
            } else {
               throw new RuntimeException("This shouldn't happen");
            }
         }
      } catch (DataFormatException e) {
         throw new RuntimeException("This shouldn't happen");
      }

      if (Debug.on) {
         Debug.println(Debug.ADAPTOR, Debug.INFO, 
                       "Decompressed " + inflater.getTotalIn() + 
                       " bytes to " + inflater.getTotalOut() + " bytes");
      }

      inflater.reset();
   }

   public void handleFlow() {

      if (Debug.on) {
         Debug.print(Debug.ADAPTOR, Debug.INFO, "(");
      }

      BlockModPointer pointer = new BlockModPointer();
      window.insertPointer(pointer);

      try {
         while (true) {
            try {
               getNextLine(pointer);               /* get query/result id */
//System.out.println("Handling a query/result record");
               pointer.movePastByte((byte) '\n');
               window.contract(pointer);           /* push it out */

               getNextLine(pointer);               /* get # of samples */
               int numSamples = readInt(pointer);
               pointer.movePastByte((byte) '\n');
               window.contract(pointer);           /* push it out */

//System.out.println("There are " + numSamples + " images in this record");
               for (int i=0; i<numSamples; i++) {
                  getNextLine(pointer);             /* get image dimensions */
                  int size = 3;
                  size *= readInt(pointer);
                  pointer.movePastByte((byte) ' ');
                  size *= readInt(pointer);
                  pointer.movePastByte((byte) '\n');
                  window.contract(pointer);        /* push them out */
//System.out.println("Image " + i + " has " + size + " bytes");

                  pointer.blockModStart();
                  uncompressBlock(pointer, size);
                  pointer.blockModEnd();
                  window.contract(pointer);
               }

               if (powerSavingProtocol) {
                  getNextLine(pointer);               /* get adjustment */
                  pointer.movePastByte((byte) '\n');
                  window.contract(pointer);           /* push it out */
               }
//System.out.println("Finished handling a query/result record");


            } catch (InputClosedException e) {
               pointer.removeSelf();
               window.contract();
               if (Debug.on) {
                  Debug.print(Debug.ADAPTOR, Debug.INFO, ")");
               }
               window.closeOutput();
               return;
            } catch (OutputClosedException e) {
               pointer.removeSelf();
               window.closeInput();
               return;
            } catch (AsyncEventException e) {
               throw new RuntimeException("Not implemented");
            }
         }
      } catch (InterruptedException ignore) {
         throw new RuntimeException("Case not handled");
      }
   }
}
