/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.ConductorExt.adaptors.pop3;

import java.io.*;
import java.net.*;

import FMG.Conductor.dataflow.*;
import FMG.Conductor.util.*;

import FMG.ConductorExt.adaptors.tools.*;

/*
 * This adaptor looks for large attachments in Mime messages transmitted
 * by the POP protocol.  When it finds one, it saves the attachment (decoding 
 * it if necessary) in a working directory for that user and replaces the
 * attachment with text describing where the file has been stored.
 * <p>
 * This adaptor assumes that some sort of Pop3RequestAdaptor is present
 * on the same node to decipher the request, so we know what type of response 
 * to expect.
 * <p>
 * Two parameters can be specified by the planner to control the behavior 
 * of this adpator.  The first parameter is an integer that specifies the 
 * size that an attachment must be before it is saved to a file.  The 
 * second parameter is an integer that specifies the number of bytes of 
 * a large text message that will be left in the transmitted message as 
 * context.  The two parameters are separated by a space.
 * <p>
 * @author Mark Yarvis (yarvis@fmg.cs.ucla.edu)
 * @version $Revision: 1.24 $      
 */
public class MimeAttachmentSaver extends Pop3MimeFilter {

   /** 
    * The threshold size beyond which an attachment will be removed and 
    * stored rather than transmitted. 
    */
   protected int saveSizeThreshold = 10240;

   /** 
    * The number of bytes to keep in a text attachment after it has been 
    * saved.  This gives the reader some context in which to judge whether
    * or not to check out the attachment.
    */
   protected int numTextIntroBytes = 1024;

   protected static final String STORAGE_MODULE_NAME = "MailAttachments";

   protected String servername;

   protected StorageAccess storageAccess;

   protected String username;

   protected int commandsSinceAuth = -1;

   public MimeAttachmentSaver(AdaptorWindow window, String parameter) {
      super(window, parameter);

      // if any parameters are given, parse them
      if (parameter != null) {
         int spacePos = parameter.indexOf(' ');

         try {
            if (spacePos == -1) {
               // one int was specified ... assume it's the size threshold
               saveSizeThreshold = Integer.parseInt(parameter);
            } else {
               // two ints were given; first is threshold, second is intro size
               saveSizeThreshold = Integer.parseInt(
                                            parameter.substring(0, spacePos));
               numTextIntroBytes = Integer.parseInt(
                                      parameter.substring(spacePos + 1,
                                                          parameter.length()));
            }
         } catch (NumberFormatException e) {
            // just use default value
            if (Debug.on) {
               Debug.println(Debug.ADAPTOR, Debug.WARNING,
                             "WARNING: MimeAttachmentSaver failed to parse " + 
                             "parameter '" + parameter + "'");
            }
         }
      }
   }

   protected void handleMessagePart(Header header, DataAccessPointer endP)
                                               throws AsyncEventException {
      String contentType = header.getHeaderValue("Content-Type");

      // not interested in multipart content (take it appart first)
      if ((contentType != null) && (contentType.length() >= 9) &&
          contentType.substring(0, 9).equalsIgnoreCase("multipart")) {
         super.handleMessagePart(header, endP);
         return;
      }

      // not interested in message content (parse it)
      if ((contentType != null) && (contentType.length() >= 7) &&
          contentType.substring(0, 7).equalsIgnoreCase("message")) {
         super.handleMessagePart(header, endP);
         return;
      }

      // not interested in small attachments
      int length = header.getPostPointer().distanceTo(endP);
      if (length < saveSizeThreshold) {
         if (Debug.on) {
            Debug.println(Debug.ADAPTOR, Debug.INFO,
                          this + ": Skipping attachment due to small size: " + 
                          length);
         }
         super.handleMessagePart(header, endP);
         return;
      }

      // determine content handling
      WindowInputStream wis;
      InputStream is;
      byte[] bytesToSave = null;

      String contentEncoding = 
                header.getHeaderValue("Content-Transfer-Encoding");
      if ((contentEncoding == null) || 
          (contentEncoding.equalsIgnoreCase("8bit")) ||
          (contentEncoding.equalsIgnoreCase("7bit")) ||
          (contentEncoding.equalsIgnoreCase("binary"))) {
         // save as is
         wis = new WindowInputStream(window, header.getPostPointer(), length);
         is = wis;
         if (contentEncoding == null) {
            // For known text messages, keep the first part of the message
            // to give the user an idea of what was saved.
            bytesToSave = new byte[numTextIntroBytes];
         }
      } else if (contentEncoding.equalsIgnoreCase("base64")) {
         // decode and save
         wis = new WindowInputStream(window, header.getPostPointer(), length);
         is = new Base64InputStream(wis);
      } else {
         if (Debug.on) {
            Debug.println(Debug.ADAPTOR, Debug.INFO, 
                          this + ": Skipping attachment due to unknown " + 
                          "encoding: " + contentEncoding);
         }
         super.handleMessagePart(header, endP);
         return;
      }

      if (storageAccess == null) {
         storageAccess = new StorageAccess(STORAGE_MODULE_NAME);
      }

      // determine name of save file
      StorageAccess.WriteAccess saveAccess = getSaveAccess(header);

      // save the attachment
      try {
         if (Debug.on) {
            Debug.println(Debug.ADAPTOR, Debug.INFO, 
                          this + ": Saving attachment of size " + length + 
                          " and encoding " + contentEncoding + 
                          " to file " + saveAccess.filename);
         }

         saveAttachment(is, saveAccess, bytesToSave);
      } catch (IOException e) {
// TODO: we should put the data back if we fail writing the file
         throw new RuntimeException("Case not handled");
      }

      WindowOutputStream wos = new WindowOutputStream(wis);
      wos.setAutoFlush(false);
      wos.setSupressFlush(true);
      // replace attachment with one of type message/external?
      try {
         replaceAttachment(header, wos, saveAccess.filename, bytesToSave);
         is.close();
         wos.close();
      } catch (IOException e) {
         throw new RuntimeException("This shouldn't happen");
      }
   }

   protected StorageAccess.WriteAccess getSaveAccess(Header header) {
      // get filename from contentDisposition's "filename="
      String filename = null;
      String contentDisposition = header.getHeaderValue("Content-Disposition");

      int filenameStart = -1;

      if (contentDisposition != null) {
         String lowerCD = contentDisposition.toLowerCase();
         filenameStart = lowerCD.indexOf("filename=");
      }

      if (filenameStart != -1) {
         // skip the filename= part
         filenameStart += 9;

         // skip the equals sign
         if (contentDisposition.charAt(filenameStart) == '"') {
            filenameStart++;
         }

         // find end
         int filenameEnd = contentDisposition.length()-1;
         char finalChar = contentDisposition.charAt(filenameEnd);
         while ((finalChar == ' ') || (finalChar == '\t') ||
                (finalChar == '"') || (finalChar == '\r') || 
                (finalChar == '\n')) {
            finalChar = contentDisposition.charAt(--filenameEnd);
         }

         filename = contentDisposition.substring(filenameStart, 
                                                 filenameEnd + 1);
      } else {    // no filename is given, use the subject
         String subject = currentMessageHeader.getHeaderValue("Subject");
         if (subject != null) {
            filename = makeReasonableFilename(subject);
         }
      }

      // if no useful name could be found, use a generic one
      if (filename == null) {  
         filename = "attachment." + System.currentTimeMillis();
      }

      // make sure it's unique, or uniquify it
      StorageAccess.WriteAccess access = null;
      if (! storageAccess.workingFileExists(filename)) {
         access = storageAccess.writeWorkingFile(filename);
      }
      for (int i = 0; access == null; i++) {
         String unique = filename + "." + i;
         if (! storageAccess.workingFileExists(unique)) {
            access = storageAccess.writeWorkingFile(unique);
         }
      }

      return access;
   }

   private boolean isGoodFilenameChar(char c) {
      return (((c >= 'a') && (c <= 'z')) ||
              ((c >= 'A') && (c <= 'Z')) ||
              ((c >= '0') && (c <= '9')) ||
              (c == '+') || (c == '-') || (c == '@') || 
              (c == '_') || (c == '=') || (c == ':') || 
              (c == '.') || (c == ','));
   }

   protected String makeReasonableFilename(String name) {
      int origLen = name.length();
      StringBuffer sb = new StringBuffer(origLen);
      boolean lastWasReplaced = false;

      for (int i=0; i<origLen; i++) {
         char c = name.charAt(i);
         if (isGoodFilenameChar(c)) {
            sb.append(c);
            lastWasReplaced = false;
         } else {
            if (!lastWasReplaced) {
               sb.append('_');
               lastWasReplaced = true;
            }
         }
      }

      if (sb.length() > 50) {
         sb.setLength(50);
      }

      return sb.toString();
   }

   protected void saveAttachment(InputStream is, 
                                 StorageAccess.WriteAccess saveAccess,
                                 byte[] bytesToSave) throws IOException {
      OutputStream os = saveAccess.os;
      byte[] buf = new byte[10240];
      int bytesSaved = 0;

      while (true) {
         int len = is.read(buf);
         if (len < 0) {
            os.close();
            return;
         }
         os.write(buf, 0, len);

         // if requested, save the top part of the message
         if ((bytesToSave != null) && (bytesSaved < bytesToSave.length)) {
            if (len > (bytesToSave.length - bytesSaved)) {
               len = bytesToSave.length - bytesSaved;
            }
            System.arraycopy(buf, 0, bytesToSave, bytesSaved, len);
            bytesSaved += len;
         }
      }
   }

   protected void replaceAttachment(Header header, WindowOutputStream wos, 
                                    String filename, byte[] topOfMessage) {
      // get the local host name
      if (servername == null) {
         try {
            servername = InetAddress.getLocalHost().getHostName();
         } catch (UnknownHostException e) {
            servername = "unknown";
         }
      }

      // remove all headers
      header.deleteHeader("content-type");
      header.deleteHeader("content-transfer-encoding");
      header.deleteHeader("content-id");
      header.deleteHeader("content-description");
      header.deleteHeader("content-disposition");

      // write out new contents
      if (topOfMessage != null) {
         try {
            wos.write(topOfMessage);
            wos.write(' ');
            wos.write('.'); wos.write('.'); wos.write('.');
            wos.write('\n'); wos.write('\n');
         } catch (IOException e) {
            throw new RuntimeException("This shouldn't happen");
         }
      }

      PrintWriter pw = new PrintWriter(wos);

      pw.println("This attachment has been removed by Conductor and " + 
                 "stored to a server.");
      pw.println("You can retrieve this file from \n  " + 
                 servername + ":" + filename);
      pw.println();
      pw.flush();
   }

   protected boolean handleResponse(ReadOnlyRequest request,
                                    boolean isInitialHandshake)
                                                 throws AsyncEventException {
      if (commandsSinceAuth != -1) {
         commandsSinceAuth++;
      }

      if (! isInitialHandshake && response.successfulResult()) {
         if (request.getCommand().equalsIgnoreCase("USER")) {
            setUsername(request.getArg(0));
         } else if (request.getCommand().equalsIgnoreCase("PASS")) {
            setPassword(request.getArg(0));
         } else if (request.getCommand().equalsIgnoreCase("AUTH") && 
                   (request.getNumArgs() != 0)) {
            if (request.getArg(0).equalsIgnoreCase("LOGIN")) {
               commandsSinceAuth = 0;
            } else {
               if (Debug.on) {
                  Debug.println(Debug.ADAPTOR, Debug.WARNING,
                                this + "Warning: unknown auth type: " + 
                                request.getArg(0));
               }
            }
         } else if (commandsSinceAuth == 1) {
            setUsername(base64Decode(request.getCommand()));
         } else if (commandsSinceAuth == 2) {
            setPassword(base64Decode(request.getCommand()));
         }
      }

      return super.handleResponse(request, isInitialHandshake);
   }

   protected String base64Decode(String input) {
// kludge ... should really be able to access the raw bas64 decoder engine!
      InputStream in = new Base64InputStream(
                              new ByteArrayInputStream(input.getBytes()));
      byte buf[] = new byte[100];
      StringBuffer out = new StringBuffer();
      
      try {
         while (true) {
            int len = in.read(buf);
            if (len == -1) {
               break;
            }
            out.append(new String(buf, 0, len));
         }
      } catch (IOException e) { 
         throw new RuntimeException ("This shouldn't happen");
      }

      return out.toString();
   }

   protected void setUsername(String username) {
      this.username = username;
   }

   protected void setPassword(String password) {
      storageAccess = new StorageAccess(username, password, 
                                        STORAGE_MODULE_NAME);

      if (Debug.on) {
         if (storageAccess.isAuthenticated()) {
            Debug.println(Debug.ADAPTOR, Debug.INFO, 
                          this + ": Successfully authenticated user " + 
                          username);
         }
      }
   }
}
