/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.ConductorExt.adaptors.pop3;

import FMG.Conductor.dataflow.*;
import FMG.Conductor.util.*;
import FMG.ConductorExt.adaptors.ProtocolAdaptor;

/*
 * This class is a useful base class for writing Pop3 adaptors.
 * It provides useful functions and inner classes for parsing the
 * stream and communicating between request and response adpators.
 * <p>
 * @author Mark Yarvis (yarvis@fmg.cs.ucla.edu)
 * @version $Revision: 1.9 $      
 */
public abstract class Pop3Adaptor extends ProtocolAdaptor {

   private final static String DEFAULT_COMM_TAG_NAME = "POP3.CurrentRequest";

   private IAC iac = getStreamIAC();

   protected interface ReadOnlyRequest {
      public String getCommand();
      public int getNumArgs();
      public String getArg(int argNum);
   }

   protected class LineWithArgs extends ElementalLine {
      public LineWithArgs(DataAccessPointer startPointer)
                                             throws AsyncEventException {
         super(startPointer);
      }

      // argument numbering starts at 0
      public String getArg(int argNum) {
         return getElement(argNum + 1);
      }     

      public void replaceArg(int argNum, String newArg) {
         replaceArg(argNum, newArg.getBytes());
      }

      public void replaceArg(int argNum, byte[] newArg) {
         replaceElement(argNum + 1, newArg);
      }

      public int getNumArgs() {
         return getNumElements() - 1;
      }

   }

   public Pop3Adaptor(AdaptorWindow window, String parameter) {
      super(window, parameter);
   }

   // change this function if you want more than one adaptor for the same 
   // pop3 stream (particularly if it's going to modify the command)
   protected String getCommTag() {
      return DEFAULT_COMM_TAG_NAME;
   }

   protected void shareRequest(ReadOnlyRequest request) {
      if (Debug.on) {
         Debug.println(Debug.ADAPTOR, Debug.EXTRAINFO, 
                       this + " Sharing request: " + request);
      }
      iac.store(getCommTag(), request);
   }

   protected ReadOnlyRequest getRequest() {
      if (Debug.on) {
         Debug.println(Debug.ADAPTOR, Debug.EXTRAINFO, 
                       this + " Getting request");
      }
      return (ReadOnlyRequest) iac.lookup(getCommTag());
   }

}
