/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.ConductorExt.adaptors.pop3;

import FMG.Conductor.dataflow.*;
import FMG.Conductor.util.*;

/**
 * This class provides a useful base class for building adaptors that
 * operate on Mime messages in POP3 streams.  For each message, 
 * handleMessagePart() is called.  If multiple parts are contained in
 * a given message part, handleMultiPart() is called, which calls 
 * handleMessagePar() for each sub-part.  Override these methods to
 * build an adaptor that contains useful functionality (either in terms
 * of content handling or special handling for specific types of multipart 
 * messages).
 * <p>
 * This adaptor assumes that some sort of Pop3RequestAdaptor is present
 * on the same node to decipher the request, so we know what type of response 
 * to expect.
 * <p>
 * @author Mark Yarvis (yarvis@fmg.cs.ucla.edu)
 * @version $Revision: 1.19 $      
 */
public class Pop3MimeFilter extends Pop3ResponseAdaptor {

   protected Header currentMessageHeader;

   protected class BoundaryFinder extends Line {

      boolean isEnd;

      private DataAccessPointer endMarker;

      public BoundaryFinder(DataAccessPointer currentPos, String target, 
                                                 DataAccessPointer endMarker) 
                                                 throws AsyncEventException {
         this.endMarker = endMarker;

         preP = new DataAccessPointer();
         currentPos.insertPointer(preP);

         lookForNext(target.getBytes());

         postP = new DataAccessPointer();
         preP.insertPointer(postP);

         len = findEndOfLine(postP);
      }

      // more preP until it's before the desired boundary
      // NOTE: this assumes all data is present
      protected void lookForNext(byte[] target) throws AsyncEventException {
         byte[] buf = new byte[target.length];
         while (true) {
            if ((preP.getByte(0) == '-') && (preP.getByte(1) == '-') &&
                (preP.distanceTo(endMarker) >= (target.length + 2)) ) {
               // possible boundary
               preP.getBytes(buf, 0, 2, target.length);
               if (Bytes.areEqual(buf, 0, target, 0, target.length)) {
                  // definate boundary
                  if ((preP.distanceTo(endMarker) >= target.length + 4) &&
                      (preP.getByte(target.length + 2) == '-') &&
                      (preP.getByte(target.length + 3) == '-')) {
                     // end boundary
                     isEnd = true;
                  }
                  return;
               }
            } else if ((preP.getByte(0) == '.')  &&   // end of message
                       ( (preP.getByte(1) == '\n') || 
                           ((preP.getByte(1) == '\r') && 
                            (preP.getByte(2) == '\n')) )
                      ) {
               isEnd = true;
               return;
            }
            preP.movePastByte((byte) '\n');
         }
      }

      public boolean isEnd() {
         return isEnd;
      }

      public void removeSelf() {
         preP.removeSelf();
         postP.removeSelf();
      }
   }

   public Pop3MimeFilter(AdaptorWindow window, String parameter) {
      super(window, parameter);
   }

   protected boolean handleResponse(ReadOnlyRequest request,
                                    boolean isInitialHandshake)
                                                 throws AsyncEventException {
      if (! isInitialHandshake && response.successfulResult()) {

         String command = request.getCommand();
         if (command.equalsIgnoreCase("TOP") ||
             command.equalsIgnoreCase("RETR")) {

            // delineate entire message
            MultiLine data = new MultiLine(currentPos);

            int origLen = data.length();

            // mark content end (not including ending .)
            DataAccessPointer contentEnd = new DataAccessPointer();
            data.getPrePointer().insertPointer(contentEnd);
            contentEnd.advance(origLen - 3);

            // recursively handle message
            currentMessageHeader = new Header(data.getPrePointer());
            handleMessagePart(currentMessageHeader, contentEnd);

            if (Debug.on) {
               int newLen = data.length();
               Debug.println(Debug.ADAPTOR, Debug.INFO, 
                             this + ": Message size changed from " + origLen + 
                             " to " + newLen + " (" + 
                             Debug.percentChange(origLen, newLen) + "%)");
            }

            // clean up
            contentEnd.removeSelf();
            currentMessageHeader.removeSelf();

            if (command.equalsIgnoreCase("RETR") && 
                response.getArg(1).equalsIgnoreCase("octets")) {
               // adjust the response line
               int messageLen = data.length() - 3;
               response.replaceArg(0, "" + messageLen);
            }

            // write out the response line
            writeResponse();

            // move the current pointer to after the message
            currentPos.removeSelf();
            data.getPostPointer().insertPointer(currentPos);

            // clean up extra pointers
            data.getPostPointer().removeSelf();
            currentMessageHeader.getPostPointer().removeSelf();

            // write out all data
            window.contract(currentPos);

            return true;
         }
      }

      // if we didn't handle it above
      return super.handleResponse(request, isInitialHandshake);
   }

   private String getBoundaryString(String contentType) {
      String lowerCT = contentType.toLowerCase();
      int boundaryStart = lowerCT.indexOf("boundary=");

      if (boundaryStart == -1) {
         throw new RuntimeException("multipart content type must specify " + 
                                    "boundary");
      }

      boundaryStart += 9;

      if (contentType.charAt(boundaryStart)== '"') {
         boundaryStart++;
      }

      int boundaryEnd = contentType.length() - 1;
      char finalChar = contentType.charAt(boundaryEnd);
      while ((finalChar == ' ') || (finalChar == '\t') || 
             (finalChar == '"') || (finalChar == '\r') || 
             (finalChar == '\n')) {
         finalChar = contentType.charAt(--boundaryEnd);
      }

      return contentType.substring(boundaryStart, boundaryEnd+1);
   }

   // override in sub-class to handle parts of other types
   protected void handleMessagePart(Header header, DataAccessPointer endP) 
                                               throws AsyncEventException {
      try {
         String contentType = header.getHeaderValue("Content-Type");

         if (contentType != null) {
            if ((contentType.length() >= 9) && 
                contentType.substring(0, 9).equalsIgnoreCase("multipart")) {
               handleMultiPart(getBoundaryString(contentType), header, endP);
            } else if ((contentType.length() >= 7) &&
                       contentType.substring(0, 7).
                                          equalsIgnoreCase("message")) {
               Header msgHeader = new Header(header.getPostPointer());
               handleMessagePart(msgHeader, endP);
               msgHeader.removeSelf();
            }
         }
      } catch (RuntimeException e) {
         if (Debug.on) {
            Debug.println(Debug.ADAPTOR, Debug.WARNING, 
                          this + " Failure handling message part: ");
            e.printStackTrace();
         }
      }
   }

   // override in sub-class for special handling of particular 
   // sub-types of multiplart (eg, multipart/alternative)
   protected void handleMultiPart(String boundaryStr, Header header, 
                                             DataAccessPointer endP) 
                                        throws AsyncEventException {
      if (Debug.on) {
         Debug.println(Debug.ADAPTOR, Debug.EXTRAINFO, this + 
                       ": Handling multi-part attachment with boundary " + 
                       boundaryStr);
      }
      BoundaryFinder nextBound = new BoundaryFinder(header.getPostPointer(), 
                                                    boundaryStr, endP);
      BoundaryFinder prevBound;

      while (! nextBound.isEnd()) {
         prevBound = nextBound;
         nextBound = new BoundaryFinder(prevBound.getPostPointer(), 
                                                        boundaryStr, endP);

         Header subHeader = new Header(prevBound.getPostPointer());
         handleMessagePart(subHeader, nextBound.getPrePointer());
         subHeader.removeSelf();
         prevBound.removeSelf();
      }
      nextBound.removeSelf();
   }
}
