/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.ConductorExt.adaptors.pop3;

import FMG.Conductor.dataflow.*;
import FMG.Conductor.util.*;

/**
 * This class has two purposes.  It provides a base class for building
 * useful adaptations on the request part of the POP3 protocol.  Also,
 * it is capable of parsing each request and providing a read-only copy
 * to subclasses of Pop3ResponseAdaptor.  
 * <p>
 * Nearly all subclasses of Pop3ResponseAdaptor requre some subclass of
 * Pop3RequestAdaptor to be present on the same node because it is
 * difficult to know, looking only a the response stream, what kind of
 * response to expect.
 * <p>
 * @author Mark Yarvis (yarvis@fmg.cs.ucla.edu)
 * @version $Revision: 1.8 $      
 */
public class Pop3RequestAdaptor extends Pop3Adaptor {

   protected class Request extends LineWithArgs implements ReadOnlyRequest {
      public Request(DataAccessPointer startPointer) 
                                             throws AsyncEventException {
         super(startPointer);
      }

      public String getCommand() {
         return getElement(0);
      }

      public void replaceCommand(String newCommand) {
         replaceCommand(newCommand.getBytes());
      }

      public void replaceCommand(byte[] newCommand) {
         replaceElement(0, newCommand);
      }

      public void replaceAll(String newRequest) {
         replaceAll(newRequest.getBytes());
      }

      public void replaceAll(byte[] newRequest) {
         replaceBytes(newRequest);
      }

   }

   public Pop3RequestAdaptor(AdaptorWindow window, String parameter) {
      super(window, parameter);
   }

   public boolean handleRequest(Request request) {
      request.cacheSelf();
      shareRequest(request);
      return true;
   }

   public void handleProtocol() {
      boolean moreAdaptation = true;
      while (moreAdaptation) {
         try {
            if (Debug.on) {
               Debug.println(Debug.ADAPTOR, Debug.EXTRAINFO, 
                             this + " Looking for request");
            }

            Request request = new Request(currentPos);

            if (Debug.on) {
               Debug.println(Debug.ADAPTOR, Debug.MOREINFO, 
                             this + " handling request: " + request);
            }

            moreAdaptation = handleRequest(request);

            DataAccessPointer postP = request.getPostPointer();
            currentPos.removeSelf();
            postP.insertPointer(currentPos);
            postP.removeSelf();
            window.contract(currentPos);
            if (Debug.on) {
               Debug.println(Debug.ADAPTOR, Debug.EXTRAINFO, 
                             this + " forwarded request");
            }
         } catch (OutputClosedException e) {
            Debug.println(Debug.ADAPTOR, Debug.INFO,
                          this + ": output closed unexpectedly");
            return;
         } catch (InputClosedException e) {
            if (Debug.on) {
               Debug.print(Debug.ADAPTOR, Debug.INFO, ")");
            }
            window.contract();
            window.closeOutput();
            Debug.println(Debug.ADAPTOR, Debug.INFO, this + ": closed");
            return;
         } catch (PlanChangedException e) {
            throw new RuntimeException("Case not handled.");
         } catch (AsyncEventException e) {
             Debug.println(Debug.ADAPTOR, Debug.WARNING,
                           this + ": Got an unexpected exception");
             return;
         }
      }
   }

}
