/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.ConductorExt.adaptors.pop3;

import FMG.Conductor.dataflow.*;
import FMG.Conductor.util.*;
/**
 * This class provides a convenient superclass from which to build adaptors
 * that adapt a POP3 response stream.  For each response, handleResponse()
 * is called.  The default implementation forwards the response.
 * <p>
 * This adaptor assumes that some sort of Pop3RequestAdaptor is present
 * on the same node to decipher the request, so we know what type of response 
 * to expect.
 * <p>
 * Nearly all subclasses of Pop3ResponseAdaptor requre some subclass of
 * Pop3RequestAdaptor to be present on the same node because it is
 * difficult to know, looking only a the response stream, what kind of
 * response to expect.
 * <p>
 * @author Mark Yarvis (yarvis@fmg.cs.ucla.edu)
 * @version $Revision: 1.14 $      
 */

public class Pop3ResponseAdaptor extends Pop3Adaptor {

   private static final byte[] OK_RESULT = "+OK".getBytes();
   private static final byte[] ERR_RESULT = "-ERR".getBytes();

   protected Response response;
   protected DataAccessPointer preResponsePointer = new DataAccessPointer();

   protected class Response extends LineWithArgs {
      Response(DataAccessPointer startPointer) 
                                             throws AsyncEventException {
         super(startPointer);
      }

      boolean successfulResult() {
         byte[] buf = getBytes();
         if (buf[0] == OK_RESULT[0]) {
            return true;
         } else if (buf[0] == ERR_RESULT[0]) {
            return false;
         } else {
            Debug.println(Debug.ADAPTOR, Debug.WARNING, 
                          "Warning: Pop3ResponseAdaptor, result line ('" + 
                          Debug.printableString(buf) +
                          "')started with unexpected character");
            throw new RuntimeException("This shouldn't happen");
         }
      }

      String getResult() {
         return getElement(0);
      }

      void setSuccessfulResult() {
         replaceElement(0, OK_RESULT);
      }

      void setErrorResult() {
         replaceElement(0, ERR_RESULT);
      }

      void replaceResult(String result) {
         replaceResult(result.getBytes());
      }

      void replaceResult(byte[] result) {
         replaceElement(0, result);
      }
   }

   protected class MultiLine extends Line {
      public MultiLine(DataAccessPointer startPointer) 
                                             throws AsyncEventException {
         super(startPointer);
      }

      // returns distance moved
      protected int findEndOfLine(DataAccessPointer p)
                                               throws AsyncEventException {
         int distance = 0;
         boolean done = false;

         while (!done) {
            // find end of line
            int lineLen = p.distanceTo((byte) '\n');
            if (lineLen == -1) {
               myWindow.expand((byte) '\n');
               lineLen = p.distanceTo((byte) '\n');
            }

            // look for eol
            if ((lineLen == 2) && (p.getByte(0) == '.')) {
               done = true;
            }

            p.advance(lineLen + 1);

            distance += (lineLen +1);
         }
         return distance;
      }

      public int length() {
         return preP.distanceTo(postP, Direction.up);
      }

   }

   public Pop3ResponseAdaptor(AdaptorWindow window, String parameter) {
      super(window, parameter);
   }

   protected boolean hasMultiLineResponse(ReadOnlyRequest request) {
      String command = request.getCommand();
      int numArgs = request.getNumArgs();

      return (command.equalsIgnoreCase("TOP") ||
              command.equalsIgnoreCase("RETR") || 
              ((numArgs == 0) && (command.equalsIgnoreCase("LIST") || 
                                  command.equalsIgnoreCase("UIDL") || 
                                  command.equalsIgnoreCase("AUTH"))
             ));
   }

   protected int getRemainingMultiLineLen() throws AsyncEventException {
      boolean done = false;
      DataAccessPointer p = new DataAccessPointer();
      currentPos.insertPointer(p);

      while (!done) {
         // make sure there is a line in the window
         int lineLen = p.distanceTo((byte) '\n');
         if (lineLen == -1) {
            window.expand((byte) '\n');
            lineLen = p.distanceTo((byte) '\n');
         }

         // look for eol
         if ((lineLen == 2) && (p.getByte(0) == '.')) {
            done = true;
         }

         p.advance(lineLen + 1);
      }

      int ret = currentPos.distanceTo(p);
      p.removeSelf();
      return ret;
   }

   protected void forwardMultiLineData() throws AsyncEventException {
      boolean done = false;

      while (!done) {
         // make sure there is a line in the window
         int lineLen = currentPos.distanceTo((byte) '\n');
         if (lineLen == -1) {
            window.expand((byte) '\n');
            lineLen = currentPos.distanceTo((byte) '\n');
         }

         // look for eol
         if ((lineLen == 2) && (currentPos.getByte(0) == '.')) {
            done = true;
         }

         currentPos.advance(lineLen + 1);
         window.contract(currentPos);
      }
   }

   // return false when no more adaptation is needed
   protected boolean handleResponse(ReadOnlyRequest request, 
                                 boolean isInitialHandshake) 
                                             throws AsyncEventException {
      response.cacheSelf();
      writeResponse();

      if ((!isInitialHandshake) && 
          response.successfulResult() && 
          hasMultiLineResponse(request)) {
         forwardMultiLineData();
      }

      return true;
   }

   protected void readResponse() throws AsyncEventException {
      try {
         // set a pointer for the start of the response
         currentPos.insertPointer(preResponsePointer);

         // get the response
         response = new Response(preResponsePointer);

         // success, move the current position
         currentPos.removeSelf();
         response.getPostPointer().insertPointer(currentPos);
      } catch (AsyncEventException e) {
         // reset things to the way they were
         preResponsePointer.removeSelf();
         throw e;
      }
   }

   protected void writeResponse() {
      preResponsePointer.removeSelf();
      DataAccessPointer postP = response.getPostPointer();
      window.contract(postP);
      postP.removeSelf();
   }

   protected void handleProtocol() {
      boolean isInitialHandshake = true;
      boolean keepAdapting = true;
      while (keepAdapting) {
         try {
            readResponse();

            if (Debug.on) {
               Debug.println(Debug.ADAPTOR, Debug.MOREINFO, 
                             this + " got response: " + response);
            }

            keepAdapting = handleResponse(getRequest(), isInitialHandshake);

            isInitialHandshake = false;
         } catch (OutputClosedException e) {
            Debug.println(Debug.ADAPTOR, Debug.INFO,
                          this + ": output closed unexpectedly");
            return;
         } catch (InputClosedException e) {
            if (Debug.on) {
               Debug.print(Debug.ADAPTOR, Debug.INFO, ")");
            }
            window.contract();
            window.closeOutput();
            Debug.println(Debug.ADAPTOR, Debug.INFO, this + ": closed");
            return;
         } catch (PlanChangedException e) {
            throw new RuntimeException("Case not handled.");
         } catch (AsyncEventException e) {
             Debug.println(Debug.ADAPTOR, Debug.WARNING,
                           this + ": Got an unexpected exception");
             return;
         }
      }
   }

}
