/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.ConductorExt.adaptors.pop3;

import java.io.*;
import java.util.*;

import FMG.Conductor.Conductor;
import FMG.Conductor.dataflow.*;
import FMG.Conductor.util.Debug;
import FMG.Conductor.util.StorageAccess;

import FMG.ConductorExt.adaptors.tools.*;

/**
 * This adaptor determines the priority of a POP3 stream based on the contents
 * of the user's mailbox and a file that contains regular expressions 
 * describing the signature of important e-mail.  An indication of the
 * priority is appended to the response to the PASS command which is
 * interpreted by the FMG.ConductorExt.adaptors.pop3.Prioritizer class.
 * <p>
 * This adaptor should be deployed upstream of a 
 * FMG.ConductorExt.adaptors.pop3.Prioritizer adaptor.
 * <p>
 * This adaptor assumes that some sort of Pop3RequestAdaptor is present
 * on the same node to decipher the request, so we know what type of response 
 * to expect.
 * <p>
 * @author Mark Yarvis (yarvis@fmg.cs.ucla.edu)
 * @version $Revision: 1.7 $      
 */
public class PriorityDetector extends Pop3ResponseAdaptor {

   private static final String MAIL_SPOOL_DIR =
                Conductor.getFilenameProperty("Pop3.MailSpoolDir", 
                                              "/var/spool/mail");

   // configuration dir
   private static final String POP_PREFS_DIR = "POP_Prefs";

   // configuration file name
   private static final String PRIORITY_CONFIG_FILE = "PriorityConfig";

   // possible resulting priority levels
   private static final int HIGH_PRIORITY = 100;
   private static final int DEFAULT_PRIORITY = 0;
   private static final int LOW_PRIORITY = -100;

   protected StorageAccess storageAccess;

   protected String username;

   public PriorityDetector(AdaptorWindow window, String parameter) {
      super(window, parameter);
   }

   protected boolean handleResponse(ReadOnlyRequest request,
                                    boolean isInitialHandshake)
                                                 throws AsyncEventException {
      boolean keepAdapting = true;

      // determine and pass the priority level of for an authenticated user
      if (! isInitialHandshake && response.successfulResult()) {
         if (request.getCommand().equalsIgnoreCase("USER")) {
            setUsername(request.getArg(0));
         } else if (request.getCommand().equalsIgnoreCase("PASS")) {
            setPassword(request.getArg(0));

            // send this priority downstream!
            int priority = getPriority();
            int lastArg = response.getNumArgs() - 1;
            response.replaceArg(lastArg, response.getArg(lastArg) + 
                                         " priority " + priority);
            keepAdapting = false;
         }
      }

      super.handleResponse(request, isInitialHandshake);
      return keepAdapting;
   }

   protected void setUsername(String username) {
      this.username = username;
   }

   protected void setPassword(String password) {
      storageAccess = new StorageAccess(username, password, 
                                        POP_PREFS_DIR);
   }

   private RegEx[] readConfig() {
      StorageAccess.ReadAccess configAccess = null;
      try {
         configAccess = storageAccess.readWorkingFile(PRIORITY_CONFIG_FILE);
      } catch (FileNotFoundException e) { }

      if (configAccess == null) {
         if (Debug.on) {
            Debug.println(Debug.ADAPTOR, Debug.INFO, 
                          this + ": Config file not found, " + 
                          "using default priority.");
         }

         return null;
      }

      if (Debug.on) {
         Debug.println(Debug.ADAPTOR, Debug.INFO, 
                       this + ": Config file is " + 
                       configAccess.filename);
      }

      BufferedReader in = new BufferedReader(
                             new InputStreamReader(configAccess.is));

      Vector results = new Vector();

      try {
         for (String line = in.readLine(); line != null; 
                                        line = in.readLine()) {
            try {
            results.addElement(new RegEx(line));
            } catch (RegExException e) {
               if (Debug.on) {
                  Debug.println(Debug.ADAPTOR, Debug.WARNING, 
                                this + " Bad config file line:" + e);
               }
            }
         }
      } catch (IOException e) {
         if (Debug.on) {
            Debug.println(Debug.ADAPTOR, Debug.WARNING, 
                          this + " Error reading config file:" + e);
         }
      } finally {
         try {
            in.close();
         } catch (IOException ignore) {}

         RegEx[] ret = new RegEx[results.size()];
         results.copyInto(ret);
         return ret;
      }
   }

   protected int getPriority() {
      // get a set of regex objects from the config file
      RegEx[] matchers = readConfig();

      if (matchers == null) {
         return DEFAULT_PRIORITY;
      }

      // get access to the local mailbox and look for match
      StorageAccess.ReadAccess mboxAccess = null;
      try {
         mboxAccess = storageAccess.readFile(MAIL_SPOOL_DIR + "/" + username);
      } catch (FileNotFoundException e) { }

      if (mboxAccess == null) {
         if (Debug.on) {
            Debug.println(Debug.ADAPTOR, Debug.INFO, 
                          this + ": No mailbox found.");
         }

         // no mail, get done quick
         return HIGH_PRIORITY;
      }

      // grep through mailbox
      BufferedReader in = new BufferedReader(
                             new InputStreamReader(mboxAccess.is));

      try {
         for (String line = in.readLine(); line != null; 
                                        line = in.readLine()) {
            try {
               for (int i = 0; i < matchers.length; i++) {
                  if (matchers[i].match(line)) {
                     return HIGH_PRIORITY;
                  }
               }
            } catch (RegExException e) {
               if (Debug.on) {
                  Debug.println(Debug.ADAPTOR, Debug.WARNING, 
                                this + " Regex failure:" + e);
               }
            }
         }
      } catch (IOException e) {
         if (Debug.on) {
            Debug.println(Debug.ADAPTOR, Debug.WARNING, 
                          this + " Error reading mailbox: " + e);
         }
      } finally {
         try {
            in.close();
         } catch (IOException ignore) {}
      }

      return LOW_PRIORITY;
   }
}
