/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.ConductorExt.adaptors.tools;

import java.io.*;

/**
 * This class implements an input stream capable of decoding a base64 
 * encoded stream as defined in the MIME spec.
 * <p>
 * @author Mark Yarvis (yarvis@fmg.cs.ucla.edu)
 * @version $Revision: 1.5 $      
 */

public class Base64InputStream extends FilterInputStream {

   private static final int DEFAULT_BUFFER_SIZE = 10240;

   int inLen = 0;
   int inPos = 0;

   // bytes that have yet to be decoded
   byte[] inBuf;

   int outLen = 0;
   int outPos = 0;

   // bytes that were decoded but not delivered
   byte[] outBuf = new byte[3];

   public Base64InputStream(InputStream is, int bufferSize) {
      super(is);
      inBuf = new byte[bufferSize];
   }

   public Base64InputStream(InputStream is) {
      this(is, DEFAULT_BUFFER_SIZE);
   }

   // returns true if at eof
   boolean refillInputBuffer() throws IOException {
      if (inPos != 0) {
         // push remaining bytes to the start
         for (int i=0; i < inLen - inPos; i++) {
            inBuf[i] = inBuf[i+inPos];
         }
         inLen = inLen - inPos;
         inPos = 0;
      }

      int numRead = in.read(inBuf, inLen, inBuf.length - inLen);

      if (numRead == -1) {
         return false;
      }

      inLen += numRead;
      return true;
   }

   int getNextByte() {
      if (outLen - outPos > 0) {
         return ((int) outBuf[outPos++]) & 0xff;
      } else {
         if (! convertFourToThree()) {
            return -1;
         } else {
            return getNextByte();
         }
      }
   }

   int getValue(byte b) {
      if ((b >= 'A') && (b <= 'Z')) {
         return (b - 'A');
      } else if ((b >= 'a') && (b <= 'z')) {
         return (b - 'a' + 26);
      } else if ((b >= '0') && (b <= '9')) {
         return (b - '0' + 52);
      } else if (b == '+') {
         return 62;
      } else if (b == '/') {
         return 63;
      } else if (b == '=') {
         return 0; // special case!
      } else {
         return -1;
      }
   }

   // return true on success
   boolean convertFourToThree() {
      long decoder = 0;

      int numBytesDecoded = 0;
      int newInPos = inPos;
      int numEqualsSigns = 0;

      while ((numBytesDecoded < 4) && (newInPos < inLen)) {
         if (inBuf[newInPos] == '=') {
            numEqualsSigns++;
         }
         int value = getValue(inBuf[newInPos]);
         if (value != -1) {
            numBytesDecoded++;
            decoder <<= 6;
            decoder += value;
         } else {
         }

         newInPos++;
      }

      // there weren't four decodable bytes
      if (numBytesDecoded < 4) {
         return false;
      }

      inPos = newInPos;

      for (int i = 2; i >=0; i--) {
         outBuf[i] = (byte) (decoder & 0xff);
         decoder >>= 8;
      }

      outLen = 3 - numEqualsSigns;
      outPos = 0;

      return true;
   }

   public int available() throws IOException {
// TODO: this is probably wrong!!!
      return (((super.available() + outLen) * 2) / 3);
   }

   public boolean markSupported() {
      return false;
   }

   public int read() throws IOException {
      int b = getNextByte();
      while (b == -1) {
         if (!refillInputBuffer()) {
            // no data left
            return -1;
         } else {
            b = getNextByte();
         }
      }

      return b;
   }

   public int read(byte[] buf, int off, int len) throws IOException {
      if (len == 0) {
         return 0;
      }

      // try to get at least one byte
      int b = getNextByte();

      // need to refill our buffer
      while (b == -1) {
         if (! refillInputBuffer()) {
            // no data left
            return -1;
         } else {
            b = getNextByte();
         }
      }

      int numBytes = 0;
      while ((b!=-1) && (numBytes < len)) {
         buf[off+numBytes]=(byte)b;
         numBytes++;
         b = getNextByte();
      }

      return numBytes;
   }

   public long skip(long n) throws IOException {
// TODO: finish
      return 0;
   }

   // test function
   public static void main(String[] args) {
      Base64InputStream b64in = new Base64InputStream(System.in);
      try {
         byte[] buf = new byte[10240];

         while (true) {
            int len = b64in.read(buf);
            if (len == -1) {
               return;
            }
            System.out.write(buf, 0, len);
         }
      } catch (IOException e) {
         e.printStackTrace();
      }
   }
}
