/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.ConductorExt.adaptors.tools;

import java.util.*;

import FMG.Conductor.util.*;
import FMG.Conductor.dataflow.*;

/**
 * The PriorityWriter is a utility class for writing a portion of the
 * data stream while taking turns with other PriorityWriter objects.
 * Each PriorityWriter can declare a priority and a delay factor.  The highest
 * priority job that has data to write is allowed to write.  If a delay
 * factor is defined, lower priority jobs cannot start until that amount of
 * time after the higher prioirty job specifying the delay has finished.
 * <p>
 * TODO: There needs to be provision for synchronization between writers
 * that are writing to the same physical channel.  Currenlty, synchronization 
 * is global per node.
 * <p>
 * @author Mark Yarvis (yarvis@fmg.cs.ucla.edu)
 * @version $Revision: 1.12 $      
 */

public class PriorityWriter {

   protected static final int WRITE_GRANULARITY = 1024;
   protected static final String SYNC_KEY = "CommonWriteSync";

   protected AdaptorWindow window;
   protected int priority;
   protected int bytesToWrite;
   protected int bytesWritten;
   protected PrioritySync sync;
   protected int winSize;

   protected long lowerPriStartDelay = 0;
   protected int delayPriorityMargin = 0;

   protected static class PrioritySync {
      protected Vector waitList = new Vector();
      protected PriorityWriter currentJob = null;
      protected Object schedLock = new Object();

      protected long delayUntilTime = 0;
      protected int delayPriority = Integer.MIN_VALUE;
      protected PriorityWriter lastJob;

      void requestWrite(PriorityWriter myJob) throws AsyncEventException {
         waitList.addElement(myJob);

         schedule();

         while (true) {

            performDelayWait(myJob);

            if (currentJob == myJob) {
               AsyncEventException thrown = null;
               boolean res = false;

               try {
                  res = myJob.doAWriteCycle();
               } catch (AsyncEventException e) {
                  thrown = e;
               }

               if ((res == false) || (thrown != null)) {

                  synchronized (schedLock) {
                     waitList.removeElement(myJob);
                     if (currentJob == myJob) {   // if i'm still current
                        currentJob = null;
                     }

                     schedule();
                  }

                  if (thrown != null) {
                     throw thrown;
                  }

                  return;
               }
            } else {
               synchronized (schedLock) {
                  while (currentJob != myJob) {
                     if (Debug.on) {
                        Debug.println(Debug.ADAPTOR, Debug.DETAIL, 
                                      this + " waiting for " + myJob);
                     }
                     try {
                        schedLock.wait();
                     } catch (InterruptedException e) {}  // ignore, for now
                  }
                  if (Debug.on) {
                     Debug.println(Debug.ADAPTOR, Debug.DETAIL,
                                   this + " is running");
                  }
               }
            }
         }
      }

      protected final void performDelayWait(PriorityWriter myJob) {
         if (myJob.priority < delayPriority) {
            while (delayUntilTime != 0) {
               long sleepTime = delayUntilTime - System.currentTimeMillis();

               if (sleepTime < 0) {           // has delay time expired?
                  delayUntilTime = 0;
                  return;
               }
               if (Debug.on) {
                  Debug.println(Debug.ADAPTOR, Debug.DETAIL,
                                "Waiting delay time " + sleepTime);
               }
               try {
                  Thread.currentThread().sleep(sleepTime);
               } catch (InterruptedException e) {}
            }
         }
      }

      protected void schedule() {
         PriorityWriter newJob = null;

         synchronized (schedLock) {
            for (int i=waitList.size() - 1; i>=0; i--) {
               PriorityWriter other = (PriorityWriter) waitList.elementAt(i);
               if ((newJob == null) || (newJob.priority < other.priority)) {
                  newJob = other;
               }
            }

            if (newJob != currentJob) {       // has a new job been scheduled?

               /*
                * If the most recent last job is of higher priority and 
                * requires a delay beyond the current one, establish that as
                * the new delay.
                */
               if ((lastJob != null) && (lastJob.priority >= delayPriority) && 
                   (lastJob.lowerPriStartDelay != 0)) {
                  long newDelayTime = System.currentTimeMillis() + 
                                           lastJob.lowerPriStartDelay;
                  if (newDelayTime > delayUntilTime) {
                     delayUntilTime = newDelayTime;
                     delayPriority = lastJob.priority - 
                                           lastJob.delayPriorityMargin;
                  }
               }
               lastJob = newJob;

               currentJob = newJob;
               if (Debug.on) {
                  Debug.println(Debug.ADAPTOR, Debug.DETAIL, 
                                "New current job is " + currentJob);
               }

               schedLock.notifyAll();
            }
         }
      }

   }

   public PriorityWriter(AdaptorWindow window, int priority, IAC iac) {
      this(window, priority, 0, 0, iac);
   }

   public PriorityWriter(AdaptorWindow window, int priority, 
                         long lowerPriStartDelay, IAC iac) {
      this(window, priority, lowerPriStartDelay, 0, iac);
   }

   public PriorityWriter(AdaptorWindow window, int priority,
                         long lowerPriStartDelay, int delayPriorityMargin,
                         IAC iac) {
      this.window = window;
      this.priority = priority;
      this.lowerPriStartDelay = lowerPriStartDelay;
      this.delayPriorityMargin = delayPriorityMargin;

      sync = getSync(iac);
   }

   public void setPriority(int priority) {
      this.priority = priority;
   }

   protected PrioritySync getSync(IAC iac) {
      PrioritySync prioritySync = (PrioritySync) iac.lookup(SYNC_KEY);
      if (prioritySync == null) {
         PrioritySync newSync = new PrioritySync();
         iac.store(SYNC_KEY, newSync, false);

         // Note, it may be that (prioritySync != newSync), but that's ok
         prioritySync = (PrioritySync) iac.lookup(SYNC_KEY);
      }

      return prioritySync;
   }

   public void writeData(int len) throws AsyncEventException {
      if (Debug.on) {
         Debug.println(Debug.ADAPTOR, Debug.EXTRAINFO,
                       this + " writing " + len + " bytes");
      }
      bytesToWrite = len;
      bytesWritten = 0;

      winSize = window.size();

      while (bytesToWrite > bytesWritten) {
         sync.requestWrite(this);
         if (bytesToWrite > bytesWritten) {
            if (Debug.on) {
               Debug.println(Debug.ADAPTOR, Debug.EXTRAINFO,
                             this + " waiting for more data");
            }
            // wait for more data
            winSize += window.expand();
         }
      }
   }

   // how many bytes were written? ... useful in case of exception
   public int bytesWritten() {
      return bytesWritten;
   }

   protected boolean doAWriteCycle() throws AsyncEventException {

      // don't write past end of content length
      if (winSize > (bytesToWrite - bytesWritten)) {
         winSize = bytesToWrite - bytesWritten;
      }

      if (winSize > WRITE_GRANULARITY) {

         window.contract(WRITE_GRANULARITY);
         bytesWritten += WRITE_GRANULARITY;
         winSize -= WRITE_GRANULARITY;

         return true;

      } else { 
   
         // get more data if some is ready
         if (window.dataAvailable()) {
            winSize += window.expand(0);
            return doAWriteCycle();
         }

         window.contract(winSize);
         bytesWritten += winSize;
         winSize = 0;

         return false;
      }
   }

}
