/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.ConductorExt.adaptors.tools;

import java.io.*;
import FMG.Conductor.dataflow.*;
import FMG.Conductor.util.Debug;

/*
 * This class provides a glue between code that expects an InputStream and
 * the AdaptorWindow.  This allows code not written for Conductor to be
 * glued into an adaptor.  Typically this class is used together with
 * a WindowOutputStream.  Data from the window can be read via a 
 * WindowInputStream modified and then written back to the window using
 * a WindowOutputStream.  Because Conductor can't know what was done to
 * the stream, the data produced will be one big segment.
 * <p>
 * @author Mark Yarvis (yarvis@fmg.cs.ucla.edu)
 * @version $Revision: 1.12 $      
 */

public class WindowInputStream extends InputStream {

   protected AdaptorWindow window;
   protected BlockModPointer readPos;

   protected int bytesRead = 0;

   boolean isLimited = false;
   int bytesRemaining;

   WindowOutputStream os;

   public WindowInputStream(AdaptorWindow w, int pos) {
      super();
      window = w;
      readPos = new BlockModPointer();

      window.insertPointer(readPos);
      readPos.advance(pos);
      readPos.blockModStart();
   }

   void setOutputStream(WindowOutputStream os) {
      this.os = os;
   }

   public WindowInputStream(AdaptorWindow w, int pos, int distanceAllowed) {
      this(w, pos);
      isLimited = true;
      bytesRemaining = distanceAllowed;
   }

   public WindowInputStream(AdaptorWindow w, DataAccessPointer p) {
      super();
      window = w;

      readPos = new BlockModPointer();
      p.insertPointer(readPos);
      readPos.blockModStart();
   }

   public WindowInputStream(AdaptorWindow w, DataAccessPointer p, 
                                             int distanceAllowed) {
      this(w, p);
      isLimited = true;
      bytesRemaining = distanceAllowed;
   }

   BlockModPointer getPointer() {
      return readPos;
   }

   AdaptorWindow getWindow() {
      return window;
   }

   public int read() throws IOException {
      if (readPos == null) {
         return -1;
      }

      if (isLimited && (bytesRemaining <= 0)) {
         return -1;
      }

      if (window.countRemaining(readPos)==0) {
         try {
            window.expand();
         } catch (AsyncEventException e) {
            if (window.inputClosed()) {
               return -1;
            }

            if (window.switchPlanInWindow()) {
               throw new IOException();
            }
         }
      }

      byte ret = readPos.getByte(0);
      readPos.deleteBytes(1);
      bytesRead++;

      if (isLimited) {
         bytesRemaining--;

         if (Debug.on && (bytesRemaining < 0)) {
            Debug.println(Debug.ADAPTOR, Debug.WARNING, 
                          "WindowInputStream: overran input limit");
         }
      }

      return (ret & 0xff);
   }

   public int read(byte buf[], int bufOff, int bufLen) throws IOException {
      if (readPos == null) {
         return -1;
      }

      if (isLimited && (bytesRemaining <= 0)) {
         return -1;
      }

      getAvailableData();

      int dataLen = window.countRemaining(readPos);

      while (dataLen == 0) {
         if (window.inputClosed()) {
            return -1;
         }
         try {
            window.expand();
         } catch (AsyncEventException e) {
            if (window.switchPlanInWindow()) {
               throw new IOException();
            }
            if (window.inputClosed()) {
               return -1;
            }
            // if output closed, just go around again
         }

         dataLen = window.countRemaining(readPos);
      }

      // transfer the min of dataLen, bufLen, and bytesRemaining
      dataLen = (dataLen > bufLen ? bufLen : dataLen);
      if (isLimited) {
         dataLen = (dataLen > bytesRemaining ? bytesRemaining : dataLen);
      }

      readPos.getBytes(buf, bufOff, 0, dataLen);

      readPos.deleteBytes(dataLen);

      bytesRead += dataLen;

      if (isLimited) {
         bytesRemaining -= dataLen;

         if (Debug.on && (bytesRemaining < 0)) {
            Debug.println(Debug.ADAPTOR, Debug.WARNING, 
                          "WindowInputStream: overran input limit");
         }
      }

      return dataLen;
   }

   public long skip(long n) throws IOException {
      // TODO: fix this!!!
      if (n > Integer.MAX_VALUE) {
         throw new RuntimeException("Case not handled");
      }

      if (readPos == null) {
         return -1;
      }

      if (isLimited && (bytesRemaining <= 0)) {
         return -1;
      }

      if (isLimited) {
         // don't try to skip beyond where we're allowed to go
         n = (n > bytesRemaining ? bytesRemaining: n);
      }

      ensureSufficientBytes((int) n);

      int dataLen = window.countRemaining(readPos);

      // skip the min of dataLen, requested data to skip
      dataLen = (n<dataLen ? (int) n : dataLen);

      readPos.deleteBytes(dataLen);

      bytesRead += dataLen;

      if (isLimited) {
         bytesRemaining -= dataLen;

         if (Debug.on && (bytesRemaining < 0)) {
            Debug.println(Debug.ADAPTOR, Debug.WARNING, 
                          "WindowInputStream: overran input limit");
         }
      }

      return dataLen;
   }

   private final void ensureSufficientBytes(int needed) throws IOException {
      needed -= window.countRemaining(readPos);

      if (needed > 0) {
         try {
            window.expand(needed);
         } catch (AsyncEventException e) {
            if (window.switchPlanInWindow()) {
               throw new IOException();
            }
         }
      }
   }

   private final void getAvailableData() throws IOException {
      if (window.dataAvailable()) {
         try {
            window.expand(0);
         } catch (AsyncEventException e) {
            if (window.switchPlanInWindow()) {
               throw new IOException();
            }
            // don't worry if the window is closed
         }
      }
   }

   public int available() throws IOException {
      if (readPos == null) {
         return -1;
      }

      getAvailableData();

      int available = window.countRemaining(readPos);

      // don't report more data than we're allowed to read
      if (isLimited) {
         available = (available > bytesRemaining ? bytesRemaining : available);
      }

      return available;
   }

   public void close() throws IOException {
      if (os.isClosed()) {
         os.cleanup();
      }
      readPos = null;
   }

   public boolean isClosed() {
      return (readPos == null);
   }

   public int bytesRead() {
      return bytesRead;
   }

}
