/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.ConductorExt.adaptors.tools;
      
import java.io.*;
import FMG.Conductor.dataflow.*;

/*
 * This class provides a glue between code that expects an OutputStream and
 * the AdaptorWindow.  This allows code not written for Conductor to be
 * glued into an adaptor.  Typically this class is used together with
 * a WindowInputStream.  Data from the window can be read via a 
 * WindowInputStream modified and then written back to the window using
 * a WindowOutputStream.  Because Conductor can't know what was done to
 * the stream, the data produced will be one big segment.
 * <p>
 * @author Mark Yarvis (yarvis@fmg.cs.ucla.edu)
 * @version $Revision: 1.7 $      
 */

public class WindowOutputStream extends OutputStream {

   protected BlockModPointer writePos;
   protected AdaptorWindow window;

   private WindowInputStream is;

   private byte[] oneByte = new byte[1];

   protected boolean supressFlush = false;

   protected boolean autoFlush = true;

   protected int bytesWritten = 0;

   public WindowOutputStream(WindowInputStream is) {
      this.is = is;
      is.setOutputStream(this);
      writePos = is.getPointer();
      window = is.getWindow();
   }

   public void write(int b) throws IOException {
      oneByte[0]=(byte) b;
      writePos.insertBytes(oneByte, 0, 1);
      bytesWritten++;
      if (autoFlush) {
         flush();
      }
   }

   public void write(byte b[],
                     int off,
                     int len) throws IOException {
      writePos.insertBytes(b, off, len);
      bytesWritten+=len;
      if (autoFlush) {
         flush();
      }
   }

   public void flush() throws IOException {
      if (! supressFlush) {
         window.contract(writePos);
      }
   }

   public void setSupressFlush(boolean ToF) {
      supressFlush = ToF;
   }

   public void setAutoFlush(boolean ToF) {
      autoFlush = ToF;
   }

   public void close() throws IOException {
      if (is.isClosed()) {
         cleanup();
      }
      writePos = null;
   }

   void cleanup() throws IOException {
      writePos.blockModEnd();
      if (autoFlush) {
         flush();
      }
      writePos.removeSelf();
   }

   public boolean isClosed() {
      return (writePos == null);
   }

   public int bytesWritten() {
      return bytesWritten;
   }

}
