/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.ConductorExt.iface;

import java.io.*;

import FMG.Conductor.util.*;
import FMG.Conductor.iface.Interface;
import FMG.Conductor.dataflow.IOManager;

public class CommandLineInterface extends Interface {
   protected BufferedReader in;
   protected PrintStream out;

   public CommandLineInterface() {
      this(System.in, System.out);
   }

   public CommandLineInterface(InputStream in, PrintStream out) {
      this.in = new BufferedReader(new InputStreamReader(in));
      this.out = out;
   }

   private void setupThread(Thread myThread) {
      myThread.setPriority(Thread.MIN_PRIORITY);
      if (Debug.on) {
         myThread.setName("CommandLineInterface");
      }
   }

   protected void runInterface() {
      setupThread(Thread.currentThread());
      try {
         while (true) {
            out.print("Conductor> ");
            out.flush();

            String s = in.readLine();

            if (s.equals("q") || s.equals("quit")) {
               in.close();
               out.close();
               return;

            } else if (s.equals("shutdown")) {
               System.exit(0);

            } else if (s.startsWith("v") || s.startsWith("V")) {
               if (! Debug.on) {
                  out.println("Debugging not enabled, " + 
                                     "verbosity not set!"); 
                  continue;
               }

               String arg = ((s.startsWith("v ") || s.startsWith("V ")) ? 
                                       s.substring(2) : s.substring(1));
               try {
                  int equalsPos = arg.indexOf("=");
                  if (equalsPos == -1) {
                     int newVerbosity = Integer.parseInt(arg);
                     Debug.setLevel(newVerbosity);
                     Debug.println("Verbosity level is now " + newVerbosity);
                  } else {
                     String module = arg.substring(0, equalsPos);
                     int newVerbosity = Integer.parseInt(
                                          arg.substring(equalsPos+1, 
                                                        arg.length()));
                     Debug.setLevel(module, newVerbosity);
                  }
               } catch (NumberFormatException e) {
                  out.println("Couldn't parse arg.  " + "How much verbosity?");
               }

            } else if (s.startsWith("m")) {
               String arg = (s.startsWith("m ") ? s.substring(2) : 
                                                  s.substring(1));
               if (arg.equalsIgnoreCase("off")) {
                  Measure.turnOff();
                  out.println("Measurement mode off");
               } else if (arg.length() == 0) {
                  Measure.turnOn(null);   // turn everything on
                  out.println("All measurements enabled");
               } else {
                  Measure.turnOn(arg);
                  out.println("Measurement mode enabled");
               }

            } else if (s.equals("threads")) {

               Thread[] tarray = new Thread[100];
               int numThreads = Thread.enumerate(tarray);

               for (int i=0; i<numThreads; i++) {
                  out.println(">>>>> " + tarray[i].toString());
               }

            } else if (s.equals("datapaths")) {
               IOManager.dumpAllDatapaths(out);

            } else if (s.equals("gc")) {
               System.gc();
               out.println("Garbage collection done.");

            } else if (s.equals("threadkill")) {
               ThreadRecycler.killUnusedThreads();
               out.println("Done.");

            } else if (s.equals("r")) {
               Recycler.dumpRecyclingInfo(out);

            } else if (s.startsWith("?") || s.startsWith("h")) {
               out.println("Available commands:");
               out.println("   v <level>    - set verbosity to <level>");
               out.println("   m <mask>     - enable measurements");
               out.println("   m off        - disable measurements");
               out.println("   r            - display recycling status");
               out.println("   threads      - print threads");
               out.println("   threadkill   - kill all cached threads");
               out.println("   datapaths    - dump datapaths");
               out.println("   gc           - run garbage collection");
               out.println("   shutdown     - shutdown Conductor");
               out.println("   q (or quit)  - quit interface");
               out.println("   ? (or h)     - print this message");

            } else {
               out.println("Huh?");
            }
         }
      } catch (IOException e) {
         out.println("Exiting on EOF");
         System.exit(0);
      }
   }
}
