/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.ConductorExt.monitor;

import java.io.*;
import java.util.*;

import FMG.Conductor.Conductor;
import FMG.Conductor.monitor.*;
import FMG.Conductor.comm.*;
import FMG.Conductor.planning.messages.PlanningInfo;
import FMG.Conductor.util.*;
import FMG.Conductor.net.AddressEntry;

/**
 * This class implements a static monitor.  No actual node monitoring 
 * is done.  The information passed to the planner includes that passed
 * to <code>getLocalInfo()</code> and information obtained from a
 * properties file.
 *
 * @author Mark Yarvis (yarvis@fmg.cs.ucla.edu)
 * @version $Revision: 1.17 $
 */

public class StaticMonitor extends EnvMonitor {

   protected static class LinkInfo {
      AddressEntry remoteNode;
      AddressEntry localInterface;
      String source;
      Hashtable properties;
      int ruleNum;

      protected LinkInfo(AddressEntry remoteNode, AddressEntry localInterface,
                         String source, Hashtable properties, int ruleNum) {
         this.remoteNode = remoteNode;
         this.localInterface = localInterface;
         this.source = source;
         this.properties = properties;
         this.ruleNum = ruleNum;
      }

      protected boolean isMatch(NodeID remoteNode, NodeID localInterface) {
         return (this.remoteNode.isMatch(remoteNode.getBytes()) && 
                 this.localInterface.isMatch(localInterface.getBytes()));
      }

      protected void addInfo(PlanningInfo pi, String prefix) {
         if (properties == null) {
            return;
         }

         if (Debug.on) {
            Debug.println(Debug.PLANNING, Debug.MOREINFO,
                          "StaticMonitor: Adding info from " + source + 
                          " for " + prefix + " using rule " + ruleNum);
         }

         Enumeration keys = properties.keys();
         while (keys.hasMoreElements()) {
            String key = (String) keys.nextElement();
            pi.addInfo(prefix + "." + key, (String) properties.get(key));
         }
      }
   }

   protected static final NodeID localhost = NodeID.getNewLocalID();

   protected static PropsLoader propsLoader;

   protected static Hashtable defaultUserPrefs;

   protected static Hashtable nodeInfo;

   protected static LinkInfo[] linkInfo;

   protected static final String PROT_SPECIFIC_USER_PREFS_FILE_PREFIX = 
                                            "StaticMonitor.UserPrefsFile.";

   protected static final String DEFAULT_USER_PREFS_FILE = 
                     Conductor.getFilenameProperty(
                                 "StaticMonitor.UserPrefsFile", 
                                 "./config/UserPrefs");

   protected static final String NODE_INFO_FILE = 
               Conductor.getFilenameProperty("StaticMonitor.NodeInfoFile", 
                                             "./config/NodeInfo");

   protected static final String LINK_INFO_FILE = 
               Conductor.getFilenameProperty("StaticMonitor.LinkInfoFile", 
                                             "./config/LinkInfo");

   static {
      propsLoader = new PropsLoader(true);
      defaultUserPrefs = propsLoader.getProps(DEFAULT_USER_PREFS_FILE);
      if (defaultUserPrefs == null) {
         if (Debug.on) {
            Debug.println(Debug.WARNING, "WARNING: StaticMonitor failed to " + 
                          "load default user preferences file '" + 
                          DEFAULT_USER_PREFS_FILE + "'");
         }
         defaultUserPrefs = new Hashtable();
      }

      nodeInfo = propsLoader.getProps(NODE_INFO_FILE);
      if (nodeInfo == null) {
         if (Debug.on) {
            Debug.println(Debug.WARNING, "WARNING: StaticMonitor failed to " + 
                          "load node information file '" + 
                          NODE_INFO_FILE + "'");
         }
         nodeInfo = new Hashtable();
      }

      Properties linkProps = propsLoader.getProps(LINK_INFO_FILE);
      if (linkProps != null) {
         linkInfo = processLinkInfo(linkProps);
      } else {
         if (Debug.on) {
            Debug.println(Debug.WARNING, "WARNING: StaticMonitor failed to " + 
                          "load link information file '" + 
                          LINK_INFO_FILE + "'");
         }
         linkInfo = new LinkInfo[0];
      }
   }

   protected static LinkInfo[] processLinkInfo(Hashtable linkProps) {
      // get number of entries
      int numEntries;
      try {
         numEntries = Integer.parseInt((String) linkProps.get("NumEntries"));
      } catch (Exception e) {
         throw new RuntimeException("Error reading link properties: " + 
                                    "NumEntries?");
      }

      String filename = null;

      Vector keepers = new Vector(numEntries);
      for (int i=0; i<numEntries; i++) {
         try {
            AddressEntry localNode = new AddressEntry(
                               (String) linkProps.get(i + ".LocalNode"));
            // only keep entries that pertain to this node
            if (! localNode.isMatch(localhost.getBytes())) {
               continue;
            }

            AddressEntry remoteNode = new AddressEntry(
                                (String) linkProps.get(i + ".NextNode"));
            AddressEntry localInterface = new AddressEntry(
                                (String) linkProps.get(i + ".LocalInterface"));
            filename = (String) linkProps.get(i + ".File");
            Hashtable properties = null;
            if (filename != null) {
               properties = propsLoader.getProps(filename);
            }

            if (properties == null) {
               throw new RuntimeException("Error loading file '" + filename);
            }

            keepers.addElement(new LinkInfo(remoteNode, localInterface, 
                                            filename, properties, i));
         } catch (Exception e) {
            throw new RuntimeException(
                           "Unable to decipher link info entry " + i);
         }
      }

      // turn the keepers into an array
      LinkInfo[] ret = new LinkInfo[keepers.size()];
      keepers.copyInto(ret);

      return ret;
   }

   /**
    * This method returns a PlanningInfo message containing only the
    * data passed to it.  It is called by the Path object to gather
    * the status of the node and links for forwarding to the planner.
    * @param connID The id of the connection for which planning is occurring.
    * @param nodeID The local nodeID.
    * @param prevNode The id of the previous node in the planning path.
    * @param prevInterface The interface to use to talk with the previous node.
    * @param nextNode The id of the next node in the planning path. 
    * @param nextInterface The interface to use to talk with the next node.
    * @param srcNode The id of the ultimate connection source.
    * @param srcPort The port number of the ultimate connection source.
    * @param dstNode The id of the ultimate connection destination.
    * @param dstPort The port number of the ultimate connection destination.
    * @param isPlanner A boolean indicating whether this is the last node
    *                  in the planning path.  If so, nextNode is probably 
    *                  null or invalid.
    * @param isRoot A boolean indicating whether this is the first node
    *               in the planning path.  If so, nextNode is probably null
    *               or invalid.
    * @return A message object containing the planning data and the
    *         statically defined node/link info from the config file.
    */
   public PlanningInfo getLocalInfo(ConnID connID, NodeID nodeID, 
                                    NodeID prevNode, NodeID interfaceToPrev,
                                    NodeID nextNode, NodeID interfaceToNext,
                                    NodeID srcNode, int srcPort,
                                    NodeID dstNode, int dstPort,
                                    boolean isPlanner, boolean isRoot) {
      PlanningInfo pi = PlanningInfo.getNewMessage(connID, nodeID, 
                                                   prevNode, interfaceToPrev,
                                                   nextNode, interfaceToNext,
                                                   isPlanner, isRoot);
      addUserPrefs(pi, dstPort);
      addNodeInfo(pi);
      addLinkInfo(pi, prevNode, interfaceToPrev, nextNode, interfaceToNext);

      return pi;
   }

   protected void addUserPrefs(PlanningInfo pi, int dstPort) {
      String filename = Conductor.getFilenameProperty(
                              PROT_SPECIFIC_USER_PREFS_FILE_PREFIX + dstPort, 
                              DEFAULT_USER_PREFS_FILE);

      Properties props = propsLoader.getProps(filename);
      if (props != null) {
         pi.copyInfo(props);
      } else {
         if (Debug.on) {
            Debug.println(Debug.WARNING, "Failed to load '" + filename + 
                                      "', using previously loaded defaults.");
         }
         pi.copyInfo(defaultUserPrefs);
      }
   }

   protected void addNodeInfo(PlanningInfo pi) {
      pi.copyInfo(nodeInfo);
   }

   protected void addLinkInfo(PlanningInfo pi, 
                              NodeID prevNode, NodeID interfaceToPrev,
                              NodeID nextNode, NodeID interfaceToNext) {
      LinkInfo ni;

      if (! prevNode.isNull()) {
         ni = findLinkInfo(prevNode, interfaceToPrev);
         if (ni != null) {
            ni.addInfo(pi, "PrevLinkProps");
         }
      }

      if (! nextNode.isNull()) {
         ni = findLinkInfo(nextNode, interfaceToNext);
         if (ni != null) {
            ni.addInfo(pi, "NextLinkProps");
         }
      }
   }

   protected LinkInfo findLinkInfo(NodeID node, NodeID interfaceToNode) {
      for (int i=0; i<linkInfo.length; i++) {
         if (linkInfo[i].isMatch(node, interfaceToNode)) {
            return linkInfo[i];
         }
      }

      return null;
   }

   /**
    * This method is called to register a new output stream.  In this case,
    * it does nothing.
    * @param nextNode The id of the node to which this output stream flows.
    * @param outputStream The output stream that was created.
    * @return No return value.
    */
   public synchronized void registerOutputStream(NodeID nextNode, 
                                    BasicMessageOutputStream outputStream) {
   }

   /**
    * This method is called to unregister an output stream.  In this case,
    * it does nothing.
    * @param outputStream The output stream that has closed.
    * @return No return value.
    */
   public synchronized void unregisterOutputStream(
                                    BasicMessageOutputStream outputStream) {
   }

   /**
    * This method is called to register a new datapath.  In this case, it
    * does nothing.
    * @param connID The id of the connection this datapath is associated with.
    * @param nextNode The next node along the path.
    * @param request The expected data parameters for this connection.
    * @return No return value.
    */
   public synchronized void registerDatapath(ConnID connID, NodeID nextNode, 
                                             LinkParams request) {
   }

   /**
    * This method is called to unregister a datapath.  In this case, it
    * does nothing.
    * @param connID The id of the connection this datapath is associated with.
    * @param nextNode The next node along the path.
    * @return No return value.
    */
   public synchronized void unregisterDatapath(ConnID connID,  
                                                   NodeID nextNode) {
   }
}
