/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.ConductorExt.monitor.basic;

import java.util.*;
import java.io.*;
import java.net.*;

import FMG.Conductor.util.*;
import FMG.Conductor.comm.*;
import FMG.Conductor.Conductor;
import FMG.Conductor.planning.messages.PlanningInfo;
import FMG.Conductor.monitor.EnvMonitor;
import FMG.Conductor.monitor.LinkParams;

public class BasicMonitor extends EnvMonitor implements Runnable {

   private static final long SAMPLE_FREQ = 
                    Conductor.getIntProperty("BasicMonitor.SampleFreq", 5000);
   private static final long PROBE_FREQ =
                    Conductor.getIntProperty("BasicMonitor.ProbeFreq", 20000);

   private Hashtable nextNodes = new Hashtable();
   private NodeID thisNode = NodeID.getNewLocalID();

   private ByteArrayOutputStream baos = new ByteArrayOutputStream();
   private DataOutputStream dos = new DataOutputStream(baos);

   private Random rand = new Random();

   private ProbeHandler probeHandler;

   public BasicMonitor() {
      probeHandler = ProbeHandler.createAndStart(this);
      (new Thread(this)).start();
   }

   public PlanningInfo getLocalInfo(ConnID connID, NodeID nodeID, 
                                    NodeID prevNode, NodeID prevInterface,
                                    NodeID nextNode, NodeID nextInterface,
                                    NodeID srcNode, int srcPort,
                                    NodeID dstNode, int dstPort,
                                    boolean isPlanner, boolean isRoot) {
      if (nextNode != null) {
         CommStats nodeStats = (CommStats) nextNodes.get(nextNode);
         if (nodeStats==null) {
            System.out.println("No stats, can't make prediction to " + 
                               nextNode + " yet");
         } else {
            nodeStats.predictLocalCharacteristics(true);
         }
      }

      if (prevNode != null) {
         CommStats nodeStats = (CommStats) nextNodes.get(prevNode);
         if (nodeStats==null) {
            System.out.println("No stats, can't make prediction to " + 
                               prevNode + " yet");
         } else {
            nodeStats.predictLocalCharacteristics(true);
         }
      }

      return PlanningInfo.getNewMessage(connID, nodeID, 
                                        prevNode, prevInterface, 
                                        nextNode, nextInterface, 
                                        isPlanner, isRoot);
   }

   public synchronized void registerOutputStream(NodeID nextNode, 
                                    BasicMessageOutputStream outputStream) {
      nextNode = nextNode.duplicate();

      CommStats nodeStats = (CommStats) nextNodes.get(nextNode);

      if (nodeStats == null) {
         nodeStats = new CommStats(nextNode);
         nextNodes.put(nextNode, nodeStats);
      }

      nodeStats.addStream(outputStream);
   }

   public synchronized void unregisterOutputStream(
                                    BasicMessageOutputStream outputStream) {
      Enumeration statList = nextNodes.elements();
      while (statList.hasMoreElements()) {
         ((CommStats) statList.nextElement()).removeStream(outputStream);
      }
   }

   /* TODO: should allow other characteristics as well */
   public synchronized void registerDatapath(ConnID connID, NodeID nextNode, 
                                             LinkParams request) {
      nextNode = nextNode.duplicate();
      connID = connID.duplicate();

      CommStats nodeStats = (CommStats) nextNodes.get(nextNode);
      if (nodeStats == null) {
         nodeStats = new CommStats(nextNode);
         nextNodes.put(nextNode, nodeStats);
      }

      nodeStats.addDatapath(connID, request);
   }

   public synchronized void unregisterDatapath(ConnID connID, 
                                               NodeID nextNode) {
      CommStats nodeStats = (CommStats) nextNodes.get(nextNode);
      if (nodeStats != null) {
         nodeStats.removeDatapath(connID);
      }
   }

   /* rtt is measured from local time only
     forwardDelay includes clock skew between the two nodes
     interpacket time is measured with remote time only
    */
   void handleProbeResults(NodeID sourceNode, int probeId, long rtt, 
                           long forwardDelay, long interPacketTime) {
      CommStats nodeStats = (CommStats) nextNodes.get(sourceNode);
      if (nodeStats.probeCount() != probeId) {
         System.out.println("##############################################");
         System.out.println("########### ack id " + probeId + 
                            " is behind probe count " + 
                            nodeStats.probeCount());
         System.out.println("##############################################");
      }

      ((CommStats) nextNodes.get(sourceNode)).
                        updateWithProbeResults(rtt, forwardDelay, 
                                               interPacketTime);
   }

   private long approximately(long num) {
      return (num + (rand.nextLong() % (num / 10)));
   }

   public void run() {
      Random rand = new Random();
      DatagramSocket s;
      Thread myThread = Thread.currentThread();

      try {
         s = new DatagramSocket();
      } catch (IOException e) {
         throw new RuntimeException("This shouldn't happen");
      }

      long nextSampleTime = 0;
      long nextProbeTime = 0;
      while (true) {
         long wakeupTime = System.currentTimeMillis();

         Enumeration nodeStats = nextNodes.elements();

         while (nodeStats.hasMoreElements()) {
            CommStats cs = (CommStats) nodeStats.nextElement();
            if (wakeupTime > nextProbeTime) {
               cs.incProbeCount();
               probeHandler.sendProbe(cs.nodeAddress(), cs.probeCount());
            }
            if (wakeupTime > nextSampleTime) {
               cs.updateLocalInfo();
               cs.predictLocalCharacteristics(false);
            }
         }

         
         if (PROBE_FREQ == SAMPLE_FREQ) {
            /* keep them in sync */
            nextSampleTime = wakeupTime + approximately(SAMPLE_FREQ);
            nextProbeTime = nextSampleTime;
         } else {
            if (wakeupTime > nextSampleTime) {
               nextSampleTime = wakeupTime + approximately(SAMPLE_FREQ);
            }

            if (wakeupTime > nextProbeTime) {
               nextProbeTime = wakeupTime + approximately(PROBE_FREQ);
            }
         }

         try {
            myThread.sleep(Math.min(nextProbeTime, nextSampleTime) - 
                           wakeupTime);
         } catch (InterruptedException ignore) {}
      }
   }
}
