/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.ConductorExt.monitor.basic;

import java.net.*;
import java.util.*;

import FMG.Conductor.comm.*;
import FMG.Conductor.monitor.*;

class CommStats {

   private class StreamInfo {
      BasicMessageOutputStream stream;
      long bytesSoFar;

      private StreamInfo(BasicMessageOutputStream stream) {
         bytesSoFar = 0;
         this.stream = stream;
      }
   }

   Hashtable streamInfoTable = new Hashtable();
   NodeID nextNode = new NodeID();
   long lastLocalUpdate;
   Hashtable datapathParamTable = new Hashtable();

   protected int probeCount;

   long clockSkew;
   boolean isCongested;
   long minRTT = Long.MAX_VALUE;
   double maxBW = 0;

   double recentBW;
   long recentForwardDelay;

   CommStats(NodeID nextNode) {
      this.nextNode.copy(nextNode);
      lastLocalUpdate = System.currentTimeMillis();
   }

   void addStream(BasicMessageOutputStream newStream) {
      streamInfoTable.put(newStream, new StreamInfo(newStream));
   }

   void removeStream(BasicMessageOutputStream stream) {
      streamInfoTable.remove(stream);
   }

   void addDatapath(ConnID connID, LinkParams request) {
      datapathParamTable.put(connID, request);
   }

   void removeDatapath(ConnID connID) {
System.out.println("***** Removing connection: " + connID);
      datapathParamTable.remove(connID);
   }

   void predictLocalCharacteristics(boolean withNewConnection) {

      System.out.println("Predictions for connections through " + 
                         nextNode + "(" + System.currentTimeMillis() + "):" + 
                         (withNewConnection ? 
                             " (including new connection)" : ""));

      System.out.println("  Link is " + (isCongested ? "" : "NOT ") + 
                         "congested");

      System.out.println("  Expected latency : " + recentForwardDelay + " ms");

      int numConnections = datapathParamTable.size() +  
                                             (withNewConnection ? 1 : 0);
      double expectedBW = (isCongested ? recentBW : maxBW);
      double averageBW = expectedBW / numConnections;

      double excessBW = 0;
      int numBelowAverage = 0;
      Enumeration e = datapathParamTable.elements();
      while (e.hasMoreElements()) {
         LinkParams lp = (LinkParams) e.nextElement();
         if (lp.bandwidth < averageBW) {
            excessBW += averageBW - lp.bandwidth;
            numBelowAverage++;
         }
      }

      System.out.println("  Expected BW :");
      e = datapathParamTable.keys();
      while (e.hasMoreElements()) {
         ConnID connID = (ConnID) e.nextElement();
         LinkParams lp = (LinkParams) datapathParamTable.get(connID);
         if (lp.bandwidth < averageBW) {
            System.out.println("      " + connID + ": " + lp.bandwidth + 
                               " bytes/ms");
         } else {
            System.out.println("      " + connID + ": " + 
                               (averageBW + 
                                excessBW/(numConnections - numBelowAverage)) +
                               " bytes/ms");
         }
      }
   }

   void updateLocalInfo() {
      long currentTime = System.currentTimeMillis();
      long bytesWritten = 0;
      boolean recordableValue = true;

      // fetch number of bytes written from all streams
      Enumeration e = streamInfoTable.elements();
      while (e.hasMoreElements()) {
         StreamInfo sti = (StreamInfo) e.nextElement();
         long current = sti.stream.bytesSent();
         bytesWritten += current - sti.bytesSoFar;

         /* don't record max BW if any stream is starting up this cycle */
         recordableValue &= ((sti.bytesSoFar != 0) || (current == 0));

         sti.bytesSoFar = current;
      }

      double measuredBW = (double) bytesWritten / 
                                    (currentTime - lastLocalUpdate);

      if ((measuredBW > maxBW) && (recordableValue)) {
         maxBW = measuredBW;
      }

      recentBW = (recentBW/2) + (measuredBW/2);

//      System.out.println("Monitor: BW used to " + nextNode + " was " + 
//                         measuredBW + " (" + maxBW + ")");

      lastLocalUpdate = currentTime;
   }

   /* rtt is measured from local time only
     forwardDelay includes clock skew between the two nodes
     interpacket time is measured with remote time only
    */
   void updateWithProbeResults(long rtt, long forwardDelay, 
                               long interPacketTime) {
      System.out.println("Monitor: got probe results for " + nextNode + ":");
      System.out.println("      delay: " + rtt + 
                         ", forwardDelay: " + forwardDelay +
                         ", interpacket time: " + interPacketTime);

      if (rtt < minRTT) {
         minRTT = rtt;
      }

      if (isNear(rtt, minRTT)) {
         isCongested = false;
         /* assumes that latency is normally symmetric */
         clockSkew = (rtt/2) - forwardDelay;
//System.out.println("Recorded new clock Skew as: " + clockSkew);
      }

      long currentForwardDelay = forwardDelay + clockSkew;

      recentForwardDelay = (recentForwardDelay / 2) + 
                           (currentForwardDelay / 2);

//System.out.println("CurrentForwardDelay = " + currentForwardDelay);
//System.out.println("minRTT = " + minRTT);
      if (isMuchGreater(currentForwardDelay, minRTT) || 
          isMuchGreater(interPacketTime, minRTT)) {
          isCongested = true;
      }
   }

   private boolean isNear(long num, long target) {
      return ((num < target * 2) && (num > target / 2));
   }

   private boolean isMuchGreater(long num, long target) {
      return (num > (5 * target));
   }

   int probeCount() {
      return probeCount;
   }

   void incProbeCount() {
      probeCount++;
   }

   InetAddress nodeAddress() {
      return nextNode.getAddress();
   }

}
