/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.ConductorExt.monitor.basic;

import java.io.*;
import java.net.*;
import java.util.Hashtable;

import FMG.Conductor.comm.NodeID;
import FMG.Conductor.Conductor;
import FMG.Conductor.util.Debug;

class ProbeHandler extends Thread {

   private class FirstPacketInfo {
      long sendTime;
      long receiveTime;
      int seqNum;

      FirstPacketInfo(int seqNum, long sendTime, long receiveTime) {
         this.sendTime = sendTime;
         this.receiveTime = receiveTime;
         this.seqNum = seqNum;
      }
   }

   Hashtable firstPacketHistory = new Hashtable();
   BasicMonitor monitor;

   NodeID thisNode = NodeID.getNewLocalID();

   private DatagramSocket s;
   private ByteArrayOutputStream baos = new ByteArrayOutputStream();
   private DataOutputStream dos = new DataOutputStream(baos);

   static final int MONITOR_PORT = 
                Conductor.getIntProperty("BasicMonitor.MonitorPort", 4447);
   static final int MAX_PACKET_LEN = 200;

   static ProbeHandler createAndStart(BasicMonitor monitor) {
      ProbeHandler ph = new ProbeHandler(monitor);
      ph.start();
      return ph;
   }

   protected ProbeHandler(BasicMonitor monitor) {
      this.monitor = monitor;

      try {
         s = new DatagramSocket(MONITOR_PORT);
      } catch (IOException e) {
         throw new RuntimeException("Monitor couldn't get port");
      }
   }

   protected void handleFirstProbe(DataInputStream dis, long receiveTime) {
      try {
         NodeID sourceNode = new NodeID();
         sourceNode.readSelf(dis);
         int probeId = dis.readInt();
         long sentTime = dis.readLong();

// System.out.println("Got first probe from " + sourceNode);

/* TODO: if it wasn't for me, I should forward it */
         firstPacketHistory.put(sourceNode, 
                                new FirstPacketInfo(probeId, 
                                                    sentTime, receiveTime));
      } catch (IOException e) {
         if (Debug.on) {
            Debug.println(Debug.WARNING, 
                          "Warning: failed reading probe packet");
         }
      }
   }

   protected void handleSecondProbe(DataInputStream dis, long receiveTime) {
      try {
         NodeID sourceNode = new NodeID();
         sourceNode.readSelf(dis);
         int probeId = dis.readInt();
         long sentTime = dis.readLong();

// System.out.println("Got second probe from " + sourceNode);
/* TODO: if it wasn't for me, I should forward it */

         FirstPacketInfo savedInfo = 
                     (FirstPacketInfo) firstPacketHistory.remove(sourceNode);
         if ((savedInfo != null) && (probeId == savedInfo.seqNum)) {
            synchronized (this) {
               baos.reset();
               dos.write(3);                           /* response pkt */
               thisNode.writeSelf(dos);                /* local node id */
               dos.writeInt(probeId);                  /* the probe id */
               dos.writeLong(sentTime);                /* 2nd pkt send time */
               dos.writeLong(receiveTime);             /* 2nd pkt recv time */
               dos.writeLong(receiveTime - savedInfo.receiveTime);
                                                       /* inter-pkt recv time */

// System.out.println("Sending ack to " + sourceNode);

               byte[] buf = baos.toByteArray();
               s.send(new DatagramPacket(buf, buf.length, 
                                         sourceNode.getAddress(), 
                                         MONITOR_PORT));
            }
         }
      } catch (IOException e) {
         if (Debug.on) {
            Debug.println(Debug.WARNING, 
                          "Warning: failed reading probe packet");
         }
      }
   }

   protected void handleProbeAck(DataInputStream dis, long receiveTime) {
      try {
         NodeID sourceNode = new NodeID();
         sourceNode.readSelf(dis);
         int probeId = dis.readInt();
         long origSentTime = dis.readLong();
         long remoteReceiveTime = dis.readLong();
         long interPacketTime = dis.readLong();

// System.out.println("Got probe ack from " + sourceNode);
/* TODO: if it wasn't for me, I should forward it */

         //     give results to monitor
         monitor.handleProbeResults(sourceNode, probeId,
                                    receiveTime - origSentTime, 
                                    remoteReceiveTime - origSentTime,
                                    interPacketTime);
      } catch (IOException e) {
         if (Debug.on) {
            Debug.println(Debug.WARNING, 
                          "Warning: failed reading probe packet");
         }
      }
   }

   protected DatagramPacket createProbe(InetAddress targetAddress, 
                                        int probeId, int probeNum) {
      try {
         baos.reset();
         dos.write(probeNum);
         thisNode.writeSelf(dos);
         dos.writeInt(probeId);
         dos.writeLong(System.currentTimeMillis());
         byte[] buf = baos.toByteArray();
         return new DatagramPacket(buf, buf.length, targetAddress, 
                                   MONITOR_PORT);
      } catch (IOException e) {
         throw new RuntimeException("This shouldn't happen");
      }
   }

   public synchronized void sendProbe(InetAddress targetAddress, int probeId) {
      try {
         s.send(createProbe(targetAddress, probeId, 1));
         s.send(createProbe(targetAddress, probeId, 2));
      } catch (IOException e) {
         if (Debug.on) {
            Debug.println(Debug.WARNING, "Monitor: probe send failed.");
         }
      }
   }

   public void run() {
      byte[] buf = new byte[MAX_PACKET_LEN];
      ByteArrayInputStream bais = new ByteArrayInputStream(buf);
      DataInputStream dis = new DataInputStream(bais);
      DatagramPacket pack;

      while (true) {
         try {
            /* unfortunately, we can't re-use packets */
            pack = new DatagramPacket(buf, MAX_PACKET_LEN);

            bais.reset();
            s.receive(pack);

            long receiveTime = System.currentTimeMillis();

            switch (dis.read()) {
               case 1:
                  handleFirstProbe(dis, receiveTime);
                  break;
               case 2:
                  handleSecondProbe(dis, receiveTime);
                  break;
               case 3:
                  handleProbeAck(dis, receiveTime);
                  break;
            }
         } catch (IOException e) {
            if (Debug.on) {
               Debug.println(Debug.WARNING, 
                             "Warning: failure receiving probe packet");
            }
         }
      }
   }

}
