/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.ConductorExt.planners.template;

import FMG.Conductor.planning.messages.PlanDistro;
import FMG.Conductor.util.Direction;
import FMG.Conductor.util.Debug;

class EffectiveNetwork {

   protected int numNetworkNodes;
   protected int numTemplateNodes;

   protected int nodeToTemplateNode[];
   protected int templateNodeToNode[];
   protected int linkToTemplateLink[];

   protected NetworkModel network;
   protected Template template;

   EffectiveNetwork(NetworkModel net, Template t, int nodeAssignments[]) {
      numNetworkNodes = net.getNumNodes();
      numTemplateNodes = t.getNumNodes();

      /*
       * Build an array that gives us the corresponding template node for
       * each physical node.  If there is none, -1 is recorded.
       */
      nodeToTemplateNode = new int[numNetworkNodes];
      for (int i=0; i<numNetworkNodes; i++) {
         nodeToTemplateNode[i]=-1;
      }
      for (int i=0; i<numTemplateNodes; i++) {
         nodeToTemplateNode[nodeAssignments[i]] = i;
      }

      /*
       * Save the input array which tells us which physical node has been
       * assigned to each template node.
       */
      templateNodeToNode = new int[numTemplateNodes];
      for (int i=0; i<numTemplateNodes; i++) {
         templateNodeToNode[i] = nodeAssignments[i];
      }

      /*
       * Record the correspondance between physical links and template links.
       * The physical link number is the index and the template link number
       * is the recorded value.  -1 is recorded if the link is not between 
       * adapting nodes.
       *
       * Note that * the reverse mapping array doesn't exist because it 
       * would have a one-to-many relationship.
       */
      linkToTemplateLink = new int[numNetworkNodes-1];
      for (int i=0; i<numNetworkNodes-1; i++) {
         linkToTemplateLink[i] = -1;
      }
      for (int i=0; i<numTemplateNodes-1; i++) {
         for (int j=nodeAssignments[i]; j<nodeAssignments[i+1]; j++) {
            linkToTemplateLink[j]=i;
         }
      }

      network = net;
      template = t;
   }

   /*
    * Verify that the list of adaptors present contains all adaptors in the
    * list of required adaptors, taking into account the fact that items in 
    * the required list may include both an adaptor name and a parameter 
    * (separated by a space).
    */
   protected static boolean verifyRequiredAdaptors(ItemList present, 
                                                   ItemList required) {
      int numRequired = required.size();

      for (int i=0; i < numRequired; i++) {
         String adaptor = required.item(i);
         int endOfClassName = adaptor.indexOf(' ');
         if (endOfClassName != -1) {
            adaptor = adaptor.substring(0, endOfClassName);
         }

         if (! present.contains(adaptor)) {
            return false;
         }
      }

      return true;
   }

   protected boolean hasSufficientNodeResources() {
      // for each node in the template
      for (int i = 0; i<numTemplateNodes; i++) {

         if (TemplatePlanner.VERIFY_ADAPTORS) {
            // verify required adaptors are present on this node
            ItemList adaptors = 
                         network.getAvailableAdaptors(templateNodeToNode[i]);
            if (! (verifyRequiredAdaptors(adaptors, 
                                          template.getUpAdaptorsForNode(i)) &&
                   verifyRequiredAdaptors(adaptors,
                                          template.getDownAdaptorsForNode(i))
               )  ) {
                  return false;
            }
         }

         // verify this node has sufficient resources
         if (! template.getNodeRequirements(i).verify(
                    network.getNodeResources(templateNodeToNode[i]))) {
            return false;
         }
      }

      return true;
   }

   // verify that the effective link qualities match user requirements
   protected boolean meetsUserRequirements(Requirements requirements) {
      AttributeList[] linkQuals = new AttributeList[numNetworkNodes-1];
      for (int i=0; i<numNetworkNodes-1; i++) {
         linkQuals[i] = network.getLinkQualities(i);
      }

      AttrModList[] linkMods = new AttrModList[numNetworkNodes-1];
      for (int i=0; i<numNetworkNodes-1; i++) {
         if (linkToTemplateLink[i] != -1) {
            linkMods[i] = template.getLinkMods(linkToTemplateLink[i]);
         } else {
            linkMods[i] = null;
         }
      }

      return requirements.getNetRequirements().verify(linkQuals, linkMods);
   }

   // TODO: improve scoring to be less bogus and to be based on the resulting
   //       constraints (see generatePlan)
   int evaluate(Requirements requirements) {
      /*
       * If this employment of the template is not possible, return the
       * minimum score.
       */
      if (! hasSufficientNodeResources()) {
         if (Debug.on) {
            Debug.println(Debug.PLANNING, Debug.MOREINFO, 
                          this + " rejected due to insufficient " + 
                                 "node resources");
         }
         return Integer.MIN_VALUE;
      }

      if (! meetsUserRequirements(requirements)) {
         if (Debug.on) {
            Debug.println(Debug.PLANNING, Debug.MOREINFO, 
                          this + " rejected for not meeting user requirements");
         }
         return Integer.MIN_VALUE;
      }

      /*
       * Determine the score of this particular use of the template.
       */
      int score = Integer.MAX_VALUE/2;

      // lose 100 points for each adaptor
      for (int i=0; i<numNetworkNodes; i++) {
         if (nodeToTemplateNode[i]!=-1) {
            score -= (template.getUpAdaptorsForNode(
                                       nodeToTemplateNode[i]).size() + 
                      template.getDownAdaptorsForNode(
                                       nodeToTemplateNode[i]).size()) * 100;
         }
      }

      // lose (20 * ranking) points per data modification
      ItemList mods = template.getDataMods();
      for (int i=0; i<mods.size(); i++) {
         score -= requirements.getDataModRanking(mods.item(i))*20;
      }

      // lose 20 points per conductor node (plus end nodes)
      score -= 20 * numTemplateNodes;
      if (nodeToTemplateNode[0] == -1) {
         score -= 20;
      }
      if (nodeToTemplateNode[numNetworkNodes-1] == -1) {
         score -= 20;
      }

      // adjust per the template's self reported value
      score += template.value();

      return score;
   }

// TODO: use requirements parameter to generate plan constraints (ie, if 
//       bw becomes less than X, this plan is invalid ... replan)
   void generatePlan(PlanDistro plan, Requirements requirements) {
      for (int i=0; i<numTemplateNodes; i++) {
         plan.addNode(network.getNodeID(templateNodeToNode[i]), 
                      template.getUpAdaptorsForNode(i).getItems(), 
                      Direction.up);
         plan.insertNode(network.getNodeID(templateNodeToNode[i]),
                      template.getDownAdaptorsForNode(i).getItems(),
                      Direction.down);
      }
   }

   Template getTemplate() {
      return template;
   }

   public String toString() {
      StringBuffer sb = new StringBuffer(template.getName());
      sb.append("{");
      for (int i=0; i<numTemplateNodes; i++) {
         sb.append(templateNodeToNode[i]);
         if (i != numTemplateNodes-1) {
            sb.append(",");
         }
      }
      sb.append("}");

      return sb.toString();
   }
}
