/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.ConductorExt.planners.template;

import java.util.*;

import FMG.Conductor.comm.NodeID;
import FMG.Conductor.planning.messages.PlanningInfo;
import FMG.Conductor.util.Debug;

class NetworkModel {

   protected class NodeInfo {
      NodeID nodeID;
      protected AttributeList resources;
      protected ItemList adaptors;

      protected NodeInfo(NodeID nodeID, Hashtable info) {
         try {
            this.nodeID = nodeID.duplicate();
            resources = new AttributeList(info, "NodeProps.Resource.");
            adaptors = new ItemList(info, "NodeProps.Adaptor.");
         } catch (FieldFormatException e) {
            throw new RuntimeException("Case not handled");
// TODO: need a parameter in the AttributedList constructor
//       that instructs it to ignore any exceptions
         }
      }

      public String toString() {
         StringBuffer sb = new StringBuffer();
         sb.append('(');
         sb.append(nodeID.toString());
         sb.append(";resources:");
         sb.append(resources.toString());
         sb.append(";adaptors:");
         sb.append(adaptors.toString());
         sb.append(')');
         return sb.toString();
      }
   }

   protected class LinkInfo {
      protected AttributeList qualities;

      protected LinkInfo(Hashtable prevNodeInfo, Hashtable nextNodeInfo) {
         int prevPriority = 0;
         int nextPriority = 0;
         String priority = "";

         try {
            priority = (String) prevNodeInfo.get("NextLinkProps.Priority");
            if (priority != null) {
               prevPriority = Integer.parseInt(priority);
            }
         } catch (NumberFormatException ignore) {
            Debug.println(Debug.WARNING, 
                          "WARNING: improper number format '" + 
                          priority + "' in NextLinkProps.Priority.");
         }

         try {
            priority = (String) nextNodeInfo.get("PrevLinkProps.Priority");
            if (priority != null) {
               nextPriority = Integer.parseInt(priority);
            }
         } catch (NumberFormatException ignore) {
            Debug.println(Debug.WARNING, 
                          "WARNING: improper number format '" + 
                          priority + "' in PrevLinkProps.Priority.");
         }

         try {
            if (prevPriority > nextPriority) {
               qualities = new AttributeList(prevNodeInfo, 
                                             "NextLinkProps.Quality.");
            } else if (nextPriority > prevPriority) {
               qualities = new AttributeList(nextNodeInfo, 
                                             "PrevLinkProps.Quality.");
            } else {
               /*
                * Priorities were equal.  So, just find one.  Prefer 
                * client-side data over server-side data.
                */
               qualities = new AttributeList(prevNodeInfo, 
                                             "NextLinkProps.Quality.");

               if (qualities.size() == 0) {
                  qualities = new AttributeList(nextNodeInfo, 
                                                "PrevLinkProps.Quality.");
               }
            }
         } catch (FieldFormatException e) {
            throw new RuntimeException("Case not handled");
// TODO: need a parameter in the AttributedList constructor
//       that instructs it to ignore any exceptions
         }
      }

      public String toString() {
         StringBuffer sb = new StringBuffer();
         sb.append('(');
         sb.append(qualities.toString());
         sb.append(')');
         return sb.toString();
      }
   }

   protected int numNodes;

   protected NodeInfo[] nodes;
   protected LinkInfo[] links;

   NetworkModel(Vector infoList) {
      numNodes = infoList.size();

      PlanningInfo[] infoMessages = new PlanningInfo[numNodes];
      infoList.copyInto(infoMessages);

      nodes = new NodeInfo[numNodes];
      links = new LinkInfo[numNodes-1];

      for (int i=0; i<numNodes; i++) {
         nodes[i] = new NodeInfo(infoMessages[i].node, infoMessages[i].info);
         if (i < numNodes - 1) {
            links[i] = new LinkInfo(infoMessages[i].info, 
                                    infoMessages[i+1].info);
         }
      }
   }

   int getNumNodes() {
      return numNodes;
   }

   ItemList getAvailableAdaptors(int n) {
      return nodes[n].adaptors;
   }

   AttributeList getNodeResources(int n) {
      return nodes[n].resources;
   }

   AttributeList getLinkQualities(int n) {
      return links[n].qualities;
   }

   NodeID getNodeID(int n) {
      return nodes[n].nodeID;
   }

   Enumeration getEffectiveNetworks(Template t) {
      Vector results = new Vector();

      // Are there enough network nodes to apply this template?
      if (numNodes >= t.getNumNodes()) {
         // make all possible assignments
         assignNodes(t, results);
      }

      return results.elements();
   }

   private void assignNodes(Template t, Vector results) {
      int numTemplateNodes = t.getNumNodes();

      int a[] = new int[numTemplateNodes];  // current assignment vector

      assignANode(a, 0, numTemplateNodes, t, results);
   }

   /**
    * A recursive function that assigns all possible network nodes to
    * template positions.  For each possible combination, a new 
    * EffectiveNetwork object is created and added to the results vector.
    * @param a The current node assignment vector.
    * @param pos The position we're currently assigining.  When pos equals 
    *            numTemplateNodes, one complete assignment combination has 
    *            been reached.
    * @param numTemplateNodes The number of nodes specified in the template.
    * @param t The template.
    * @param results A vector in which to place the resulting EffectiveNetwork 
    *                objects
    * @returns No return value.
    */
   private void assignANode(int[] a, int pos, int numTemplateNodes, 
                            Template t, Vector results) {
      if (pos == numTemplateNodes) {
         results.addElement(new EffectiveNetwork(this, t, a));
      } else {
         for (a[pos]=(pos==0 ? 0 : a[pos-1] + 1);
              a[pos]<numNodes-(numTemplateNodes-(pos+1));
              a[pos]++) {
            assignANode(a, pos+1, numTemplateNodes, t, results);
         }
      }
   }

   public String toString() {
      StringBuffer sb = new StringBuffer("NetworkModel[nodes:");
      for (int i=0; i<numNodes; i++) {
         sb.append(nodes[i].toString());
         if (i < numNodes-1) {
            sb.append(",");
         }
      }
      sb.append(";links:");
      for (int i=0; i<numNodes-1; i++) {
         sb.append(links[i].toString());
         if (i < numNodes-2) {
            sb.append(",");
         }
      }
      sb.append(']');
      return sb.toString();
   }
}
