/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.ConductorExt.planners.template;

import java.io.*;
import java.util.*;

import FMG.Conductor.util.PropsLoader;

class Template {

   protected class NodeInfo {
      AttrCompList nodeRequirements;
      ItemList upAdaptors;
      ItemList downAdaptors;

      protected NodeInfo(Hashtable source, int nodeNum) 
                                               throws FieldFormatException {
         String nodePrefix = "Node." + nodeNum + ".";

         upAdaptors = new ItemList(source, nodePrefix + "UpAdaptor.");
         downAdaptors = new ItemList(source, nodePrefix + "DownAdaptor.");

         nodeRequirements = new AttrCompList(source,
                                             nodePrefix + "Requirement.");
      }

      public String toString() {
         StringBuffer sb = new StringBuffer();
         sb.append("(reqs:");
         sb.append(nodeRequirements.toString());
         sb.append(";upAdaptors:");
         sb.append(upAdaptors.toString());
         sb.append(";downAdaptors:");
         sb.append(downAdaptors.toString());
         sb.append(")");
         return sb.toString();
      }
   }

   protected class LinkInfo {
      AttrModList linkMods;

      protected LinkInfo(Hashtable source, int nodeNum) 
                                                throws FieldFormatException {
         linkMods = new AttrModList(source, "Link." + nodeNum + ".Effect.");
      }

      public String toString() {
         StringBuffer sb = new StringBuffer();
         sb.append("(mods:");
         sb.append(linkMods.toString());
         sb.append(")");
         return sb.toString();
      }
   }

   String name;

   ItemList dataMods;

   RangeList protocols;

   int numNodes;

   NodeInfo[] nodeList;
   LinkInfo[] linkList;

   int value;

   Template(File templateFile) throws BadTemplateException {
      name = templateFile.getName();

      Properties source = PropsLoader.readProps(templateFile, true);

      if (source == null) {
         throw new BadTemplateException("Couldn't load template " + 
                                        templateFile);
      }

      try {
         protocols = new RangeList(source, "Protocol.");
      } catch (FieldFormatException e) {
         throw new BadTemplateException("Bad protocol range list: " + e);
      }

      dataMods = new ItemList(source, "DataMod.");

      try {
         String tmp = source.getProperty("NumNodes");
         if (tmp == null) {
            throw new BadTemplateException("Can't find required property " + 
                                           "NumNodes");
         }
         numNodes = Integer.parseInt(source.getProperty("NumNodes"));
      } catch (NumberFormatException e) {
         throw new BadTemplateException("Could not decipher NumNodes");
      }

      try {
         String tmp = source.getProperty("Value");
         if (tmp == null) {
            value = 0;
         } else {
            value = Integer.parseInt(tmp);
         }
      } catch (NumberFormatException e) {
         throw new BadTemplateException("Could not decipher Value");
      }

      try {
         nodeList = new NodeInfo[numNodes];

         for (int i=0; i<numNodes; i++) {
            nodeList[i] = new NodeInfo(source, i);
         }

         if (numNodes > 0) {
            linkList = new LinkInfo[numNodes-1];
         } else {
            linkList = null;
         }

         for (int i=0; i<numNodes-1; i++) {
            linkList[i] = new LinkInfo(source, i);
         }
      } catch (FieldFormatException e) {
         throw new BadTemplateException(e.toString());
      }
   }

   boolean handlesProtocol(int prot) {
      return protocols.isInRange(prot);
   }

   ItemList getDataMods() {
      return dataMods;
   }

   AttrModList getLinkMods(int n) {
      return linkList[n].linkMods;
   }

   int getNumNodes() {
      return numNodes;
   }

   ItemList getUpAdaptorsForNode(int n) {
      return nodeList[n].upAdaptors;
   }

   ItemList getDownAdaptorsForNode(int n) {
      return nodeList[n].downAdaptors;
   }

   AttrCompList getNodeRequirements(int n) {
      return nodeList[n].nodeRequirements;
   }

   int value() {
      return value;
   }

   public String toString() {
      StringBuffer sb = new StringBuffer(name);
      sb.append("[prot:");
      sb.append(protocols.toString());
      sb.append(",mods:");
      sb.append(dataMods.toString());
      sb.append(",nodes:(");
      for (int i=0; i<numNodes; i++) {
          sb.append(i);
          sb.append(':');
          sb.append(nodeList[i].toString());
          if (i < numNodes-1) {
             sb.append(';');
          }
      }
      sb.append("),links:(");
      for (int i=0; i<numNodes-1; i++) {
          sb.append(i);
          sb.append(':');
          sb.append(linkList[i].toString());
          if (i < numNodes-2) {
             sb.append(';');
          }
      }
      sb.append(")]");
      return sb.toString();
   }

   String getName() {
      return name;
   }
}
