/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.ConductorExt.planners.template;

import java.util.*;
import java.io.File;
import java.io.IOException;

import FMG.Conductor.Conductor;
import FMG.Conductor.util.Debug;
import FMG.Conductor.util.Direction;
import FMG.Conductor.planning.Planner;
import FMG.Conductor.planning.messages.PlanDistro;
import FMG.Conductor.planning.messages.PlanningInfo;
import FMG.Conductor.comm.*;

/*
 * TODO:
 * (1) We should be able to assign more than one (adjacent) virtual node
 *     to the same physical node.  Note that the node would have to
 *     satisfy all the virtual node requirements.  Also, the link benefit
 *     between two of those virtual nodes would not be attained.
 * (2) We should handle upstream and downstream node assignments
 *     separately and allow a template to specify when a pair of adaptors
 *     must be colocated.  Note: It may be the case that if I implement
 *     #1 (above), I no longer need this (just colocate the things that
 *     must be, and don't colocated the ones that don't.
 * (3) Should be able to describe upstream and downstream link characteristcs
 *     independently (when they are) or specify them as a shared resource
 *     (when they are).
 *       One solution is to allow optional specification of link resources:
 *           bw=...
 *             or
 *           bw:up=...
 *           bw:down=...
 *       Link mods would always be component
 *           bw:down/=2   (but perhaps up is unchanged)
 *       User mods could be either way.
 */

public class TemplatePlanner extends Planner {

   protected int protocol;
   protected NodeID clientNode;
   protected NodeID serverNode;

   protected Vector infoList = new Vector();

   protected static Template[] availableTemplates;

   protected static final String TEMPLATE_DIRECTORY_PROP = 
                                      "TemplatePlanner.TemplateDir";

   protected static final String DEFAULT_TEMPLATE_DIR = "./config/templates";

   protected static final String TEMPLATE_FILE_EXTENSION = ".t";

   static final boolean VERIFY_ADAPTORS = 
         Conductor.getBooleanProperty("TemplatePlanner.VerifyAdaptors", true);

   static {
      Vector templates = new Vector();
      File templateDir = new File(Conductor.getFilenameProperty(
                                               TEMPLATE_DIRECTORY_PROP, 
                                               DEFAULT_TEMPLATE_DIR));
      String[] files = templateDir.list();

      for (int i=0; i<files.length; i++) {
         try {
            if (files[i].endsWith(TEMPLATE_FILE_EXTENSION)) {
               File f = new File(templateDir, files[i]);

               Template newTemplate = new Template(f);
               templates.addElement(newTemplate);

               if (Debug.on) {
                  Debug.println(Debug.PLANNING, Debug.MOREINFO, 
                                "TemplatePlanner: Loaded " + newTemplate);
               }
            }
         } catch (BadTemplateException e) {
            if (Debug.on) {
               Debug.println(Debug.PLANNING, Debug.WARNING, 
                             "Failed to load template '" + files[i] + 
                                  "': " + e);
            }
         }
      }

      if (templates.size() == 0) {
         throw new RuntimeException("TemplatePlanner: no templates loaded!");
      }

      availableTemplates = new Template[templates.size()];
      templates.copyInto(availableTemplates);
   }

   public TemplatePlanner(ConnID id, NodeID sourceAddr, int sourcePort, 
                                     NodeID destAddr, int destPort) {
      super(id);
      protocol = destPort;
      clientNode = sourceAddr.duplicate();
      serverNode = destAddr.duplicate();
   }

   public void addInfo(PlanningInfo pi) {
      infoList.addElement(pi);
      pi.retain();             // don't let this message be recycled
   }

   public PlanDistro formulatePlan() {
      if (! sortPlanningInfo(infoList)) {
         throw new RuntimeException("This shouldn't happen");
      }

      Requirements requirements = 
            new Requirements(((PlanningInfo) infoList.firstElement()).info);

      NetworkModel network = new NetworkModel(infoList);

      if (Debug.on) {
         Debug.println(Debug.PLANNING, Debug.MOREINFO, 
                       "TemplatePlanner: Planning for network " + network);
         Debug.println(Debug.PLANNING, Debug.MOREINFO, 
                       "TemplatePlanner: Requirements are " + requirements);
      }

      PlanDistro plan = findBestPlan(requirements, network);

      /**
       * Post process: insert real addresses for nodes whose node id is not
       * the same as the interface to use to contact them.
       */
      setUpstreamAddresses(plan, infoList);

      for (int i=0; i<infoList.size(); i++) {
         ((PlanningInfo) infoList.elementAt(i)).recycle();
      }
      infoList.removeAllElements();

      if (Debug.on && (Debug.getLevel(Debug.PLANNING) > Debug.MOREINFO)) {
         Debug.println(Debug.PLANNING, Debug.MOREINFO, 
                       "TemplatePlanner: returning plan: " + plan);
      }

      return plan;
   }

// TODO: cache the last EffectiveNetwork for each host/protocol.  First
//       try to reuse the EffNet, then try to reuse the template, then
//       plan from scratch
   protected PlanDistro findBestPlan(Requirements requirements, 
                                     NetworkModel network) {
      EffectiveNetwork bestEffNet = null;
      int bestScore = Integer.MIN_VALUE;

      for (int i=0; i<availableTemplates.length; i++) {
         Template t = availableTemplates[i];

         if (Debug.on) {
            Debug.println(Debug.PLANNING, Debug.MOREINFO,
                          "TemplatePlanner: Evaluating template: " + 
                          t.getName());
         }

         // Are there enough nodes present to support this template?
         if (t.getNumNodes() > network.getNumNodes()) {
            if (Debug.on) {
               Debug.println(Debug.PLANNING, Debug.MOREINFO, 
                             "TemplatePlanner: template requires more nodes " +
                             "than are physically present.");
            }
            continue;
         }

         // Does the current protocol match?
         if (! t.handlesProtocol(protocol)) {
            if (Debug.on) {
               Debug.println(Debug.PLANNING, Debug.MOREINFO, 
                             "TemplatePlanner: template doesn't support " + 
                             "protocol " + protocol);
            }
            continue;
         }

         // Are there any unacceptable data mods in this template?
         if (! requirements.hasAcceptableDataMods(t)) {
            if (Debug.on) {
               Debug.println(Debug.PLANNING, Debug.MOREINFO,
                             "TemplatePlanner: template uses unacceptable " + 
                             "data mods protocol " + protocol);
            }
            continue;
         }

         // Generate a list of possible applicaitons of this template
         Enumeration effectiveNets = network.getEffectiveNetworks(t);

         // Evaluate them all and find the best one
         while (effectiveNets.hasMoreElements()) {
            EffectiveNetwork effNet = 
                      (EffectiveNetwork) effectiveNets.nextElement();

            int score = effNet.evaluate(requirements);

            if (Debug.on) {
                Debug.println(Debug.PLANNING, Debug.MOREINFO,
                              "TemplatePlanner: score for " + effNet + 
                              " was " + score);
            }

            if (score > bestScore) {
               bestEffNet = effNet;
               bestScore = score;
            }
         }
      }

      PlanDistro plan = PlanDistro.getNewMessage(connID);
      if (bestEffNet != null) {
         bestEffNet.generatePlan(plan, requirements);
      } else {
         // TODO: need some sort of default plan?
      }

      // if first node in upstream direction isn't the client, add it
      addEndpointIfNotPresent(plan, clientNode, Direction.down);
      // if first node in downstream direction isn't the client, add it
      addEndpointIfNotPresent(plan, serverNode, Direction.up);

      if (Debug.on && (Debug.getLevel(Debug.PLANNING) > Debug.MOREINFO)) {
         if (bestEffNet == null) {
            Debug.println("TemplatePlanner: Found no matching plans!!!");
         } else {
            Debug.println("TemplatePlanner: Selected plan from template: " + 
                          bestEffNet.getTemplate());
         }
      }

      return plan;
   }

   protected void addEndpointIfNotPresent(PlanDistro plan, NodeID nodeID, 
                                     Direction dir) {
      String[] emptyList = new String[0];

      Enumeration nodes = plan.getNodes(dir.opposite());

      if ((! nodes.hasMoreElements()) || 
          (! nodeID.equals(nodes.nextElement()))) {
          plan.addNode(nodeID, emptyList, dir);
          plan.insertNode(nodeID, emptyList, dir.opposite());
      }
   }
}
