/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.ConductorExt.planners.template.numbers;

public class ValueComp {
   protected Value value;
   protected CompFunc comparison;

   protected Class aggregatorClass;

   private static ValueComp dummy = new ValueComp();

   public abstract class CompFunc {
      abstract boolean test(boolean x1, boolean x2);
      abstract boolean test(long x1, long x2);
      abstract boolean test(double x1, double x2);
      public abstract String toString();
   }

   public static CompFunc CMP_EQ = dummy. new CompFunc() {
      boolean test(boolean x1, boolean x2) { return (x1 == x2); }
      boolean test(long x1, long x2) { return (x1 == x2); }
      boolean test(double x1, double x2) { return (x1 == x2); }
      public String toString() { return "=="; }
   };

   public static CompFunc CMP_NEQ = dummy. new CompFunc() {
      boolean test(boolean x1, boolean x2) { return (x1 != x2); };
      boolean test(long x1, long x2) { return (x1 != x2); }
      boolean test(double x1, double x2) { return (x1 != x2); }
      public String toString() { return "!="; }
   };

   public static CompFunc CMP_GT = dummy. new CompFunc() {
      boolean test(boolean x1, boolean x2) { return x2; }
      boolean test(long x1, long x2) { return (x2 > x1); }
      boolean test(double x1, double x2) { return (x2 > x1); }
      public String toString() { return ">"; }
   };

   public static CompFunc CMP_GE = dummy. new CompFunc() {
      boolean test(boolean x1, boolean x2) { return (x2 | !x1); }
      boolean test(long x1, long x2) { return (x2 >= x1); }
      boolean test(double x1, double x2) { return (x2 >= x1); }
      public String toString() { return ">="; }
   };

   public static CompFunc CMP_LT = dummy. new CompFunc() {
      boolean test(boolean x1, boolean x2) { return x2; }
      boolean test(long x1, long x2) { return (x2 < x1); }
      boolean test(double x1, double x2) { return (x2 < x1); }
      public String toString() { return "<"; }
   };

   public static CompFunc CMP_LE = dummy. new CompFunc() {
      boolean test(boolean x1, boolean x2) { return (x1 | !x2); }
      boolean test(long x1, long x2) { return (x2 <= x1); }
      boolean test(double x1, double x2) { return (x2 <= x1); }
      public String toString() { return "<="; }
   };

   private ValueComp() { }

   public ValueComp(Value value, CompFunc comparison) {
      this(value, comparison, null);
   }

   public ValueComp(Value value, CompFunc comparison, Class aggregatorClass) {
      this.value = value;
      this.comparison = comparison;
      this.aggregatorClass = aggregatorClass;
   }

   public boolean test(Value target) {
      return test(target, null);
   }

   public boolean test(Value[] targets, ValueMod[] mods) {
      if (aggregatorClass == null) {
         throw new RuntimeException("This shouldn't happen");
      }

      try {
         AggregateValue aggValue = (AggregateValue) 
                                          aggregatorClass.newInstance();

         for (int i=0; i<targets.length; i++) {
            if (targets[i] == null) {
               return false;
            }
            aggValue.includeValue(targets[i], mods[i]);
         }

         return test(aggValue);

      } catch (InstantiationException e) {
         throw new RuntimeException("This shouldn't happen");
      } catch (IllegalAccessException e) {
         throw new RuntimeException("This shouldn't happen");
      }
   }

   public boolean test(Value target, ValueMod mod) {
      if (target == null) {
         return false;
      }

      int type1 = value.dataType();
      int type2;
      if (mod != null) {
         type2 = mod.resultType(target);
      } else {
         type2 = target.dataType();
      }
// Should I type based on the comparison value type?
      int compareType = (type1 > type2 ? type1 : type2);

      switch (compareType) {
         case Value.TYPE_BOOL:
            if (mod != null) {
               return comparison.test(value.booleanValue(), 
                                      mod.booleanResult(target));
            } else {
               return comparison.test(value.booleanValue(), 
                                      target.booleanValue());
            }
         case Value.TYPE_LONG:
            if (mod != null) {
               return comparison.test(value.longValue(), 
                                      mod.longResult(target));
            } else {
               return comparison.test(value.longValue(), target.longValue());
            }
         case Value.TYPE_DOUBLE:
            if (mod != null) {
               return comparison.test(value.doubleValue(), 
                                      mod.doubleResult(target));
            } else {
               return comparison.test(value.doubleValue(), 
                                      target.doubleValue());
            }
         default:
            throw new RuntimeException("This should't happen");
      }
   }

   public String toString() {
      StringBuffer sb = new StringBuffer();
      sb.append(comparison.toString());
      sb.append(value.toString());
      if (aggregatorClass != null) {
         sb.append('{');
         sb.append(aggregatorClass.toString());
         sb.append('}');
      }
      return sb.toString();
   }

}
