/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.ConductorExt.security;

import java.util.*;
import java.net.*;
import java.io.*;
import java.security.*;
import xjava.security.Cipher;

import FMG.Conductor.comm.*;
import FMG.Conductor.util.*;
import FMG.Conductor.security.*;
import FMG.Conductor.planning.messages.*;

import FMG.ConductorExt.security.util.*;
import FMG.ConductorExt.security.messages.*;

// be specific here, so there's no confusion with java.net.Authenticator
import FMG.ConductorExt.security.messages.Authenticator;

import FMG.Conductor.Conductor;

public abstract class ExtensionSecurityBox extends SecurityBox {

    /** The default name for the public key algorithm */
    private static final String  DEFAULT_PKEY_ALGO   = "RSA";

    /** The default name for the signing algorithm */
    private static final String  DEFAULT_SIGN_ALGO   = "SHA-1/RSA/PKCS#1";

    /** The default name for the cipher  algorithm */
    private static final String  DEFAULT_CRYP_ALGO   = "RSA/ECB/OneAndZeroes";

    /** The object describing the characteristics of this security box  */
    protected BoxCharacteristics boxCharacteristics;

    /** The specified properties propagated from the client */
    private        Properties    params = new Properties();

    /* The message buffer */
    private        MessageBuffer msgBuf = new MessageBuffer(this);


    static {
        registerMessages();
    }

    /*
     * Abstract Methods
     */

    public abstract  void authentication(Authenticator auth, boolean isPlanner);
    public abstract  void authenticationDone(boolean Planner);
    public abstract  PublicKey getVerifiedPublicKey(NodeID node);
    public abstract  Authenticator createNewAuthenticator(ConnID connID,
							   Direction dir,
							   String pkeyAlgo,
							   String signAlgo);
    public abstract  String getProperty(String propName);
    public abstract  String configDirName();

    /*
     * Constructor
     */
    protected ExtensionSecurityBox(SecurityCoordinator coordinator, String param) {
	super(coordinator, param);
        setParameters(param);
    }

    /*
     * Public functions
     */

    /**
     * Determines the public algorithm name used by the security box.
     * If not specified by the client, nor by the property file, the
     * default will be returned.
     *
     * @return  The public algorithm name used by the security box.
     */
    public String     getPubKeyAlgo() {
        return getParameter("PublicKeyAlgo", DEFAULT_PKEY_ALGO);
    }
    
    /**
     * Determines the signing algorithm name used by the security box.
     * If not specified by the client, nor by the property file, the
     * default will be returned.
     *
     * @return  The signing algorithm name used by the security box.
     */
    public String     getSignAlgo() {
        return getParameter("SignAlgo", DEFAULT_SIGN_ALGO);
    }

    /**
     * Determines the cipher algorithm name used by the security box.
     * If not specified by the client, nor by the property file, the
     * default will be returned.
     *
     * @return  The cipher algorithm name used by the security box.
     */
    public String     getCipherAlgo() {
        return getParameter("CipherAlgo", DEFAULT_CRYP_ALGO);
    }

    /**
     * Determines the private key of the node itself.
     * The private key is in terms of the public key algorithm used hereof.
     *
     * @return  The local private key
     */
    public PrivateKey getOwnPrivateKey() {
        return (PrivateKey)getKeyProperty("OwnPrivateKey");
    }

    /**
     * Determines the public key of the certificate server, following the
     * public key algorithm used hereof 
     *
     * @return the public key of the certificate server
     */
    public PublicKey  getCertServerPublicKey() {
        return (PublicKey)getKeyProperty("CertServerPublicKey");
    }

    /**
     * Determines the target node
     *
     * @return the target node of the current connection
     */
    public NodeID targetNode() {
	return coordinator.targetNode();
    }

    /**
     * Handle messages that come in from the network.
     * <p>
     * There are three cases to handle with:
     * <ul>
     *    <li>  this security box on a non-planner node and the message is
     *          towards the server
     *    <li>  this security box on a non-planner node and the message is
     *          towards the client
     *    <li>  this security box on the planner node
     *</ul>
     *
     * Note: We assume from each node all security related messages
     *       follows the PlanningInfo or the PlanDistro message. 
     *
     * @see handleOutgoingToServer and handleOutgoingToClient.
     * @param m The message that has been received from the network.
     * @return No return value.
     */
    public void handleIncoming(PlanningMessage m) {
        if (Debug.on) {
            Debug.println(Debug.SECURITY, Debug.MOREINFO,
                          "[secbox] handleIncoming(" + m + ") ");
        }

	if ( !m.connID.equals(coordinator.connectionID()) ) {
	    System.out.println("Replay attack with message: "+m);
            m.recycle();
	}

	msgBuf.setcurrentNode(m);

	/* Authentication Work */

	if ( m instanceof Authenticator )
	    authentication((Authenticator)m, coordinator.isPlannerNode());
	if ( m instanceof PlanningDone )
	    authenticationDone(coordinator.isPlannerNode());

	/* Message delivery to Path or buffering or recycling */

        if ( ! coordinator.isPlannerNode() && m.flowDir().isUp()) {
	    // non-planner node,  m is towards the server
            if ( m.handledByPath() )
                coordinator.deliverToPath(m);  // deliver to Path objet
	    else
		msgBuf.put(m);
	} else if ( coordinator.isPlannerNode() ) {
	    if ( m instanceof PlanningInfo ) {
		msgBuf.put(m);
	    } else if ( m instanceof Authenticator ) {
		m.recycle();
	    } else if ( m instanceof SignInfo ) {
		handleIncomingSignInfo((SignInfo)m);
	    } else if ( m instanceof PlanningDone ) {
		Enumeration msgs = msgBuf.releaseAllVerifiedMsgsForPath();
		while ( msgs != null && msgs.hasMoreElements() )
		    coordinator.deliverToPath( (PlanningMessage)msgs.nextElement() );    
		coordinator.deliverToPath(m);
	    } else {
		if ( ! m.handledByPath() )
		    throw new RuntimeException("The message should be handled by Path");
		System.out.println("Warning: no knowledge of this message before");
		coordinator.deliverToPath(m);
	    }
	} else {
	    // ! coordinator.isPlannerNode() && m.flowDir().isDown()

	    if ( m instanceof PlanDistro || m instanceof KeyDistro ) {
		msgBuf.put(m);
	    } else if ( m instanceof Authenticator ) {
		if ( coordinator.isRootNode() )
		    m.recycle();
		else
		    msgBuf.put(m);
	    } else if ( m instanceof SignInfo ) {
		handleIncomingSignInfo((SignInfo)m);
	    } else  {
		if ( ! m.handledByPath() )
		    throw new RuntimeException("The message should be handled by Path");
		coordinator.deliverToPath(m);
		if ( ! ( m instanceof PlanPathInit ) )
		    System.out.println("Warning: no knowledge of this message before");
	    }
	}
    }

    private void handleIncomingSignInfo(SignInfo m) {

	if ( m instanceof PlanningInfoSignInfo ) {
	    /* This only happens on planner */

	    PlanningInfo pi = (PlanningInfo)msgBuf.releaseVerifiedMsg((SignInfo)m);
	    if ( pi != null ) {
		coordinator.deliverToPath(pi);
		m.recycle();
	    } else {
		msgBuf.put(m);
	    }

	    return;
	} 

	/* 
	 * The following happens on a non-planner conductor node where the
	 * SignInfo message is flowing towards the client.
	 */

	PublicKey plannerPubKey = getVerifiedPublicKey(coordinator.targetNode());
	if ( plannerPubKey == null ) {
	    System.out.println("Warning: The public key of the planner can not be verified");
	    return;
	}
	if ( m instanceof PlanDistroSignInfo ) {
	    PlanDistro pd =  (PlanDistro)msgBuf.releaseMsg((SignInfo)m);
	    
	    if ( pd == null ) {
		throw new RuntimeException("No PlanDistro released from buffer");
	    }
	    
	    if ( ((SignInfo)m).verify(pd, getSignAlgo(), plannerPubKey)) {
		coordinator.deliverToPath(pd);
	    } else {
		System.out.println("Warning: the plan is not signed correctly");
		System.out.println("Warning: plan distribution failed");
		pd.recycle();
		m.recycle();
		return;
	    }
	    
	    if ( coordinator.isRootNode())
		m.recycle();
	    else
		msgBuf.put(m);
	    
	} else if ( m instanceof KeyDistroSignInfo ) {
	    KeyDistro kd = (KeyDistro)msgBuf.releaseMsg((SignInfo)m);
	    
	    if ( kd != null ) {
		// This kd is for this node
		if ( ((SignInfo)m).verify(kd, getSignAlgo(), plannerPubKey)) {
		    getStreamIAC().store(kd.getKeyLabel(),
					 kd.getKey(getCipherAlgo(), getOwnPrivateKey()));
		} else {
		    System.out.println("Warning: keydistro for this node not signed correctly");
		}
		kd.recycle();
		m.recycle();
	    } else {
		// This m signs for other node's KeyDistro
		msgBuf.put(m);
	    }
	    
	} else if ( m instanceof BoxCharacteristicsSignInfo ) {
	    if ( ((SignInfo)m).verify(boxCharacteristics, getSignAlgo(), plannerPubKey) ) {
		if (  ! coordinator.isRootNode() ) {
		    msgBuf.put(m);
		} else {
		    m.recycle();
		}
	    } else {
		System.out.println("Warning: Box characteristics for this node not signed correctly");
		m.recycle();
	    }
	}
	
    }

    public boolean handleOutgoingToServer(PlanningMessage m) {
        if (Debug.on) {
            Debug.println(Debug.SECURITY, Debug.MOREINFO,
                          "[secbox] handleOutgoingToServer(" + m + ")");
        }
	if ( m instanceof PlanningInfo ) {
	    PlanningInfo         pi   = (PlanningInfo)m;
	    PlanningInfoSignInfo si   = null;
	    Authenticator        auth = null;

	    if ( ! coordinator.sendSToNet(m) ) {
		return false;
	    }

	    if (  pi.node.isLocalNode() ) {
		auth  = createNewAuthenticator(coordinator.connectionID(),
					       Direction.up,
					       getPubKeyAlgo(),
					       getSignAlgo());
		si    = (PlanningInfoSignInfo)getNewSignInfo(pi, Direction.up);

		/*
		 * PlanningInfoSignInfo is the last of the whole family messages
		 * associated with PlanningInfo.
		 */
		if ( ! coordinator.sendSToNet(auth) || ! coordinator.sendSToNet(si) )
		    return false;
	    } else {
		Enumeration msgs = msgBuf.releaseMsgs(pi.node);
		//TODO: wait until both auth and si arrive.
		while ( msgs.hasMoreElements() ) {
		    PlanningMessage pm = (PlanningMessage)msgs.nextElement();
		    if (  ! coordinator.sendSToNet(pm) )
			return false;
		}
	    }
	} else {
	    if ( ! coordinator.sendSToNet(m) ) {
		return false;
	    }
	}

	return true;
    }

    public boolean handleOutgoingToClient(PlanningMessage m) {
        if (Debug.on) {
            Debug.println(Debug.SECURITY, Debug.MOREINFO,
                          "[secbox] handleOutgoingToClient(" + m + ")");
        }

	if (  m instanceof PlanDistro ) {
	    PlanDistro                 pd = (PlanDistro)m;

	    Authenticator auth = null;

	    PlanDistroSignInfo         pdsig  = null;
	    BoxCharacteristicsSignInfo boxsig = null;

            if ( !coordinator.sendCToNet(m) ) {
                return false;
            }

	    if (  coordinator.isPlannerNode() ) {
		auth   = createNewAuthenticator(coordinator.connectionID(),
						Direction.down,
						getPubKeyAlgo(),
						getSignAlgo());
		boxsig = (BoxCharacteristicsSignInfo)getNewSignInfo(boxCharacteristics, 
								    Direction.down);
		pdsig  = (PlanDistroSignInfo)getNewSignInfo(pd, Direction.down);

		if (  !coordinator.sendCToNet(auth)  )
		    return false;

		if (  !coordinator.sendCToNet(boxsig) )
		    return false;

		if ( ! distributeKeys(getRequiredSessionKeys(pd)) )
		    return false;

		/*
		 * PlanDistroSignInfo is the last of the whole family messages
		 * associated with Plandistro.
		 */
		if (  !coordinator.sendCToNet(pdsig)  )
		    return false;
	    } else {
		Enumeration msgs = msgBuf.releaseMsgs(coordinator.targetNode());
		while ( msgs.hasMoreElements() ) {
		    PlanningMessage pm = (PlanningMessage)msgs.nextElement();
		    if (  ! coordinator.sendCToNet(pm) )
			return false;
		}
	    }
	} else {
	    // Not a PlanDistro message
	    if ( !  coordinator.sendCToNet(m) )
		return false;
	}

	return true;
    }

    /*
     * Static Functions
     */

    /**
     * Registers security messages.
     * The following three security messages needs to be registered:
     * <ul>
     *    <li> <code>PlanningInfoSignInfo</code>
     *    <li> <code>PlanDistroSignInfo</code>
     *    <li> <code>KeyDistroSignInfo</code>
     *    <li> <code>BoxCharacteristicsSignInfo</code>
     *    <li> <code>KeyDistro</code>
     *    <li> <code>BoxCharacteristics</code>
     * </ul>
     *
     * @return No return value.
     */
    private static void registerMessages() {
        PlanningInfoSignInfo.registerSelf();
        PlanDistroSignInfo.registerSelf();
        KeyDistroSignInfo.registerSelf();
        BoxCharacteristicsSignInfo.registerSelf();
        KeyDistro.registerSelf();
        BoxCharacteristics.registerSelf();
    }

    /**
     * Reads the property file.
     *
     * @param filename  The name of the property file
     *
     * @return  The Properties variable initialized by reading the file
     */
    protected static Properties readPropertiesFile(String filename) {
	Properties      props = new Properties();

        /* if the user specified a null filename, don't read any properties */
        if (filename.length() == 0) {
            return null;
      }
        
        /* read the properties file */
        try {
            File propFile = new File(filename);
            FileInputStream fis = new FileInputStream(propFile);
            props.load(fis);
	    return props;
        } catch (IOException e) { 
            // Verbosity is not set yet, so just go ahead and warn the user
            System.err.println("Warning: error reading properties file (" + 
                               e + ")");
	    return null;
        }
    }

    /* Reads the name of a host from an input stream.
     * Each line contains the name of one host. And only one line of the
     * stream is read.
     *
     * @param is the input stream
     * @return a String of the first line from the stream
     */
    protected static String readHostname(InputStream is) throws IOException {
        byte[] buf = new byte[100];
        int len = 0;
        int b;
        
        b = is.read();
        
        if (b == -1) {
            return null; 
        }
        
        while ((b!='\n') && (b != -1)) { 
            buf[len++]=(byte)b;
            b = is.read();
        }  
        
        return new String(buf,0,len);
    }  


    /*
     * Private Functions
     */

    private boolean distributeKeys(Enumeration sessionKeys) {
	SessionKey            sk = null;
	Enumeration        nodes = null;
	NodeID              node = null;
	KeyDistro             kd = null;
	KeyDistroSignInfo  kdsig = null;

	while ( sessionKeys.hasMoreElements() ) {
	    sk = (SessionKey)sessionKeys.nextElement();
	    nodes = sk.getNodes();
	    while ( nodes.hasMoreElements() ) {
		node = (NodeID)nodes.nextElement();
		if ( node.isLocalNode() ) {
		    // This is the planner node itself
		    getStreamIAC().store(sk.getKeyLabel(), sk.getKey());
		} else {
		    kd = getKeyDistro(node,
				      sk.getKeyLabel(),
				      sk.getKey(),
				      coordinator.connectionID());
		    if ( kd != null ) {
			kdsig = (KeyDistroSignInfo)getNewSignInfo(kd, Direction.down);
			if ( !coordinator.sendCToNet(kd) || !coordinator.sendCToNet(kdsig))
			    return false;
                    } else {
                        System.out.println("Warning: no key distribution message formed for node "+node);
                    }
                }
            }
	}

	return true;
    }

    /**
     * Sets the <code>params</code> variable.
     * The format of paramString is "A=a B=b C=c D", where A, B, C, D, etc
     * are names of the parameters and a, b, c, "true", etc, are values,
     * respectively.
     *
     * @return No return value
     */

    private void setParameters(String paramString) {
        int from  = 0;
        int index = paramString.indexOf(' ');
        
        while ( index != -1 ) {
            updateParameter(paramString.substring(from, index));
            from  = index +1;
            index = paramString.indexOf(' ', from);
        }
        updateParameter(paramString.substring(from));
    }
    
    /**
     * Updates the value of a property using the given string.  
     * The string should either be the name of a property or in the form
     * "propname=value". If only the name of a property is given (ie, no
     * equals sign), then the value is set to "true".
     *
     * @param propString A string containing a property name or both a 
     *                   property name and value.
     * @return No return value
     */
    private void updateParameter(String propString) {
        int index = propString.indexOf('=');
        
        if (index == -1) {
            params.put(propString, "true");
        } else {
            String property = propString.substring(0, index);
            String value    = propString.substring(index + 1, propString.length());
            params.put(property, value);
        }
    }


    /**
     * Gets the string value of the property
     * <p>
     * It checks the <code>params</code> first to see if a value can be
     * returned; otherwise check with getProperty(). The default will be
     * returned if nothing can be returned from both. This way the
     * property with same name from <code>params</code> overwrites
     * what can be got from getProperty().
     * <p> 
     * getProperty() returns class level properties read from local
     * property file, while the <code>params</code> contains instance level
     * properties propagated from client. 
     */
    private String getParameter(String paraName, String defaultValue) {
        String value = params.getProperty(paraName);
        if ( value == null )
            value = getProperty(paraName);
        if ( value == null ) 
            value = defaultValue;

        return value;
    }

    /**
     * Gets the private key of its own or the public key of the
     * certificate server, in terms of the public key type.
     * <p> 
     * It will first determine the name of corresponding property,
     * such as <code>RSAOwnPrivateKey</code> or
     * <code>RSACertServerPublicKey</code>, then from the property value
     * the filename stores the key can be determined.
     *   
     * @param  whosewhat "OwnPrivateKey" or "CertServerPublicKey"
     * @return the key as requested.
     */
    protected Key getKeyProperty(String whosewhat) {
        String keyType = getPubKeyAlgo();
        String propName = new String(keyType + whosewhat);
        String keyFileName = configDirName() + "/" + getProperty(propName);

        Key    ret = null ;
        if ( keyFileName == null ) {
            return null;
        } 

        try {
            FileInputStream fis = new FileInputStream(new File(keyFileName));
            if ( whosewhat.indexOf("PrivateKey") != -1 )
                ret = KeySerializer.readPrivKey(fis, keyType); 
            else if ( whosewhat.indexOf("PublicKey") != -1 )
                ret = KeySerializer.readPubKey(fis, keyType);
            else 
		throw new RuntimeException(whosewhat+" should contain \"PrivateKey\" or \"PublicKey\".");
            fis.close();
        } catch (IOException e) {
            System.err.println("[ExtensionSecurityBox] Error reading key file: "+keyFileName+": "+e);
            System.exit(1);
            return null;
        }

        return ret;
    }


    /**
     * Gets a <code>SignInfo</code> message for an object.
     *
     * @param o the object to be signed
     * @param dir the direction the <code>SignInfo</code> message will go
     * @return the <code>SignInfo</code> messages for the object with
     *         specific direction.
     */
    private SignInfo getNewSignInfo(Object o, Direction dir) {
	if ( o instanceof PlanningInfo )
	    return  PlanningInfoSignInfo.getNewMessage(coordinator.connectionID(),
						       dir,
						       o,
						       getSignAlgo(),
						       getOwnPrivateKey());
	else if ( o instanceof PlanDistro )
	    return  PlanDistroSignInfo.getNewMessage(coordinator.connectionID(),
						     dir,
						     o,
						     getSignAlgo(),
						     getOwnPrivateKey());
	else if ( o instanceof KeyDistro )
	    return  PlanDistroSignInfo.getNewMessage(coordinator.connectionID(),
						     dir,
						     o,
						     getSignAlgo(),
						     getOwnPrivateKey());
	else if ( o instanceof BoxCharacteristics )
	    return  BoxCharacteristicsSignInfo.getNewMessage(coordinator.connectionID(),
							     dir,
							     o,
							     getSignAlgo(),
							     getOwnPrivateKey());
	throw new RuntimeException("Type Error");
    }


    /**
     * Gets the KeyDistro message intended for a specific node
     *
     * @param node   the KeyDistro message to send to
     * @param label  the key label
     * @param skey   the byte array of the secret key to be distributed
     * @param connID the identification of this connection
     * @return the KeyDistro message with a specific node as the
     *         recipient. If the node does not have the verified public key,
     *         null is returned.
     */
    private KeyDistro getKeyDistro(NodeID node, String label, byte[] skey, ConnID connID) {

        /* The public key of the node */
        PublicKey pk       = getVerifiedPublicKey(node);

        if ( pk == null ) {
            if (Debug.on) {
                Debug.println(Debug.SECURITY, Debug.MOREINFO,
                              "[secbox] No public key verified for node "+node);
            }
            return null;   
        }
  
        return KeyDistro.getNewMessage(connID, node, NodeID.getNewLocalID(),
                                       skey, label,
                                       getCipherAlgo(), pk);
    }



}
