/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.ConductorExt.security.adaptors;

import java.security.*;
import xjava.security.Cipher;
import java.io.*;

import cryptix.provider.key.RawKey;

import FMG.Conductor.util.*;
import FMG.Conductor.comm.*;
import FMG.Conductor.dataflow.*;

public class DecryptAdaptor extends Adaptor {

   private static final String DEFAULT_KEYLABEL="DES";
   private static final String DEFAULT_MODE    ="CFB";

   DataAccessPointer pointer = new DataAccessPointer();

   Cipher cipher; 

   Key    key;

   public DecryptAdaptor(AdaptorWindow window, String parameter) {
      super(window, parameter);
      if (Debug.on) {
         Debug.println(Debug.ADAPTOR, Debug.INFO,
                       "[Decryptor] New DecryptAdaptor instantiated");
      }
   }

   public static String getKeyLabel(String parameters) {
      return EncryptAdaptor.getKeyLabel(parameters);
   }

   public static byte[] generateKey(String parameters) {
      return EncryptAdaptor.generateKey(parameters);
   }

   protected void initCipher() {
      key = getKey();

      if (Debug.on) {
         Debug.print(Debug.ADAPTOR, Debug.INFO, "[Decryptor] Got my key");
      }

      try {
	  cipher = Cipher.getInstance(getKeyLabel(null) + "/" + DEFAULT_MODE);
	  cipher.initDecrypt(key);
      } catch ( NoSuchAlgorithmException e ) {
	  throw new RuntimeException("No such algorithm as "+
				     getKeyLabel(null) + "/" + DEFAULT_MODE);
      } catch ( KeyException e ) {
	  throw new RuntimeException("KeyException in cipher initialization");
      }

      if ( Debug.on && (Debug.getLevel(Debug.ADAPTOR) > Debug.MOREINFO )) {
	 Debug.println("[Decryptor] Cipher Properties of this Decryptor:");
	 Debug.println("[Decryptor]       Algorithm:" + cipher.getAlgorithm());
	 Debug.println("[Decryptor]       Mode     :" + cipher.getMode());
	 Debug.println("[Decryptor]       Padding  :" + cipher.getPadding());
	 Debug.println("[Decryptor]       Provider :" + cipher.getProvider());
	 Debug.println("[Decryptor]       Blocksize:" + cipher.blockSize());
      }

   }

   public void handleFlow() {
      if (Debug.on) {
         Debug.print(Debug.ADAPTOR, Debug.INFO, "[Decryptor] (");
      }

      window.insertPointer(pointer);

      while (true) {
         try {
            int len = window.expand();
            if (Debug.on) {
               Debug.println(Debug.ADAPTOR, Debug.DETAIL, 
                             "[Decryptor] moving " + len + " bytes");
            }
            processWindow(len);
            window.contract(pointer);
                  } catch (InputClosedException e) {
            if (window.size() != 0) {
               Debug.println(this +
                             " WARNING: shutting down with non-empty window");
            }
   
            pointer.removeSelf();
            window.contract();
            window.closeOutput();
            if (Debug.on) {
               Debug.println(Debug.ADAPTOR, Debug.INFO,
                             this + " input closed");
            }
            return;
         } catch (OutputClosedException e) {
            if (window.size() != 0) {
               Debug.println(this +
                             " WARNING: shutting down with non-empty window");
            }
            
            if (Debug.on) {
               Debug.println(Debug.ADAPTOR, Debug.INFO,
                             this + " output closed");
            }
            window.closeInput();
            return;
         } catch (PlanChangedException e) {
            if (Debug.on) {
               Debug.println(Debug.ADAPTOR, Debug.INFO,
                             this + " plan changed");
            }
            pointer.removeSelf();
            window.contract();
            window.advanceSwitchPlan();
            window.insertPointer(pointer);
         } catch (AsyncEventException e) {
            throw new RuntimeException("Case not handled");
         }
      }
   }

   private Key getKey() {
       return getRawKey();
   }

   private Key getRawKey() {
      byte[] keyBytes;

      try {
	  keyBytes =  (byte[])getStreamIAC().waitFor(getKeyLabel(null));
      } catch ( InterruptedException e ) {
	  throw new RuntimeException("Interrupted");
      } 

      if (Debug.on && (Debug.getLevel(Debug.ADAPTOR) > Debug.MOREINFO )) {
	  Debug.print("[Decryptor] Decryption key: ["+keyBytes.length +"|");
	  Debug.println(Debug.printableString(keyBytes, 0, keyBytes.length-1)+"]");
      }

      return new RawKey(getKeyLabel(null), keyBytes);
   }

   private void processWindow(int len) {
      byte[] in = new byte[len];
      byte[] out;

      pointer.getBytes(in, 0, 0, len);
      
      if (Debug.on && (Debug.getLevel(Debug.ADAPTOR) > Debug.DETAIL)) {
	  Debug.print("[Decryptor] before decryption: ["+len+"|");
	  Debug.println(Debug.printableString(in , 0, len-1)+"]");
      }

      if (cipher == null) {
        initCipher();
      }

      out = cipher.crypt(in);

      if ( out.length != len )
	  throw new RuntimeException("Length change after decryption");

      if (Debug.on && (Debug.getLevel(Debug.ADAPTOR) > Debug.DETAIL)) {
	  Debug.print("[Decryptor] after  decryption: ["+out.length+"|");
	  Debug.println(Debug.printableString(out, 0, out.length-1)+"]");
      }

      for (int i = 0; i<len; i++) {
         pointer.replaceByte(i, out[i]);
      }
      
      pointer.advance(out.length);
   }

}
