/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.ConductorExt.security.cert;


import java.net.*;
import java.io.*;
import java.security.*; 
import xjava.security.Cipher;

import FMG.Conductor.util.*;
import FMG.Conductor.comm.*;
import FMG.Conductor.util.Debug;

/**
 * The CertClient establishes connection with a certificate server and it
 * then can ask for the retrieval of a certificate from the server, given
 * the certificate quest.
 *
 * @author Jun Li (lijun@fmg.cs.ucla.edu)
 * @version $Revision: 1.9 $
 */
public class CertClient {

    /** The socket that connects the certificate client and the server */
    Socket s;

    /** The input stream from which a certificate from server is read */
    DataInputStream dis;

    /** Th output stream through which a request is sent to the server */
    DataOutputStream dos;
    
    /**
     * Creates a new certificate client connected with a specific
     * certificate server at particular port number.
     * 
     * @param cs the String representing the certificate server and the port
     */
    public CertClient(String  cs) {
	String      serverName = null;
	InetAddress serverAddr = null;
	int         port;
	int         index = cs.indexOf(':');

	if ( index == -1 ) {
	    throw new RuntimeException("The Cert Server name must contain port number");
	} else {
	    serverName = new String(  cs.substring(0, index) );
	    port       = new Integer( cs.substring(index+1)  ).intValue();
	}

        try {
            serverAddr = InetAddress.getByName(serverName);
        } catch ( UnknownHostException e ) {
	    throw new RuntimeException("Unknown machine "+serverName);
        }

        try {
	    if ( Debug.on )
		Debug.println(Debug.SECURITY, Debug.MOREINFO, 
                              "[certclnt] Contacting certificate server "+
			      serverAddr+"("+port +")");

            s = new Socket(serverAddr, port);

	    if ( Debug.on)
		Debug.println(Debug.SECURITY, Debug.MOREINFO, 
                              "[certclnt] Conection establised: "+s);

            dos = new DataOutputStream(s.getOutputStream());
            dis = new DataInputStream(s.getInputStream());
        } catch (IOException e) {
            done();
            throw new RuntimeException("Can not establish connection with certificate server at " + 
                               serverName + "[" + e + "]");
        }
    }
    
    /**
     * Gets a certificate for a specific certificate request
     *
     * @param req the certificate request
     * @return the certificate based on the given request.
     *         null if no certificate can be provided from the server.
     */
    public Certificate getCertificate(CertRequest req ) 
        throws IOException {
        req.writeSelf(dos);
	req.recycle();       // recycling this certificate request

	if ( ! dis.readBoolean() )
	    return null;

        Certificate cert = Certificate.make(dis);
	if ( Debug.on)
	    Debug.println(Debug.SECURITY, Debug.MOREINFO, 
                          "[certclnt] Get a certificate as "+cert);
        return cert;
    }
    
    /**
     * Cleans the environment when the certificate client quits.
     *
     * @return no return value
     */
    public void done() {
        try {
            if (s != null) {
                s.close();
            }
        } catch (IOException ignore) {}
    }
    
    /**
     * Tests the CertClient works correctly.
     */
    public static final void main(String[] args) {
        String serverinfo = null;
        String testhost   = null;

        switch (args.length) {
        case 1:
            serverinfo = args[0];
            testhost   = "";
            break;
        case 2:
            serverinfo = args[0];
            testhost   = args[1];
            break;
        default:
            System.out.println("Usage:\n" + "java FMG.ConductorExt.security.cert.CertClient <server> [ <test-host> ] ");
            System.exit(1);
        }

        try {
            self_test(serverinfo, testhost);
        } catch ( Exception e) {
            e.printStackTrace();

        }
    }
    
    /**
     * Self-test procedure
     *
     * @param serverName the name of the certificate server
     * @param testName   the name of the host whose certificate is needed
     * @return no return value
     * @exception java.io.IOException if error happens when getting
     *                                certificate from server
     */
    private static void self_test(String serverInfo, String testName)
        throws IOException {
        
        String  TEST_KEY_ALGO = "RSA";
        String  TEST_SIGN_ALGO = "SHA-1/RSA/PKCS#1";
        
        CertClient cc = new CertClient(serverInfo);
        if ( testName == "" ) 
	    testName = InetAddress.getLocalHost().getHostName();

        cc.getCertificate(CertRequest.make(testName, TEST_KEY_ALGO, TEST_SIGN_ALGO));
        
    }
}
