/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.ConductorExt.security.cert;

import java.net.*;
import java.io.*;

import FMG.Conductor.comm.*;
import FMG.Conductor.util.*;

/**
 * This CertRequest class defines the format of the certificate request,
 * which is to be sent to a certificate server to retrive the certificate
 * for a particular node.
 *
 * @author Jun Li (lijun@fmg.cs.ucla.edu)
 * @version $Revision: 1.6 $
 */

public class CertRequest extends Recyclable{

    /** The identification of the node whose certificate is needed */
    String      node;

    /** The public key algorithm name to be used by the certificate server */
    String      keyType;

    /** The signing algorithm name to be used by the certificate server */
    String      signType;

    /** The recycler used by this CertRequest */
    private static Recycler recycler = new Recycler(CertRequest.class);

    /**
     * Gets a certificate request by reading from stream.
     *
     * @param is the stream to read from
     * @return the certificate request coming from the stream
     * @exception java.io.IOException if an error occurs while reading from
     *                                the stream.
     */
    public static CertRequest make(DataInputStream is) throws IOException {
        CertRequest req = (CertRequest)recycler.getOne();
        req.readSelf(is);
        return req;
    }

    /**
     * Makes a certificate request with given parameters.
     *
     * @param id whose certificate to be retrieved
     * @param k  what public key algorithm to ask the server to use
     * @param s  what signing algorithm to ask the server to use
     * @return a certificate request ready to send out
     */
    public static CertRequest make(String id, String k, String s) {
        CertRequest req = (CertRequest)recycler.getOne();
        req.init(id, k, s);
        return req;
    }
    
    /**
     * Writes the information contained in this CertRequest to a stream
     * Presumably the information is to be read by readSelf().
     *
     * @param os the output stream to write the elements of this CertRequest
     * @return no return value
     */
    public void writeSelf(DataOutputStream os) throws IOException{
        os.writeBytes(node+"\n");
        os.writeBytes(keyType+"\n");
        os.writeBytes(signType+"\n");
    }
    
    /**
     * Updates this CertRequest with the given information from stream
     * Presumably the infomation is written by writeSelf().
     *
     * @param is the stream from which to get information
     * @return no return value
     * @exception java.io.IOException if an error occurs while reading from the
     *                                stream.
     */
    public void readSelf(DataInputStream is) throws IOException{
	BufferedReader br = new BufferedReader(new InputStreamReader(is));
	node = br.readLine();
        keyType = br.readLine();
        signType = br.readLine();
    }

    /**
     * Gets the string representation of this CertRequest.
     *
     * @return the string value of this CertRequest
     */
    public String toString() {
        return node+"|"+keyType+"|"+signType;
    }

    /*
     * private functions
     */

    /**
     * Initializes this CertRequest with given parameters.
     *
     * @return no return value.
     */
    private void init(String id, String k, String s){
	node     = new String(id);
        keyType  = new String(k);
        signType = new String(s);
    }

}
