/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.ConductorExt.security.cert;

import java.security.*;
import xjava.security.Cipher;
import java.io.*;
import java.util.*;
import java.net.*;

import FMG.Conductor.comm.NodeID;
import FMG.ConductorExt.security.util.*;

/**
 * This CertServer class implements a certificate server as a running
 * thread. It listens to connection requests from certificate clients, and
 * when one request is received, a new certificate server will be started up
 * to communicate with the certificate client.
 * <p> 
 * It is the certificate request (@see CertRequest) that a certificate
 * client sends to the server, and based on that, the certificate server will
 * form a certificate satisfying the request, and send back to the client. 
 * <p> 
 * When a certificate server starts up, it will build the in-memory
 * storage of all certificates by reading from the database(@see
 * multiSigningMatrix).
 *
 * @author Jun Li (lijun@fmg.cs.ucla.edu)
 * @version $Revision: 1.9 $
 */

public class CertServer extends Thread {

    /** The port number on which the server listens to the connection request */
    public  static final int    DEFAULT_ACCEPTOR_PORT = 4000;

    /** The default name for the properties file. */
    private static final String DEFAULT_PROPERTIES_FILE = "./certserver.properties";

    /** The default public key type that this certitificate server certifies */
    private static final String DEFAULT_PUBLIC_KEY_TYPE = "RSA";

    /** The default signing methods used by this certificate server */
    private static final String DEFAULT_SIGN_METHODS = "SHA-1/RSA/PKCS#1:RSA";

    /** The list of properties read from the properties file */
    private static Properties props              = new Properties();

    /**
     *  The hashtable of certificate server private keys, with the public
     *  key algorithm name as the key
     */
    private static Hashtable  csPrivateKeys      = new Hashtable();

    /**
     *  The storage of the certifying information of all the nodes, with
     *  the public key algorithm name as the key; furthermore, each
     *  element is also a hashtable of certifying information, with the id of the node(@see
     *  FMG.Conductor.comm.NodeID) as the key. The certifying information
     *  is of class MultiSigning(@see MultiSigning).
     */
    private static Hashtable  multiSigningMatrix = new Hashtable();

    /**
     * The port on which the certificate server is listening
     */
    private static int        port;

    /**
     * The socket with which this CertServer communicates with a certificate client
     */
    private Socket            client;
    
    /**
     * Creates an empty CertServer.
     *
     * @return no return value.
     */
    CertServer() {}

    /**
     * Creates a CertServer communicating with a certificate client using
     * the given socket.
     *
     * @param s the socket to communicate with a specific certificate client
     * @return no return value.
     */
    CertServer(Socket s) {
        client = s;
    }
    
    /**
     * The main function executed when a certificate server starts up.
     *
     * @param args the arguments passed from command line.
     * @return no return value. The function may never return.
     */
    public static void main(String[] args) {
	int    port     = CertServer.DEFAULT_ACCEPTOR_PORT;;
	String propFile = new String(DEFAULT_PROPERTIES_FILE);

	int i = 0;
	while ( i < args.length ) {
	    if (args[i].equals("-f")) {
		i++;
		if ( args[i] == "" )
		    throw new RuntimeException("Usage error. No prop file specified");
		propFile = new String(args[i]);
	    } else if (args[i].equals("-p")) {
		i++;
		if ( args[i] == "" )
		    throw new RuntimeException("Usage error. No port number specified");
		port = new Integer(args[i]).intValue();
	    }

	    ++i;
	}

	readPropertiesFile(propFile);
	setPort(port);
        getCSPrivateKeys();
        buildCertificates();

        (new CertServer()).runServer();
    }

    /**
     * Runs the certificate server.
     * <p> 
     * It creates a server socket first and then listens on that
     * socket. Whenever a new connection comes, a new certificate server
     * will be created and starts running.
     *
     * @return no return value.
     */
    public void runServer() {
        try {
            System.out.println("Listening on port "+port);

            ServerSocket s = new ServerSocket(port);
            System.out.println("Waiting for connections");
            CertServer t;
            Socket newSocket;
            int i=0;
            while (true) {
                newSocket = s.accept();
                System.out.println("Got new connection: "+newSocket);
                t = new CertServer( newSocket );
                t.setName(String.valueOf(++i));
                t.start();
                this.yield();
                System.out.println("Waiting for another connection");
            }
        } catch (IOException e) {
            System.err.println("Error accepting connections: " + e);
            System.exit(1);
        }
    }

    /**
     * This function implements the run() function, since the CertServer
     * is a subclass of Thread.
     *
     * @return no return value.
     */
    public void run() {
        DataInputStream  dis      = null;
        DataOutputStream dos      = null;
        CertRequest      request  = null;

        try {
            dis = new DataInputStream(client.getInputStream());
            dos = new DataOutputStream(client.getOutputStream());

            while (true) {
                request = CertRequest.make(dis);
                System.out.println("       "+ getName()+" Got request for "+ request.node);

                Hashtable mslist=(Hashtable)multiSigningMatrix.get(request.keyType);
                MultiSigning ms  = (MultiSigning)mslist.get(request.node);
                if ( ms == null ) {
                    System.out.println("       "+ getName()+" Can not certify "+ request.node );
		    dos.writeBoolean(false);
                } else {
		    dos.writeBoolean(true);
		    ms.genCert(request.signType+request.keyType).writeSelf(dos);
		}
                System.out.println("       "+ getName()+" Response sent");
		request.recycle();    // recycle the certificate request
            }
        } catch (EOFException e) {
            System.out.println("       "+ getName()+" Connection closing.");
        } catch (Exception e) {
            System.out.println("       "+ getName()+" Problem handling request("+request+"):" + e);
        }
        
        try {
            if (dis != null) {
                client.close();
            }
        } catch (IOException e) {}
    }

    /*
     * private functions
     */

    /**
     * Reads the properties for this certificate server from a file.
     * 
     * @param filename the name of the property file
     * @return no return value.
     */
    private static void readPropertiesFile(String filename) {
        /* if the user specified a null filename, don't read any properties */
        if (filename.length() == 0) {
            return;
        }
        
        /* read the properties file */
        try {
            File propFile = new File(filename);
            FileInputStream fis = new FileInputStream(propFile);
            props.load(fis);
        } catch (IOException e) { 
            // Verbosity is not set yet, so just go ahead and warn the user
            System.err.println("Warning: error reading properties file (" + 
                               e + ")");
        }
    }

    /**
     * Sets the port on which the cert server listens to cert client request.
     *
     * @param port the port number on which the cert server listens.
     * @return no return value.
     */
    private static void setPort(int port) {
	CertServer.port = port;
    }

    /**
     * Prepares the certifying information by reading the public key database.
     * <p>
     * This certificate server may provide different type of public keys,
     * and it can also sign a certificate using different signing methods,
     * that is, the different public key algorithm and signing
     * algorithm. So the prepared certifying information includes
     * signatures with different methods.
     * <p>
     * The result is stored in <code>multiSigningMatrix</code>(@see
     * multiSigningMatrix).
     */
    private static void buildCertificates() {
        Enumeration keyTypes   = getKeyTypes();
        Vector      signTokens = getSignTokens();
        String      aNode;
	byte[]      pubKeyBytes;
        String      myself; 
        String      authorityForMyself = getAuthorityForMyself();

	try {
	    myself = InetAddress.getLocalHost().getHostName() + ":" 
		     + new Integer(port).toString();
	} catch ( UnknownHostException e ) {
	    throw new RuntimeException("Unknown host exception");
	}

        while (keyTypes.hasMoreElements()) {
            String    ktype            = (String)keyTypes.nextElement();
            String    keyFileName      = props.getProperty("PublicKeyFile_"+ktype);

            Hashtable multiSigningList = new Hashtable();

	    System.out.println(ktype + " public keys maintained for:");

            try{
                FileInputStream fis = new FileInputStream(new File(keyFileName));
		BufferedReader   br = new BufferedReader(new InputStreamReader(fis));

                boolean done = false;
                
                while ( !done) {
                    
                    try {
                        aNode = readHost(br);
                    } catch (EOFException e) {
                        fis.close();
                        done = true;
                        break;     // jump out of while (!done)  loop
                    }
                    
                    pubKeyBytes  = readKeyBytes(br);

                    MultiSigning ms = new MultiSigning(aNode,
                                                       myself,
                                                       authorityForMyself,
                                                       pubKeyBytes,
                                                       signTokens);
                    multiSigningList.put(aNode, ms);
		    System.out.println("     " + aNode );
                }
                multiSigningMatrix.put(ktype, multiSigningList);
            }  catch (IOException e) {
                System.err.println("[CertServer] Error reading key file: " + 
                                   keyFileName + ": " + e);
                System.exit(1);
            } 
        }
    }

    /**
     * Gets the types of public keys that this certificate server can
     * certify.
     * <p>
     * The "PublicKeyType" property tells the types delimited by space.
     * 
     * @return the enumeration list of public key types that the
     *         certificate server can certify.
     */
    private static Enumeration getKeyTypes() {
        Vector keyTypes = new Vector();

        String types = props.getProperty("PublicKeyType", DEFAULT_PUBLIC_KEY_TYPE);
        int    from  = 0;
        int    index = types.indexOf(' ');

        while ( index != -1 ) {
          keyTypes.addElement(types.substring(from, index));
          from  = index +1;
          index = types.indexOf(' ', from);
        }
        keyTypes.addElement(types.substring(from));

        return keyTypes.elements();
    }

    /**
     * Gets all the private keys of the certificate server, in terms of
     * different public key type that this server certifies. The result is
     * put into <code>csPrivateKeys</code>(@see csPrivateKeys).
     *
     * @return no return value.
     */
    private static void getCSPrivateKeys() {

        Enumeration keyTypes = getKeyTypes();
        while (keyTypes.hasMoreElements()) {
            String ktype    = (String)keyTypes.nextElement();
            String filename = props.getProperty("PrivateKeyFile_"+ktype);
            try {
                FileInputStream fis = new FileInputStream(new File(filename));
                PrivateKey  privKey = KeySerializer.readPrivKey(fis, ktype);
                csPrivateKeys.put(ktype, privKey);
                fis.close();
            } catch (IOException e) {
                System.err.println("[CertServer] Error reading" +
                                   " private key file: " +  filename + ": " + e);
                System.exit(1);
            }
        }
    }

    /**
     * Gets the vector of <code>SignToken</code>s, that is, the different
     * signing methods that a certificate server is capable of using.
     * <p>
     * The "SignMethods" property is a string of different signing methods
     * delimited by space. For example: 
     *           SignMethod=SHA-1/RSA/PKCS#1:RSA SHA-1/XXX/padding:XXX
     * @see SignToken
     */
    private static Vector getSignTokens() {
        Vector tokens = new Vector();
        String signMethods = props.getProperty("SignMethods", DEFAULT_SIGN_METHODS);

        int from =0;
        int index = signMethods.indexOf(' ');
        while ( index != -1 ) {
            tokens.addElement(new SignToken(signMethods.substring(from, index), csPrivateKeys));
            from = index + 1;
            index = signMethods.indexOf(' ', from);
        }
        tokens.addElement( new SignToken(signMethods.substring(from), csPrivateKeys) );

        return tokens;

    }

    private static String getAuthorityForMyself() {
	return props.getProperty("MyAuthority");
    }

    /**
     * Reads the name of a host from a BufferedReader.
     *
     * @param br the BufferedReader to read from
     *
     * @return the name of a host
     *
     * @exception java.io.IOException  if error happens when reading
     *            java.io.EOFException if end of file is reached when reading
     */
    private static String readHost(BufferedReader br) 
        throws EOFException, IOException {
        byte[] buf = new byte[100];
        int b;
        int i = 0;
        
        for (b = br.read(); (b>0) && (b!=' '); b = br.read()) {
            buf[i++] = (byte) b;
        }
        if (b<0) {
            throw new EOFException("Done with input file");
        }
        return new String(buf,0,i);
    }


    /**
     * Reads the bytes of a public or private key from a BufferedReader.
     *
     * @param br the BufferedReader to read from
     *
     * @return The bytes representing the key, which are already converted
     *         from the readable form.
     *
     * @exception java.io.IOException  if error happens when reading
     *
     */
    private static byte[] readKeyBytes(BufferedReader br) 
	throws IOException {

	byte[] linebuf = br.readLine().getBytes();

	ByteArrayInputStream    bais = new ByteArrayInputStream(linebuf);
	ReadableBytesInputStream ris = new ReadableBytesInputStream(bais);
	byte[]                   buf = new byte[linebuf.length/2];

	ris.read(buf, 0, buf.length);

	return buf;
    }
}
