/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.ConductorExt.security.cert;

import java.io.*;
import java.net.*;
import java.security.*;
import xjava.security.Cipher;

import FMG.Conductor.comm.*;
import FMG.Conductor.util.*;

import FMG.ConductorExt.security.util.*;
import FMG.ConductorExt.security.util.Signer;

/**
 * This Certificate class defines the format of the certificate used in
 * Condunctor security implementation. It consists of: the certificate
 * data, the certificate signature, as well as the status.
 * <p>
 * The Certificate class is a subclass of Recyclable class.
 *
 * @author Jun Li (lijun@fmg.cs.ucla.edu)
 * @version $Revision: 1.9 $
 */
public class Certificate extends Recyclable {
    /** A status meaning the authenticity of a certificate is not known */
    private static int      UNKNOWN   = 0;

    /** A status meaning a certificate is known as authentic */
    private static int      AUTHENTIC = 1;

    /** A status meaning a certificate is known as not authentic */
    private static int      UNAUTHENTIC = 2;

    /** The authenticity status of a certificate */
    private int             status;

    /** The data part of the certificate */
    private CertificateData certData;

    /** The signature on the certificate data by the certificate server */
    private byte[]          sig;

    /** The recycler used for this Certificate */
    private static Recycler recycler = new Recycler(Certificate.class);

    /** 
     * Makes a certificate with given certificate data and signature.
     *
     * @param certData the certificate data
     * @param sig the signature of the certificate data by a certificate server
     * @return a certificate composed of the given certificate data and signature
     * @see FMG.ConductorExt.security.cert.CertificateData
     */
    public static Certificate make(CertificateData certData, byte[] sig) {
        Certificate cert = (Certificate)recycler.getOne();
        cert.init(certData, sig);
        return cert;
    }

    /** 
     * Makes a certificate by reading from an input stream
     *
     * @param dis the input stream to read certificate contents
     * @return a certificate
     * @exception java.io.IOException if an error occurs while reading from
     *                                the stream.
     */
    public static Certificate make(DataInputStream dis) throws IOException {
        Certificate cert = (Certificate)recycler.getOne();
        cert.init(dis);
        return cert;
    }

    /**
     * Verifies the signature in this certificate
     *
     * @param sigName     the name of the signing algorithm to use
     * @param pk          the public key used to verify
     * @param checkStatus whether to firstly check the status of this cert or not
     * @return <code>true</code> if verified; <code>false</code> otherwise
     */
    public boolean verify(String sigName, PublicKey pk, boolean checkStatus) {
	boolean res;

	if ( status == UNKNOWN || ! checkStatus) {
	    res = Signer.verify(certData, sig, sigName, pk);
	    status = res ? AUTHENTIC : UNAUTHENTIC;
	} else {
	    res = ( status == AUTHENTIC );
	}

	return res;
    }

    /**
     * Verifies the signature in this certificate
     *
     * @param sigName the name of the signing algorithm to use
     * @param pk      the public key used to verify
     * @return <code>true</code> if verified; <code>false</code> otherwise
     */
    public boolean verify(String sigName, PublicKey pk) {
	return verify(sigName, pk, false);
    }

   /**
    * This method is used to read the data elements of this object from
    * the input stream.  Presumably the information was generated by
    * <code>writeSelf()</code>.
    * @param is The input stream to read the message's data from.
    * @return No return value.
    * @exception java.io.IOException if an error occurs while reading from the
    *                                stream.
    */
    public void readSelf(DataInputStream is) throws IOException {
        certData = CertificateData.make(is);
        sig      = BytesReader.getFrom(is);
    }

   /**
    * This method is used to write the data elements of this object to
    * an output stream.  Presumably the information will be read by
    * <code>readSelf()</code>.
    * @param os The output stream to write the message's data to.
    * @return No return value.
    * @exception java.io.IOException if an error occurs while writing to the
    *                                stream.
    */
    public void writeSelf(DataOutputStream os) throws IOException {
        certData.writeSelf(os);
        if ( sig == null )
            os.writeInt(0);
        else {
            os.writeInt(sig.length);
            os.write(sig, 0, sig.length);
        }
    }

    /**
     * Gets the node  that this certificate belongs
     *
     * @return the node associated with this certificate
     */
    public String getNode() {
        return certData.node;
    }
  
    /**
     * Gets the id of the certificate server
     *
     * @return the certificate server associated with this certificate
     */
    public String getAuthority() {
        return certData.authority;
    }
  
    /**
     * Gets the id of the certificate server for the certificate server
     *
     * @return the certificate server for the certificate server. It is
     *         null when chain strategy is employed.
     */
    public String getAuthorityForAuthority() {
        return certData.authorityForAuthority;
    }
    
    /**
     * Gets the public key that is being certified
     *
     * @return the public key in byte array form associated with this
     certificate
    */
    public byte[] getPubKey() {
        return certData.pubKey;
    }

    /**
     * Perform cleanup before allowing recycling to occur.  This method
     * is called by the Recycler.
     * @return No return value.
     * @see FMG.Conductor.util.Recycler
     */
    public void recycling() {
        super.recycling();
        if (certData != null ) {
            certData.recycle();
        }
    }

    /**
     * Generates the representative string of this certificate
     *
     * @return the string representation
     */
    public String toString() {
	//        return certData.toString()+ "["+sig+"]";
	return certData.toString();
    }

    /*
     * private functions
     */

    /**
     * Initializes this certificate with given certificate data and signature
     *
     * @param certData the certificate data
     * @param sig      the corresponding signature of the certificate data
     * @return no return value
     */
    private void init(CertificateData certData, byte[] sig){
        this.certData = certData;
        this.sig      = sig;
	status = UNKNOWN;
    }
    
    /**
     * Initializes this certificate by reading from an input stream
     *
     * @param dis the input stream to read certificate information
     * @return no return value
     * @exception java.io.IOException if an error occurs while reading from
     *                                the stream.
     */
    private void init(DataInputStream dis) throws IOException {
        this.readSelf(dis);
	status = UNKNOWN;
    }


}
