/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.ConductorExt.security.cert;

import FMG.Conductor.comm.NodeID; 

import java.io.*;
import java.net.InetAddress;

import FMG.Conductor.util.*;
import FMG.ConductorExt.security.util.*;

/**
 * The CertificateData class manages the data part of a certificate, that
 * is, the node whose public key is being certified, the certificate server
 * who certifies the public key and its own certificate server, and the
 * public key itself.
 * <p>
 * The Certificate class is a subclass of Recyclable class.
 *
 * @author Jun Li (lijun@fmg.cs.ucla.edu)
 * @version $Revision: 1.7 $
 */
public class CertificateData extends Recyclable{
    /** the node whose public key is being certified */
    String node;

    /** the certificate server who certifies the public key */
    String authority;

    /** the certificate server who certifies the above certificate server */
    String authorityForAuthority;

    /** the public key being certified */
    byte[] pubKey;

    /** The recycler used for this CertificateData */
    private static Recycler recycler = new Recycler(CertificateData.class);

    /**
     * Creates a new (or recycled) CertificateData with given parameters
     *
     * @param node                  whose certificate data
     * @param authority             who certifies the public key
     * @param authorityForAuthority who certifies the public key
     * @param pk                    the public key to certify
     * @return a CertificateData initilized with given parameters
     */
    public static CertificateData make(String node, 
                                       String authority,
                                       String authorityForAuthority,
                                       byte[] pk) {
        CertificateData cd = (CertificateData)recycler.getOne();
        cd.init(node, authority, authorityForAuthority, pk);
        return cd;
    }

    /**
     * Creates a new(or recycled) CertificateData by reading from stream
     *
     * @param is the input stream to read certificate data information
     * @return a CertificateData initialized from the stream
     * @exception java.io.IOException if an error occurs while reading from
     *                                the stream.
     */
     public static CertificateData make(DataInputStream is) 
        throws IOException {
        CertificateData cd = CertificateData.create();
        cd.readSelf(is);
        return cd;
    }

    /**
     * Updates this CertificateData with the given information from stream
     * Presumably the infomation is written by writeSelf().
     *
     * @param is the stream from which to get information
     * @return no return value
     * @exception java.io.IOException if an error occurs while reading from the
     *                                stream.
     */
    public void readSelf(DataInputStream is) throws IOException {
	node       = new String(BytesReader.getFrom(is));
	authority  = new String(BytesReader.getFrom(is));
        if ( is.readBoolean() )
	    authorityForAuthority = new String(BytesReader.getFrom(is));
        pubKey = BytesReader.getFrom(is);
    }

    /**
     * Writes the information contained in this CertificateData to a stream
     * Presumably the information is to be read by readSelf().
     *
     * @param os the output stream to write the elements of this CertificateData
     * @return no return value
     */
    public void writeSelf(DataOutputStream os) throws IOException {
        os.writeInt(node.length());
        os.write(node.getBytes());

        os.writeInt(authority.length());
        os.write(authority.getBytes());
	
        os.writeBoolean(authorityForAuthority != null );

        if ( authorityForAuthority != null ) {
	    os.writeInt(authorityForAuthority.length());
	    os.write(authorityForAuthority.getBytes());
	}

        os.writeInt(pubKey.length);
        os.write(pubKey, 0, pubKey.length);
    }
    
    /**
     * Gets the string representation of this CertificateData
     *
     * @return no return value
     */
    public String toString() {
        String str = new String (pubKey);
	//        return ( "["+node + "] by [" + authority +"] pub key is (" + str + ")" );
        return ( "["+node + "] by [" + authority +"]" );
    } 

    /*
     * private functions 
     */

    /**
     * Creates a new (or recycled) CertificateData
     */
    private static CertificateData create() {
        CertificateData cd = (CertificateData)recycler.getOne();
        cd.init();
        return cd;
    }
    
    /**
     * Initializes this CertificateData without any parameters.
     *
     * @return no return value.
     */
    private void init() {
	node = null;
	authority = null;
	authorityForAuthority = null;
        pubKey    = null;
    }

    /**
     * Initializes this CertificateData with given parameters.
     *
     * @return no return value.
     */
    private void init(String node, 
                      String authority, 
                      String authorityForAuthority, 
                      byte[] pk) {
        this.node                  = new String(node);
        this.authority             = new String(authority);
	if ( authorityForAuthority != null )
	    this.authorityForAuthority = new String(authorityForAuthority);
	else
	    this.authorityForAuthority = null;

        this.pubKey                = pk;
    }
}
