/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.ConductorExt.security.cert;

import java.util.*;
import java.io.*;
import java.net.*;

import FMG.Conductor.comm.*;
import FMG.Conductor.util.*;

/**
 * This CertificateList class handles a list of Certificates. 
 *
 * @author Jun Li (lijun@fmg.cs.ucla.edu)
 * @version $Revision: 1.8 $
 *
 * @see Certificate
 */
public class CertificateList extends Recyclable {

    /** the vector with each element a Certificate */
    Vector list = new Vector();

    /** The recycler used for this CertificateList class */
    private static Recycler recycler = new Recycler(CertificateList.class);

    /** 
     * Returna a new (or recycled) CertificateList
     * @return a new(or recycled) CertificateList
     */    
    public static CertificateList create() {
        return (CertificateList)recycler.getOne();
    }
    
    /**
     * Returns a new(or recycled) CertificateList by reading from stream
     *
     * @param is the stream to read from
     * @return a new(or recycled) CertificateList
     */
    public static CertificateList make(DataInputStream is) 
        throws IOException {

        CertificateList cl = (CertificateList)recycler.getOne();
        cl.readSelf(is);
        return cl;
    }
    
    /**
     * Gets all the certificates contained in this CertificateList.
     *
     * @return a enumertation of all certificate elements
     */
    public Enumeration getAll() {
        return list.elements();
    }
    
    /**
     * Adds a new certificate into this CertificateList
     *
     * @param cert the new certificate to add
     * @return no return value
     */
    public void addCertificate(Certificate cert) {
        if (cert != null) {
	    //            list.addElement(cert);    // like a queue
	    list.insertElementAt(cert, 0);          // like a stack
        }
    }

    /**
     * Removes the oldest certificate in the list 
     * @return no return value
     */
    public void removeOldestCertificate() {
	if ( size() == 0 )
	    return;

	Certificate cert = (Certificate)list.firstElement();
	list.removeElementAt(0);
	cert.recycle();
    }

    /**
     * Finds those certificates associated with a specific node.
     *
     * @param node the id of the node whose certificates to be retrieved
     * @return all certificates certifying this node
     */
    public Enumeration findCertsForNode(String node) {
        Vector sublist = new Vector();
        Enumeration certs = list.elements();
        
        while (certs.hasMoreElements()) {
            Certificate c = (Certificate) certs.nextElement();
            if (c.getNode().equals(node)) {
                sublist.addElement(c);
            }
        }
        
        return sublist.elements();
    }
    
    /**
     * Finds one certificate associated with a specific node.
     *
     * @param node the id of the node whose certificates to be retrieved
     * @return the certificate for this specific node in the list
     */
    public Certificate findCertForNode(String node) {
        Enumeration certs = list.elements();
        
        while (certs.hasMoreElements()) {
            Certificate c = (Certificate) certs.nextElement();
            if (c.getNode().equals(node)) {
                return c;
            }
        }
        
        return null;
    }
    
    /**
     * Finds one certificate associated with a specific node.
     *
     * @param node the id of the node whose certificates to be retrieved
     * @return the certificate for this specific node in the list
     */
    public Certificate findCert(String node, String cs) {
        Enumeration certs = list.elements();
        
        while (certs.hasMoreElements()) {
            Certificate c = (Certificate) certs.nextElement();
            if (c.getNode().equals(node) && c.getAuthority().equals(cs)) {
                return c;
            }
        }
        return null;
    }

    /**
     * Determines whether a certificate by a particular server is contained
     *
     * @return <code>true</code> if contains; <code>false</code> otherwise
     */
    public boolean containsCertBy(String cs) {
        Enumeration certs = list.elements();
        
        while (certs.hasMoreElements()) {
            Certificate c = (Certificate) certs.nextElement();
            if (c.getAuthority().equals(cs)) {
                return true;
            }
        }

	return false;
    }
    
    /**
     * Updates this CertificateList with every certificate read from stream
     * Presumably the infomation is written by writeSelf().
     *
     * @param is the stream from which to get information
     * @return no return value
     * @exception java.io.IOException if an error occurs while reading from the
     *                                stream.
     */
    public void readSelf(DataInputStream is) throws IOException {
        int size= is.readInt();
        for ( int i=0; i< size; i++ ) {
            Certificate c = Certificate.make(is);
            addCertificate(c);
        }
    }
    
    /**
     * Writes the information contained in this CertificateList to a stream
     * Presumably the information is to be read by readSelf().
     *
     * @param os the output stream to write the elements of this CertificateList
     * @return no return value
     */
    public void writeSelf(DataOutputStream os) throws IOException {
        os.writeInt(list.size());
        
        Enumeration certs = list.elements();
        while (certs.hasMoreElements()) {
            Certificate c = (Certificate) certs.nextElement();
            c.writeSelf(os);
        }
    }

    /**
     * Gets the size of the list
     *
     * @return the size of the list
     */
    public int size() {
	return list.size();
    }

    /**
     * Gets the string representation of this CertificateList
     *
     * @return no return value
     */
    public String toString() {
        String s="";
        Enumeration certs = list.elements();
        while (certs.hasMoreElements()) {
            Certificate c = (Certificate) certs.nextElement();
            s += c.toString() + "\n";
        }
        return s;
    }
    
    /**
     * Perform cleanup before allowing recycling to occur.  This method
     * is called by the Recycler. 
     * <p>
     * Note:  Each certificate element needs separate recycling.
     * @return No return value.
     * @see FMG.Conductor.util.Recycler
     */
    public void recycling() {
        super.recycling();
        Enumeration certs = list.elements();
        while (certs.hasMoreElements()) {
            Certificate c = (Certificate) certs.nextElement();
	    c.recycle();
	}
        list.removeAllElements();
    }
}
