/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.ConductorExt.security.cert;

import java.io.*;
import java.net.*;
import java.util.*;
import java.security.*;
import xjava.security.Cipher;

import FMG.Conductor.comm.NodeID;
import FMG.ConductorExt.security.util.Signer; 

/**
 * This MultiSigning class is used by a certificate server to keep
 * multiple signatures for each CertificateData instance. In detail, when a
 * certificate server can certify the public key of a node, it can sign the
 * information in serveral different ways, depending on the public key
 * algorithm and signing algorithm to use. 
 *
 * @author Jun Li (lijun@fmg.cs.ucla.edu)
 * @version $Revision: 1.7 $
 */
public class MultiSigning {
    /** The certificate data associated with this MultiSigning */
    private CertificateData certData;

    /**
     * The hashtable where the signatures are stored, with the sum of the
     * signing and public key algorithm name as the key
     */
    private Hashtable       sigs = new Hashtable();

    /**
     * Creates a new MultiSigning object.
     *
     * @param node                   the machine whose public key is being handled
     * @param authority              the certificate server
     * @param authorityForAuthority  the certificate server of the certificate server
     * @param pubkey                 the public key of the node in byte array
     * @param signTokens             a list of signtoken, each signtoken tells how to
     *                               sign the certificate data
     */
    public MultiSigning(String      node, 
                        String      authority,
                        String      authorityForAuthority,
                        byte[]      pubkey, 
                        Vector      signTokens) {

        int i=0;

        certData = CertificateData.make(node, authority, authorityForAuthority, pubkey);
        
        while ( i < signTokens.size() ) {
            SignToken  token   = (SignToken)(signTokens.elementAt(i++));
            String     signAlg = token.getSignAlg();
            String     pkeyAlg = token.getPkeyAlg();
            PrivateKey privKey = token.getPrivKey();
            byte[]     sig     = Signer.calcSig(certData, signAlg, privKey);
            sigs.put(signAlg+pkeyAlg, sig);
        }
    }

    /**
     * Generates a certificate knowing how to sign the certificate.
     *
     * @param howToSign the way to sign the certificate, in fact the sum
     *                  of the public key algorithm name and the signing
     *                  algorithm name
     * @return a certificate derived from this MultiSigning using the
     *         given signing method
     */
    //TODO: if sig is null.
    public Certificate genCert(String howToSign) {

        byte[] sig = (byte[])sigs.get(howToSign);
        /**
         *  if ( sig == null ) 
         *     System.out.println("not found");
         */

        return Certificate.make(certData, sig);
    }

    

}

