/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.ConductorExt.security.cert;

import java.io.*;
import java.util.*;
import java.security.*;
import xjava.security.Cipher;

import FMG.Conductor.util.*;
import FMG.ConductorExt.security.util.KeySerializer;

/**
 * This SignToken class defines how to sign a certificate, particularly it
 * includes the name of the publick key algorithm, the name of the signing
 * algorithm, and the private key of the certificate server.
 *
 * @author Jun Li (lijun@fmg.cs.ucla.edu)
 * @version $Revision: 1.5 $
 *
 * @see MultiSigning
 */
public class SignToken {
    /** the default public key algorithm used to sign a certificate */
    private static final String  DEFAULT_PKEY_ALGO   = "RSA";

    /** the default signing algorithm used to sign a certificate */
    private static final String  DEFAULT_SIGN_ALGO   = "SHA-1/RSA/PKCS#1";

    /** the name of the public key algorithm name used to sign */
    String     pkeyAlgo;

    /** the name of the signing algorithm name used to sign */
    String     signAlgo;

    /** the private key used to sign */
    PrivateKey privKey;

    /**
     * Gets the name of the public key algorithm to sign a certificate
     *
     * @return the public key algorithm name
     */
    String getPkeyAlg() {
        return pkeyAlgo;
    }

    /**
     * Gets the name of the signing algorithm to sign a certificate
     *
     * @return the signing algorithm name
     */
    String getSignAlg() {
        return signAlgo;
    }

    /**
     * Gets the private key used to sign a certificate
     *
     * @return the private key used to sign
     */
    PrivateKey getPrivKey() {
        return privKey;
    }

    /**
     * Creates a SignToken.
     *
     * @param signMethod the string telling the signing algorithm name and
     *                   the public key algorithm name, separated by ":".
     * @param csPrivateKeys the hashtable of the certificate server
     *                      private keys, with the public key algorithm
     *                      name as the key
     */
    public SignToken(String signMethod, Hashtable csPrivateKeys) {
        int index = signMethod.indexOf(':');
        if ( index == -1 ) {
            System.err.println("ERROR of signing method from the property file");
            System.exit(1);
        }

        signAlgo = new String( signMethod.substring(0, index) );
        pkeyAlgo = new String( signMethod.substring(index+1)  );
        privKey  = (PrivateKey)csPrivateKeys.get(pkeyAlgo);
    }

    /*
    public SignToken(PrivateKey priv) {
        SignToken( DEFAULT_PKEY_ALGO, DEFAULT_SIGN_ALGO, priv);
    }

    public SignToken(String pkeyAlgo, String signAlgo, Privatekey priv) {
        this.pkeyAlgo = pkeyAlgo;
        this.signAlgo = signAlgo;
        this.privKey  = priv;
    }
    */
}


