/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.ConductorExt.security.chain;

import java.util.*;
import java.net.*;
import java.security.*;
import xjava.security.Cipher;

import FMG.Conductor.comm.*;
import FMG.Conductor.util.*;

import FMG.ConductorExt.security.util.*;
import FMG.ConductorExt.security.messages.*;
import FMG.ConductorExt.security.cert.*;
import FMG.ConductorExt.security.cert.Certificate;
import FMG.ConductorExt.security.chain.messages.ChainAuthenticator;

/**
 * The ChainMasterAuthenticator is the place where the public keys are
 * authenticated according to chain strategy.
 * <p>
 * There are two situations: one is for the planner to verify the public keys
 * of other conductors; the other is for a conductor to verify the public
 * key of the planner in the reverse direction.
 * <p> 
 * Each node whose public key needs verification has its own authenticator
 * collected within the ChainMasterAuthenticator object. 
 * <p>
 * When a "chain" strategy is to be enforced, a chain of certificates has
 * to be formed to verify the public key of a conductor. In a chain, the
 * first certificate certifies the public key of this conductor, and every
 * following certificate certifies the public key of the certificate server
 * in the preceding certificate. The certificate server of the last
 * certificate is the local certificate server associated with the local
 * conductor.
 * <p>
 * After a ChainMasterAuthenticator collects all the necessary authenticators,
 * the public keys will be verified. The verified public keys will be
 * maintained for later usage.
 *
 * @author Jun Li (lijun@fmg.cs.ucla.edu)
 * @version $Revision: 1.6 $
 */
public class ChainMasterAuthenticator {
    /** The boolean whether to check the status of a certificate for verification*/
    private static final boolean checkCertStatus = true;

    /** The associated ChainSecurityBox object */
    private ChainSecurityBox sbox;

    /** Collected authenticators */
    private Vector           authList           = new Vector();          

    /** Gathered certs from all authenticators */
    private CertificateList  gatheredCerts      = new CertificateList();

    /** Those public keys that are verified */
    private Hashtable        verifiedPubkeyPool = new Hashtable();

    /** Create a new ChainMasterAuthenticator associated with a ChainSecurityBox */
    public ChainMasterAuthenticator(ChainSecurityBox sbox) {
        this.sbox = sbox;
    }

    /**
     * Gets the public key of a specific conductor.
     * The verified public keys are all kept in the hashtable
     * <code>verifiedPubkeyPool</code>. If the verification is not done yet,
     * the public key verification procedure will be started.
     *
     * @param node The conductor node whose verified public key to be returned   
     * @return the public key of the conductor node
     */
    synchronized public PublicKey getVerifiedPublicKey(NodeID node) {
        return (PublicKey)verifiedPubkeyPool.get(node);
    }


    /*
     * Protected Functions
     */

    /**
     * Update this CHainMasterAuthenticator by adding a new authenticator.
     * Two things: add the authenticator itself; 
     *             add the certificates for the certificate server of 
     *             the owner of the authenticator.
     * 
     * @param auth A received ChainAuthenticator of other conductor
     * @return No return value
     */
    synchronized protected void update(ChainAuthenticator auth) {
	auth.retain();  // to avoid recycled by caller. Self-recycle later.
        addAuthenticator(auth);
        addCertificates(auth.getServerCerts());
    }

    /**
     * Generates and store those verified public keys.
     * <p>
     * Looping through each authenticator received so far, if a chain of
     * certificates can be formed, the first certificate in the chain is
     * authentic. Then the public key of the node in question is stored
     * based on this certificate.
     * <p>
     * At a planner node, this function generates a pool of verified public
     * keys; and at a non-planner node, it generates the verified public
     * key of the planner which does the key distribution.
     *
     * @return no return value.
     */
    synchronized protected void generateVerifiedPubkeys() {
        for (int i=0; i<authList.size(); i++) {
            ChainAuthenticator auth = (ChainAuthenticator)authList.elementAt(i);

            if ( Debug.on)
                Debug.println(Debug.SECURITY, Debug.MOREINFO, 
                              "[secbox] "+auth + " is being processed");

            Certificate cert = getAuthenticCertificate(auth);

            if (cert != null) {
                NodeID    node;
		try {
		    node = new NodeID(InetAddress.getByName(cert.getNode()));
		} catch ( UnknownHostException e ) {
		    throw new RuntimeException("Unknow node from cert "+cert);
		}

                PublicKey pk   = KeySerializer.getPubKey(cert.getPubKey(), 
							 sbox.getPubKeyAlgo());
		verifiedPubkeyPool.put(node, pk);
                if ( Debug.on)
                    Debug.println(Debug.SECURITY, Debug.MOREINFO, 
                                  "[secbox] pub key for "+node+" is authenticated and added");
            }
            auth.recycle();
        }
    }


    /*
     * Private Functions
     */


    /**
     * Add an authenticator into this ChainMasterAuthenticator.
     *
     * @param auth the ChainAuthenticator to be added
     * @return no return value
     */
    private void addAuthenticator(ChainAuthenticator auth) {
        if ( auth != null )
            authList.addElement(auth);
        // TODO : duplicate checking
    }
    
    /*
     *  Methods of adding more certificate to repository
     */
    
    /**
     * Add a certificate into the repository of gathered certificates
     *
     * @param cert the certificate to be added
     * @return no return value
     */
    private void addCertificate(Certificate cert) {
        gatheredCerts.addCertificate(cert);
    }

    /**
     * Add multiple certificates into the repository of gathered certificates
     *
     * @param certs the certificates to be added
     * @return no return value
     */
    private void addCertificates(Enumeration certs) {
        while (certs.hasMoreElements()) {
            gatheredCerts.addCertificate((Certificate) certs.nextElement());
        }
    }

    /**
     * Add a list of certificates into the repository of gathered certificates
     *
     * @param certs the list of certificates to be added
     * @return no return value
     */
    private void addCertificates(CertificateList certs) {
        addCertificates(certs.getAll());
    }


    /**
     * Gets the authentic certificate with respect to a ChainAuthenticator.
     * <p>
     * The authenticator owner must be trusted at first. Then, looping
     * through the certificates for the owner signed by different
     * certificate servers, if one of them can be verified as authentic,
     * this certificate is returned; otherwise, this authenticator can not
     * be verified as an authentic authentictor.
     * 
     * @param auth the ChainAuthenticator to be checked
     * @return the authentic certificate if it exists; null otherwise
     */
    private Certificate getAuthenticCertificate(ChainAuthenticator auth) {
        /* don't bother if we don't trust this host */
        if (!ChainSecurityBox.trusted(auth.getNode())) {
            if ( Debug.on)
                Debug.println(Debug.SECURITY, Debug.MOREINFO, 
                              "[secbox] "+auth.getNode()+"not trusted");
            return null;
        }
        
        Enumeration certs = auth.getNodeCerts();
        while (certs.hasMoreElements()) {
            Certificate c = (Certificate) certs.nextElement();

	    if ( Debug.on)
		Debug.println(Debug.SECURITY, Debug.MOREINFO, 
                              "[secbox] Verifying node cert "+c);
	    
	    if (verifyCertificate(c)) {
                return c;
            }
        }
        return null;
    }
    
    /** 
     * Verifies a certificate.
     * <p> 
     * If the certificate is signed by local certificate server, then
     * it is one-step verification. Otherwise, the public key of the
     * certificate server of this certificate has to be verified first,
     * which is done if at least one of the certificates of the server can
     * be verified using the same procedure recursively.
     *
     * <p> This is a recursive function.
     *
     * @param cert the certificate to be verified
     * @return <code>true</code> if the certificate is verified as
     *         authentic; <code>false</code> otherwise.
     */
    private boolean verifyCertificate(Certificate cert) {

        if ( Debug.on)
            Debug.println(Debug.SECURITY, Debug.MOREINFO, 
                          "[secbox] Verifying "+cert);

        if (! ChainSecurityBox.trusted(cert.getNode())) {
            if ( Debug.on)
                Debug.println(Debug.SECURITY, Debug.MOREINFO, 
                              "[secbox] "+cert.getNode()+" not trusted");
            return false;
        }
        
        try { 
            // The address of the local certificate server
            String    localCS   = sbox.getCertServer();
            // the public key of the local certificate server
            PublicKey localCSKey= sbox.getCertServerPublicKey();
            
            if (cert.getAuthority().equals(localCS)) {
                if ( Debug.on)
                    Debug.println(Debug.SECURITY, Debug.MOREINFO, 
                                  "[secbox] Local certificate server");

                boolean res = cert.verify(sbox.getSignAlgo(), localCSKey, checkCertStatus);

                if ( Debug.on)
                    Debug.println(Debug.SECURITY, Debug.MOREINFO, 
                                  "[secbox] Verification: " + res);

                return res;
            } else {
                Enumeration list = 
                    gatheredCerts.findCertsForNode(cert.getAuthority());
                while (list.hasMoreElements()) {
                    Certificate authCert = (Certificate) list.nextElement();
                    if (verifyCertificate(authCert)) {
                        // Get the public key contained in var[authCert]
                        PublicKey pk =
                            KeySerializer.getPubKey(authCert.getPubKey(), 
						    sbox.getPubKeyAlgo());
                        if ( cert.verify(sbox.getSignAlgo(), pk, checkCertStatus) ) 
                            return true;
                    }
                }
                return false;
            }
        } catch ( Exception e ) {
            System.err.println("ChainMasterAuthenticator.verifyCertificate()" +
                               "[" + e + "]");
            System.exit(1);
            return false;
        }
    }

    
}


