/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.ConductorExt.security.chain;

import java.util.*;
import java.io.*;
import java.security.*;
import xjava.security.Cipher;

import FMG.Conductor.comm.*;
import FMG.Conductor.util.*;
import FMG.Conductor.security.*;

import FMG.Conductor.Conductor;

import FMG.ConductorExt.security.*;
import FMG.ConductorExt.security.messages.*;
import FMG.ConductorExt.security.chain.messages.*;

public class ChainSecurityBox extends ExtensionSecurityBox {

    /** The directory where all security related configuration information resides */
    private static final String  CONFIG_DIR          = configDir();

    /** The default name for the file listing the trusted hosts */
    private static final String  DEFAULT_TRUSTED_HOSTS_FILE = CONFIG_DIR+"/trusted_hosts";

    /** The vector of trusted hosts by this node */
    private static Vector        trustedHosts  = new Vector();

    /** The class level list of properties read from the property file */
    private static Properties    props;

    /** The object doing all detailed "security" work */
    private ChainMasterAuthenticator  master;

    /**
     *  Reads the properties file into <code>props</code>, reads the file
     *  of trusted hosts, and register security messages.
     */
    static {
        props = readPropertiesFile(getPropsFilename());
        readTrustedHosts();
	registerMessages();
    }

    /**
     * Create a new chain security box object.
     * @param coordinator The coordinator for this object.
     * @param param The parameter specified by the user in the config file
     *              used to customize this instance of the class.
     */
    public ChainSecurityBox(SecurityCoordinator coordinator, String param) {
        super(coordinator, param);
        master = new ChainMasterAuthenticator(this);
        boxCharacteristics = BoxCharacteristics.getNewMessage(
                                coordinator.connectionID(),
                                "FMG.ConductorExt.security.chain.ChainSecurityBox", 
                                param);
    }

    /* 
     * Public Static Functions
     */



    /** 
     * Determine if a node is trusted
     *
     * @return <code>true</code> if trusted; <code>false</code> if not
     */
    public static boolean trusted(String node) {
        for (int i=0; i<trustedHosts.size(); i++) {
            if (((String)trustedHosts.elementAt(i)).equals(node)) {
                return true;
            }
        }
        return false;
    }


    /**
     * Indicates whether the boolean property "CertificateCacheEnabled" was set.
     *
     * @return True iff the property "CertificateCacheEnabled" was true according
     *              to the <code>getBooleanProperty()</code> method.
     */
    public static boolean certificateCacheEnabled() {
        return getBooleanProperty("CertificateCacheEnabled", true);
    }


    /**
     * Determines who is the certificate server.
     *
     * @return  The identification of the certificate server
     */
    public static String getCertServer() {
	return props.getProperty("CertServer");
    }


    /* 
     * Implementation of those abstract functions defined in ExtensionSecurityBox    
     */
    public void authentication(Authenticator authenticator, boolean isPlanner) {
	ChainAuthenticator auth;

	if ( ! (authenticator instanceof ChainAuthenticator) )
	    throw new RuntimeException("Not a ChainAuthenticator");

	auth = (ChainAuthenticator)authenticator;

	auth.updateSelf(getPubKeyAlgo(), getSignAlgo());

	if ( isPlanner || auth.flowDir().isDown() ) {
	    master.update(auth);
	    if ( auth.flowDir().isDown() )
		master.generateVerifiedPubkeys();
	} 
    }

    /* 
     * This functions signals that all authenticators have been received
     */
    public void authenticationDone( boolean isPlanner) {
	if ( !isPlanner ) 
	    return;

	master.generateVerifiedPubkeys();
    }

    public PublicKey getVerifiedPublicKey(NodeID node) {
	return master.getVerifiedPublicKey(node);
    }

    public Authenticator createNewAuthenticator(ConnID connID,
						Direction dir,
						String pkeyAlgo,
						String signAlgo) {
	return ChainAuthenticator.getNewMessage(connID, 
						dir, 
						pkeyAlgo, 
						signAlgo, 
						getCertServer());
    }
    

    /**
     * Gets the string value of a property
     * It checks the <code>pros</code> to get the value of a property.
     *
     * @param propName the name of the property
     * @return the value of the property
     */
    public String getProperty(String propName) {
        return props.getProperty(propName);
    }

    /**
     * Gets the configuration directory.
     *
     * @return The configuration direction for the chain strategy
     */
    public String configDirName() {
	return CONFIG_DIR;
    }


    /*
     * Private Static Functions
     */


    /**
     * Gets the configuration directory.
     *
     * @return a directory under current directory based on hostname of
     *         local machine
     */
    private static String configDir() {
      return Conductor.getFilenameProperty("ChainSecurityBox.ConfigDir",
                                           "./config/chainsec.cfg");
    }

    /**
     * Registers chain strategy related security messages.
     *
     * @return No return value.
     */
    private static void registerMessages() {
        ChainAuthenticator.registerSelf();
    }


    /**
     * Gets the name of the property file for the security box
     * We use the "chain.properties" under the configuration directory.
     *
     * @return the name of the property file
     */
    private static String getPropsFilename() {
        return CONFIG_DIR + "/chain.properties";
    }

    /**
     * Reads the file of trusted hosts into memory.
     *
     * @return No return   value  
     */
    private static void readTrustedHosts() {
        String filename = props.getProperty("TrustedHostsFile");
        String hostname;

	if ( filename == null )
	    filename = DEFAULT_TRUSTED_HOSTS_FILE;

        try {
            FileInputStream fis = new FileInputStream(new File(filename));
        
            while ((hostname = readHostname(fis))!=null) {
                trustedHosts.addElement(hostname);
            }
        } catch ( IOException e ) {
            System.err.println("Error when reading trusted hosts file");
            System.exit(1);
        }
    }
    
    /**
     * Returns the boolean value of a property from the security box property
     * file.
     * @param name The name of the property to test.
     * @param defaultValue The default value to return if the property is 
     *                     not defined.
     * @return If the property is defined, returns <code>true</code> iff the 
     *         property value is "true", "1", or "yes" (case insensitive).
     *         If the property is not defined, returns the value passed
     *         as defaultValue.
     */
    private static boolean getBooleanProperty(String name, 
					      boolean defaultValue) {
        String value = props.getProperty(name);
        if (value == null) {
            return defaultValue;
        } else {
            return (value.equalsIgnoreCase("true") || 
                    value.equals("1") || 
                    value.equalsIgnoreCase("yes"));
        }
    }

}
