/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.ConductorExt.security.chain.messages;

import  java.io.*;
import  java.net.*;
import  java.util.*;

import  FMG.Conductor.comm.*;
import  FMG.Conductor.util.*;

import  FMG.ConductorExt.security.cert.*;
import  FMG.ConductorExt.security.util.*;
import  FMG.ConductorExt.security.chain.*;
import  FMG.ConductorExt.security.messages.Authenticator;

/**
 * The ChainAuthenticator class implements the authenticator used by the
 * "chain" strategy. It passes necessary authentication information to
 * facilitate the authentication of the public key of the owner of the
 * authenticator. A ChainAuthenticator mainly includes the direction if
 * flows, the NodeID of the owner and the local certificate server, the
 * certificates certifying the node (the owner of the authenticator) and the
 * certificates in a separate list certifying the local certificate server.
 *
 * @author Jun Li (lijun@fmg.cs.ucla.edu)
 * @version $Revision: 1.8 $
 */ 

public class ChainAuthenticator extends Authenticator {
    /** Boolean indicating whether to cache certificates or not */
    private static boolean    certCacheEnabled = ChainSecurityBox.certificateCacheEnabled();
    
    /** The certificate client connecting to the local certificate server */
    private static CertClient cc = new CertClient(ChainSecurityBox.getCertServer());

    /** the negative cache */
    private static Hashtable  negativeCache = new Hashtable();

    /** The recycler used for messages of this type */
    private static Recycler   recycler = new Recycler(ChainAuthenticator.class);

    /** The message id assigned to messages of this type */
    private static int        messageID = -1;

    /** The NodeID of the owner of this ChainAuthenticator */
    private String            node;

    /** The NodeID of the local certificate server */
    private String            certServer;

    /** The list of certificates certifying this node */
    private CertificateList   nodeCerts;

    /** The list of certificates certifying the local certificate server */
    private CertificateList   serverCerts;

    /**
     * Register a prototype of this message class with the BasicMessageInputStream.
     * @return No return value.
     * @see FMG.Conductor.comm.BasicMessageInputStream
     */
    public static void registerSelf() {
        messageID = BasicMessageInputStream.
            registerMessagePrototype((Message) recycler.getOne());
    }

    /**
     * This method returns the message id that was assigned to messages of this
     * type by BasicMessageInputStream.
     * @return The message id for this type of message.
     * @see FMG.Conductor.comm.BasicMessageInputStream
     */
    public int getMessageID() {
        return messageID;
    }
    
    /**
     * This method returns a new (or recycled) ChainAuthenticator object 
     * initialized from the input stream.  Normally, this method is only 
     * called on the prorotype given to BasicMessageInputStream.
     * @param is The input stream to read the message's data from
     * @return a new object initialized from the data stream.
     * @exception java.io.IOException if an error occurs while reading from
     *                                the stream.
     * @see FMG.Conductor.comm.BasicMessageInputStream
     */
    public Message readNewMessage(DataInputStream is) throws IOException {
        return ChainAuthenticator.getNewMessage(is);
    }

    /**
     * Generate a new (or recycled) ChainAuthenticator message with the appropriate
     * data members initialized.
     *
     * @param connID The id of the connection for which we are planning.
     * @param dir    The direction that this message will flow
     * @param pkeyAlgo The public key algorithm to use
     * @param signAlgo The signing algorithm to use
     * @param certServer The NodeID of the certificate server
     * @return A new (or recycled) and initialized ChainAuthenticator message.
     */

    public static ChainAuthenticator getNewMessage(ConnID connID,
                                                   Direction dir,
                                                   String pkeyAlgo,
                                                   String signAlgo,
                                                   String certServer) {
        ChainAuthenticator m = (ChainAuthenticator) recycler.getOne();
        m.init(connID, dir, pkeyAlgo, signAlgo, certServer);
        return m;
    }


   /**
    * This method returns a new (or recycled) ChainAuthenticator object 
    * initialized from the input stream.
    * @param is The input stream to read the message's data from
    * @return A new (or recycled) and initialized ChainAuthenticator message.
    * @exception java.io.IOException if an error occurs while reading from
    *                                the stream.
    */
    public static ChainAuthenticator getNewMessage(DataInputStream is) 
        throws IOException {
       ChainAuthenticator m = (ChainAuthenticator) recycler.getOne();
       m.receiveSelf(is);
       return m;
    }
    
    /**
     * Read the data elements of this object from the input stream.  Presumably
     * this information was generated by <code>sendSelf()</code>.
     * @param is The input stream to read the message's data from.
     * @return No return value.
     * @exception java.io.IOException if an error occurs while reading from 
     *                                the stream
     */
    public void receiveSelf(DataInputStream is) throws IOException {
      super.receiveSelf(is);

      node        = new String(BytesReader.getFrom(is));
      certServer  = new String(BytesReader.getFrom(is));
      nodeCerts   = CertificateList.make(is);
      serverCerts = CertificateList.make(is);
    }

    /**
     * Write the data elements of this object to an output stream.  Presumably
     * the information will be read by <code>receiveSelf()</code>.
     * @param os The output stream to write the message's data to.
     * @return No return value.
     * @exception java.io.IOException if an error occurs while writing to the
     *                                stream.
     */
    public void sendSelf(DataOutputStream os) throws IOException {
       super.sendSelf(os);

       os.writeInt(node.length());
       os.write(node.getBytes());

       os.writeInt(certServer.length());
       os.write(certServer.getBytes());

       nodeCerts.writeSelf(os);

       serverCerts.writeSelf(os);
    }

    /**
     * Update this ChainAuthenticator by trying to add a certificate for the
     * node and the certificate server associated with this ChainAuthenticator,
     * respectively.
     *
     * @param pkeyAlgo the public key algorithm used for certificate retrieval
     * @param signAlgo the signing algorithm used for certificate retrieval
     */
    public void updateSelf(String pkeyAlgo, String signAlgo) {
        /*
         * If the local certificate server is same as the certificate
         * server associated with one of the certificates in the
	 * var[nodeCerts], it means this certificate server already
	 * contributed its certificate, for both the node and the
	 * certificate server of this ChainAuthenticator, so simply return;
         */
        if (nodeCerts.containsCertBy(ChainSecurityBox.getCertServer()))
            return;
        // Add more certificates for the owner and its associated cert server
	addNodeCert(getCertificate(this.node, pkeyAlgo, signAlgo));
	addServerCert(getCertificate(this.certServer, pkeyAlgo, signAlgo));
	
	// if reverse direction, add more certificates potentially useful
	if ( direction.isDown() ) {
	    Enumeration allCerts=getAllCerts();
	    while ( allCerts.hasMoreElements() ) {
		Certificate c = (Certificate) allCerts.nextElement();
		if ( serverCerts.findCert(c.getAuthority(),
					  ChainSecurityBox.getCertServer()) == null )
		    addServerCert(getCertificate(c.getAuthority(), pkeyAlgo, signAlgo));
	    }
	}
    }

    /**
     * Gets the node associated with this ChainAuthenticator.
     *
     * @return the node associated with this ChainAuthenticator
     */
    public String getNode() {
        return node;
    }

    /** Gets the certificates certifying the node associated with this ChainAuthenticator
     *
     * @return the enumeration of those certificates for the associated node
     */
    public Enumeration getNodeCerts() {
        return nodeCerts.getAll();
    }
    
    /** Gets the certificates certifying the certificate server associated
     *  with this ChainAuthenticator
     *
     * @return the enumeration of those certificates for the associated
     *         certificate server
     */
    public Enumeration getServerCerts() {
        return serverCerts.getAll();
    }

    public Enumeration getAllCerts() {
	Vector allCerts = new Vector();
	
        Enumeration certs = getNodeCerts();
        while (certs.hasMoreElements())
	    allCerts.addElement(certs.nextElement());
 
        certs = getServerCerts();
        while (certs.hasMoreElements())
	    allCerts.addElement(certs.nextElement());
 
	return allCerts.elements();
    }

    /**
     * Gets the string representation of this ChainAuthenticator
     * 
     * @return the string representation of this ChainAuthenticator
     */    
    public String toString() {
        return "ChainAuthenticator("+node+" BY "+certServer+")";
    }

    
    /**
     * Perform cleanup before allowing recycling to occur.  This method
     * is called by the Recycler. 
     * @return No return value.
     * @see FMG.Conductor.util.Recycler
     */
    public void recycling() {
        super.recycling();
        nodeCerts.recycle();
        serverCerts.recycle();
    }

    /*
     * private functions
     */

    /**
     * Initializes a ChainAuthenticator message.
     * <p>
     * It assigns the connection ID, the flow direction, the NodeID of the
     * node and its local certificate server. In particular, a certificate
     * for this node by the local certificate server will be retrieved, and
     * added to the list keeping the certificates for the node.
     *
     * @param connID  The id of the connection for which we are planning.
     * @param dir the direction that this ChainAuthenticator will flow
     * @param pkeyAlgo The public key algorithm to use
     * @param signAlgo The signing algorithm to use
     * @param certServer The NodeID of the certificate server
     * @return no return value.
     */
    private void init(ConnID connID, 
                      Direction dir, 
                      String pkeyAlgo,
                      String signAlgo, 
                      String certServer) {
	super.init(connID, dir);

	this.node        = localHostName();
        this.certServer  = new String(certServer);

        this.nodeCerts   = CertificateList.create();
        this.serverCerts = CertificateList.create();
        
	addNodeCert(getCertificate(this.node, pkeyAlgo, signAlgo));
    }

    /** 
     * Adds a new certificate for the node associated with this ChainAuthenticator
     * 
     * @param cert the new certificate for the associated node
     * @return no return value
     */
    private void addNodeCert(Certificate cert) {
       if (cert != null) {
	   nodeCerts.addCertificate(cert);
       }
    }
  
    /** 
     * Adds a new certificate for the certificate server associated with
     * this ChainAuthenticator.
     * 
     * @param cert the new certificate for the associated certificate server
     * @return no return value
     */
    private void addServerCert(Certificate cert) {
       if (cert != null) {
           serverCerts.addCertificate(cert);
       }
    }


    private Certificate getCertificate(String aNode, 
				       String pkeyAlgo, 
				       String signAlgo) {
	Certificate cert = getLocalCertificate(aNode, pkeyAlgo, signAlgo);

	if ( cert == null ) {
	    if ( unavailable(aNode, pkeyAlgo, signAlgo) ) {
		return null;
	    }
	}

	if ( cert == null ) {
	    try {
		CertRequest req = CertRequest.make(aNode, pkeyAlgo, signAlgo);
		cert = cc.getCertificate(req);
	    } catch (IOException e) {
		throw new RuntimeException("ChainAuthenticator: Certificate I/O Error" + e);
	    }

	    if ( cert != null ) {
		addLocalCertificate(cert, pkeyAlgo, signAlgo);
	    } else {
		addNegativeCache(aNode, pkeyAlgo, signAlgo);
	    }
	}

	if ( cert != null )
	    if ( certCacheEnabled ||
		 cert.getNode().equals(localHostName()) )
	    cert.retain();

	return cert;
    }


    private static String localHostName() {
	try {
	    return InetAddress.getLocalHost().getHostName();
	} catch ( UnknownHostException e ) {
	    throw new RuntimeException("Unknown host exception");
	}
    }
    
    synchronized private static boolean unavailable(String theNode, 
						    String pkeyAlgo, 
						    String signAlgo) {
	String format = new String(pkeyAlgo+signAlgo);
	Vector list;

	if ( ! certCacheEnabled )
	    return false;
       
	list =  (Vector)negativeCache.get(format);
	if ( list == null )
	    return false;

        Enumeration nodes = list.elements();
        
        while (nodes.hasMoreElements()) {
            String aNode = (String) nodes.nextElement();
            if (aNode.equals(theNode)) {
		return true;
            }
        }

	return false;
    }


    synchronized private static void addNegativeCache(String aNode, 
						      String pkeyAlgo, 
						      String signAlgo) {
	String format = new String(pkeyAlgo+signAlgo);
	Vector list;

	if ( ! certCacheEnabled )
	    return;

	list =  (Vector)negativeCache.get(format);
	if ( list == null ) {
	    list = new Vector();
	    negativeCache.put(format, list);
	}
	list.addElement(aNode);
    }

    synchronized private static Certificate getLocalCertificate(String principal,
								String pkeyAlgo,
								String signAlgo) {
	String          format;
	CertificateList list;
	Certificate     cert   = null;

	if ( ! certCacheEnabled && ! principal.equals(localHostName()) ) 
	    return null;

	format = new String(pkeyAlgo+signAlgo);
	list   = (CertificateList)localCerts.get(format);

	if ( list != null ) {
	    cert = list.findCertForNode(principal);
	}
	return cert;
    }

    synchronized private static void addLocalCertificate(Certificate cert,
							 String pkeyAlgo,
							 String signAlgo) {
	String          format; 
	CertificateList list; 

	/*
	 * By not adding any certificates at all, cache is useless. This
	 * way the cache is diabled. All other code unchanged.
	 */
	if ( ! certCacheEnabled && ! cert.getNode().equals(localHostName()) ) 
	    return;

	/*
	 *  Assuming a conductor node can not be its own certificate server;
	 *  Otherwise, the cache may contain two certificates with same principal,
	 *  A(conductor) by A(CS), and A(CS) by B(next level CS).
	 */
	if ( getLocalCertificate(cert.getNode(), pkeyAlgo, signAlgo) != null ) {
	    throw new RuntimeException("Warning: certificate cache already contains the certificate to add");
	}
	
	format = new String(pkeyAlgo+signAlgo);
	list   = (CertificateList)localCerts.get(format);
	
	if ( list == null ) {
	    list = CertificateList.create();
	    localCerts.put(format, list);
	}
	list.addCertificate(cert);
    }

    synchronized private static void printLocalCerts() {
	Enumeration all = localCerts.elements();
	while ( all.hasMoreElements() ) {
	    CertificateList cl = (CertificateList)all.nextElement();
	    System.out.println("============================\n"+cl);
	}
    }

                                                                   
}
