/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.ConductorExt.security.hier;

import java.io.*;
import java.util.*;
import java.net.*;
import java.security.*;
import xjava.security.Cipher;

import FMG.Conductor.planning.messages.*;
import FMG.Conductor.security.*;
import FMG.Conductor.comm.*;
import FMG.Conductor.util.*;
import FMG.Conductor.dataflow.*;

import FMG.ConductorExt.security.messages.*;
import FMG.ConductorExt.security.util.*;
import FMG.ConductorExt.security.cert.*;
import FMG.ConductorExt.security.hier.messages.*;

import FMG.Conductor.Conductor;

/**
 * The HierSecurityBox is a security box implementation that provides
 * "hier" style security. That is, the public key of each conductor
 * node is verified by the planner node, if a hierarchy of certificates can
 * be formed. In the reverse direction, the public key of the planner is
 * verified by other conductors by the "hier" strategy too.
 * <p>
 * The HierSecurityBox handles incoming and outgoing messages. See 
 * below for more details.
 * <p>
 * The client can specify more parameters for the hier strategy, including
 * the public key algorithm and the signing algorithm;
 *
 * @author Jun Li (lijun@fmg.cs.ucla.edu)
 * @version $Revision: 1.9 $
 */

public class HierSecurityBox extends SecurityBox {
    /** TODO: The directory where all security related configuration information resides */
    private static final String  CONFIG_DIR          = configDirName();

    /** The default name for the public key algorithm */
    private static final String  DEFAULT_PKEY_ALGO   = "RSA";

    /** The default name for the signing algorithm */
    private static final String  DEFAULT_SIGN_ALGO   = "SHA-1/RSA/PKCS#1";

    /** The default name for the cipher  algorithm */
    private static final String  DEFAULT_CRYP_ALGO   = "RSA/ECB/OneAndZeroes";

    /** The default name for the file listing the trusted hosts */
    private static final String  DEFAULT_TRUSTED_HOSTS_FILE = CONFIG_DIR+"/trusted_hosts";

    /** The vector of trusted hosts by this node */
    private static Vector        trustedHosts  = new Vector();

    /** The class level list of properties read from the property file */
    private static Properties    props  = new Properties();

    /** The instance level list of properties propagated from the client */
    private        Properties    params = new Properties();

    /** The object doing all detailed "security" work */
    private MasterAuthenticator  master;

    /** The random number to prevent replay attack */
    private int                  random;

    /** The certificate client connecting to a certificate server */
    private CertClient           cc;

    /** The buffer of non-SignInfo planning messages */
    private Vector               pbuffer  = new Vector();

    /** The buffer of SignInfo messages */
    private Vector               sbuffer  = new Vector();

    /** The object describing the characteristics of this security box  */
    private BoxCharacteristics   boxCharacteristics;
    
    /**
     *  Reads the properties file into <code>props</code>, reads the file
     *  of trusted hosts, and register all security messages.
     */
    static {
        readPropertiesFile(getPropsFilename());
        readTrustedHosts();
        registerMessages();
    }

    /**
     * Determines who is the certificate server.
     *
     * @return  The identification of the certificate server
     */
    public static NodeID     getCertServer() {
        InetAddress addr = null;
        try {
            addr = InetAddress.getByName(getProperty("CertServer"));
        } catch ( UnknownHostException e ) {
            System.err.println("Unknown cert server address");
            System.exit(1);
            return null;
        }
        
        return new NodeID(addr);
    }

    /**
     * Determines who is the root of the certificate hierarchy.
     *
     * @return  The identification of the root certificate server
     */
    public static NodeID     getRootCertServer() {
        InetAddress addr = null;
        try {
            addr = InetAddress.getByName(getProperty("RootCertServer"));
        } catch ( UnknownHostException e ) {
            System.err.println("Unknown root cert server address");
            System.exit(1);
            return null;
        }
        
        return new NodeID(addr);
    }

    /**
     * Indicates whether the boolean property "CertificateCacheEnabled" was set.
     *
     * @return True iff the property "CertificateCacheEnabled" was true according
     *              to the <code>getBooleanProperty()</code> method.
     */
    public static boolean certificateCacheEnabled() {
        return getBooleanProperty("CertificateCacheEnabled", true);
    }

    /** 
     * Determine if a node is trusted
     *
     * @return <code>true</code> if trusted; <code>false</code> if not
     */
    public static boolean trusted(NodeID node) {
        for (int i=0; i<trustedHosts.size(); i++) {
            if (((InetAddress)trustedHosts.elementAt(i)).equals(node.getAddress())) {
                return true;
            }
        }

        return false;
    }

    /**
     * Create a new hier security box object.
     * @param coordinator The coordinator for this object.
     * @param param The parameter specified by the user in the config file
     *              used to customize this instance of the class.
     */
    public HierSecurityBox(SecurityCoordinator coordinator, String param) {
        super(coordinator, param);
        setParameters(param);
        master = new MasterAuthenticator(this);
        random = (new SecureRandom()).nextInt();
        cc     = new CertClient(getCertServer());
        boxCharacteristics = BoxCharacteristics.getNewMessage(
                                coordinator.connectionID(),
                                "FMG.ConductorExt.security.hier.HierSecurityBox", 
                                param);
    }
    
    /**
     * Determines the public algorithm name used by the security box.
     * If not specified by the client, nor by the property file, the
     * default will be returned.
     *
     * @return  The public algorithm name used by the security box.
     */
    public String     getPubKeyAlgo() {
        return getParameter("PublicKeyAlgo", DEFAULT_PKEY_ALGO);
    }
    
    /**
     * Determines the signing algorithm name used by the security box.
     * If not specified by the client, nor by the property file, the
     * default will be returned.
     *
     * @return  The signing algorithm name used by the security box.
     */
    public String     getSignAlgo() {
        return getParameter("SignAlgo", DEFAULT_SIGN_ALGO);
    }

    /**
     * Determines the cipher algorithm name used by the security box.
     * If not specified by the client, nor by the property file, the
     * default will be returned.
     *
     * @return  The cipher algorithm name used by the security box.
     */
    public String     getCipherAlgo() {
        return getParameter("CipherAlgo", DEFAULT_CRYP_ALGO);
    }

    /**
     * Determines the private key of the node itself.
     * The private key is in terms of the public key algorithm used hereof.
     *
     * @return  The local private key
     */
    public PrivateKey getOwnPrivateKey() {
        return (PrivateKey)getKeyProperty("OwnPrivateKey");
    }

    /**
     * Determines the public key of the certificate server, following the
     * public key algorithm used hereof 
     *
     * @return the public key of the certificate server
     */
    public PublicKey  getCertServerPublicKey() {
        return (PublicKey)getKeyProperty("CertServerPublicKey");
    }

    /**
     * Determines the public key of the root certificate server, following the
     * public key algorithm used hereof 
     *
     * @return the public key of the certificate server
     */
    public PublicKey  getRootCertServerPublicKey() {
        return (PublicKey)getKeyProperty("RootCertServerPublicKey");
    }
    
    /**
     * Handle messages that come in from the network.
     * <p>
     * There are three cases to handle with:
     * <ul>
     *    <li>  this security box on a non-planner node and the message is
     *          towards the server
     *    <li>  this security box on a non-planner node and the message is
     *          towards the client
     *    <li>  this security box on the planner node
     *</ul>
     *
     * @param m The message that has been received from the network.
     * @return No return value.
     */
    public void handleIncoming(PlanningMessage m) {
        if (Debug.on)
            Debug.println(Debug.SECURITY, Debug.MOREINFO, 
                          "[secbox] handleIncoming(" + m + ") ");
        }

        // non-planner node,  m is towards the server
        if ( ! coordinator.isPlannerNode() && m.flowDir().isUp()) {
            if ( m.handledByPath() )
                coordinator.deliverToPath(m);  // deliver to Path objet
            else
                coordinator.sendSToNet(m);     // send to network directly              
            /*
             * The above might cause the message delivery order
             * crossed. For instance, the further processing by Path
             * object of PlanningInfo may cause it out of this node to
             * network AFTER its corresponding SignInfo message.
             */

            if (Debug.on) {
                Debug.println(Debug.SECURITY, Debug.MOREINFO, "[secbox] done");
            }
            
            return;
        }
        
        // non-planner node,  m is towards the client
        if ( ! coordinator.isPlannerNode() && m.flowDir().isDown()) {

            if ( m instanceof HierAuthenticator ) {   // the authenticator of the planner
                master.update((HierAuthenticator)m);
                if ( ! coordinator.isRootNode() ) {
                    coordinator.sendCToNet(m);
                } else {
                    m.recycle();
                }
            } else if ( m instanceof BoxCharacteristics ||
                        m instanceof KeyDistro          ||
                        m instanceof PlanDistro         ||
                        m instanceof SignInfo   ) {
                hold(m); 
            } else {
                // For other message types, deliver it to Path as is
                coordinator.deliverToPath(m);   
            }

            if (Debug.on) {
                Debug.println(Debug.SECURITY, Debug.MOREINFO, "[secbox] done");
            }

            return;
        } 
        
        // This is the planner.
        if ( m instanceof HierAuthenticator ) {
            master.update((HierAuthenticator)m);
        } else if ( m instanceof PlanningInfo || m instanceof SignInfo ) {
            hold(m);
        } else {
            // For other messages, including PlanningDone, simply forward to Path as is
            coordinator.deliverToPath(m);
        }

        if (Debug.on) {
            Debug.println(Debug.SECURITY, Debug.MOREINFO, "[secbox] done");
        }

        return;
    }
    
    /**
     * Handle messages that the path object is trying to send over the
     * network toward the server.  In particular, if the message is of
     * <code>PlanningInfo</code> from the local node, two aditional
     * security messages will be sent out to the network towards the
     * server:
     * <ul>
     *    <li> the <code>SignInfo</code> message containing the signature 
     *         of the <code>PlanningInfo</code> message, and 
     *    <li> the authenticator of the local node
     * </ul>
     * All the other messages will be transmitted as-is.
     *
     * @param m The message that is to be transmitted.
     * @return False if a failure occurs while sending.  True if it appears
     *         that transmission was successful (not guaranteed).
     */
    public boolean handleOutgoingToServer(PlanningMessage m) {
        
        if (Debug.on) {
            Debug.println(Debug.SECURITY, Debug.MOREINFO,  
                          "[secbox] handleOutgoingToServer(" + m + ")");
        }

        if ( m instanceof PlanningInfo && ((PlanningInfo)m).node.isLocalNode()) {
            SignInfo           sig  = getNewSignInfo(m, Direction.up);
            HierAuthenticator auth = HierAuthenticator.getNewMessage(coordinator.connectionID(),
                                                                     Direction.up,
                                                                     getPubKeyAlgo(),
                                                                     getSignAlgo(),
                                                                     cc,
                                                                     random);
            
            // send authenticator, signature of own PlanningInfo, and the PlanningInfo
            if ( ! coordinator.sendSToNet(auth)  ||
                 ! coordinator.sendSToNet(m)   ||
                 ! coordinator.sendSToNet(sig)       )    {
                if ( Debug.on)
                    Debug.println(Debug.SECURITY, Debug.MOREINFO,
                                  "[secbox] failed");
                //TODO: the case when successful in sending var[auth, m] but not var[sig]
                return false;
            } else {
                if ( Debug.on && (Debug.getLevel(Debug.SECURITY) > 
                                                       Debug.MOREINFO )) {
                    Debug.println("[secbox]  successfully sent out to the wire towards server:" );
                    Debug.println(m+"\n"+sig+"\n"+auth);
                }
            }
        } else {
            if ( ! coordinator.sendSToNet(m) ) {
                if ( Debug.on) {
                    Debug.println(Debug.SECURITY, Debug.MOREINFO, 
                                  "[secbox] failed");
                return false;
            }
            if ( Debug.on)
                Debug.println(Debug.SECURITY, Debug.MOREINFO,
                              "[secbox]  successfully sent out to the wire towards server:\n"+m );
        }

        if ( Debug.on)
            Debug.println(Debug.SECURITY, Debug.MOREINFO, "[secbox] finished");
        
        return true;
    }
    
    /**
     * Handle messages that the path object is trying to send over the
     * network toward the client.  In particular, all messages will
     * be transmitted as-is, except when the planner is transmitting the
     * plan towards the client.
     * <p>
     * In case that the planner is distributing the plan, the following
     * additional messages will be sent out towards the clients, besides
     * the plan itself as-is:
     * <ul>
     *    <li> the authenticator of the planner
     *    <li> the security box description message
     *    <li> the key distribution message
     *    <li> the signing information message containing the signature of the
     *         plan or the key distribution message or the security box
     *         description message. Each such message immediately follows
     *         the message it signs. 
     * </ul>
     *
     * @param m The message that is to be transmitted.
     * @return False if a failure occurs while sending.  True if it appears
     *         that transmission was successful (not guaranteed).
     */
    public boolean handleOutgoingToClient(PlanningMessage m) {
        if (Debug.on) {
            Debug.println(Debug.SECURITY, Debug.MOREINFO,
                          "[secbox] handleOutgoingToClient(" + m + ")");
        }

        if (  m instanceof PlanDistro && coordinator.isPlannerNode() ) {
            
            HierAuthenticator auth;
            SignInfo           sig;
            boolean            success;

            auth = HierAuthenticator.getNewMessage(coordinator.connectionID(),
                                                    Direction.down,
                                                    getPubKeyAlgo(),
                                                    getSignAlgo(),
                                                    cc,
                                                    0);// no random needed for reverse direction
            if ( !coordinator.sendCToNet(auth) ) {
                if ( Debug.on)
                    Debug.println(Debug.SECURITY, Debug.MOREINFO, 
                                  "[secbox] failed");
                return false;
            }

            if ( !coordinator.sendCToNet(boxCharacteristics) ) {
                if ( Debug.on )
                    Debug.println(Debug.SECURITY, Debug.MOREINFO, 
                                  "[secbox] failed");
                return false;
            }
            
            sig = getNewSignInfo(boxCharacteristics, Direction.down);
            if ( !coordinator.sendCToNet(sig)) {
                if ( Debug.on)
                    Debug.println(Debug.SECURITY, Debug.MOREINFO, 
                                  "[secbox] failed");
                return false;
            }

            Enumeration sessionKeys = getRequiredSessionKeys((PlanDistro)m);
            while ( sessionKeys.hasMoreElements() ) {
                SessionKey  sk = (SessionKey)sessionKeys.nextElement();
                Enumeration nodes = sk.getNodes();
                while ( nodes.hasMoreElements() ) {
                    NodeID node = (NodeID)nodes.nextElement();
                    if ( Debug.on)
                        Debug.println(Debug.SECURITY, Debug.MOREINFO,
                                      "[secbox] prepare key for "+node);
                    if ( node.isLocalNode() ) {
                        // This is the planner node itself
                        Debug.println("[secbox] this is the local node itself");
                        getStreamIAC().store(sk.getKeyLabel(), sk.getKey());
                        continue;
                    }
                    KeyDistro kd = master.getKeyDistro(node,
                                                       sk.getKeyLabel(),
                                                       sk.getKey(),
                                                       coordinator.connectionID());
                    if ( kd != null ) {
                        if ( !coordinator.sendCToNet(kd)) {
                            if ( Debug.on) 
                                Debug.println(Debug.SECURITY, Debug.MOREINFO,
                                              "[secbox] failed");
                            return false;
                        }

                        sig = getNewSignInfo(kd, Direction.down);
                        if ( !coordinator.sendCToNet(sig) ) {
                            if ( Debug.on) 
                                Debug.println(Debug.SECURITY, Debug.MOREINFO,
                                                    "[secbox] failed");
                            return false;
                        }
                    } else {
                        System.out.println("Warning: no key distribution message formed for node "+node);
                    }
                }
            }
            
            if ( !coordinator.sendCToNet(m) ) {
                if ( Debug.on)
                    Debug.println(Debug.SECURITY, Debug.MOREINFO, 
                                  "[secbox] failed");
                return false;
            }

            sig = getNewSignInfo(m, Direction.down);
            if ( !coordinator.sendCToNet(sig) ) {
                if ( Debug.on)
                    Debug.println(Debug.SECURITY, Debug.MOREINFO, 
                                  "[secbox] failed");
                return false;
            }
            
            if ( Debug.on)
                Debug.println(Debug.SECURITY, Debug.MOREINFO, 
                              "[secbox] finished");
            return true;
        } else {
            // just send to network towards the client if not a plan
            boolean success = coordinator.sendCToNet(m);  
            if ( Debug.on)
                if ( success)
                    Debug.println(Debug.SECURITY, Debug.MOREINFO, 
                                  "[secbox] finished");
                else
                    Debug.println(Debug.SECURITY, Debug.MOREINFO, 
                                  "[secbox] failed");

            return success;
        } 
        
    }


    /*
     * Private functions
     */
    
    /**
     * Registers security messages.
     * The following three security messages needs to be registered:
     * <ul>
     *    <li> <code>SignInfo</code>
     *    <li> <code>KeyDistro</code>
     *    <li> <code>HierAuthenticator</code>
     *    <li> <code>BoxCharacteristics</code>
     * </ul>
     *
     * @return No return value.
     */
    private static void registerMessages() {
        SignInfo.registerSelf();
        KeyDistro.registerSelf();
        HierAuthenticator.registerSelf();
        BoxCharacteristics.registerSelf();
    }

    /**
     * Gets the configuration directory.
     *
     * @return a directory under current directory based on hostname of
     *         local machine
     */
    private static String configDirName() {
      // Get the file name from the properties
      return Conductor.getFilenameProperty("HierSecurityBox.ConfigDir", 
                                           "./config/hiersec.cfg");
    }
    
    /**
     * Gets the name of the property file for the security box
     * We use the "hier.properties" under the configuration directory.
     *
     * @return the name of the property file
     */
    private static String getPropsFilename() {
        return CONFIG_DIR + "/hier.properties";
    }

    /**
     * Reads the property file to set <code>props</code> variable
     *
     * @return No return value
     */
    private static void readPropertiesFile(String filename) {
        /* if the user specified a null filename, don't read any properties */
        if (filename.length() == 0) {
            return;
      }
        
        /* read the properties file */
        try {
            File propFile = new File(filename);
            FileInputStream fis = new FileInputStream(propFile);
            props.load(fis);
        } catch (IOException e) { 
            // Verbosity is not set yet, so just go ahead and warn the user
            System.err.println("Warning: error reading properties file (" + 
                               e + ")");
        }
    }

    /**
     * Reads the file of trusted hosts into memory.
     *
     * @return No return value
     */
    private static void readTrustedHosts() {
        String filename = getProperty("TrustedHostsFile", DEFAULT_TRUSTED_HOSTS_FILE);
        String hostname;
        try {
            FileInputStream fis = new FileInputStream(new File(filename));
        
            while ((hostname = readHostname(fis))!=null) {
                trustedHosts.addElement(InetAddress.getByName(hostname));
            }
        } catch ( IOException e ) {
            System.err.println("Error when reading trusted hosts file");
            System.exit(1);
        }
    }

    /* Reads the name of a host from an input stream.
     * Each line contains the name of one host. And only one line of the
     * stream is read.
     *
     * @param is the input stream
     * @return a String of the first line from the stream
     */
    private static String readHostname(InputStream is) throws IOException {
        byte[] buf = new byte[100];
        int len = 0;
        int b;
        
        b = is.read();
        
        if (b == -1) {
            return null; 
        }
        
        while ((b!='\n') && (b != -1)) { 
            buf[len++]=(byte)b;
            b = is.read();
        }  
        
        return new String(buf,0,len);
    }  

    /**
     * Gets the string value of a property
     * It checks the <code>pros</code> to get the value of a property.
     *
     * @param propName the name of the property
     * @return the value of the property
     */
    private static String getProperty(String propName) {
        return props.getProperty(propName);
    }

    /**
     * Gets the string value of a property
     * It checks the <code>pros</code> only to get the value of a property. If
     * null is returned after this, the default will be returned.
     *
     * @param propName the name of the property
     * @param defaultValue the default value of the property
     * @return the value of the property
     */
    private static String getProperty(String propName, String defaultValue) {
        String value = props.getProperty(propName);
        if ( value == null ) 
            value = defaultValue;

        return value;
    }


    /**
     * Returns the boolean value of a property from the security box property
     * file.
     * @param name The name of the property to test.
     * @param defaultValue The default value to return if the property is 
     *                     not defined.
     * @return If the property is defined, returns <code>true</code> iff the 
     *         property value is "true", "1", or "yes" (case insensitive).
     *         If the property is not defined, returns the value passed
     *         as defaultValue.
     */
    private static boolean getBooleanProperty(String name, 
                                              boolean defaultValue) {
        String value = getProperty(name);
        if (value == null) {
            return defaultValue;
        } else {
            return (value.equalsIgnoreCase("true") || 
                    value.equals("1") || 
                    value.equalsIgnoreCase("yes"));
        }
    }

    
    /**
     * Sets the <code>params</code> variable.
     * The format of paramString is "A=a B=b C=c D", where A, B, C, D, etc
     * are names of the parameters and a, b, c, "true", etc, are values,
     * respectively.
     *
     * @return No return value
     */

    private void setParameters(String paramString) {
        int from  = 0;
        int index = paramString.indexOf(' ');
        
        while ( index != -1 ) {
            updateParameter(paramString.substring(from, index));
            from  = index +1;
            index = paramString.indexOf(' ', from);
        }
        updateParameter(paramString.substring(from));
    }

    
    /**
     * Updates the value of a property using the given string.  
     * The string should either be the name of a property or in the form
     * "propname=value". If only the name of a property is given (ie, no
     * equals sign), then the value is set to "true".
     *
     * @param propString A string containing a property name or both a 
     *                   property name and value.
     * @return No return value
     */
    private void updateParameter(String propString) {
        int index = propString.indexOf('=');
        
        if (index == -1) {
            params.put(propString, "true");
        } else {
            String property = propString.substring(0, index);
            String value    = propString.substring(index + 1, propString.length());
            params.put(property, value);
        }
    }

    /**
     * Gets the string value of the property
     * <p>
     * It checks the <code>params</code> first to see if a value can be
     * returned; otherwise check the <code>props</code>. The default will be
     * returned if nothing can be returned from both. This way the
     * property with same name from <code>params</code> overwrites
     * <code>props</code>.
     * <p> 
     * The <code>props</code> contains class level properties read from local
     * property file, while the <code>params</code> contains instance level
     * properties propagated from client. 
     */
    private String getParameter(String paraName, String defaultValue) {
        String value = params.getProperty(paraName);
        if ( value == null )
            value = props.getProperty(paraName);
        if ( value == null ) 
            value = defaultValue;

        return value;
    }
    
    /**
     * Gets the private key of its own or the public key of the
     * certificate server, in terms of the public key type.
     * <p> 
     * It will first determine the name of corresponding property,
     * such as <code>RSAOwnPrivateKey</code> or
     * <code>RSACertServerPublicKey</code>, then from the property value
     * the filename stores the key can be determined.
     *   
     * @param  whosewhat "OwnPrivateKey" or "CertServerPublicKey" or "RootCertServerPublicKey"
     * @return the key as requested.
     */
    private Key getKeyProperty(String whosewhat) {
        String keyType = getPubKeyAlgo();
        String propName = new String(keyType + whosewhat);
        String keyFileName = CONFIG_DIR + "/" + getProperty(propName);

        Key    ret = null ;
        if ( keyFileName == null ) {
            return null;
        } 

        try {
            FileInputStream fis = new FileInputStream(new File(keyFileName));
            if ( whosewhat.equals("OwnPrivateKey") )
                ret = KeySerializer.readPrivKey(fis, keyType); 
            else if ( whosewhat.equals("CertServerPublicKey") ||
                      whosewhat.equals("RootCertServerPublicKey"))
                ret = KeySerializer.readPubKey(fis, keyType);
            else 
                System.err.println("ERROR");
            fis.close();
        } catch (IOException e) {
            System.err.println("[secbox] Error reading key file: "+keyFileName+": "+e);
            System.exit(1);
            return null;
        }

        return ret;
    }


    /**
     * Gets a <code>SignInfo</code> message for an object.
     *
     * @param o the object to be signed
     * @param dir the direction the <code>SignInfo</code> message will go
     * @return the <code>SignInfo</code> messages for the object with
     *         specific direction.
     */
    private SignInfo   getNewSignInfo(Object o, Direction dir) {
        return  SignInfo.getNewMessage(coordinator.connectionID(),
                                       dir,
                                       o,
                                       getParameter("SignAlgo", DEFAULT_SIGN_ALGO), 
                                       getOwnPrivateKey());
    }

    /** 
     * Holds a planning message in buffer; and Releases a message if possible.
     *
     * @return No return value
     */
    private void hold( PlanningMessage m ) {
        if ( m instanceof SignInfo )
            sbuffer.addElement(m);
        else
            pbuffer.addElement(m);

        if ( sbuffer.size() >= 1 && pbuffer.size() >= 1 ) {
            SignInfo       sig =        (SignInfo)sbuffer.elementAt(0);
            PlanningMessage pm = (PlanningMessage)pbuffer.elementAt(0);
            sbuffer.removeElementAt(0);
            pbuffer.removeElementAt(0);
            releaseBy(pm, sig);
        }
    }

    /**
     * Releases the buffered message in var[pbuffer] at index 0, with the given
     * <code>SignInfo</code> message.
     * <p>
     * The message could be a <code>PlanningInfo</code>, <code>PlanDistro</code>,
     * <code>KeyDistro</code> or <code>BoxCharacteristics</code>.
     * <ul>
     *    <li> <code>PlanningInfo</code><br>
     *         <table width="90%" aligh=center><tr><td>
     *         The <code>PlanningInfo</code> message will be delivered to
     *         <code>Path</code> object if verified, and dropped on the floor
     *         otherwise.
     *         </td></tr></table>
     *    <li> <code>PlanDistro</code><br>
     *         <table width="90%" aligh=center><tr><td>
     *         The <code>PlanDistro</code> message will be delivered to
     *         <code>Path</code> object if verified, and dropped on the floor
     *         otherwise.
     *         </td></tr></table>
     *    <li> <code>KeyDistro</code><br>
     *         <table width="90%" aligh=center><tr><td>
     *         The random number will be compared and the verification
     *         will be done first. If not successful, the
     *         <code>KeyDistro</code> will be dropped on the
     *         floor. Otherwise, the distributed key will be stored in the
     *         IAC(@see FMG.Conductor.dataflow.IAC) object.
     *         </td></tr></table>
     *    <li> <code>BoxCharacteristics</code><br>
     *         <table width="90%" aligh=center><tr><td>
     *         The <code>BoxCharacteristics</code> message will be sent to
     *         the network towards the client, if it is consistent with the security
     *         box running on local node and its signature is verified; otherwise,
     *         it will be dropped on the floor.
     *         </td></tr></table>
     * </ul>
     * <p> 
     *  TODO: The bad thing is that a man in the middle can forge a SignInfo to
     *  cause legal messages held in cache be purged.
     *
     * @return No return value
     */
    private void releaseBy(PlanningMessage m, SignInfo si) {

        NodeID theNode = ( m instanceof PlanningInfo ) ? ((PlanningInfo)m).node 
                                                       : coordinator.targetNode();
        PublicKey    pk = master.getVerifiedPublicKey(theNode);

        if ( pk == null ) {
            System.out.println("Warning: The public key of "+theNode+" can not be verified");
            m.recycle();
            si.recycle();
        } else if ( m instanceof PlanningInfo ) {
            PlanningInfo pi = (PlanningInfo)m;
            if ( si.verify(pi, getSignAlgo(), pk ) ) {
                coordinator.deliverToPath(pi);
            } else {
                pi.recycle();
            }
            si.recycle();
        } else if ( m instanceof BoxCharacteristics ) {
            BoxCharacteristics bc = (BoxCharacteristics)m;
            if ( this.boxCharacteristics.equals(bc) && si.verify(bc, getSignAlgo(), pk) ) {
                if ( ! coordinator.isRootNode() ) {
                    coordinator.sendCToNet(bc);
                    coordinator.sendCToNet(si);
                } else {
                    bc.recycle();
                    si.recycle();
                }
                if ( Debug.on)
                    Debug.println(Debug.SECURITY, Debug.MOREINFO, 
                                  "[secbox] Consistent security box is deployed at planner node");
                }
            } else {
                throw new RuntimeException("[secbox] The planner is perhaps using an inconsistent security box");
            }
        } else if ( m instanceof PlanDistro ) {
            PlanDistro pd = (PlanDistro)m;
            if ( si.verify(pd, getSignAlgo(), pk) ) {
                coordinator.deliverToPath(pd);
                if (  ! coordinator.isRootNode() ) {
                    coordinator.sendCToNet(si); 
                } else {
                    si.recycle();
                }
            } else {
                pd.recycle();
                si.recycle();
            }
        } else if ( m instanceof KeyDistro ) {
            KeyDistro kd = (KeyDistro)m;
            if ( ! kd.getNode().equals(NodeID.getNewLocalID()) ) {
                // the message is not for this node
                if ( ! coordinator.isRootNode() ) {
                    coordinator.sendCToNet(kd);  // forwards the key distro as is to the client
                    coordinator.sendCToNet(si);
                } else {
                    throw new RuntimeException("[secbox] The message "+kd+ " has no recipient");
                }
            } else {
                //get the session key if there
                if ( kd.getRandom() == this.random )
                    if ( si.verify(kd, getSignAlgo(), pk)) {
                        getStreamIAC().store(kd.getKeyLabel(),
                                             kd.getKey(getCipherAlgo(), getOwnPrivateKey()));
                    }
                kd.recycle();
                si.recycle();
            }
        } else {
            throw new RuntimeException("[secbox] Unrecognizable PlanningMessage: "+
				       m);
        }
    }
}
