/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.ConductorExt.security.hier;

import java.util.*;
import java.net.*;
import java.security.*;
import xjava.security.Cipher;

import FMG.Conductor.comm.*;
import FMG.Conductor.util.*;

import FMG.ConductorExt.security.util.*;
import FMG.ConductorExt.security.messages.*;
import FMG.ConductorExt.security.cert.*;
import FMG.ConductorExt.security.cert.Certificate;
import FMG.ConductorExt.security.hier.messages.HierAuthenticator;

/**
 * The MasterAuthenticator is the place where the public keys are
 * authenticated.
 * <p>
 * There are two situations: one is for the planner to verify the public keys
 * of other conductors; the other is for a conductor to verify the public
 * key of the planner in the reverse direction.
 * <p> 
 * Each node whose public key needs verification passes its own authenticator
 * to the MasterAuthenticator object. If there is already an authentic
 * certificate cached verifying the public key of this certificate, then we
 * simply maintain the public key of this node and other information;
 * otherwise, certificate obtainment and verification has to be done.
 * <p>
 * When a "hier" strategy is to be enforced, certificate servers are
 * arranged hierarchically. A node verifies a certificate as following:
 * <ul>
 *    <li> check whether the node knows the public key of the signing
 *         certificate server. The public key certificate of the signing
 *         certificate server may already be cached. In particular, every
 *         node knows the public key of the certificate hierarchy root.
 *    <li> if yes, run verification algorithm;
 *    <li> otherwise, try to get the verified public key of the
 *         certificate server, which is done by recursively verifying the
 *         certificate for the certificate server public key. THIS
 *         CERTIFICATE NEEDS TO BE RETURNED BY CONSULTING THE NEXT LEVEL
 *         CERTIFICATE SERVER, OR AN AUTHENTIC CERTIFICATE MIGHT CACHED
 *         ALREADY. 
* </ul>
 * <p>
 * The verified public keys for each conductor node is maintained for later usage.
 *
 * @author Jun Li (lijun@fmg.cs.ucla.edu)
 * @version $Revision: 1.7 $
 */
public class MasterAuthenticator {
    /** The maximum number of certificates that can be cached */
    private static final int CACHE_MAX_SIZE = 20;

    /** Boolean indicating whether to cache certificates or not */
    private static boolean   certCacheEnabled = HierSecurityBox.certificateCacheEnabled();

    /**
     * The cache where the certificates are kept. It is a Hashtable with
     * each element a CertificateList
     */
    private static Hashtable cachedCerts = new Hashtable(CACHE_MAX_SIZE);

    /** The size of the certificate cache */
    private static int       cacheSize = 0;

    /** The HierSecurityBox object with which this MasterAuthenticator is associated */
    private HierSecurityBox  sbox;

    /** Those public keys that are verified, and corresponding random number */
    private Hashtable        verifiedPubkeyPool = new Hashtable();

    public class VerifiedUnit {
        PublicKey pkey;
        int random;
        VerifiedUnit(PublicKey pkey, int random) {
            this.pkey = pkey;
            this.random = random;
        }
    }

    /** Create a new MasterAuthenticator associated with a HierSecurityBox */
    public MasterAuthenticator(HierSecurityBox sbox) {
        this.sbox = sbox;
    }

    /**
     * Update this MasterAuthenticator based on the given authenticator.
     * <p>
     * The authenticator contains certificates used for verifying the
     * public key of the node in question. If the public key is verified,
     * the public key will be added.
     * 
     * @param auth A received HierAuthenticator of other conductor
     * @return No return value
     */
    public void update(HierAuthenticator auth) {

        if ( Debug.on)
            Debug.println(Debug.SECURITY, Debug.MOREINFO, 
                          "[secbox] "+auth + " is being processed");            
        
        Certificate cert = getAuthenticCertificate(auth);
        
        if (cert != null) {
            NodeID    node = cert.getNode();
            PublicKey pk   = KeySerializer.toPubKey(cert.getPubKey(), 
                                                    sbox.getPubKeyAlgo());
            int       rand = auth.getRandom();
            
            addVerifiedPubkey(node, pk, rand);
            if ( Debug.on)
                Debug.println(Debug.SECURITY, Debug.MOREINFO,
                              "[secbox] pub key for "+node+
                              " is authenticated and added");
        } else {
            if ( Debug.on)
                Debug.println(Debug.SECURITY, Debug.MOREINFO,
                              "Authentication not passed: "+auth);
        }
    }

    /**
     * Gets the public key of a specific conductor.
     * The verified public keys are all kept in the hashtable
     * <code>verifiedPubkeyPool</code>. If the verification is not done yet,
     * the public key verification procedure will be started.
     *
     * @param node The conductor node whose verified public key to be returned   
     * @return the public key of the conductor node
     */
    public PublicKey getVerifiedPublicKey(NodeID node) {
        VerifiedUnit vu = (VerifiedUnit) verifiedPubkeyPool.get(node);
        if ( vu==null)
            return null;      // No verified public key is found
        return vu.pkey;
    }

    /**
     * Gets the KeyDistro message intended for a specific node
     *
     * @param node   the KeyDistro message to send to
     * @param label  the key label
     * @param skey   the byte array of the secret key to be distributed
     * @param connID the identification of this connection
     * @return the KeyDistro message with a specific node as the
     *         recipient. If the node does not have the verified public key,
     *         null is returned.
     */
    public KeyDistro getKeyDistro(NodeID node, String label, byte[] skey, ConnID connID) {
        if ( Debug.on && (Debug.getLevel(Debug.SECURITY) > Debug.MOREINFO )) {
            Debug.print("[secbox] getKeyDistro("+node+","+label+",");
            Debug.print("["+skey.length+"|"+Debug.printableString(skey, 0, skey.length-1));
            Debug.println("],"+connID+")");
        }

        KeyDistro kd;

        /* The public key of the node */
        PublicKey pk       = getVerifiedPublicKey(node);

        /*
         * The random number is to prevent the replay attack. Otherwise,
         * an attacker can replay a KeyDistro from previous time to cause
         * different key distributed which the attacker knows from
         * previous session.
         */
        int       rand     = getNodeRandom(node);

        if ( pk == null ) {
            /* 
             * No KeyDistro can be formed if the node has no verified public key. 
             * TODO: one node can get KeyDistro generated but his peer can not.
             */
            if ( Debug.on) {
                Debug.println(Debug.SECURITY, Debug.MOREINFO, 
                              "[secbox] No public key verified for node "+node);
            }
            return null;   
        }
  
        return KeyDistro.getNewMessage(connID, node, NodeID.getNewLocalID(),
                                       rand, skey, label,
                                       sbox.getCipherAlgo(), pk);
    }



    /*
     * private functions
     */

    /**
     * Gets the authentic certificate with respect to a HierAuthenticator.
     * <p>
     * The authenticator owner must be trusted at first. Then, check if
     * the certificate with this authenticator can be verified or not. If
     * verified as authentic, this certificate is returned; otherwise, this
     * authenticator can not be verified as an authentic authentictor.
     * 
     * @param auth the HierAuthenticator to be checked
     * @return the authentic certificate if it exists; null otherwise
     */
    private Certificate getAuthenticCertificate(HierAuthenticator auth) {
        Certificate cert = getCachedCertificate(auth.getNode(), 
                                                sbox.getPubKeyAlgo(),
                                                sbox.getSignAlgo());
        if ( cert != null )
            return cert;

        /* don't bother if we don't trust this host */
        if (!HierSecurityBox.trusted(auth.getNode())) {
            if ( Debug.on) 
                Debug.println(Debug.SECURITY, Debug.MOREINFO, 
                              "[secbox] "+auth.getNode()+"not trusted");
            return null;
        }
        
        cert = auth.getNodeCert();

        if (verifyCertificate(cert)) {
            //      addCachedCert(cert, sbox.getPubKeyAlgo(), sbox.getSignAlgo());
            return cert;
        } else {
            return null;
        }
    }

    
    /** 
     * Verifies a certificate.
     * <p> 
     * First check if this certificate is signed by a known certificate
     * server. A certificate server is "known" if it is the "root" of the
     * certificate hierarchy, or its own certificate is cached
     * locally. Otherwise, we have to ask the parent of this certificate
     * server for a certificate certifying this certificate server. And of
     * course the public key of the parent has to be verified in the same
     * way recursively, unless the parent is the "root".
     *
     * <p> This is a recursive function.
     *
     * @param cert the certificate to be verified
     * @return <code>true</code> if the certificate is verified as
     *         authentic; <code>false</code> otherwise.
     */
    private boolean verifyCertificate(Certificate cert) {

        boolean   verified;
        PublicKey csPubKey;

        if ( Debug.on)
            Debug.println(Debug.SECURITY, Debug.MOREINFO, 
                          "[secbox] Verifying "+cert);

        csPubKey = getCSPublicKey(cert.getAuthority());

        try {
            if ( csPubKey != null ) {
                verified = cert.verify(sbox.getSignAlgo(), csPubKey);
                
                if ( Debug.on)
                    Debug.println(Debug.SECURITY, Debug.MOREINFO,
                                  "[secbox] Verification: " + verified);
            } else {
                CertClient   cc =  new CertClient(cert.getAuthorityForAuthority());
                CertRequest req =  CertRequest.make(cert.getAuthority(),
                                                    sbox.getPubKeyAlgo(),
                                                    sbox.getSignAlgo());
                Certificate newCert = cc.getCertificate(req);
                if (verifyCertificate(newCert)) {
                    // Get the public key contained in var[authCert]
                    PublicKey pk = KeySerializer.toPubKey(newCert.getPubKey(), 
                                                          sbox.getPubKeyAlgo());
                    verified = cert.verify(sbox.getSignAlgo(), pk);
                } else {
                    verified = false;
                }
            }
        } catch ( Exception e ) {
            System.err.println("MasterAuthenticator.verifyCertificate()" +
                               "[" + e + "]");
            System.exit(1);
            return false;
        }

        if ( verified )
            addCachedCert(cert, sbox.getPubKeyAlgo(), sbox.getSignAlgo());

        return verified;
    }

    /**
     * Gets the public key of a certificate server.
     * If the certificate server is the root of the certificate hierarchy,
     * the root public key will be returned directly. Otherwise, the cached
     * certificates for certificate servers will be searched for this
     * specific certificate server.
     *
     * @param cs  the certificate server whose public key is being questioned.
     * @return the public key of this certificate server.
     */

    PublicKey getCSPublicKey(NodeID cs) {
        NodeID rootCS = HierSecurityBox.getRootCertServer();
        if ( cs.equals(rootCS) )
            return sbox.getRootCertServerPublicKey();

        Certificate csCert = getCachedCertificate(cs,
                                                  sbox.getPubKeyAlgo(),
                                                  sbox.getSignAlgo());
        if ( csCert == null )
            return null;
        else
            return  KeySerializer.toPubKey(csCert.getPubKey(), sbox.getPubKeyAlgo());
    }

    /**
     * Adds a verified public key into the pool for authentic public keys,
     * together with the random number.
     *
     * @param node the node whose public key is verified and being added   
     * @param pkey the verified public key of the specific node
     * @param rand the random number associated with the specific node
     * @return no return value.
     */
    private void addVerifiedPubkey(NodeID node, PublicKey pkey, int rand){
        VerifiedUnit vu = new VerifiedUnit(pkey, rand);
        verifiedPubkeyPool.put(node, vu);
    }
    
    /**
     * Gets the random number associated with a specific node.
     * 
     * @param node the node whose random number is to be determined
     * @return the random number associated with the node if it exists;
     *         -1 is non-existing.
     */
    private int getNodeRandom(NodeID node) {
        VerifiedUnit vu = (VerifiedUnit) verifiedPubkeyPool.get(node);
        if ( vu == null)
            return -1;  // A random number is >=0. -1 means non-existence.
        return vu.random;
    }

    synchronized private static Certificate getCachedCertificate(NodeID principal,
                                                                 String pkeyAlgo,
                                                                 String signAlgo) {
        
        String          format = new String(pkeyAlgo+signAlgo);
        CertificateList list   = (CertificateList)cachedCerts.get(format);
        Certificate     cert   = null;

        if ( list != null )
            cert = list.findCertForNode(principal);

        return cert;
    }

  
    synchronized private static void addCachedCert(Certificate cert,
                                                   String pkeyAlgo,
                                                   String signAlgo) {
        String          format;
        CertificateList list;
        
	/*
	 * By not adding any certificates at all, cache is useless. This
	 * way the cache is diabled. All other code unchanged.
	 */
	if ( !certCacheEnabled )
	    return;  

 	/*
 	 *  Assuming a conductor node can not be its own certificate server;
 	 *  Otherwise, the cache may contain two certificates with same principal,
 	 *  A(conductor) by A(CS), and A(CS) by B(next level CS).
 	 */
        if (getCachedCertificate(cert.getNode(), pkeyAlgo, signAlgo) != null)
	    throw new RuntimeException("Warning: certificate cache already contains the certificate to add");

        if ( cacheSize == CACHE_MAX_SIZE )
            shrinkCache();

        format = new String(pkeyAlgo+signAlgo);
        list   = (CertificateList)cachedCerts.get(format);

        if ( list == null ) {
            list = CertificateList.create();
            cachedCerts.put(format, list);
        }
        list.addCertificate(cert);
        cacheSize ++;
    }

    synchronized private static void shrinkCache() {
        Enumeration formats   = cachedCerts.keys();
        String      aFormat   = (String)formats.nextElement();
        CertificateList aList = (CertificateList)cachedCerts.get(aFormat);

        if ( aList.size() == 0 )
            throw new RuntimeException("Cached CertificateList size is zero");

        aList.removeOldestCertificate();

        if ( aList.size() == 0 ) {
            aList.recycle();
            cachedCerts.remove(aFormat);
        }

        cacheSize --;
    }
    
}
