/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.ConductorExt.security.hier.messages;

import  java.io.*;
import  java.net.*;
import  java.util.*;
import  java.security.SecureRandom;

import  FMG.Conductor.comm.*;
import  FMG.Conductor.util.*;
import  FMG.Conductor.planning.messages.PlanningMessage;


import  FMG.ConductorExt.security.cert.*;
import  FMG.ConductorExt.security.hier.*;

/**
 * The HierAuthenticator class implements the authenticator used by the
 * "hier" strategy. It passes necessary authentication information to
 * facilitate the authentication of the public key of the owner of the
 * authenticator. A HierAuthenticator mainly includes the direction if
 * flows, the NodeID of the owner, the certificate certifying the node 
 * (the owner of the authenticator) by the associated certificate server,
 * and a random number.
 *
 * @author Jun Li (lijun@fmg.cs.ucla.edu)
 * @version $Revision: 1.4 $
 */ 

public class HierAuthenticator extends PlanningMessage {
    /** Boolean indicating whether to cache certificates or not */
    private static boolean   certCacheEnabled = HierSecurityBox.certificateCacheEnabled();

    /** The certificates useful for forming the HierAuthenticator */
    private static Hashtable localCerts = new Hashtable();

    /** The recycler used for messages of this type */
    private static Recycler  recycler = new Recycler(HierAuthenticator.class);

    /** The message id assigned to messages of this type */
    private static int       messageID = -1;

    /** The direction that this HierAuthenticator flows */
    private Direction        direction;

    /** The NodeID of the owner of this HierAuthenticator */
    private NodeID           node;

    /** The certificate certifying this node by local certificate server */
    private Certificate      nodeCert;

    /** The random number associated with this node in this session */
    private int              random;
    
    /**
     * Register a prototype of this message class with the BasicMessageInputStream.
     * @return No return value.
     * @see FMG.Conductor.comm.BasicMessageInputStream
     */
    public static void registerSelf() {
        messageID = BasicMessageInputStream.
            registerMessagePrototype((Message) recycler.getOne());
    }

    /**
     * This method returns the message id that was assigned to messages of this
     * type by BasicMessageInputStream.
     * @return The message id for this type of message.
     * @see FMG.Conductor.comm.BasicMessageInputStream
     */
    public int getMessageID() {
        return messageID;
    }
    
    /**
     * This method returns a new (or recycled) HierAuthenticator object 
     * initialized from the input stream.  Normally, this method is only 
     * called on the prorotype given to BasicMessageInputStream.
     * @param is The input stream to read the message's data from
     * @return a new object initialized from the data stream.
     * @exception java.io.IOException if an error occurs while reading from
     *                                the stream.
     * @see FMG.Conductor.comm.BasicMessageInputStream
     */
    public Message readNewMessage(DataInputStream is) throws IOException {
        return HierAuthenticator.getNewMessage(is);
    }

    /**
     * Generate a new (or recycled) HierAuthenticator message with the appropriate
     * data members initialized.
     *
     * @param connID The id of the connection for which we are planning.
     * @param dir    The direction that this message will flow
     * @param pkeyAlgo The public key algorithm to use
     * @param signAlgo The signing algorithm to use
     * @param cc  The certificate client to asks for certificates
     * @param random The random number associated with this  node during this planning
     * @return A new (or recycled) and initialized HierAuthenticator message.
     */

    public static HierAuthenticator getNewMessage(ConnID connID,
                                                  Direction dir,
                                                  String pkeyAlgo,
                                                  String signAlgo,
                                                  CertClient cc,
                                                  int random) {
        HierAuthenticator m = (HierAuthenticator) recycler.getOne();
        m.init(connID, dir, pkeyAlgo, signAlgo, cc, random);
        return m;
    }


   /**
    * This method returns a new (or recycled) HierAuthenticator object 
    * initialized from the input stream.
    * @param is The input stream to read the message's data from
    * @return A new (or recycled) and initialized HierAuthenticator message.
    * @exception java.io.IOException if an error occurs while reading from
    *                                the stream.
    */
    public static HierAuthenticator getNewMessage(DataInputStream is) 
        throws IOException {
       HierAuthenticator m = (HierAuthenticator) recycler.getOne();
       m.receiveSelf(is);
       return m;
    }
    
    /**
     * Read the data elements of this object from the input stream.  Presumably
     * this information was generated by <code>sendSelf()</code>.
     * @param is The input stream to read the message's data from.
     * @return No return value.
     * @exception java.io.IOException if an error occurs while reading from 
     *                                the stream
     */
    public void receiveSelf(DataInputStream is) throws IOException {
      super.receiveSelf(is);

      direction   = Direction.readDir(is);

      node        = new NodeID(is);
      nodeCert    = Certificate.make(is);
      random      = is.readInt();
    }

    /**
     * Write the data elements of this object to an output stream.  Presumably
     * the information will be read by <code>receiveSelf()</code>.
     * @param os The output stream to write the message's data to.
     * @return No return value.
     * @exception java.io.IOException if an error occurs while writing to the
     *                                stream.
     */
    public void sendSelf(DataOutputStream os) throws IOException {
       super.sendSelf(os);

       direction.writeSelf(os);

       node.writeSelf(os);
       nodeCert.writeSelf(os);
       os.writeInt(random);
    }

    /**
     * Gets the node associated with this HierAuthenticator.
     *
     * @return the node associated with this HierAuthenticator
     */
    public NodeID getNode() {
        return node;
    }

    /** Gets the certificates certifying the node associated with this HierAuthenticator
     *
     * @return the enumeration of those certificates for the associated node
     */
    public Certificate getNodeCert() {
        return nodeCert;
    }

    /** Gets the random number associated with the node during this
     *  planning session    
     *
     * @return the random number from this HierAuthenticator
     */
    public int getRandom() {
        return random;
    }

    /**
     * Determine the direction of motion for this object.
     * @return returns <code>Direction.up</code> if towards the server;
     *                 <code>Direction.down</code> if towards the client.
     */
    public Direction flowDir() {
        return direction;
    }
    
    /**
     * Determine whether this object is understood by the Path object.
     * @return Always returns false.
     */
    public boolean handledByPath() {
        return false;
    }

    /**
     * Gets the string representation of this HierAuthenticator
     * 
     * @return the string representation of this HierAuthenticator
     */    
    public String toString() {
        return "HierAuthenticator("+node+" WITH "+nodeCert+" random="+random+")";
    }
    
    /**
     * Perform cleanup before allowing recycling to occur.  This method
     * is called by the Recycler. 
     * @return No return value.
     * @see FMG.Conductor.util.Recycler
     */
    public void recycling() {
        super.recycling();
	nodeCert.recycle();
    }

    /**
     * This is for test purpose only 
     */
    public static final void main(String[] args) {
        if ( args.length != 1 ) {
            System.out.println("Usage:\n"+"java FMG.ConductorExt.security.hier.messages.HierAuthenticator <certserver>");
            System.exit(1);
        }

        try {
            self_test(args[0]);
        } catch ( Exception e) {
            e.printStackTrace();

        }
    }

    /**
     * Initializes a HierAuthenticator message.
     * <p>
     * It assigns the connection ID, the flow direction, the NodeID of the
     * node and its local certificate server. In particular, a certificate
     * for this node by the local certificate server will be retrieved, and
     * added to the list keeping the certificates for the node.
     *
     * @param connID  The id of the connection for which we are planning.
     * @param dir the direction that this HierAuthenticator will flow
     * @param pkeyAlgo The public key algorithm to use
     * @param signAlgo The signing algorithm to use
     * @param certServer The NodeID of the certificate server
     * @param cc  The certificate client to asks for certificates
     * @param random The random number associated with this  node during this planning
     * @return no return value.
     */
    private void init(ConnID connID, 
                      Direction dir, 
                      String pkeyAlgo,
                      String signAlgo, 
                      CertClient cc,
                      int random) {
      initMessage(connID);

      this.direction   = dir;

      this.node        = NodeID.getNewLocalID();
      this.random      = random;
      this.nodeCert    = deriveNodeCert(cc,  pkeyAlgo, signAlgo);
    }

    private Certificate deriveNodeCert(CertClient cc, 
				       String pkeyAlgo,
				       String signAlgo) {
	Certificate cert = getLocalCertificate(this.node, pkeyAlgo, signAlgo);
	if ( cert == null ) {
	    cert = getRemoteCertificate(this.node, cc, pkeyAlgo, signAlgo);
	    addLocalCert(cert, pkeyAlgo, signAlgo);
	}

	if ( cert == null )
	    throw new RuntimeException("No Certificate Returned From Authority");

	return cert;
    }

    private Certificate getRemoteCertificate(NodeID who, 
					     CertClient cc, 
					     String pkeyAlgo, 
					     String signAlgo) {
      try {
          CertRequest req = CertRequest.make(who, pkeyAlgo,signAlgo);
          return cc.getCertificate(req);
      } catch (IOException e) {
	  throw new RuntimeException("HierAuthenticator: Certificate I/O Error" + e);
      }
    }

    synchronized private static Certificate getLocalCertificate(NodeID principal,
								String pkeyAlgo,
								String signAlgo) {
	String          format = new String(pkeyAlgo+signAlgo);
	return (Certificate)localCerts.get(format);
    }

    synchronized private static void addLocalCert(Certificate cert,
						  String pkeyAlgo,
						  String signAlgo) {
	String          format; 

	/*
	 * By not adding any certificates at all, cache is useless. This
	 * way the cache is diabled. All other code unchanged.
	 */
	if ( ! certCacheEnabled ) 
	    return;

	if ( getLocalCertificate(cert.getNode(), pkeyAlgo, signAlgo) != null )
	    throw new RuntimeException("Warning: certificate cache already contains the certificate to add");
	
	format = new String(pkeyAlgo+signAlgo);
	localCerts.put(format, cert);
    }

    /**
     * Self-tests procedure
     *
     * @param serverName the name of the certificate server
     * @return no return value
     */
    private static void self_test(String serverName) {

        InetAddress serverAddr = null;
        CertClient  certClient;
        int rand = (new SecureRandom()).nextInt();

        try {
            serverAddr = InetAddress.getByName(serverName);
        } catch ( UnknownHostException e ) {
            System.err.println("Unknown cert server address");
            System.exit(1);
        }

        certClient = new CertClient(new NodeID(serverAddr));

        HierAuthenticator auth = 
            HierAuthenticator.getNewMessage(new ConnID(),
                                             Direction.up,
                                             "RSA",
                                             "SHA-1/RSA/PKCS#1",
                                             certClient,
                                             rand);
        System.out.println("auth is " + auth );
    }                                        
                                                                   
}
