/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.ConductorExt.security.keydistro;

import java.util.*;
import java.security.*;
import xjava.security.Cipher;
import java.io.*;

import FMG.ConductorExt.security.util.KeySerializer;
import FMG.Conductor.security.*;
import FMG.Conductor.planning.messages.*;
import FMG.Conductor.dataflow.IAC;
import FMG.Conductor.util.Debug;
import FMG.Conductor.comm.NodeID;
import FMG.Conductor.Conductor;

/**
 * This class does key distribution only, using public/private keys stored
 * locally.
 * <p>
 * @author Mark Yarvis (yarvis@fmg.cs.ucla.edu)
 * @version $Revision: 1.10 $      
 */
public class KeyDistroSecurityBox extends SecurityBox {

   static final String PUBLIC_KEY_TYPE = "RSA";
   static final String SIGNATURE_ALGORITHM = "SHA-1/RSA/PKCS#1";
   static final String ENCRYPTION_ALGORITHM = "RSA/ECB/OneAndZeroes";

   static final String PUBLIC_KEY_FILE = 
          Conductor.getFilenameProperty("KeyDistroSecurityBox.PublicKeyFile",
                                        "./config/PublicKeys");
   static final String PRIVATE_KEY_FILE = 
          Conductor.getFilenameProperty("KeyDistroSecurityBox.PrivateKeyFile",
                                        "./config/PrivateKey");

   protected PrivateKey privateKey;

   private Hashtable publicKeys = new Hashtable();

   protected static NodeID localNode = NodeID.getNewLocalID();

   protected Vector keyMessagesToBeSent = new Vector();

   protected IAC iac;

   private boolean alreadyMadeKeys = false;

   static {
      if ( Security.getProvider("Cryptix") == null ) {
         Security.insertProviderAt(new cryptix.provider.Cryptix(), 1);
      }   

      KeyMessage.registerSelf();
   }

   public KeyDistroSecurityBox(SecurityCoordinator coordinator, 
                               String param) {
      super(coordinator, param);

      privateKey = readPrivateKey();
      readPublicKeys(publicKeys);

      iac = getStreamIAC();
   }

   private PrivateKey readPrivateKey() {
      FileInputStream is = null;
      try {
         is = new FileInputStream(PRIVATE_KEY_FILE);
         return KeySerializer.readPrivKey(is, PUBLIC_KEY_TYPE);
      } catch (IOException e) {
         throw new RuntimeException("Failure reading private key from file " +
                                    PRIVATE_KEY_FILE + ":" + e);
      } finally {
         try {
            if (is != null) {
               is.close();
            }
         } catch (IOException ignore) {}
      }
   }

   private void readPublicKeys(Hashtable keyTable) {
      InputStream is = null;
      try {
         is = new FileInputStream(PUBLIC_KEY_FILE);
         StringBuffer sb = new StringBuffer();

         while (true) {
            sb.setLength(0);
            while (true) {
               int c = is.read();
               if (c == ' ') {
                  break;
               }
               if (c == -1) {  // end of stream
                  return;
               }
               sb.append((char) c);
            }

            NodeID node = new NodeID(sb.toString());
            if (Debug.on) {
               Debug.println(Debug.SECURITY, Debug.EXTRAINFO, 
                             "Reading a public key for node " + node);
            }
            PublicKey key = KeySerializer.readPubKey(is, PUBLIC_KEY_TYPE);

            publicKeys.put(node, key);

            int eol = is.read();
            if (eol == -1) {
               return;
            }
         }

      } catch (IOException e) {
         throw new RuntimeException("Failure reading private key from file " +
                                    PRIVATE_KEY_FILE + ":" + e);
      } finally {
         try {
            if (is != null) {
               is.close();
            }
         } catch (IOException ignore) {}
      }
   }

   protected PublicKey getPublicKey(NodeID node) {
      return (PublicKey) publicKeys.get(node);
   }

   public void handleIncoming(PlanningMessage m) {
      if (Debug.on) {
         Debug.println(Debug.SECURITY, Debug.MOREINFO, 
                       "KeyDistroSecurityBox: handling incoming " + m);
      }

      // if it's a key message
      if (m instanceof KeyMessage) {
          KeyMessage km = (KeyMessage) m;

          // if it's local, deliver it
          if (km.getTargetNode().isLocalNode()) {
             // verify, decrypt, and deliver it
             if (km.unlockAndVerify(privateKey, 
                                    getPublicKey(km.getSourceNode()))) {
                iac.store(km.getKeyLabel(), km.getKey());
             } else {
                if (Debug.on) {
                   Debug.println(Debug.SECURITY, Debug.WARNING,
                                 this + " Unable to verify key");
                }
             }
             km.recycle();
          } else {   // if it's nonlocal, cache it
             // try to send, but if we fail, store it for later
             if (!coordinator.sendCToNet(km)) {
                keyMessagesToBeSent.addElement(km);
             }
          }
      } else {
         if (Debug.on) {
            Debug.println(Debug.SECURITY, Debug.MOREINFO, 
                          "KeyDistroSecurityBox: delivering message to path");
         }

         coordinator.deliverToPath(m);
      }
   }

   public boolean handleOutgoingToServer(PlanningMessage m) {
      if (Debug.on) {
         Debug.println(Debug.SECURITY, Debug.MOREINFO, 
                       "KeyDistroSecurityBox: forwarding " + 
                       "outgoing to server " + m);
      }

      return coordinator.sendSToNet(m);
   }

   public boolean handleOutgoingToClient(PlanningMessage m) {
      if (Debug.on) {
         Debug.println(Debug.SECURITY, Debug.MOREINFO, 
                       "KeyDistroSecurityBox: handling " + 
                       "outgoing to client " + m);
      }

      // if it's a PlanDistro
      if (m instanceof PlanDistro) {
         // if I'm the planner
         if (coordinator.isPlannerNode() && !alreadyMadeKeys) {
            alreadyMadeKeys = true;

            // find out what keys are needed ...
            Enumeration sessionKeys = getRequiredSessionKeys((PlanDistro) m);

            // for each key that needs to be distributed
            while (sessionKeys.hasMoreElements()) {
               SessionKey sk = (SessionKey) sessionKeys.nextElement();
               byte[] key = sk.getKey();
               String keyLabel = sk.getKeyLabel();

               // for each node it needs to go to
               Enumeration nodes = sk.getNodes();
               while (nodes.hasMoreElements()) {
                  NodeID targetNode = (NodeID) nodes.nextElement();
                  if (targetNode.isLocalNode()) {
                     // deliver locally
                     iac.store(keyLabel, key);
                  } else {
                     // create a signed and encrypted message
                     KeyMessage km = 
                          KeyMessage.getNewMessage(m.connID, key, keyLabel, 
                                                   localNode, targetNode, 
                                                   privateKey,
                                                   getPublicKey(targetNode));
                     // try to send, otherwise store it for later
                     if (!coordinator.sendCToNet(km)) {
                        keyMessagesToBeSent.addElement(km);
                     }
                  }
               }
            }
         }

         // send remaining key messages (on failure, return false)
         while (keyMessagesToBeSent.size() != 0) {
            if (Debug.on) {
               Debug.println(Debug.SECURITY, Debug.MOREINFO, 
                             "KeyDistroSecurityBox: sending key toward client");
            }

            KeyMessage km = (KeyMessage) keyMessagesToBeSent.firstElement();

            if (!coordinator.sendCToNet(km)) {
               return false;
            }

            keyMessagesToBeSent.removeElementAt(0);
         }
      }

      if (Debug.on) {
         Debug.println(Debug.SECURITY, Debug.MOREINFO, 
                       "KeyDistroSecurityBox: forwarding toward client");
      }

      if (coordinator.sendCToNet(m)) {
         alreadyMadeKeys = false;   // reset for next time
         return true;
      } else {
         return false;
      }
   }
}
