/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.ConductorExt.security.keydistro;

import java.io.*;
import java.security.*;
import xjava.security.Cipher;

import FMG.Conductor.comm.*;
import FMG.Conductor.util.*;

// this class presumes that the key label length can be encoded in 4 bytes
public class KeyMessage extends SignedEncryptedPlanningMessage {

   /** The recycler used for messages of this type. */
   private static Recycler recycler = new Recycler(KeyMessage.class);

   // transmitted values
   NodeID sourceNode = new NodeID();
   NodeID targetNode = new NodeID();

   // cached values, not transmitted
   byte[] key;
   String keyLabel;

   private static final int messageID = 120;

   public static void registerSelf() {
      // since we're a dynamic module, use a static message id
      BasicMessageInputStream.registerMessagePrototype(
                                  (Message) recycler.getOne(), messageID);
   }

   public int getMessageID() {
      return messageID;
   }

   public Message readNewMessage(DataInputStream is) throws IOException {
      return KeyMessage.getNewMessage(is);
   }

   public static KeyMessage getNewMessage(DataInputStream is)
                                       throws IOException {
      KeyMessage m = (KeyMessage) recycler.getOne();
      m.receiveSelf(is);
      return m;
   }

   public static KeyMessage getNewMessage(ConnID connID, 
                                          byte[] key, String keyLabel, 
                                          NodeID sourceNode, NodeID targetNode,
                                          PrivateKey signKey, 
                                          PublicKey encryptKey) {
      KeyMessage m = (KeyMessage) recycler.getOne();
      m.initMessage(connID, key, keyLabel, sourceNode, targetNode, 
                                              signKey, encryptKey);
      return m;
   }

   public void initMessage(ConnID connID, byte[] key, String keyLabel, 
                           NodeID sourceNode, NodeID targetNode,
                           PrivateKey signKey, PublicKey encryptKey) {
      byte[] data = packageData(key, keyLabel);

      initMessage(connID, data, signKey, encryptKey);

      this.key = key;
      this.keyLabel = keyLabel;
      this.sourceNode.copy(sourceNode);
      this.targetNode.copy(targetNode);

   }

   protected byte[] packageData(byte[] key, String keyLabel) {
      byte[] labelBytes = keyLabel.getBytes();
      int labelLen = labelBytes.length;
      byte[] buf = new byte[4 + labelLen + key.length];

      // encode the label length into the first four bytes
      for (int i = 3; i>=0; i--) {
         buf[i] = (byte) (labelLen & 0xff);
         labelLen >>= 8;
      }
      // note, labelLen is no longer valid, and should be zero

      // sanity check ... make sure the label length fit
      if (Debug.on && (labelLen != 0)) {
         Debug.println("WARNING: The key label is too long!");
      }


      // copy the label
      for (int i=0; i<labelBytes.length; i++) {
         buf[i+4] = labelBytes[i];
      }

      // copy the key
      for (int i=0; i<key.length; i++) {
         buf[i+4+labelBytes.length] = key[i];
      }

      return buf;
   }

   public void receiveSelf(DataInputStream is) throws IOException {
      super.receiveSelf(is);
      sourceNode.readSelf(is);
      targetNode.readSelf(is);
   }

   public void sendSelf(DataOutputStream os) throws IOException {
      super.sendSelf(os);
      sourceNode.writeSelf(os);
      targetNode.writeSelf(os);
   }

   public boolean unlockAndVerify(PrivateKey decryptionKey, 
                                  PublicKey verificationKey) {
      if (super.unlockAndVerify(decryptionKey, verificationKey)) {
         byte[] data = getData();

         // decode label length
         int labelLen = 0;
         for (int i=0; i<4; i++) {
            labelLen <<= 8;
            labelLen += data[i];
         }

         // get the key label
         keyLabel = new String(data, 4, labelLen);

         // get the key
         int keyLen = data.length - labelLen - 4;
         key = new byte[keyLen];
         for (int i=0; i < keyLen; i++) {
            key[i] = data[i + labelLen + 4];
         }
         return true;
      } else {
         return false;
      }
   }

   // returns null if unlockAndVerify has not ben successfully called
   public String getKeyLabel() {
      return keyLabel;
   }

   // returns null if unlockAndVerify has not ben successfully called
   public byte[] getKey() {
      return key;
   }

   public NodeID getSourceNode() {
      return sourceNode;
   }

   public NodeID getTargetNode() {
      return targetNode;
   }

   /**
    * Cleanup secret data before recycling can occur.
    */
   public void recycling() {
      key = null;
      keyLabel = null;
   }

   public Direction flowDir() {
      return Direction.down;
   }

   public boolean handledByPath() {
      return false;
   }

}
