/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.ConductorExt.security.keydistro;

import java.net.*;
import java.io.*;
import java.security.*;
import xjava.security.Cipher;

import FMG.Conductor.comm.NodeID;
import FMG.ConductorExt.security.util.KeySerializer;

public class MakeKeys {

   static final String PUBLIC_KEY_TYPE = KeyDistroSecurityBox.PUBLIC_KEY_TYPE;

   static {
      if ( Security.getProvider("Cryptix") == null ) {
         Security.insertProviderAt(new cryptix.provider.Cryptix(), 1);
      }
   }

   public static void main(String[] args) {
      if (args.length != 3) {
         usage();
         System.exit(1);
      }

      NodeID node = null;
      try {
         node = new NodeID(InetAddress.getByName(args[0]));
      } catch (UnknownHostException e) {
         System.out.println("Host unknown: " + args[0]);
         System.exit(1);
      }

      // create a key pair
      KeyPair kp = getKeyPair();

      // write the private key
      writePrivateKey(kp.getPrivate(), args[1]);

      // write the public key
      writePublicKey(node, kp.getPublic(), args[2]);
   }

   public static KeyPair getKeyPair() {
      try {
         KeyPairGenerator kg = KeyPairGenerator.getInstance(PUBLIC_KEY_TYPE);
         SecureRandom rand = new SecureRandom();
         kg.initialize(512, rand);

         return kg.generateKeyPair();
      } catch (NoSuchAlgorithmException e) {
         System.out.println("Failure making key pair: " + e);
         System.exit(1);
         return null;   // not reached
      }
   }

   public static void writePublicKey(NodeID node, PublicKey publicKey, 
                                                      String filename) {
      FileOutputStream os = null;
      try {
         os = new FileOutputStream(filename,true);
         os.write(node.toString().getBytes());
         os.write(' ');
         KeySerializer.writePubKey(os, publicKey, PUBLIC_KEY_TYPE);
         os.write('\n');
      } catch (IOException e) {
         System.out.println("Failure writing private key file (" + 
                            filename + "): " + e);
         System.exit(1);
      } finally {
         try {
            if (os != null) {
               os.close();
            }
         } catch (IOException ignore) {}
      }
   }

   public static void writePrivateKey(PrivateKey privateKey, String filename) {
      FileOutputStream os = null;
      try {
         os = new FileOutputStream(filename);
         KeySerializer.writePrivKey(os, privateKey, PUBLIC_KEY_TYPE);
      } catch (IOException e) {
         System.out.println("Failure writing private key file (" + 
                            filename + "): " + e);
         System.exit(1);
      } finally {
         try {
            if (os != null) {
               os.close();
            }
         } catch (IOException ignore) {}
      }
   }

   public static void usage() {
      System.out.println("MakeKeys host private-key-file public-key-file");
   }
}
