/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.ConductorExt.security.keydistro;

import java.io.*;
import java.security.*;
import xjava.security.Cipher;

import cryptix.provider.key.*;

import FMG.Conductor.planning.messages.PlanningMessage;
import FMG.Conductor.comm.*;

public abstract class SignedEncryptedPlanningMessage extends PlanningMessage {

   // this is kind of a hack
   protected String sigName = KeyDistroSecurityBox.SIGNATURE_ALGORITHM;
   protected String cipherName = KeyDistroSecurityBox.ENCRYPTION_ALGORITHM;

   // the algs never change, so we're going to create them once
   Cipher cipherAlg;
   Signature sigAlg;

   {
     try {
         cipherAlg = Cipher.getInstance(cipherName);
         sigAlg = Signature.getInstance(sigName);
      } catch (NoSuchAlgorithmException e) {
         throw new RuntimeException("Unexpected signature failure");
      }
   }

   // transmitted values
   private byte[] encryptedData;
   private byte[] signature;

   // cached values, not transmitted
   private byte[] plaintextData;

   protected byte[] getData() {
      return plaintextData;
   }

   public void initMessage(ConnID connID, byte[] data, 
                           PrivateKey signKey, PublicKey encryptKey) {
      initMessage(connID);

      plaintextData = data;

      // encrypt the data
      try {
         cipherAlg.initEncrypt(encryptKey);
         encryptedData = cipherAlg.crypt(plaintextData, 0, 
                                         plaintextData.length);
      } catch (KeyException e) {
         throw new RuntimeException("Unexpected encryption failure");
      }

      // sign the data
      try {
         sigAlg.initSign(signKey);
         sigAlg.update(encryptedData, 0, encryptedData.length);
         signature = sigAlg.sign();
      } catch (InvalidKeyException e) {
         throw new RuntimeException("Unexpected signature failure");
      } catch (SignatureException e) {
         throw new RuntimeException("Unexpected signature failure");
      }
   }

   public void receiveSelf(DataInputStream is) throws IOException {
      super.receiveSelf(is);

      int dataLen = is.readInt();
      encryptedData = new byte[dataLen];
      is.readFully(encryptedData);

      int sigLen = is.readInt();
      signature = new byte[sigLen];
      is.readFully(signature);
   }

   public void sendSelf(DataOutputStream os) throws IOException {
      super.sendSelf(os);

      os.writeInt(encryptedData.length);
      os.write(encryptedData);

      os.writeInt(signature.length);
      os.write(signature);
   }

   public boolean unlockAndVerify(PrivateKey decryptionKey, 
                                  PublicKey verificationKey) {
      // verify the signature
      try {
         sigAlg.initVerify(verificationKey);
         sigAlg.update(encryptedData);
         if (! sigAlg.verify(signature)) {
            return false;
         }
      } catch (InvalidKeyException e) {
         throw new RuntimeException("Unexpected signature failure");
      } catch (SignatureException e) {
         throw new RuntimeException("Unexpected signature failure");
      }

      // decrypt the data
      try {
         cipherAlg.initDecrypt(decryptionKey);
         plaintextData = cipherAlg.crypt(encryptedData, 0, 
                                         encryptedData.length);
      } catch (KeyException e) {
         throw new RuntimeException("Unexpected decryption failure");
      }
      
      return true;
   }

   /**
    * Cleanup secret data before recycling can occur.
    */
   public void recycling() {
      plaintextData = null;
   }
}
