/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.ConductorExt.security.messages;

import java.io.*;

import FMG.Conductor.comm.*;
import FMG.Conductor.util.*;

import FMG.ConductorExt.security.util.*;
import FMG.Conductor.planning.messages.PlanningMessage;

/**
 * This BoxCharacteristics class defines the characteristics of a security
 * box. 
 * <p> 
 * It is intended to be used by the planner to confirm back to other
 * conductor node that the planner is indeed deploying a security box that
 * it is supposed to.
 * @author Jun Li      ( lijun@fmg.cs.ucla.edu)
 * @version $Revision: 1.4 $
 */
public class BoxCharacteristics extends PlanningMessage {

    /** The recycler used for messages of this type */
    private static Recycler recycler = new Recycler(BoxCharacteristics.class);

    /** The message id assigned to messages of this type */
    private static int      messageID = -1;

    /** The name of the security box class */
    private String boxClass;

    /** 
     *  The parameters used by the security box. They are propagated from
     *  the client node and represented by a String.
     */
    private String parameter;

    /**
     * Register a prototype of this message class with the 
     * BasicMessageInputStream.
     * @return No return value.
     * @see FMG.Conductor.comm.BasicMessageInputStream
     */
    public static void registerSelf() {
        messageID = BasicMessageInputStream.
            registerMessagePrototype((Message) recycler.getOne());
    }
    
    /**
     * This method returns the message id that was assigned to messages of this
     * type by BasicMessageInputStream.
     * @return The message id for this type of message.
     * @see FMG.Conductor.comm.BasicMessageInputStream
     */
    public int getMessageID() {
        return messageID;
    }
    
    
    /**
     * This method returns a new (or recycled) BoxCharacteristics object 
     * initialized from the input stream.  Normally, this method is only 
     * called on the prototype given to BasicMessageInputStream.
     * @param is The input stream to read the message's data from
     * @return a new object initialized from the data stream.
     * @exception java.io.IOException if an error occurs while reading from
     *                                the stream.
     * @see FMG.Conductor.comm.BasicMessageInputStream
     */
    public Message readNewMessage(DataInputStream is) throws IOException {
        return BoxCharacteristics.getNewMessage(is);
    }
    

    /**
     * Generates a new(or recycled) BoxCharacteristics object.
     *
     * @param connID       The id of the connection for which we are planning
     * @param boxClassName The name of the security box class
     * @param parameter    The parameter string propagated from client
     * @return a new(or recycled) initialized BoxCharacteristics message.
     */
    public static BoxCharacteristics getNewMessage(ConnID connID,
                                                   String boxClassName,
                                                   String parameter) {
        BoxCharacteristics m = (BoxCharacteristics)recycler.getOne();
        m.init(connID, boxClassName, parameter);
        return m;
    }

    /**
     * This method returns a new (or recycled) BoxCharacteristics object 
     * initialized from the input stream.
     * @param is The input stream to read the message's data from
     * @return A new (or recycled) and initialized BoxCharacteristics message.
     * @exception java.io.IOException if an error occurs while reading from
     *                                the stream.
     */
    public static BoxCharacteristics getNewMessage(DataInputStream is) throws IOException {
        BoxCharacteristics m = (BoxCharacteristics) recycler.getOne();
        m.receiveSelf(is);
        return m;
    }

    
    /**
     * Write the data elements of this object to an output stream.  Presumably
     * the information will be read by <code>receiveSelf()</code>.
     * @param os The output stream to write the message's data to.
     * @return No return value.
     * @exception java.io.IOException if an error occurs while writing to the
     *                                stream.
     */
    public void sendSelf(DataOutputStream os) throws IOException {
        super.sendSelf(os);

        os.writeInt(boxClass.length());
        os.write(boxClass.getBytes());
        os.writeInt(parameter.length());
        os.write(parameter.getBytes());
    }

    /**
     * Read the data elements of this object from the input stream.  Presumably
     * this information was generated by <code>sendSelf()</code>.
     * @param is The input stream to read the message's data from.
     * @return No return value.
     * @exception java.io.IOException if an error occurs while reading from 
     *                                the stream
     */
    public void receiveSelf(DataInputStream is) throws IOException {
        super.receiveSelf(is);

        boxClass  = new String(BytesReader.getFrom(is));
        parameter = new String(BytesReader.getFrom(is));
    }

    /**
     * Determine the direction of motion for this object.
     * @return always <code>Direction.down</code> since it flows towards the client.
     */
     public Direction flowDir() {
        return Direction.down;
    }

    /**
     * Determine whether this object is understood by the Path object.
     * @return Always returns false.
     */
    public boolean handledByPath() {
        return false;
    }
    
    /**
     * Gets the string representation of this BoxCharacteristics.
     * 
     * @return the string representation of this BoxCharacteristics.
     */    
    public String toString() {
        return "BoxCharacteristics("+boxClass+" "+parameter+")";
    }

    /**
     * Determines if this BoxCharacteristics object represents the same
     * security box characteristics as another object.
     *
     * @param o An object to compare with.
     * @return True iff the object is a BoxCharacteristics object that
     *                  represents the same characteristics as this one.
     */
    public boolean equals(Object o) {
        if ( o == null ) {
            return false;
        }

        if ( o instanceof BoxCharacteristics ) {
            BoxCharacteristics bc = (BoxCharacteristics) o;
            return ( this.boxClass.equals(bc.boxClass) &&
                     this.parameter.equals(bc.parameter) );
        } else
            return false;
    }

    /*
     * private functions
     */

    /**
     *  Initializes this BoxCharacteristics object.
     *
     * @param connID    The id of the connection for which we are planning
     * @param boxClass  The name of the security box class
     * @param parameter The parameter string propagated from client
     * @return no return value.
     */
    private void init(ConnID connID, 
                      String boxClass, 
                      String parameter) {
        initMessage(connID);
        this.boxClass = boxClass;
        this.parameter = parameter;
    }

}
