/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.ConductorExt.security.messages;

import java.io.*;
import java.net.*;
import java.security.*;
import xjava.security.Cipher;

import cryptix.provider.key.*;

import FMG.Conductor.planning.messages.PlanningMessage;
import FMG.Conductor.comm.*;
import FMG.Conductor.util.*;
import FMG.ConductorExt.security.util.*;

/**
 *  This KeyDistro class is used to distribute a secret key to a
 *  particular node by another node.
 *  <p>
 *  This class tells the destination node to distribute and the origin
 *  node of the key. The key to distribute is inside this KeyDistro class,
 *  with a label identifying the type of the secret key.
 *
 * @author Jun Li (lijun@fmg.cs.ucla.edu)
 * @version $Revision: 1.8 $
 */
public class KeyDistro extends SecurityMessage {

    /** The recycler used for messages of this type */
    private static Recycler recycler = new Recycler(KeyDistro.class);

    /** The message id assigned to messages of this type */
    private static int      messageID = -1;

    /** The destination node to distribute the key */
    NodeID          node    = new NodeID();

    /** The origin node where the key is generated */
    NodeID          fromNode = new NodeID();

    /**
     *  The key to distribute, encrypted by the public key of the
     *  destination node 
     */
    byte[]          encryptedKey;
    
    /**
     * The label identifying the type of the secret key.
     */
    String          keyLabel;

    /**
     * Register a prototype of this message class with the 
     * BasicMessageInputStream.
     * @return No return value.
     * @see FMG.Conductor.comm.BasicMessageInputStream
     */
    public static void registerSelf() {
        messageID = BasicMessageInputStream.
            registerMessagePrototype((Message) recycler.getOne());
    }
    
    /**
     * This method returns the message id that was assigned to messages of this
     * type by BasicMessageInputStream.
     * @return The message id for this type of message.
     * @see FMG.Conductor.comm.BasicMessageInputStream
     */
    public int getMessageID() {
        return messageID;
    }
    
    
    /**
     * This method returns a new (or recycled) KeyDistro object 
     * initialized from the input stream.  Normally, this method is only 
     * called on the prototype given to BasicMessageInputStream.
     * @param is The input stream to read the message's data from
     * @return a new object initialized from the data stream.
     * @exception java.io.IOException if an error occurs while reading from
     *                                the stream.
     * @see FMG.Conductor.comm.BasicMessageInputStream
     */
    public Message readNewMessage(DataInputStream is) throws IOException {
        return KeyDistro.getNewMessage(is);
    }
    

    /**
     * Generates a new KeyDistro message.
     * <p> 
     * The encrypted key particularly is generated by encrypting the key
     * using the specified cipher algorithm with the destination node's
     * public key. Other data members are also initialized appropriately.
     *
     * @param connID   The id of the connection for which we are planning.
     * @param node     The destination node
     * @param fromNode The origin node
     * @param key      The secret key to distribute
     * @param keyLabel The type of the secret key
     * @param cipherName The name of the cipher algorithm to encrypt the secrey key
     * @param epk        The public key of the destination node
     * @return A new (or recycled) and initialized KeyDistro message.
     */
    public static KeyDistro getNewMessage(ConnID connID,
                                          NodeID node, 
                                          NodeID fromNode, 
                                          byte[] key,
                                          String keyLabel,
                                          String cipherName, 
                                          PublicKey epk) {
        KeyDistro m = (KeyDistro)recycler.getOne();
        m.init(connID, node, fromNode, key, keyLabel, cipherName, epk);
        return m;
    }

   /**
    * This method returns a new (or recycled) KeyDistro object 
    * initialized from the input stream.
    * @param is The input stream to read the message's data from
    * @return A new (or recycled) and initialized KeyDistro message.
    * @exception java.io.IOException if an error occurs while reading from
    *                                the stream.
    */
     public static KeyDistro getNewMessage(DataInputStream is) throws IOException {
        KeyDistro m = (KeyDistro) recycler.getOne();
        m.receiveSelf(is);
        return m;
    }
 
    /**
     * Gets the destination node.
     *
     * @return the destination node.
     */
    public NodeID getNode() {
        return node;
    }
    
    /**
     * Gets the secret key distributed by this KeyDistro.
     * Decrypting using the specified cipher algorithm with the provided private key.
     * 
     * @param cipherName the name of the cipher algorithm name for decryption
     * @param esk        the private key for decryption
     * @return the distributed secrey key.
     */
    public byte[] getKey(String cipherName, PrivateKey esk) {
        return decrypt(encryptedKey, cipherName, esk);
    }

    /**
     * Gets the label of the secret key, that is, the key type.
     *
     * @return the label identifying the type of the secret key.
     */
    public String getKeyLabel() {
        return keyLabel;
    }

    /**
     * Write the data elements of this object to an output stream.  Presumably
     * the information will be read by <code>receiveSelf()</code>.
     * @param os The output stream to write the message's data to.
     * @return No return value.
     * @exception java.io.IOException if an error occurs while writing to the
     *                                stream.
     */
    public void sendSelf(DataOutputStream os) throws IOException {
        super.sendSelf(os);
        
        node.writeSelf(os);
        fromNode.writeSelf(os);

        os.writeInt(encryptedKey.length);
	//System.out.println("Key length is "+encryptedKey.length);
        os.write(encryptedKey, 0, encryptedKey.length);

        os.writeInt(keyLabel.length());
	//System.out.println("Key label is ("+keyLabel+") length is "+ keyLabel.length());
        os.write(keyLabel.getBytes());
    }

    /**
     * Read the data elements of this object from the input stream.  Presumably
     * this information was generated by <code>sendSelf()</code>.
     * @param is The input stream to read the message's data from.
     * @return No return value.
     * @exception java.io.IOException if an error occurs while reading from 
     *                                the stream
     */
    public void receiveSelf(DataInputStream is) throws IOException {
        super.receiveSelf(is);

        node.readSelf(is);
        fromNode.readSelf(is);
        int len = is.readInt();
        encryptedKey = new byte[len];
        is.readFully(encryptedKey, 0, len);
        len = is.readInt();
        byte[] buf = new byte[len];
        is.readFully(buf, 0, len);
        keyLabel = new String(buf);
    }


    /**
     * Determine the direction of motion for this object.
     * @return always <code>Direction.down</code> since it flows towards the client.
     */
    public Direction flowDir() {
        return Direction.down;
    }

    /**
     * Determine whether this object is understood by the Path object.
     * @return Always returns false.
     */
    public boolean handledByPath() {
        return false;
    }
    
    /**
     * Gets the string representation of this KeyDistro
     * 
     * @return the string representation of this KeyDistro
     */    
    public String toString() {
        String s = "KeyDistro[node="+node+"\n";
        s+=        "          fromNode="+fromNode+"\n";
        s+=        "          encryptedKey="+
            Debug.printableString(encryptedKey, 0, encryptedKey.length-1) +"\n";
        s+=        "          keyLabel="+keyLabel+"]\n";
        return s;
    }

    /*
     * private functions
     */

    /**
     * Initialize this keyDistro class. 
     * <p>
     * Assigns the arguments as is, except the encrypted key needs to be
     * calculated using the given cipher with the given public key.
     * 
     * @param connID   The id of the connection for which we are planning.
     * @param node     The destination node
     * @param fromNode The origin node
     * @param key      The secret key to distribute
     * @param keyLabel The type of the secret key
     * @param cipherName The name of the cipher algorithm to encrypt the secrey key
     * @param epk        The public key of the destination node
     * @return A new (or recycled) and initialized KeyDistro message.
     */
    private void init(ConnID connID, 
                      NodeID node, 
                      NodeID fromNode, 
                      byte[] key, 
                      String keyLabel, 
                      String cipherName,
                      PublicKey epk)  {
        initMessage(connID);
        
        this.node.copy(node);
        this.fromNode.copy(fromNode);
        this.keyLabel = keyLabel;
        this.encryptedKey = encrypt(key, cipherName, epk);
    }

    /**
     * Encrypts a buffer using the specified cipher with the given public key.
     * 
     * @param cleartext  The buffer where the clear text is stored
     * @param cipherName The name of the cipher algorithm
     * @param pk         The public key used to do encryption
     * @return a buffer with encrypted data.
     */
    private byte[] encrypt(byte[] cleartext, String cipherName, PublicKey pk) {
        Cipher alg = getCipherAlg(cipherName);
        try {
            alg.initEncrypt(pk);
        } catch (KeyException e) {
            System.out.println("Key exception happens when initilizating a cipher for encryption");
            System.exit(1);
        }

        return alg.crypt(cleartext, 0, cleartext.length);
    }

    /**
     * Decrypts a buffer using the specified cipher with the given private key.
     *
     * @param ciphertext The cipher text to be decrypted
     * @param cipherName The name of the cipher algorithm
     * @param sk         The private key use to do decryption
     * @return a buffer with clear text.
     */
    private byte[] decrypt(byte[] ciphertext, String cipherName, PrivateKey sk) {
        Cipher alg = getCipherAlg(cipherName);
        try {
            alg.initDecrypt(sk);
        } catch (KeyException e) {
            System.out.println("Key exception happens when initilizating a cipher for encryption");
            System.exit(1);
        }

        return alg.crypt(ciphertext, 0, ciphertext.length);
    }


    /**
     * Gets the cipher algorithm for the given name of the cipher.
     *
     * @param cipherName the name of the cipher
     * @return the cipher algorithm of the cipher
     */
    private Cipher getCipherAlg(String cipherName) {
        try {
            return Cipher.getInstance(cipherName);
        } catch (NoSuchAlgorithmException e) {
            System.err.println("EncryptedKey: Couldn't get crypt alg " + cipherName
                               + "[" + e + "]");
            System.exit(1);
            return null;    // Not reached. Needed for compile.
        }
    }

}
