/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.ConductorExt.security.messages;

import java.io.*;
import java.security.*;
import xjava.security.Cipher;

import FMG.Conductor.comm.*;
import FMG.Conductor.util.*;


/**
 * This PlanDistroSignInfo class is a message containing the signature
 * for a PlanDistro message.
 *
 * @author Jun Li (lijun@fmg.cs.ucla.edu)
 * @version $Revision: 1.4 $
 */

public class PlanDistroSignInfo extends SignInfo {

    /** The recycler used for messages of this type */
    private static Recycler recycler = new Recycler(PlanDistroSignInfo.class);

    /** The message id assigned to messages of this type */
    private static int      messageID = -1;

    /**
     * Register a prototype of this message class with the BasicMessageInputStream.
     * @return No return value.
     * @see FMG.Conductor.comm.BasicMessageInputStream
     */
     public static void registerSelf() {
        messageID =
            BasicMessageInputStream.registerMessagePrototype((Message)
                                                             recycler.getOne());
    }

    /**
     * This method returns the message id that was assigned to messages of this
     * type by BasicMessageInputStream.
     * @return The message id for this type of message.
     * @see FMG.Conductor.comm.BasicMessageInputStream
     */
    public int getMessageID() {
        return messageID;
    }


    /**
     * This method returns a new (or recycled) PlanDistroSignInfo object 
     * initialized from the input stream.  Normally, this method is only 
     * called on the prorotype given to BasicMessageInputStream.
     * @param is The input stream to read the message's data from
     * @return a new object initialized from the data stream.
     * @exception java.io.IOException if an error occurs while reading from
     *                                the stream.
     * @see FMG.Conductor.comm.BasicMessageInputStream
     */
    public Message readNewMessage(DataInputStream is) throws IOException {
        return getNewMessage(is);
    }
    
    /**
     * Generate a new (or recycled) PlanDistroSignInfo message with the appropriate
     * data members initialized.
     *
     * @param connID  The id of the connection for which we are planning.
     * @param dir     The direction that this message will flow
     * @param o       The object to be signed
     * @param sigName The name of the signing algorithm
     * @param sk      The private key used to sign
     * @return A new (or recycled) and initialized PlanDistroSignInfo message.
     */
    public static PlanDistroSignInfo getNewMessage(ConnID connID, 
						   Direction dir,
						   Object o, 
						   String sigName,
						   PrivateKey sk ) {
        PlanDistroSignInfo m = (PlanDistroSignInfo)recycler.getOne();
        m.init(connID, dir, o, sigName, sk);
        return m;
    }

   /**
    * This method returns a new (or recycled) PlanDistroSignInfo object 
    * initialized from the input stream.
    * @param is The input stream to read the message's data from
    * @return A new (or recycled) and initialized PlanDistroSignInfo message.
    * @exception java.io.IOException if an error occurs while reading from
    *                                the stream.
    */
    public static PlanDistroSignInfo getNewMessage(DataInputStream is) throws IOException {
        PlanDistroSignInfo m = (PlanDistroSignInfo) recycler.getOne();
        m.receiveSelf(is);
        return m;
    }
    
    /**
     * Gets the string representation of this PlanDistroSignInfo
     * 
     * @return the string representation of this PlanDistroSignInfo
     */    
    public String toString() {
        return "PlanDistroSignInfo["+sig.length+"|"+Debug.printableString(sig, 0, sig.length-1)+"]";
    }
}
