/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.ConductorExt.security.messages;

import java.io.*;
import java.security.*;
import xjava.security.Cipher;

import FMG.Conductor.planning.messages.PlanningMessage;
import FMG.Conductor.comm.*;
import FMG.Conductor.util.*;
import FMG.ConductorExt.security.util.Signer;

/**
 * This SignInfo class is a message containing the signature for a certain
 * object. 
 *
 * @author Jun Li (lijun@fmg.cs.ucla.edu)
 * @version $Revision: 1.6 $
 */
public abstract class SignInfo extends SecurityMessage {

    /** The signature data */
    byte[]          sig;  

    /** The direction that this SignInfo flows */
    Direction       dir;

    /**
     * Verifies whether this SignInfo contains the signature for an object.
     * 
     * @param o       The object to verify
     * @param sigName The name of the signature algorithm
     * @param pk      The public key used to verify the signature
     * @return <code>true</code> if verified as correct;
     *         <code>false</code> otherwise.
     */
    public boolean verify(Object o, String sigName, PublicKey pk) {
        return Signer.verify(o, sig, sigName, pk);
        
    }

    /**
     * Read the data elements of this object from the input stream.  Presumably
     * this information was generated by <code>sendSelf()</code>.
     * @param is The input stream to read the message's data from.
     * @return No return value.
     * @exception java.io.IOException if an error occurs while reading from 
     *                                the stream
     */
    public void receiveSelf(DataInputStream is) throws IOException {
        super.receiveSelf(is);
        dir = Direction.readDir(is);
        int len = is.readInt();
        sig = new byte[len];
        is.readFully(sig, 0, len);
    }
    
    /**
     * Write the data elements of this object to an output stream.  Presumably
     * the information will be read by <code>receiveSelf()</code>.
     * @param os The output stream to write the message's data to.
     * @return No return value.
     * @exception java.io.IOException if an error occurs while writing to the
     *                                stream.
     */
    public void sendSelf(DataOutputStream os) throws IOException {
        super.sendSelf(os);
        dir.writeSelf(os);
        os.writeInt(sig.length);
        os.write(sig, 0, sig.length);
    }

    /**
     * Determine the direction of motion for this object.
     * @return returns <code>Direction.up</code> if towards the server;
     *                 <code>Direction.down</code> if towards the client.
     */
    public Direction flowDir() {
        return dir;
    }


    /*
     * private functions
     */

    /**
     * Initializes this SignInfo class.
     * <p>
     * It assigns the connection id, the flow direction of this
     * message, and calculated the signature for a particular object using
     * specified signing algorithm and private key.
     *
     * @param connID  The id of the connection for which we are planning.
     * @param dir     The direction that this ChainAuthenticator will flow
     * @param o       The object to generate the signature
     * @param sigName The name of the signing algorithm
     * @param sk      The private key
     * @return no return value.
     */
    protected void init(ConnID connID,
			Direction dir,
			Object o, 
			String sigName,
			PrivateKey sk ) {
        initMessage(connID);
        this.dir = dir;
        sig = Signer.calcSig(o, sigName, sk);
    }
}
