/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.ConductorExt.security.tree;

import java.util.*;
import java.net.*;
import java.security.*;
import xjava.security.Cipher;

import FMG.Conductor.comm.*;
import FMG.Conductor.util.*;

import FMG.ConductorExt.security.util.*;
import FMG.ConductorExt.security.messages.*;
import FMG.ConductorExt.security.cert.*;
import FMG.ConductorExt.security.cert.Certificate;
import FMG.ConductorExt.security.tree.messages.TreeAuthenticator;

/**
 * The TreeMasterAuthenticator is the place where the public keys are
 * authenticated.
 * <p>
 * There are two situations: one is for the planner to verify the public keys
 * of other conductors; the other is for a conductor to verify the public
 * key of the planner in the reverse direction.
 * <p> 
 * Each node whose public key needs verification passes its own authenticator
 * to the TreeMasterAuthenticator object.
 * <p>
 * When a "tree" strategy is to be enforced, certificate servers are
 * arranged hierarchically. A node verifies a certificate as following:
 * <ul>
 *    <li> check whether the node knows the public key of the signing
 *         certificate server.  In particular, every node knows the public
 *         key of the certificate hierarchy root.
 *    <li> if yes, run verification algorithm;
 *    <li> otherwise, try to get the verified public key of the
 *         certificate server, which is done by recursively verifying the
 *         certificate for the certificate server public key. THIS
 *         CERTIFICATE IS CONTAINED IN THE AUTHENTICATOR ALREADY.
 * </ul>
 * <p>
 * The verified public keys for each conductor node is maintained for later usage.
 *
 * @author Jun Li (lijun@fmg.cs.ucla.edu)
 * @version $Revision: 1.6 $
 */
public class TreeMasterAuthenticator {

    /** The TreeSecurityBox object with which this TreeMasterAuthenticator is associated */
    private TreeSecurityBox  sbox;

    /** Those public keys that are verified */
    private Hashtable        verifiedPubkeyPool = new Hashtable();

    /** Create a new TreeMasterAuthenticator associated with a TreeSecurityBox */
    public TreeMasterAuthenticator(TreeSecurityBox sbox) {
        this.sbox = sbox;
    }

    /**
     * Update this TreeMasterAuthenticator based on the given authenticator.
     * <p>
     * The authenticator contains certificates used for verifying the
     * public key of the node in question. If the public key is verified,
     * the public key will be added.
     * 
     * @param auth A received TreeAuthenticator of other conductor
     * @return No return value
     */
    public void update(TreeAuthenticator auth) {

        if ( Debug.on)
            Debug.println(Debug.SECURITY, Debug.MOREINFO,
                          "[secbox] "+auth + " is being processed");            
        
        Certificate cert = getAuthenticCertificate(auth);
        
        if (cert != null) {
	    NodeID    node;
	    try {
		node = new NodeID(InetAddress.getByName(cert.getNode()));
	    } catch ( UnknownHostException e ) {
		throw new RuntimeException("Unknow node from cert "+cert);
	    }

            PublicKey pk   = KeySerializer.getPubKey(cert.getPubKey(), 
						     sbox.getPubKeyAlgo());
            addVerifiedPubkey(node, pk);
            if ( Debug.on) 
                Debug.println(Debug.SECURITY, Debug.MOREINFO,
                              "[secbox] pub key for "+node+" is authenticated and added");
        } else {
            if ( Debug.on)
                Debug.println(Debug.SECURITY, Debug.MOREINFO,
                              "Authentication not passed: "+auth);
        }
    }

    /**
     * Gets the public key of a specific conductor.
     * The verified public keys are all kept in the hashtable
     * <code>verifiedPubkeyPool</code>. 
     *
     * @param node The conductor node whose verified public key to be returned   
     * @return the public key of the conductor node
     */
    public PublicKey getVerifiedPublicKey(NodeID node) {
	return (PublicKey) verifiedPubkeyPool.get(node);
    }



    /*
     * private functions
     */

    /**
     * Gets the authentic certificate with respect to a TreeAuthenticator.
     * <p>
     * The authenticator owner must be trusted at first. Then, check if
     * the certificate with this authenticator can be verified or not. If
     * verified as authentic, this certificate is returned; otherwise, this
     * authenticator can not be verified as an authentic authentictor.
     * 
     * @param auth the TreeAuthenticator to be checked
     * @return the authentic certificate if it exists; null otherwise
     */
    private Certificate getAuthenticCertificate(TreeAuthenticator auth) {

        /* don't bother if we don't trust this host */
        if (!TreeSecurityBox.trusted(auth.getNode())) {
            if ( Debug.on)
                Debug.println(Debug.SECURITY, Debug.MOREINFO,
                              "[secbox] "+auth.getNode()+"not trusted");
            return null;
        }
        
        Certificate nodeCert = auth.getNodeCert();

        if (verifyCertificate(nodeCert, auth.getServerCerts())) {
            return nodeCert;
        } else {
            return null;
        }
    }
    
    /** 
     * Verifies a certificate.
     * <p> 
     * <p> This is a recursive function.
     *
     * @param cert the certificate to be verified
     * @param list the list of certificates collectively verifying a public key
     * @return <code>true</code> if the certificate is verified as
     *         authentic; <code>false</code> otherwise.
     */
    private boolean verifyCertificate(Certificate cert, CertificateList list) {

        if ( Debug.on)
            Debug.println(Debug.SECURITY, Debug.MOREINFO, 
                          "[secbox] Verifying "+cert);

        PublicKey   csPubKey;
        Certificate authCert;
        boolean     verified;

        authCert =  list.findCertForNode(cert.getAuthority());

        if ( authCert == null ) {
            if ( ! cert.getAuthority().equals(TreeSecurityBox.getRootCertServer())) {
                throw new RuntimeException("Certificate Hierarchy Root Not Consistent");
            }
            csPubKey = sbox.getRootCertServerPublicKey();
        } else {
            if ( verifyCertificate(authCert, list) ) {
                csPubKey = KeySerializer.getPubKey(authCert.getPubKey(), sbox.getPubKeyAlgo());
            } else {
                return false;
            }
        }

        verified = cert.verify(sbox.getSignAlgo(), csPubKey);

        return verified;
    }

    /**
     * Adds a verified public key into the pool for authentic public keys.
     *
     * @param node the node whose public key is verified and being added   
     * @param pkey the verified public key of the specific node
     * @return no return value.
     */
    private void addVerifiedPubkey(NodeID node, PublicKey pkey){
        verifiedPubkeyPool.put(node, pkey);
    }
}
