/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.ConductorExt.security.tree.messages;

import  java.io.*;
import  java.net.*;
import  java.util.*;

import  FMG.Conductor.comm.*;
import  FMG.Conductor.util.*;
import  FMG.Conductor.planning.messages.PlanningMessage;


import  FMG.ConductorExt.security.cert.*;
import  FMG.ConductorExt.security.util.*;
import  FMG.ConductorExt.security.tree.*;
import  FMG.ConductorExt.security.messages.Authenticator;

/**
 * The TreeAuthenticator class implements the authenticator used by the
 * "tree" strategy. It passes necessary authentication information to
 * facilitate the authentication of the public key of the owner of the
 * authenticator. A TreeAuthenticator mainly includes the direction if
 * flows, the name of the owner, the certificate certifying the node 
 * (the owner of the authenticator) by the associated certificate server,
 * and other certificates from the certificate hierardhy.
 *
 * @author Jun Li (lijun@fmg.cs.ucla.edu)
 * @version $Revision: 1.6 $
 */ 

public class TreeAuthenticator extends Authenticator {
    /** Boolean indicating whether to cache certificates or not */
    private static boolean   certCacheEnabled = TreeSecurityBox.certificateCacheEnabled();

    /** The recycler used for messages of this type */
    private static Recycler  recycler = new Recycler(TreeAuthenticator.class);

    /** The message id assigned to messages of this type */
    private static int       messageID = -1;

    /** The name of the owner of this TreeAuthenticator */
    private String           node;

    /** The certificate certifying this node by the local cert server */
    private Certificate      nodeCert;

    /** The list of certificates from the cert hierarchy  */
    private CertificateList  serverCerts;

    /**
     * Register a prototype of this message class with the BasicMessageInputStream.
     * @return No return value.
     * @see FMG.Conductor.comm.BasicMessageInputStream
     */
    public static void registerSelf() {
        messageID = BasicMessageInputStream.
            registerMessagePrototype((Message) recycler.getOne());
    }

    /**
     * This method returns the message id that was assigned to messages of this
     * type by BasicMessageInputStream.
     * @return The message id for this type of message.
     * @see FMG.Conductor.comm.BasicMessageInputStream
     */
    public int getMessageID() {
        return messageID;
    }
    
    /**
     * This method returns a new (or recycled) TreeAuthenticator object 
     * initialized from the input stream.  Normally, this method is only 
     * called on the prorotype given to BasicMessageInputStream.
     * @param is The input stream to read the message's data from
     * @return a new object initialized from the data stream.
     * @exception java.io.IOException if an error occurs while reading from
     *                                the stream.
     * @see FMG.Conductor.comm.BasicMessageInputStream
     */
    public Message readNewMessage(DataInputStream is) throws IOException {
        return TreeAuthenticator.getNewMessage(is);
    }

    /**
     * Generate a new (or recycled) TreeAuthenticator message with the appropriate
     * data members initialized.
     *
     * @param connID The id of the connection for which we are planning.
     * @param dir    The direction that this message will flow
     * @param pkeyAlgo The public key algorithm to use
     * @param signAlgo The signing algorithm to use
     * @param cc  The certificate client to asks for certificates
     * @return A new (or recycled) and initialized TreeAuthenticator message.
     */

    public static TreeAuthenticator getNewMessage(ConnID connID,
                                                  Direction dir,
                                                  String pkeyAlgo,
                                                  String signAlgo) {
        TreeAuthenticator m = (TreeAuthenticator) recycler.getOne();
        m.init(connID, dir, pkeyAlgo, signAlgo);
        return m;
    }


   /**
    * This method returns a new (or recycled) TreeAuthenticator object 
    * initialized from the input stream.
    * @param is The input stream to read the message's data from
    * @return A new (or recycled) and initialized TreeAuthenticator message.
    * @exception java.io.IOException if an error occurs while reading from
    *                                the stream.
    */
    public static TreeAuthenticator getNewMessage(DataInputStream is) 
        throws IOException {
       TreeAuthenticator m = (TreeAuthenticator) recycler.getOne();
       m.receiveSelf(is);
       return m;
    }
    
    /**
     * Read the data elements of this object from the input stream.  Presumably
     * this information was generated by <code>sendSelf()</code>.
     * @param is The input stream to read the message's data from.
     * @return No return value.
     * @exception java.io.IOException if an error occurs while reading from 
     *                                the stream
     */
    public void receiveSelf(DataInputStream is) throws IOException {
      super.receiveSelf(is);

      node        = new String(BytesReader.getFrom(is));
      nodeCert    = Certificate.make(is);
      serverCerts = CertificateList.make(is);
    }

    /**
     * Write the data elements of this object to an output stream.  Presumably
     * the information will be read by <code>receiveSelf()</code>.
     * @param os The output stream to write the message's data to.
     * @return No return value.
     * @exception java.io.IOException if an error occurs while writing to the
     *                                stream.
     */
    public void sendSelf(DataOutputStream os) throws IOException {
       super.sendSelf(os);

       os.writeInt(node.length());
       os.write(node.getBytes());

       nodeCert.writeSelf(os);
       serverCerts.writeSelf(os);
    }

    /**
     * Gets the node associated with this TreeAuthenticator.
     *
     * @return the node associated with this TreeAuthenticator
     */
    public String getNode() {
        return node;
    }

    /**
     * Gets the certificate for the node by its local certificate server.
     *
     * @return the certificate by the local cert server of this node.
     */
    public Certificate getNodeCert() {
	return nodeCert;
    }

    /** Gets the certificates certifying the node associated with this TreeAuthenticator
     *
     * @return the enumeration of those certificates for the associated node
     */
    public CertificateList getServerCerts() {
        return serverCerts;
    }

    /**
     * Gets the string representation of this TreeAuthenticator
     * 
     * @return the string representation of this TreeAuthenticator
     */    
    public String toString() {
        return "TreeAuthenticator("+node+" WITH "+nodeCert + "|" +serverCerts+")";
    }
    
    /**
     * Perform cleanup before allowing recycling to occur.  This method
     * is called by the Recycler. 
     * @return No return value.
     * @see FMG.Conductor.util.Recycler
     */
    public void recycling() {
        super.recycling();
	serverCerts.recycle();
    }

    /**
     * Initializes a TreeAuthenticator message.
     * <p>
     * It assigns the connection ID, the flow direction, the NodeID of the
     * node and its local certificate server. In particular, a certificate
     * for this node by the local certificate server will be retrieved, and
     * added to the list keeping the certificates for the node.
     *
     * @param connID  The id of the connection for which we are planning.
     * @param dir the direction that this TreeAuthenticator will flow
     * @param pkeyAlgo The public key algorithm to use
     * @param signAlgo The signing algorithm to use
     * @param certServer The NodeID of the certificate server
     * @param cc  The certificate client to asks for certificates
     * @return no return value.
     */
    private void init(ConnID connID, 
                      Direction dir, 
                      String pkeyAlgo,
                      String signAlgo) {

	super.init(connID, dir);

	try {
	    this.node = InetAddress.getLocalHost().getHostName();
	} catch ( UnknownHostException e ) {
	    throw new RuntimeException("Unknown host exception");
	}

	this.nodeCert    = deriveNodeCert(pkeyAlgo, signAlgo);
	this.serverCerts = deriveserverCerts(this.nodeCert.getAuthority(),
					     this.nodeCert.getAuthorityForAuthority(),
					     pkeyAlgo, 
					     signAlgo);
    }


    private Certificate deriveNodeCert( String pkeyAlgo, String signAlgo) {

	Certificate cert = obtainCertificate(this.node,
					     TreeSecurityBox.getCertServer(),
					     pkeyAlgo,
					     signAlgo);
	if ( cert == null )
	    throw new RuntimeException("No certificate returned from local authority");

	cert.retain();

	return cert;
    }

    private CertificateList deriveserverCerts(String local_cs, 
					      String parent_cs,
					      String pkeyAlgo,
					      String signAlgo) {
	
	CertificateList list = CertificateList.create();
	Certificate     cert;

	String cs     = local_cs;
	String parent = parent_cs;

	while ( parent != null ) {
	    cert = obtainCertificate(cs, parent, pkeyAlgo, signAlgo);
	    if ( cert == null )
		throw new RuntimeException("No Certificate Returned From Authority");

	    cert.retain();
	    list.addCertificate(cert);

	    cs     =  cert.getAuthority();
	    parent =  cert.getAuthorityForAuthority();
	}

	return list;
    }

    private static Certificate obtainCertificate(String principal,
						 String cs,
						 String pkeyAlgo,
						 String signAlgo) {

	// For each principal, there is only one certificate in the cache at most;
	Certificate cert = getLocalCertificate(principal, cs, pkeyAlgo, signAlgo);

	if ( cert != null )
	    return cert;
	else 
	    return getRemoteCertificate(principal, cs, pkeyAlgo, signAlgo);
    }

    private static Certificate getRemoteCertificate(String who, 
						    String authority,
						    String pkeyAlgo, 
						    String signAlgo) {
	CertClient  cc = new CertClient(authority);
	Certificate cert=null;

	CertRequest req = CertRequest.make(who, pkeyAlgo,signAlgo);
	try {
	    cert = cc.getCertificate(req);
	} catch (IOException e) {
	    throw new RuntimeException("TreeAuthenticator: Certificate I/O Error" + e);
	}
	cc.done();

	if ( cert != null )
	    addLocalCertificate(cert, pkeyAlgo, signAlgo);

	return cert;
    }
    
    synchronized private static Certificate getLocalCertificate(String principal,
								String authority,
								String pkeyAlgo,
								String signAlgo) {
	String          format;
	CertificateList list;
	Certificate     cert=null;
	
	if ( ! certCacheEnabled ) 
	    return null;

	format = new String(pkeyAlgo+signAlgo);
	list   = (CertificateList)localCerts.get(format);

	if ( list != null ) {
	    cert = list.findCert(principal, authority);
	}

	return cert;
    }

    synchronized private static void addLocalCertificate(Certificate cert,
							 String pkeyAlgo,
							 String signAlgo) {
	String          format; 
	CertificateList list; 

	if ( ! certCacheEnabled ) 
	    return;

	if ( getLocalCertificate(cert.getNode(), 
				 cert.getAuthority(), 
				 pkeyAlgo, 
				 signAlgo) != null )
	    throw new RuntimeException("Warning: certificate already locally available");
	
	format = new String(pkeyAlgo+signAlgo);

	list   = (CertificateList)localCerts.get(format);
	if ( list == null ) {
	    list = CertificateList.create();
	    localCerts.put(format, list);
	}

	list.addCertificate(cert);
    }

}
