/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.ConductorExt.security.util;

import java.security.*;
import xjava.security.Cipher;
import java.io.*;

import cryptix.util.core.*;
import cryptix.provider.elgamal.*;

import FMG.Conductor.util.*; 

/**
 * This class ElGamalKeySerializer handles reading ElGamal public or private key from a
 * stream, and writing ElGamal public or private key to a stream.
 *
 * @author Jun Li (lijun@fmg.cs.ucla.edu)
 * @version $Revision: 1.6 $
 */
public class ElGamalKeySerializer {


    /**
     * Reads an ElGamal private key from a stream.
     *
     * @param is   the stream to read the ElGamal private key from.
     * @return an ElGamal private key read from the stream.
     * @exception java.io.IOException if an error occurs while reading from
     *                                the stream.
     */
    static public PrivateKey readPrivKey(InputStream is) throws  IOException  {
        
        ReadableBytesInputStream ris = new ReadableBytesInputStream(is);
        
        // Read an ElGamal private key from stream
        BaseElGamalPrivateKey begp = 
            new BaseElGamalPrivateKey(BI.fromStream(is),
                                      BI.fromStream(is), 
                                      BI.fromStream(is), 
                                      BI.fromStream(is));
        return (PrivateKey)begp;
    }
    

    /**
     * Writes an ElGamal private key to a stream.
     *
     * @param os   the stream to write the ElGamal private key to
     * @param priv the private key to write to the stream
     * @return no return value
     * @exception java.io.IOException if an error occurs while writing to
     *                                the stream.
     */
    static public void writePrivKey(OutputStream os, PrivateKey priv) 
        throws  IOException{
        ReadableBytesOutputStream rbos = new ReadableBytesOutputStream(os);
        
        // Write an ElGamal private key to stream
        BaseElGamalPrivateKey begp = (BaseElGamalPrivateKey)priv;
        
        BI.toStream(begp.getP(),rbos);
        BI.toStream(begp.getG(),rbos);
        BI.toStream(begp.getX(),rbos);
        BI.toStream(begp.getY(),rbos);
        
    }
    
    /**
     * Reads an ElGamal public  key from a stream.
     *
     * @param is   the stream to read the ElGamal public key from.
     * @return an ElGamal public key read from the stream.
     * @exception java.io.IOException if an error occurs while reading from
     *                                the stream.
     */
    static public PublicKey readPubKey(InputStream is) 
        throws IOException {
        ReadableBytesInputStream rbis = new ReadableBytesInputStream(is);
        // Read an ElGamal public key from stream
        BaseElGamalPublicKey begpub = 
            new BaseElGamalPublicKey(BI.fromStream(rbis),
                                     BI.fromStream(rbis),
                                     BI.fromStream(rbis));
        return (PublicKey)begpub;
    }
    
    /**
     * Writes an ElGamal public key to a stream.
     *
     * @param os   the stream to write the ElGamal public key to
     * @param pub  the public key to write to the stream
     * @return no return value
     * @exception java.io.IOException if an error occurs while writing to
     *                                the stream.
     */
    static public void writePubKey(OutputStream os, PublicKey pub)  
        throws IOException{
        
        ReadableBytesOutputStream rbos = new ReadableBytesOutputStream(os);
        
        // Write an ElGamal public key to stream
        
        BaseElGamalPublicKey begpub = (BaseElGamalPublicKey)pub;
        BI.toStream(begpub.getP(), rbos);
        BI.toStream(begpub.getG(), rbos);
        BI.toStream(begpub.getY(), rbos);
        
    }


    /**
     * Gets an ElGamal public key represented by a byte array.
     *
     * @param buf  the byte array to convert to an ElGamal public key
     * @return the ElGamal public key represented by the byte array
     */
    public static PublicKey getPubKey(byte[] buf)
	throws IOException {

	ByteArrayInputStream bais = new ByteArrayInputStream(buf);

        // Read an ElGamal public key from stream
        BaseElGamalPublicKey begpub = 
            new BaseElGamalPublicKey(BI.fromStream(bais),
                                     BI.fromStream(bais),
                                     BI.fromStream(bais));
        return (PublicKey)begpub;
    }

     
    /**
     * Converts an ElGamal public key to its byte array form.
     *
     * @param pub the ElGamal public key to convert.
     *
     * @return the buffer holds the bytes converted from the public key.
     */
    public static byte[] pubKeyBytes(PublicKey pub)
	throws IOException {

        BaseElGamalPublicKey begpub = (BaseElGamalPublicKey)pub;

	ByteArrayOutputStream baos = new ByteArrayOutputStream();
	DataOutputStream dos = new DataOutputStream(baos);

	baos.reset();

        BI.toStream(begpub.getP(), dos);
        BI.toStream(begpub.getG(), dos);
        BI.toStream(begpub.getY(), dos);
        
	dos.flush();

	return baos.toByteArray();
    }


}
