/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.ConductorExt.security.util;

import java.security.*;
import xjava.security.Cipher;
import java.io.*;
import java.lang.reflect.*;

import FMG.Conductor.util.*;

/**
 * This class KeySerializer handles reading public or private key from a
 * stream, and writing public or private key to a stream, as well as the
 * conversion between a public key and its byte array form.
 * <p>
 * The actual implementation is done in specific key serializer class,
 * whose name is in the form of "XXXKeySerializer", where "XXX" is the name
 * of the key algorithm name, such as "RSAKeySerilizer" or "ElGamalKeySerialier".
 *
 * @author Jun Li (lijun@fmg.cs.ucla.edu)
 * @version $Revision: 1.7 $
 */
public class KeySerializer {

    /**
     * Reads a specific type private key from a stream.
     *
     * @param is   the stream to read the private key from.
     * @param type the type of the private key.
     * @return a private key read from the stream.
     */
    public static PrivateKey readPrivKey(InputStream is, String type) 
        throws  IOException  {
        
        String className = getKeySerializerClassName(type);
        Class[] parameters = { InputStream.class };
        Method m = getMethod(className, "readPrivKey", parameters);
        Object[] args={is};
        return (PrivateKey)invoke(m, null, args);
    }

    /**
     * Writes a specific type private key to a stream.
     *
     * @param os   the stream to write the private key to.
     * @param priv the private key to write to the stream.
     * @param type the type of the private key.
     * @return no return value.
     */
     public static void writePrivKey(OutputStream os, PrivateKey priv, String type) 
        throws  IOException{
        
        String className = getKeySerializerClassName(type);
        Class[] parameters = { OutputStream.class, PrivateKey.class };
        Method m = getMethod(className, "writePrivKey", parameters);
        Object[] args={os, priv};
        invoke(m, null, args);
    }


    /**
     * Reads a specific type public key from a stream.
     *
     * @param is   the stream to read the public key from.
     * @param type the type of the public key.
     * @return a public key read from the stream.
     */
    public static PublicKey readPubKey(InputStream is, String type) 
        throws IOException {
        
        String className = getKeySerializerClassName(type);
        Class[] parameters = { InputStream.class };
        Method m = getMethod(className, "readPubKey", parameters);
        Object[] args={is};
        return (PublicKey)invoke(m, null, args);
    }


    /**
     * Writes a specific type public key to a stream.
     *
     * @param os   the stream to write the public key to.
     * @param pub  the public key to write to the stream.
     * @param type the type of the public key.
     * @return no return value.
     */
    public static void writePubKey(OutputStream os, PublicKey pub, String type)  
        throws IOException{
        
        String className = getKeySerializerClassName(type);
        Class[] parameters = { OutputStream.class, PublicKey.class };
        Method m = getMethod(className, "writePubKey", parameters);
        Object[] args={os, pub};
        invoke(m, null, args);
    }
    
    /**
     * Gets a public key of a specific type represented by a byte array.
     *
     * @param buf  the byte array to convert to a public key
     * @param type the type of the public key to return
     * @return the public key converted from the byte array
     */
    public static PublicKey getPubKey(byte[] buf, String type){

        String className = getKeySerializerClassName(type);
        Class[] parameters = { byte[].class };
        Method m = getMethod(className, "getPubKey", parameters);
        Object[] args={buf};
        return (PublicKey)invoke(m, null, args);
    }

     
    /**
     * Converts a public key to its byte array form.
     *
     * @param pub the public key to convert.
     * @param type the typf of the public key.
     * @return the buffer holds the bytes converted from the public key.
     */
    public static byte[] pubKeyBytes(PublicKey pub, String type) {
    
        String className = getKeySerializerClassName(type);
        Class[] parameters = { PublicKey.class };
        Method m = getMethod(className, "pubKeyBytes", parameters);
        Object[] args={pub};
        return (byte[])invoke(m, null, args);
    }

    /**
     * private functions
     */

    /**
     * Gets the name of the class doing real work of specific type key
     * serialization.
     *
     * @param type the type of the public key algorithm
     * @return the name of the class implementing serialization for the
     *         given type key.
     */
    private static String getKeySerializerClassName(String type) {

        return  "FMG.ConductorExt.security.util."+type + "KeySerializer";
    }

    /**
     * Gets the method with given name and arguments type in a specific class.
     *
     * @param className   the name of the class to search the method
     * @param methodName  the name of the method
     * @param parameters  the array of the argument names of the method 
     * @return the method in the class with the same method name and arguments
     */
    private static Method getMethod(String className, 
                                    String methodName,
                                    Class[] parameters) {
        Class theClass;

        try {
            theClass = Class.forName(className);
        } catch ( ClassNotFoundException e ) {
            throw new RuntimeException("The class "+className+" is not found");
        }

        Method m = null;
        try {
            m = theClass.getMethod(methodName, parameters);
        } catch (NoSuchMethodException e) {
            throw new RuntimeException("No method "+methodName+" in class "+className);
        }

        return m;
    }

    /**
     * Invokes a particular method under a particular object with given parameters.
     *
     * @param m    the method to invoke
     * @param o    the object with which the method is to be invoked
     * @param args the arguments to be passed to the method
     * @return the object returned from running the method
     */
    private static Object invoke(Method m,
                                 Object o,
                                 Object[] args) {
        try {
            return m.invoke(o, args);
        } catch ( InvocationTargetException e ) {
            throw new RuntimeException("Method invocation error "+
                                        e.getTargetException());
        } catch ( IllegalAccessException e ) {
            throw new RuntimeException("Illegal access error "+e);
        }
    }
    
}
