/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.ConductorExt.security.util;

import java.security.*;
import xjava.security.Cipher;
import java.io.*;

/**
 * This MakeKeys class generates public and private keys for a given
 * public key algorithm. 
 * <p>
 * It contains main() routine, and can be called from the command line.
 *
 * @author Mark Yarvis (yarvis@fmg.cs.ucla.edu)
 * @author Jun Li      ( lijun@fmg.cs.ucla.edu)
 * @version $Revision: 1.8 $
 */
public class MakeKeys {
    
    /**
     * The main() routine to start running this MakeKeys class.
     *
     * @param args the arguments from the command line. It can be a list
     *             of public key algorithm names.
     */
    public static void main(String[] args) {
        (new MakeKeys()).go(args);
    }
    
    /**
     * Starts key generation one type after another.
     *
     * @param args the arguments from the command line. It can be a list
     *             of public key algorithm names.
     */
    void go(String[] args) {
        if ( args.length == 0 ) {
            System.err.println("Usage: java MakeKeys KEY_TYPE ...");
            System.exit(1);
        }
        
        for ( int i=0; i<args.length; i++ )
            makeKeys(args[i]);
    }
    
    /**
     * Makes public and private keys for a given public key algorithm.
     *
     * @param keyAlgoName the name of the public key algorithm.
     * @return no return value.
     */
    void makeKeys(String keyAlgoName) {
	//        System.out.println("====== "+ keyAlgoName +"=======");

        if ( Security.getProvider("Cryptix") == null ) {
	    //            System.out.println("Adding a new provider: cryptix.provider.Cryptix");
            Security.insertProviderAt(new cryptix.provider.Cryptix(), 1);
        }
        
        try {
            KeyPairGenerator kg = KeyPairGenerator.getInstance(keyAlgoName);
            SecureRandom rand = new SecureRandom();
            kg.initialize(512, rand);
            
            KeyPair kp = kg.generateKeyPair();
            
            //System.out.print("Public key:");
            
            KeySerializer.writePubKey(System.out, kp.getPublic(), keyAlgoName);
            
            System.out.println("");
            //System.out.print("Private key:");
            
            KeySerializer.writePrivKey(System.out, kp.getPrivate(),  keyAlgoName);
            
            System.out.println("");
        } catch (NoSuchAlgorithmException e) {
            System.err.println("MakeKeys: Couldn't get key gen alg " + 
                               keyAlgoName + "[" + e + "]");
            System.exit(1);
        } catch (IOException e) {}  // Shouldn't happen
    }
    
}
