/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.ConductorExt.security.util;

import java.util.*;
import java.security.*;
import xjava.security.Cipher;

import FMG.Conductor.planning.messages.*;
import FMG.Conductor.comm.*;

import FMG.ConductorExt.security.*;
import FMG.ConductorExt.security.messages.*;

public class MessageBuffer {
    private Hashtable            buffer = new Hashtable();
    private NodeID               currentNode;

    private ExtensionSecurityBox sbox;

    public MessageBuffer(ExtensionSecurityBox sbox) {
	this.sbox = sbox;
    }

    public void setcurrentNode(PlanningMessage m) {
	if ( m instanceof PlanningInfo )
	    currentNode = ((PlanningInfo)m).node;
	else if ( m instanceof PlanDistro )
	    currentNode = sbox.targetNode();
    }

    synchronized public void put(PlanningMessage m) {
	Vector ml;

	ml = (Vector)buffer.get(currentNode);

	if ( ml == null ) {
	    ml = new Vector();
	    buffer.put(currentNode, ml);
	}

	ml.addElement(m);

	if ( m instanceof PlanningInfoSignInfo || m instanceof PlanDistroSignInfo )
	    notify();
    }

    synchronized public Enumeration releaseMsgs(NodeID node) {
	Vector ml;

	ml = (Vector)buffer.get(node);
	while ( ! allMsgsFromTheNodeRecved(ml)) {
	    try {
		wait();
	    } catch (InterruptedException ignore) {}
	    ml = (Vector)buffer.get(node);
	}

	return ml.elements();
    }

    synchronized public PlanningMessage releaseMsg(SignInfo si) {
	Vector      ml; 
	Enumeration list;

	if ( ! (si instanceof PlanDistroSignInfo) &&
	     ! (si instanceof KeyDistroSignInfo )    )
	    throw new RuntimeException("Type Error");

	ml = (Vector)buffer.get(currentNode);
	if ( ml == null )
	    return null;

	list = ml.elements();
	while ( list.hasMoreElements() ) {
	    PlanningMessage m = (PlanningMessage)list.nextElement();

	    if ( si instanceof PlanDistroSignInfo ) {
		if (m instanceof PlanDistro) {
		    ml.removeElement(m);
		    return m; // there is only one PlanDistro
		}
	    } else if ( si instanceof KeyDistroSignInfo ) {
		if ( m instanceof KeyDistro ) {
		    if (((KeyDistro)m).getNode().equals(NodeID.getNewLocalID())) {
			ml.removeElement(m);
			return m;
		    }
		}
	    }
	}

	return null;
    }

    /**
     * Try to release the last added PlanningInfo from the buffer, if the
     * message can be verified with the given signature. "Release" means
     * this message will be removed from the buffer and returned to the
     * caller.
     */

    synchronized public PlanningMessage releaseVerifiedMsg(SignInfo si) {
	Vector ml = (Vector)buffer.get(currentNode);
	PlanningMessage m = (PlanningMessage)ml.elementAt(ml.size()-1);
        PublicKey    pk = sbox.getVerifiedPublicKey(currentNode);

	if ( pk == null )
	    return null;

	if ( si.verify(m, sbox.getSignAlgo(), pk) ) {
	    //	release this message and return it;
	    ml.removeElement(m);
	    return m;
	} else {
	    return null;
	}
    }

    synchronized public Enumeration releaseAllVerifiedMsgsForPath() {
	NodeID node;
	Vector ml;
	PublicKey pk;
	PlanningInfo pi;
	SignInfo si;
	Vector allVerified = new Vector();

	Enumeration nodes = buffer.keys();
	while ( nodes.hasMoreElements() ) {
	    node = (NodeID)nodes.nextElement();
	    ml   = (Vector)buffer.remove(node);

	    // This is running on the planner, and ml should be either
	    // nothing or containing a PlanningInfo  and a SignInfo.
	    if ( ml.size() == 0 )
		continue;

	    pk = sbox.getVerifiedPublicKey(node);
	    pi = (PlanningInfo)ml.elementAt(0);
	    si = (SignInfo)ml.elementAt(1);
	    ml.removeAllElements();
	    if ( si.verify(pi, sbox.getSignAlgo(), pk) ) {
		allVerified.addElement(pi);
	    } else {
		pi.recycle();
	    }
	    si.recycle();
	}
	return allVerified.elements();
    }



    private boolean allMsgsFromTheNodeRecved(Vector ml) {
	PlanningMessage msg;

	if ( ml == null ||  ml.size() == 0 )
	    return false;

	msg = (PlanningMessage)ml.elementAt(ml.size()-1);
	return ( msg instanceof PlanningInfoSignInfo ||
		 msg instanceof PlanDistroSignInfo );
    }

}
