/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.ConductorExt.security.util;

import java.io.*;
import java.lang.reflect.*;

import FMG.Conductor.comm.Message;

/**
 * This ObjSerializer class serializes an object to byte array.
 *
 * @author Jun Li (lijun@fmg.cs.ucla.edu)
 * @version $Revision: 1.4 $
 */
public class ObjSerializer {

    private static ByteArrayOutputStream baos = new ByteArrayOutputStream();

    /**
     * Serializes an object.
     * <p>
     * The object must implements the method "sendSelf" or "writeSelf" to
     * be serialized.
     *
     * @param o the object to serialize.
     * @return a buffer containing bytes converted from the object.
     */
    static public byte[] serialize(Object o) {

        baosSetup(o);

        return baos.toByteArray();
        // Internally copy baos.count bytes from baos.buf to a new allocated buffer

    }

    /**
     * Serializes an object into a given buffer.
     * <p>
     * The object must implements the method "sendSelf" or "writeSelf" to
     * be serialized.
     *
     * @param o the object to serialize.
     * @param outBuf the buffer where the serialization result should go
     * @param offset the offset of the buffer to start with
     * @param max    the maximum number of bytes that the buffer can accomodate
     * @return the number of bytes transfered into the buffer after serialization
     */
    static public int serialize(Object o, byte[] outBuf, int offset, int max) {
        byte[] tmp = serialize(o);

        if ( offset + tmp.length > max ) {
            System.err.println("[ObjSerializer] Insufficient memory");
            System.exit(1);
        }
        System.arraycopy(tmp, 0, outBuf, offset, tmp.length);
        return tmp.length;
    }

    /** 
     * Serializes an array of objects into a given buffer.
     * <p>
     * The object must implements the method "sendSelf" or "writeSelf" to
     * be serialized.
     *
     * @param objs   the array of objects to serialize.
     * @param outBuf the buffer where the serialization result should go
     * @param offset the offset of the buffer to start with
     * @param max    the maximum number of bytes that the buffer can accomodate
     * @return the number of bytes transfered into the buffer after serialization
     */
    static public int serialize(Object[] objs, byte[] outBuf, int offset, int max) {
        int total = 0;
        for ( int i=0; i<objs.length; i++ )
            total+= serialize(objs[i], outBuf, offset+total, max);

        return total;
    }

    /*
     * private functioins
     */

    /**
     * Gets the method used for serialization for a given object.
     * <p>
     * The object has to have the "sendSelf" or "writeSelf" implemented.
     *
     * @param o  the object for which a serialization method is to find
     * @return   the serialization method for the object.
     */
    private static Method getSerialMethod(Object o) {

        Class c = o.getClass();

        String methodname;
        if ( o instanceof Message ) {
            methodname = "sendSelf";
        } else {
            methodname = "writeSelf";
        }

        Method m = null;
        Class[] parameters = { DataOutputStream.class };
        try {
            m=c.getMethod(methodname, parameters);
        } catch (NoSuchMethodException e) {
            System.err.println("[ObjSerializer] "+e);
            System.exit(1);
        }    

        return m;
    }

    /**
     * Sets up a byte array stream for serialization of a particular
     * object.
     * <p>
     * When this method exits, the byte array stream should already
     * contain the serialized results of the object, which is done by
     * outputing the object to the stream with the serialization method of
     * the object.
     *
     * @param o the object to serialize, used for setting up the byte
     *          array stream.
     * @return no return value.
     */
    private static void baosSetup(Object o) {

        Method m = getSerialMethod(o);
        try {
            baos.reset();
            DataOutputStream dos = new DataOutputStream(baos);
            Object[] args = {dos};
            m.invoke( o, args );
            dos.flush();
            dos.close();
        } catch (IllegalAccessException e) {
            System.err.println("[ObjSerialier] "+e);
            System.exit(1);
        } catch (InvocationTargetException e) {
            System.err.println("[ObjSerialier] "+e);
            System.exit(1);
        } catch (IOException e) {
            System.err.println("[ObjSerialier] Error serializing[" + e + "]");
            System.exit(1);
        }
    }

}
