/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.ConductorExt.security.util;

import java.security.*;
import xjava.security.Cipher;
import java.io.*;

import cryptix.util.core.*;
import cryptix.provider.rsa.*;

import FMG.Conductor.util.*; 

/**
 * This class RSAKeySerializer handles reading RSA public or private key from a
 * stream, and writing RSA public or private key to a stream.
 *
 * @author Jun Li (lijun@fmg.cs.ucla.edu)
 * @version $Revision: 1.6 $
 */
public class RSAKeySerializer {

    /**
     * Reads an RSA private key from a stream.
     *
     * @param is   the stream to read the RSA private key from.
     * @return an RSA private key read from the stream.
     * @exception java.io.IOException if an error occurs while reading from
     *                                the stream.
     */
    static public PrivateKey readPrivKey(InputStream is) throws IOException  {
        
        ReadableBytesInputStream ris = new ReadableBytesInputStream(is);
        RawRSAPrivateKey rrp = new RawRSAPrivateKey(ris);
        return (PrivateKey)rrp;
    }

    /**
     * Writes an RSA private key to a stream.
     *
     * @param os   the stream to write the RSA private key to
     * @param priv the private key to write to the stream
     * @return no return value
     * @exception java.io.IOException if an error occurs while writing to
     *                                the stream.
     */
    static public void writePrivKey(OutputStream os, PrivateKey priv) 
        throws  IOException{
        
        ReadableBytesOutputStream rbos = new ReadableBytesOutputStream(os);
        
        // Write a RSA private key to stream
        RawRSAPrivateKey rrp = (RawRSAPrivateKey)priv;
        BI.toStream(rrp.getExponent(), rbos);
        BI.toStream(rrp.getP(), rbos);
        BI.toStream(rrp.getQ(), rbos);
        BI.toStream(rrp.getInverseOfQModP(), rbos);
    }

    /**
     * Reads an RSA public  key from a stream.
     *
     * @param is   the stream to read the RSA public key from.
     * @return an RSA public key read from the stream.
     * @exception java.io.IOException if an error occurs while reading from
     *                                the stream.
     */
    static public PublicKey readPubKey(InputStream is) 
        throws IOException {

        ReadableBytesInputStream rbis = new ReadableBytesInputStream(is);
        // Read a RSA public key from stream
        RawRSAPublicKey rrpub = new RawRSAPublicKey(rbis);
        return (PublicKey)rrpub;
    }

    /**
     * Writes an RSA public key to a stream.
     *
     * @param os   the stream to write the RSA public key to
     * @param pub  the public key to write to the stream
     * @return no return value
     * @exception java.io.IOException if an error occurs while writing to
     *                                the stream.
     */
    static public void writePubKey(OutputStream os, PublicKey pub)  
        throws IOException{

        ReadableBytesOutputStream rbos = new ReadableBytesOutputStream(os);
        // Write a RSA public key to stream
        RawRSAPublicKey rrpub = (RawRSAPublicKey)pub;
        BI.toStream(rrpub.getExponent(), rbos);
        BI.toStream(rrpub.getModulus(), rbos);
    }

    /**
     * Gets an RSA public key represented by a byte array.
     *
     * @param buf  the byte array to convert to an RSA public key
     * @return the RSA public key represented by the byte array
     */
    public static PublicKey getPubKey(byte[] buf)
	throws IOException {

	ByteArrayInputStream bais = new ByteArrayInputStream(buf);
	RawRSAPublicKey     rrpub = new RawRSAPublicKey(bais);

        return (PublicKey)rrpub;
    }

     
    /**
     * Converts an RSA public key to its byte array form.
     *
     * @param pub the RSA public key to convert.
     *
     * @return the buffer holds the bytes converted from the public key.
     */
    public static byte[] pubKeyBytes(PublicKey pub) 
	throws IOException {

	RawRSAPublicKey rrpub = (RawRSAPublicKey)pub;

	ByteArrayOutputStream baos = new ByteArrayOutputStream();
	DataOutputStream dos = new DataOutputStream(baos);

	baos.reset();

        BI.toStream(rrpub.getExponent(), dos);
        BI.toStream(rrpub.getModulus(),  dos);

	dos.flush();

	return baos.toByteArray();
    }


}
