/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.ConductorExt.security.util;
import java.io.*;

/**
 * This ReadableBytesInputStream class implements a stream after which two
 * readable bytes convert to one byte. For instance, "9F" will be converted
 * to "9*16+15", which is 159, assuming "9" is the high byte and "F" the low
 * byte. The high byte is read before the low byte.
 *
 * @author Mark Yarvis (yarvis@fmg.cs.ucla.edu)
 * @author Jun  Li     ( lijun@fmg.cs.ucla.edu)  (mainly adding comments)
 * @version $Revision: 1.4 $
 * @see ReadableBytesOutputStream
 */
public class ReadableBytesInputStream extends FilterInputStream {
    /** The internal buffer used by the stream */
    byte[] buf;

    /** The number of bytes in the buffer */
    int bytesInBuf;

    /** 
     * Creates a new ReadableBytesInputStream from a given input stream.
     *
     * @param in the input stream on which this stream will be operating
     */
    public ReadableBytesInputStream(InputStream in) {
        super(in);
        buf = new byte[10];
        bytesInBuf = 0;
    }
    /**
     * Read an integer from this ReadableBytesInputStream.
     * <p>
     * It works by converting every two bytes read from the origin input
     * stream into one integer.
     *
     * @return no return value.
     * @exception java.io.IOException if an error occurs while reading from
     *                                the stream.
     */
    public int read() throws IOException {
        byte hi;
        byte lo;
        
        if (bytesInBuf == 1) {
            hi = buf[0];
            bytesInBuf = 0;
        } else if (bytesInBuf == 0) {
            int tmp = in.read();
            if (tmp < 0) {
                return -1;
            }
            hi = (byte) tmp;
        } else {
            throw new RuntimeException("read: Inconsistent state.");
        }
        int tmp = in.read();
        if (tmp < 0) {
            throw new RuntimeException("unexpected EOF");
        }
        lo = (byte) tmp;
        
        return makeInt(hi,lo);
    }
    
    /**
     * Reads bytes from this ReadableBytesInputStream into a buffer.
     *
     * @param b   the buffer to store the bytes read from this
     *            ReadableBytesInputStream.
     * @param off the beginning offset in the buffer to put the data
     * @param len the number of bytes supposed to read from this
     *            ReadableBytesInputStream.
     */
    public int read(byte b[], int off, int len) throws IOException {
        insureBufSpace(len);
        
        int bytesRead = in.read(buf, bytesInBuf, len*2);
        
        if (bytesRead < 0) {
            if (bytesInBuf > 0) {
                throw new RuntimeException("unexpected EOF");
            }
            return -1;
        }
        
        bytesInBuf += bytesRead;
        
        int i;
        int bytesReallyRead=0;
        for (i=0; i+1<bytesInBuf; i=i+2) {
            b[i/2] = (byte) makeInt(buf[i],buf[i+1]);
            bytesReallyRead++;
        }
        if (i < bytesInBuf) {
            for (int j=0; j<bytesInBuf - i; i++) {
                buf[j] = buf[i+j];
            }
        }
        bytesInBuf = bytesInBuf - i;
        
        if (bytesInBuf > 1) {
            throw new RuntimeException("inconsistend state");
        }
        
        return bytesReallyRead;
    }
    
    /**
     * Skips bytes from this ReadableBytesInputStream.
     *
     * @param n the number of bytes to skip.
     * @return the number of bytes skipped.
     */
    public long skip (long n) throws IOException {
        in.skip((n*2) - bytesInBuf);
        bytesInBuf = 0;
        return n;
    }

    /**
     * Determines the number of bytes that can be read without blocking.
     *
     * @return the number of bytes that can be read without blocking.    
     * @exception java.io.IOException if an error occurs while reading from
     *                                the stream.
     */
    public int available() throws IOException {
        return (in.available() + bytesInBuf)/2;
    }
    
    /**
     * Not supported
     */
    public synchronized void mark(int readlimit) {
        throw new RuntimeException("mark not supproted");
    }
    

    /**
     * Not supported
     */
    public synchronized void reset() throws IOException {
        throw new RuntimeException("reset not supproted");
    }

    /**
     * Determines whether this ReadableBytesInputStream supports mark/reset.  
     *
     * @return always return <code>false</code>.
     */
    public boolean markSupported() {
        return false;
    }

    /*
     * private functions
     */
    
    /**
     * Converts a hex representation to an integer
     *
     * @param n a hex which is in [0-9a-fA-F].
     * @return the corresponding integer value represented by the hex
     */
    private int hexToInt(byte n) {
        if ((n>='0') && (n<='9')) {
            return (n - '0');
        } else if ((n>='A') && (n<='F')) {
            return (n - 'A' + 10);
        } else {
            return (n - 'a' + 10);
        }
    }
    
    /**
     * Makes an integer from two bytes, one is in 4-bits higher order.
     *
     * @param hi the high order byte
     * @param lo the low order byte
     * @return the integer value represented by the two bytes.
     */
    private int makeInt(byte hi, byte lo) {
        //System.out.println("("+hi+","+lo+") -> " + 
        //                   "("+(hexToInt(hi)<<4)+","+hexToInt(lo)+") -> " +
        //                     ((byte) ((hexToInt(hi) << 4) + hexToInt(lo))));
        return (hexToInt(hi) << 4) + hexToInt(lo);
    }

    
    /**
     * Insures that there are space in the internal buffer for
     * returning specific number of bytes.
     *
     * @param len the number of bytes to be able to get from this
     *            ReadableBytesInputStream.
     * @return no return value.
     */
    private void insureBufSpace(int len) {
        if (buf.length - bytesInBuf < len*2) {
            byte[] newBuf = new byte[(len*2) + bytesInBuf];
            for (int i=0; i<bytesInBuf; i++) {
                newBuf[i]=buf[i];
            }
            buf = newBuf;
        }
    }


}
