/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.ConductorExt.security.util;
import java.io.*;

/**
 * This ReadableBytesOutputStream class implements a stream after which
 * every one byte converts into two printable bytes. For instance, the byte
 * 0x9F will be converted to byte '9' and 'F'(with '9' first).
 *
 * @author Mark Yarvis (yarvis@fmg.cs.ucla.edu)
 * @author Jun  Li     ( lijun@fmg.cs.ucla.edu)  (mainly adding comments)
 * @version $Revision: 1.4 $
 * @see ReadableBytesInputStream
 */

public class ReadableBytesOutputStream extends FilterOutputStream {

    /**
     * Creates a new ReadableBytesOutputStream from a given output stream.
     *
     * @param out the output stream on which this stream works.
     */
    public ReadableBytesOutputStream(OutputStream out) {
        super(out);
    }
    
    /**
     * Writes a byte to this ReadableBytesOutputStream.
     * <p>
     * In fact two separate bytes will be written.
     *
     * @param b the byte to write
     * @return no return value.
     * @exception java.io.IOException if an error occurs while writing to
     *                                the stream.
     */
    public void write(int b) throws IOException {
        writePrintable(b);
    }
    
    /**
     * Writes <code>len</code> bytes from the byte array <code>b</code>,
     * starting at index </code>off</code>, to this
     * ReadableBytesOutputStream.
     *
     * @param b   The byte array containing bytes to be written
     * @param off The index in <code>b</code> at which to start getting bytes to
     *            be written.
     * @param len The number of bytes to write.
     * @exception java.io.IOException if an error occurs while writing to
     *                                the stream.
     */
    public void write(byte b[], int off, int len) throws IOException {
        for (int i = 0 ; i < len ; i++) {
            writePrintable(b[off + i]);
        }
    }
    
    /*
     * private functions
     */
    
    /**
     * Converts a positive integer(less than 15) to the corresponding hex byte.
     *
     * @param n the integer to convert, 0<=n<=15
     * @return the hex byte
     */
    private int intToHex(int n) {
        return (n>9 ? n+'a'-10 : n+'0');
    }
    
    /**
     * Writes an integer to output stream.
     *
     * @param b the integer to write
     * @return no return value
     * @exception java.io.IOException if error happens when attempting to
     *                                write to the output stream.
     */
    private void writePrintable(int b) throws IOException {
        out.write(intToHex((b>>4)&0xf));
        out.write(intToHex(b & 0xf));
    }
    
}
