/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package FMG.ConductorExt.security.util;

import java.security.*;
import xjava.security.Cipher;

import FMG.Conductor.comm.*;
import FMG.Conductor.util.*;
import FMG.Conductor.planning.messages.PlanningInfo;

/**
 * This Signer class provides facilities in calculating and verifying signatures.
 *
 * @author Jun Li (lijun@fmg.cs.ucla.edu)
 * @version $Revision: 1.9 $
 */
public class Signer {

    /**
     * Verifies whether a byte array is the signature of an object.
     *
     * @param o    the object to verify
     * @param sig  the signature data to check
     * @param sigName the signature algorithm name to use
     * @param pk      the public key to use for signature verification
     * @return <code>true</code> if the signature is verified as true;
     *         <code>false</code> otherwise.
     */
    public static boolean verify(Object o, byte[] sig, String sigName, PublicKey pk) {

        Signature alg = getSignature(sigName);
        byte[]    data = ObjSerializer.serialize(o);

        try {
            alg.initVerify(pk);
            alg.update(data);
            return alg.verify(sig);
        } catch (SignatureException e) {
            System.err.println("Signer: Error in signature verification" +
                               "[" + e + "]");
            System.exit(1);
            return false;   // Never executed.  Needed for compilation.
        } catch (InvalidKeyException e) {
            System.err.println("Signer: Error in signature verification" +
                               "[" + e + "]");
            System.exit(1);
            return false;   // Never executed.  Needed for compilation.
        }
        
    }

    /**
     * Calculates the signature of a given object using specific signing
     * algorithm with given private key.
     *
     * @param o   the object to calculate the signature
     * @param sigName the name of the signing algorithm to use
     * @param sk      the private key to calculate the signature
     * @return the byte array containing the signature of the object.
     */
    public static byte[] calcSig( Object o, 
                                  String sigName,
                                  PrivateKey sk ) {
        byte[]    data = ObjSerializer.serialize(o);
        Signature alg  = getSignature(sigName);

        try {
            alg.initSign(sk);
            alg.update(data, 0, data.length);
            return alg.sign();
        } catch (SignatureException e) {
            System.err.println("[Signer] Error creating signature" + 
                               "[" + e + "]");
            System.exit(1);
            return null;
        } catch (InvalidKeyException e) {
            System.err.println("[Signer] Invalid key exception" + 
                               "[" + e + "]");
            System.exit(1);
            return null;
        }
    }

    /**
     * Tests the signing and verifying procedure works correctly.
     *
     * @param args the arguments from the command line.
     */
    public static final void main(String[] args) {
        String sigName=null;
        String keyName=null;

        switch ( args.length ) {
        case 0:
            sigName = "SHA-1/RSA/PKCS#1";
            keyName = "RSA";
            break;
        case 2:
            sigName = args[0];
            keyName = args[1];
            break;
        default:
            System.out.println("Usage:");
            System.out.println("      java FMG.ConductorExt.security.util.Signer [ signature-name public-key-name ]");
        }

        self_test(sigName, keyName);
    }


    /*
     * private functions
     */

    /**
     * Gets the signature algorithm for a given signing algorithm name.
     *
     * @param sigName the name of the signature algorithm
     * @return the signature algorithm
     */
    private static Signature getSignature(String sigName) {
        if ( Security.getProvider("Cryptix") == null ) {
            Security.insertProviderAt(new cryptix.provider.Cryptix(), 1);
        }

        try {
            return Signature.getInstance(sigName);
        } catch (NoSuchAlgorithmException e) {
            System.err.println("[Signer] Couldn't get sig alg " + 
                               sigName + "[" + e + "]");
            System.exit(1);
            return null;  // for compilation reason
        } 
    }


    /**
     * Gets the signature algorithm from a particular provider.
     *
     * @param sigName the name of the signature algorithm.
     * @param provider the name of the provider
     * @return the signature algorithm
     */
    private static Signature getSignature(String sigName, String provider) {
        try {
            return Signature.getInstance(sigName, provider);
        } catch (NoSuchAlgorithmException e) {
            System.err.println("[Signer] Couldn't get sig alg " + 
                               sigName + "[" + e + "]");
            System.exit(1);
            return null;  // for compilation reason
        } catch (NoSuchProviderException e) {
            System.err.println("[Signer] no provider "+provider+" can be found");
            System.exit(1);
            return null;
        }
    }

    /**
     * Self-testing.
     * 
     * @param sigName the name of the signature algorithm name.
     * @param publicKeyAlgo the name of the public key algorithm.
     *
     */
    private static void self_test(String sigName, String publicKeyAlgo)  {
        KeyPair kp;

        try {
            KeyPairGenerator kg = KeyPairGenerator.getInstance(publicKeyAlgo);
            SecureRandom rand = new SecureRandom();
            kg.initialize(512, rand);
            kp = kg.generateKeyPair();
        } catch (NoSuchAlgorithmException e) {
            throw new RuntimeException("No "+publicKeyAlgo+" exists");
        }

        //Just making an object
        PlanningInfo pi = PlanningInfo.getNewMessage(new ConnID(),
                                                     NodeID.getNewLocalID(),
                                                     NodeID.getNewLocalID(),
                                                     NodeID.getNewLocalID(),
                                                     NodeID.getNewLocalID(),
                                                     NodeID.getNewLocalID(),
                                                     true,
                                                     true);
        byte[] sig = Signer.calcSig(pi, sigName, kp.getPrivate());
        //      sig[0]++; //making verification return wrong
        if ( verify(pi, sig, sigName, kp.getPublic()) ) {
            System.out.println("Signature verified as correct");
        } else {
            System.out.println("Signature verified as wrong");
        }
    }

}
