/*
 * Copyright (c) 2001 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Los Angeles and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include "config.h"

#include <netdb.h>
#include <netinet/in.h>
#include <pwd.h>
#include <string.h>
#define __USE_XOPEN
#include <unistd.h>
#include "FMG_Conductor_util_ExternalInfo.h"

#include "md5crypt.h"

#if defined(HAVE_SHADOW_PASSWORDS)
#include <shadow.h>
#endif

/*
 * Class:     FMG_Conductor_util_ExternalInfo
 * Method:    doServiceLookup
 * Signature: (Ljava/lang/String;)I
 */
JNIEXPORT jint JNICALL Java_FMG_Conductor_util_ExternalInfo_doServiceLookup
  (JNIEnv *env, jclass thisClass, jstring serviceName) {

   // convert String to char*
   const char *name = env->GetStringUTFChars(serviceName, 0);

   // lookup the entry
   struct servent * entry = getservbyname(name, "tcp");

   // tell VM I no longer need "name"
   env->ReleaseStringUTFChars(serviceName, name);

   // return the results
   if (entry == NULL) {
      return -1;
   } else {
      return ntohs(entry->s_port);
   }
}

/*
 * Class:     FMG_Conductor_util_ExternalInfo
 * Method:    getUserHome
 * Signature: (Ljava/lang/String;)Ljava/lang/String;
 */
JNIEXPORT jstring JNICALL Java_FMG_Conductor_util_ExternalInfo_getUserHomedir
  (JNIEnv *env, jclass thisClass, jstring username) {
   // convert String to char*
   const char *name = env->GetStringUTFChars(username, 0);

   // lookup the entry
   struct passwd * entry = getpwnam(name);

   // tell VM I no longer need "name"
   env->ReleaseStringUTFChars(username, name);

   if (entry == NULL) {
      // return null
      return 0;
   }

   return env->NewStringUTF(entry->pw_dir);
}

char * getEncryptedPassword(const char * username) {
#if defined(HAVE_SHADOW_PASSWORDS)

   // lookup the entry in the shadow file
   struct spwd * entry = getspnam(username);

   if (entry != NULL) {
      return entry->sp_pwdp;
   } else {
      return NULL;
   }

#else /* defined(HAVE_SHADOW_PASSWORDS) */

   // lookup the entry
   struct passwd * entry = getpwnam(username);

   if (entry != NULL) {
      // get the encrypted password
      return entry->pw_passwd;
   } else {
      return NULL;
   }

#endif
}

int validatePassword(const char *candidate_password, 
                     const char *encrypted_password) {
   char * encrypted_candidate;

#if defined(HAVE_MD5_PASSWORDS)
   if (is_md5_salt(encrypted_password)) {
      encrypted_candidate = md5_crypt(candidate_password, encrypted_password);
      return (strcmp(encrypted_candidate, encrypted_password)==0);
   }
#endif

   // encrypt candidate password
   char salt[3];
   salt[0] = encrypted_password[0];
   salt[1] = encrypted_password[1];
   salt[2] = '\0';
   encrypted_candidate = crypt(candidate_password, salt);

   return (strcmp(encrypted_candidate, encrypted_password)==0);
}

/*
 * Class:     FMG_Conductor_util_ExternalInfo
 * Method:    verifyPassword
 * Signature: (Ljava/lang/String;Ljava/lang/String;)Z
 */
JNIEXPORT jboolean JNICALL Java_FMG_Conductor_util_ExternalInfo_verifyPassword
  (JNIEnv *env, jclass thisClass, jstring username, jstring password) {
   // convert username String to char*
   const char *name = env->GetStringUTFChars(username, 0);

   // get the user's password from the system
   char * encrypted_password = getEncryptedPassword(name);
   if (encrypted_password == NULL) {
      return JNI_FALSE;  // no such user!!!
   }

   // tell VM I no longer need username
   env->ReleaseStringUTFChars(username, name);

   // convert password String to char*
   const char *candidate_password = env->GetStringUTFChars(password, 0);

   // test the candidate password against the previously encrypted password
   int matches = validatePassword(candidate_password, encrypted_password);

   // tell VM I no longer need password
   env->ReleaseStringUTFChars(password, candidate_password);

   // do the two encrypted passowrds match?
   if (matches) {
      return JNI_TRUE;
   } else {
      return JNI_FALSE;
   }
}

/* 
 * Class:     FMG_Conductor_util_ExternalInfo
 * Method:    getUsersID
 * Signature: (Ljava/lang/String;)I
 */
JNIEXPORT jint JNICALL Java_FMG_Conductor_util_ExternalInfo_getUsersID
  (JNIEnv *env, jclass thisClass, jstring username) {
   // convert username String to char*
   const char *name = env->GetStringUTFChars(username, 0);

   // lookup the entry
   struct passwd * entry = getpwnam(name);

   // tell VM I no longer need username
   env->ReleaseStringUTFChars(username, name);

   return (jint) (entry->pw_uid);
}

/* 
 * Class:     FMG_Conductor_util_ExternalInfo
 * Method:    getUsersDefaultGroup
 * Signature: (Ljava/lang/String;)I
 */
JNIEXPORT jint JNICALL Java_FMG_Conductor_util_ExternalInfo_getUsersDefaultGroup
  (JNIEnv * env, jclass thisClass, jstring username) {
   const char *name = env->GetStringUTFChars(username, 0);

   // lookup the entry
   struct passwd * entry = getpwnam(name);

   // tell VM I no longer need username
   env->ReleaseStringUTFChars(username, name);

   return (jint) (entry->pw_gid);
}

